c------------------------------------------------------------------------
c   program sda <c
c>  \version{version 6.00 (2010)}
c> <c
c>  Copyright (c) 2009, 2010
c>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
c>        formerly EML Research gGmbH (EML-R www.eml-research.de)
c>  Schloss-Wolfsbrunnenweg 35
c>  69118 Heidelberg, Germany
c>
c>  Copyright (c) 2000, 2003 <c
c>  European Molecular Biology Laboratory <c
c>  Meyerhofstr. 1, Postfach 10.2209 <c
c>  D-69012, Heidelberg, Germany <c
c>  <c
c>  Please send your contact address to get information on updates and
c>  new features to "mcmsoft@h-its.org". Questions will be
c>  answered as soon as possible.
c>
c>  References:
c>     R.R. Gabdoulline, R.C. Wade.
c>  Brownian dynamics simulation of protein-protein diffusional encounter.
c>  (1998) Methods, 14, 329-341.
c>  ProMetCS: an atomistic force field for modeling protein-metal surface interactions
c>  in a comntinuum aqueous solvent
c>  (submitted in January 2010) JCTC
c>
c>  for additional references see http://mcm.h-its.org/mcmsoft/sda/6.00/#ref
c>
c>  \authors{Authors: Winn P. (2007), Kokh D.B., Berynskyy M. Jan. 2010 }
c------------------------------------------------------------------------
c
cccc
c  Razif's declarations
          implicit none
          integer nsmx, ngmx
	  parameter ( nsmx=30000, ngmx=500)
	  character*2 dummy(30)
          character*69 string
	  real xs(3,nsmx), r(nsmx)
	  real a3(nsmx),a2(nsmx)
	  real x(3),xgc(3),xgo(3)
	  real g(ngmx,ngmx,ngmx)
c	  dimension g(ngmx,ngmx,ngmx)
	  character*40 pfile,efile,eefile
C         character*1 an        CCCCCCCC Commented out by peter
          real h, ais2db, fkcal
          integer ndimx, ndimy, ndimz
          real deb, factor
	  common/paramt/deb,factor
	  data ais2db/1.0807858E-04/,fkcal/332.12793/
          data xgc/0.0,0.0,0.0/


c Symbol meanings
c h - grid spacing
c ndimx number of x grid points
c ndimy number of y grid points
c ndimz number of z grid points
c pfile - input pdb (protein or inorganic surface)
c iform - binary or ascii output grid (0/1)
c ns - in first loop: counter/index indicating atom number, afterwards the total
c number of atoms (not rezeroed after incremented in the first loop)
c rave ??????????
c an - atom name (e.g. Au, N, C, etc.)
c (xs(k,ns),k=1,3) x,y,z for atom ns
c a3 -  cube of atomic radius a3(ns)=r(ns)**3
c a2 - square of atomic radius plus 0.3 a2(ns)=(r(ns)+0.3d0)**2
c          0.3 is added, as a probe radius. a2 is used to calculate the
c           exclusion grid. (Square of distance calculated by Pythagoras
c            theorem, in dipole() subroutine, and compared directly.
c x is an array with the coordinates of the origin of the grid
c    
c

C Peter's declarations
       integer BreakOne,BreakTwo,KeywordIndex,ns,nprobe,nprobeinteract
          parameter (nprobe=50, nprobeinteract=500)
       integer ProbeCounter,InteractCounter,ProbeTotal
       character*100 SEARCH
       character*1 split(100)
       character*10 ProbeName, CharSigma, CharEpsilon, CharOverlapEnergy
C   Store probes in sequential order.
C      character*2 probe(30)
       character*8 probe(nprobe)
       character*3 probetype(nprobe), ProbeNumber
C   For each probe store the atoms for which interactions are defined
C      character*8 interact(30,10)
       character*8 interact(nprobe,nprobeinteract)
C   Store the matching sigma and epsilon data for the Lennard Jones
C   function
C      real sigma(30,10)
C      real epsilon(30,10)
       real sigma2
       real sigma(nprobe,nprobeinteract)
       real sigma6(nprobe,nprobeinteract)
       real sigma12(nprobe,nprobeinteract)
       real epsilon(nprobe,nprobeinteract)
       real OverlapEnergy(nprobe)
c      real temp
C  Declare Functions
       character*10 GetProbeName
       character*3 GetProbeNumber
       character*8 GetInteractName
C      real GetSigma
C      real GetEpsilon
       character*10 GetOverlapEnergy
       character*10 GetSigma
       character*10 GetEpsilon
       character*5 an(nsmx)
       character*5 resname(nsmx)
C      character*2 AtomType(nsmx), GetAtomType
       character*10 AtomType(nsmx), GetAtomType
       character*40 ProbeFile 
       integer iform, i, j ,k, verbose
       real epssol,rion,rave, radius, axhalf, ayhalf, azhalf
       real enmax
       real Cutoff

C     interactcount is an array listing the number of 
C     interaction types read from file for each probe
       integer interactcount(nprobe)

c
c.
          print *, "reading data from the input file, as follows:"
	  read(5,*)
      read(5,*) h,ndimx,ndimy,ndimz
          print *, h,ndimx,ndimy,ndimz
c	  read(5,*)
c      read(5,*) epssol,rion
c          print *, epssol,rion
	  read(5,*)
      read(5,500) pfile
          print *, pfile
	  read(5,*)
c    Razif's latest SDA requires grid file names in the format
c    apY-XXX.grd where Y is 1,2 (p1 (surface) or p2 (protein))
c    efile read here should be 1 or 2
      read(5,500) efile
	print *, efile
	  read(5,*) iform
          print *, iform
	  read(5,*)
      read(5,500) ProbeFile
          print *, ProbeFile
	  read(5,*) verbose
          print *, verbose

          if (verbose.ne.1.and.verbose.ne.0) then
           print *, "verbose flag is incorrectly set, should be 1 or 0)"
           stop
          endif

	  read(5,*)
      read(5,*) Cutoff
          if(Cutoff <= 0) Cutoff=100.
          print *, Cutoff


c.
	  write(6,*)'---------------------------------'
      write(6,503) pfile
c    Razif's latest SDA requires grid file names in the format
c    apY-XXX.grd where Y is 1,2 (p1 (surface) or p2 (protein))
      write(6,504) "ap"//efile//"-XXX.grd where XXX is probe type"
      write(unit=6,fmt='(A40)'), "Probe data in: "// ProbeFile
c
cccc     Could Use UHBD subroutines to read qtable
cccc      but instead wrote the next section

C from program ReadProbesAndParameters
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C  Peter J Winn 03 07 07
C  To Read The Probe Data To Create A Lennard Jones Grid
C 
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

C   Initialise the counter to keep track of the probes
       ProbeCounter=0
C      open(1,file='ProbeList.txt',status='old')
       open(1,file=ProbeFile,status='old')

C Read the title lines
10     read(1,fmt='(a)',end=5000) SEARCH
C      check for comments
       if(index(SEARCH,'#').eq.1) then
          print*,SEARCH
          goto 10
       else if(index(SEARCH,'PROBE').ne.0) then
          ProbeCounter=ProbeCounter+1
          if (ProbeCounter.gt.nprobe) then
             print*, "parameter nprobe is "
             print*, "less than the number of probes"
             print*, "nprobe is: ", nprobe
             print*, "please increase this in the source code and"
             print*, "recompile before proceeding."
             stop
             endif
          InteractCounter=0
c         print*, "Before GetProbeName efile is:", efile
          probe(ProbeCounter) = GetProbeName(SEARCH)
c         print*, "After GetProbeName efile is:", efile
          probetype(ProbeCounter) = GetProbeNumber(SEARCH)
c         print*, "After GetProbeNumber efile is:", efile
C   Next line calls subroutine that assigns the variable ProbeName and
C   ProbeNumber
C         call GetProbeNameNNumber(SEARCH)
C         print *, "Return from GetProbeNameNNumber, ProbeName:",
C    +    ProbeName, " ProbeNumber", ProbeNumber
C         probe(ProbeCounter) = ProbeName
C         probetype(ProbeCounter) = ProbeNumber

c          write(iout,*), "PROBE NAME, type number: ", probe(ProbeCounter),
c     +    probetype(ProbeCounter)
c
C  Read the interacting atoms data
20        read(1,fmt='(a)',end=5000) SEARCH
          if(index(SEARCH,'INTERACT').ne.0) then
              InteractCounter=InteractCounter+1
              if (InteractCounter.gt.nprobeinteract) then
                 print*, "parameter nprobeinteract is "
                 print*, "less than the number of interactions"
                 print*, "that the probe will make"
                 print*, "nprobeinteract is: ", nprobeinteract
                 print*, "please increase this in the source code and"
                 print*, "recompile before proceeding."
                 stop
                 endif
              interact(ProbeCounter,InteractCounter)
     +                        =GetInteractName(SEARCH)
c             print*, "Interact: ", 
c    +                       interact(ProbeCounter,InteractCounter)
C             sigma(ProbeCounter,InteractCounter)=GetSigma(SEARCH)
              CharSigma=GetSigma(SEARCH)
              read(CharSigma,*)sigma(ProbeCounter,InteractCounter)
c             print *, " CharSigma ", CharSigma
              sigma2=sigma(ProbeCounter,InteractCounter)*
     +                         sigma(ProbeCounter,InteractCounter)
              sigma6(ProbeCounter,InteractCounter)=sigma2*sigma2*sigma2
              sigma12(ProbeCounter,InteractCounter)=
     +           sigma6(ProbeCounter,InteractCounter)
     +           *sigma6(ProbeCounter,InteractCounter)
c             print*, "Sigma: ", sigma(ProbeCounter,InteractCounter)
              CharEpsilon=GetEpsilon(SEARCH)
              read(CharEpsilon,*)epsilon(ProbeCounter,InteractCounter)
c             write(iout,*) "AtomType Sigma Epsilon: ",
c     +                         interact(ProbeCounter,InteractCounter), 
c     +                         sigma(ProbeCounter,InteractCounter),
c     +                         epsilon(ProbeCounter,InteractCounter)
c             write(iout,*) "After Reading interaction efile: ", efile
C           Read Next line of PROBE entry 
             goto 20
          else if(index(SEARCH,'OVERLAPENERGY').ne.0) then
             CharOverlapEnergy=GetOverlapEnergy(SEARCH)
            print *, "CharOverlapEnergy= ", CharOverlapEnergy
             read(CharOverlapEnergy,*)OverlapEnergy(ProbeCounter)
            print *, "OverlapEnergy= ", OverlapEnergy(ProbeCounter)
C           Read Next line of PROBE entry 
             goto 20
          else if(index(SEARCH,'ENDPROBE').ne.0) then
             interactcount(ProbeCounter)=InteractCounter
             print*, "Number of interaction types : ", InteractCounter
             print*, "ENDPROBE ", probe(ProbeCounter)
C            Jump back to top of first search loop
c             print*, "After Reading ENDPROBE card, efile: ", efile
             goto 10
          else 
C           Invalid entry after PROBE card
             goto 5000
          endif
       else if(index(SEARCH,'ENDFILE').ne.0) then
C        Leave the loop
           goto 30
       else
C         Not a valid line
          goto 5000
       endif
30     continue
          print*, "After ReadingProbe data efile: ", efile
       ProbeTotal=ProbeCounter


CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C  Peter J Winn 03 07 07
C  End Reading The Probe Data To Create A Lennard Jones Grid
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC




c         open(1,file=pfile)
c.. read sites
C         ns=0
C         rave=0.0
c101  continue
c         read(1,'(a)',end=97) string
c         if(string(1:4).ne.'ATOM') goto 101
c         ns=ns+1
c	  read(string,501) an,(xs(k,ns),k=1,3)
c        print*, "Before call ParsePDB efile is:", efile
        call ParsePDB(an, xs, ns,pfile, nsmx, ngmx,resname,verbose)
c        print*, "After call ParsePDB efile is:", efile
         rave=0.0
c next loop takes atomic data and calculates centre of geometry, atomic
C radius etc. Razif originally had this loop as part of the loop to read
C the atomic data from file, which would be slightly quicker, but to my
C mind produces a more convoluted code structure - Peter July18th 2007
c       write(iout,*) "AtomType, LJ_radius, (LJ_radius+0.3)^2"
       do 97 i=1,ns
C          AtomType(i)=GetAtomType(an(i),resname(i),nsmx)
           AtomType(i)=GetAtomType(an(i),resname(i))
C          AtomType(i)=an(i)//"_"//resname(i)
c          print *, an(i), resname(i), "Atom type: ", AtomType(i),
c    +     (xs(k,i),k=1,3)
C          print *, resname(i)
C          print *, (xs(k,i),k=1,3)
c calculate geometric centre - first summate all x,y,z - later find average
           xgc(1)=xgc(1)+xs(1,i)
           xgc(2)=xgc(2)+xs(2,i)
           xgc(3)=xgc(3)+xs(3,i)


	  call vans(an(i),radius)
	  r(i)=radius
	  rave=rave+r(i)
c.. factor 0.5 to count vor a volume intersection (almost exact !!)
c	  a3(ns)=r(ns)**3/1.93
c. discarded, since must be larger
	  a3(i)=r(i)**3
	  a2(i)=(r(i)+0.3d0)**2
c	  if(r(i).NE.0.0) write(iout,*) i, AtomType(i),radius,a2(i)
          if(r(1).EQ.0.0) write(6,*) " no LJ-radius parameter for",
     +                AtomType(i)
C goto 101
 97   continue


           rave=rave/float(ns)
c calculate geometric centre - find average of all atomic coordinates
           xgc(1)=xgc(1)/float(ns)
           xgc(2)=xgc(2)/float(ns)
           xgc(3)=xgc(3)/float(ns)
c
c   Electrostatics for Razif's Desolvation - Deleted
c
c  Calculate half distance along x, y, z dimensions
c                    used below to place grid centre
	  axhalf=float(ndimx-1)/2.*h
	  ayhalf=float(ndimy-1)/2.*h
	  azhalf=float(ndimz-1)/2.*h
c
	  write(6,*) 'read in x, y, z dimensions, gridspacing:  '
     +	                    ,ndimx,ndimy,ndimz,h

c calculate position for the origin of the grid
c molecular centre minus half x,y,z dimensions
	  xgo(1)=xgc(1)-axhalf-h
	  xgo(2)=xgc(2)-ayhalf-h
	  xgo(3)=xgc(3)-azhalf-h
	  write(6,*) 'computed center:',xgc
	  write(6,*) 'computed postion of grid point 0 0 0:  ',xgo
          write(6,*) '(grid starts at 1 1 1, increment is by ', h,')'
c      write(iout,*) ' Strtng w',ns,' ave radius:',rave
c
c. grdidding begins   .............
c
	  enmax=0.0
c     do 1 i=1,ndimx
c	  write(6,*) 'x-layer #',i
c       open(88,file='layer.doin')
c         write(88,*) i
c         close(88)
c         x(1)=xgo(1)+i*h
c         do 1 j=1,ndimy
c         x(2)=xgo(2)+j*h
c         do 1 k=1,ndimz
c         x(3)=xgo(3)+k*h
c  Call the function to check if grid point is inside the molecule, and if not
c   calculates the energy
C         call dipole(xs,a2,a3,ns,x,energy)
c	   write(6,*) x,energy
C         g(i,j,k)=energy
C          if(energy.gt.enmax) enmax=energy
c 1    continue
C     write(6,*)' max energy outside probe acc surface', enmax
c
C        call writg
C    ,(efile,g,h,xgo(1),xgo(2),xgo(3),ndimx,ndimy,ndimz,iform)
c
C  Cludge to check if output files exist before calculating LJ grids. Loop over probes.
C  program will crash out here if output files already exist.
C   CheckFiles tries to open a new file of the given name.
C    The later write routine will thus find a file that exists,
C    therefore writg writes to 'old' files.
       do 999  ProbeCounter=1, ProbeTotal
           eefile="ap"//efile(1:index(efile,' ')-1)//"-"//
     +        probetype(ProbeCounter)//".grd"
           call CheckFiles(eefile,iform)
 999   continue
C  Loop over probes and calculate LJ grids and write.
C  For each probe:
C        Calculate the energy on the grid
C        Write the grid
       do 1000 ProbeCounter=1, ProbeTotal
           print *, "Calculating energy for probe ", probe(ProbeCounter)
           print *, "Number of interaction types ",
     +             interactcount(ProbeCounter)
           call LJProbe(probe,interact,
C          call LJProbe(interact,
     + sigma,sigma6,sigma12,epsilon,OverlapEnergy(ProbeCounter),
     +   ProbeCounter,xs,
     +  a2,a3,ns,x,AtomType,interactcount(ProbeCounter),
     +    ndimx,ndimy,ndimz,xgo,
     +          nsmx,ngmx,nprobe,nprobeinteract,g,h,verbose,Cutoff)
C          g(i,j,k)=energy
c          print*, "efile(1:index(efile,' ')-1)", 
c    +     efile(1:index(efile,' ')-1)
C          eefile=efile(1:index(efile,' ')-1)
C    +       //probe(ProbeCounter)
C    +      (1:index(probe(ProbeCounter),' ')-1)//"_LJGrid.grd"
c          print*, "eefile probe, probetype:- ", probe(ProbeCounter), 
c    +     probetype(ProbeCounter)
c          eefile="ap"//efile
c          eefile=eefile//"-"
c          eefile=eefile//probetype(ProbeCounter)
c          eefile=eefile//".grd"
C          eefile="ap"//"-"//
           eefile="ap"//efile(1:index(efile,' ')-1)//"-"//
     +        probetype(ProbeCounter)//".grd"
           print*, "eefile --- ", eefile
C          print *, "efile: ", efile, "probe", probe(ProbeCounter)
C          print *, "eefile", eefile
           call writg
     +      (eefile,g,h,xgo(1),xgo(2),xgo(3),ndimx,ndimy,
     +           ndimz,iform,nsmx,ngmx)
 1000  continue
      stop
 500  format(a)
 501  format(13x,a1,16x,3f8.3,f15.8)
 503  format('sites to be read from:',a40)
 504  format('will write potentials:',a40)
C Peter's error reporting
5000     print*,"error reading probe file"
         print*,"last line read"
         print*,SEARCH
         stop
C
c
      end


C----------------Razif's subroutines-------------------

C-------------Dipole ---- desovlation calculation routine
      subroutine dipole(xs,a2,a3,ns,x,energy)
	  parameter ( nsmx=9000, ngmx=500)
      dimension xs(3,nsmx),a2(nsmx),a3(nsmx)
	  dimension x(3) 
	  dimension ds2(nsmx)
	  common/paramt/deb,factor
	  double precision enerd
	  logical inside
c
	  energy=0.
	  enerd=0.d0
	  inside=.false.
c
c..   tabulate first distances and not compute inside
c    checks if grid point is inside the vdw surface
c
	  do 10 is=1,ns
      ds2(is)=(xs(1,is)-x(1))**2+
     +(xs(2,is)-x(2))**2+
     +(xs(3,is)-x(3))**2
	  if(ds2(is).lt.a2(is)) return
 10   continue

c...Run over all sites  ...
c  evaluate the energy at the grid point.
	  do 20 is=1,ns
	  dist2=ds2(is)
      dist4=dist2*dist2
      debr=deb*sqrt(dist2)
	  expf=exp(-debr*2.)
	  enerd=enerd+(1.+debr)**2*expf/dist4*a3(is)
 20   continue
      energy=enerd*factor
c 	  write(6,*) enerd
c
	  return
	  end

C--------------Check Grid Files-----------------------------
      subroutine CheckFiles(fnam,iform)
          implicit none
	  character*40 fnam
          integer iform, iounit
c
	  iounit=7
c
      IF(iform.eq.0) THEN
	  open(iounit,file=fnam,form='unformatted',status='new')
      ELSE
	  open(iounit,file=fnam,form='formatted',status='new')
      END IF
   77 format(a72/2e12.6,5i7/3i7,4e12.6/4e12.6/2e12.6,2i7)
   78 format(3i7)
   79 format( (6(1x,e12.6)) )
c          
	  close (iounit)
	  return
	  end
C--------------writg ---- write grid to fileCroutine--------------
      subroutine writg(fnam,grid,h,ox,oy,oz,im,jm,km,iform,nsmx,ngmx)
C   CheckFiles() tries to open a new file of the given name.
C    This write routine will thus find a file that exists,
C    therefore writg writes to 'old' files.
	  character*40 fnam
	  character*72 title
	  real grid(ngmx,ngmx,ngmx)
c	  dimension grid(ngmx,ngmx,ngmx)
	  real*4 dum2, dum3, dum4, dum5, dum6, dum7, dum8
	  integer idum1, idum2, idum3, idum4, idum5, idum6, idum7, iform
	  integer grdflg
c
      idum=0
	  dum=0.0
	  iounit=7
      title='Output from Lennard-Jones calculations'
c. it is necessary to define scale as in UHBD if the file is 
c. going to be imported to UHBD !  Here anything is in kcal/mole
	  scale=0.239006E-02
c
      IF(iform.eq.0) THEN
c         open(iounit,file=fnam,form='unformatted',status='new')
	  open(iounit,file=fnam,form='unformatted',status='old')
      write(iounit) title, scale, dum, idum, idum,
     &                   idum, idum, idum, im, jm, km,
     &                   h, ox, oy, oz,
     &                   dum, dum, dum, dum, dum, dum,
     &                   idum, idum
      do  10  k = 1 , km
	  write(iounit) k, im, jm
	  write(iounit) ((grid(i,j,k),i=1,im),j=1,jm)
 10   continue
      ELSE
c         open(iounit,file=fnam,form='formatted',status='new')
	  open(iounit,file=fnam,form='formatted',status='old')
      write(iounit,77) title, scale, dum, idum, idum,
     &                   idum, idum, idum, im, jm, km,
     &                   h, ox, oy, oz,
     &                   dum, dum, dum, dum, dum, dum,
     &                   idum, idum
      do  11  k = 1 , km
	  write(iounit,78) k, im, jm
	  write(iounit,79) ((grid(i,j,k),i=1,im),j=1,jm)
 11   continue
      END IF
   77 format(a72/2e12.6,5i7/3i7,4e12.6/4e12.6/2e12.6,2i7)
   78 format(3i7)
   79 format( (6(1x,e12.6)) )
c          
	  close (iounit)
	  return
	  end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine vans(a,v)
      character*5 a
      character*1 split(5)
c..These are OPLS parameters used in UHBD
      V=-1.0
      read(unit=a,fmt='(5a1)')split
      write(44,500) a,split(3)
c..if statement for atom names containing 4 letters:
      IF(split(2).ne." ") THEN
      split(3) = a(2:2)
      END IF
c..if statement for atom names containing 5 letters:
      IF(split(1).ne." ") THEN
      split(3) = a(1:1)
      END IF

 500  format(a,'  ',a)
      IF(split(3).EQ.'H')V=1.20d0
      IF(split(3).EQ.'E')V=1.20d0
      IF(split(3).EQ.'C')V=1.90d0
      IF(split(3).EQ.'N')V=1.20d0  !1.625d0 for vdW
      IF(split(3).EQ.'O')V=1.48d0
      IF(split(3).EQ.'P')V=1.90d0
      IF(split(3).EQ.'S')V=1.85d0
      IF(split(3).EQ.'F')V=0.64d0  !split(3)SSUMES FE
      IF(split(3).EQ.'Z')V=1.38d0  !split(3)SSUMES ZN
      IF(split(3).EQ.'B')V=1.95d0  !split(3)SSUMES BR
      IF(split(3).EQ.'I')V=2.15d0
      IF(split(3).EQ.'W')V=2.00d0
      IF(split(3).EQ.'V')V=2.50d0
      IF(split(3).EQ.'T')V=2.80d0
      IF(split(3).EQ.'X')V=5.00d0
      IF(split(3).EQ.'Y')V=7.50d0
      IF(split(3).EQ.'Z')V=10.0d0

      if(V < 0) then
        V=1.70  !DEFAULT VDW RADIUS
        print *,"no vdw radius for atom ", A ,
     +". Default value for vdw radius assigned."
      end if


      RETURN
      END

C Peter's Functions
C-----------------------Functions--------------------------------
       character*8  function GetInteractName(SEARCH)
       implicit none
       integer BreakOne, BreakTwo,i,KeywordIndex,InteractionCounter
       character*100 SEARCH
       character*1 split(100)
       character*10 InteractName 
C Read the interaction data
       KeywordIndex=index(SEARCH,'INTERACT')
       if(KeywordIndex.gt.0) then
C         print*,'Found INTERACT'
          read(unit=SEARCH,fmt='(100a1)')split
C  Search for white space
          do 30 i=KeywordIndex+8,100
             if(split(i).eq.' ') then
                BreakOne=i
C               Break out of loop.
                goto 32 
             else
C               Goto the error call at the end of the program
                goto 5000
             endif
30        continue
32        continue
C  Search for end of white space
          do 40 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 42
             endif
40        continue
42        continue
C  Read Probe Name
          do 50 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 52
             endif
50        continue
52        continue
C  Define GetInteractName as the interactname value, which is the return value
C  of the function.
          GetInteractName=SEARCH(BreakOne:BreakTwo)
C         print*, "GetInteractName: ", GetInteractName
C         print*, split(BreakOne:BreakTwo), "ProbeName"
C         print*, BreakOne, BreakTwo, "BreakOne, BreakTwo"
C         print*,split(BreakOne)
C         print*,split(BreakTwo)
C         return to the read statement (goto 20)
       endif
C      print*, ProbeName
       return 
5000     print*,"error reading probe file"
         print*,"last line read"
         print*,SEARCH
         stop
       end

C-----------------------------------------------------------------------------------

       character*10 function GetProbeName(SEARCH)
       implicit none
       integer BreakOne, BreakTwo,i,KeywordIndex,ProbeCounter
       character*100 SEARCH
       character*100 SEARCH110
       character*1 split(100)
       character*10 ProbeName 
C Read the probe data
       KeywordIndex=index(SEARCH,'PROBE')
       if(KeywordIndex.gt.0) then
          print*,'Found PROBE'
          read(unit=SEARCH,fmt='(100a1)')split
C  Search for white space
          do 30 i=KeywordIndex+5,100
             if(split(i).eq.' ') then
                BreakOne=i
C               Break out of loop.
                goto 32 
             else
C               Goto the error call at the end of the program
                goto 5000
             endif
30        continue
32        continue
C  Search for end of white space
          do 40 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 42
             endif
40        continue
42        continue
C  Read Probe Name
          do 50 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 52
             endif
50        continue
52        continue
C  Define GetProbeName as the probename value, which is the return value
C  of the function.
          GetProbeName=SEARCH(BreakOne:BreakTwo)
C         print*, "ProbeName: ",ProbeName
C         print*, split(BreakOne:BreakTwo), "ProbeName"
C         print*, BreakOne, BreakTwo, "BreakOne, BreakTwo"
C         print*,split(BreakOne)
C         print*,split(BreakTwo)
C         return to the read statement (goto 20)
       endif
C      print*, ProbeName
       return 
5000     print*,"error reading probe file"
         print*,"last line read"
         print*,SEARCH
         stop
       end

C-----------------------------------------------------------------------------------
C-----------------------------------------------------------------------------------


       subroutine GetProbeNameNNumber(SEARCH)
       implicit none
       integer BreakOne, BreakTwo,i,KeywordIndex,ProbeCounter
       character*100 SEARCH
       character*100 SEARCH110
       character*1 split(100)
       character*3 ProbeNumber
       character*10 ProbeName 
C Read the probe data
       KeywordIndex=index(SEARCH,'PROBE')
       if(KeywordIndex.gt.0) then
          print*,'Found PROBE'
          read(unit=SEARCH,fmt='(100a1)')split
C  Search for white space
          do 30 i=KeywordIndex+5,100
             if(split(i).eq.' ') then
                BreakOne=i
C               Break out of loop.
                goto 32 
             else
C               Goto the error call at the end of the program
                goto 5000
             endif
30        continue
32        continue
C  Search for end of white space
          do 40 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 42
             endif
40        continue
42        continue
C  Read Probe Name
          do 50 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 52
             endif
50        continue
52        continue
C  Define GetProbeName as the probename value, which is the return value
C  of the function.
          ProbeName=SEARCH(BreakOne:BreakTwo)
C  Search for end of white space
          BreakOne=BreakTwo
          do 80 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 82
             endif
80        continue
82        continue
C  Search for type number 
          do 90 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 92
             endif
90        continue
92        continue
          ProbeNumber=SEARCH(BreakOne:BreakTwo)
c         print*, "ProbeName: ",ProbeName
c         print*, "Probe Number: ", ProbeNumber
C         print*, split(BreakOne:BreakTwo), "ProbeName"
C         print*, BreakOne, BreakTwo, "BreakOne, BreakTwo"
C         print*,split(BreakOne)
C         print*,split(BreakTwo)
C         return to the read statement (goto 20)
       endif
c      print*, "ProbeName: ",ProbeName
c      print*, "Probe Number: ", ProbeNumber
       return 
5000     print*,"error reading probe file"
         print*,"last line read"
         print*,SEARCH
         stop
       end

C-----------------------------------------------------------------------------------
C-----------------------------------------------------------------------------------


       character*3 function GetProbeNumber(SEARCH)
       implicit none
       integer BreakOne, BreakTwo,i,KeywordIndex,ProbeCounter
       character*100 SEARCH
       character*100 SEARCH110
       character*1 split(100)
       character*3 ProbeNumber
       character*10 ProbeName 
C Read the probe data
       KeywordIndex=index(SEARCH,'PROBE')
       if(KeywordIndex.gt.0) then
c         print*,'Found PROBE'
          read(unit=SEARCH,fmt='(100a1)')split
C  Search for white space
          do 30 i=KeywordIndex+5,100
             if(split(i).eq.' ') then
                BreakOne=i
C               Break out of loop.
                goto 32 
             else
C               Goto the error call at the end of the program
                goto 5000
             endif
30        continue
32        continue
C  Search for end of white space
          do 40 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 42
             endif
40        continue
42        continue
C  Read Probe Name
          do 50 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 52
             endif
50        continue
52        continue
C  Define GetProbeName as the probename value, which is the return value
C  of the function.
C         GetProbeName=SEARCH(BreakOne:BreakTwo)
C  Search for end of white space
          BreakOne=BreakTwo
          do 80 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 82
             endif
80        continue
82        continue
C  Search for type number 
          do 90 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 92
             endif
90        continue
92        continue
          GetProbeNumber=SEARCH(BreakOne:BreakTwo)
c         print*, "ProbeName: ",ProbeName
c         print*, "Probe Number: ", ProbeNumber
C         print*, split(BreakOne:BreakTwo), "ProbeName"
C         print*, BreakOne, BreakTwo, "BreakOne, BreakTwo"
C         print*,split(BreakOne)
C         print*,split(BreakTwo)
C         return to the read statement (goto 20)
       endif
C      print*, ProbeName
C      print*, "GetProbeNameNNumber, pre-return:"
C      print*, "ProbeName: ",ProbeName
C      print*, "Probe Number: ", ProbeNumber
       return 
5000     print*,"error reading probe file"
         print*,"last line read"
         print*,SEARCH
         stop
       end

C-----------------------------------------------------------------------------------


       character*10 function GetOverlapEnergy(SEARCH)
       implicit none
       integer BreakOne, BreakTwo,i,KeywordIndex,ProbeCounter
       character*100 SEARCH
       character*100 SEARCH110
       character*1 split(100)
       character*10 OverlapEnergy
C Read the probe data
       KeywordIndex=index(SEARCH,'OVERLAPENERGY')
       if(KeywordIndex.gt.0) then
         print*,'Found overlap energy'
          read(unit=SEARCH,fmt='(100a1)')split
C  Search for white space
          do 30 i=KeywordIndex+13,100
             if(split(i).eq.' ') then
                BreakOne=i
C               Break out of loop.
                goto 32 
             else
C               Goto the error call at the end of the  subroutine
                goto 5000
             endif
30        continue
32        continue
C  Search for end of white space
          do 40 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 42
             endif
40        continue
42        continue
C  Read Probe Name
          do 50 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 52
             endif
50        continue
52        continue
C  Define GetProbeName as the probename value, which is the return value
C  of the function.
          GetOverlapEnergy=SEARCH(BreakOne:BreakTwo)
       endif
       return 
5000     print*,"error reading probe file, OVERLAPENERGY"
         print*,"last line read"
         print*,SEARCH
         stop
       end

C--------------------------------------------------------------
C--------------------------------------------------------------
       character*10 function GetSigma(SEARCH)
C      real function GetSigma(SEARCH)
       implicit none
       integer BreakOne, BreakTwo,i,KeywordIndex,InteractionCounter
       character*100 SEARCH
       character*1 split(100)
       character*10 InteractName 
C Read the interaction data
       KeywordIndex=index(SEARCH,'INTERACT')
       if(KeywordIndex.gt.0) then
C         print*,'Found INTERACT'
          read(unit=SEARCH,fmt='(100a1)')split
C  Search for white space
          do 30 i=KeywordIndex+8,100
             if(split(i).eq.' ') then
                BreakOne=i
C               Break out of loop.
                goto 32 
             else
C               Goto the error call at the end of the program
                goto 5000
             endif
30        continue
32        continue
C  Search for end of white space
          do 40 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 42
             endif
40        continue
42        continue
C  Search for white space again - end of interacting atom name
          do 50 i=BreakOne,100
             if(split(i).eq.' ') then
                 BreakOne=i
C               Break out of the loop.
                 goto 52
             endif
50        continue
52        continue
C  Search for end of white space
          do 60 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 62
             endif
60        continue
62        continue
C  Search for sigma parameter
          do 70 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 72
             endif
70        continue
72        continue
C  Define GetInteractName as the interactname value, which is the return value
C  of the function.
C         read(SEARCH(BreakOne:BreakTwo),*)GetSigma
C         print*, SEARCH(BreakOne:BreakTwo)
          GetSigma=SEARCH(BreakOne:BreakTwo)
C         print*, "GetSigma: ", GetSigma
       endif
       return 
5000     print*,"error reading probe file"
         print*,"last line read"
         print*,SEARCH
         stop
       end

C--------------------------------------------------------------
       character*10 function GetEpsilon(SEARCH)
C      real function GetSigma(SEARCH)
       implicit none
       integer BreakOne, BreakTwo,i,KeywordIndex,InteractionCounter
       character*100 SEARCH
       character*1 split(100)
       character*10 InteractName 
C Read the interaction data
       KeywordIndex=index(SEARCH,'INTERACT')
       if(KeywordIndex.gt.0) then
C         print*,'Found INTERACT'
          read(unit=SEARCH,fmt='(100a1)')split
C  Search for white space
          do 30 i=KeywordIndex+8,100
             if(split(i).eq.' ') then
                BreakOne=i
C               Break out of loop.
                goto 32 
             else
C               Goto the error call at the end of the program
                goto 5000
             endif
30        continue
32        continue
C  Search for end of white space
          do 40 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 42
             endif
40        continue
42        continue
C  Search for white space again - end of interacting atom name
          do 50 i=BreakOne,100
             if(split(i).eq.' ') then
                 BreakOne=i
C               Break out of the loop.
                 goto 52
             endif
50        continue
52        continue
C  Search for end of white space
          do 60 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 62
             endif
60        continue
62        continue
C  Search for End of sigma parameter
          do 70 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakOne=i
C               Break out of the loop.
                 goto 72
             endif
70        continue
72        continue
C  Search for end of white space
          do 80 i=BreakOne,100
             if(split(i).ne.' ') then
                BreaKOne=i
C               Break out of the loop.
                 goto 82
             endif
80        continue
82        continue
C  Search for Epsilon parameter
          do 90 i=BreakOne,100
             if(split(i).eq.' '.or.split(i).eq.'#') then
                 BreakTwo=i
C               Break out of the loop.
                 goto 92
             endif
90        continue
92        continue

C  Define GetInteractName as the interactname value, which is the return value
C  of the function.
C         read(SEARCH(BreakOne:BreakTwo),*)GetSigma
C         print*, SEARCH(BreakOne:BreakTwo)
          GetEpsilon=SEARCH(BreakOne:BreakTwo)
C         print*, "GetEpsilon: ", GetEpsilon
       endif
       return 
5000     print*,"error reading probe file"
         print*,"last line read"
         print*,SEARCH
         stop
       end

C-----------------------Functions--------------------------------
C   THIS WOULD HAVE BEEN MORE ELEGANT BUT COULDN"T GET TO WORK!!
C       character*10 function GetEpsilon(SEARCH,CharSigma)
CC      real function GetEpsilon(SEARCH,CharSigma)
C       implicit none
C       integer BreakOne, BreakTwo,i,KeywordIndex,InteractionCounter
C       character*100 SEARCH
C       character*1 split(100)
C       character*10 InteractName 
C       character*10 CharSigma
C       real sigma
CC Read the interaction data
C       print*, "looking for Epsilon"
C       print*, CharSigma, SEARCH
C       KeywordIndex=index(SEARCH,CharSigma)
C       print*, KeywordIndex
C       if(KeywordIndex.gt.0) then
C          print*,'Found Epsilon'
C          read(unit=SEARCH,fmt='(100a1)')split
CC  Search for white space
C          do 30 i=KeywordIndex+8,100
C             if(split(i).eq.' ') then
C                BreakOne=i
CC               Break out of loop.
C                goto 32 
C             else
CC               Goto the error call at the end of the program
C                goto 5000
C             endif
C30        continue
C32        continue
CC  Search for end of white space
C          do 40 i=BreakOne,100
C             if(split(i).ne.' ') then
C                BreaKOne=i
CC               Break out of the loop.
C                 goto 42
C             endif
C40        continue
C42        continue
CC  Read Probe Name
C          do 50 i=BreakOne,100
C             if(split(i).eq.' '.or.split(i).eq.'#') then
C                 BreakTwo=i
CC               Break out of the loop.
C                 goto 52
C             endif
C50        continue
C52        continue
CC  Define GetInteractName as the interactname value, which is the return value
CC  of the function.
C          read(SEARCH(BreakOne:BreakTwo),*)GetEpsilon
C       endif
C       return 
C5000     print*,"error reading probe file"
C         print*,"last line read"
C         print*,SEARCH
C         stop
C       end


CCCCCCC--------------------------------------------------------
       character*10 function GetAtomType(an,resname)
           implicit none
           integer Start, Start2, End, End2, i
           character*5 an
           character*5 resname
C           if (index(an,"Au").gt.0.or.index(an,"AU").gt.0)
C    +      GetAtomType="Au"
C              Assume this is gold
C           if (index(an,"Au").gt.0.or.index(an,"AU").gt.0)
C    +      GetAtomType="Au"
C              Assume this is gold
C              Search for white space
C        Remove white space from atom name 
C        Find first character
           do 30 i=1,6
             Start=i
             if(an(i:i).ne.' ') then
C               Break out of loop.
                goto 32 
             endif
30        continue
32        continue
C  Search for end of white space
          do 40 i=Start,6
             End=i
             if(an(i:i).eq.' ') then
C               Break out of the loop.
                goto 42
             endif
40        continue
42        continue

C        Remove white space from atom name 
C        Find first character
           do 50 i=1,6
             Start2=i
             if(resname(i:i).ne.' ') then
C               Break out of loop.
                goto 52 
             endif
50        continue
52        continue
C  Search for end of white space
          do 60 i=Start2,6
             End2=i
             if(resname(i:i).eq.' ') then
C               Break out of the loop.
                 goto 62
             endif
60        continue
62        continue
C  Read Probe Name
C  Read Probe Name
            GetAtomType=an(Start:End-1)//"_"//resname(Start2:End2-1)
       return
       end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC----------------------------------------
       subroutine ParsePDB(an, xs, ns, pfile, 
     +             nsmx, ngmx, resname, verbose)

C parameter ( nsmx=9000,ngmx=110)
           implicit none
          integer nsmx, ngmx, k, verbose
          character*69 string
	  character*40 pfile
	  real xs(3,nsmx)
          integer ns
          character*5 an(nsmx)
          character*5 resname(nsmx)


         open(1,file=pfile)
c.. read sites
         ns=0
         print *, "In parse pdb"
5101   continue
       read(1,'(a)',end=97) string
       if (verbose.eq.1) print *, string
       if(string(1:4).ne.'ATOM') goto 5101
       ns=ns+1
       if (ns.gt.nsmx) then
          print *, "number of atoms exceeds array size: "
          print *, nsmx
          print *, "please increase the size of parameter nsmx "
          print *, " in the source code."
          stop
       endif
C      print *, "ns in ParsePDB ", ns
       an(ns)="     "
       read(string,5501) an(ns), resname(ns), (xs(k,ns),k=1,3)
C   Now read the next line by returning to the top of this loop
       goto 5101
97     print *, "Number of atoms in pdb: ", ns
       return
5501  format(11x,a5,a5,9x,3f8.3,f15.8)
C5501  format(11x,a5,a5,9x,3f8.3)
       end


CCCCCCCCCCCCCCCCCC------------------------------------------------------

      subroutine LJProbe(probe, interact,
     + sigma,sigma6,sigma12,epsilon,OverlapEnergy,
     +     ProbeCounter,xs,
     +          a2,a3,ns,x,AtomType,interactcount,
     +    ndimx,ndimy,ndimz,xgo,
     + nsmx,ngmx,nprobe,nprobeinteract,g,h,verbose,Cutoff)

           
          implicit none
          integer nsmx, ngmx, nprobe, nprobeinteract, is
          real xs(3,nsmx),a2(nsmx),a3(nsmx)
	  real ds2(nsmx),dist6(nsmx),dist12(nsmx)
	  common/paramt/deb,factor
	  double precision enerd
          real g(ngmx,ngmx,ngmx)
C         character*8 probe
          real sigma(nprobe,nprobeinteract)
          real sigma6(nprobe,nprobeinteract)
          real sigma12(nprobe,nprobeinteract)
          real epsilon(nprobe,nprobeinteract)
          real OverlapEnergy
	  real x(3),xgc(3),xgo(3)
c         character*2 AtomType(nsmx)
          character*10 AtomType(nsmx)
          character*8 interact(nprobe,nprobeinteract)
          character*8 probe(nprobe)
          real h
          real deb, factor
          integer ndimx, ndimy, ndimz, i, j, k, verbose
          integer interactcount,ProbeCounter,InteractCounter,ns
	  logical inside
	  real Cutoff
c
	  inside=.false.


       print*, "Enter LJ probe routine XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
       print*, "Probe Counter, # Interaction types, overlap energy:", 
     + ProbeCounter, interactcount, OverlapEnergy
c     print*, "Epsilon: ", epsilon(ProbeCounter,1)
C   LOOP over grid points
c--------------check parameters for all atoms
      DO  is=1,ns
             i=1
             DO  InteractCounter=1,interactcount
                        if((interact(ProbeCounter,InteractCounter)
     +                               .eq.AtomType(is))) then
c                            write(iout,*) "Atom-in pdb ",AtomType(is),
c     +                       epsilon(ProbeCounter,InteractCounter),
c     +                       sigma(ProbeCounter,InteractCounter)
                            i=i+1
                        endif
             END DO
          if(i.EQ.1) write(0,*) "no parameters for ",AtomType(is)    
      END DO
c--------------
      do 1 i=1,ndimx
	  write(6,*) 'x-layer #',i
          open(88,file='layer.doin')
          write(88,*) i
          close(88)
          x(1)=xgo(1)+i*h
          do 1 j=1,ndimy
              x(2)=xgo(2)+j*h
                do 1 k=1,ndimz
                   x(3)=xgo(3)+k*h
                   g(i,j,k)=0.0
c
C  Loop nested in first loop(over grid points)
C  Loop over atoms. Calculate distance from atom to grid point
C                      checks if less than vdw radius and more than CutOff
	           do 10 is=1,ns
                    ds2(is)=(xs(1,is)-x(1))**2+
     +              (xs(2,is)-x(2))**2+
     +              (xs(3,is)-x(3))**2
                     if(ds2(is).gt.Cutoff*Cutoff) goto 10
C                    print*, "ds2 ", ds2(is)
                     dist6(is)=ds2(is)*ds2(is)*ds2(is)
                     dist12(is)=dist6(is)*dist6(is)
C                    if(ds2(is).lt.a2(is)) return
C                    small distances will cause v. large energies. Set
C                    to default for probe
                     if(ds2(is).lt.a2(is)) then
c                     print *, "Grid point inside VDW,assign energy"
c     + ,is,AtomType(is),ds2(is),a2(is)
                      g(i,j,k)=OverlapEnergy
C         Goto end of loop, so can read next grid point
                      goto 1 
                     endif
C      Loop nested in first loop(over grid points) For each atom evaluate the energetic contribution to the grid
C      point
                    do 30 InteractCounter=1,interactcount
c                    print *, "ProbeCounter,InteractCounter ",
c    +                        ProbeCounter,InteractCounter
c                    print *, "interact(ProbeCounter,InteractCounter)",
c     +                   interact(ProbeCounter,InteractCounter)
c                    print *,"AtomType", AtomType(is)

                        if((interact(ProbeCounter,InteractCounter)
     +                               .eq.AtomType(is))) then
c                    print *, "Atom=Interact, epsilon is",
c    +               epsilon(ProbeCounter,InteractCounter)
             g(i,j,k)=g(i,j,k)+
     +                 (
     +                  4*epsilon(ProbeCounter,InteractCounter)*
     +                  ( 
     +                   ( sigma12(ProbeCounter,InteractCounter)
     +                                              /dist12(is) )
     +                   - ( sigma6(ProbeCounter,InteractCounter)
     +                                            /dist6(is) )
     +                                                        )
     +                                                         )
c             if(g(i,j,k).gt.OverlapEnergy) g(i,j,k)=OverlapEnergy
c                   print *, "Evaluated Energy", g(i,j,k)
C                   print *, "Epsilon ",
C    +              epsilon(ProbeCounter,InteractCounter), 
C    +              " sigma12 ", sigma12(ProbeCounter,InteractCounter),
C    +              " distance dist12", dist12(is)
                        endif
30                  continue
C                   print *, "ProbeCounter, InteractCounter ",
C    +                  ProbeCounter, InteractCounter
C                   print *, "Epsilon ",
C    +              epsilon(ProbeCounter,InteractCounter-1), 
C    +            " sigma12 ", sigma12(ProbeCounter,InteractCounter-1),
C    +              " distance dist12", dist12(is)
                   if (verbose.eq.1) then
                    print *, "**************************"
                    print *, " Probe ", probe(ProbeCounter)
                    print *, "Energy at ", i, j, k, " is ", g(i,j,k)
                    print *, "Co-ordinates at ", 
     +              i, j, k, " are", x(1), x(2), x(3)
                    print *, "Distance squared ", ds2(is) 
                    print *, "Epsilon ",
     +              epsilon(ProbeCounter,InteractCounter-1), 
     +            " sigma6 ", sigma6(ProbeCounter,InteractCounter-1),
     +            " distance dist6", dist6(is),
     +            " sigma12 ", sigma12(ProbeCounter,InteractCounter-1),
     +              " distance dist12", dist12(is)
                    print *, "*************************"
                  endif
10                continue
1     continue

c desolvation energy calculation
cc...Run over all sites  ...
cc  evaluate the energy at the grid point.
c     do 20 is=1,n
c        dist2=ds2(is)
c     dist4=dist2*dist2
c     debr=deb*sqrt(dist2)
c          expf=exp(-debr*2.)
c      enerd=enerd+(1.+debr)**2*expf/dist4*a3(is)
c20   continue
c     energy=enerd*factor
c 	  write(6,*) enerd
c
	  return
	  end


