/*
 * UHBDgrd.h
 *
 *  Created on: Dec 23, 2011
 *  Author: oezboyma
 *  Modified: 05.04.12
 */

#ifndef UHBDGRD_H_
#define UHBDGRD_H_

#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <math.h>
#include "ProteinDataBankStructure.h"

#define MAX 10000
using namespace std;

template<typename T1>
class UHBD
{

public:
	UHBD (char * Input);
	~UHBD ();

	void loadpdb(char *);
	//void Exclusion();
	//void Extended_Surface;
	void Generate_Skin(float, float);
	//void Skin_Fill();

// DO PRIVATE ???
	ifstream file; 
	float scale;
	//float dum2;
	int grdflag;
	//int idum2;
	int one;
	float spacing;
	const char * Title;
	//Dummy variables - start
	float dummy1;
	float dummy2;
	float dummy3;
	float dummy4;
	float dummy5;
	float dummy6;
	float dummy7;
	int idummy1;
	int idummy2;
	int idummy3;
	//Dummy variables - end
	struct dimensions
	{
		int i;
		int j;
		int k;
	}dim;
	struct origin
	{
		float x;
		float y;
		float z;
	}o;
	
	T1 *** Grid;
	//Boolean for Skin! No need for other types?
	bool *** Skin;
	//bool *** ExGrid;
	//bool *** ExtGrid;
	PDB * structure;
	
	void Read_Header();
	void Read_Data();

	template <typename type>
	type *** Allocate();

	template <typename type>
	void Deallocate(type ***);

	template <typename type>
	void Initialize(type ***);

};

template <typename T1> template <typename type>
type *** UHBD<T1>::Allocate()
{
	type *** grd;
	grd = new type ** [dim.i];
	//printf ("%d %d %d \n",dim.i,dim.j,dim.k);
	for (int x=0; x<dim.i; x++){
		grd[x] = new type * [dim.j];
		for (int y=0; y<dim.j; y++)
			grd[x][y] = new type [dim.k];
	}
	return grd;
}

template <typename T1> template <typename type>
void UHBD<T1>::Deallocate(type *** grd)
{
	for (int x=0; x<dim.i; x++){
		for (int y=0; y<dim.j; y++)
			delete[] grd[x][y];
		delete[] grd[x];
   }
}

template <typename T1> template <typename type>
void UHBD<T1>::Initialize(type *** grd)
{
	for (int x=0; x<dim.i; x++){
		for (int y=0; y<dim.j; y++){
			for (int z=0; z<dim.k; z++){
				grd[x][y][z] = 0;
			}
		}
	}
}

template <typename T1>
UHBD<T1>::UHBD(char * Input)
{
	cout << "Opening the UHBD file..." << endl;
	file.open(Input);
	Read_Header();
	Grid = Allocate <T1> ();
	Initialize <T1> (Grid);
	Read_Data();
	file.close();
}

template <typename T1>
UHBD<T1>::~UHBD()
{
	Deallocate <T1> (Grid);
	Deallocate <bool> (Skin);
	delete[] Grid;
	delete[] Skin;
	delete structure;
}

//FORMER UHBDgrd.cpp PART

// format of header is 72 character title, followed by:
// scale dum2 grdflag, idum2 km one km im jm km h ox oy oz
template <typename T1>
void UHBD<T1>::Read_Header()
{
	string title_str, line;
	string scale_str, dum1_str, grdflag_str, idum1_str, one_str, i_str, j_str, k_str;
	string spacing_str, ox_str, oy_str, oz_str;
	string dum2_str, dum3_str, dum4_str, dum5_str;
	string dum6_str, dum7_str, idum2_str, idum3_str;
	getline(file,line);
	title_str = line.substr(0,72);
	getline(file,line);
	scale_str = line.substr(1,11);
	dum1_str = line.substr(13,11);
	grdflag_str = line.substr(25,6);
	idum1_str = line.substr(32,6);
	one_str = line.substr(46,6);
	getline(file,line);
	i_str = line.substr(3,4);
	j_str = line.substr(10,4);
	k_str = line.substr(17,4);
	spacing_str = line.substr(22,11);
	ox_str = line.substr(33,12);
	oy_str = line.substr(45,12);
	oz_str = line.substr(57,12);
	getline(file,line);
	dum2_str = line.substr(1,11);
	dum3_str = line.substr(13,11);
	dum4_str = line.substr(25,11);
	dum5_str = line.substr(37,11);
	getline(file,line);
	dum6_str = line.substr(1,11);
	dum7_str = line.substr(13,11);
	idum2_str = line.substr(25,6);
	idum3_str = line.substr(32,6);

	Title = title_str.c_str();

	stringstream (scale_str) >> scale;
	stringstream (dum1_str) >> dummy1;
	stringstream (grdflag_str) >> grdflag;
	stringstream (idum1_str) >> idummy1;
	stringstream (one_str) >> one;
	stringstream (i_str) >> dim.i;
	stringstream (j_str) >> dim.j;
	stringstream (k_str) >> dim.k;
	stringstream (spacing_str) >> spacing;
	stringstream (ox_str) >> o.x;
	stringstream (oy_str) >> o.y;
	stringstream (oz_str) >> o.z;
	stringstream (dum2_str) >> dummy2;
	stringstream (dum3_str) >> dummy3;
	stringstream (dum4_str) >> dummy4;
	stringstream (dum5_str) >> dummy5;
	stringstream (dum6_str) >> dummy6;
	stringstream (dum7_str) >> dummy7;
	stringstream (idum2_str) >> idummy2;
	stringstream (idum3_str) >> idummy3;
}

template <typename T1>
void UHBD<T1>::Read_Data()
{
	cout << "Reading the UHBD grid ..." << endl;
	cout << "Grid dimensions     : " << dim.i << "  " << dim.j << "  " << dim.k << endl;
	cout << "Grid spacing        : " << spacing << endl;
	cout << "Grid origin         : " << o.x << "  " << o.y << "  " << o.z << endl;

	string line;
	//int ij_length = dim.i * dim.j;
	//T1 value;

	getline(file,line);
	for (int i = 0; i<dim.i; i++)
	{
		for (int j = 0; j<dim.j; j++)
		{
			for (int k = 0; k<dim.k; k++)
			{
				file >> Grid[i][j][k];
			}
		}
		getline(file,line);
		getline(file,line);
	}
}

template <typename T1>
void UHBD<T1>::loadpdb(char *pdbfilename)
{
	cout << "Loading the PDB file ..." << endl;
	ifstream pdbfile (pdbfilename);
	structure = new PDB;
	structure->Read_PDB_File(pdbfile,MAX);
}

template <typename T1>
void UHBD<T1>::Generate_Skin(float skin, float probes)
{
	cout << "Generating the molecule skin points ..." << endl;
	float dexcl, hdexcl, rexmax, xexh, yexh, zexh, rexcl;
	float rexcl2, x1a, x2a,x3a,zk, yj, xi, ztemp, ytemp, xtemp, Dist2;
	int ih, jh, kh, KK, JJ, II;
	int klim;

	dexcl = spacing;
	rexmax = 1.9f + probes;
	hdexcl = dexcl / 2.0f;
	klim = (int) (rexmax + dexcl*1.5f) / dexcl;
	xexh = dim.i;
	yexh = dim.j;
	zexh = dim.k;

	bool *** IEX;
	IEX = Allocate<bool>();
	Initialize<bool>(IEX);

	for (int ix = 0; ix < structure->Number_of_Residues; ix++){
		for (int jx = 0; jx < structure->Residues[ix].Number_of_Atoms; jx++){
			rexcl = structure->Residues[ix].Atoms[jx].Mass_of_Atom() + probes;
			rexcl2 = rexcl * rexcl;
			x1a = structure->Residues[ix].Atoms[jx].Coord.X  + hdexcl;
			x2a = structure->Residues[ix].Atoms[jx].Coord.Y  + hdexcl;
			x3a = structure->Residues[ix].Atoms[jx].Coord.Z  + hdexcl;

			ih = ((x1a - o.x) / dexcl) - 1;
			jh = ((x2a - o.y) / dexcl) - 1;
			kh = ((x3a - o.z) / dexcl) - 1;

			for (int i = -klim; i <= klim; i++){
				KK = i + kh;
				if (KK < 0 or KK >= dim.k){
					continue;
				}
				else{
					zk = KK * dexcl + o.z;
					ztemp = zk - structure->Residues[ix].Atoms[jx].Coord.Z;
					if (ztemp > rexcl){
						continue;
					}
					else{
						for (int j = -klim; j <= klim; j++){
							JJ = j + jh;
							if (JJ < 0 or JJ >= dim.j){
								continue;
							}
							else{
								yj = JJ * dexcl + o.y;
								ytemp = yj - structure->Residues[ix].Atoms[jx].Coord.Y;
								if (ytemp > rexcl){
									continue;
								}
								else{
									for (int k = -klim; k <= klim; k++){
										II = k + ih;
										if(II < 0 or II >= dim.i){
											continue;
										}
										else{
											if (IEX[II][JJ][KK] == 1){
												continue;
											}
											else{
												xi = II * dexcl + o.x;
												xtemp = xi - structure->Residues[ix].Atoms[jx].Coord.X;
												Dist2 = xtemp * xtemp + ytemp * ytemp + ztemp * ztemp;
												if (Dist2 > rexcl2){
													continue;
												}
												else{
													IEX[II][JJ][KK] = 1;
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}

	float probsk;
	bool *** ISK;
	ISK = Allocate<bool>();
	Initialize<bool>(ISK);

	probsk = probes + skin;
	rexmax = 1.9f + probsk;

	klim = (int) (rexmax + dexcl * 1.5f) / dexcl;

	for (int ix = 0; ix < structure->Number_of_Residues; ix++){
		for (int jx = 0; jx < structure->Residues[ix].Number_of_Atoms; jx++){
			rexcl = structure->Residues[ix].Atoms[jx].Mass_of_Atom() + probsk;
			rexcl2 = rexcl * rexcl;
			x1a = structure->Residues[ix].Atoms[jx].Coord.X  + hdexcl;
			x2a = structure->Residues[ix].Atoms[jx].Coord.Y  + hdexcl;
			x3a = structure->Residues[ix].Atoms[jx].Coord.Z  + hdexcl;
			ih = ((x1a - o.x) / dexcl) - 1;
			jh = ((x2a - o.y) / dexcl) - 1;
			kh = ((x3a - o.z) / dexcl) - 1;
			for (int i = -klim; i <= klim; i++){
				KK = i + kh;
				if (KK < 0 or KK >= dim.k){
					continue;
				}
				else{
					zk = KK * dexcl + o.z;
					ztemp = zk - structure->Residues[ix].Atoms[jx].Coord.Z;
					if (ztemp > rexcl){
						continue;
					}
					else{
						for (int j = -klim; j <= klim; j++){
							JJ = j + jh;
							if (JJ < 0 or JJ >= dim.j){
								continue;
							}
							else{
								yj = JJ * dexcl + o.y;
								ytemp = yj - structure->Residues[ix].Atoms[jx].Coord.Y;
								if (ytemp > rexcl){
									continue;
								}
								else{
									for (int k = -klim; k <= klim; k++){
										II = k + ih;
										if(II < 0 or II >= dim.i){
											continue;
										}
										else{
											if (IEX[II][JJ][KK] == 1){
												continue;
											}
											else{
												if (ISK[II][JJ][KK] == 1){
													continue;
												}
												else{
													xi = II * dexcl + o.x;
													xtemp = xi - structure->Residues[ix].Atoms[jx].Coord.X;
													Dist2 = xtemp * xtemp + ytemp * ytemp + ztemp * ztemp;
													if (Dist2 > rexcl2){
														continue;
													}
													else{
														ISK[II][JJ][KK] = 1;
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}

	//SKIN FILL PART
	Skin = Allocate <bool> ();
	Initialize(Skin);
	for (int i = 0; i < dim.i; i ++){
		for (int j = 0; j < dim.j; j++){
			for (int k = 0; k < dim.k; k++){
				if (IEX[i][j][k]== true){
					Skin[i][j][k]=false;
				}
				else
					if (ISK[i][j][k]== true)
						Skin[i][j][k]=true;
					else
						Skin[i][j][k]=false;
			}
		}
	}
	Deallocate<bool>(IEX);
	Deallocate<bool>(ISK);
	delete[] IEX;
	delete[] ISK;
}

#endif /* UHBDGRD_H_ */
