!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file MainLoop_sda_2proteins.f90
!!
!! Version of test, without OpenMP parralellisation
!!
!! Should use \ref mainloop_sda_2proteins_omp::mainloop_sda_2proteins_omp "MainLoop_sda_2proteins_omp"\n
!!
!!
!! Make Brownian dynamics trajectory with 2 solutes\n
!! Solute 1 is always kept fixed at the center

!> Main loop function, test without OpenMP
!!
!! Second version of MainLoop, movable atoms are copied to proteins if needed\n
!! rhit is computed here, no need in other module, but need probep\n
!! @param tab_protein : contains all proteins, instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param rcriteria : instance of \ref mod_reactioncriteria::react_criter "react_criter"
!! @param o_complexes, trajectories : for writing complexes and trajectories
!! @param geom : geometry instance
!! @param p_restart : pointer on a restart record
!! @param resid_time : instance of \ref mod_residencetime::residence_time "residence_time"
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param param_timestep : instance of \ref com_sda::parameter_timestep "parameter_timestep"
!! @param param_probe : instance of \ref com_sda::probe_type "probe_type"
!! @param param_force_energy : instance of  \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param param_analytic : structure \ref com_sda::parameter_analytic "parameter_analytic"
!! @param param_metaldesolv : instance of \ref com_sda::parameter_metaldesolv "parameter_metaldesolv"
!! @param nmax_overlap : maximum overlap before boost
!! @param rboost : distance of boost
!! @param timemax : maximum time for a trajectory, 0 means infinite
!! @param nrun : number of runs
subroutine MainLoop_sda_2proteins ( tab_protein, rcriteria, &
                                    o_complexes, trajectories, geom, p_restart, resid_time, &
                                    type_calc, param_timestep, param_probe, param_force_energy, &
                                    param_analytic, param_metaldesolv, nmax_overlap, &
                                    rboost, timemax, nrun )

     USE mod_array_protein ! include mod_protein, which includes mod_setofgrid, which includes mod_grid...
     USE mod_gridtype
     USE mod_ReactionCriteria
     USE mod_geometry
     USE mod_record
     USE mod_ResidenceTime
     USE AllTimer
     USE com_sda
     USE mod_force_energy
     USE mod_compute_force
     USE mod_compute_energy
     USE mod_compute_energy_flex
     USE mod_compute_image_charge
     USE mod_compute_metaldesolv
     USE mod_compute_analytic_2proteins
     USE mod_set_parameters
     USE mod_init_position
     USE mod_hydrodynamics
     IMPLICIT NONE

     interface

            subroutine convert_to_record_format ( array_energy, array_energy_1d, max_bit_energy )
                IMPLICIT NONE
                real(kind=8), dimension(:,:), intent ( in ) :: array_energy
                real(kind=8), dimension(:,:), intent ( out ) :: array_energy_1d
                integer, intent ( in ) :: max_bit_energy
            end subroutine convert_to_record_format

            subroutine make_bd_move_2prot_timer_excl ( prot1, prot2, geom, force, torque, torque1, &
                                               dtnow, dist, rhit, hit, &
                                               accept, escape, total_overlap, timer_exchk )
                USE mod_protein
                USE mod_geometry
                USE AllTimer
                IMPLICIT NONE
                type ( protein ) :: prot1, prot2
                type ( geometry ), intent ( in ) :: geom
                real ( kind=8 ), dimension ( 3 ), intent ( in ) :: force, torque, torque1
                real ( kind=8 ), intent ( in ) :: dtnow
                real ( kind=4 ) :: dist, rhit
                integer :: total_overlap
                ! accept true if at least one motion has been accepted
                logical, intent ( out ) :: hit, escape, accept
                type ( mytimer ) :: timer_exchk
            end subroutine make_bd_move_2prot_timer_excl

            subroutine change_conformation ( prot, tab_prot, geom, param_force_energy, param_metal, type_calc, dummy_array, &
                                             hit, accepted, nb_overlap_flex, ptrf_energy )
              USE mod_array_protein
              USE mod_geometry
              USE mod_type_calc
              USE mod_force_energy
              USE mod_compute_energy_flex
              USE com_sda
              IMPLICIT NONE

              interface
                subroutine ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                                         dummy_array, energy, energy_imgchg )
                  USE mod_force_energy
                  USE mod_array_protein
                  USE mod_geometry
                  USE com_sda
                  implicit none
                  type ( protein ), intent(in), target :: prot
                  type ( array_protein_type ), intent( in ), target :: tab_prot
                  type ( geometry ), intent( in ) :: geom
                  type ( type_force_energy ), intent ( in ) :: param_force_energy
                  type ( parameter_metaldesolv ), intent ( in ) :: param_metal
                  real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
                  real ( kind=8 ) :: energy, energy_imgchg
                  integer, intent ( in ) :: nb_prot
                end subroutine ptrf_energy
              end interface

              type ( protein ) :: prot
              type ( array_protein_type ), intent( in ), target :: tab_prot
              type ( geometry ), intent( in ) :: geom
              type ( type_force_energy ), intent ( in ) :: param_force_energy
              type ( parameter_metaldesolv ), intent ( in ) :: param_metal
              type ( type_calculation ), intent ( in ) :: type_calc
              real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
              integer :: nb_overlap_flex
              logical, intent ( in ) :: hit
              logical, intent ( out ) :: accepted

           end subroutine change_conformation

            subroutine check_reaction_pair ( prot1, prot2, rcriteria, visited_windows, escape, sum_time, &
                                             first_time, time, dtnow, close )
                USE mod_protein
                USE mod_ReactionCriteria ! include mod_rate_calculation
                IMPLICIT NONE
                type ( protein ), intent ( in ) :: prot1, prot2
                type ( react_criter ) :: rcriteria
                real ( kind=4 ), dimension(:,:) :: sum_time, first_time
                real ( kind=8 ), dimension (:,:) :: escape
                integer, dimension (:) :: visited_windows
                real ( kind=4 ) :: time, dtnow
                logical, intent( out ) :: close
            end subroutine check_reaction_pair

            subroutine init_time ( param_timestep, geom, dist, rhit, dm30, &
                       dtnow, hit )

                USE com_sda
                USE mod_geometry
                IMPLICIT NONE
                type ( parameter_timestep ), intent ( in ) :: param_timestep
                type ( geometry ), intent ( in ) :: geom
                real ( kind=8 ), intent ( in ) :: dm30
                real ( kind=8 ), intent ( out ) :: dtnow
                real ( kind=4 ), intent( in ) :: rhit
                real ( kind=4 ), intent( in ) :: dist
                logical, intent ( out ) :: hit
            end subroutine init_time

     end interface

     type ( array_protein_type ), target :: tab_protein
     type ( react_criter ) :: rcriteria
     type ( record ) :: o_complexes, trajectories
     type ( record ), pointer :: p_restart
     type ( geometry ) :: geom
     type ( residence_time ) :: resid_time
     type ( type_calculation ) :: type_calc
     type ( parameter_timestep ) :: param_timestep
     type ( probe_type ) :: param_probe
     type ( type_force_energy ) :: param_force_energy
     type ( parameter_analytic ) :: param_analytic
     type ( parameter_metaldesolv ) :: param_metaldesolv
     real ( kind=8 ) :: timemax
     real ( kind=4 ) :: rboost
     integer :: nrun, nmax_overlap

     ! local timer to this subroutine
     type ( mytimer ) :: timer_force, timer_energy, timer_reactpair
     type ( mytimer ) :: timer_exchk, timer_bd, timer_complexes, timer_flex
     type ( mytimer ) :: timer_update_rate

     ! local variables, used for simplicity
     type ( protein ), pointer :: protein1, protein2

     ! to fit for record
     !real ( kind=8 ), dimension(:,:), allocatable:: array_energy
     integer :: max_energy_terms
     real ( kind=8 ), dimension(:,:), pointer:: array_energy
     real ( kind=8 ) :: total_energy
     ! new, fixe format for entry to add_record. Expect 2 dimen in add_record, but only 1 if filled
     ! allocated as ( 20, 1 )
     integer :: size_array_energy
     real ( kind=8 ), dimension(:,:), pointer :: array_energy_1d
     logical :: correct_format_energy

     ! variable for force and torques
     real ( kind=8 ), dimension(3) :: f, t, t1

     !! seems ok with gfortran, maybe problem with intel, may need a pointer ?? certianly problem with v2.5 OpenMP
     ! now used only for electron transfert
     real (kind=8), dimension(:,:), pointer :: local_escape
     ! equivalent of local_escape with new implementation
     integer, dimension(:), pointer :: local_visited_windows
     real ( kind=4 ), dimension(:,:), pointer :: local_sumtime, local_fpt

     !! record of the number of step, allocated here
     !! size known only when nrun is read
     integer, dimension(:), allocatable :: steps_record

     real ( kind=8 ) :: dm, dm30
     real ( kind=8 ) :: time, time_all, time_per_run, timehit, dtnow
     real ( kind=4 ) :: Cut2, dist, rhit

     integer :: current_run, run_done, nstep_all, current_step, current_conf
     integer :: total_overlap, nboost_all, total_overlap_flex
     integer :: k, nb_overlap, ierr !iw

     !! for sda-koff, need internal p_current_node in mod_record if want to avoid explicit declaration here
     !! to see with all tools as well
     integer :: current_occur, n_repeat_occur, total_occur
     type ( one_complexe ), pointer :: p_tmp_onecomplexe

     logical :: overlap, bool_escape, print_step
     logical :: clos1, hit, accept
     !! add isurf, maybe not necessary, but could be used by compute distance, general sda, sdamm
     logical :: iflex, accepted_new_conf

     !! for printing steps.log
     character*32 :: steps_str

    !!used for lennardjones, contains only 1.0, comes from sdamm, needed for common flex energy evaluation
    ! and new version of force
    ! with version fast, no need of dummy array
    real ( kind=8 ), dimension( : ), pointer :: dummy_array
    integer :: status, io_err

    !! could be in mod_record ?? problen
    nullify ( p_tmp_onecomplexe )
    nullify ( local_visited_windows, local_escape, local_sumtime, local_fpt )
    nullify ( dummy_array, array_energy_1d )

    steps_str = 'steps.log'
    if ( type_calc % bool_par ) then
        !!! keep the suffix, add to filename the integer as string
        call create_filename_flex ( steps_str, type_calc % nb_par )
        write(*,*) 'new log ', steps_str
    end if

    size_array_energy = 0
    max_energy_terms = 0
    iflex = .false.
    accepted_new_conf = .false.
    correct_format_energy = .false.
    current_occur = 0
    n_repeat_occur = 0
    ! count done really finished, for correct normalization in association rate
    run_done = 0

    !! initialise timer
    call init_mytimer ( timer_force )
    call init_mytimer ( timer_bd )
    call init_mytimer ( timer_exchk )
    call init_mytimer ( timer_reactpair )
    call init_mytimer ( timer_energy )
    call init_mytimer ( timer_complexes )
    call init_mytimer ( timer_flex )
    call init_mytimer ( timer_update_rate )

    write (*,*)
    write (*,*) '========================================================================='
    write (*,*) ' Entry main loop sda 2 proteins '
    write (*,*) ' Nrun ', nrun
    !write (*,*) 'timemax ', timemax

    ! intialize the array, use only protein 1 information
    ! call init_force_energy ( param_force_energy, tab_protein % array ( 1 ) % p_sogrid )
    call init_force_energy ( param_force_energy, tab_protein, type_calc, param_metaldesolv % imetaldesolv )
    ! no hydrodynamic interactions for 2 proteins
    ! JRom: turning on hydrodynamic interactions
    if ( param_analytic % hydro ) then
        write(*,*) 'Set param_force_energy % ihi to true'
        param_force_energy % ihi = .true.
    end if
    ! summary data, should group with analytical
    !call print_info_param_fe ( param_force_energy )

    ! for test if force asymetric, uncomment to force asymetric energy and force
    ! BEBUG to uncomment
    ! param_force_energy % fast_computation = .false.

    ! not necessary, just convenient to refer 1 and 2
    protein1 => tab_protein % array ( 1 )
    protein2 => tab_protein % array ( 2 )

    ! modified only if iflex, needed in add_record
    current_conf = 1
    ! only sda_koff
    total_occur = 1

    !! new function, set up only shared variable
    call set_dummy_array_and_logical ( tab_protein, dummy_array, iflex )
    !write(*,*) 'iflex ', iflex
    !write(*,*) 'size dummy_array ', size(dummy_array)

    ! initialize array, always 2 proteins here, fixe max size to 20 terms
    if ( ( associated ( o_complexes % pcomplexe ) ) .or. ( associated ( trajectories % ptrajectory ) ) ) then

        call set_array_energy ( type_calc, o_complexes, size ( tab_protein % array ), &
                                array_energy, max_energy_terms ) !opt_force_large_array
        !! specific to sda_2proteins, need the larger one as intermediate
        size_array_energy = get_nbterms_energy_compatible ( o_complexes, .true. )
        !! need also max_energy_terms, should be in function
        ! max_energy_terms = get_max_bit_energy( o_complexes )

#ifdef DEBUG
        write(*,*) 'Allocate array_1d ', size_array_energy
        write(*,*) 'Allocate energy_1d with size_array_energy ', size_array_energy
        write(*,*) 'max_energy_terms ', max_energy_terms
#endif
        allocate ( array_energy_1d ( size_array_energy, 1 ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In allocation of array_energy size ', size_array_energy
            stop 1
        endif
        array_energy = 0.d0
        array_energy_1d = 0.d0

    ! can be commented, all is fine, not used and already nullified
    else
        write(*,*)
        write(*,*) 'WARNING No complexe / no trajectory what happens for array_energy ?? '
        ! not sure thread copies are correctly setup
        nullify ( array_energy )
        nullify ( array_energy_1d )
    endif

    ! if sda-koff need to change nrun, if we want to use occurence
    ! maybe not fixed is better
    if ( type_calc %  enum_type_calc == sda_koff ) then
        write(*,*)
        write(*,*) 'sda_koff modify nrun before', nrun

        if ( associated ( p_restart ) ) then
            !write(*,*) 'p_restart associated'
            call compute_all_occurencies ( p_restart, total_occur )
            write(*,*) 'number complexe ', p_restart % nb_max_complexe
            write(*,*) 'total_occurencies ', total_occur
        !! no restart given, should be in read_input_file !!
        else
            write(*,*) 'ERROR p_restart not associated, a restart is needed by sda_koff'
            stop 1
        end if
        !save value before overwritting
        n_repeat_occur = nrun
        nrun = total_occur * nrun
        write(*,*) 'nrun = total_occur * nrun = ', nrun
        write(*,*) 'n_repeat_occur ', n_repeat_occur

        ! reset to first data
        call reset_position ( p_restart )
        p_tmp_onecomplexe => p_restart % p_onecompl
        !write(*,*) 'first conf ', p_tmp_onecomplexe % int_ocompl ( get_indice_nconf ( p_restart ) )
        !write(*,*) 'ok is associated ', associated ( p_tmp_onecomplexe )

        if ( associated ( p_tmp_onecomplexe ) .eqv. .false. ) then
            write(*,*) 'ERROR Could not read restart data'
            stop 1
        end if
        !intialisation, fixe to 0 because updated first
        current_occur = 0

    end if

    !! need to modify allocation, nrun is modified, should move previous group much before
    allocate ( steps_record ( nrun ), stat = ierr )
    if ( ierr /= 0 ) then
        write ( io6, * ) 'ERROR In allocation of steps_record, size =', nrun
    end if
    !! open file
    ! test if parrallel modify name
    open ( io_steps, file = steps_str, form='formatted', status='replace', iostat = io_err )
    if ( io_err /= 0 ) then
        write ( io6, * ) 'ERROR In opening steps.log '
        stop 1
    end if

    ! global array for bootstrap, avoid a critical section
    if ( associated ( rcriteria % p_assoc_rate ) .and. (  rcriteria % p_assoc_rate % bootstrap .eqv. .true. ) ) then
        ! array for assoc_rate bootstrap
        if ( rcriteria % type_reaction == assoc_rate ) then
            ! array for assoc_rate bootstrap, record min_windows
            allocate ( rcriteria % p_assoc_rate % array_bootstrap ( nrun, rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating array_bootstrap '
                stop 1
            end if
            rcriteria % p_assoc_rate % array_bootstrap = 0

        else if ( rcriteria % type_reaction == elec_transfer ) then
           ! for electron transfer, only 1 contact, can use the same array
            allocate( rcriteria % p_assoc_rate % array_bootstrap_real ( nrun, rcriteria % p_assoc_rate % nb_window ),stat = ierr)
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating array_bootstrap_real '
                stop 1
            end if
            rcriteria % p_assoc_rate % array_bootstrap_real = 0.d0
        end if

        !write(*,*) 'allocated rcriteria % p_assoc_rate % array_bootstrap ', allocated ( rcriteria % p_assoc_rate % array_bootstrap )
        !write(*,*) 'allocated this % array_bootstrap_real ', allocated ( rcriteria % p_assoc_rate % array_bootstrap_real )

        !! if fpt and bootstrap
        if ( rcriteria % p_assoc_rate % fpt ) then
            allocate ( rcriteria % p_assoc_rate % array_boot_fpt ( nrun, rcriteria % p_assoc_rate % nb_window, &
                       rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating array_bootstrap '
                stop 1
            end if
            rcriteria % p_assoc_rate % array_boot_fpt = 0.
#ifdef DEBUG
            write(*,*) 'size array_boot_fpt ', size ( rcriteria % p_assoc_rate % array_boot_fpt )
#endif
        end if

    end if ! if bootstrap

    ! global variable for bd
    rhit = 0.

    ! setup rhit, according to geom, surface
    ! in case of flex, take distance_max from plist_flex == the largest from all,
    ! will make extra exclusion check, but otherwise need to setup all the time ( and in every test in change-conformation )
    call set_rhit ( tab_protein, param_probe, rhit )

    !make warning
    if ( ( tab_protein % array ( 1 ) % surface .eqv. .true. ) .and. ( geom %  type == enum_sphere ) ) then
        write(*,*) 'WARNING you specified the first Solute_Grid as a surface with a sphere geometry'
        write(*,*) 'WARNING only the z distance to the surface will be computed'
        write(*,*) 'WARNING this combination was developed for sdamm, check your rhit is correct !'
        write(*,*) 'WARNING the safe way is to set for the first Solute_Grid :'
        write(*,*) '          surface = 0, diffusion_trans = 0, diffusion_rot = 0, rotate = 0'
        write(*,*)
        write(*,*) 'WARNING you may need to delete the accessibility file (solute1.access ) also '
        write(*,*)
    endif

    !! try to use the same cutoff than sdamm, will speed when close to escape !!!
    !! need to initialize position, restart option, function in MainLoop_sdamm.f90
    !call set_cutoff ( tab_protein, geom, Cut2 )
    !write(*,*) 'Interaction cutoff square, Cut2 ', Cut2
    !write(*,*) 'Cutoff ', sqrt ( Cut2 )

    ! MIC: add set_cut_array in noomp version
    !NJB: call set_cut_array - not yet used
    call set_cut_array ( tab_protein ,geom, Cut2 )
    write(*,*) 'Interaction cutoff square, Cut2 ', Cut2

    ! JRom: setting cutoff for HI, as in sdamm
    ! could be done in set_cutoff
    if ( param_force_energy % ihi ) then
        write(*,*) 'Set cutoff for hydrodynamic interaction 2* sqrt(Cut2) ', sqrt(Cut2)*2.
        param_force_energy % cutoff_hi = sqrt(Cut2) * 2.
        
        if ( geom % type .eq. enum_box ) then
            write(*,*) 'min_sizebox = ', MINVAL ( geom % pbox % sizebox )
            if ( param_force_energy % cutoff_hi .gt. MINVAL ( geom % pbox % sizebox )/2. ) then
                write(*,*) 'INFO Local volume fraction cutoff too large. Rescaled to 1/2 of the simulation box size.'
                write(*,*) 'INFO Cutoff HI ', MINVAL ( geom % pbox % sizebox ) / 2.
                param_force_energy % cutoff_hi = MINVAL ( geom % pbox % sizebox ) / 2.
            end if
        end if
        write(*,*) 'Final cutoff for hydrodynamic interaction: ', param_force_energy % cutoff_hi
    end if

    if ( param_timestep % swd1 < rhit ) then
        write(*,*)
        write(*,*) 'WARNING Your parameters for the timestep are not adequat '
        write(*,*) 'WARNING We advice to setup swd1 > maximum size of solute 1 + solute 2 ', &
                   ' ( swd1 > ',rhit,')'
    end if

    ! new, check if there will be interaction close to the exit
    ! if no interaction fast move, otherwise computation of force, energy...
    ! write(*,*) 'sqrt(Cut2) C-x1', sqrt(Cut2), param_timestep % optimize_exit
    if ( sqrt(Cut2) > param_timestep % optimize_exit ) then
        write(*,*)
        Write(*,*) 'WARNING Cutoff :', sqrt(Cut2), ',optimize_exit : ', param_timestep % optimize_exit
        write(*,*) 'WARNING Your parameters for the timestep are not adequat '
        write(*,*) 'WARNING Close to the c surface there is still interactions between the solutes  '
        write(*,*) 'WARNING It would be faster if there is no interactions after the distance optimize_exit.'
        write(*,*) 'WARNING Try to increase the c surface.'
    endif

    ! need more data to init param_force_energy
    if ( param_analytic % ianalytic ) then
#ifdef DEBUG
        write(*,*)
        write(*,*) 'call init_analytic from Main loop'
        write(*,*) 'with cutoff ', sqrt(Cut2)
        write(*,*) 'use geometry % c-surf ', geom % stop_position
#endif

        if ( param_analytic % debyeh ) then
            ! sda 2proteins, use the c-surf as limit
            call init_analytic ( param_force_energy, tab_protein, &
                                 geom % stop_position, param_analytic % ionic )
        endif
    endif

    ! summary of all forces/energies to compute
    call print_info_param_fe ( param_force_energy )

    !!!!!!!!!! initialisation of the global counter !!!!!!!
    nstep_all = 0
    total_overlap = 0
    total_overlap_flex = 0
    nboost_all = 0
    time_all = 0.

    time_per_run = 0.
    hit = .false.

    !!!!!!!!! initialisation of diffusion parameters
    !! used in initt.f
    dm = protein1 % p_sogrid % diffusion_trans + protein2 % p_sogrid % diffusion_trans
    write(*,*)
    write(*,*) 'SDA case : the relative translational diffusion is the sum of the absolute diffusion coefficients'
    write(*,*) 'Solute 1 : ', protein1 % p_sogrid % diffusion_trans
    write(*,*) 'Solute 2 : ', protein2 % p_sogrid % diffusion_trans
    write(*,*) 'Relative diffusion coefficient : ', dm
    write(*,*)
    dm30 = dm * 30.

    if ( param_force_energy % ihi .eqv. .false. ) then
        protein2 % p_sogrid % diffusion_trans = dm
        protein2 % diff_trans = dm
    end if


    if ( associated ( rcriteria % p_assoc_rate ) ) then
        !write (*,*) 'rate calc array '
        !write (*,*) nrun, ( rcriteria % p_assoc_rate % window ( iw ), &
        !                   iw= 1, rcriteria % p_assoc_rate % nb_window )

        ! change logic because of OpenMP
        ! only electron transfert !
        if ( rcriteria % type_reaction == elec_transfer ) then
            allocate ( local_escape ( rcriteria % p_assoc_rate % nb_window, rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocation of local_escape '
            end if
        else
            allocate ( local_escape( 1, 1 ) )
        endif

        ! for association rate
        if ( rcriteria % type_reaction == assoc_rate ) then
            allocate ( local_visited_windows ( rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocation of local_visited_windows '
            end if
        else
            allocate ( local_visited_windows ( 1 ) )
        endif

        allocate ( local_sumtime ( rcriteria % p_assoc_rate % nb_window, rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In allocation of local_sumtime '
        end if

        ! local array for first time passage
        if ( rcriteria % p_assoc_rate % fpt ) then
            allocate ( local_fpt ( rcriteria % p_assoc_rate % nb_window, rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocation of local_fpt '
            end if
        else
            ! necessary ?
            allocate ( local_fpt ( 1, 1 ) )
        end if
    ! need an allocation for check reaction pair
    else
        allocate ( local_escape ( 1, 1 ) )
        allocate ( local_visited_windows ( 1 ) )
        allocate ( local_sumtime( 1, 1 ) )
        allocate ( local_fpt ( 1, 1 ) )
    end if
    local_escape = 1
    ! now init -1, set X if visited
    local_visited_windows = -1
    local_sumtime = 0.
    local_fpt = -1.

    write (*,*)
    write (*,*) '========================================================================='
    write (*,*) ' Begin run '
    write (*,*) ' nrun ', nrun
    write (*,*)
    flush ( io6 )

!! Start of BD runs, do-loop
LOOP_RUN: &
    do current_run = 1, nrun

        !write(*,*) 'current run, nrun ', current_run, nrun
        if ( MOD ( current_run, 100 ) == 0 ) then
           !write(*,*) 'current run, nrun ', current_run, nrun
           !! write in log file
           rewind ( io_steps )
           write ( io_steps, fmt='(i10)' ) current_run - 1
           write ( io_steps, fmt='(15i10)' ) ( steps_record ( k ), k=1,current_run-1 )
           flush ( io_steps )
           !! added omp
           call write_record ( o_complexes )
           !flush ( o_complexes )
        end if

        time = 0.d0
        timehit = 0.d0
        current_step = 0
        correct_format_energy = .false.

        print_step = .false.
        overlap = .false.
        bool_escape = .false.

        !!set correctly in init_time if overlap as soon as the beginning, to check !!
        hit = .false.

        call init_position_protein ( tab_protein, type_calc, geom, dist, iflex, current_conf, &
                                     p_restart, current_occur, n_repeat_occur )

#ifdef VERBOSE
write(*,*)
write(*,*) '=============================================='
write(*,*) 'init pos of prot2 after random_pos or initial if sda-koff'
    if ( iflex ) then
        write(*,*) 'conformation ', current_conf
    end if

    !protein2 % position = (/ -10.6, -10.4, -24.0 /)
    !protein2 % position = (/ 0.0, 0.0, 45.0 /)
    !write(*,*) 'pos relat ', protein2 % position - protein1 % position
    !protein2 % orientation_x =(/1.0, 0.0, 0.0/)
    !protein2 % orientation_y =(/0.0, 1.0, 0.0/)
    !protein2 % orientation_z =(/0.0, 0.0, 1.0/)
    write(*,*) 'pos ', protein2 % position
    write(*,*) 'x ', protein2 % orientation_x
    write(*,*) 'y ', protein2 % orientation_y
    write(*,*) 'z ', protein2 % orientation_z

    write(*,*) 'pos atom surf1', protein1 % p_list_atom_surface(:,1), &
            protein1 % p_list_atom_surface( :,protein1 % nb_atom_surface )
    write(*,*) 'pos atom surf2', protein2 % p_list_atom_surface(:,1) , &
            protein2 % p_list_atom_surface( :,protein2 % nb_atom_surface )
#endif

    ! for first_time passage, need a first test, cannot reproduce exactly sda6 runs anymore (??)
    if ( ( rcriteria % type_reaction == docking ) .or. ( rcriteria % type_reaction == assoc_rate ) .or. &
        ( rcriteria % type_reaction == elec_transfer ) ) then
        ! local copy and only read should not be a problem, time = 0. here
        ! dtnow not validated, set to 0 ??
        dtnow = 0.d0
        !call check_reaction_pair_old ( protein1, protein2, rcriteria, local_escape, local_sumtime, local_fpt, &
        !                           real(time,4), real(dtnow,4), clos1 )
        call check_reaction_pair ( protein1, protein2, rcriteria, local_visited_windows, local_escape, local_sumtime, &
                                   local_fpt, real(time,4), real(dtnow,4), clos1 )
        !write(*,*) 'after check_reaction_pair and current_step ', clos1, current_step

    else if ( rcriteria % type_reaction == all ) then
        clos1 = .true.
    ! else off_rc nothing to do
    endif

!!! begin BD loop
LOOP_STEP: &
        do

            ! signal handling, stop the job
            if ( EXIT_PROG .eqv. .true. ) then
                !write(*,*) 'exit prog is true'
                CYCLE LOOP_RUN ! do the job, because make the loop until the endbut
            ! not a nice way to quit ! still makes all the loop
            end if

            ! choose timestep, variable in sda_2proteins type
            !if ( param_timestep % variable ) then
            call init_time ( param_timestep, geom, dist, rhit, dm30, &
                             dtnow, hit )
            !write(*,*) 'out init_time dist dtnow rhit, hit ', dist, dtnow, rhit, hit
            ! by default never used in sda_2proteins and related, but should work
            !else
            !    dtnow = param_timestep % dt1
            !endif

            current_step = current_step + 1
#ifdef VERBOSE
            write(*,*) 'current_step', current_step
#endif
            print_step = .false.
            ! need to reset at every steps
            correct_format_energy = .false.

            !! check for printing trajectory, will need to compute energy
            if ( associated ( trajectories % ptrajectory ) ) then
                !write(*,*) 'ptraj associated print_step ', print_step, trajectories % freq_print, trajectories % ntraj_rec, current_run
                if ( ( trajectories % ntraj_rec < 0 ) .or. ( trajectories % ntraj_rec == current_run ) ) then
                    !! if freq_print == 0, print every step, otherwise modulo
                    if ( ( trajectories % freq_print == 0 ) .or. &
                         ( ( trajectories % freq_print /= 0 ) .and.  (MOD ( current_step, trajectories % freq_print ) == 0) ) ) then
                    print_step = .true.
                    end if
                end if
            end if


        !! comes back here if boost, need to recompute the force
    2   continue

!!!!!!!!!!!!!!!!!!!!!!!!!!! Compute force and torques
    call start_mytimer( timer_force )

            f =  0.d0
            t =  0.d0
            t1 = 0.d0

#ifdef VERBOSE
            write(*,*)
            write(*,*) '=== Compute force ==='
#endif

            if ( param_force_energy % fast_computation ) then

                call force_epedhdlj_2proteins_fast ( protein1, protein2, param_force_energy, dist, &
                                                     f, t, t1, dummy_array )
            else
                call force_epedhdlj_2proteins ( protein1, protein2, param_force_energy, dist, &
                                                f, t, t1, dummy_array )
            end if

            ! force is smaller than SDAMM,
            ! should be the sum in sda_2proteins
#ifdef VERBOSE
            !write(*,*) 'after uhbd forces '
            ! pometcs elctro. desolvation, test with input = 1.67 * 2 = 3.34 works at 70 A
            !write(*,*) 'f ', f
            !write(*,*) 't ', t
            !write(*,*) 't1 ', t1
#endif

            if ( param_force_energy % iimg_chg ) then
                !write(*,*) 'Will compute image_charge'
                ! apply only for protein 2 here
                ! total force is updated inside the function
                call force_image_charge ( tab_protein, param_force_energy, f, t )
            endif

            ! if metal_desolv...not implemented, strange results
            if ( param_force_energy % ianalytic ) then

                if ( param_force_energy % idebye ) then
                    ! compute and add the force inside the function
                    call force_analytic_dh ( tab_protein, param_force_energy, dist, f, t, t1 )

                endif
                ! if others...
            endif


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#ifdef VERBOSE
    write(*,*) 'Total FORCE '
    write (*,'(''Test new force f: '',3E14.5 )') f
    write (*,'(''Test new torque t: '',3E14.5 )') t
    write (*,'(''Test new torque t1: '',3E14.5 )') t1
    !stop 1
#endif
  call stop_mytimer( timer_force )
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!! Make BD move, try to move and rotate, check for overlap,
           !   come back directly to LOOP_OVERLAP if move is rejected

    call start_mytimer ( timer_bd )
        accept = .false.

    ! JRom: re-calculating the diffusion coefficient
        if ( param_force_energy % ihi ) then
#ifdef VERBOSE
            !write(*,*) 'ihi activated,call scale diff coeff with cutoff_hi ', param_force_energy % cutoff_hi
            write(*,*)
            write(*,*) '=== Before t_diff_coeff, hydrodynamics part ==='
#endif
    ! rescale diffusion coeeficient with Weinstein method
    ! keep cutoff as 4 in memory, like other if later regrouped in one array
          call scale_diffusion_coefficients ( tab_protein, geom, real(param_force_energy % cutoff_hi,8) )
        end if

LOOP_OVERLAP: &
        do nb_overlap = 1, nmax_overlap+1

            call make_bd_move_2prot_timer_excl ( protein1, protein2, geom, &
                                         f, t, t1, &
                                         dtnow, dist, rhit, &
                                         hit, accept, bool_escape, total_overlap, &
                                         timer_exchk )

            !write(*,*) 'after make_bd_move dist ', dist

            !! finish trajectory if protein  escapes
            !! could finalize and CYCLE BDLOOP to avoid GO TO
            if ( bool_escape ) then
                call stop_mytimer ( timer_bd )
                !! go out of this run / traj
                !write(*,*) 'will escape'
                GO TO 333
            endif
            !! if move accepted exit LOOP_OVERLAP
            !!    continue in LOOP_STEP
            if ( accept .eqv. .true. ) exit

        end do LOOP_OVERLAP

        !! if accept is still false at this time, means nb_overlap == nmax_overlap => call boost
        if ( accept .eqv. .false. ) then
            ! boost check if surface, increase only z
            call boost( protein2 % position, geom, dist, rboost)
            !write (*,*) 'rboost after', rboost

            !! here need to update dist
            call compute_distance ( protein1, protein2, dist )
            !! test in case of boost, to delete
            !write(*,*) 'new pos relat', protein2 % position
            !write(*,*) ' new dist ', dist

            !! increment boost counter
            nboost_all = nboost_all + 1
            !! recompute forces with the new position, and renter overlap loop
            call stop_mytimer ( timer_bd )
            !! goes back just beforce force calculation, end of overlap and boost
            GO TO 2
        end if

    call stop_mytimer ( timer_bd )
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        ! update time
        time = time + dtnow
        ! why timehit ?? hit only to check when protein may be excluded
        ! hit is a test for not doing exchk because sure to be false
        ! only timehit is compared to timemax
        if ( hit ) then
            timehit = timehit + dtnow
            !write(*,*) 'update timehit ', timehit
            !write(*,*) 'time ', time
        end if

        ! which distance, surface or receptor on membrane ??
        ! for plan no problem, only for box
        ! for dist, default dist. May be a spot in a surface ?? need other parameters
        !write(*,*) 'update residence_time ', dist, dtnow
        call update_residence_time ( resid_time, protein2 % position, dist, dtnow )

        !! just for test, to be independent of timehit of normal sda
        !!write(*,*) 'current_step ', current_step
        !if ( current_step == 20 ) GO TO 333

!!!!!!!!!!!!!!!!!!!!!! Reaction Criteria check
        !! setup clos1 to true if contact is detected ??? could be at the start
        clos1=.false.
    call start_mytimer ( timer_reactpair )

        !! no problem of distance here, reaction pairs defined explicitely
        !! may add new type, as a z distance
        if ( ( rcriteria % type_reaction == docking ) .or. ( rcriteria % type_reaction == assoc_rate ) .or. &
             ( rcriteria % type_reaction == elec_transfer ) ) then
            !!to add escape
            !call check_reaction_pair_old ( protein1, protein2, rcriteria, local_escape, local_sumtime, local_fpt, &
            !                           real(time,4), real(dtnow,4), clos1 )
            call check_reaction_pair ( protein1, protein2, rcriteria, local_visited_windows, local_escape, local_sumtime, &
                                       local_fpt, real(time,4), real(dtnow,4), clos1 )
            !write(*,*) 'after check_reaction_pair and current_step ', clos1, current_step

        else if ( rcriteria % type_reaction == all ) then
                !write(*,*) 'always clos1 to true'
                clos1 = .true.
        !!else of_rc nothing to do
        endif

        !write(*,*) 'close ', clos1
        ! should check if associated before printing
        ! write(*,*) 'resc 1', rcriteria % p_assoc_rate % escape(1,1)
    call stop_mytimer ( timer_reactpair )

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Complexes to record AND no trajectories to print  => no energy to compute

        if ( ( associated ( o_complexes % pcomplexe ) .and. ( clos1 ) ) .or. &
             ( print_step .eqv. .true. ) ) then

!!!!!!!!!  Compute energy if needed
     call start_mytimer ( timer_energy )

#ifdef VERBOSE
            write(*,*)
            write(*,*) '=== Compute energy ==='
#endif

            ! new array energy  real ( kind=8 ), dimension(10) :: array_energy
            ! initialize all terms to 0
            array_energy = 0.d0

            if ( param_force_energy % fast_computation ) then

                call energy_epedhdlj_2proteins_fast ( protein1, protein2, param_force_energy, &
                                                      array_energy, dummy_array, dist )

            else
                call energy_epedhdlj_2proteins ( protein1, protein2, param_force_energy, dist , &
                                                 array_energy, dummy_array )
            end if

            ! test image charge, could outside the loop if image_charge without lennard-jones
            ! would need to add inside change_conformation
            if ( param_force_energy % iimg_chg ) then
                    call energy_image_charge ( tab_protein, param_force_energy, array_energy )
            endif

            if ( param_force_energy % imetaldesolv ) then
                !write(*,*) 'Will compute metal desolvation energy'
                 !! energy updated inside
                 call compute_energy_metaldesolv_2prot( tab_protein, param_metaldesolv, array_energy )
            end if

            if ( param_force_energy % ianalytic ) then

                if ( param_force_energy % idebye ) then
                    ! compute and add the energy to the electrostatic energy
                    !write(*,*) 'normal electrostatic energy ', array_energy ( 2 , 1 ), array_energy (2, 2)
                    call energy_analytic_dh ( tab_protein, param_force_energy, dist, array_energy )
                endif
            end if

            ! same definition than sdamm, now sum up half of potential energy, to use with epfct = 0.5
            total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
            !! put total energy in first position
            array_energy (1,2) = total_energy
            !array_energy (1,1) = 0.d0
            !write(*,*) 'array_energy', array_energy (:,2)

#ifdef VERBOSE
    write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
    write(*,*) ' ENERGY '
    write(*,*) ' total energy ', array_energy ( 1, 2 )
    write(*,*) ' potential energy', array_energy ( 2 , 1 ), array_energy (2, 2)
    write(*,*) ' desol energy', array_energy( 3, 1 ), array_energy( 3, 2 )
    !write(*,*) ' hydro. energy', array_energy( 4, 1 ), array_energy ( 4 , 2 )
    !write(*,*) ' ljrep energy', array_energy( 5, 1 ), array_energy ( 5 , 2 )
    !write(*,*) ' lj energy', array_energy( 6, 1 ), array_energy ( 6 , 2 )
    !write(*,*) ' img_chg energy', array_energy( 7, 1 ), array_energy ( 7 , 2 )
    !write(*,*) ' metal desol energy', array_energy( 8, 1 ), array_energy( 8, 2 )
#endif

    call stop_mytimer ( timer_energy )
!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!! case of complexes
    !! check max energy in complexes, if total_energy is higher, nothing to do

            !!! case of complexe to record
            if ( ( associated (o_complexes % pcomplexe ) ) .and. ( clos1 ) ) then
                !write(*,*) 'current_conf protein1 ', tab_protein % array ( 1 ) % p_flex % current_conf
                !write(*,*) 'current_conf protein1 ', current_conf

            call start_mytimer ( timer_complexes )

               !! sda 2 proteins only one flexible, use variable current_conf default 1 modifed only iflex and change conformation
               !write(*,*) 'Before add record in complexe run,step,time,conf',  current_run, current_step, time, current_conf

               ! for sda_2proteins need to make the format compatible into 1dimension array
                ! need additional, max_index to check, easy to get only once from complexe ? trajectory ?
                call convert_to_record_format ( array_energy, array_energy_1d, max_energy_terms )
                ! set variable to say it has been done already
                correct_format_energy = .true.

               call add_record ( o_complexes, tab_protein % array ( 2:2 ), current_run, current_step, &
                                 array_energy_1d, real(time,4), current_conf )
            !write(*,*) 'Print complexe '
            !call print_listcomplexe( o_complexes % pcomplexe, .true. )
            call stop_mytimer ( timer_complexes )
            !! end if record complexe
            end if

            !!! case of trajectory to write
            if ( print_step ) then
                !write(*,*) 'add record in trajectories freq_print, ntraj_rec, current_run, current_step ', &
                !  trajectories % freq_print, trajectories % ntraj_rec, current_run, current_step

                ! not sure the format has already been converted
                ! or add a variable correct_format_energy : true / false
                if ( correct_format_energy .eqv. .false. ) &
                    call convert_to_record_format ( array_energy, array_energy_1d, max_energy_terms )

                call add_record ( trajectories, tab_protein % array (2:2 ), current_run, current_step, &
                                  array_energy, real(time,4), current_conf )
            !write(*,*) 'record traj done'
            end if

        ! end if energy to compute
        end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Test timemax and finish trajectory

        !! exit trajectory if maximum time is exceeded or infinite if 0.
        !! why timehit ?? could be time_all for docking
        !write(*,*) 'timehit ', timehit

        ! for testing, force only 5 steps
        !if ( current_step > 5. ) then
        if ( (timemax > 0. ) .and. ( timehit > timemax) ) then
            !write(*,*) 'will finish trajectory'
            GO TO 333
        end if

!!!!  Look for changing conformation
        if ( iflex ) then

        call start_mytimer ( timer_flex )

            if ( protein1 % iflex ) then
                !!update statistic
                call update_visited_conf ( protein1, real(dtnow,4) )

                !! decrease time flex until 0 or negative
                protein1 % p_flex % time_next_flex = protein1 % p_flex % time_next_flex &
                                                     - real(dtnow,4)
                !write(*,*) 'update time next flex ', protein1 % p_flex % time_next_flex

                if ( protein1 % p_flex % time_next_flex <= 0. ) then
                    !write ( *,*) 'Will make change conf for protein1 '
                    call change_conformation ( protein1, tab_protein, &
                                               geom, param_force_energy, param_metaldesolv, type_calc, dummy_array, &
                                               hit, accepted_new_conf, total_overlap_flex, &
                                               energy_epedhdlj_flex  )
                    !time_flex = 0.
                    ! update variable current_conf
                    current_conf = protein1 % p_flex % current_conf
                    !! fisrt developped for sdamm, should be valid here also
                    !! here should not be problem of multiple change, execept prot1 and prot2 flex at the same time
                    call init_flexible_freq ( tab_protein % array ( 1 ) % p_flex )
                end if
            end if

            !! check here because only 1 move accepted by timestep
            !! commented because both flexible is not yet implemented
            !! not good need to update time for protein2
            !if ( accept_new_conf .eqv. .true. ) then
            !    write(*,*) 'prot1 new conf accepted, reinit time '
            !    call init_flexible_freq ( protein1 % p_flex )
            !    exit
            !endif

            !write(*,*) 'prot1 not accepted test prot2 '
            !! check for protein2
            if ( protein2 % iflex ) then
                call update_visited_conf ( protein2, real(dtnow,4) )
                !write(*,*) 'After update visited conf'
                protein2 % p_flex % time_next_flex = protein2 % p_flex % time_next_flex &
                                                     - real(dtnow,4)
                !write(*,*) 'update p2 time next flex ', protein2 % p_flex % time_next_flex

                if ( protein2 % p_flex % time_next_flex <= 0. ) then
                    !write ( *,*) 'Will make change conf for protein2 '
                    call change_conformation ( protein1, tab_protein, &
                                               geom, param_force_energy, param_metaldesolv, type_calc, dummy_array, &
                                               hit, accepted_new_conf, total_overlap_flex, &
                                               energy_epedhdlj_flex  )
                    ! update variable current_conf
                    current_conf = protein2 % p_flex % current_conf
                    call init_flexible_freq ( tab_protein % array ( 2 ) % p_flex )
                end if
                ! end test protein 2
            end if

        call stop_mytimer ( timer_flex )
        end if ! end if iflex

    end do LOOP_STEP

!..here when the trajectory is finished.... timemax exceeded or escape
!..for statistics...
  333   continue


    if ( ( rcriteria % type_reaction == assoc_rate ) .or. ( rcriteria % type_reaction == elec_transfer ) ) then

 call start_mytimer ( timer_update_rate )
        ! need to add new variable escape
        call update_escape ( rcriteria % p_assoc_rate, current_run, local_visited_windows, local_escape, &
                             local_sumtime, local_fpt )
        local_escape = 1.0d0
        local_visited_windows = -1
        local_sumtime = 0.
        local_fpt = -1.
 call stop_mytimer ( timer_update_rate )
    end if

    ! increment only when the trajectory has finished, important for association rate    
    run_done = run_done + 1

    ! update step and time
    nstep_all = nstep_all + current_step
    time_all = time_all + time
    ! fill record of the number of steps
    !if ( current_run <= nrun ) steps_record ( current_run ) = current_step
    steps_record ( current_run ) = current_step

    ! use default value true, recompute dispersion
    call write_record ( o_complexes )

!!! end loop over nrun
  end do LOOP_RUN

!!!! Here could be in the main code !!!!!!!! with a test on existence of resid_array
!!!! Easier if belong to geom
!!!!!!!!!!!!!!!!!!!!!!!    Print final data !!!!!!!!!!!!!!!!!!!!!!!
  time_per_run = time_all / float ( nrun )
  WRITE (io6,621)  nstep_all, time_all, time_per_run, &
     +                 total_overlap, total_overlap_flex, nboost_all

  621 FORMAT (/,'stat o:    total steps these runs:',i12,/, &
             'stat o:      total simulated time:',f13.0,' ps',/, &
             'stat o:    simulated time per run:',f13.0,' ps',/, &
             'stat o: total overlaps these runs:',i12,/, &
             'stat o:        total overlap flex:',i12,/, &
             'stat o:   total boosts these runs:',3i12,/)

    !! write the last one and close the file
    rewind ( io_steps )
    write ( io_steps, fmt='(i10)' ) current_run - 1
    write ( io_steps, fmt='(15i10)' ) ( steps_record ( k ), k=1,nrun )
    close ( io_steps )

    !! print residence time
    status = print_residence_time ( resid_time, geom, nrun )
    if ( resid_time % enum_rt == rt_resid3d ) then
        call write_resid_3d ( resid_time, protein1 % center_protein )
    end if
    !! print info about visited conformation
    if ( iflex ) then
        if ( protein1 % iflex ) then
            write ( io6, * )
            write ( io6, * ) 'Protein1'
            call print_visited_conf ( protein1 )
        end if
        if ( protein2 % iflex ) then
            write ( io6, * ) 'Protein2'
            call print_visited_conf ( protein2 )
        end if
    end if

    !! sda-koff not implemented ??

    ! print info about association rat
    if ( rcriteria % type_reaction == assoc_rate ) then

        if ( rcriteria % p_assoc_rate % bootstrap .eqv. .true. ) then

          ! write output to file only at the end of the simulation
          !call write_boot_file ( rcriteria % p_assoc_rate, nrun )
          call write_boot_file ( rcriteria % p_assoc_rate, run_done )

          ! can suppress the deallocation, no problem because only one copy
          ! delete here because allocated in this function, not in the module
          if ( allocated ( rcriteria % p_assoc_rate % array_bootstrap ) ) then
            deallocate ( rcriteria % p_assoc_rate % array_bootstrap, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocating array_bootstrap'
                STOP 1
            end if
        end if

        ! used for electron_transfer
        if ( allocated ( rcriteria % p_assoc_rate % array_bootstrap_real ) ) then
            deallocate ( rcriteria % p_assoc_rate % array_bootstrap_real, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocating array_bootstrap_real'
                STOP 1
            end if
        end if

!       ! test for fpt
        if ( rcriteria % p_assoc_rate % fpt .eqv. .true. ) then
             ! write to file, done before
            deallocate ( rcriteria % p_assoc_rate % array_boot_fpt, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocating array_boot_fpt'
                STOP 1
            end if
        end if
      end if

        if ( type_calc %  enum_type_calc == sda_koff ) then
            call print_table_rate( rcriteria % p_assoc_rate, nrun, .true. )
        else
            !call print_table_rate( rcriteria % p_assoc_rate, nrun, .false. )
            call print_table_rate( rcriteria % p_assoc_rate, run_done, .false. )
        end if
    end if

    !! Print timer data, only local timer
    write(*,*)
    write(*,*) 'Mytimer force'
    call print_mytimer (timer_force)
    write(*,*) 'Mytimer energy'
    call print_mytimer (timer_energy)
    write(*,*) 'Mytimer bd'
    call print_mytimer (timer_bd)
    write(*,*) 'Mytimer reaction'
    call print_mytimer (timer_reactpair)
    write(*,*) 'Mytimer update rate'
    call print_mytimer (timer_update_rate )
    write(*,*) 'Mytimer exchk'
    call print_mytimer (timer_exchk)
    write(*,*) 'Mytimer complexes'
    call print_mytimer (timer_complexes)
    write(*,*) 'Mytimer flex'
    call print_mytimer (timer_flex)

    !! need to deallocate ?? Not really, no complain from valgrind.
    if ( associated ( dummy_array ) ) &
         deallocate ( dummy_array )


    if ( associated ( local_escape ) ) deallocate ( local_escape )
    nullify ( local_escape )

    if ( associated ( local_visited_windows ) ) then
        deallocate ( local_visited_windows, stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR In deallocation of local_visited_windows'
        end if
        nullify ( local_visited_windows )
    end if

    !! same for forgoten other deallocation, deleted automatically when quitting function ??
    deallocate ( steps_record, stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR In deallocating steps_records '
    end if

    !deallocate ( array_energy, stat = ierr )
    if ( associated ( array_energy ) ) then
        deallocate ( array_energy, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy '
        end if
    end if
    if ( associated ( array_energy_1d ) ) then 
        deallocate ( array_energy_1d, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy_1d '
        end if
    end if
    write(*,*) 'Exit main loop sda 2 proteins'

end subroutine MainLoop_sda_2proteins


