!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file MainLoop_sda_2proteins_omp.f90
!! Parralel version of sda 2proteins with OpenMP
!!
!! Main loop for BD with 2 solutes\n
!! Used for normal sda or sda_koff\n
!! Can compute association rate, docking or electron transfert calculation with 2 proteins,\n
!! Residence time, first passage time\n

! here only code comment
! 2nd version omp, use local_complexe for OpenMP.
! every thread has a local copy of complexe, to merge with original every nsteps.

!> MainLoop_sda_2proteins_omp
!!
!! Main function for 2 solutes computation
!!
!! @param tab_protein_unique  : contains all proteins, instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param rcriteria : instance of \ref mod_reactioncriteria::react_criter "react_criter"
!! @param o_complexes : for writing complexes
!! @param trajectories : for writing trajectory
!! @param geom : geometry instance
!! @param p_restart : pointer on a restart record
!! @param resid_time : instance of \ref mod_residencetime::residence_time "residence_time"
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param param_timestep : instance of \ref com_sda::parameter_timestep "parameter_timestep"
!! @param param_probe : instance of \ref com_sda::probe_type "probe_type"
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy
!! @param param_analytic : structure \ref com_sda::parameter_analytic "parameter_analytic"
!! @param param_metaldesolv : instance of \ref com_sda::parameter_metaldesolv "parameter_metaldesolv"
!! @param option_omp : option for openmp
!! @param nmax_overlap : could with other parameter
!! @param rboost : distance of boost
!! @param timemax : maximum time for a trajectory, 0 means infinite
!! @param nrun : number of runs
subroutine MainLoop_sda_2proteins_omp ( tab_protein_unique, rcriteria, &
                                        o_complexes, trajectories, geom, p_restart, resid_time, &
                                        type_calc, param_timestep, param_probe, param_force_energy, &
                                        param_analytic, param_metaldesolv, option_omp, &
                                        nmax_overlap, rboost, timemax, nrun )

     USE mod_array_protein ! include mod_protein, which includes mod_setofgrid, which includes mod_grid
     USE mod_gridtype
     USE mod_ReactionCriteria
     USE mod_geometry
     USE mod_record
     USE mod_ResidenceTime
     USE AllTimer
     USE com_sda
     USE mod_force_energy
     USE mod_compute_force
     USE mod_compute_energy
     USE mod_compute_energy_flex
     USE mod_compute_image_charge
     USE mod_compute_metaldesolv
     USE mod_compute_analytic_2proteins
     USE mod_set_parameters
     USE mod_init_position
     USE mod_hydrodynamics
!    Only used if OpenMP
!$   USE omp_lib
     IMPLICIT NONE

     interface

            subroutine convert_to_record_format ( array_energy, array_energy_1d, max_bit_energy )
                IMPLICIT NONE
                real(kind=8), dimension(:,:), intent ( in ) :: array_energy
                real(kind=8), dimension(:,:), intent ( out ) :: array_energy_1d
                integer, intent ( in ) :: max_bit_energy
            end subroutine convert_to_record_format

            !> called if flexible
            subroutine change_conformation ( prot, tab_prot, geom, param_force_energy, param_metal, type_calc, dummy_array, &
                                             hit, accepted, nb_overlap_flex, ptrf_energy )
                USE mod_force_energy
                USE mod_array_protein
                USE mod_geometry
                USE mod_type_calc
                USE mod_compute_energy_flex
                USE com_sda
                IMPLICIT NONE
                interface

                subroutine ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                                         dummy_array, energy, energy_img_chg )
                  USE mod_force_energy
                  USE mod_array_protein
                  USE mod_geometry
                  USE com_sda
                  implicit none
                  type ( protein ), intent(in), target :: prot
                  type ( array_protein_type ), intent( in ), target :: tab_prot
                  type ( geometry ), intent( in ) :: geom
                  type ( type_force_energy ), intent ( in ) :: param_force_energy
                  type ( parameter_metaldesolv ), intent ( in ) :: param_metal
                  real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
                  real ( kind=8 ) :: energy, energy_img_chg
                  integer, intent ( in ) :: nb_prot
                end subroutine ptrf_energy

                end interface

                type ( protein ) :: prot
                type ( array_protein_type ), intent( in ), target :: tab_prot
                type ( geometry ), intent( in ) :: geom
                type ( type_force_energy ), intent ( in ) :: param_force_energy
                type ( parameter_metaldesolv ), intent ( in ) :: param_metal
                type ( type_calculation ), intent ( in ) :: type_calc
                real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
                integer :: nb_overlap_flex
                logical, intent ( in ) :: hit
                logical, intent ( out ) :: accepted
            end subroutine change_conformation

            subroutine check_reaction_pair ( prot1, prot2, rcriteria, visited_windows, escape, sum_time, &
                                             first_time, time, dtnow, close )
                USE mod_protein
                USE mod_ReactionCriteria ! include mod_rate_calculation
                IMPLICIT NONE
                type ( protein ), intent ( in ) :: prot1, prot2
                type ( react_criter ) :: rcriteria
                real ( kind=4 ), dimension(:,:) :: sum_time, first_time
                real ( kind=8 ), dimension (:,:) :: escape
                integer, dimension (:) :: visited_windows
                real ( kind=4 ) :: time, dtnow
                logical, intent( out ) :: close
            end subroutine check_reaction_pair

            subroutine make_bd_move_2prot ( prot1, prot2, geom, force, torque, torque1, &
                                            dtnow, dist, rhit, hit, &
                                            accept, escape, total_overlap )
                USE mod_protein
                USE mod_geometry
                IMPLICIT NONE
                type ( protein ) :: prot1, prot2
                type ( geometry ), intent ( in ) :: geom
                real ( kind=8 ), dimension ( 3 ), intent ( in ) :: force, torque, torque1
                real ( kind=8 ), intent ( in ) :: dtnow
                real ( kind=4 ) :: dist, rhit
                integer :: total_overlap
                ! accept true if at least one motion has been accepted
                logical, intent ( out ) :: hit, escape, accept
            end subroutine make_bd_move_2prot

            ! has been moved to check the validity of input
!            subroutine check_timestep( geom, param_timestep, dm30 )
!                USE mod_geometry
!                USE com_sda
!                IMPLICIT NONE
!                type ( geometry ) :: geom
!                type ( parameter_timestep ) :: param_timestep
!                real ( kind=8 ) :: dm30
!            end subroutine check_timestep

            subroutine init_time ( param_timestep, geom, dist, rhit, dm30, &
                       dtnow, hit )

                USE com_sda
                USE mod_geometry
                IMPLICIT NONE
                type ( parameter_timestep ), intent ( in ) :: param_timestep
                type ( geometry ), intent ( in ) :: geom
                real ( kind=8 ), intent ( in ) :: dm30
                real ( kind=8 ), intent ( out ) :: dtnow
                real ( kind=4 ), intent( in ) :: rhit
                real ( kind=4 ), intent( in ) :: dist
                logical, intent ( out ) :: hit
            end subroutine init_time

     end interface

     !! should be the reference, whatever OpenMP or not
     type ( array_protein_type ), target :: tab_protein_unique
     type ( react_criter ) :: rcriteria
     type ( record ), target :: o_complexes, trajectories
     type ( record ), pointer :: p_restart
     type ( geometry ) :: geom
     type ( residence_time ) :: resid_time
     type ( type_calculation ) :: type_calc
     type ( parameter_timestep ) :: param_timestep
     type ( probe_type ) :: param_probe
     type ( type_force_energy ) :: param_force_energy
     type ( parameter_analytic ) :: param_analytic
     type ( parameter_metaldesolv ) :: param_metaldesolv
     !! general structure to group all omp option, only complexe for now
     type ( type_option_omp ) :: option_omp
     real ( kind=8 ) :: timemax
     real ( kind=4 ) :: rboost
     integer :: nrun, nmax_overlap

     !Only for OpenMP
!$   integer :: tid

!$   integer :: merge_step, merge_step_shared, size_thread_complexe

     ! make a general pointer ptab_protein, allows to use one variable only, needed for openmp
     type ( array_protein_type ), pointer :: ptab_protein
     ! local variables, used for simplicity
     type ( protein ), pointer :: protein1, protein2

     ! same for complexes
     type ( record ), pointer :: p_local_complexe

     ! to record energy, array_energy will be private, specification 2.5 gcc 4.3.2 cannot be allocatable
     ! use a pointer instead
     integer :: max_energy_terms
     real ( kind=8 ), dimension(:,:), pointer:: array_energy
     real ( kind=8 ) :: total_energy
     ! new, fixe format for entry to add_record. Expect 2 dimen in add_record, but only 1 if filled
     ! allocated as ( 20, 1 )
     integer :: size_array_energy
     real ( kind=8 ), dimension(:,:), pointer :: array_energy_1d
     logical :: correct_format_energy

     !! for recording steps, in this version global_steps_record is global
     integer, dimension(:), allocatable :: global_steps_record

     ! variable for force and torques
     real ( kind=8 ), dimension(3) :: f, t, t1

     ! local needed for array for every thread
     real (kind=8), dimension(:,:), pointer :: local_escape
     ! equivalent of local_escape with new implementation
     integer, dimension(:), pointer :: local_visited_windows

     real ( kind=4 ), dimension(:,:), pointer :: local_sumtime, local_fpt

     real ( kind=8 ) :: dm, dm30
     real ( kind=8 ) :: time, time_all, time_per_run, timehit, dtnow
     real ( kind=4 ) :: Cut2, dist, rhit

     integer :: current_run, run_done, nstep_all, current_step, current_conf
!$   integer :: thread_nrun, i
     integer :: total_overlap, nboost_all, total_overlap_flex
     integer :: k, nb_overlap, ierr

     ! for sda-koff,
     ! need internal p_current_node in mod_record if want to avoid explicit declaration here, done to delete if ok
     ! to see with all tools as well
     integer :: current_occur, n_repeat_occur, total_occur

     logical :: overlap, bool_escape, print_step
     logical :: clos1, hit, accept
     ! for conformation change
     logical :: iflex, accepted_new_conf
     integer :: total_test_change_conformation, total_accepted_new_conf

     ! for printing steps.log
     character*32 :: steps_str

    ! used for lennardjones, contains only 1.0, comes from sdamm, needed for common flex energy evaluation
    ! and new version of force
    ! with version fast, no need of dummy array ?? to check
    real ( kind=8 ), dimension( : ), pointer :: dummy_array
    integer :: status, io_err

    nullify ( ptab_protein )
    nullify ( protein1, protein2 )
    nullify ( p_local_complexe )
    !! used for sda_koff, reading starting position, need to be private, not sure, shared rather
    !! must be shared !! should use the internal pointer of restart
    !nullify ( p_tmp_onecomplexe )
    nullify ( dummy_array )
    nullify ( array_energy_1d )

    !! create steps.log, take into account if fparrellel
    steps_str = 'steps.log'
    if ( type_calc % bool_par ) then
    !!! keep the suffix, add to filename the integer as string
        call create_filename_flex ( steps_str, type_calc % nb_par )
        !write(*,*) 'new log ', steps_str
    end if

    !! open file only for master
    ! test if parrallel modify name
    open ( io_steps, file = steps_str, form='formatted', status='replace', iostat = io_err )
    if ( io_err /= 0 ) then
        write ( io6, * ) 'ERROR In opening steps.log '
        stop 1
    end if

    size_array_energy = 0
    max_energy_terms = 0
    iflex = .false.
    accepted_new_conf = .false.
    correct_format_energy = .false.
    total_test_change_conformation = 0
    total_accepted_new_conf = 0
    run_done = 0

    ! option for OpenMP
!$  merge_step = 0
!$  merge_step_shared = option_omp %  merge_complexe
!$  size_thread_complexe = option_omp % maxsize_thread_complexe
!! !$    write(*,*) 'before parallel option openmp ', merge_step, merge_step_shared, size_thread_complexe

    write (*,*)
    write (*,*) '========================================================================='
    write (*,*) ' Entry main loop sda 2 proteins open mp '
    write (*,*) ' Nrun ', nrun
!! !$    write (*,*) 'tid ', tid



    ! debye huckel, need maximum size of the array to store
    ! 2 proteins, it is fixed to 8* size ep_grid
    ! SDAMM, same or half the box: maybe use as default  half - 20 A shoulmd be ok
    call init_force_energy ( param_force_energy, tab_protein_unique, type_calc, param_metaldesolv % imetaldesolv )
    ! JRom: turning on hydrodynamic interactions
    if ( param_analytic % hydro ) then
        write(*,*) 'Set param_force_energy % ihi to true'
        param_force_energy % ihi = .true.
    end if
    ! summary data, should group with analytical
    !call print_info_param_fe ( param_force_energy )

    ! for test if force asymetric, uncomment to force asymetric energy and force
    ! BEBUG to uncomment
    ! param_force_energy % fast_computation = .false.

    !! new function, set up only shared variable
    call set_dummy_array_and_logical ( tab_protein_unique, dummy_array, iflex )
    !write(*,*) 'iflex ', iflex
    !write(*,*) 'size dummy_array ', size(dummy_array)

    ! deal with koff before parallel
    if ( type_calc %  enum_type_calc == sda_koff ) then
        write(*,*)
        !write (*,*)'sda_koff modify nrun before', nrun

        ! need to compute all runs, false if not want to count occurencies, if true will send back nb_complexe
        if ( associated ( p_restart ) ) then
            !write(*,*) 'p_restart associated'
            call compute_all_occurencies ( p_restart, total_occur )
#ifdef DEBUG
            write(*,*) 'number complexe ', p_restart % nb_max_complexe
            write(*,*) 'total_occurencies ', total_occur
#endif
        else
            write(*,*) 'ERROR p_restart not associated, a restart is needed by sda_koff'
            stop 1
        end if
        !save value before overwritting
        n_repeat_occur = nrun
        nrun = total_occur * nrun
#ifdef DEBUG
        write(*,*) 'nrun = n_occur * nrun = ', nrun
        write(*,*) 'n_repeat_occur ', n_repeat_occur
#endif
        ! init pointer, first position after compute_all_occurency
        ! reset to first data, lost after OpenMP loop to do again, restart shared ok
        call reset_position ( p_restart )
        ! intialisation, fixe to 0 because will be updated first
        current_occur = 0
    end if

    ! need to modify allocation, nrun is modified, should move previous group much before
    ! one global array for all threads, shoudl not be a problem for OpenMP

    ! global_steps_record is a global array to be created before parallel region
    ! write(*,*) 'allocate step record before parrallel'
    allocate ( global_steps_record ( nrun ), stat = ierr )
    if ( ierr /= 0 ) then
        write ( io6, * ) 'ERROR In allocation of steps_record, size =', nrun
        STOP 1
    end if
    global_steps_record = 0

    !! global array for bootstrap, avoid a critical section because one traj by thread
    if ( associated ( rcriteria % p_assoc_rate ) .and. (  rcriteria % p_assoc_rate % bootstrap .eqv. .true. ) ) then

        !write(*,*) 'bootstrap yes,', rcriteria % type_reaction
        if ( rcriteria % type_reaction == assoc_rate ) then
            ! array for assoc_rate bootstrap, record min_windows
            !write(*,*) 'allocate array_bootstrap ', rcriteria % type_reaction
            allocate ( rcriteria % p_assoc_rate % array_bootstrap ( nrun, rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating array_bootstrap '
                stop 1
            end if
            rcriteria % p_assoc_rate % array_bootstrap = 0

        else if ( rcriteria % type_reaction == elec_transfer ) then
            !write(*,*) 'allocate bootstrap_real ', rcriteria % type_reaction
            ! for electron transfer, only 1 contact, can use the same array
            allocate( rcriteria % p_assoc_rate % array_bootstrap_real ( nrun, rcriteria % p_assoc_rate % nb_window ),stat = ierr)
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating array_bootstrap_real '
                stop 1
            end if
            rcriteria % p_assoc_rate % array_bootstrap_real = 0.d0
        end if

        ! if fpt and bootstrap
        if ( rcriteria % p_assoc_rate % fpt ) then
            allocate ( rcriteria % p_assoc_rate % array_boot_fpt ( nrun, rcriteria % p_assoc_rate % nb_window, &
                       rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating array_boot_fpt '
                stop 1
            end if
            rcriteria % p_assoc_rate % array_boot_fpt = 0.
#ifdef DEBUG
            write(*,*) 'size array_boot_fpt ', size ( rcriteria % p_assoc_rate % array_boot_fpt )
#endif
        end if

    end if ! if bootstrap

    ! global variable for bd, only 2 proteins
    rhit = 0.
    ! setup rhit, according to geom, surface
    ! in case of flex, take distance_max from plist_flex == the largest from all,
    ! will make extra exclusion check, but otherwise need to setup all the time ( and in every test in change-conformation )
    call set_rhit ( tab_protein_unique, param_probe, rhit )

    ! make warning
    if ( ( tab_protein_unique % array ( 1 ) % surface .eqv. .true. ) .and. ( geom %  type == enum_sphere ) ) then
        write(*,*) 'INFO You specified the first Solute_Grid as a surface with a sphere geometry'
        write(*,*) 'INFO Only the z distance to the surface will be computed'
        write(*,*) 'INFO This combination was developed for sdamm, check your rhit is correct !'
        write(*,*) 'INFO The safe way is to set for the first Solute_Grid :'
        write(*,*) '          surface = 0, diffusion_trans = 0, diffusion_rot = 0, rotate = 0'
        write(*,*)
        write(*,*) 'INFO You may need to delete the accessibility file (solute1.access ) also '
        write(*,*)
    endif

    !NJB: call set_cut_array - cutoff for each interaction, Cut2 global the largest one
    call set_cut_array(tab_protein_unique,geom, Cut2)
    write(*,*) 'Interaction cutoff square, Cut2 ', Cut2

    ! JRom: setting cutoff for HI, as in sdamm
    ! could be done in set_cutoff
    if ( param_force_energy % ihi ) then
        write(*,*) 'Set cutoff for hydrodynamic interaction: 2.*sqrt( Cut2 ) = ', sqrt(Cut2)*2.
        param_force_energy % cutoff_hi = sqrt(Cut2) * 2.
        
        if ( geom % type .eq. enum_box ) then
            write(*,*) 'min_sizebox = ', MINVAL ( geom % pbox % sizebox )
            if ( param_force_energy % cutoff_hi .gt. MINVAL ( geom % pbox % sizebox )/2. ) then
                write(*,*) 'INFO Local volume fraction cutoff too large. Rescaled to 1/2 of the simulation box size.'
                write(*,*) 'INFO Cutoff HI ', MINVAL ( geom % pbox % sizebox ) / 2.
                param_force_energy % cutoff_hi = MINVAL ( geom % pbox % sizebox ) / 2.
            end if
        end if

        write(*,*) 'Final cutoff for hydrodynamic interaction: ', param_force_energy % cutoff_hi
    end if

    if ( param_timestep % swd1 < rhit ) then
        write(*,*)
        write(*,*) 'WARNING Your parameters for the timestep are not adequat '
        write(*,*) 'WARNING We advice to setup swd1 > maximum size of solute 1 + solute 2 ', &
                   ' ( swd1 > ',rhit,')'
        write(*,*)
    end if

    !! case of PrometCS, run are extremly long with the default rhit, make a special case
    !! to set after the test on timestep
    if ( (param_force_energy % imetaldesolv) .and. (param_force_energy % iimg_chg_desolv) ) then
        rhit = rhit * 3.
        write(*,*)
        write(*,*) 'INFO MetalDesolvation and Image Charge elctrostatic desolvation are selected'
        write(*,*) 'INFO You certainly want to use the PrometCS force field'
        write(*,*) 'INFO To avoid very long trajectories, the rhit distance is multiplied by 3 rhit= ', rhit
        write(*,*)
    end if

    !rhit = 93.5
    !write(*,*) 'Force rhit to large value ',rhit

    ! new, check if there will be interaction close to the exit
    ! if no interaction fast move, otherwise computation of force, energy...
    ! write(*,*) 'sqrt(Cut2) C-x1', sqrt(Cut2), param_timestep % optimize_exit
    if ( sqrt(Cut2) > param_timestep % optimize_exit ) then
        write(*,*)
        Write(*,*) 'WARNING Cutoff :', sqrt(Cut2), ',optimize_exit : ', param_timestep % optimize_exit
        write(*,*) 'WARNING Your parameters for the timestep are not adequat '
        write(*,*) 'WARNING Close to the c surface there is still interactions between the solutes  '
        write(*,*) 'WARNING It would be faster if there is no interactions after the distance optimize_exit.'
        write(*,*) 'WARNING Try to increase the c surface.'
        write(*,*)
    endif

    if ( param_analytic % ianalytic ) then
#ifdef DEBUG
        write(*,*) 'call init_analytic from Main loop'
        write(*,*) 'use geometry % c-surf ', geom % stop_position
#endif
        if ( param_analytic % debyeh ) then
            ! sda 2proteins, use the c-surf as limit
            !call init_analytic ( param_force_energy, tab_protein_unique, sqrt(Cut2), param_analytic % ionic )
            call init_analytic ( param_force_energy, tab_protein_unique, &
                                 geom % stop_position, param_analytic % ionic )
        end if
    endif

    ! summary of all forces/energies to compute
    call print_info_param_fe ( param_force_energy )

! new version, set correct relative diffusion before the loop
! update p_sogrid if needed ( no hydrodynamic )
    dm = tab_protein_unique % array ( 1 ) % p_sogrid % diffusion_trans + &
            tab_protein_unique % array ( 2 ) % p_sogrid % diffusion_trans

    write(*,*)
    write(*,*) 'SDA case : the relative translational diffusion is the sum of the absolute diffusion coefficients'
    write(*,*) 'Solute 1 : ', tab_protein_unique % array ( 1 ) % p_sogrid % diffusion_trans
    write(*,*) 'Solute 2 : ', tab_protein_unique % array ( 2 ) % p_sogrid % diffusion_trans
    write(*,*) 'Relative diffusion coeefficient : ', dm
    write(*,*)

    ! in this version fixed, shared variable
    dm30 = dm * 30.d0

    ! send dm is bd_motion or modify diffusion_trans
    ! do not work with OpenMP, because dm is refered in so_grid, so one unqie copy !!
    ! every threads increment the unique value !
    ! solution only the master assign the new value, then only read by the threads, seems ok

    ! @todo check best method, with case of hydrodynamics interaction. delete if unused variable
    ! first version
    ! protein2 % p_sogrid % diffusion_trans = dm
    ! second version if hydrodynamics
    !protein2 % diff_trans = dm
    ! MIC: p_sogrid is used in rescale, need to set up correctly ?
    ! MIC: Ok, not needed, because the sum is recomputed if ihi
    ! so_grid updated correctly

    ! update diffisoin coefficient in p_sogrid only if no hydrodynamic
    ! if false p2 % sogrid contains the relative diffusion coeeficient
    ! if true  p2 % sogrid keeps its original value, relative will be recomputed every time
    if ( param_force_energy % ihi .eqv. .false. ) &
        tab_protein_unique % array ( 2 ) % p_sogrid % diffusion_trans = dm
    ! otherwise keep default and recompute everytime the rescaled relative diffusion coeff.

    write (*,*)
    write (*,*) '========================================================================='
    write (*,*) ' Begin run '
    write (*,*) 'before parrallel nrun', nrun
    flush ( io6 )

!$OMP PARALLEL DEFAULT ( NONE ) &
!$OMP SHARED ( tab_protein_unique, param_force_energy, param_metaldesolv, rcriteria, o_complexes, trajectories, geom, p_restart, &
!$OMP          resid_time, type_calc, param_timestep, param_probe, size_thread_complexe, merge_step_shared, &
!$OMP          nmax_overlap, rhit, Cut2, &
!$OMP          total_test_change_conformation, total_accepted_new_conf, &
!$OMP          rboost, timemax, nrun, run_done, &
!$OMP          global_steps_record, dummy_array, &
!$OMP          dm30, & !dm
!$OMP          ierr, iflex, EXIT_PROG, &
!$OMP          n_repeat_occur, current_occur ) &
!$OMP PRIVATE ( array_energy, array_energy_1d, max_energy_terms, size_array_energy, total_energy ) &
!$OMP PRIVATE (  p_local_complexe, merge_step, thread_nrun ) &
!$OMP PRIVATE ( ptab_protein, protein1, protein2 ) &
!$OMP PRIVATE ( k, current_run, current_conf, current_step, nb_overlap, overlap, hit, time, timehit ) & ! iw
!$OMP PRIVATE ( f, t, t1, dtnow, dist, accept, correct_format_energy, bool_escape, total_overlap, total_overlap_flex ) &
!$OMP PRIVATE ( local_escape, local_visited_windows, local_sumtime, local_fpt, time_per_run ) &
!$OMP PRIVATE ( nstep_all, nboost_all, time_all ) &
!$OMP PRIVATE ( tid ) &
!$OMP PRIVATE ( accepted_new_conf, print_step, clos1 )

!$  tid = omp_get_thread_num()
!! !$  write(*,*) 'tid ', tid

    ! initialize variable after parallel, particular for private ones are all undefined
    ! need to do again ?? undefined from specification
    nullify ( protein1, protein2 )
    nullify ( ptab_protein )
    nullify ( array_energy )
    nullify ( array_energy_1d )
    nullify ( local_escape, local_visited_windows, local_sumtime, local_fpt )
    nullify ( p_local_complexe )

    ! modified only if iflex, needed in add_record
    current_conf = 1
    current_occur = 0

    !assign local value, used only if OpenMP
!$  merge_step = merge_step_shared
!$  thread_nrun = 0

    if ( ( associated ( o_complexes % pcomplexe ) ) .or. ( associated ( trajectories % ptrajectory ) ) ) then

          if ( associated ( o_complexes % pcomplexe ) ) then
              call set_array_energy ( type_calc, o_complexes, size ( tab_protein_unique % array ), &
                                array_energy, max_energy_terms ) !opt_force_large_array
            !! specific to sda_2proteins, need the larger one as intermediate
            size_array_energy = get_nbterms_energy_compatible ( o_complexes, .true. )
            !! need also max_energy_terms, should be in function
            ! max_energy_terms = get_max_bit_energy( o_complexes )
          else
              call set_array_energy ( type_calc, trajectories, size ( tab_protein_unique % array ), &
                                      array_energy, max_energy_terms )
              size_array_energy = get_nbterms_energy_compatible( trajectories, .true. )
          end if

        !write(*,*) 'Allocate array_1d ', size_array_energy
        !write(*,*) 'Allocate energy_1d with size_array_energy ', size_array_energy
        !write(*,*) 'max_energy_terms ', max_energy_terms
        allocate ( array_energy_1d ( size_array_energy, 1 ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In allocation of array_energy size ', size_array_energy
            stop 1
        endif
        array_energy = 0.d0
        array_energy_1d = 0.d0

    else
        !write(*,*)
        !write(*,*) 'Warning: No complexe / no trajectory what happens for array_energy ?? '
        ! not sure thread copies are correctly setup
        ! cannot assing array_energy = 0 if null
        nullify ( array_energy )
        nullify ( array_energy_1d )

    endif

    ! need a fictive allocation ? never called if not o_complexes

    ! initialize local_escape to each thread from the pointer, if association rate et electron transfert
    if ( associated ( rcriteria % p_assoc_rate ) ) then
        ! allocate array visited_windows and escape, should be independent for each thread
        if ( rcriteria % type_reaction == elec_transfer ) then
            allocate ( local_escape ( rcriteria % p_assoc_rate % nb_window, rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocation of local_escape '
                STOP 1
            end if
        else
            allocate ( local_escape( 1, 1 ) )
        endif

        !! for association rate
        if ( rcriteria % type_reaction == assoc_rate ) then
            allocate ( local_visited_windows ( rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocation of local_visited_windows '
                STOP 1
            end if
        else
            allocate ( local_visited_windows ( 1 ) )
        endif

        ! for both assoc_rate and electron_transfert ??
        allocate ( local_sumtime ( rcriteria % p_assoc_rate % nb_window, rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In allocation of local_sumtime '
            STOP 1
        end if

        !! local array for first time passage
        if ( rcriteria % p_assoc_rate % fpt ) then
            allocate ( local_fpt ( rcriteria % p_assoc_rate % nb_window, rcriteria % p_assoc_rate % nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocation of local_fpt'
                STOP 1
            end if
        else
            ! necessary ?
            allocate ( local_fpt ( 1, 1 ) )
        end if
    !! need an allocation for check reaction pair
    else
        allocate ( local_escape ( 1, 1 ) )
        allocate ( local_visited_windows ( 1 ) )
        allocate ( local_sumtime( 1, 1 ) )
        !! necessary ??
        allocate ( local_fpt ( 1, 1 ) )
    end if

    ! init 1 set to 0 if visited
    local_escape = 1.d0
    ! now init -1, set X if visited
    local_visited_windows = -1
    local_sumtime = 0.
    ! default -1. not yet visited
    local_fpt = -1.
    !write(*,*) 'size local_escape ', size ( local_escape )
    !write(*,*) 'size local_sumtime ', size ( local_sumtime )
    !write(*,*) 'size local_fpt ', size ( local_fpt )

    ! local copy for tab_protein, master points to the shared
    ! associate pointer ptab_protein to a copy of initial tab_protein if OpenMP, master points to the original
!$  if ( tid > 0 ) then
!! !$      write (*,*) 'openmp activated make copy_array_protein for tid', tid
!$      call copy_array_protein ( tab_protein_unique, ptab_protein, type_calc, param_metaldesolv )
    !! normal if not OpenMP or master
!$  else
        !write (*,*) 'Associate ptab_protein to tab_protein_unique'
        ptab_protein => tab_protein_unique
!$  end if


    !! all threads, even master have a local copy of the complexe
!$  if ( ( tid >= 0 ) .and. ( associated ( o_complexes % pcomplexe ) ) ) then
!$      call copy_record ( o_complexes, p_local_complexe, .false., size_thread_complexe )

     !! else normal if not openMP
!$  else
    !! associate the global pointer to o_complexe if not openmp
      if ( associated ( o_complexes % pcomplexe ) ) then
          p_local_complexe => o_complexes
      end if
!$  end if

    ! not necessary but convenient, just to refer to solute 1 and 2, for simplicity
    protein1 => ptab_protein % array ( 1 )
    protein2 => ptab_protein % array ( 2 )

    !!!!!!!!!! initialisation of the local counter if openMP !!!!!!!
    nstep_all = 0
    total_overlap = 0
    total_overlap_flex = 0
    nboost_all = 0
    time_all = 0.d0
    time_per_run = 0.d0

    !!!!!!!!! initialisation of diffusion parameters
    ! added here if depends on conformation, should be private in this case !!
    !! \bug solved (to check), dm was used in initt.f, but not used in bd motion (sda6 ??)

    ! update coefficient diffusion for all threads
    ! if false p2 % diff_trans will be the relative diffusion coefficient, previously setup before the omp loop
    ! if true, they keep their initial value, relative will be recomputed every time
    if ( param_force_energy % ihi .eqv. .false. ) then
        protein2 % diff_trans = ptab_protein % array ( 2 ) % p_sogrid % diffusion_trans
    endif

!! !$  write(*,*) 'tid ', tid, ' reachs barrier before loop '
!$OMP BARRIER

! add for tests all threads
#ifdef VERBOSE
!$  write(*,*) 'tid ', tid
    write(*,*) 'dm for protein2 so_grid ', protein2 % p_sogrid % diffusion_trans
    write(*,*) 'dm for protein2 diff_trans ', protein2 % diff_trans
    write(*,*) 'dm30 ', dm30
!$OMP BARRIER
#endif


! problem flush and OpenMP ?? here should be ok
!$  if ( tid == 0 ) then
        flush(io6)
!$  end if

! may test other schema, option dynamic or guided
!$OMP DO SCHEDULE( DYNAMIC )

!! Start of BD runs, do-loop
LOOP_RUN: &
    do current_run = 1, nrun

!$      thread_nrun = thread_nrun + 1
!! !$   write (*,*) 'current_run ', current_run,' tid=',tid

! here could be any thread, normally only one ,problem if concurrent run 100 and 200 for instance, avoided by critical section
#ifdef WEBSDA
        if ( MOD ( current_run, 10 ) == 0 ) then
#else
        if ( MOD ( current_run, 100 ) == 0 ) then
#endif

!$OMP CRITICAL ( write_steps )
            rewind ( io_steps )
            write ( io_steps, fmt='(i10)' ) current_run - 1
            write ( io_steps, fmt='(15i10)' ) ( global_steps_record ( k ), k=1,current_run-1 )
            flush ( io_steps )
!$OMP END CRITICAL ( write_steps )

            !! one thread will update the writting of record, but do not force to be merged before
            if ( associated ( p_local_complexe ) ) then
!$OMP CRITICAL ( crit_merge_p_complexe )
                !write(*,*) 'just before write o_complexes'
                call write_record ( o_complexes )
!$OMP END CRITICAL ( crit_merge_p_complexe )
            end if

        end if

! new version, need to merge the complexe
! only merge the current thread with o_complexes, do not need to block the other threads, merge_step is local
! default fixed to 50 trajectories
!$      if ( associated ( p_local_complexe ) ) then
!$          merge_step = merge_step - 1
!$          if ( merge_step == 0 ) then
!! !$              write(*,*) 'merge thread ', tid

! need critical here, to block other threads
!$OMP CRITICAL ( crit_merge_p_complexe )
!$              call merge_complexe ( o_complexes, p_local_complexe )
!$OMP END CRITICAL ( crit_merge_p_complexe )
                ! reset local merge_step, use an input parameter
!$              merge_step = merge_step_shared
!$          end if
!$      end if

        timehit = 0.d0
        current_step = 0
        time = 0.d0

        !! if sda-koff should check overlap during initialisation
        !! initialize conformation at the beginning of each run
        print_step = .false.
        overlap = .false.
        bool_escape = .false.
        !! default not correct if overlap as soon as the beginning
        !! no, set correctly in init_time
        hit = .false.
        correct_format_energy = .false.

        ! if sda_koff, initial position is given by the restart file
        call init_position_protein ( ptab_protein, type_calc, geom, dist, iflex, current_conf, &
                                     p_restart, current_occur, n_repeat_occur )


#ifdef VERBOSE
write(*,*)
write(*,*) '=============================================='
write(*,*) 'init pos of prot2 after random_pos or initial if sda-koff'
    if ( iflex ) then
        write(*,*) 'conformation ', current_conf
    end if

    !protein2 % position = (/ -10.6, -10.4, -24.0 /)
    !protein2 % position = (/ 0.0, 0.0, 45.0 /)
    !write(*,*) 'pos relat ', protein2 % position - protein1 % position
    !protein2 % orientation_x =(/1.0, 0.0, 0.0/)
    !protein2 % orientation_y =(/0.0, 1.0, 0.0/)
    !protein2 % orientation_z =(/0.0, 0.0, 1.0/)
    write(*,*) 'pos ', protein2 % position
    write(*,*) 'x ', protein2 % orientation_x
    write(*,*) 'y ', protein2 % orientation_y
    write(*,*) 'z ', protein2 % orientation_z

    write(*,*) 'pos atom surf1', protein1 % p_list_atom_surface(:,1), &
            protein1 % p_list_atom_surface( :,protein1 % nb_atom_surface )
    write(*,*) 'pos atom surf2', protein2 % p_list_atom_surface(:,1) , &
            protein2 % p_list_atom_surface( :,protein2 % nb_atom_surface )
#endif
        ! for first_time passage, need a first test, cannot reproduce exactly sda6 runs anymore (??)
        if ( ( rcriteria % type_reaction == docking ) .or. ( rcriteria % type_reaction == assoc_rate ) .or. &
             ( rcriteria % type_reaction == elec_transfer ) ) then

            ! local copy and only read should not be a problem, time = 0. here
            ! dtnow not validated, set to 0 ??
            dtnow = 0.d0
            call check_reaction_pair ( protein1, protein2, rcriteria, local_visited_windows, local_escape, local_sumtime, &
                                       local_fpt, real(time,4), real(dtnow,4), clos1 )
            !write(*,*) 'after check_reaction_pair and current_step ', clos1, current_step
            !write(*,*) 'return etrxncrt'
            !write (*,*) 'resc ', local_escape

        else if ( rcriteria % type_reaction == all ) then
            clos1 = .true.
        ! else off_rc nothing to do, clos always false and not record the complex
        endif
!!!!!!!!!!!!!!! begin BD loop
LOOP_STEP: &
        do

            ! signal handling, stop the job
            if ( EXIT_PROG .eqv. .true. ) then
!! !$              write(*,*) 'thread got the message exit tid', tid
                CYCLE LOOP_RUN ! do the job, because make the loop until the end but not so nice...
            !not nice way to quit
            end if
            
            ! JRom: for hydrodynamic interactions - since the local diff.transl.coeff. changes in each step
            !if ( param_force_energy % ihi ) then
            !    dm = protein1 % diff_trans + protein2 % diff_trans
            !    dm30 = dm * 30.d0
            !end if
            ! MIC: Cannot work, dm and dm30 are SHARED, not PRIVATE
            ! MIC: Many complication for few, dm30 influences only init_time ( far from each other => no rescaling )
            ! MIC: Keep default dm30 in init_time
            
            ! choose timestep, variable in sda_2proteins type
            ! setup hit acording to dist and rhit and param_timestep
            ! update hit with dist, first call ok because setup to start_position
            !if ( param_timestep % variable ) then
            call init_time ( param_timestep, geom, dist, rhit, dm30, dtnow, hit )
            !write(*,*) 'out init_time dist dtnow rhit, hit ', dist, dtnow, rhit, hit
            current_step = current_step + 1
#ifdef VERBOSE
            write(*,*) 'current_step', current_step
            write(*,*) 'dist after init_time ', dist
            write(*,*) 'dist from relative position ',real(sqrt(dot_product(protein2%position,protein2%position)),4)
#endif
            print_step = .false.
            ! need to reset at every steps
            correct_format_energy = .false.

            ! check for printing trajectory, will need to compute energy
            if ( associated ( trajectories % ptrajectory ) ) then
                !write(*,*) 'ptraj associated print_step ', print_step, trajectories % freq_print, trajectories % ntraj_rec, current_run
                ! if ntraj_rec < 0, print all trajectories, otherwise only the required
                if ( ( trajectories % ntraj_rec < 0 ) .or. ( trajectories % ntraj_rec == current_run ) ) then
                    ! if freq_print == 0, print every step, otherwise modulo
                    if ( ( trajectories % freq_print == 0 ) .or. &
                         ( ( trajectories % freq_print /= 0 ) .and.  (MOD ( current_step, trajectories % freq_print ) == 0) ) ) then
                    print_step = .true.
                    end if
                end if
            end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!! comes back here if boost, need to recompute the force
    2   continue

!!!!!!!!!!!!!!!!!!!!!!!!!!! Compute force and torques

! seem to slow down ompp profiler, to comment later
!! $POMP INST BEGIN ( compute_force )
            f =  0.d0
            t =  0.d0
            t1 = 0.d0

#ifdef VERBOSE
            write(*,*)
            write(*,*) '=== Compute force ==='
#endif

            ! fast_computation if interaction are symetric, and no lennard-jones
            if ( param_force_energy % fast_computation ) then
                !write(*,*) 'call force fast version'
                call force_epedhdlj_2proteins_fast ( protein1, protein2,  param_force_energy, dist, &
                                                     f, t, t1, dummy_array )
            ! more general case, deal with asymetry, do not optimise computation
            else
                !write(*,*) 'compute general forces asym / lennard-jones included '
                call force_epedhdlj_2proteins ( protein1, protein2, param_force_energy, dist, &
                                                f, t, t1, dummy_array )
            endif

            ! force is smaller than SDAMM,
            ! should be the sum in sda_2proteins
!#ifdef VERBOSE
            !write(*,*) 'after uhbd forces '
            ! pometcs elctro. desolvation, test with input = 1.67 * 2 = 3.34 works at 70 A
            !write(*,*) 'f ', f
            !write(*,*) 't ', t
            !write(*,*) 't1 ', t1
!#endif

            ! move follow order sda6
            ! test image charge, could outside the loop if image_charge without lennard-jones
            if ( param_force_energy % iimg_chg ) then
                !write(*,*) 'Will compute image_charge'
                ! apply only for protein 2 here
                ! total force are updated inside the function
                call force_image_charge ( ptab_protein, param_force_energy, f, t )

            endif

            ! implemented but strange results, not existing in sda6
            !if ( param_force_energy % imetaldesolv ) then
                 !write(*,*) 'Will compute metal desolvation forces activated'
                 ! seems to be really strong
                 !call compute_force_metaldesolv_2prot( ptab_protein, param_metaldesolv, f,t )
            !end if

            ! Analytical interactions
            if ( param_force_energy % ianalytic ) then

                if ( param_force_energy % idebye ) then
                    ! compute and add the force inside the function
                    call force_analytic_dh ( ptab_protein, param_force_energy, dist, f, t, t1 )
                endif
                ! if others...
            endif

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#ifdef VERBOSE
    write (*,*) 'Total FORCE '
    write (*,'(''Test new force f: '',3E14.5 )') f
    write (*,'(''Test new torque t: '',3E14.5 )') t
    write (*,'(''Test new torque t1: '',3E14.5 )') t1
#endif

!! $POMP INST END ( compute_force )

!!!!!!!!!!!! Make BD move, try to move and rotate, check for overlap,
           !   come back directly to LOOP_OVERLAP if move is rejected

!! $POMP INST BEGIN ( bd_move )

    ! JRom: re-calculating the diffusion coefficient
    if ( param_force_energy % ihi ) then
#ifdef VERBOSE
        !write(*,*) 'ihi activated,call scale diff coeff with cutoff_hi ', param_force_energy % cutoff_hi
        write(*,*)
        write(*,*) '=== Before t_diff_coeff, hydrodynamics part ==='
#endif
        ! rescale diffusion coefficient with Weinstein method
        ! keep cutoff as 4 in memory, like other if later regrouped in one array
        call scale_diffusion_coefficients ( ptab_protein, geom, real(param_force_energy % cutoff_hi,8) )
    end if

LOOP_OVERLAP: &
        do nb_overlap = 1, nmax_overlap+1

            call make_bd_move_2prot ( protein1, protein2, geom, &
                                      f, t, t1, &
                                      dtnow, dist, rhit, &
                                      hit, accept, bool_escape, total_overlap )
            !write(*,*) 'after make_bd_move dist ', dist

            ! finish trajectory if protein  escapes
            ! could finalize and CYCLE BDLOOP to avoid GO TO
            if ( bool_escape ) then
                ! go out of this run / traj
                !write(*,*) 'will escape'
!! $POMP INST ALTEND( bd_move )
                GO TO 333
            endif
            if ( accept .eqv. .true. ) exit
            !if ( accept .eqv. .true. ) then
                !write(*,*) 'move accepted'
                !write(*,*) 'no overlap apply rtry'
            !    exit
            !end if
            !else
                !write(*,*) 'accept false tid ', tid
                !write(*,*) 'tid nb/tot overlap ', tid, nb_overlap, total_overlap
            !end if
            !write (*,*) 'accept false test an other move'
        end do LOOP_OVERLAP

        !! if accept is still false at this time, means nb_overlap == nmax_overlap => call boost
        if ( accept .eqv. .false. ) then
            ! boost check if surface, increase only z
            call boost( protein2 % position, geom, dist, rboost)
            !write (*,*) 'rboost after ', rboost
            !write(*,*) 'after boost nb_overlap ', nb_overlap
            !write (*,*) 'boosts current step / run ', current_step, current_run
            !write(*,*) 'force: ', f
            !write(*,*) 'torque: ', t, t1

            ! here need to update dist, check if a surface is present
            call compute_distance ( protein1, protein2, dist )
            !! test in case of boost, to delete
            !write(*,*) 'new pos relat', protein2 % position
            !write(*,*) ' new dist ', dist

            ! increment boost counter
            nboost_all = nboost_all + 1

            ! recompute forces with the new position, and renter overlap loop
            ! goes back just beforce force calculation, end of overlap and boost
            GO TO 2
        end if

!! $POMP INST END( bd_move )

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        ! update time
        time = time + dtnow
        ! why timehit ?? hit only to check when protein may be excluded
        ! it is a test for not doing exchk because sure to be false
        ! nothing to see with timestep, but implemented like this in original sda
        if ( hit ) then
            timehit = timehit + dtnow
            !write(*,*) 'update timehit ', timehit
            !write(*,*) 'time ', time
        end if

        ! to delete
        !stop


        ! which distance, surface or receptor on membrane ??
        ! for plan no problem, only for box
        ! for dist, default dist. May be a spot in a surface ?? need other parameters
        !write(*,*) 'update residence_time ', dist, dtnow
!! use atomic inside
        call update_residence_time ( resid_time, protein2 % position, dist, dtnow )

        !! just for test, to be independent of timehit of normal sda
        !!write(*,*) 'current_step ', current_step
        !if ( current_step == 20 ) GO TO 333

 !!!!!!!!!!!!!!!!!!!!! Reaction Criteria check
        !! setup clos1 to true if contact is detected ??? could be at the start
        clos1=.false.

        !! no problem of distance here, reaction pairs defined explicitely
        !! may add new type, as a z distance
        if ( ( rcriteria % type_reaction == docking ) .or. ( rcriteria % type_reaction == assoc_rate ) .or. &
             ( rcriteria % type_reaction == elec_transfer ) ) then
!! $POMP INST BEGIN ( check_reaction_pair )

            !local copy and only read should not be a problem
            !call check_reaction_pair_old ( protein1, protein2, rcriteria, local_escape, local_sumtime, local_fpt, &
            !                           real(time,4), real(dtnow,4), clos1 )
            call check_reaction_pair ( protein1, protein2, rcriteria, local_visited_windows, local_escape,  &
                                       local_sumtime, local_fpt, real(time,4), real(dtnow,4), clos1 )
!! $POMP INST END ( check_reaction_pair )

        else if ( rcriteria % type_reaction == all ) then
                clos1 = .true.
        !!else off_rc nothing to do
        endif

        !write(*,*) 'close ', clos1
        ! should check if associated before printing

!!!!!!!!!!!!! no complexes to record AND no trajectories to print  => no energy to compute
! DEBUG to comment for prod !!
        !clos1 = .true.

        if ( ( associated ( p_local_complexe ) .and. ( clos1 ) ) .or. &
             ( print_step .eqv. .true. ) ) then

!!!!!!!!!!!!!!!!!  Compute energy

!! $POMP INST BEGIN ( compute_energy )
#ifdef VERBOSE
            write(*,*)
            write(*,*) '=== Compute energy ==='
#endif

            ! initialize all terms to 0
            array_energy = 0.d0

            ! in sda 2proteins,
            if ( param_force_energy % fast_computation ) then
                !write(*,*) 'call fast_computation energy '
                call energy_epedhdlj_2proteins_fast ( protein1, protein2, param_force_energy, &
                                                      array_energy, dummy_array, dist )
            else
                !write(*,*) 'call asymetric computation energy'
                call energy_epedhdlj_2proteins ( protein1, protein2, param_force_energy, dist, &
                                                 array_energy, dummy_array )
           end if

            ! test image charge, could outside the loop if image_charge without lennard-jones
            if ( param_force_energy % iimg_chg ) then
                !write(*,*) 'compute energy image charge'
                call energy_image_charge ( ptab_protein, param_force_energy, array_energy )
            endif

            ! compute metal desolvation
            if ( param_force_energy % imetaldesolv ) then
                 !write(*,*) 'Will compute metal desolvation energy'
                 !! energy updated inside
                 call compute_energy_metaldesolv_2prot( ptab_protein, param_metaldesolv, array_energy )
            end if

            if ( param_force_energy % ianalytic ) then

                if ( param_force_energy % idebye ) then
                    ! compute and add the energy to the electrostatic energy
                    !write(*,*) 'normal electrostatic energy ', array_energy ( 2 , 1 ), array_energy (2, 2)
                    call energy_analytic_dh ( ptab_protein, param_force_energy, dist, array_energy )
                endif
            end if

            ! same definition than sdamm, now sum up half of potential energy, to use with epfct = 0.5
            total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
            !! put total energy here, could change, in energy of prot2
            array_energy (1,2) = total_energy
            !array_energy (1,1) = 0.d0
            !write(*,*) 'array_energy', array_energy (:,2)

#ifdef VERBOSE
    write(*,*) '       Z  ' , ptab_protein % array (2) % position (3)
    write(*,*) ' total energy ', array_energy ( 1, 2 )
    write(*,*) ' electro. energy', array_energy ( 2 , 1 ), array_energy (2, 2)
    write(*,*) ' desol energy', array_energy( 3, 1 ), array_energy( 3, 2 )
    !write(*,*) ' hydro. energy', array_energy( 4, 1 ), array_energy ( 4 , 2 )
    !write(*,*) ' ljrep energy', array_energy( 5, 1 ), array_energy ( 5 , 2 )
    !write(*,*) ' lj energy', array_energy( 6, 1 ), array_energy ( 6 , 2 )
    !write(*,*) ' img_chg energy', array_energy( 7, 1 ), array_energy ( 7 , 2 )
    !write(*,*) ' metal desol energy', array_energy( 8, 1 ), array_energy( 8, 2 )
#endif
!! $POMP INST END ( compute_energy )


!!!!!!!!!!!!!!!!!  Case of complexes

            if ( ( associated ( p_local_complexe ) ) .and. ( clos1 ) ) then

                !write(*,*) 'current_conf protein1 ', ptab_protein % array ( 1 ) % p_flex % current_conf
                !write(*,*) 'current_conf protein1 ', current_conf

                ! for sda_2proteins need to make the format compatible into 1dimension array
                ! need additional, max_index to check, easy to get only once from complexe ? trajectory ?
                ! seems ok here always convert when a new energy has been computed
                call convert_to_record_format ( array_energy, array_energy_1d, max_energy_terms )
                ! set variable to say it has been done already
                correct_format_energy = .true.

                ! new format for input, add for test the number of entry, not needed
                call add_record ( p_local_complexe, ptab_protein % array ( 2:2 ), current_run, current_step, &
                                  array_energy_1d, real(time,4), current_conf ) !, 1 )

                !write(*,*) 'Print complexe '
                !call print_listcomplexe( p_local_complexe % pcomplexe, .true. )

            end if !record complexe
            !stop 1

!!!!!!!!!!!!! Case of trajectory to write
!            print_step = .true.
            if ( print_step ) then
                !write(*,*) 'add record in trajectories freq_print, ntraj_rec, current_run, current_step ', &
                !  trajectories % freq_print, trajectories % ntraj_rec, current_run, current_step

! only one thread will print the trajectory, no need of critical
! need to check before in read_input_file, not yet done
!$OMP CRITICAL ( crit2 )
!! !$ write(*,*) 'entry critical trajectory tid', tid
!! !$              write(*,*) 'current_run ', current_run
!! !$              write(*,*) 'current_step ', current_step
!! !$              write(*,*) 'array_energy ', array_energy

                ! not sure the format has already been converted,
                ! already done when inserting a complexe but no reset for the next loop
                ! certianly error in trajectories
                if ( correct_format_energy .eqv. .false. ) &
                    call convert_to_record_format ( array_energy, array_energy_1d, max_energy_terms )

                call add_record ( trajectories, ptab_protein % array (2:2 ), current_run, current_step, &
                                  array_energy_1d, real(time,4), current_conf )!, 1 )
!! !$ write(*,*) 'exit critical tid', tid
!$OMP END CRITICAL ( crit2 )

            end if ! print trajectory

        end if ! energy to compute


!!!!!!!!!!!!!!!!!!!!!!!! Test timemax to finish trajectory
        !! exit trajectory if maximum time is exceeded or infinite if 0.
        !! why timehit ?? could be time_all for docking
        !! Should make before test flexibility
        !write(*,*) 'timehit ', timehit

        ! for testing, force only 5 steps
        !if ( current_step > 5. ) then
        if ( (timemax > 0.d0 ) .and. ( timehit > timemax) ) then
            !write(*,*) 'will finish trajectory'
            GO TO 333
        end if

!!!!  Look for changing conformation and update visited_conf
        if ( iflex ) then

            if ( protein1 % iflex ) then
                !!update statistic
                call update_visited_conf ( protein1, real(dtnow,4) )

                !! decrease time flex until 0 or negative
                protein1 % p_flex % time_next_flex = protein1 % p_flex % time_next_flex &
                                                     - real(dtnow,4)
                !write(*,*) 'update time next flex ', protein1 % p_flex % time_next_flex

                if ( protein1 % p_flex % time_next_flex <= 0. ) then
                    !write ( *,*) 'Will make change conf for protein1 '
                    !$ATOMIC
                    total_test_change_conformation = total_test_change_conformation + 1

                    call change_conformation ( protein1, ptab_protein,  &
                                               geom, param_force_energy, param_metaldesolv, type_calc, dummy_array, &
                                               hit, accepted_new_conf, total_overlap_flex, &
                                               energy_epedhdlj_flex  )
                    ! update variable current_conf
                    current_conf = protein1 % p_flex % current_conf
                    !! fisrt developped for sdamm, should be valid here also
                    !! here should not be problem of multiple change, execept prot1 and prot2 flex at the same time
                    if ( accepted_new_conf .eqv. .true. ) then
                        !$ATOMIC
                        total_accepted_new_conf = total_accepted_new_conf + 1
                        call init_flexible_freq ( ptab_protein % array ( 1 ) % p_flex )

                    end if
                end if
            end if

            !write(*,*) 'prot1 not accepted test prot2 '

            !! check for protein2
            if ( protein2 % iflex ) then
                !write(*,*) 'prot2 change conformation '
                call update_visited_conf ( protein2, real(dtnow,4) )
                !write(*,*) 'After update visited conf'
                protein2 % p_flex % time_next_flex = protein2 % p_flex % time_next_flex &
                                                     - real(dtnow,4)
                !write(*,*) 'update p2 time next flex ', protein2 % p_flex % time_next_flex

!! to uncoment for production
                !protein2 % p_flex % time_next_flex = -1.

                if ( protein2 % p_flex % time_next_flex <= 0. ) then
                    !write ( *,*) 'Will make change conf for protein2 '

                    !$ATOMIC
                    total_test_change_conformation = total_test_change_conformation + 1

                    call change_conformation ( protein2, ptab_protein, &
                                               geom, param_force_energy, param_metaldesolv, type_calc, dummy_array, &
                                               hit, accepted_new_conf, total_overlap_flex, &
                                               energy_epedhdlj_flex  )

                    ! update variable current_conf
                    current_conf = protein2 % p_flex % current_conf

                    !! fisrt developped for sdamm, should be valid here also
                    if ( accepted_new_conf .eqv. .true. ) then
                        !write(*,*) 'accept_new_conf is true update'
                        !$ATOMIC
                        total_accepted_new_conf = total_accepted_new_conf + 1
                        call init_flexible_freq ( ptab_protein % array ( 2 ) % p_flex )
                    end if
                end if

                !!just test to delete, to move into change_conformation
                !array_energy = 0.d0
                !call energy_epedhdlj ( ptab_protein, array_energy, &
                !                      dummy_array, Cut2 )
                !total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( 3:20 , 1 ) )
                !array_energy (1,2) = total_energy
                !array_energy (1,1) = 0.d0
                !write(*,*) 'array_energy', array_energy (:,2)
!!!!!!!!!
!    write(*,*) ' ENERGY '
!    write(*,*) ' total energy ', array_energy ( 1, 2 )
!    write(*,*) ' potential energy', array_energy ( 2 , 1 ), array_energy (2, 2)
!    write(*,*) ' desol energy', array_energy( 3, 1 ), array_energy( 3, 2 )
!    write(*,*) ' hydro. energy', array_energy( 4, 1 ), array_energy ( 4 , 2 )
!    write(*,*) ' lj energy', array_energy( 5, 1 ), array_energy ( 5 , 2 )
!!!!!!!!!!!!!!!!!!!!!!!!!!!
                ! end test protein 2
            end if

        end if ! end if iflex

    end do LOOP_STEP

!..here when the trajectory is finished.... timemax exceeded or escape
!..for statistics...
  333   continue

!! !$  write(*,*) 'end of run tid ',tid
    if ( ( rcriteria % type_reaction == assoc_rate ) .or. ( rcriteria % type_reaction == elec_transfer ) ) then

!! $POMP INST BEGIN ( update_rate )
        ! use atomic in the function update_escape, no checkpoint possible
        call update_escape ( rcriteria % p_assoc_rate, current_run, local_visited_windows, local_escape, &
                             local_sumtime, local_fpt )
        ! reinitialize local_escape, could be in update_escape
        local_escape = 1.0d0
        local_visited_windows = -1
        local_sumtime = 0.
        local_fpt = -1.
!! $POMP INST END ( update_rate )
    end if

    ! more tricky here, take the smallest one ? or the biggest one ?
    if ( EXIT_PROG .eqv. .false. ) then
        if ( current_run > run_done ) then
            ! necessary to protect here ? ATOMIC rejected at compilation
            run_done = current_run
        end if
    endif
    ! update step and time
    nstep_all = nstep_all + current_step
    time_all = time_all + time

    ! update global variable, no concurrency possible: each thread is running an unique trajectory
    global_steps_record ( current_run ) = current_step

!!! end loop over nrun
  end do LOOP_RUN

!$OMP END DO

!! merge complexe and visited conf

!$  if ( associated ( p_local_complexe ) ) then
!$OMP CRITICAL ( crit_merge_p_complexe )
!! !$     write (*,*) 'make last merging tid', tid
!$     call merge_complexe  ( o_complexes, p_local_complexe )
!$OMP END CRITICAL ( crit_merge_p_complexe )
!$  end if

!!  if OpenMP need to sum up into tab_protein_unique, this one is the original and not destroyed
!$  if ( tid > 0 ) then
!$       if ( iflex ) then
!$         do i = 1, size ( tab_protein_unique % array )
!$           if ( tab_protein_unique % array ( i ) % iflex ) then

!$OMP CRITICAL ( crit_test )
!$             call print_visited_conf ( ptab_protein % array ( i ) )
!$OMP END CRITICAL ( crit_test )

               !! atomic inside
!$             call merge_visited_conf ( tab_protein_unique % array ( i ), &
!$                                       ptab_protein % array ( i ) )
!$           end if
!$         end do

!$       end if
!$  end if

!!!!!!!!!!!!!!!!!!!!!!!    Print final data, done by all threads !!!!!!!!!!!!!!!!!!!!!!!

!! this can be printed by all threads
  time_per_run = time_all / float ( nrun )

!$OMP CRITICAL ( print_stat )
!$ write (*,*)
!$ write (*,*) 'Stat for thread ',tid

  WRITE (io6,621)  nstep_all, &
!$                 thread_nrun, &
                   time_all, time_per_run, total_overlap, total_overlap_flex, nboost_all

  621 FORMAT (/,'stat o:    total steps these runs:',i12,/, &
!$           'stat o:       total number of run:',i12,/, &
             'stat o:      total simulated time:',f13.0,' ps',/, &
             'stat o:    simulated time per run:',f13.0,' ps',/, &
             'stat o: total overlaps these runs:',i12,/, &
             'stat o:        total overlap flex:',i12,/, &
             'stat o:   total boosts these runs:',3i12,/)
!$OMP END CRITICAL ( print_stat )

!! they belong to each thread, need to be deleted by all
    if ( associated ( local_escape ) ) then
        deallocate ( local_escape, stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR In deallocation of local_escape'
            STOP 1
        end if
        nullify ( local_escape )
    end if

    if ( associated ( local_visited_windows ) ) then
        deallocate ( local_visited_windows, stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR In deallocation of local_visited_windows'
            STOP 1
        end if
        nullify ( local_visited_windows )
    end if

    if ( associated ( local_sumtime ) ) then
        deallocate ( local_sumtime, stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR In deallocation of local_sumtime'
            STOP 1
        end if
        nullify ( local_escape )
    end if

    if ( associated ( local_fpt ) ) then
        deallocate ( local_fpt, stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR In deallocation of local_fpt'
            STOP 1
        end if
        nullify ( local_fpt )
    end if

!!    if array_energy allocated explicitely inside parrallel
    if ( associated ( array_energy ) ) then
        deallocate ( array_energy, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy '
            STOP 1
        end if
        nullify ( array_energy )
    end if

    if ( associated ( array_energy_1d ) ) then
        !write(*,*) 'deallocate array_energy_1d '
        deallocate ( array_energy_1d, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy '
            STOP 1
        end if
        nullify ( array_energy_1d )
    end if

!! delete copy of loacal_tab_protein only for threads > 0
!! if thread == 0 or no OpenMP it will be deleted at the end

!! !$  write (*,*) 'delete local array tid ', tid
nullify ( protein1, protein2 )
!$  if ( tid > 0 ) then
!!      !option true for NOT deleting the grid
!$      call delete_array_protein ( ptab_protein, .true. )
!! now need explicit deallocation of the pointer
!$      deallocate ( ptab_protein,  stat = ierr )
!$      if ( ierr /= 0 ) then
!$          write (*,*) 'ERROR In allocating ptab_dest '
!$          STOP 1
!$      end if
!$      nullify ( ptab_protein )
!$  end if

!! delete p_local_complexe, for all threads
!$    if ( associated ( p_local_complexe ) ) then
!$        call delete_record ( p_local_complexe )

!! now need explicit deallocation of the pointer
!$        deallocate ( p_local_complexe,  stat = ierr )
!$        if ( ierr /= 0 ) then
!$            write (*,*) 'ERROR In allocating p_local_complexe '
!$            STOP 1
!$        end if
!$        nullify ( p_local_complexe )
!$    end if

!! can close the parrallel region, should be implicit barrier at the close
!$OMP END PARALLEL


    ! global array bootstrap
    if ( associated ( rcriteria % p_assoc_rate ) ) then
        if ( rcriteria % p_assoc_rate % bootstrap .eqv. .true. ) then

          ! write output to file only at the end of the simulation
          !call write_boot_file ( rcriteria % p_assoc_rate, nrun )
          call write_boot_file ( rcriteria % p_assoc_rate, run_done )

          ! can suppress the deallocation, no problem because only one copy
          ! delete here because allocated in this function, not in the module
          if ( allocated ( rcriteria % p_assoc_rate % array_bootstrap ) ) then
            deallocate ( rcriteria % p_assoc_rate % array_bootstrap, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocating array_bootstrap'
                STOP 1
            end if
        end if
      
        ! used for electron_transfer
        if ( allocated ( rcriteria % p_assoc_rate % array_bootstrap_real ) ) then
            deallocate ( rcriteria % p_assoc_rate % array_bootstrap_real, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocating array_bootstrap_real'
                STOP 1
            end if
        end if

!       ! test for fpt
        if ( rcriteria % p_assoc_rate % fpt .eqv. .true. ) then
             ! write to file, done before
            deallocate ( rcriteria % p_assoc_rate % array_boot_fpt, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocating array_boot_fpt'
                STOP 1
            end if
        end if
      end if
    end if

!! can be printed outside, tab_protein_unique has been update before end parallell
    if ( iflex ) then
        write ( *,* )
        write ( *, fmt='("stat o:    total test change conf:",i12,/,"stat o:     total accepted change:",i12)' ) &
                 total_test_change_conformation, total_accepted_new_conf

        if ( tab_protein_unique % array ( 1 ) % iflex ) then
            write ( io6, * )
            write ( io6, * ) 'Protein1'
            call print_visited_conf ( tab_protein_unique % array ( 1 ) )
        end if
        if ( tab_protein_unique % array ( 2 ) % iflex ) then
            write ( io6, * ) 'Protein2'
            call print_visited_conf ( tab_protein_unique % array ( 2 ) )
        end if
    end if

    if ( associated ( o_complexes % pcomplexe ) ) then
        !write(*,*) 'final writing of complexes'
        call write_record ( o_complexes )
    end if

!! write the last update of steps.log and close the file
    rewind ( io_steps )
    write ( io_steps, fmt='(i10)' ) nrun
    write ( io_steps, fmt='(15i10)' ) ( global_steps_record ( k ), k=1,nrun )
    close ( io_steps )

    ! delete global array, seems not necessary for valgrind
    deallocate ( global_steps_record, stat = ierr )
    if ( ierr /= 0 ) then
        write ( io6, * ) 'ERROR In deallocation of global_steps_record'
        STOP 1
    end if


    ! print residence time, if resid3d will print distance and plan and write 3ddata to a file
    status = print_residence_time ( resid_time, geom, nrun )
    if ( resid_time % enum_rt == rt_resid3d ) then
        call write_resid_3d ( resid_time, tab_protein_unique % array ( 1 ) % center_protein )
    end if

    ! if needed print the association rate data
    if ( ( rcriteria % type_reaction == assoc_rate ) .or. ( rcriteria % type_reaction == elec_transfer ) ) then

        if ( type_calc %  enum_type_calc == sda_koff ) then
            !call print_table_rate( rcriteria % p_assoc_rate, nrun, .true. )
            call print_table_rate( rcriteria % p_assoc_rate, run_done, .true. )
        else
            !call print_table_rate( rcriteria % p_assoc_rate, nrun, .false. )
            call print_table_rate( rcriteria % p_assoc_rate, run_done, .false. )
        end if
    end if

    !! need to deallocate ?? Not really, no complain from valgrind ??
    !! now is pointer, so yes
    if ( associated ( dummy_array ) ) &
         deallocate ( dummy_array )

    !! now a shared variable
    if ( allocated ( global_steps_record ) ) then
        deallocate ( global_steps_record , stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR In deallocation of global_steps_record'
            STOP 1
        end if
    end if

    !nullify ( p_tmp_onecomplexe )

    write (*,*)
    write (*,*) 'Exit main loop sda 2 proteins'

end subroutine MainLoop_sda_2proteins_omp

!> Make format compatible to call add_record.\n
!!
!! could be in record, and get bit information from there.
!! need a parameter, maximum max_bit_energy to limit the loop of the array.
subroutine convert_to_record_format ( array_energy, array_energy_1d, max_bit_energy )
    IMPLICIT NONE
    real(kind=8), dimension(:,:), intent ( in )  :: array_energy
    real(kind=8), dimension(:,:), intent ( out ) :: array_energy_1d
    integer, intent ( in ) :: max_bit_energy

    integer :: counter_1d, i_1d

    !write(*,*)
    !write(*,*) 'transform array for add_record specification max_bit_energy ', max_bit_energy
    !write(*,*) 'size_array_energy ', size(array_energy)
    !write(*,*) 'size array_energy_1d ' , size ( array_energy_1d )

    array_energy_1d = 0.d0
    ! copy the total energy
    !just convention for total energy, associated to protein 2
    array_energy_1d ( 1, 1 ) = array_energy ( 1, 2 )
    !write(*,*) 'array_energy_1d total ', array_energy_1d ( 1, 1 )

    ! all other terms splited, begin at 2
    counter_1d = 2
    !do i = 1, 8
    !here should be better max number ?? available in param_force_energy + imgchg
    do i_1d = 1, max_bit_energy
        array_energy_1d ( counter_1d, 1 ) = array_energy ( i_1d + 1, 1 )
        array_energy_1d ( counter_1d + 1, 1 ) = array_energy ( i_1d + 1, 2 )
        counter_1d = counter_1d + 2
    end do

!    write(*,*) 'array 1D ', array_energy_1d

end subroutine convert_to_record_format

