!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file MainLoop_sda_energy.f90
!! Loop for sda_energy

!> Main function for sda_energy
!!
!! Notifications for sda_energy:\n
!! This function is the less tested for the moment\n
!! Apply only with 2 proteins
!!
!! Use one complex file for restart ( trajectory restart to write )\n
!! Use the colomn for occurence as a number of restart\n
!! Each line will be started n_occur * nrun\n
!!
!! Flexibility partly implemented ( to check )
!!
!! Output may be a new complex, and/or a new trajectory, as asked from the input
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param o_complexes : instance of \ref mod_record::record "record" for complexes output
!! @param trajectories : instance of \ref mod_record::record "record" for trajectory output
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param param_analytic : structure \ref com_sda::parameter_analytic "parameter_analytic"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param p_restart : pointer to the restart file
subroutine MainLoop_sda_energy ( tab_protein, o_complexes, trajectories, &
                                 param_force_energy, param_analytic, param_metaldesolv, geom, type_calc, p_restart )

    USE mod_array_protein ! include mod_setofgrid, which includes mod_grid
    USE mod_compute_image_charge
    USE mod_compute_metaldesolv
    USE mod_gridtype
    USE mod_geometry
    USE mod_record
    USE mod_force_energy
    USE mod_compute_energy
    USE mod_set_parameters
    USE com_sda
    USE AllTimer
    IMPLICIT NONE

    interface

            subroutine convert_to_record_format ( array_energy, array_energy_1d, max_bit_energy )
                IMPLICIT NONE
                real(kind=8), dimension(:,:), intent ( in ) :: array_energy
                real(kind=8), dimension(:,:), intent ( out ) :: array_energy_1d
                integer, intent ( in ) :: max_bit_energy
            end subroutine convert_to_record_format

    end interface

    type ( array_protein_type ), target :: tab_protein
    type ( record ) :: o_complexes, trajectories
    type ( type_force_energy ) :: param_force_energy
    type ( parameter_analytic ) :: param_analytic
    type ( record ), pointer :: p_restart
    type ( geometry ) :: geom
    type ( type_calculation ) :: type_calc
    type ( parameter_metaldesolv ) :: param_metaldesolv

    real ( kind=8 ), dimension(:,:), pointer:: array_energy
    real ( kind=8 ), dimension( : ), pointer :: dummy_array

    integer :: max_energy_terms, size_array_energy
    real ( kind=8 ), dimension(:,:), pointer :: array_energy_1d
    
    real ( kind=8 ) :: total_energy
    real ( kind=4 ) :: Cut2, dist
    type ( NodeComplexe ), pointer :: tmp_nodecomplexe
    integer :: nb_conf, nconf, ierr
    logical :: iflex

    iflex = .false.
    size_array_energy = 0
    max_energy_terms = 0
    nullify ( tmp_nodecomplexe )
    Cut2 = 0.

    write(*,*) 'Entry MainLoop_sda_energy'

    !! intialize the array, use only protein 1 information
    !call init_force_energy ( param_force_energy, tab_protein_unique, param_metaldesolv % imetaldesolv )
    call init_force_energy ( param_force_energy, tab_protein, type_calc, param_metaldesolv % imetaldesolv )

    !write(*,*) 'associated complexe ', associated ( o_complexes % pcomplexe )
    !write(*,*) 'associated traj ', associated ( trajectories % ptrajectory )

    ! new setup of energy_terms for output
    if ( type_calc % enum_type_calc == sda_energy ) then
        if ( ( associated ( o_complexes % pcomplexe ) ) .or. ( associated ( trajectories % ptrajectory ) ) ) then

            if ( associated ( o_complexes % pcomplexe ) ) then
                call set_array_energy ( type_calc, o_complexes, size ( tab_protein % array ), &
                                        array_energy, max_energy_terms )
                !! specific to sda_2proteins, need the larger one as intermediate
                size_array_energy = get_nbterms_energy_compatible ( o_complexes, .true. )

            else
                call set_array_energy ( type_calc, trajectories, size ( tab_protein % array ), &
                                        array_energy, max_energy_terms )
                size_array_energy = get_nbterms_energy_compatible( trajectories, .true. )
            end if

#ifdef DEBUG
            write(*,*) 'Allocate energy_1d with size_array_energy ', size_array_energy
#endif
            allocate ( array_energy_1d ( size_array_energy, 1 ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'Error in allocation of array_energy size ', size_array_energy
                stop 1
            endif
            array_energy = 0.d0
            array_energy_1d = 0.d0

        !else not associated, not possible here
        else
            write(*,*) 'Error : neither complexes or trajectory file output are defined, will stop'
            stop 1
        end if

        !! new function, set up only shared variable
        call set_dummy_array_and_logical ( tab_protein, dummy_array, iflex )
        !write(*,*) 'iflex ', iflex
        !write(*,*) 'size dummy_array ', size(dummy_array)

        !NJB: Setup cut array
        call set_cut_array (tab_protein, geom, Cut2)
        write(*,*) 'Interaction cutoff square, Cut2 ', Cut2

        ! summary data
        call print_info_param_fe ( param_force_energy )

        ! if sda_energy, check for p_restart
        if ( associated ( p_restart ) ) then

            ! check if complexe or trajectory
            if ( associated ( p_restart % pcomplexe ) ) then
                !! error in nb_complexe !! corrected fine now
                nb_conf = p_restart % pcomplexe % nb_complexe
                !! associate tmp_nodecomplexe to the first node
                tmp_nodecomplexe =>  p_restart % pcomplexe % p_first

                write(*,*) 'sda_energy : ',nb_conf, 'complexes read from restart complexes'
            ! if trajectory
            else if ( associated ( p_restart % ptrajectory ) ) then
                nb_conf = p_restart % ptrajectory % size_array
                write(*,*) 'sda_energy : ', nb_conf, 'complexes read from restart trajectory'
            else
                write (*,*) 'p_restart is not associated, problem'
                stop 1
            end if

        ! else error
        else
            write(*,*) 'ERROR Problem, sda_energy but no p_resart associated'
            stop 1
        end if

        !!!!!!! Main loop

        ! sda_2protein , prot1 fixed, whatever restart complexe or trajectory
        !if ( type_calc % enum_type_calc == sda_energy ) .and. ( p_resart % record_type_calc /= sdamm ) then

        if ( p_restart % record_type_calc % enum_type_calc /= sdamm ) then
            write(*,*) 'not restart from sdamm, so sda_energy prot 1 is fixed, default values are correct '

            write(*,*) 'fix position of solute 1'
            write(*,*) 'pos ', tab_protein % array ( 1 ) %  position
            write(*,*) 'orientx ', tab_protein % array ( 1 ) % orientation_x
            write(*,*) 'orienty ', tab_protein % array ( 1 ) % orientation_y
            write(*,*) 'orientz ', tab_protein % array ( 1 ) % orientation_z
            ! could normalize

            ! in case of protein1, do nothing here
            ! call copy_movable_array ( tab_protein % array ( 1 ) )
        end if

        write(*,*)

        write(*,*) 'Begin loop'
CONFLOOP : &
        do nconf = 1, nb_conf

#ifdef VERBOSE
            write(*,*) 'nconf ',nconf
#endif
        !!!!!!!!!! set up position of protein and grid

        !!! if not sdamm, will read only for protein 2
        !if ( p_restart % record_type_calc % enum_type_calc /= sdamm ) then
        !if ( type_calc % enum_type_calc == sda_energy ) then

            !write(*,*) 'type_calc == sda_energy load pos prot2 nconf=', nconf

            !! test sdamm should be done inside this loop

            !! if a complexe, read from listcomplexe
            if ( associated ( p_restart % pcomplexe ) ) then

                tmp_nodecomplexe => tmp_nodecomplexe % p_next
                !! just a check, coan delete
                if ( associated ( tmp_nodecomplexe ) ) then
                    !write(*,*) 'ok to the next complexe'
                else
                    write(*,*) 'tmp_nodecomplexe NOT associated error'
                    stop 1
                end if

                !write(*,*) 'load data from list_complexe '
                tab_protein % array ( 2 ) %  position = tmp_nodecomplexe % p_data %  position
                tab_protein % array ( 2 ) %  orientation_x = tmp_nodecomplexe % p_data % orient_x
                tab_protein % array ( 2 ) %  orientation_y = tmp_nodecomplexe % p_data % orient_y
                !! should read conformation_nb here

            !! else if a trajectory, read from array_trajectory
            else if ( p_restart % type_record == trajectory_enum ) then
                tab_protein % array ( 2 ) %  position = p_restart % ptrajectory % array_trajectory ( nconf ) % position
                tab_protein % array ( 2 ) %  orientation_x = p_restart % ptrajectory % array_trajectory ( nconf ) % orient_x
                tab_protein % array ( 2 ) %  orientation_y = p_restart % ptrajectory % array_trajectory ( nconf ) % orient_y
                !! shoud read conformation_nb here
            end if

            ! could normalize
            ! all cases, need orient_z by cross poduct
            call cross ( tab_protein % array ( 2 ) % orientation_z, &
                         tab_protein % array ( 2 ) %  orientation_x, tab_protein % array ( 2 ) %  orientation_y )
            ! here should setup conformation as well
#ifdef VERBOSE
            write(*,*) 'read position of prot2'
            write(*,*) 'pos ', tab_protein % array ( 2 ) %  position
            write(*,*) 'orientx ', tab_protein % array ( 2 ) % orientation_x
            write(*,*) 'orienty ', tab_protein % array ( 2 ) % orientation_y
            write(*,*) 'orientz ', tab_protein % array ( 2 ) % orientation_z
#endif
            ! make copy of movable atom, only needed if flexible protein2 or sdamm
            call copy_movable_array ( tab_protein % array ( 2 ) )

        !else
        !    write(*,*) 'ERROR Error in sda_energy, type_calc not known'
        !    stop 1
        !end if

        !!!! compute_distance
        call compute_distance ( tab_protein % array ( 1 ) , tab_protein % array ( 2 ) , dist )
#ifdef VERBOSE
        write(*,*) 'dist', dist
#endif

        !if ( type_calc % enum_type_calc == sda_energy ) then
            !!!!!! Compute energy
            ! initialize all terms to 0
            array_energy = 0.d0
            
            ! not applicable to all cases, missing test fast computation, image charge ( and lennard jones )
            call energy_epedhdlj_2proteins_fast ( tab_protein % array ( 1 ), tab_protein % array ( 2 ), param_force_energy, &
                                                  array_energy, dummy_array, dist )

            ! same definition than sdamm, now sum up half of potential energy, to use with epfct = 0.5
             total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
             !! put total energy here, could change, in energy of prot2
             array_energy (1,2) = total_energy

!!!!!!!!!!!!!!!!!
#ifdef VERBOSE
        write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
        write(*,*) ' ENERGY '
        write(*,*) ' total energy ', array_energy ( 1, 2 )
        write(*,*) ' potential energy', array_energy ( 2 , 1 ), array_energy (2, 2)
        write(*,*) ' desol energy', array_energy( 3, 1 ), array_energy( 3, 2 )
        write(*,*) ' hydro. energy', array_energy( 4, 1 ), array_energy ( 4 , 2 )
        !write(*,*) ' lj energy', array_energy( 5, 1 ), array_energy ( 5 , 2 )
#endif
!!!!!!!!!!!!!!!!!

            !!!!! write result in complexe and trajectory
            !write(*,*) 'Before add record run,step,time,conf',  current_run, current_step, time, current_conf

            ! for sda_2proteins need to make the format compatible into 1dimension array
            ! need additional, max_index to check, easy to get only once from complexe ? trajectory ?
            call convert_to_record_format ( array_energy, array_energy_1d, max_energy_terms )

            !! do not record, but should
            if ( associated ( o_complexes % pcomplexe ) ) then
            ! if ( o_complexes % type_record == complexe_enum ) then

            
                call add_record ( o_complexes, tab_protein % array ( 2:2 ), 1, nconf, &
                                  array_energy_1d, 0., 1 )
            end if

            !! add to trajectory, work without problem
            if ( associated ( trajectories % ptrajectory ) ) then
            !if ( trajectories % type_record == trajectory_enum ) then
                call add_record ( trajectories, tab_protein % array (2:2 ), 1, nconf, &
                                  array_energy_1d, 0., 1 )
            end if
	
        end do CONFLOOP

    else
        write(*,*) 'ERROR Error in sda_energy, type_calc not known'
        stop 1
    end if

    !! NEED to explicitely call write_record, not need for trajectory done when deleted, could do the same for complexe ( but option comput_dispersion or not )
    !! use default value true, recompute dispersion
    call write_record ( o_complexes )

    if ( associated ( array_energy_1d ) ) then
        deallocate ( array_energy_1d, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'Error in deallocating array_energy_1d'
            stop 1
        end if
        nullify ( array_energy_1d )
    end if

    deallocate ( dummy_array )
    if ( associated ( array_energy ) ) then
        deallocate ( array_energy, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'Error in deallocating array_energy'
            stop 1
        end if
        nullify ( array_energy )
    end if

    nullify ( tmp_nodecomplexe )
    write(*,*) 'Exit MainLoop_sda_energy'

end subroutine MainLoop_sda_energy
