!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file alltimer.f90
!! Module to define the timers for time performance

!> Module AllTimer\n
!! Utility, independent of the rest of sda.\n
!! Implemtent 2 methods : cpu time and clock time
MODULE AllTimer
    type mytimer

       !> Clock information
       integer :: clock_max, clock_rate, clock_start, clock_stop
       !> Value at start for cpu clock
       real (kind=8) :: cpu_start, cpu_stop
       !> Delta between start and stop for cpu
       real (kind=8) :: delta_cpu, min_cpu
       !> Delta for clock
       integer :: delta_clock, min_clock
       !> count the total number of call for normalization
       integer :: counter

   end type mytimer

   contains

      !> Initialize the timer, combine the 2 methods.\n
      !! mandatory before starting
      !! @param this : instance of \ref mytimer
      subroutine init_mytimer(this)
         IMPLICIT NONE
         type ( mytimer ) :: this

         this % delta_clock = 0
         ! an initial large number
         this % min_clock = 1000000
         this % delta_cpu = 0.d0
         ! an initial large number
         this % min_cpu = 10E10
         this % counter = 0

      end subroutine init_mytimer

      !!> Start the timer.
      !! @param this : instance of \ref mytimer
      subroutine start_mytimer(this)
         IMPLICIT NONE
         type ( mytimer ) :: this

         ! initialize
         call system_clock ( this % clock_start, this % clock_rate, this % clock_max )
         call cpu_time ( this % cpu_start )

      end subroutine start_mytimer

      !!> Stop the timer and increment the total time and the counter.
      !! @param this : instance of \ref mytimer
      subroutine stop_mytimer(this)
         IMPLICIT NONE
         type ( mytimer ) :: this

         call system_clock ( this%clock_stop, this%clock_rate, this%clock_max )
         call cpu_time ( this%cpu_stop )

         ! update minimum
         if ( ( this % clock_stop - this % clock_start ) < this % min_clock ) &
            this % min_clock = this % clock_stop - this % clock_start
         if ( ( this % cpu_stop - this % cpu_start ) < this % min_cpu ) &
            this % min_cpu = this % cpu_stop - this % cpu_start

         ! increment delta
         this % delta_clock = this % delta_clock + ( this%clock_stop - this%clock_start )
         this % delta_cpu = this % delta_cpu + ( this%cpu_stop - this%cpu_start )

         ! increment counter
         this % counter = this % counter + 1

     end subroutine stop_mytimer

     !> Print information on standard output
     !! @param this : instance of \ref mytimer
     subroutine print_mytimer(this)
         IMPLICIT NONE
         type ( mytimer ) :: this
         real ( kind=4 ) :: dclock, dcpu_time

         integer :: hour, min, sec, msec

         dclock = real( this % delta_clock ) / real( this % clock_rate )
         dcpu_time = real ( this % delta_cpu, 4 )

         hour = 0
         min = 0
         sec = 0
         msec = 0

         if ( this % counter > 0 ) then
            write(*,*) '===================================='
            write(*,*) 'clock time :', dclock, ' sec count ', this % counter, ' each ', dclock/ this % counter

            hour = floor(dclock/3600.)
            min =  modulo( floor(dclock/60.), 60)
            sec = floor ( dclock - hour * 3600 - min * 60 )
            msec = int ( ( dclock - floor (dclock )) * 1000 )
            write(*,*) '      soit :', hour, 'h ', min, 'min. and ', &
                             sec, 'sec and ', msec, 'ms'
            ! add minimum
            write(*,*) 'min.clock time: ', real ( this % min_clock ) / real ( this % clock_rate ), ' sec'

            write(*,*)'cpu time (F95 library):', dcpu_time, ' sec, each: ', dcpu_time/ this%counter
            hour = floor(dcpu_time/3600.)
            min =  modulo( floor(dcpu_time/60.), 60)
            sec = floor ( dcpu_time - hour * 3600 - min * 60 )
            msec = int ( ( dcpu_time - floor (dcpu_time )) * 1000 )
            write(*,*)'        soit:', hour, 'h ', min, 'min. and ', &
                             sec, 'sec and ', msec, 'ms'
            ! add minimum
            write(*,*) 'min.cpu time: ', real( this % min_cpu, 4 ), ' sec'
        else
            write(*,*) 'counter 0 '
        endif

    end subroutine print_mytimer

    !> Return the clock time.\n
    !! Only used in tes_module programs
    !! @param this : instance of \ref mytimer
    !! @return dclock : the actual time clock
    function get_time(this) result (dclock)
        implicit none
        type ( mytimer ) :: this
        real(kind=4) :: dclock

        dclock = real( this % delta_clock ) / real( this % clock_rate )

    end function get_time

    !> Test function, gives information about the system_clock implementation (system dependent)
    !! @param this : instance of \ref mytimer
    subroutine system_clock_info(this)
        IMPLICIT NONE
        type ( mytimer ) :: this

        real :: tmax

        call system_clock ( this%clock_start, this%clock_rate, this%clock_max )
        write ( *, * ) ' '
        write ( *, * ) '  SYSTEM_CLOCK with INTEGER arguments reports:'
        write ( *, * ) ' '
        write ( *, * ) '    The current clock count is    ', this%clock_start
        write ( *, * ) '    The clock count per second is ', this%clock_rate
        write ( *, * ) '    The maximum clock count is    ', this%clock_max
        write ( *, '(a)' ) ' '
        write ( *, '(a)' ) '  The minimum measurable time:'
        write ( *, '(a)' ) ' '
        write ( *, * ) 1.0E00 / ( this%clock_rate ), ' seconds'
        write ( *, '(a)' ) ' '
        write ( *, '(a)' ) '  The maximum measurable time:'
        write ( *, '(a)' ) ' '
        tmax = real ( this%clock_max ) / real ( this%clock_rate )
        write ( *, * ) '    Seconds: ', tmax
        tmax = tmax / 60.0E+00
        write ( *, * ) '  = Minutes: ', tmax
        tmax = tmax / 60.0E+00
        write ( *, * ) '  = Hours:   ', tmax
        tmax = tmax / 24.0E+00
        write ( *, * ) '  = Days:    ', tmax

    end subroutine system_clock_info

END MODULE AllTimer
