!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file check_exclusion.f90
!! Check if protein 2 overlaps with protein 1\n
!!
!! Only for sda_2proteins and related. Sdamm implements a repulsive lennard-jones interaction


!> Check exclusion with dynamic exclusion grid\n
!! Only check protein 2 in exclusion grid of protein 1, default in sda\n
!! As soon a first overlap is found, we exit from the function\n
!! The new position and new orientation are tested to check if there is an overlap.\n
!! New position are not applied to the solute in this function, because a serie of transformation will be applied at the end.\n
!! Maybe not the best method
!! @param prot1 : solute 1
!! @param prot2 : solute 2
!! @param rtry : new position of the protein to check
!! @param rot_try : new orientation of the protein to check
!! @param overlap : true if the move is rejected
!! @param do_rot : indicates if the rotation need to be applied
subroutine check_exclusion ( prot1, prot2, rtry, rot_try, overlap, do_rot )

    USE mod_protein
    IMPLICIT NONE

    type ( protein ), intent ( in ), target :: prot1, prot2
    real ( kind=8 ), dimension (3), intent(in) :: rtry
    real ( kind=8  ), dimension (3,3), intent ( in ) :: rot_try
    logical , intent ( out ) :: overlap
    logical, intent (in) :: do_rot

    real ( kind=8 ), dimension(:,:), pointer :: p_lat_surf
    type ( type_grid ), pointer :: ptgrid_excl

    real ( kind = 8 ), dimension ( 3 ) :: shift_excl
    real ( kind=8 ), dimension(3) :: tmp_pos
    real ( kind=8 ) :: x1, x2, x3, hexcl
    integer :: i, i1,i2,i3, nacc

    !write(*,*) 'Entry check_exclusion'
    !write(*,*) 'rtry ', rtry

    !! assign pointer to the exclusion grid and local variables
    ptgrid_excl => prot1 % p_sogrid % p_excl_grid % pgrid
    hexcl = ptgrid_excl % h
    shift_excl = ptgrid_excl % origin - hexcl / 2.d0
    !write(*,*) 'shift_excl ', shift_excl

    !! assign pointer and variable for protein 2
    p_lat_surf => prot2 % p_list_atom_surface
    nacc = prot2 % nb_atom_surface
    !write(*,*) 'nacc ',nacc
    !write(*,*) 'size ', ptgrid_excl % size_x, ptgrid_excl % size_y, ptgrid_excl % size_z

    overlap = .false.

ATOMS: &
    do i = 1, nacc
        !write(*,*) 'nat do rot', i, do_rot

!! version  1.10974943E-04 -g
        if ( do_rot ) then
!   write(*,*) 'p_lat ',  p_lat_surf(1,i)
            tmp_pos(1) = rot_try(1,1) * p_lat_surf(1,i) + &
                         rot_try(1,2) * p_lat_surf(2,i) + &
                         rot_try(1,3) * p_lat_surf(3,i)
            tmp_pos(2) = rot_try(2,1) * p_lat_surf(1,i) + &
                         rot_try(2,2) * p_lat_surf(2,i) + &
                         rot_try(2,3) * p_lat_surf(3,i)
            tmp_pos(3) = rot_try(3,1) * p_lat_surf(1,i) + &
                         rot_try(3,2) * p_lat_surf(2,i) + &
                         rot_try(3,3) * p_lat_surf(3,i)
        else
            tmp_pos(1) = p_lat_surf(1,i)
            tmp_pos(2) = p_lat_surf(2,i)
            tmp_pos(3) = p_lat_surf(3,i)
        endif

        x1 = rtry(1) + tmp_pos(1) - shift_excl ( 1 )
        x2 = rtry(2) + tmp_pos(2) - shift_excl ( 2 )
        x3 = rtry(3) + tmp_pos(3) - shift_excl ( 3 )
        !write(*,*) 'x1 x2 x3 ',x1,x2,x3

        i1 = int ( x1/hexcl )
        i2 = int ( x2/hexcl )
        i3 = int ( x3/hexcl )
        !write(*,*) 'i ', i1, i2, i3
        if ( ( i1  < 1 ) .or. ( i1 > ptgrid_excl % size_x ) ) CYCLE ATOMS
        if ( ( i2  < 1 ) .or. ( i2 > ptgrid_excl % size_y ) ) CYCLE ATOMS
        if ( ( i3  < 1 ) .or. ( i3 > ptgrid_excl % size_z ) ) CYCLE ATOMS

        !! here need test or 2 function
        !if ( associated ( ptgrid_excl % p_grid ) ) then
        ! should stop before if not associated
            if ( ptgrid_excl % p_grid % pgrid_logical ( i1, i2, i3 ) .eqv. .true. ) then
                !write(*,*) 'overlap true at i', i, i1, i2, i3
                overlap = .true.
                return
            end if
        !else
            !write(*,*) 'to write for dtgrid'
            ! or function
            !call get_values_excl ( ptgrid_excl % p_grid % p_dtgrid_logical ( i1, i2, i3, value )
        !end if
    end do ATOMS

end subroutine check_exclusion
