!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file eschk.f90
!! check escape of solute 2.\n

!> Check if protein 2 escapes.\n
!! It escapes at the c-surface or on z crteria if box.\n
!! Called after the BD move.
!! @param geom : instance of \ref mod_geometry
!! @param dtnow : instantaneous value of the timestep
!! @param disto4 : distance before the BD move
!! @param dist4 : distance to test
!! @param dm : translational diffusion coeff.
!! @param escape : return if escape or not
subroutine eschk ( geom, dtnow, disto4, dist4, dm, escape )
    USE mod_geometry
    IMPLICIT NONE
    type ( geometry ), intent ( in ) :: geom
    real ( kind=8 ), intent ( in ) :: dtnow, dm
    real ( kind=4 ), intent ( in ) :: disto4, dist4
    logical, intent ( out ) :: escape

    real ( kind=8 ) :: disp, p0, p1, w, unif
    real ( kind=8 ) :: dist_escape, disto, dist

    escape = .false.
    !! use all variable as real 8, exponential
    disto = disto4
    dist = dist4

    !! by default if box
    !! can use z_max, relative to the center of the system, cm of protein1, not exctly the distance to the surface
    !! if send as parameter avoid repetitive test...
    if ( geom % type == enum_box ) then
        dist_escape = geom % pbox % z_max
        !write(*,*) 'dist_escape box', dist_escape
    else
        dist_escape = geom % pshere % c_surface
        !write(*,*) 'dist_escape sphere ', dist_escape
    end if

    !write(*,*) 'dist_escape ', dist_escape

! outside c-surface = immediate escape
! this test could be done by the calling function ??
    if (dist > dist_escape ) then
       escape = .true.
       !write(*,*) 'already outside'
       return
    end if
! still inside - decide if will escape after corrections
! this is necessary for the case if the protein is very close to the surface
! in this case the dtnow is close to zero and no movement of the protein is done
! so that it would never cross the c-surface and the trajectory would be endless
    disp = dm*dtnow
    p0 = dexp(- (dist-disto)**2/disp)
    !p1 = dexp(- (c-disto+c-dist)**2/disp)
    p1 = dexp ( -( dist_escape - disto + dist_escape - dist )**2 / disp )
    w = (p0-p1)/p0
    !call ggubs(1,unif)
    call ggubs_thread_safe( unif )

    if ( unif > w ) escape = .true.

end subroutine eschk
