!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file init_time.f90
!! Define the timestep ( variable for sda_2proteins )

!> Define instantaneous timestep if it is variable ( only for sda_2proteins )\n
!!
!! Depends on position and geometry\n
!! Reduce time step near outer boundary\n
!! Small difference with sda 6 implementation (value of dthet)\n
!!
!! @todo full check of dthet modifications with sda6
!!
!! @param param_timestep : structure of timestep parameters
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param dist : distance between solutes
!! @param rhit : distance where overlap is possible
!! @param dm30 : diff. coeff. * 30
!! @param dtnow : return timestep
!! @param hit :  return if overlap is possible at this distance
subroutine init_time ( param_timestep, geom, dist, rhit, dm30, &
                       dtnow, hit )

    USE com_sda
    USE mod_geometry
    IMPLICIT NONE
    type ( parameter_timestep ), intent ( in ) :: param_timestep
    type ( geometry ), intent ( in ) :: geom
    real ( kind=8 ), intent ( in ) :: dm30
    real ( kind=8 ), intent ( out ) :: dtnow
    real ( kind=4 ), intent( in ) :: rhit
    real ( kind=4 ), intent( in ) :: dist
    logical, intent ( out ) :: hit

    real ( kind=8 ) :: dt1, swd1, dt2f
    real ( kind=8 ) :: dthmin, dthet, dtf, dto, dtt
    real ( kind=8 ) :: c

    dthmin = 0.d0
    dthet = 0.d0
    dtt = 0.d0
    dto = 0.d0
    c = 0.d0

#ifdef VERBOSE
    write(*,*) "entry init_time dist ", dist
    write(*,*) 'dm30 ', dm30
    write(*,*) 'rhit ', rhit
#endif

    dt1 = param_timestep % dt1
    swd1 = param_timestep % swd1
    dt2f = param_timestep % dt2f
    c= geom % stop_position

    hit = .false.
    if ( dist < rhit ) hit = .true.

! SHN choice: dtr is 3 degree + (dist-rhit) but 90 degree outside rswd
! used only for dteth variable
    dthmin = 3.

    ! linear interpolation of dt1, works when positions are closed
    dtf = dt2f * (dist-swd1) + dt1
    IF ( dist < swd1 ) dtf = dt1

    !write(*,*) "hit ",hit
    !write(*,*) "dtf ",dtf
    !write(*,*) "dthet ",dthet
    !write(*,*) "dthmin ",dthmin

    ! means dist < rhit
    if ( hit ) then
        dthet = dthmin
        !dthet = 3.
        !write(*,*) "dthet < rhit",dthet
    else

    ! slight difference with the original in sda6 (where rswd is taken into account)
    ! rhit very different for aubs, makes a large difference for dthet and for the timestep
    !IF ( dist.GT.rhit) THEN
#ifdef VERBOSE
        write(*,*) 'dist < rswd and > rhit dthet=', (dist-rhit) + dthmin
        !write(*,*) "dthet > rhit",dthet
#endif
        dthet = (dist-rhit) + dthmin
        !GO TO 98
    endif
    ! default value of rswd ?? to check
    ! never called !!! if hit go to 98, if not go to 98
!    write(*,*) 'never called'
    !dthet = 90.
 !  98 CONTINUE

    ! parabolic close to c surface
    dto = (c-dist)**2/dm30 ! decrease with the distance

    ! keep for sda6 comparison, but do not get this variable
    ! parabolic, very strange, could delete ?
    ! to comment later
    ! dthet = 3. ! force like if hit was called, for aubs comaprison
    dtt = 3.9*dthet**2 ! increase with dist, but smaller than dtf if dthet = dthmin = 3
#ifdef VERBOSE
    write(*,*) 'dtt dtf dto', dtt, dtf, dto
#endif

    ! apply the linear interpolation, they are close, MIN(dtt,dtf) MIN(dtf,dt0)= dtt
    IF (dtt.GT.dtf) dtt = dtf
    ! if dto small, close to escape c, dtt is large
    IF (dtt.GT.dto) THEN
    ! Min ( ddt,dto )
        dtt = dto
    END IF
    dtnow = dtt

    ! to delete
    !! do not unserstand this, should not use rotime, it is a counter ??
    !! maybe apply rotation when a treshold is passed
    !rotime = rotime + dtnow
    ! to test if used with variable tworot
    !IF (rotime.GT.dtr) rotate = .true.
    !RETURN

#ifdef VERBOSE
   write(*,*) 'dthet ', dthet
   write(*,*) 'return dist / dtnow ', dist, dtnow
#endif

end subroutine init_time

!> Compute basic property about the maximum timestep...
!! @param geom : instance of \ref mod_geometry
!! @param param_timestep : structure for the timestep varaibles
!! @param tab_protein : intance of \ref mod_array_protein
subroutine check_timestep( geom, param_timestep, tab_protein )
    USE mod_array_protein
    USE mod_geometry
    USE com_sda
    IMPLICIT NONE

    type ( geometry ) :: geom
    type ( parameter_timestep ) :: param_timestep
    type ( array_protein_type ) :: tab_protein

    real ( kind=8 ) :: dm30
    ! a,b of ax+b, C is c_surface
    real ( kind=8 ) :: dt1, dt2, s1, s2, det, sqdet, a, b, C
    ! solutions, always 2 real, x1, y1 to keep
    real ( kind=8 ) :: x1, x2, y1, y2, rmsnow

    !write(*,*) "entry check_timestep "
    dm30 = ( tab_protein % array ( 1 ) % p_sogrid % diffusion_trans + &
             tab_protein % array ( 2 ) % p_sogrid % diffusion_trans ) * 30.

    !write(*,*) "dm30 ", dm30
    !write(*,*) "C ", geom % stop_position
    !write(*,*) "C zmax ", geom % pbox % z_max
    !write(*,*) "dt1 dt2 ", param_timestep % dt1, param_timestep % dt2
    !write(*,*) "s1 s2 ", param_timestep % swd1, param_timestep % swd2
    !write(*,*) "a=dt2f ", param_timestep % dt2f

    dt1 = param_timestep % dt1
    dt2 = param_timestep % dt2
    s1 = param_timestep % swd1
    s2 = param_timestep % swd2
    a = param_timestep % dt2f
    C = geom % stop_position

    ! solve s1 = dt1*a + b
    b = ( dt1*s2 - dt2*s1 ) / ( s2 - s1 )

    !write(*,*) "b ", b
    !write(*,*) "check b ", a*s1 + b

    ! solve
    ! y = ax+b
    ! y = (c-x)^2 / dm30
    ! <=> x^2 - ( 2c + dm a ) x + ( c^2 - dm b ) = 0
    ! determinant
    det = dm30 * ( 4.d0 * C * a + dm30 * a**2 + 4.d0 * b ) !* dm30 )
    !write(*,*) "2 sol, always positive det ", det
    sqdet = sqrt( det )
    !write(*,*) "square det ", sqdet

    x1 =  ( ( 2.d0*c+dm30*a ) - sqdet ) / 2.
    y1 = a * x1 + b
    !write(*,*) "solution 1 ", x1, y1
    !write(*,*) "check ", ( C - x1 )**2 / dm30

    x2 =  ( ( 2.*c+dm30*a ) + sqdet ) / 2.
    y2 = a * x2 + b
    !write(*,*) "solution 2 ", x2, y2
    !write(*,*) "check ", ( C - x2 )**2 / dm30
    !write(*,*) "Value maximum of the timestep ", y1
    !write(*,*) "dm30 / 30 ", dm30 / 30.d0

    ! compute max distance possible, take 95 % proproba. stddev 1.68
    ! rmsnow
    rmsnow = sqrt ( 2.d0 * dm30 / 30.d0 * y1 )
    !write(*,*) "rmsnow ", rmsnow
    !write(*,*) "max step distance ", 1.d0 * rmsnow
    !write(*,*) "C - max ", C - x1

    write(*, fmt='("maximum timestep:    ",f8.3," ps",/, &
    &        "at the distance:           ",f8.3," A",/, &
    &        "maximum step (m):          ",f8.3," A",/, &
    &        "dist. from c surface(d):   ",f8.3," A",/, &
    &        "nb of std.deviation (d/m): ",f8.3,/ &
    &         )') y1, x1, rmsnow,C-x1, (C-x1)/rmsnow

    ! 2 standard deviation, not so confident
    if ( ( ( C-x1 ) / rmsnow ) .le. 3.d0 ) then
        Write(*,*)
        write(*,*) 'WARNING Your setup of the timestep is dangerous, decrease dt2 and/or increase swd2'
        write(*,*)
    end if

    ! always positive, timestep begin to decrease exponentialy after x1
    param_timestep % optimize_exit = x1

end subroutine check_timestep
