!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!>\mainpage Developer Documentation
!> This site contains documentation for sda developers. In case you search for user documentation
!> please look in the doc/doc_sda7/ directory in your distribution or at <a href="http://mcm.h-its.org/sda7/doc/doc_sda7/index.html">SDA 7 User docummentation</a>
!<

!> \file main.f90
!! Main program

!> Program main
!!
!! Read input\n
!! Call a main subroutine \n
!! Delete object and exits\n

program main

    USE mod_array_protein ! include mod_protein which includes mod_setofgrid, which includes mod_grid...
    USE mod_gridtype
    USE mod_ReactionCriteria
    USE mod_geometry
    USE mod_record
    USE mod_ResidenceTime
    USE mod_type_calc
    USE mod_force_energy
    USE com_sda
    USE AllTimer

!$  USE omp_lib
!$  USE Par_Zig_mod

    IMPLICIT NONE
    external warning_sigusr2 ! ! Must declare as external ?
    external warning_sigterm
    external warning_sigint
    interface

          subroutine read_input_file ( tab_protein, &
             rcriteria, resid_time, o_complexes, trajectories, geom, p_restart, &
             mtype_calc, param_timestep, param_probe, param_force_energy, param_analytic, &
             param_metaldesolv, param_pmf, option_omp, timemax, rboost, nrun, nmax_overlap, &
             account_occurency )
            USE mod_array_protein
            USE mod_ReactionCriteria
            USE mod_rate_calculation
            USE mod_type_calc
            USE mod_geometry
            USE mod_record
            USE mod_input_parameter
            USE mod_interface
            USE mod_flexibility
            USE mod_ResidenceTime
            USE mod_force_energy
            USE com_sda
            type ( array_protein_type ) :: tab_protein
            type ( react_criter ), target :: rcriteria
            type ( residence_time ) :: resid_time
            type ( record ) :: o_complexes, trajectories
            type ( record ), pointer :: p_restart
            type ( geometry ) :: geom
            type ( type_calculation ) :: mtype_calc
            type ( parameter_timestep ) :: param_timestep
            type ( probe_type ) :: param_probe
            type ( type_force_energy ) :: param_force_energy
            type ( parameter_analytic ) :: param_analytic
            type ( parameter_metaldesolv ) :: param_metaldesolv
            type ( parameter_pmf ) :: param_pmf
            type ( type_option_omp ) :: option_omp
            real ( kind = 4 ) :: rboost
            real ( kind=8 ), intent ( out ) :: timemax
            integer :: nrun, nmax_overlap
            logical :: account_occurency

          end subroutine read_input_file


          ! used only for testing, use AllTimer for many functions which are run by different threads in omp version
          subroutine MainLoop_sda_2proteins ( tab_protein, rcriteria, &
                   o_complexes, trajectories, geom, p_restart, resid_time, &
                   mtype_calc, param_timestep, param_probe, param_force_energy, &
                   param_analytic, param_metaldesolv, nmax_overlap, &
                   rboost, timemax, nrun )
                  USE mod_array_protein
                  USE mod_gridtype
                  USE mod_ReactionCriteria
                  USE mod_geometry
                  USE mod_record
                  USE mod_ResidenceTime
                  USE mod_force_energy
                  USE AllTimer
                  USE com_sda
                  IMPLICIT NONE
                  type ( array_protein_type ), target :: tab_protein
                  type ( react_criter ) :: rcriteria
                  type ( record ) :: o_complexes, trajectories
                  type ( record ), pointer :: p_restart
                  type ( geometry ) :: geom
                  type ( residence_time ) :: resid_time
                  type ( type_calculation ) :: mtype_calc
                  type ( parameter_timestep ) :: param_timestep
                  type ( probe_type ) :: param_probe
                  type ( type_force_energy ) :: param_force_energy
                  type ( parameter_analytic ) :: param_analytic
                  type ( parameter_metaldesolv ) :: param_metaldesolv
                  real (kind=8) :: timemax
                  real ( kind=4 ) :: rboost
                  integer :: nrun, nmax_overlap
          end subroutine MainLoop_sda_2proteins

          subroutine MainLoop_sda_2proteins_omp ( tab_protein, rcriteria, &
                   o_complexes, trajectories, geom, p_restart, resid_time, &
                   mtype_calc, param_timestep, param_probe, param_force_energy, param_analytic, param_metaldesolv, &
                   option_omp, nmax_overlap, rboost, timemax, nrun )
                  USE mod_array_protein
                  USE mod_gridtype
                  USE mod_ReactionCriteria
                  USE mod_geometry
                  USE mod_record
                  USE mod_ResidenceTime
                  USE mod_force_energy
                  USE AllTimer
                  USE com_sda
                  IMPLICIT NONE
                  type ( array_protein_type ), target :: tab_protein
                  type ( react_criter ) :: rcriteria
                  type ( record ), target :: o_complexes, trajectories
                  type ( geometry ) :: geom
                  type ( record ), pointer :: p_restart
                  type ( residence_time ) :: resid_time
                  type ( type_calculation ) :: mtype_calc
                  type ( parameter_timestep ) :: param_timestep
                  type ( probe_type ) :: param_probe
                  type ( type_force_energy ) :: param_force_energy
                  type ( parameter_analytic ) :: param_analytic
                  type ( parameter_metaldesolv ) :: param_metaldesolv
                  type ( type_option_omp ) :: option_omp
                  real ( kind=8 ) :: timemax
                  real ( kind=4 ) :: rboost
                  integer :: nrun, nmax_overlap
          end subroutine MainLoop_sda_2proteins_omp

          subroutine MainLoop_sdamm ( tab_protein, rcriteria, o_complexes, trajectories, geom, resid_time, &
                 mtype_calc, param_timestep, param_force_energy, param_analytic, param_metaldesolv, timemax )
                USE mod_array_protein
                USE mod_ReactionCriteria
                USE mod_geometry
                USE mod_record
                USE mod_ResidenceTime
                USE mod_force_energy
                USE AllTimer
                USE com_sda
                IMPLICIT NONE
                type ( array_protein_type ) :: tab_protein
                type ( react_criter ) :: rcriteria
                type ( record ) :: o_complexes, trajectories
                type ( geometry ) :: geom
                type ( residence_time ) :: resid_time
                type ( type_calculation ) :: mtype_calc
                type ( parameter_timestep ) :: param_timestep
                type ( type_force_energy ) :: param_force_energy
                type ( parameter_analytic ) :: param_analytic
                type ( parameter_metaldesolv ) :: param_metaldesolv
                real ( kind=8 ) :: timemax
          end subroutine MainLoop_sdamm

          subroutine MainLoop_sda_energy ( tab_protein, o_complexes, trajectories, param_force_energy, &
                                           param_analytic, param_metaldesolv, geom, mtype_calc, p_restart )
                USE mod_array_protein ! include mod_setofgrid, which includes mod_grid
                USE mod_gridtype
                USE mod_geometry
                USE mod_record
                USE mod_force_energy
                USE mod_compute_energy
                USE mod_set_parameters
                USE com_sda
                USE AllTimer
                type ( array_protein_type ), target :: tab_protein
                type ( record ) :: o_complexes, trajectories
                type ( type_force_energy ) :: param_force_energy
                type ( parameter_analytic ) :: param_analytic
                type ( record ), pointer :: p_restart
                type ( geometry ) :: geom
                type ( type_calculation ) :: mtype_calc
                type ( parameter_metaldesolv ) :: param_metaldesolv
          end subroutine MainLoop_sda_energy

          subroutine MainLoop_sda_pmf ( tab_protein, o_complexes, trajectories, &
                                 param_force_energy, param_analytic, param_metaldesolv, param_pmf, geom, type_calc )

                USE com_sda
                USE mod_set_parameters
                USE mod_gridtype
                USE mod_array_protein
                USE mod_force_energy
                USE mod_compute_energy
                USE mod_compute_image_charge
                USE mod_compute_metaldesolv
                USE mod_record
                USE mod_geometry
                USE AllTimer
                type ( array_protein_type ), target :: tab_protein
                type ( record ) :: o_complexes, trajectories
                type ( type_force_energy ) :: param_force_energy
                type ( parameter_analytic ) :: param_analytic
                type ( geometry ) :: geom
                type ( type_calculation ) :: type_calc
                type ( parameter_metaldesolv ) :: param_metaldesolv
                type ( parameter_pmf ) :: param_pmf
            end subroutine MainLoop_sda_pmf

    end interface

    ! timer for testing speed
    type ( mytimer ) :: timer_total, timer_prep

    type ( array_protein_type ) :: tab_protein_unique

    !! all related to reaction criteria, will create rate_calculation object if needed
    type ( react_criter ) :: rcriteria
    !! for recording residence time, distance and surface in the same module
    !! could be in geometry or other ?? could be used for rdf, resid_time 1D, 2D...
    type ( residence_time ) :: resid_time
    !! for geometry
    type ( geometry ) :: geom
    !! for record : complexes and trajectories
    type ( record ) :: o_complexes, trajectories
    !! if use restart for sda_koff, sda_energy. for sdamm update coordinates in read_input_file
    type ( record ), pointer :: p_restart

    ! type_calculation: sda_2proteins, sda_koff, sdamm,...
    type ( type_calculation ) :: mtype_calc
    ! simple type defined in sda_com
    type ( parameter_timestep ) :: param_timestep
    type ( probe_type ) :: param_probe
    type ( type_force_energy ) :: param_force_energy
    type ( parameter_analytic ) :: param_analytic
    type ( parameter_metaldesolv ) :: param_metaldesolv
    type ( parameter_pmf ) :: param_pmf
    ! general structure to group all omp option, only complexe for now
    type ( type_option_omp ) :: option_omp

    real ( kind=8 ) :: timemax
    real ( kind=4 ) :: rboost
    integer :: nrun, nmax_overlap
    logical :: account_occurency

    ! new arguments
    integer :: nb_argument, count_arg, io_err
    character*128 :: input_filename, str_arg
    logical :: ex
    integer :: dummy_rng

    ! signal handling, will stop gracefully the run
    integer, parameter :: SIGUSR2 = 12 , SIGTERM = 15, SIGINT = 2 !SIGUSR1 = 10

    !! init some, could be here or in read input
    nullify ( p_restart )
    resid_time % enum_rt = rt_off

    ! for argument
    nb_argument = 0
    count_arg = 0
    input_filename = ''
    str_arg = ''
    ex = .false.

    !! init type calc for false parrallel or other
    mtype_calc % bool_par = .false.
    mtype_calc % nb_par = 1
    !! this code does not exist, like this sure to be read before testing
    mtype_calc % enum_type_calc = 0

    dummy_rng = 0

    ! record signal, send end of main.f90, SIGINT 2, SIGUSR 12 for cluster
    ! add sigclear, to modify default behaviour
    call sigclear (SIGUSR2)
    call signal(SIGUSR2, warning_sigusr2)
    call sigclear (SIGINT)
    call signal(SIGINT, warning_sigint)
    call sigclear (SIGTERM)
    call signal(SIGTERM, warning_sigterm)

    call print_logo ()
    !flush(io6)

!! Only for profiling with ompP
!$POMP INST INIT


    !!Test OpenMP parameters
!$  write(*,*) 'OpenMP activated'

!! give available on hte machine, but not used here, only thread used
!$  write(*,*) 'nb processor on this machine ', OMP_get_num_procs()

!! give only the thread at this point
!$  write(*,*) 'nb of threads ', OMP_get_num_threads()
!$  write(*,*) 'max nb of threads ', OMP_get_max_threads()
!$  write(*,*) 'dynamic activated ', OMP_get_dynamic()

!! not sure what dynamics, but slow down a lot
!! !$  call OMP_SET_DYNAMIC( .true. )
!!  Required to get 'pmf_com' subroutine in 'Mainloop_sda_pmf.f90' work
!$  call OMP_SET_NESTED ( .true. )
!$  write(*,*) 'is dynamic activated ', OMP_get_dynamic()
!$  write(*,*) 'is nested activated ', OMP_get_nested()

#ifdef DEBUG
    write(*,*)
    write(*,*) 'Compiled with DEBUG flag'
#endif
#ifdef VERBOSE
    write(*,*)
    write(*,*) 'Compiled with VERBOSE flag'
#endif


    !! need to init before using it
    call init_mytimer ( timer_total )
    call init_mytimer ( timer_prep )
    !! start timers
    call start_mytimer ( timer_total )
    call start_mytimer ( timer_prep )


    !! Treat argument
    nb_argument = command_argument_count()
!    nb_argument = iargc ()
    !write(*,*) 'nb_argument ', nb_argument

    if ( nb_argument == 0 ) then
        write(*,*)
        write(*,*) 'help, NO MORE REDIRECTION WITH "<", classical argument sda_flex sda.in'
        write(*,*) 'sda_flex sda.input [-fparrallel nb(1..X)]'
        write(*,*)
        write(*,*) '-fparallel X: used when 1 ( and only 1! ) solute has a list of conformations '
        write(*,*) '               will run trajectories with the Xth conformation, without flexibility '
        write(*,*) '               can be run in a loop, all outputs will be in the same directory with an additional suffix'
        write(*,*) '               can be used with a sge job array, ( but not on magny ) '
        write(*,*) '               independent of OpenMP '
        write(*,*)
        write(*,*) 'No input file will stop'
        flush(6)
        stop 1
    end if

    call getarg ( 1, input_filename )
    write(*,*) 'read input file ',trim ( input_filename )

     !write(*,*) 'length of input file ', len(trim(input_filename))
    if ( (trim(input_filename) == "") .or. (len(trim(input_filename))== 0) ) then
        write(*,*) "No input file will stop "
        flush(io6)
        stop 1
    end if

    inquire ( file=input_filename, exist=ex )
    if ( .not.ex ) then
        write (*,*) "input file does not exist: ", trim ( input_filename )
        flush(io6)
        stop 1
    end if

    !! open the file, the new io is inputf in new_sda_com
    open ( inputf, file=input_filename, form='formatted', action = 'read', status = 'old', iostat = io_err )
    if ( io_err /= 0 ) then
        write (*,*) 'error in opening input file, but exists ??'
        stop 1
    end if

    !! read optional argument
    count_arg = 2
    do while ( count_arg <= nb_argument )
        call getarg ( count_arg, str_arg )
        !write(*,*) 'str_agr ', str_arg

        if ( str_arg(1:11) == '-fparrallel' ) then
            write(*,*) 'The spelling has been corrected, run -fparallel instead'
            stop 1
        else if ( str_arg(1:10) == '-fparallel' ) then
            mtype_calc % bool_par = .true.
            call getarg( count_arg +1, str_arg )
            read ( str_arg, fmt='(i4)' ) mtype_calc % nb_par
            !write(*,*) 'read number ', mtype_calc % nb_par
            count_arg = count_arg + 1

        else
            write(*,*) 'argument not known ', trim ( str_arg )
            stop 1
        endif

        count_arg = count_arg + 1

    end do

    !!move initialization of random generator before read input. Can be used in initialisation of flexible

    !! need to read the input file before, in order to know the type of calculation
    !! initialize all grids, proteins...  when reading input
    call read_input_file ( tab_protein_unique, rcriteria, resid_time, &
      o_complexes, trajectories, geom, p_restart, &
      mtype_calc, param_timestep, param_probe, param_force_energy, param_analytic, param_metaldesolv, &
      param_pmf, option_omp, timemax, rboost, nrun, nmax_overlap, account_occurency )

    !write(*,*) 'close input file'
    close ( inputf )

    !write(*,*)
    !write(*,*) 'after read input file '
    !write(*,*) 'param_force_energy ', param_force_energy % epsilon_scale, param_force_energy % iimg_chg_desolv
    !write(*,*) 'param_metaldesolv %  an_integer ', param_metaldesolv %  an_integer
    !write(*,*) 'param_metaldesolv %  a_real ', param_metaldesolv %  a_real
    !write(*,*)

    write (*,*)
    write (*,*) 'Type of calculation ', mtype_calc % type_calc_str, ':', mtype_calc % enum_type_calc
    write (*,*) 'Type of calculation false parrallel ', mtype_calc % bool_par, 'nb_par ', mtype_calc % nb_par

    call stop_mytimer ( timer_prep )


    !! will be the same loop for sda_koff, just the init position is different
    !! is sda_koff needed, better to set rate_calculation to dissassociation and use restart file ??
    if ( ( mtype_calc % enum_type_calc == sda_2proteins ) .or. ( mtype_calc % enum_type_calc == sda_koff ) ) then

        ! particular because assymetry: exclusion grid, rotation of protein 2 only
        ! position always refer to position_relative

! OMP version
! second version, parrallel region inside the function, global steps to delete for both version
! seems faster from initial tests
! works both with or without openmp
      call MainLoop_sda_2proteins_omp ( tab_protein_unique, &
               rcriteria, o_complexes, trajectories, geom, p_restart, resid_time, &
               mtype_calc, param_timestep, param_probe, param_force_energy, param_analytic, param_metaldesolv, option_omp, &
               nmax_overlap, rboost, timemax, nrun )

! Keep only for testing the performance of each function, should always be updated to the last version !!
! Usefull for debugging and optimisation ( timing better without OMP )
!        call MainLoop_sda_2proteins ( tab_protein_unique, &
!                 rcriteria, o_complexes, trajectories, geom, p_restart, resid_time, &
!                 mtype_calc, param_timestep, param_probe, param_force_energy, param_analytic, param_metaldesolv, nmax_overlap, &
!                 rboost, timemax, nrun )

    ! multiple proteins : 3 or more. position are absolute here
    else if ( mtype_calc % enum_type_calc == sdamm ) then

        call MainLoop_sdamm ( tab_protein_unique, rcriteria, o_complexes, trajectories, geom, resid_time, &
                              mtype_calc, param_timestep, param_force_energy, param_analytic, param_metaldesolv, timemax )

    ! No dynamics, read position from restart file and compute energy ( force ?? )
    ! PMF is a loop over predefined position and orientation
    ! Could be extended to multiple proteins ??

    else if ( mtype_calc % enum_type_calc == sda_energy ) then
        call MainLoop_sda_energy ( tab_protein_unique, o_complexes, trajectories, param_force_energy, &
                                   param_analytic, param_metaldesolv, geom, mtype_calc, p_restart )

    else if ( mtype_calc % enum_type_calc == sda_pmf ) then
        call MainLoop_sda_pmf ( tab_protein_unique, o_complexes, trajectories, param_force_energy, &
                                param_analytic, param_metaldesolv, param_pmf, geom, mtype_calc )

    end if


    !! now delete all
    !write(*,*) 'delete all'

    !! for random generator
!$  call delete_par_zig ( dummy_rng ) !0

    !! NEEDED explictely for valgrind memcheck
    !! need to deallocate derived type
    call delete_geometry ( geom )
    call delete_resid_time ( resid_time )

    write(*,*) 'INFO All your SDA runs/trajectories finished successfully.'

    flush ( io6 )

    !! if sda_energy, sda_koff, p_restart has been allocated in read_input_file
    if ( associated ( p_restart ) ) then
        !write(*,*) 'p_restart associated delete it'
        !! true, do not want to save data in memory before deleting
        call delete_record ( p_restart, .true. )
        deallocate ( p_restart )
        nullify ( p_restart )
    end if

    !! only one function for all deletion of the array and the elements
    !write(*,*) 'will delete complexe'
    call delete_record ( o_complexes )
    call delete_record ( trajectories )
    call delete_react_criter ( rcriteria )

    !! delete all proteins and associated sogrid and flexible
    call delete_array_protein ( tab_protein_unique )

    call stop_mytimer ( timer_total )
    write(*,*) 'Mytimer total'
    call print_mytimer (timer_total)
    write(*,*) 'Mytimer prep'
    call print_mytimer (timer_prep)

end program

!> Print the logo of sda_flex
subroutine print_logo ()
    IMPLICIT NONE

    write (*,*)
    write ( *, fmt = '( " Entry sda_flex" )' )
    write (*,*)

end subroutine print_logo

subroutine warning_sigusr2
    USE com_sda
    EXIT_PROG = .true.
    write(*,*) 'exit prog ', EXIT_PROG
    write(*,*) 'Process interrupted by (SIGUSR2 12), exit gracefully...'
    write(*,*) 'WARNING We did not finish all your SDA runs/trajectories, but the output data from the performed ones are saved.'
    flush (io6)
end subroutine warning_sigusr2

subroutine warning_sigint
    USE com_sda
    !flush (io6)
    EXIT_PROG = .true.
    write(*,*) 'exit prog ', EXIT_PROG
    write(*,*) 'Process interrupted by (SIGINT 2), exit gracefully...'
    write(*,*) 'WARNING We did not finish all your SDA runs/trajectories, but the output data from the performed ones are saved.'
    flush (io6)
end subroutine warning_sigint

subroutine warning_sigterm
    USE com_sda
    EXIT_PROG = .true.
    write(*,*) 'exit prog ', EXIT_PROG
    write(*,*) 'Process interrupted by (SIGTERM 15), exit gracefully...'
    write(*,*) 'WARNING We did not finish all your SDA runs/trajectories, but the output data from the performed ones are saved.'
    flush (io6)
end subroutine warning_sigterm

