!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file make_exclusion_grid.f90
!!
!! Compute the exclusion grid, only for sda_2proteins and solute 1

! trick in original code, hexcl already defined in sdacommon
! here use the common excl1v, change the name in dexcl and
! create new hdexcl which is the old hexcl divided / 2
! should be 0.25 ?? seems !!


!>
!! Use intermediate module \ref mod_exclusion_grid "mod_exclusion_grid"
!!
!! The grid is stored in a \ref mod_setofgrid::sogrid "sogrid" object, like all properties belonging to the protein conformation\n
!! It is stored as an UHBD grid. It can be written to file and visualised\n
!! Use probep the protein probe.
!!
!! @param sgrid : instance of  \ref mod_setofgrid::sogrid "sogrid"
!! @param tmp_atom_pos : position of the atom
!! @param tmp_nat : total number of atoms
!! @param center : center of the solute ( used as input here ?? )
!! @param atom_vdw : list of Van der Walls parameters for all atoms
!! @param max_vdw : maximum value
!! @param probep : protein probe
!! @param hexclusion : size of cells for the exclusion grid
subroutine make_exclusion_grid ( sgrid, tmp_atom_pos, tmp_nat, center, atom_vdw, max_vdw, probep, hexclusion )

    USE com_sda
    ! include mod_exclusion_grid
    USE mod_setofgrid

    IMPLICIT NONE
    type ( sogrid ) :: sgrid
    real ( kind=8 ), dimension(:) :: atom_vdw
    real ( kind=8 ), dimension ( 3 ) :: center
    real ( kind=8 ) :: max_vdw
    real ( kind=4 ) :: probep, hexclusion
    !! original use all atoms
    real ( kind=8 ), dimension ( :, : ) :: tmp_atom_pos
    integer :: tmp_nat

    real ( kind=8 ) :: rexmax, hdexcl, rexcl, rexcl2
    real ( kind=8 ), dimension ( 3 ) :: xv, xa
    real ( kind=8 ) :: oex, oey, oez

    integer, dimension ( 3 ) :: cell_3dim
    !! integer, cell min / max
    integer :: xmin, xmax, ymin, ymax, zmin, zmax
    integer :: ii, jj, kk, i, j, k
    real ( kind=8 ) :: zk, ztemp, yj, ytemp, xi, xtemp, dist2
    !! final size nxex = xmax - xmin + 2
    integer :: nxex, nyex, nzex
    !! klim, account for extension of atom radius over different grid points
    integer :: klim, iat, ierr, nb_true_excl

    xmin = 0
    xmax = 0
    ymin = 0
    ymax = 0
    zmin = 0
    zmax = 0
    nb_true_excl = 0

    rexmax = max_vdw + probep
    hdexcl = hexclusion / 2.d0

    klim = int ( ( rexmax + hexclusion * 1.5d0 )/ hexclusion )

    !write(*,*) 'klim ', klim
    !write(*,*) 'nb _atoms ', tmp_nat
    !write(*,*) 'hexclusion, hdexcl, klim, rexmax ',hexclusion, hdexcl,klim,rexmax
    !write(*,*) 'center ', center

    !write(*,*)
    write(*, fmt='("Generate exclusion grid with hexclusion=", f8.3, " porobep= ", f8.3)' ) hexclusion, probep
!    write(*, fmt='("Generate exclusion grid with hexclusion=", f8.3, ) hexclusion, probep
    !! First, need to find the size of the grid
    !! Loop over atoms
ATOMS:do iat = 1, tmp_nat

        !! recenter
        xv = tmp_atom_pos ( :, iat ) - center

        !! find home grid cell, ih, jh, kh, cell_3dim
        !! here center to zero, truncate 1.8 = 1
        cell_3dim = int ( xv / hexclusion )

        !write (*,*) 'cell_3dim ', cell_3dim
        if ( cell_3dim ( 1 ) < xmin ) xmin = cell_3dim ( 1 )
        if ( cell_3dim ( 1 ) > xmax ) xmax = cell_3dim ( 1 )
        if ( cell_3dim ( 2 ) < ymin ) ymin = cell_3dim ( 2 )
        if ( cell_3dim ( 2 ) > ymax ) ymax = cell_3dim ( 2 )
        if ( cell_3dim ( 3 ) < zmin ) zmin = cell_3dim ( 3 )
        if ( cell_3dim ( 3 ) > zmax ) zmax = cell_3dim ( 3 )

    end do ATOMS

    !write (*,*) '1 min-max value ', xmin, xmax, ymin, ymax, zmin, zmax
    ! which value to add ?? add maximum value
    xmin = xmin - klim
    xmax = xmax + klim
    ymin = ymin - klim
    ymax = ymax + klim
    zmin = zmin - klim
    zmax = zmax + klim
#ifdef DEBUG
    write (*,*) ' min-max grid points ', xmin, xmax, ymin, ymax, zmin, zmax
#endif

    ! + 1 account for 0,
    ! + 1 account for only positive number now and want to avoid 0 the origin of the grid
    nxex = xmax - xmin + 2
    nyex = ymax - ymin + 2
    nzex = zmax - zmin + 2
#ifdef DEBUG
    write(*,*) ' grid size: ', nxex, nyex, nzex
#endif

    ! could be an array ( 3 ) // -0.5 fixed or should be hexcl ??
    oex = hexclusion * ( xmin - 0.5 ) - hexclusion !work for reproducing log file k=53
    oey = hexclusion * ( ymin - 0.5 ) - hexclusion ! -hexclusion needed for avoiding i = 0
    oez = hexclusion * ( zmin - 0.5 ) - hexclusion
#ifdef DEBUG
    write (*,fmt='("origin: ",3f8.3)' ) oex, oey, oez
#endif
    ! allocate grid, nxex
    allocate ( sgrid % p_excl_grid, stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot allocate pointer on excl grid'
        stop
    else
        call allocate_excl_grid ( sgrid % p_excl_grid, nxex, nyex, nzex )
        ! only compute in uhbd at this point
        sgrid % p_excl_grid % pgrid % p_grid % pgrid_logical = .false.
        ! set up the origin and hexcl
        sgrid % p_excl_grid % pgrid % origin = (/ oex, oey, oez /)
        sgrid % p_excl_grid % pgrid % h = hexclusion
    end if

    !! make 2nde loop like original version
    !write(*,*) 'before loop2 tmp_nat', tmp_nat, ' size(atom_vdw) ', size(atom_vdw)

ATOMS2: &
    do iat = 1, tmp_nat

        !! recenter
        !write(*,*) 'iat ', iat, xv, atom_vdw(iat)
        xv = tmp_atom_pos ( :, iat ) - center
        xa = xv + hdexcl

        rexcl = atom_vdw ( iat ) + probep
        rexcl2 = rexcl**2

        !! find home grid cell, ih, jh, kh, cell_3dim
        !! now shift from origin, should be positive
        cell_3dim ( 1 )  = int ( (xa(1)-oex) / hexclusion )
        cell_3dim ( 2 )  = int ( (xa(2)-oey) / hexclusion )
        cell_3dim ( 3 )  = int ( (xa(3)-oez) / hexclusion )
        !write (*,*) '2 cell_3dim ', cell_3dim

LOOPZ: &
        do kk = -klim, klim
            k = kk + cell_3dim ( 3 )
            !write(*,*) 'kk ',kk

            !! should not happen, can delete when sure
            if ( k <= 0 .or. k > nzex ) then
                write(*,*) ' k out of range', k
            end if
! from original code
!..here and further we correct the shifting by hdexcl:
!       ZK=dfloat(K)*dexcl-hdexcl+oez
!..here is a concerted change with change in exchk.f
            zk = dfloat ( k ) * hexclusion + oez
            ztemp = zk - xv ( 3 )

            !write(*,*) 'k zk, ztemp ', k, zk, ztemp
            if ( ztemp > rexcl ) CYCLE LOOPZ

     LOOPY: do jj = -klim, klim
                j = jj + cell_3dim ( 2 )
                ! should not happen
                if ( j <= 0 .or. j > nyex ) then
                    write(*,*) ' j out of range', j
                end if

                yj = dfloat ( j ) * hexclusion + oey
                ytemp = yj - xv ( 2 )

                !write(*,*) 'j, yj, ytemp ', j, yj, ytemp
                if ( ytemp > rexcl ) CYCLE LOOPY

         LOOPX: do ii = -klim,klim
                    i = ii + cell_3dim ( 1 )
                    ! should not happen
                    if ( i <= 0 .or. i > nxex ) then
                        write(*,*) ' i out of range', i
                    end if

                    if ( sgrid % p_excl_grid % pgrid % p_grid % pgrid_logical ( i, j, k ) .eqv. .true. ) then
                        !write(*,*) 'exclusion already 1 ', i, j, k
                        CYCLE LOOPX
                    end if
                    xi = dfloat ( i ) * hexclusion + oex
                    xtemp = xi - xv ( 1 )
                    !write(*,*) 'i, xi, xtemp ', i, xi, xtemp

                    dist2 = xtemp*xtemp + ytemp*ytemp + ztemp*ztemp
                    if ( dist2 > rexcl2 ) then
                        CYCLE LOOPX
                    end if
                    !!if here assign true
                    !write(*,*) 'assign true',i,j,k, nb_true_excl + 1
                    sgrid % p_excl_grid % pgrid % p_grid % pgrid_logical ( i, j, k ) = .true.
                    nb_true_excl = nb_true_excl + 1

                end do LOOPX
            end do LOOPY
        end do LOOPZ

    end do ATOMS2

#ifdef DEBUG
    write(*,602) tmp_nat, nb_true_excl
    write(*,*)

602 FORMAT ('no of atoms used for prot 1 excl grid:',i10,/, &
            'no of points assigned value 1        :',i20)
#endif

    !! print grid, should be apart, can have a flag, do not want all grid when flexible
    ! call print_excl_grid_log ( sgrid % p_excl_grid, iopoti )

end subroutine make_exclusion_grid
