!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file maths.f
!! group general mathematical functions



      REAL*8 FUNCTION fmax3(x1,x2,x3)
c maximal among 3 real*8
      IMPLICIT REAL*8 (a-h,o-z)

      fm = x1
      IF (fm.LT.x2) fm = x2
      IF (fm.LT.x3) fm = x3
      fmax3 = fm
      RETURN
      END

      subroutine cross(ab,a,b)
c cross-product of 2 vectors
      IMPLICIT NONE
      real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
      real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b

      ab(1) = a(2)*b(3) - a(3)*b(2)
      ab(2) = a(3)*b(1) - a(1)*b(3)
      ab(3) = a(1)*b(2) - a(2)*b(1)

      end subroutine cross


      subroutine cross4 (ab,a,b)
      IMPLICIT NONE
      !real ( kind=4 ), dimension( 3 ) :: ab, a, b
      real ( kind=4 ), dimension ( 3 ), intent ( out ) :: ab
      real ( kind=4 ), dimension ( 3 ), intent ( in ) :: a, b

      ab(1) = a(2)*b(3) - a(3)*b(2)
      ab(2) = a(3)*b(1) - a(1)*b(3)
      ab(3) = a(1)*b(2) - a(2)*b(1)
      RETURN
      end subroutine cross4

      subroutine matrix ( om, rot )
c converts rotation angle to rotation matrix
      IMPLICIT NONE
      real ( kind=8 ), dimension ( 3 ), intent ( in ) :: om
      real ( kind=8 ), dimension ( 3,3 ), intent ( out ) :: rot
      !IMPLICIT REAL*8 (a-h,o-z)
      !DIMENSION rot(3,3),om(3)
      real ( kind=8 ) :: c1, c2, c3, s1, s2, s3

      c1 = dcos(om(1))
      s1 = dsin(om(1))
      c2 = dcos(om(2))
      s2 = dsin(om(2))
      c3 = dcos(om(3))
      s3 = dsin(om(3))
      rot(1,1) = c2*c3
      rot(1,2) = s1*s2*c3 - c1*s3
      rot(1,3) = s1*s3 + c1*s2*c3
      rot(2,1) = c2*s3
      rot(2,2) = c1*c3 + s1*s2*s3
      rot(2,3) = c1*s2*s3 - s1*c3
      rot(3,1) = -s2
      rot(3,2) = s1*c2
      rot(3,3) = c1*c2
      !RETURN
      end subroutine matrix

      subroutine tr(vo,vn,ex,ey,ez)
c given vo() returns vn() - coordinates of vo() in
c a new coordinate system defined by ex,ey,ez axes
      IMPLICIT NONE
      real ( kind=8 ), dimension ( 3 ), intent ( out ) :: vn
      real( kind=8 ),dimension( 3 ), intent ( in ) :: vo,ex,ey,ez

      !gfortran faster with -O2
      vn(1) = dot_product(vo,ex)
      vn(2) = dot_product(vo,ey)
      vn(3) = dot_product(vo,ez)
      !CALL dot(vn(1),vo,ex)
      !CALL dot(vn(2),vo,ey)
      !CALL dot(vn(3),vo,ez)
      !RETURN
      end subroutine tr

      !! input and output are real 4
      subroutine tr4(vo,vn,ex,ey,ez)
c given vo() returns vn() - coordinates of vo() in
c a new coordinate system defined by ex,ey,ez axes
      !IMPLICIT REAL*8 (a-h,o-z)
      IMPLICIT NONE
      !DIMENSION vo(3),vn(3),ex(3),ey(3),ez(3)
      real ( kind = 8 ), dimension ( 3 ) :: ex, ey, ez
      real ( kind = 4 ), dimension ( 3 ) :: vo, vn

      real ( kind = 4 ), dimension ( 3 ) :: ex4, ey4, ez4
      ex4 = real (ex,4)
      ey4 = real (ey,4)
      ez4 = real (ez,4)

      !gfortran faster with -O2
      vn(1)=dot_product(vo,ex4)
      vn(2)=dot_product(vo,ey4)
      vn(3)=dot_product(vo,ez4)

      !RETURN
      end subroutine tr4

      subroutine tr_vector ( vo, vn, ex, ey, ez, nat )
        IMPLICIT NONE
        real ( kind = 8 ), dimension ( 3,nat ), intent ( in ) :: vo
        real ( kind = 8 ), dimension ( 3,nat ), intent ( out) :: vn
        real ( kind = 8 ), dimension ( 3 ), intent ( in ) :: ex, ey, ez
        integer, intent ( in ) :: nat

        integer :: i

        do i = 1, nat
            vn( 1, i ) = dot_product ( vo(:,i), ex )
            vn( 2, i ) = dot_product ( vo(:,i), ey )
            vn( 3, i ) = dot_product ( vo(:,i), ez )
        end do
      end subroutine tr_vector


c     scalar product of 2 vectors
      SUBROUTINE dot(ab,a,b)

      implicit none
      real(kind=8), DIMENSION(3), intent(in) :: a,b
      real(kind=8), intent(out) ::ab

      ab = a(1)*b(1) + a(2)*b(2) + a(3)*b(3)
c      RETURN
      END

cc add subroutine for multiplication, 3x3 with a vector(3)
cc make multiplication c = rot * b
      subroutine rotate31( c, rot, b)
      IMPLICIT NONE
      real( kind=8 ), dimension(3,3), intent(in) :: rot
      real (kind=8 ), dimension(3), intent(in) :: b
      real (kind=8 ), dimension(3), intent(out) :: c

      c(1) = rot(1,1)*b(1) + rot(1,2)*b(2) + rot(1,3)*b(3)
      c(2) = rot(2,1)*b(1) + rot(2,2)*b(2) + rot(2,3)*b(3)
      c(3) = rot(3,1)*b(1) + rot(3,2)*b(2) + rot(3,3)*b(3)

      end subroutine rotate31

      !! update b
      !!function rotate31( rot, b )


      subroutine rotate3n ( c, rot, b, n )
      IMPLICIT NONE
      real( kind=8 ), dimension(3,3), intent(in) :: rot
      real (kind=8 ), dimension(3,n), intent(in) :: b
      real (kind=8 ), dimension(3,n), intent(out) :: c

      integer :: i,n

      do i = 1,n
      !c(:,i) = rotate31(
      ! why slighlty faster ??
      call rotate31 ( c(:,i), rot, b(:,i) )
      ! seem slightly slower for one short run ??
      !c(1,i) = rot(1,1)*b(1,i) + rot(1,2)*b(2,i) + rot(1,3)*b(3,i)
      !c(2,i) = rot(2,1)*b(1,i) + rot(2,2)*b(2,i) + rot(2,3)*b(3,i)
      !c(3,i) = rot(3,1)*b(1,i) + rot(3,2)*b(2,i) + rot(3,3)*b(3,i)
      end do

      end subroutine rotate3n

       subroutine rotate33 ( c, rot, b )
      IMPLICIT NONE
      real( kind=8 ), dimension(3,3), intent(in) :: rot
      real (kind=8 ), dimension(3,3), intent(in) :: b
      real (kind=8 ), dimension(3,3), intent(out) :: c

      integer :: i

      do i = 1,3
      !c(:,i) = rotate31(
      call rotate31 ( c(:,i), rot, b(:,i) )
      end do
      end subroutine rotate33

      function determinant ( mat ) result ( det )
        IMPLICIT NONE
        real ( kind=8 ), dimension(3,3), intent ( in ) :: mat
        real ( kind=8 ) :: det

        det = 0.d0
        det = mat(1,1)*( mat(2,2)*mat(3,3)-mat(2,3)*mat(3,2) ) -
     +        mat(1,2)*( mat(2,1)*mat(3,3)-mat(2,3)*mat(3,1) ) +
     +        mat(1,3)*( mat(2,1)*mat(3,2)-mat(2,2)*mat(3,1) )

      end function determinant

c original version, gauss fixed max to 3
      SUBROUTINE ggnml(n,gaus)
! generates normally distributed random numbers
! mean 0 and std. deviation 1
!! !$    USE omp_lib
      IMPLICIT REAL*8 (a-h,o-z)
      DIMENSION gaus(3)
      !DOUBLE PRECISION u1,u2,rand,dum
      !DOUBLE PRECISION u1,u2,rand,dum
      DOUBLE PRECISION u1,u2,dum

      !write(*,*) 'entry ggnml ', dum
      pi2 = 8.d0*datan(1.d0)

!! !$OMP CRITICAL ( call_zbql )
      DO 1 i = 1,n
    2     u1 = ZBQLU01(dum)
      !write(*,*) 'dum u1', dum, u1
          IF (u1.EQ.0) GO TO 2
          u2 = ZBQLU01(dum)
          x = dlog(u1)
          x = x*(-2)
          x = dsqrt(x)
          y = pi2*u2
          y = dcos(y)
          gaus(i) = x*y
    1 CONTINUE
!! !$OMP END CRITICAL ( call_zbql )
      RETURN
      END

!! new version, n is fully used, tested in flexible
!! send back real*4
      SUBROUTINE ggnml4_new ( n, gaus )
c generates normally distributed random numbers
      IMPLICIT REAL*8 (a-h,o-z)
      ! need to declare block data from randgen for ZBQLU01
      !    IMPLICIT NONE
      !DIMENSION gaus(3)
        real ( kind=4 ), dimension ( n ) :: gaus
        integer :: n

      !DOUBLE PRECISION u1,u2,dum

          real ( kind=8 ) :: u1, u2, dum, pi2, x, y
          integer :: i

          !write(*,*) 'entry ggnml ', dum
          pi2 = 8.d0 * datan ( 1.d0 )

          do i = 1, n
    2       u1 = ZBQLU01 ( dum )
      !write(*,*) 'dum u1', dum, u1
              IF ( u1.EQ.0 ) GO TO 2
              u2 = ZBQLU01 ( dum )
              x = dlog( u1 )
              x = x*(-2)
              x = dsqrt(x)
              y = pi2*u2
              y = dcos(y)
              gaus(i) = real( x*y )
          end do

      end subroutine ggnml4_new

c      SUBROUTINE ggubs(n,unif)
      subroutine ggubs ( unif )
c gererates uniformly distributed random numbers
      IMPLICIT REAL*8 (a-h,o-z)

c      n = 0
      u1 = ZBQLU01(dum)
      unif = u1
c      write(99,500) unif
c 500  format(f10.8)
      RETURN
      END

      !! test with rand(), not thread safe
c      SUBROUTINE ggubs_thread_safe (n,unif)
c gererates uniformly distributed random numbers
c      IMPLICIT REAL*8 (a-h,o-z)

c       unif = RAND ()
c      RETURN
c      END

      subroutine meanvar( input, mean, var )
        IMPLICIT NONE
        real(kind=8), dimension(:), intent ( in ) :: input
        real(kind=8), intent ( out ) :: mean, var

        integer :: size_input, i
        real ( kind=8 ) :: delta, M2
        mean = 0.d0
        var = 0.d0
        M2 = 0.d0
        size_input = size ( input )
        !write(*,*) 'entry meanvar size=', size_input
        !write(*,*) 'data ', input

        do i = 1, size_input
            delta = input( i ) - mean
            mean = mean + delta / real(i,8)
            M2 = M2 + delta * ( input (i) - mean )
        end do

        var = M2 / real( size_input - 1, 8 )

        !write(*,*) 'mean var', mean, var

      end subroutine meanvar

!> Linear least square.\n
!! The input data set is X(m), Y(m).\n
!! The number of data points is n (n must be > 2).\n
!! The returned parameters are:\n
!!    a,b, coefficients of equation\n
!!    Y = a + b X, and d, standard deviation of fit.\n
!! original function label array from 0 to n-1 !!
      subroutine least_square(n,X,Y,a,b,d,r2)
      implicit none

      interface
        subroutine meanvar( input, mean, var )
          IMPLICIT NONE
          real(kind=8), dimension(:), intent ( in ) :: input
          real(kind=8), intent ( out ) :: mean, var
        end subroutine meanvar
      end interface
      !include 'sizes.i'
      integer :: n
      !real*8  a,b,d,X(maxvecsize),Y(maxvecsize)
      real ( kind=8 ), dimension (n), intent ( in ) :: X,Y
      real*8  a,b,d, r2

      real*8  a1,a2,b0,b1,d1
      real*8  mean,var ! sample mean and variance
      real*8  sstot ! total sum of square
      !real*8  r2
      integer m

      mean=0
      var=0
      a1 = 0
      a2 = 0
      b0 = 0
      b1 = 0
      do m = 0, n-1
        a1 = a1+X(m)
        a2 = a2+X(m)*X(m)
        b0 = b0+Y(m)
        b1 = b1+Y(m)*X(m)
      enddo
      a1 = a1/n
      a2 = a2/n
      b0 = b0/n
      b1 = b1/n
      d = a1*a1-a2
      a = a1*b1-a2*b0
      a = a/d
      b = a1*b0-b1
      b = b/d
      !  Evaluation of standard deviation d (unbiased estimate)
      d = 0
      do m = 0, n-1
        d1 = Y(m)-a-b*X(m)
        d  = d+d1*d1
      enddo
      d = dsqrt(d/(n-2))

      !  Evaluation of R**2 (R**2=1-sserr/sstot)
      !call meanvar(Y,n,mean,var)
      call meanvar ( Y, mean, var )
      sstot = var*(n-1)
      r2 = 1 - ((d**2)*n-2)/sstot

      !return
      end subroutine least_square

      !> same function, but label array from 1 to N
      subroutine least_square2(n,X,Y,a,b,d,r2)
      implicit none

      interface
        subroutine meanvar( input, mean, var )
          IMPLICIT NONE
          real(kind=8), dimension(:), intent ( in ) :: input
          real(kind=8), intent ( out ) :: mean, var
        end subroutine meanvar
      end interface
      !include 'sizes.i'
      integer, intent ( in ) :: n
      !real*8  a,b,d,X(maxvecsize),Y(maxvecsize)
      real ( kind=8 ), dimension (n), intent ( in ) :: X,Y
      real ( kind=8 ), intent ( out ) ::  a,b,d, r2

      real*8  a1,a2,b0,b1,d1
      real*8  mean,var ! sample mean and variance
      real*8  sstot ! total sum of square
      !real*8  r2
      integer m

      mean=0
      var=0
      a1 = 0
      a2 = 0
      b0 = 0
      b1 = 0
      do m = 1, n
        a1 = a1+X(m)
        a2 = a2+X(m)*X(m)
        b0 = b0+Y(m)
        b1 = b1+Y(m)*X(m)
      enddo
      a1 = a1/n
      a2 = a2/n
      b0 = b0/n
      b1 = b1/n
      d = a1*a1-a2
      a = a1*b1-a2*b0
      a = a/d
      b = a1*b0-b1
      b = b/d
      !  Evaluation of standard deviation d (unbiased estimate)
      d = 0
      do m = 1, n !-1
        d1 = Y(m)-a-b*X(m)
        d  = d+d1*d1
      enddo
      d = dsqrt(d/(n-2))

      !  Evaluation of R**2 (R**2=1-sserr/sstot)
      !call meanvar(Y,n,mean,var)
      call meanvar ( Y, mean, var )
      sstot = var*(n-1)
      r2 = 1 - ((d**2)*n-2)/sstot

      !return
      end subroutine least_square2

      ! centered "real" derivative
      subroutine derivative ( input, output, dt, opt_factor )
        IMPLICIT NONE
        real ( kind=8 ), dimension(:), intent ( in ) :: input
        real ( kind=8 ), dimension(:), intent( out ) :: output
        real ( kind=8 ) :: dt
        real ( kind=8 ), optional :: opt_factor


        real ( kind=8 ) :: factor
        integer :: size_input, i

        !default
        factor = 1.d0
        size_input = size ( input, 1 )
        write(*,*) 'entry derivative size_input ', size_input

        if ( present ( opt_factor ) ) factor = opt_factor
        write(*,*) 'factor ', factor

        !! first point use only forward value
        output ( 1 ) = ( input ( 2 ) - input ( 1 ) ) / ( factor * dt )

        !! put all in factor
        factor = 2.d0 * factor * dt
        !! rest use centered difference
        do i = 2, size_input - 1
            output(i) = ( input ( i + 1 ) - input ( i - 1 ) ) / factor
        end do

        !do i = 1, size_input - 1
        !  write(123,*) output ( i )
        !end do
      end subroutine derivative

      !> kind of derivative, simplified version, 
      !! gives a smoother approximation
      subroutine simple_derivative ( input, output, dt, opt_factor )
        IMPLICIT NONE
        real ( kind=8 ), dimension(:), intent ( in ) :: input
        real ( kind=8 ), dimension(:), intent( out ) :: output
        real ( kind=8 ) :: dt
        real ( kind=8 ), optional :: opt_factor


        real ( kind=8 ) :: factor
        integer :: size_input, i

        !default
        factor = 1.d0
        size_input = size ( input, 1 )
        write(*,*) 'entry simple derivative size_input ', size_input

        if ( present ( opt_factor ) ) factor = opt_factor
        !write(*,*) 'factor ', factor

        !! put all in factor
        factor = factor * dt
        ! derivative should have one point less.
        do i = 1, size_input - 1
            output(i) = input ( i + 1 ) / ( real( i, 8 ) * factor )
        end do

        !do i = 1, size_input - 1
        !  write(123,*) output ( i )
        !end do

      end subroutine simple_derivative

      subroutine norm( wa )
      !IMPLICIT REAL*8 (a-h,o-z)
      IMPLICIT NONE
      real ( kind = 8 ), dimension ( 3 ) :: wa
      real ( kind = 8 ) :: sum_vec
!       normalize:
      sum_vec = sqrt ( dot_product ( wa, wa ) )
      wa = wa / sum_vec

      end subroutine norm

      !include 'randgen.f'
C*PL*ERROR* Missing END statement
