!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file maths_thread_safe.f90
!!
!! Implement thread-safe mathematical subroutines which deals with RGN
!!
!! see \ref RNG "Random Number Generator section
!!
!! Theses subroutine call different generator depending if the compilation was done with OpenMP.\n
!! Allow the same function call, but a different implementation

      !> Generate 3 gaussian random number
      !!
      !! Gaussian with mean 0 and variance 1
      !! Safe function to call, in all cases
      !!
      !! @result gauss : array of 3 random generated numbers
      subroutine ggnml_thread_safe ( gaus )
!$      USE Par_Zig_mod
!$      USE omp_lib
        IMPLICIT NONE

        real ( kind=8 ), dimension ( 3 ), intent ( out ) :: gaus
        integer, parameter :: n = 3
!$      integer :: i, thd_id
        ! should not write thd_id = 0, otherwise considered as global data
        ! may i can in this case ?? if not openmp ??

!$      thd_id = omp_get_thread_num()
!       write(*,*) 'thread safe ggnml thd_id ', thd_id
!$      do i = 1, 3
!$          gaus ( i ) = par_rnor( thd_id )
!$      end do
!       write(*,*) 'thread safe ggnml thd_id ', thd_id, gaus
!$      return

        !! else call original version, later recopy the method ??
        call ggnml ( n, gaus )

      end subroutine ggnml_thread_safe

	  ! impossible to call directly ZBQLU01 !! crash on 64 bits, inside use double precision ??
	  ! let code for later test, but how coud it work ??

	  !> Generate an uniform distributed number between 0 and 1
	  !!
	  !! Thread-safe function
	  !! @result unif : random number
      subroutine ggubs_thread_safe ( unif )
!$      USE Par_Zig_mod
!$      USE omp_lib
        IMPLICIT NONE
        real ( kind=8 ), intent ( out ) :: unif
!        integer, parameter :: n = 3

!$      integer :: thd_id

        !write(*,*) 'thread safe thd_id ', thd_id
        !thd_id = 0
        !unif = 0

!! if OpenMP use Ziggurat and return directly
!$      thd_id = omp_get_thread_num()
! !$      write(*,*) 'thread safe thd_id ', thd_id
!$      unif = par_uni( thd_id )
!$      return
        !write(*,*) 'thread safe ggubs thd_id ', thd_id, unif

!! else original generator
		!dum = 0.d0
        !u1 = ZBQLU01(dum)
        !unif = u1
        !write(*,*) 'unif ggubs ', unif, u1
!! for 32 and 64 bits only working version ( did not find which type to send and/or receive in fact )
!        call ggubs ( n, unif )
        call ggubs ( unif )

      end subroutine ggubs_thread_safe
