!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_analytic_2proteins.f90
!! Compute force and energies for analytic interaction (DH for now) for sda_2proteins.

! Last dev. mainly in sdamm, recopy the method to sda_2proteins
MODULE mod_compute_analytic_2proteins
     USE mod_grid
     USE mod_array_protein

    contains

!> Compute force with analytic function.\n
!! Values precomputed and stored in an array
!! Seems necessary to make separate for dh ? easier
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param dist : distance between solutes, in Angs.
!! @param f,t,t1 : force and torques on the solute 2, torque on 1
subroutine force_analytic_dh( tab_protein, param_force_energy, dist, f, t, t1 )
    USE mod_array_protein
    USE mod_force_energy
    IMPLICIT NONE
    type ( array_protein_type ), target, intent(in) :: tab_protein
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real( kind=4 ), intent(in) :: dist
    ! will update the values
    real ( kind=8 ), dimension ( 3 ), intent(inout) :: f, t, t1

    ! all setup to 0.d0 in functions
    real ( kind=8 ), dimension ( 3 ) :: ff2_analytic, ff1_analytic, ff2, tt2, ff1, tt1
    integer :: type_p1, type_p2
    type ( protein ), pointer :: prot1, prot2

#ifdef VERBOSE
    write(*,*)
    write(*,*) '=================================='
    write(*,*) 'Entry force analytic dh 2 proteins'
    write(*,*) 'dist =', dist
#endif

    prot1 => tab_protein % array ( 1 )
    prot2 => tab_protein % array ( 2 )


    type_p1 = param_force_energy % array_prot_type ( 1 )
    type_p2 = param_force_energy % array_prot_type ( 2 )

    ! cut-off tested inside the function
    ! no torque in full analytic
    call pair_force_full_analytic_2proteins( prot1, prot2, type_p1, type_p2, dist, &
                                             ff2_analytic, ff1_analytic )
#ifdef VERBOSE
    write(*,*) 'After full analytic'
    write(*,*) 'Force ff2_analytic DH ', ff2_analytic
    write(*,*) 'Force ff1_analytic DH ', ff1_analytic
#endif
    ! sda_2proteins done inside function usually
    ! here ff2 = -ff1 , f = 2*ff2
    !f = f + 2.d0 * ff2_analytic !== +(ff2 - ff1)
    !write(*,*) 'Update total f ', f

    call pair_force_partial_analytic_2proteins( prot1, prot2, type_p1, type_p2, dist, ff2, ff1, tt2, tt1 )
#ifdef VERBOSE
    write(*,*) 'After partial analytic'
    write(*,*) 'Force ff2 DH ', ff2
    write(*,*) 'Force ff1 DH ', ff1
    ! should be correctly updated inside the function
    write(*,*) 'Torque tt2 DH ', tt2
    write(*,*) 'Torque tt1 DH ', tt1
#endif

    ! update of f
    !f = f + ff2 - ff1
    ! factor 2 times larger than SDAMM,
    ! but SDA_2proteins, did not divide ep by 2.
    f = f + (ff2_analytic + ff2) - (ff1_analytic + ff1)

    ! update the torque, to check
    t = t + tt2
    t1 = t1 + tt1

    !write(*,*)
    !write(*,*) '== End force_analytic_dh =='
    !write(*,*) 'f ', f
    !write(*,*) 't ', t
    !write(*,*) 't1 ', t1
    !write(*,*) '######################'


end subroutine force_analytic_dh

!> Compute force if solute and grids do not overlap
!! p2 solute in DH "grid" of p1, and opposite.\n
!!
!! @param prot1, prot2 : solute 1 (fixed) and 2 , instance of \ref mod_protein::protein "protein"
!! @param type_p1, type_p2 : integer for the solute type
!! @param dist : distance between the 2 solutes
!! @param ff2, ff1 : force on solute 2 and 1, should be exactly the same with full analytic
subroutine pair_force_full_analytic_2proteins( prot1, prot2, type_p1, type_p2, dist, ff2, ff1 )
    USE mod_array_protein
    IMPLICIT NONE
    type ( protein ), intent(in) :: prot1, prot2
    real( kind=4 ), intent(in) :: dist
    real ( kind=8 ), dimension ( 3 ), intent(out) :: ff2, ff1
    integer :: type_p1, type_p2

    real ( kind=8 ), dimension ( 3 ) :: pos_relat
    real ( kind=8 ) :: ff_tmp
    ! some need to be extracted only once, depends 2 prot or sdamm
    integer :: bin
    real(kind=8) :: size_epgrid_p1, size_epgrid_p2

    !write(*,*) '==============================='
    !write(*,*) 'Entry pair force full analytic '
    !write(*,*) 'type p1 ', type_p1
    !write(*,*) 'type p2 ', type_p2
    !write(*,*) 'tot_charge *tot_charge ', prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total
    
    ! certainly not needed pos_relat, only dist
    ! needed to indicate the direction
    pos_relat = prot2 % position
    !write(*,*) 'pos_relat ', pos_relat
    !write(*,*) 'dist ', dist

    ff2 = 0.d0
    ff1 = 0.d0

    size_epgrid_p1 = prot1 % p_sogrid % pset_grid ( 1 ) % rp
    size_epgrid_p2 = prot2 % p_sogrid % pset_grid ( 1 ) % rp

    ! compute energy2 , total charge prot2 in analytic grid of prot1
    ! only if full outside of prot1 el.grid
    if ( dist > ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'outside size_epgrid_p1 ', size_epgrid_p1, ' +dist_max p2 ', prot2 % p_sogrid % distance_max
        !write(*,*) 'Compute ff2'

        bin = int ( dist / prot1 % p_sogrid % pset_analytic_grid( 1 ) % h )
        !write(*,*) 'bin ', bin
        ! should be get value
        !write(*,*) 'Value of DH for prot 1 at bin=', bin, ', value= ', &
        !            prot1 % p_sogrid % pset_analytic_grid( 1 ) % parray_force ( bin, type_p2 )

        ! norm of the force for this interaction
        !ff_tmp = prot1 % p_sogrid % pset_analytic_grid( 1 ) % parray_force ( bin, type_p2 )
        ff_tmp = get_force_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
        !> Convert to kT/A
        ff_tmp = ff_tmp * 1.d-10
        !write(*,*) 'ff_tmp ', ff_tmp !, 'norm ', sqrt( dot_product(ff_tmp,ff_tmp) )

        ! adjust the correct orientation
        ff2 = prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total * ff_tmp * ( pos_relat / dist )
        !write(*,*) 'ff2 after charge multiplication ', ff2
        !write(*,*) 'ff2 norm ', sqrt( dot_product(ff2,ff2) )
    !else
    !    write(*,*) 'Partialy or full inside ep_grid1, see later '
    end if

    ! compute energy1, no need of change of refererential here
    ! bin certainly identical for both, because dist is identical
    ! only if full outside if prot2 el.grid
    if ( dist > ( size_epgrid_p2 + prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'outside size_epgrid_p2 ', size_epgrid_p2, ' +dist_max p1 ', prot1 % p_sogrid % distance_max
        !write(*,*) 'Compute ff1, should be similar if full DH and symetric protein'
        pos_relat = -pos_relat
        !write(*,*) 'Value of DH for prot 2 at bin=', bin, ', value= ', &
        !            prot2 % p_sogrid % pset_analytic_grid( 1 ) % parray_force ( bin, type_p1 )

        ! norm of the force for this interaction
        ff_tmp = get_force_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
        !> Convert to kT/A
        ff_tmp =ff_tmp * 1.d-10
        !write(*,*) 'ff_tmp ', ff_tmp !, 'norm ', sqrt( dot_product(ff_tmp,ff_tmp) )

        ! adjust the correct orientation
        ff1 =  prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total * ff_tmp *( pos_relat / dist )
        !write(*,*) 'ff1 ', ff1, 'norm ff1 ', sqrt( dot_product( ff1, ff1 ) )
    !else
    !    write(*,*) 'Partialy or full inside ep_grid2, see later '
    end if

end subroutine pair_force_full_analytic_2proteins

!> Only partial analytic, loop over all atoms in this case
!! Analytic is set to zero if inside the cutoff
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param type_p1, type_p2 : integer to get the type of solute
!! @param dist : distance between the solutes
!! @param ff2, ff1, tt2, tt1 : return the torque and force for the 2 solutes
subroutine pair_force_partial_analytic_2proteins( prot1, prot2, type_p1, type_p2, dist, ff2, ff1, tt2, tt1 )
    USE mod_array_protein
    IMPLICIT NONE

    type ( protein ), intent(in) :: prot1, prot2
    real ( kind=8 ), dimension ( 3 ), intent(out) :: ff2, ff1, tt2, tt1
    real ( kind=4 ) :: dist
    integer :: type_p1, type_p2

    interface
          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr

          subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross
    end interface

    real ( kind=8 ), dimension ( 3 ) :: pos_relat, tmp1, tmp2, tt2_copy
    real ( kind=8 ), dimension(3) :: x, p1, p2, rto, force
    real ( kind=8 ) :: ff_tmp, dist_atom !q
    real(kind=8) :: size_epgrid_p1, size_epgrid_p2
    integer :: bin, nq1, nq2, j

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2

    ff1 = (/0.d0, 0.d0, 0.d0/)
    tt1 = (/0.d0, 0.d0, 0.d0/)
    ff2 = (/0.d0, 0.d0, 0.d0/)
    tt2 = (/0.d0, 0.d0, 0.d0/)
    tmp1 = (/0.d0, 0.d0, 0.d0/)
    tmp2 = (/0.d0, 0.d0, 0.d0/)
    pos_relat = prot2 % position

    size_epgrid_p1 = prot1 % p_sogrid % pset_grid ( 1 ) % rp
    size_epgrid_p2 = prot2 % p_sogrid % pset_grid ( 1 ) % rp

    p_charge1 =>  prot1 % p_sogrid % charge_grid
    p_charge2 =>  prot2 % p_sogrid % charge_grid
    p_charge_position1 => prot1 % p_pos_charge
    p_charge_position2 => prot2 % p_pos_charge
    nq1 = prot1 % nb_charge
    nq2 = prot2 % nb_charge

    !write(*,*)
    !write(*,*) 'Entry partial analytic 2proteins'

    ! compute force2 , partial charge prot2 in analytic grid of prot1
    ! discard case fully outside, already treated
    if ( dist > ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'dist > size_epgrid_p1 + prot2 % p_sogrid % distance_max =', &
        !    size_epgrid_p1 + prot2 % p_sogrid % distance_max, ' GOTO partial force1'
        GOTO 100
    end if

    ! only if paritally in prot1 el.grid
    if ( dist > ( size_epgrid_p1 - prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'Partially inside tje grid/analytic of ep1 ', size_epgrid_p1 - prot2 % p_sogrid % distance_max

        ! same as normal force 2 proteins, 0 inside the epgrid
        FORCE2: do j=1,nq2

            ! got position to the center of protein 1
            p2 = p_charge_position2 (:,j)
            ! needed here ? yes certainly
            x = p2 + pos_relat

            dist_atom = sqrt ( dot_product( x,x ) )
            !write(*,*) 'x =', x, 'norm x= ', sqrt( dot_product( x,x ) )
            !write(*,*) 'dist  atom', dist_atom

            bin = int ( dist_atom / prot1 % p_sogrid % pset_analytic_grid( 1 ) % h )
            !write(*,*) 'bin ', bin
            !write(*,*) 'Value of DH for prot 1 at bin=', bin, ', value= ', &
            !    prot1 % p_sogrid % pset_analytic_grid( 1 ) % parray_force ( bin, type_p2 )

            ff_tmp = get_force_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
            !> Convert to kT/A
            ff_tmp = ff_tmp * 1.d-10

            ! pos relat, associated to center
            force = ff_tmp * p_charge2(j) * prot1 % p_sogrid % charge_total * ( x / dist_atom )
            !write(*,*) 'force parttial DH', force

            call cross (tmp2, p2, force )
            tt2 = tt2 + tmp2 !tt2 = tt2 + ti2
            ff2 = ff2 + force !ff2 = ff2 + qf2

          end do FORCE2
    !else
    !    write(*,*) 'Fully inside epgrid1, already accounted'
    end if

100 continue

    ! compute force1, partial charge of prot1 in grid of prot2
    ! discard case fully inside, already treated
    if ( dist > ( size_epgrid_p2 + prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'dist > size_epgrid_p2 + prot1 % p_sogrid % distance_max =', &
        !    size_epgrid_p2 + prot1 % p_sogrid % distance_max, ' goto 200'
        ! need torque correction...cannot exit
        goto 200
    end if

    !write(*,*)
    !write(*,*) 'Compute force on p1'
    ! only if paritally in prot1 el.grid
    if ( dist > ( size_epgrid_p2 - prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'Partially inside the grid/analytic of ep1 ', size_epgrid_p1 - prot2 % p_sogrid % distance_max

        ! not necessary, see test/comment in analytic_sdamm
        ! torques and force for the first protein
        !ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
        !ey = prot2 % orientation_y
        !ez = prot2 % orientation_z
        !xe(1) = ex(1)
        !xe(2) = ey(1)
        !xe(3) = ez(1)
        !ye(1) = ex(2)
        !ye(2) = ey(2)
        !ye(3) = ez(2)
        !ze(1) = ex(3)
        !ze(2) = ey(3)
        !ze(3) = ez(3)

    FORCE1: do j=1,nq1

               p1 = p_charge_position1(:,j) ! p1 = x1(:,j)
               rto = - pos_relat + p1 !rto = -r + x1(:,j) !!position charge protein1

               ! identical to previous, only distance and correct relative position are needed
               dist_atom = sqrt ( dot_product( rto,rto ) )
               !write(*,*) 'dist atom', dist_atom

               bin = int ( dist_atom / prot2 % p_sogrid % pset_analytic_grid( 1 ) % h )
               !write(*,*) 'bin ', bin
               !write(*,*) 'Value of DH for prot 2 at bin=', bin, ', value= ', &
               !     prot2 % p_sogrid % pset_analytic_grid( 1 ) % parray_force ( bin, type_p1 )

               ff_tmp = get_force_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
               !> Convert to kT/A
               ff_tmp = ff_tmp * 1.d-10

               ! can use normal/original referential, will be always pointing to the center of the sphere
               force = ff_tmp * p_charge1(j) * prot2 % p_sogrid % charge_total * ( rto / dist_atom )
               ! update force
               ff1 = ff1 + force
               ! update torque
               call cross(tmp1, p1, force) !call cross(ti1, p1, qf1)
               tt1 = tt1 + tmp1

            end do FORCE1
    !else
    !    write(*,*) 'Fully inside epgrid2, already accounted'
    end if

200 continue
!   apply correction, like uhbd case
    !torque computed on ff1 and ff2, will be asymetric
    !torque on 2   (-r)x(-ff1), CALL cross(tmp1,r,ff1)
    call cross ( tmp1, pos_relat, ff1 )
    !torque on 1    (r)x(-ff2), CALL cross(tmp2,r,ff2)
    call cross ( tmp2, pos_relat, ff2 )
    !write(*,*) 'tt1 tmp1 ',tt1, tmp1
    !write(*,*) 'tt2 tmp2 ',tt2, tmp2

    tt2_copy = tt2

    ! should not correct if asymetric, why not ??
    ! see remark in sdamm equivalent
    !ff2 = ff2 - ff1
    tt2 = tt2 - tt1 + tmp1
    tt1 = tt1 - tt2_copy - tmp2


end subroutine pair_force_partial_analytic_2proteins

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!! Compute energies 2 proteins !!!!!!!!!!!!!!!!!!!!!!!!!!!!

!> Compute energy with analytic function
!! Seems necessary to make separate for dh ? easier.. to see later
subroutine energy_analytic_dh( tab_protein, param_force_energy, dist, array_energy )
    USE mod_array_protein
    USE mod_force_energy
    IMPLICIT NONE
    type ( array_protein_type ), target, intent(in) :: tab_protein
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension (:,:), intent ( inout ) :: array_energy
    real( kind=4 ), intent(in) :: dist

    integer :: type_p1, type_p2
    type ( protein ), pointer :: prot1, prot2
    real(kind=8) :: energy1, energy2, energy1_analytic, energy2_analytic

#ifdef VERBOSE
    write(*,*)
    write(*,*) '=== Entry energy analytic dh 2 proteins ==='
    write(*,*) 'dist =', dist
#endif

    prot1 => tab_protein % array ( 1 )
    prot2 => tab_protein % array ( 2 )
    type_p1 = param_force_energy % array_prot_type ( 1 )
    type_p2 = param_force_energy % array_prot_type ( 2 )

    ! if try to make a common call sda / sdamm
    call pair_energy_full_analytic_2proteins( prot1, prot2, type_p1, type_p2, dist, &
                energy1_analytic, energy2_analytic )
#ifdef VERBOSE
    write(*,*) 'After energy full analytic'
    write(*,*) 'energy1_analytic ', energy1_analytic
    write(*,*) 'energy2_analytic ', energy2_analytic
#endif

    !write(*,*) 'Compute partial analytic energy'
    call pair_energy_partial_analytic( prot1, prot2, type_p1, type_p2, dist, energy1, energy2 )
#ifdef VERBOSE
    write(*,*) 'After energy partial analytic'
    write(*,*) 'energy1_partial_analytic ', energy1
    write(*,*) 'energy2_partial_analytic ', energy2
#endif

    ! add energies to electrostatic part, grid_number(1) + 1(total)=2
    array_energy ( 2, 1 ) = array_energy ( 2, 1 ) + energy1 + energy1_analytic
    array_energy ( 2, 2 ) = array_energy ( 2, 2 ) + energy2 + energy2_analytic

end subroutine energy_analytic_dh

!> Compute Full DH interactions.\n
!! Need to split, cutoffs may be different/ asymetric between p1 and p2.\n
!! Only compute El-DH prot1 in analytical grid of protein 2 ( all grids stored in prot2 % p_sogrid % pset_analytical)
!! Not good for 2 proteins because asymetry between static/mobil solute
!! Need to test cutoffs for each interaction
!! @param prot1, prot2 : olute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param type_p1, type_p2 : integer to get the type of solute
!! @param dist : distance between the solutes
!! @param energy1, energy2 : return the energy of solute 1 and 2
subroutine pair_energy_full_analytic_2proteins ( prot1, prot2, type_p1, type_p2, dist, energy1, energy2 )
    USE mod_array_protein
    IMPLICIT NONE
    type ( protein ), intent(in) :: prot1, prot2
    real( kind=4 ), intent ( in ) :: dist
    real ( kind=8 ), intent(out) :: energy1, energy2
    integer :: type_p1, type_p2

    !real(kind=8), dimension(3) :: pos_relat
    real(kind=8) :: energy1_tmp, energy2_tmp
    integer :: bin
    real(kind=8) :: size_epgrid_p1, size_epgrid_p2

    !write(*,*) 'Entry pair_energy_full_analytic'
    energy1 = 0.d0
    energy2 = 0.d0

    size_epgrid_p1 = prot1 % p_sogrid % pset_grid ( 1 ) % rp
    size_epgrid_p2 = prot2 % p_sogrid % pset_grid ( 1 ) % rp

    ! compute energy2 , total charge prot2 in analytic grid of prot1
    ! only if full outside of prot1 el.grid
    if ( dist > ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'outside size_epgrid_p1 ', size_epgrid_p1, ' +dist_max p2 ', prot2 % p_sogrid % distance_max
        bin = int ( dist / prot1 % p_sogrid % pset_analytic_grid( 1 ) % h )

        !write(*,*) 'bin ', bin
        !write(*,*) 'Value of DH for prot 1 at bin=', bin, ', value= ', &
        !    get_energy_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
        energy2_tmp =  get_energy_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
        energy2 = prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total * energy2_tmp
    !else
    !    write(*,*) 'Partialy or full inside ep_grid1, see later '
    end if

    ! compute energy1, no need of change of refererential here
    ! bin certainly identical for both, because dist is identical
    ! only if full outside if prot2 el.grid
    if ( dist > ( size_epgrid_p2 + prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'outside size_epgrid_p2 ', size_epgrid_p2, ' +dist_max p1 ', prot1 % p_sogrid % distance_max
        bin = int ( dist / prot2 % p_sogrid % pset_analytic_grid( 1 ) % h )
        !write(*,*) 'Value of DH for prot 2 at bin=', bin, ', value= ', &
        !   get_energy_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
        energy1_tmp =  get_energy_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
        !write(*,*) 'energy1_tmp ', energy1_tmp
        energy1 = prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total * energy1_tmp
    !else
    !    write(*,*) 'Partially or full inside ep_grid2, see later '
    end if

end subroutine pair_energy_full_analytic_2proteins

!> Only partial analytic, loop over all atoms in this case.\n
!! Analytic has been set to zero if inside the grid
!! Asymetry due to cutoffs
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param type_p1, type_p2 : integer to get the type of solute
!! @param dist : distance between the solutes
!! @param energy1, energy2 : return the energy of solute 1 and 2
subroutine pair_energy_partial_analytic( prot1, prot2, type_p1, type_p2, dist, energy1, energy2 )
    USE mod_array_protein
    IMPLICIT NONE
    type ( protein ), intent(in) :: prot1, prot2
    real( kind=4 ), intent ( in ) :: dist
    real ( kind=8 ), intent(out) :: energy1, energy2
    integer :: type_p1, type_p2

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2

    real(kind=8), dimension(3) :: pos_relat
    real ( kind=8 ), dimension(3) :: x, p1, p2, rto
    real(kind=8) :: energy1_tmp, energy2_tmp, dist_atom
    integer :: j, bin, nq1, nq2
    real(kind=8) :: size_epgrid_p1, size_epgrid_p2

    energy1 = 0.d0
    energy2 = 0.d0

    !write(*,*)
    !write(*,*) '== Pair energy partial analytic =='
    ! does not work in use for test_force_energy, dist is still correct
    !pos_relat = prot2 % position
    pos_relat = prot2 % position - prot1 % position
    ! now pos relat given, not need for energy
    !write(*,*) 'pos_relat ', pos_relat
    !dist = real ( sqrt ( dot_product( pos_relat, pos_relat ) ), 4 )
    !write(*,*) 'dist ', dist

    size_epgrid_p1 = prot1 % p_sogrid % pset_grid ( 1 ) % rp
    size_epgrid_p2 = prot2 % p_sogrid % pset_grid ( 1 ) % rp

    p_charge1 =>  prot1 % p_sogrid % charge_grid
    p_charge2 =>  prot2 % p_sogrid % charge_grid
    p_charge_position1 => prot1 % p_pos_charge
    p_charge_position2 => prot2 % p_pos_charge
    nq1 = prot1 % nb_charge
    nq2 = prot2 % nb_charge

    ! compute energy2 , parital charge prot2 in analytic grid of prot1
    ! discard case fully outside, already treated
    if ( dist > ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'dist > size_epgrid_p1 + prot2 % p_sogrid % distance_max =', &
        !    size_epgrid_p1 + prot2 % p_sogrid % distance_max, ' GOTO partial energy1'
        GOTO 100
    end if

    ! only if paritally in prot1 el.grid
    if ( dist > ( size_epgrid_p1 - prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'Partially  in analytic/epgrid of protein1 ', size_epgrid_p1 - prot2 % p_sogrid % distance_max
        !write(*,*) 'Compute energy2'
        !write(*,*) ' nq2= ', nq2

        NRJ2: do j= 1, nq2
            !write(*,*) '======='
            !write(*,*) 'j ', j

            ! position charge solute 2
            p2 = p_charge_position2 (:,j)
            !for relative position of this atom
            x = pos_relat + p2
            !write(*,*) 'x= ', x, ' dist_atom= ', sqrt( dot_product( x, x) )

            ! version analytic
            bin = int ( sqrt ( dot_product ( x, x ) ) / prot1 % p_sogrid % pset_analytic_grid( 1 ) % h )
            !write(*,*) 'bin ', bin
            energy2_tmp = get_energy_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
            energy2 = energy2 + p_charge2 (j) * prot1 % p_sogrid % charge_total * energy2_tmp

        end do NRJ2
    !else
    !    write(*,*) 'Fully inside epgrid1, already accounted'
    end if

100 continue

    ! compute energy1, partial charge in prot
    ! discard case fully inside, already treated
    if ( dist > ( size_epgrid_p2 + prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'dist > size_epgrid_p2 + prot1 % p_sogrid % distance_max =', &
        !    size_epgrid_p2 + prot1 % p_sogrid % distance_max, ' return from here'
        !can exit energy1 is 0
        return
    end if

    ! only if paritally in prot1 el.grid
    if ( dist > ( size_epgrid_p2 - prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'Partially inside the grid/analytic of ep2 ', size_epgrid_p2 - prot1 % p_sogrid % distance_max
        !write(*,*) 'compute energy1 '

  NRJ1: do j = 1, nq1

           p1 = p_charge_position1(:,j)
           !write(*,*) 'p1 ', p1
           ! position charge solute 1
           rto = - pos_relat + p1
           !write(*,*) 'rto ', rto
           dist_atom = sqrt ( dot_product( rto, rto ) )
           ! version analytic
           bin = int ( dist_atom / prot2 % p_sogrid % pset_analytic_grid( 1 ) % h )
           !write(*,*) 'bin ', bin
           energy1_tmp = get_energy_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
           energy1 = energy1 + p_charge1( j ) * prot2 % p_sogrid % charge_total * energy1_tmp

        end do NRJ1

    !else
    !    write(*,*) 'Fully inside epgrid2, already accounted'
    end if

end subroutine pair_energy_partial_analytic

END MODULE mod_compute_analytic_2proteins
