!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_energy.f90
!! group energy computation for sda_2proteins

!> Group all function to compute the energy for sda 2 proteins/ sda koff
MODULE mod_compute_energy

! To make a ref to a page with fast algo.

! naming convention
! energy 2 is the charge of protein 2 in the field of protein 1
! energy 1 the opposite
contains

!> Compute energy for sda_2proteins
!!
!! standard algorithm, loop over all the interactions
!!
!! @param prot1 : solute 1
!! @param prot2 : solute 2
!! @param param_force_energy : instance of type_force_energy in mod_force_energy
!! @param dummy_array : used for the computation of repulsive lennard-jones interaction
!! @param dist : distance between the solutes
!! @param array_energy : return energies
subroutine  energy_epedhdlj_2proteins ( prot1, prot2, param_force_energy, dist, array_energy, &
                                        dummy_array )
     USE mod_protein
     USE mod_force_energy
     IMPLICIT NONE

     type ( protein ), intent ( in ) :: prot1, prot2
     type ( type_force_energy ), intent ( in ) :: param_force_energy
     real ( kind=8 ), dimension (:,:), intent ( out ) :: array_energy
     real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
     real ( kind=4 ), intent ( in ) ::  dist

     real ( kind=8 ) :: energy1, energy2, tmp_energy1, tmp_energy2
     integer :: i, k, grid_number !grid_type

     !NJB: protein type
     integer :: p1_type , p2_type
     !NJB: square of separation used for cut off calculations
     real ( kind=4 ) :: test_dist2

     ! NJB: Find square of distance
     test_dist2 = dist*dist

     ! to be sure
     array_energy = 0.d0

     ! If cut off larger than cut off of largest grid of both proteins, skip
     if ((test_dist2).GT.max(prot1%p_sogrid%max_cut2,prot2%p_sogrid%max_cut2)) then
        return
     end if

     ! NJB: Find type of both proteins
     ! Trivial for 2 protein
     p1_type = param_force_energy%array_prot_type(prot1%id)
     p2_type = param_force_energy%array_prot_type(prot2%id)

     !write(*,*) 'compute energy nb_grid ',param_force_energy % nb_grid

     !! now loop over gird, sure to be associated
     !! should include lennard-jones here
GRID: do i = 1, param_force_energy % nb_grid

          ! needed only for asymetric and multiple subgrids ( lennard-jones )
          energy1 = 0.d0
          energy2 = 0.d0

          grid_number =  param_force_energy % array_grid_calc ( i, 1 )

          if ( param_force_energy % array_grid_calc ( i, 2 ) == symetric ) then
              !write(*,*) 'compute a symetric interaction '
              ! here pos_relat prot2 % position
              call compute_energy_sda_2proteins ( prot1, prot2, grid_number, dummy_array, &
                                                  energy1, energy2 , &
                                                  p1_type , p2_type , test_dist2) !NJB: Add protein types and separation squared

          else
              !write(*,*) 'compute an asymetric interaction ', param_force_energy % array_grid_calc ( i, 2 )
              !write(*,*) 'number of sub-grids ', param_force_energy % array_grid_calc ( i, 3 )

              do k = 1, param_force_energy % array_grid_calc ( i, 3 )

                  call energy_asymetric ( prot1, prot2, grid_number, dummy_array, tmp_energy1, tmp_energy2, &
                               ! reverse,
                               param_force_energy % array_grid_calc ( i, 2 ), &
                               ! subgrid
                               k , p1_type , p2_type, test_dist2) !NJB: Add protein types and separation squared
                  !write(*,*) 'add tmp_energy for subgrid '
                  energy1 = energy1 + tmp_energy1
                  energy2 = energy2 + tmp_energy2
              end do
          !! need to sum up energies here
          end if

          !! shift 1 because total energy
          !write(*,*) 'add energies to grid_number ', grid_number + 1
          !write(*,*) 'energy1 ', energy1
          !write(*,*) 'energy2 ', energy2
          array_energy ( grid_number + 1, 1 ) = energy1
          array_energy ( grid_number + 1, 2 ) = energy2

      end do GRID

      RETURN
end subroutine energy_epedhdlj_2proteins

!> Version with asymetric terms.\n
!! Can deal with all uhbd grids
subroutine energy_asymetric ( prot1, prot2, grid_number, dummy_array, energy1, energy2, reverse, subgrid , &
                               type_p1, type_p2, test_dist2 )  !NJB: Add protein types
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE
    type ( protein ), intent ( in ) :: prot1, prot2
    integer, intent ( in ) :: grid_number
    real ( kind=8 ), dimension ( : ), intent ( in ) :: dummy_array
    !real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ff1, ff2, tt1, tt2 !first comput could out only
    real ( kind=8 ), intent( out ) :: energy1, energy2
    integer :: reverse, subgrid

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2
    type ( type_grid ), pointer :: p_grid1, p_grid2

    real ( kind=8 ), dimension(3) :: pos_relat, ex, ey, ez
    real ( kind=8 ), dimension(3) :: x, p1, p2, rto, rtn
    real( kind=8 ) :: potential
    real ( kind=8 ) :: q
    integer :: j, nq1, nq2

    ! NJB: Protein types
    integer , intent(in):: type_p1, type_p2
    ! NJB: Square of protein separation
    real ( kind=4 )  , intent(in) :: test_dist2

    nullify ( p_grid1, p_grid2 )

    energy1 = 0.d0
    energy2 = 0.d0
    pos_relat = prot2 % position ! ( -prot1 % position for sdamm ) or send as parameter

    !write(*,*)
    !write(*,*) 'compute energy_asymetric_new2, reverse ', reverse
    !write(*,*) 'subgrid ', subgrid
    !write(*,*)

    call select_charge2 ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                          p_charge_position1, p_charge_position2, dummy_array, &
                          ! added specifically for lennard-jones need the grid
                          ! no need to make optional, not used it is all, make optional in select_charge only
                          p_grid1, p_grid2, subgrid )

!write(0,*) "ASYMMETRIC"

    ! "normal " p2 charges in grid of p1, less geometrical transformation to do
    if ( reverse == c2g1 ) then
       !write(*,*) 'case c2g1 '
       !write(*,*) 'opt_reverse ', reverse
       !write(*,*) 'op_subgrid ', subgrid
       !write(*,*) 'nq2 ', nq2
       !write(*,*) 'associated p_charge2 ', associated ( p_charge2 )
       !write(*,*) 'associated p_charge_position2 ', associated ( p_charge_position2 )
       !write(*,*) 'associated p_grid1 ', is_associated ( p_grid1 )
       !write(*,*) 'size dummy_array p_charge2 ', size ( p_charge2 )

       !NJB: apply cut off
       if (test_dist2.LT.prot1%p_sogrid%cut2_array(grid_number,type_p2)) then

  
NRJ2:     do j=1,nq2
             !write(*,*) '======='
             !write(*,*) 'j ', j

             ! for electrostatic, force and torques divided by 2 at the end
             ! here q == q2
             q = p_charge2 ( j )
             !! for hydrophobic should be list_atom_surface ! before nq2p1,ns2l
             p2 = p_charge_position2 (:,j) ! !p2 = y(:,j)
             !for force
             x = pos_relat + p2 ! x = r + p2 !y(:,j)

             call potential_atom_uhbd ( x, potential, p_grid1 )
             !write(*,*) 'force ', force
             energy2 = energy2 + potential * q !!qf2 = f2 * q2
             !write(*,*) 'x ', x
             !write(*,*) 'q ', q
             !write(*,*) 'elect potential ', potential

          end do NRJ2

          !write(*,*) 'ff2, tt2 ', ff2, tt2
       end if

    ! inverse order p1 charges in grid of p2, need more change of referential
    else
       !write(*,*) 'case c1g2 '
       !write(*,*) 'nq2 ', nq2
       !write(*,*) 'associated p_charge2 ', associated ( p_charge2 )
       !write(*,*) 'associated p_charge2 ', associated ( p_charge_position2 )
       ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
       ey = prot2 % orientation_y
       ez = prot2 % orientation_z

       !NJB: apply cut off
       if (test_dist2.LT.prot2%p_sogrid%cut2_array(grid_number,type_p1)) then

  NRJ1:   do j = 1, nq1


             p1 = p_charge_position1(:,j) ! p1 = x1(:,j)
             !write(*,*) 'p1 ', p1
             rto = - pos_relat + p1 !rto = -r + x1(:,j) !!position charge protein1
             !write(*,*) 'rto ', rto
             call tr(rto, rtn, ex, ey, ez )
             !write(*,*) 'rtn ', rtn
             call potential_atom_uhbd ( rtn, potential, p_grid2 )

             energy1 = energy1 + potential * p_charge1( j )

          end do NRJ1

       end if

    end if ! opt_reverse

    !write(*,*) 'intern energy1 ', energy1
    !write(*,*) 'intern energy2 ', energy2

end subroutine energy_asymetric

!!to keep common sda/sdamm need to pass pos_relat

!> Compute only one interaction term, only symetric interactions.\n
!!
!! The type is given by the grid number
!! @param prot1 : solute 1
!! @param prot2 : solute 2
!! @param grid_number : type of grid to treat
!! @param dummy_array : used for the computation of repulsive lennard-jones interaction
!! @param energy1 : energy of protein 1 in the grid of protein 2
!! @param energy2 : energy of protein 2 in the grid of protein 1
!! @param type_p1, type_p2 : integer type of the solutes
!! @param test_dist2 : distance square between the solutes
subroutine compute_energy_sda_2proteins ( prot1, prot2, grid_number, dummy_array, energy1, energy2 , &
                                          type_p1, type_p2 , test_dist2) !NJB: Add protein types
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    interface
          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr
    end interface

    type ( protein ), intent( in ) :: prot1, prot2
    integer, intent ( in ) :: grid_number
    real ( kind=8 ), dimension ( : ), intent ( in ) :: dummy_array
    real ( kind=8 ), intent ( out ) :: energy1, energy2

    real ( kind=8 ), dimension(:), pointer :: p_charge1
    real ( kind=8 ), dimension(:), pointer :: p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position2

    ! tmp orientation, need inversion for forces on protein1
    real ( kind=8 ), dimension(3) :: x, rto, rtn, pos_relat

    !! q for q1 and q2
    real ( kind=8 ) :: potential1, potential2
    integer :: nq1, nq2, j

    ! NJB: Protein types
    integer , intent(in) :: type_p1, type_p2
    ! NJB: Square of protein separation
    real ( kind=4 ) , intent (in) :: test_dist2

    !write(*,*) 'compute energy'
    nq1 = 0
    nq2 = 0

    ! need to initialize ?? YES
    energy1 = 0.d0
    energy2 = 0.d0

    ! needed if called from sdamm
    pos_relat = prot2 % position - prot1 % position
    !write(*,*) 'pos_relat modified ', pos_relat

    !! assign the correct pointer
    call select_charge ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                         p_charge_position1, p_charge_position2, dummy_array )

    !NJB: Check if distance is less than relevant cut off
    if (test_dist2.LT.prot1%p_sogrid%cut2_array(grid_number,type_p2)) then

!! energy2 is the energy of the charge of protein2 in the grid of protein1
ENERG2: do j=1,nq2

            ! for electrostatic, force and torques divided by 2 at the end
            x = pos_relat + p_charge_position2 (:,j)

            call potential_atom_uhbd( x, potential1, prot1 % p_sogrid % pset_grid ( grid_number) )
            energy2 = energy2 + potential1 * p_charge2 ( j )

        end do ENERG2
    end if

    !NJB: Check if distance is less than relevant cut off
    if (test_dist2.LT.prot2%p_sogrid%cut2_array(grid_number,type_p1)) then

ENERG1: do j = 1,nq1

            rto = - pos_relat + p_charge_position1(:,j)
            call tr (rto, rtn, prot2 % orientation_x, prot2 % orientation_y, &
                     prot2 % orientation_z )
            call potential_atom_uhbd( rtn, potential2, prot2 % p_sogrid % pset_grid ( grid_number) )
            energy1 = energy1 + potential2 * p_charge1( j )

        end do ENERG1
    end if

end subroutine compute_energy_sda_2proteins

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  fast version  !!!!!!!!!!!!!!!!!!!

   !> Compute energy for 2 proteins with the fast algorithm
   !!
   !! Fast algorithm, check if it possible to compute couple of interactions at the same time
   !! Apply only with symetric interactions
   !!
   !! @param prot1 : solute 1
   !! @param prot2 : solute 2
   !! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
   !! @param dummy_array : used for the computation of repulsive lennard-jones interaction
   !! @param dist : distance between the solutes
   !! @param array_energy : to store energies
   subroutine  energy_epedhdlj_2proteins_fast ( prot1, prot2, param_force_energy, &
                                                array_energy, dummy_array, dist )
      USE mod_protein
      USE mod_force_energy
      IMPLICIT NONE

      type ( protein ), intent ( in ) :: prot1, prot2
      type ( type_force_energy ), intent ( in ) :: param_force_energy
      real ( kind=8 ), dimension( :, :), intent ( out ) :: array_energy
      real ( kind = 8 ), dimension ( : ), intent ( in ) :: dummy_array
      real ( kind=4 ), intent ( in ) :: dist

      real ( kind=8 ) :: energy11, energy12, energy21, energy22
      integer :: counter

      !NJB: protein type
      integer :: p1_type , p2_type
      !NJB: square of separation used for cut off calculations
      real ( kind=4 ) :: test_dist2

      array_energy = 0.d0

      !write(*,*) 'compute energy fast'
      ! missing test with surface !
      ! \todo add tests for surface

      ! NJB: Find square of distance
      test_dist2 = dist*dist

      ! If cut off larger than cut off of largest grid of both proteins, skip
      if (test_dist2.GT.max(prot1%p_sogrid%max_cut2,prot2%p_sogrid%max_cut2)) then
         return
      end if

      ! NJB: Find type of both proteins
      ! Trivial for 2 protein
      p1_type = param_force_energy%array_prot_type(prot1%id)
      p2_type = param_force_energy%array_prot_type(prot2%id)

      !!see force for other algorithm
      counter = 1
      if ( param_force_energy % couple_ElDes .eqv. .true. ) then

          call compute_energy_couple ( prot1, prot2, param_force_energy % array_grid_calc ( 1, 1 ), &
                                       param_force_energy % array_grid_calc ( 2, 1 ), &
                                       1, dummy_array, energy11, energy12, energy21, energy22 , &
                                       p1_type , p2_type , test_dist2) !NJB: Add protein types and separation squared
          counter = 3

          array_energy ( 2, 1 ) = array_energy ( 2, 1 ) + energy11
          array_energy ( 2, 2 ) = array_energy ( 2, 2 ) + energy21
          array_energy ( 3, 1 ) = array_energy ( 3, 1 ) + energy12
          array_energy ( 3, 2 ) = array_energy ( 3, 2 ) + energy22

      else if ( ( param_force_energy % array_grid_calc ( 1, 1 ) == 1 ) .or. &
                ( param_force_energy % array_grid_calc ( 1, 1 ) == 2 ) ) then

              call compute_energy_sda_2proteins (  prot1, prot2, &
                                  param_force_energy % array_grid_calc ( 1, 1 ), &
                                  dummy_array, energy11, energy21 , &
                                  p1_type , p2_type , test_dist2) !NJB: Add protein types and separation squared
              counter = 2
              !! shift one for total energy
              array_energy ( param_force_energy % array_grid_calc ( 1, 1 ) + 1 , 1 ) = &
                                  array_energy ( 2, 1 ) + energy11
              array_energy ( param_force_energy % array_grid_calc ( 1, 1 ) + 1, 2 ) = &
                                  array_energy ( 2, 2 ) + energy21
      end if

      !need to test size, before continuing ??
      if ( param_force_energy % nb_grid < counter ) then
        !write(*,*) 'in energy no more grid return '
        return
      end if

    !! wrong for fast and both, independent it is fine
    if ( param_force_energy % couple_ApDesLJrep .eqv. .true. ) then

        call compute_energy_couple ( prot1, prot2, &
               param_force_energy % array_grid_calc ( counter, 1 ), &
               param_force_energy % array_grid_calc ( counter + 1, 1 ), &
               2, dummy_array, energy11, energy12, energy21, energy22, &
               p1_type , p2_type , test_dist2) !NJB: Add protein types and separation squared

       ! do not find the same result than normal energy
       !! fill array protein 1
       array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 1, 1 ) = &
         array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 1, 1 ) + energy11
       array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 2, 1 ) = &
         array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 2, 1 ) + energy12
       !! fill array protein 2
       array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 1, 2 ) = &
         array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 1, 2 ) + energy21
       array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 2, 2 ) = &
         array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 2, 2 ) + energy22

     !! else only one the grid to compute
     else if ( ( param_force_energy % array_grid_calc ( counter, 1 ) == 3 ) .or. &
               ( param_force_energy % array_grid_calc ( counter, 1 ) == 4 ) ) then

            call compute_energy_sda_2proteins (  prot1, prot2, &
                           param_force_energy % array_grid_calc ( counter, 1 ), &
                           dummy_array, energy11, energy21  , &
                           p1_type , p2_type , test_dist2) !NJB: Add protein types and separation squared
            ! protein 1
            !! shift one for total energy
            array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 1 , 1 ) = &
              array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 1, 1 ) + energy11
            ! protein 2
            array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 1, 2 ) = &
              array_energy ( param_force_energy % array_grid_calc ( counter, 1 ) + 1, 2 ) + energy21

    end if

   end subroutine energy_epedhdlj_2proteins_fast

   !> Compute 2 energies terms at the same time, use the same "charges"
   !!
   !! Compute electrostatic / desolvation or hydrophobic / repulsive lennard-jones
   !! @param prot1 : solute 1
   !! @param prot2 : solute 2
   !! @param nb_grid1 : grid to use for interaction 1
   !! @param nb_grid2 : grid to use for interaction 2
   !! @param gridtype : type of the couple of grids (0 = Ep/Des or 1 == Hd/RepLj )
   !! @param dummy_array : : used for the computation of repulsive lennard-jones interaction
   !! @param energy11 : energy of protein 1 in the grid of protein 2 ( with type grid 1 )
   !! @param energy12 : energy of protein 1 in the grid of protein 2 ( with type grid 2 )
   !! @param energy21 : energy of protein 2 in the grid of protein 1 ( with type grid 1 )
   !! @param energy22 : energy of protein 2 in the grid of protein 1 ( with type grid 2 )
   !! @param type_p1, type_p2 : integer type of the solutes
   !! @param test_dist2 : distance square between the solutes
   subroutine compute_energy_couple ( prot1, prot2, nb_grid1, nb_grid2, gridtype, dummy_array, &
                                      energy11, energy12, energy21, energy22 , &
                                      type_p1, type_p2 , test_dist2) !NJB: Add type and separation squared
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    interface
        subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
        end subroutine tr
    end interface

    type ( protein ), intent( in ) :: prot1, prot2
    real ( kind=8 ), dimension ( : ), intent ( in ), target :: dummy_array
    real ( kind=8 ), intent ( out ) :: energy11, energy12, energy21, energy22
    integer, intent ( in ) :: nb_grid1, nb_grid2, gridtype

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2, p2_charge1, p2_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_pos_charge1, p_pos_charge2
    real ( kind=8 ), dimension(3) :: rto, rtn, pos_relat, ex, ey, ez
    real ( kind=8 ) :: potential
    integer :: nq1, nq2, j

    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: relevent distance to test for cut off 
    real (kind=4) , intent (in) :: test_dist2

    ! case El. and Desolv.
    if ( gridtype == 1 ) then
        nq1 = prot1 % nb_charge
        nq2 = prot2 % nb_charge
        p_pos_charge1 => prot1 % p_pos_charge
        p_pos_charge2 => prot2 % p_pos_charge
        p_charge1 => prot1 % p_sogrid % charge_grid
        p_charge2 => prot2 % p_sogrid % charge_grid
        p2_charge1 => prot1 % p_sogrid % charge_square_grid
        p2_charge2 => prot2 % p_sogrid % charge_square_grid
     !! case Hyd. and lennardjones
     else
        nq1 = prot1 % nb_atom_surface
        nq2 = prot2 % nb_atom_surface
        p_pos_charge1 => prot1 % p_list_atom_surface
        p_pos_charge2 => prot2 % p_list_atom_surface
        p_charge1 => prot1 % p_sogrid % accessibility
        p_charge2 => prot2 % p_sogrid % accessibility
        p2_charge1 => dummy_array
        p2_charge2 => dummy_array
     end if

     pos_relat = prot2 % position
     ex = prot2 % orientation_x
     ey = prot2 % orientation_y
     ez = prot2 % orientation_z

     energy11 = 0.d0
     energy12 = 0.d0
     energy21 = 0.d0
     energy22 = 0.d0

!write(0,*) "ISCOUPLE"

     ! NJB: If distance less than both P2 cut offs, calc both 
     if ((test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid1,type_p1)).AND. &
        (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid2,type_p1))) then

        !! energy of charge of protein1 in grid of protein2
        do j = 1, nq1
           rto =  - pos_relat + p_pos_charge1 ( :, j )
           call tr (rto, rtn, ex, ey, ez )
           call potential_atom_uhbd ( rtn, potential, prot2 % p_sogrid % pset_grid ( nb_grid1 )  )
           energy11 = energy11 + potential * p_charge1 ( j )

           call potential_atom_uhbd ( rtn, potential, prot2 % p_sogrid % pset_grid ( nb_grid2 )  )
           energy12 = energy12 + potential * p2_charge1 ( j )
        end do

     else if (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid1,type_p1)) then

        do j = 1, nq1
           rto =  - pos_relat + p_pos_charge1 ( :, j )
           call tr (rto, rtn, ex, ey, ez )
           call potential_atom_uhbd ( rtn, potential, prot2 % p_sogrid % pset_grid ( nb_grid1 )  )
           energy11 = energy11 + potential * p_charge1 ( j )
        end do

     ! NJB: If only smaller than the second calc that
     else if (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid2,type_p1)) then 

        do j = 1, nq1
           rto =  - pos_relat + p_pos_charge1 ( :, j )
           call tr (rto, rtn, ex, ey, ez )
           call potential_atom_uhbd ( rtn, potential, prot2 % p_sogrid % pset_grid ( nb_grid2 )  )
           energy12 = energy12 + potential * p2_charge1 ( j )
        end do

     end if

     !! energy of charge of protein2 in grid of protein1
     ! NJB: If distance less than both P1 cut offs, calc both
     if ((test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid1,type_p2)).AND. &
         (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid2,type_p2))) then 
     
        do j = 1, nq2
           rto = pos_relat + p_pos_charge2 ( :, j )

           call potential_atom_uhbd ( rto, potential, prot1 % p_sogrid % pset_grid ( nb_grid1 )  )
           energy21 = energy21 + potential * p_charge2 ( j )

           call potential_atom_uhbd ( rto, potential, prot1 % p_sogrid % pset_grid ( nb_grid2 )  )
           energy22 = energy22 + potential * p2_charge2 ( j )
        end do

     ! NJB: Else if only smaller than the first calc that
     else if (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid1,type_p2)) then
        do j = 1, nq2
           rto = pos_relat + p_pos_charge2 ( :, j )

           call potential_atom_uhbd ( rto, potential, prot1 % p_sogrid % pset_grid ( nb_grid1 )  )
           energy21 = energy21 + potential * p_charge2 ( j )
        end do

     ! NJB: Else if only smaller than the second calc that
     else if (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid2,type_p2)) then

       do j = 1, nq2
           rto = pos_relat + p_pos_charge2 ( :, j )

          call potential_atom_uhbd ( rto, potential, prot1 % p_sogrid % pset_grid ( nb_grid2 )  )
           energy22 = energy22 + potential * p2_charge2 ( j )
        end do

     end if

end subroutine compute_energy_couple


END MODULE mod_compute_energy
