!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_energy_flex.f90
!! used for evaluating energy when there is a change of conformation

!> Compute energy when changing conformation for sda_2proteins.\n
!! \todo maybe need to add the versions fast (really necessary ?)
MODULE mod_compute_energy_flex

contains

! made very similar to energy_sdamm ( not fast )

!deal simply with asymetry, param_force_energy not fully extended to sdamm
!\todo not working with lennard-jones need to add
!interface must be similar in both cases to send the function as a parameter

!> Version for sda_flex, evaluate energy of prot in the grids of tab_prot\n
!!
!! group same function for sda_2protein and sdamm, maybe loose efficiency\n
!! maybe slower but works for both ( and no problem, hydro/lennardjones )
!!
!! add parameter energy_img_chg, compute energy apart used only for the tests
!! include Debye-Huckel interactions
!!
!! @param prot : protein to which we want to evaluate the interaction energy
!! @param tab_prot : all other proteins with which prot interacts
!! @param geom : instance of geometry
!! @param nb_prot : size of array_protein
!! @param param_force_energy : instance of param_force_energy in mod_apram_force_energy
!! @param param_metal : instance of param_metal_desolv in com_sda
!! @param dummy_array : filled only with 1.0d
!! @param energy_img_chg : specific to image-charge and metaldesolv (included for unit-tests only)
!! @param energy : double value of the total energy of interaction prot with all other proteins
subroutine energy_epedhdlj_flex ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                                  dummy_array, energy, energy_img_chg )

      USE mod_force_energy
      USE mod_array_protein
      USE mod_geometry
      USE mod_compute_image_charge
      USE mod_compute_metaldesolv
      USE mod_compute_analytic_2proteins
      USE mod_compute_analytic_sdamm
      USE com_sda
      IMPLICIT NONE

      interface

          subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
        end subroutine compute_dist_pbc

      end interface

      type ( protein ), intent( in ), target :: prot
      type ( array_protein_type ), intent( in ), target :: tab_prot
      type ( geometry ), intent( in ) :: geom
      type ( type_force_energy ), intent ( in ) :: param_force_energy
      type ( parameter_metaldesolv ), intent ( in ) :: param_metal
      real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
      real ( kind=8 ) :: energy, energy_img_chg, energy_met_des
      integer, intent ( in ) :: nb_prot

      type ( protein ), pointer :: pprot1, pprot2
      real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
      real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2
      type ( type_grid ), pointer :: p_grid1, p_grid2

      ! tmp orientation, need inversion for forces on protein1
      real ( kind=8 ), dimension(3) :: rto, rtn, pos_relat
      ! q for q1 and q2
      real ( kind=8 ) :: potential
      real ( kind=4 ) :: dist, test_dist2
      integer :: n2, nq1, nq2, i, j, k, grid_number
      ! added for DH, mainly tests can sum up the common energy term
      real ( kind=8 ) :: energy_dh1, energy_dh2, energy_analytic_dh1, energy_analytic_dh2
      integer :: type_p1, type_p2 ! used also for grid based cut. Will be reset in DH loop

      ! here, just want the total energy, do not split into terms
      energy = 0.d0
      energy_img_chg = 0.d0
      ! to delete, to include in energy_img_chg
      energy_met_des = 0.d0
      ! to delete, for test only
      energy_dh1 = 0.d0
      energy_dh2 = 0.d0
      energy_analytic_dh1 = 0.d0
      energy_analytic_dh2 = 0.d0
      dist = 0.

      !write(*,*) 'Energy flex nb_prot ', nb_prot
      !write(*,*) 'img_chg ', prot % iimg_chg
      !write(*,*) 'size array_other_prot ', size ( tab_prot % array )
      !write(*,*) 'param nb_grid ', param_force_energy % nb_grid

! protein 1 is fixed
      pprot1 => prot

! loop only over second protein
PROT2:  do  n2 = 1, nb_prot

            !write(*,*) 'fill array_energy n2 nb_prot', n2, nb_prot
            pprot2 => tab_prot % array ( n2 )

            ! check they are different proteins
            if ( associated ( pprot1, pprot2 ) ) then
#ifdef VERBOSE_ENERGY_FLEX
                write(*,*) 'same protein in mod_compute_energy_flex:2proteins discard '
#endif
                CYCLE PROT2
            end if

            call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
            !write(*,*) 'After compute dist pbc '
            !write (*,*) 'after dist, pos relat ',  dist, pos_relat, dot_product ( pos_relat, pos_relat )
            test_dist2 = dist*dist

            if (test_dist2.GT.max(pprot1%p_sogrid%max_cut2,pprot2%p_sogrid%max_cut2)) then
                cycle PROT2
            end if

            ! NJB: Find type of both proteins
            ! Trivial for 2 protein
            type_p1 = param_force_energy%array_prot_type(pprot1%id)
            type_p2 = param_force_energy%array_prot_type(pprot2%id)

      GRID: do i = 1, param_force_energy % nb_grid

               ! should use this function, need to test with lj ?
               grid_number =  param_force_energy % array_grid_calc ( i, 1 )
               !write(*,*) 'grid_number ', grid_number

               ! modify for lennard-jones, may create a fast version , easier with pointer function
               do k = 1,  param_force_energy % array_grid_calc ( i, 3 )
                  call select_charge2 ( grid_number, pprot1, pprot2, nq1, nq2, p_charge1, p_charge2, &
                               p_charge_position1, p_charge_position2, dummy_array, &
                               p_grid1, p_grid2, k )

                  ! test if asymetric
                  if ( ( nq2 > 0 ) .and. (associated ( p_grid1 ) ) ) then
                  !write(*,*) 'on protein2'

                     !NJB: Test cut off for charges of P2 in grid of P1
                     if (test_dist2.LT.pprot1%p_sogrid%cut2_array(grid_number,type_p2)) then

                ENERG2: do j = 1, nq2
                           rto = pos_relat + p_charge_position2 (:,j)
                           call tr (rto, rtn, pprot1 % orientation_x, pprot1 % orientation_y, &
                               pprot1 % orientation_z )
                           call potential_atom_uhbd( rtn, potential, p_grid1 ) !pprot1 % p_sogrid % pset_grid ( grid_number ) )
                           ! case of electrostatic, should be the same than q2 in E1.
                           energy = energy + potential * p_charge2(j)
                        end do ENERG2
                     end if
                  end if

                  if ( ( nq1 > 0 ) .and. ( associated ( p_grid2 ) ) ) then

                     !NJB: Test cut off for charges of P1 in grid of P2
                     if (test_dist2.LT.pprot2%p_sogrid%cut2_array(grid_number,type_p1)) then

                        ! energy1 is the energy of the charge of protein1 in the grid of protein2
                        ! sdamm both protein rotate, need to change referentiel
                        ! energy is invariant, do not need to come back to the original frame
                ENERG1: do j = 1 , nq1
                           rto = -pos_relat + p_charge_position1(:,j)
                           call tr (rto, rtn, pprot2 % orientation_x, pprot2 % orientation_y, &
                                 pprot2 % orientation_z )
                           call potential_atom_uhbd( rtn, potential, p_grid2 )
                           energy = energy + potential * p_charge1 ( j )

                        end do ENERG1
                     end if
                  end if

               end do ! endo subgrids

            end do GRID

        end do PROT2

   ! check if image charge must be computed for this protein
   if ( ( param_force_energy % iimg_chg ) .and. ( prot % iimg_chg ) ) then

       call energy_image_charge_one_protein ( prot, param_force_energy, 1, &
                         tab_prot % array ( param_force_energy % surface_id ) % p_sogrid % distance_max, &
                         energy_img_chg )
       energy = energy + energy_img_chg
   end if

   ! add if metaldesolv, should use same array_iamge charge
   if ( ( param_force_energy % imetaldesolv ) .and. ( prot % imetaldesolv ) ) then
       !write(*,*) 'energy flex, ask for metal desolvation '
       !write(*,*) 'param_metal ', param_metal % Zadw
       !write(*,*) 'id surface ', param_force_energy % surface_id
       !write(*,*) 'prot id ', prot % id
       call compute_energy_metaldesolv_1prot( prot, tab_prot % array( param_force_energy % surface_id ), &
                                               param_metal, energy_met_des )
        !write(*,*) 'energy metal desolv ', energy_met_des
        ! used mainly for unit-testing, need to add before and then dividing by two, see test_force_energy
        energy_img_chg = energy_img_chg + energy_met_des
        energy = energy + energy_met_des
   end if

   ! new added DH interaction
   if ( param_force_energy % ianalytic ) then

       if ( param_force_energy % idebye ) then
           ! compute and add the energy to the electrostatic energy
           ! write(*,*) 'normal electrostatic energy ', array_energy ( 2, 1 ), array_energy (2, 2)
           ! normal call from MainLoop
           !call energy_analytic_dh ( ptab_protein, param_force_energy, dist, array_energy )
#ifdef VERBOSE_ENERGY_FLEX
           write(*,*) 'call analytic debyeh general in energy flex 2proteins'
#endif

           ! use same trick than in normal uhbd grid, works for sda_2proteins
           ! and sdamm
PROT2_DH:  do n2 = 1, nb_prot

               pprot2 => tab_prot % array ( n2 )

               ! check they are different proteins
               if ( associated ( pprot1, pprot2 ) ) then
#ifdef VERBOSE_ENERGY_FLEX
                   write(*,*) 'same protein discard '
#endif
                   CYCLE PROT2_DH
               end if

               ! now always pos2 - pos1
               !call position_pbc ( geom, pprot1 % position, pprot2 % position, pos_relat )
               !write (*,*) 'pos relat ', ( pos_relat( k ), k=1,3 )
               !dist = real ( sqrt ( dot_product( pos_relat, pos_relat ) ), 4 )
               call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
               !write(*,*) 'After compute dist pbc '
               !write (*,*) 'after dist, pos relat ',  dist, pos_relat, dot_product ( pos_relat, pos_relat )


               !write(*,*) 'pprot1 id ', pprot1 % id
               !write(*,*) 'pprot2 id ', pprot2 % id
               type_p1 = param_force_energy % array_prot_type ( pprot1 % id )
               type_p2 = param_force_energy % array_prot_type ( pprot2 % id )
               !write(*,*) 'type_p1 ', type_p1
               !write(*,*) 'type_p2 ', type_p2
               call pair_energy_full_analytic_2proteins( pprot1, pprot2, type_p1, type_p2, dist, &
                        energy_analytic_dh1, energy_analytic_dh2 )
#ifdef VERBOSE_ENERGY_FLEX
               write(*,*) 'After energy full analytic'
               write(*,*) 'energy_analytic_dh1 ', energy_analytic_dh1
               write(*,*) 'energy_analytic_dh2 ', energy_analytic_dh2
#endif

               call pair_energy_partial_analytic( pprot1, pprot2, type_p1, type_p2, dist, energy_dh1, energy_dh2 )
#ifdef VERBOSE_ENERGY_FLEX
               write(*,*) 'After energy partial analytic'
               write(*,*) 'energy_dh1 ', energy_dh1
               write(*,*) 'energy_dh2 ', energy_dh2
#endif

               ! should update energy here, all terms ?
               energy = energy + energy_analytic_dh1 + energy_analytic_dh2 + &
                        energy_dh1 + energy_dh2

           end do PROT2_DH

#ifdef VERBOSE_ENERGY_FLEX
           write(*,*) 'after all energy flex 2proteins energy= ', energy
#endif

       end if
   endif

end subroutine energy_epedhdlj_flex

!> For sdamm simulation, similar but only difference with use of OpenMP.\n
!!
!! @param prot : protein to which we want to evaluate the interaction energy
!! @param tab_prot : all other proteins with which prot interacts
!! @param geom : instance of geometry
!! @param nb_prot : size of array_protein
!! @param param_force_energy : instance of param_force_energy in mod_apram_force_energy
!! @param param_metal : instance of param_metal_desolv in com_sda
!! @param dummy_array : filled only with 1.0d
!! @param energy : double value of the total energy of interaction prot with all other proteins
!! @param energy_img_chg : specific to image-charge and metaldesolv (included for unit-tests only)
subroutine energy_epedhdlj_flex_sdamm ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                                        dummy_array, energy, energy_img_chg )

      USE mod_force_energy
      USE mod_array_protein
      USE mod_geometry
      USE mod_compute_image_charge
      USE mod_compute_metaldesolv
      !USE mod_compute_analytic_2proteins
      USE mod_compute_analytic_sdamm
      USE com_sda
!$    USE omp_lib
      IMPLICIT NONE

      type ( protein ), intent( in ), target :: prot
      type ( array_protein_type ), intent( in ), target :: tab_prot
      type ( geometry ), intent( in ) :: geom
      type ( type_force_energy ), intent ( in ) :: param_force_energy
      type ( parameter_metaldesolv ), intent ( in ) :: param_metal
      real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
      real ( kind=8 ) :: energy, energy_img_chg
      integer, intent ( in ) :: nb_prot

      type ( protein ), pointer :: pprot1, pprot2
      real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
      real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2
      type ( type_grid ), pointer :: p_grid1, p_grid2

      !! tmp orientation, need inversion for forces on protein1
      real ( kind=8 ), dimension(3) :: rto, rtn, pos_relat
      !! q for q1 and q2
      real ( kind=8 ) :: potential, energy_met_des !, energy_img_chg !, q
      real ( kind=4 ) :: dist, test_dist2
      integer :: n2, nq1, nq2, i, j, k, grid_number
      ! added for DH, mainly tests can sum up the common energy term
      real ( kind=8 ) :: energy_dh1, energy_dh2, energy_analytic_dh1,energy_analytic_dh2
      integer :: type_p1, type_p2

      !! here, just want the total energy, do not split into terms
      energy = 0.d0
      energy_img_chg = 0.d0
      energy_met_des = 0.d0
      ! to delete, for test only
      energy_dh1 = 0.d0
      energy_dh2 = 0.d0
      energy_analytic_dh1 = 0.d0
      energy_analytic_dh2 = 0.d0
      nq1 = 1
      nq2 = 1
      dist = 0

#ifdef VERBOSE_ENERGY_FLEX
      write(*,*) 'Energy flex sdamm'
#endif
!! !$    write(*,*) 'openMP activated'
      !write(*,*) 'size array_other_prot ', size ( tab_prot % array )
      !write(*,*) 'param nb_grid ', param_force_energy % nb_grid

! delete form shared nb_types, type_prot1, type_prot2, want only the sum here
! make omp loop only on the second protein
!$OMP PARALLEL DEFAULT ( NONE ) SHARED ( prot, tab_prot, nb_prot, param_force_energy, geom, dummy_array, dist ) &
!$OMP PRIVATE ( pprot1, pprot2, pos_relat, p_charge_position1, p_charge_position2, p_charge1, p_charge2, nq1, nq2 ) &
!$OMP PRIVATE ( p_grid1, p_grid2, grid_number, rto, rtn, potential , type_p1 , type_p2, test_dist2 ) &
!$OMP REDUCTION(+:energy) ! could try manual reduction like in force..not necessary for energy in sdamm

    energy = 0.d0
    ! assign to local, better ??
    pprot1 => prot
!$OMP DO SCHEDULE( DYNAMIC )
! loop only over second protein
PROT2:  do  n2 = 1, nb_prot

            !write(*,*) 'fill array_energy n2 nb_prot', n2, nb_prot
            pprot2 => tab_prot % array ( n2 )

            !! check they are different proteins
            if ( associated ( pprot1, pprot2 ) ) then
#ifdef VERBOSE_ENERGY_FLEX
                write(*,*) 'same protein discard '
#endif
                CYCLE PROT2
            end if
 
            call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
#ifdef VERBOSE_ENERGY_FLEX
            write(*,*) 'After compute dist pbc grid '
            write (*,*) 'dist/pos relat ',  dist, pos_relat, dot_product ( pos_relat, pos_relat )
#endif
            test_dist2 = dist*dist

            if (test_dist2.GT.max(pprot1%p_sogrid%max_cut2,pprot2%p_sogrid%max_cut2)) then
                cycle PROT2
            end if

            ! NJB: Find type of both proteins
            ! Trivial for 2 protein
            type_p1 = param_force_energy%array_prot_type(pprot1%id)
            type_p2 = param_force_energy%array_prot_type(pprot2%id)



      GRID: do i = 1, param_force_energy % nb_grid

               ! should use this function, need to test with lj ?
               grid_number =  param_force_energy % array_grid_calc ( i, 1 )
               !write(*,*) 'grid_number ', grid_number
               do k = 1,  param_force_energy % array_grid_calc ( i, 3 )

                  call select_charge2 ( grid_number, pprot1, pprot2, nq1, nq2, p_charge1, p_charge2, &
                              p_charge_position1, p_charge_position2, dummy_array, &
                              p_grid1, p_grid2, k )

                  ! test if asymetric
                  !if ( ( nq2 > 0 ) .and. (associated ( pprot1 % p_sogrid % pset_grid ( grid_number ) % p_grid ) ) ) then
                  if ( ( nq2 > 0 ) .and. (associated ( p_grid1 ) ) ) then
                     !write(*,*) 'on protein2'

                     !NJB: Test cut off for charges of P2 in grid of P1
                     if (test_dist2.LT.pprot1%p_sogrid%cut2_array(grid_number,type_p2)) then

                ENERG2: do j = 1, nq2
                           rto = pos_relat + p_charge_position2 (:,j)
                           call tr (rto, rtn, pprot1 % orientation_x, pprot1 % orientation_y, &
                              pprot1 % orientation_z )
                           call potential_atom_uhbd( rtn, potential, p_grid1 ) !pprot1 % p_sogrid % pset_grid ( grid_number ) )
                           !! case of electrostatic, should be the same than q2 in E1
                           energy = energy + potential * p_charge2(j)
                        end do ENERG2
                     end if
                  end if

                  if ( ( nq1 > 0 ) .and. ( associated ( p_grid2 ) ) ) then

                     !NJB: Test cut off for charges of P1 in grid of P2
                     if (test_dist2.LT.pprot2%p_sogrid%cut2_array(grid_number,type_p1)) then

                        !! energy1 is the energy of the charge of protein1 in the grid of protein2
                        !! sdamm both protein rotate, need to change referentiel
                        !! energy is invariant, do not need to come back to the original frame
                        !write(*,*) 'on protein1 nq1', nq1
                ENERG1: do j = 1 , nq1
                           rto = -pos_relat + p_charge_position1(:,j)
                           call tr (rto, rtn, pprot2 % orientation_x, pprot2 % orientation_y, &
                           pprot2 % orientation_z )
                           call potential_atom_uhbd( rtn, potential, p_grid2 )
                           energy = energy + potential * p_charge1 ( j )

                         end do ENERG1
                     end if
                  end if
               end do ! subgrids

            end do GRID

        end do PROT2
!$OMP END DO
!$OMP END PARALLEL

   ! check if image charge must be computed
   if ( ( param_force_energy % iimg_chg ) .and. ( prot % iimg_chg ) ) then

       call energy_image_charge_one_protein ( prot, param_force_energy, 1, &
                         tab_prot % array ( param_force_energy % surface_id ) % p_sogrid % distance_max, &
                         energy_img_chg )
       energy = energy + energy_img_chg
   end if

   ! add if metaldesolv, should use same array_image charge
   if ( ( param_force_energy % imetaldesolv ) .and. ( prot % imetaldesolv ) ) then
#ifdef VERBOSE_ENERGY_FLEX
       write(*,*) 'energy flex sdamm, ask for metal desolvation '
#endif
       !write(*,*) 'param_metal ', param_metal % Zadw
       !write(*,*) 'id surface ', param_force_energy % surface_id
       !write(*,*) 'prot id ', prot % id
       call compute_energy_metaldesolv_1prot( prot, tab_prot % array(  param_force_energy % surface_id ), &
                                               param_metal, energy_met_des )
#ifdef VERBOSE_ENERGY_FLEX
        write(*,*) 'energy metal desolv sdamm ', energy_met_des
#endif
        ! used mainly for testing, need to add before dividing by two
        energy_img_chg = energy_img_chg + energy_met_des
        energy = energy + energy_met_des
   end if

   ! new added DH interaction
   if ( param_force_energy % ianalytic ) then

       ! if assign local in OpenMP, need to reassign here
       pprot1 => prot
       if ( param_force_energy % idebye ) then

#ifdef VERBOSE_ENERGY_FLEX
           write(*,*) 'call analytic debyeh general in energy flex sdamm '
#endif

! could use OMP here as well
PROT2_DH:  do n2 = 1, nb_prot

               pprot2 => tab_prot % array ( n2 )

               !! check they are different proteins
               if ( associated ( pprot1, pprot2 ) ) then
#ifdef VERBOSE_ENERGY_FLEX
                   write(*,*) 'same protein discard'
#endif
                   CYCLE PROT2_DH
               end if

               ! sdamm accounts for PBC
               call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
#ifdef VERBOSE_ENERGY_FLEX
               write (*,*) 'After pbc debyeh '
               write (*,*) 'pos relat ', pos_relat
               write (*,*) 'dist ', dist
#endif

               type_p1 = param_force_energy % array_prot_type ( pprot1 % id )
               type_p2 = param_force_energy % array_prot_type ( pprot2 % id )
               !write(*,*) 'type_p1 ', type_p1
               !write(*,*) 'type_p2 ', type_p2
               ! seems identical to 2proteins versions, maybe can merge ?
               call pair_energy_full_analytic_sdamm( pprot1, pprot2, type_p1, type_p2, dist, &
                        energy_analytic_dh1, energy_analytic_dh2 )
#ifdef VERBOSE_ENERGY_FLEX
               write(*,*) 'After energy full analytic'
               write(*,*) 'energy_analytic_dh1 ', energy_analytic_dh1
               write(*,*) 'energy_analytic_dh2 ', energy_analytic_dh2
#endif

               ! only case where pos_relat is needed, could be recomputed inside !
               call pair_energy_partial_analytic_sdamm( pprot1, pprot2, pos_relat, type_p1, type_p2, dist, energy_dh1, energy_dh2 )
#ifdef VERBOSE_ENERGY_FLEX
               write(*,*) 'After energy partial analytic'
               write(*,*) 'energy_dh1 ', energy_dh1
               write(*,*) 'energy_dh2 ', energy_dh2
#endif
                
           ! should update energy here !, which term ?
           energy = energy + energy_analytic_dh1 + energy_analytic_dh2 + &
                    energy_dh1 + energy_dh2

           end do PROT2_DH

#ifdef VERBOSE_ENERGY_FLEX
           write(*,*) 'after all energy flex sdamm energy= ', energy
#endif
           !write(*,*) 'End call analytic debyeh general in energy flex sdamm '
       end if
  end if 

end subroutine energy_epedhdlj_flex_sdamm

END MODULE
