!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_energy_sdamm.f90
!! Group energy functions for sdamm

!> Module for computing energy, case sdamm
MODULE mod_compute_energy_sdamm

contains

! Extend version to case of asymetry :
! -  case nq1 or nq2 = 0, do not make the loop, no problem
! -  case grid is not associated, problem ! so check everytime
!

! version from sdamm
! here even more approximation, only protein2 in grid 1
! but loop in done from ip=1,nprot
!                       jp=1,nprot

! make backup usual order q1 in grid2, here inverse order to make as sdamm for energy
! stupid

   !> Compute multi proteins energy\n
   !! Loop over all pairs of proteins and interactions
   !! @param tab_prot : array with all solutes
   !! @param geom : instance of mod_geometry
   !! @param nb_prot : total number of solutes
   !! @param param_force_energy : instance type_force_energy in mod_force_energy
   !! @param dummy_array : used for the computation of repulsive lennard-jones interaction
   !! @param array_energy : return, to store energies
   subroutine energy_epedhdlj_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                      array_energy, dummy_array )
      USE mod_array_protein
      USE mod_geometry
      USE mod_force_energy
      ! seems not necessary ??
!!    USE omp_lib
      IMPLICIT NONE

      interface

        subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
        end subroutine compute_dist_pbc

      end interface

      type ( array_protein_type ), target :: tab_prot
      type ( geometry ) :: geom
      type ( type_force_energy ), intent ( in ) :: param_force_energy
      !real ( kind=8 ), dimension(20, nb_prot), intent(inout) :: array_energy
      real ( kind=8 ), dimension( :, :), intent ( out ) :: array_energy
      real ( kind=8 ), dimension( : ) :: dummy_array
      integer :: nb_prot

      type ( protein ), pointer :: pprot1
      type ( protein ), pointer :: pprot2
      integer :: type_prot1, type_prot2, nb_types

      real ( kind=8 ), dimension(3) :: pos_relat
      real ( kind=8 ) :: energy1, energy2
      real ( kind=4 ) :: dist, test_dist2
      integer :: n1, n2, i, k, grid_number

      ! assign shared variable
      nb_types = tab_prot % nb_grids

      !write(*,*) 'Entry energy simple sdamm '
      !write(*,*) 'nb_types ', nb_types

!$OMP PARALLEL DEFAULT ( NONE ) SHARED ( tab_prot, nb_prot, nb_types, param_force_energy, geom, dummy_array ) &
!$OMP PRIVATE ( pprot1, pprot2, pos_relat, dist, test_dist2 ) &
!$OMP PRIVATE ( grid_number, type_prot1, type_prot2, energy1, energy2 ) &
!$OMP REDUCTION(+:array_energy) ! could try manual reduction like in force..not necessary for energy in sdamm

!$OMP DO SCHEDULE( DYNAMIC )
PROT1: do n1 = 1, nb_prot
PROT2:    do  n2 = n1 + 1, nb_prot

            !write(*,*) 'fill array_energy n1 n2', n1, n2
            pprot1 => tab_prot % array ( n1 )
            pprot2 => tab_prot % array ( n2 )

            type_prot1 = param_force_energy % array_prot_type ( n1 )
            type_prot2 = param_force_energy % array_prot_type ( n2 )
            !write(*,*) 'type_prot ', type_prot1, type_prot2

            call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
            test_dist2 = dist*dist

            if (test_dist2.GT.max(pprot1%p_sogrid%max_cut2,pprot2%p_sogrid%max_cut2)) then
                cycle PROT2
            end if

! should use param_force_energy here... lennard-jones to see later, but should work also
GRID: do i = 1, param_force_energy % nb_grid

        grid_number =  param_force_energy % array_grid_calc ( i, 1 )
        !write(*,*) 'grid_number ', grid_number

        energy1 = 0.d0
        energy2 = 0.d0

        !ok if pass pointer ??
        ! new added
        do k = 1,  param_force_energy % array_grid_calc ( i, 3 )
            call compute_energy_sdamm_2proteins ( pprot1, pprot2, pos_relat, grid_number, dummy_array, &
                                                  energy1, energy2, k , &
                                                  type_prot1 , type_prot2 , test_dist2) !NJB: Add protein types

        ! now want energy by type
        array_energy ( 1 + nb_types * (grid_number-1) + type_prot2, n1 )  = &
             array_energy ( 1 + nb_types * (grid_number-1) + type_prot2, n1 ) + energy1
        ! total shift, number of full groups type_prot1 subtype of prot1
        array_energy ( 1 + nb_types * (grid_number-1) + type_prot1, n2 )  = &
             array_energy ( 1 + nb_types * (grid_number-1) + type_prot1, n2 ) + energy2
        end do

        !write(*,*) 'prot 1, index ', 1 + nb_types * (grid_number-1) + type_prot2, energy1
        !write(*,*) 'prot 2, index', 1 + nb_types * (grid_number-1) + type_prot1, energy2
    end do GRID

   end do PROT2
 end do PROT1
!$OMP END DO
!$OMP END PARALLEL
end subroutine energy_epedhdlj_sdamm

  !> Compute the energy term for one interaction for a particular pair of solutes.\n
  !!
  !! Explicit tests are done to check asymetry
  !!
  !! @param prot1 : solute 1
  !! @param prot2 : solute 2
  !! @param pos_relat : relative position between the 2 proteins
  !! @param grid_number : specify the type of interaction
  !! @param dummy_array : used for the computation of repulsive lennard-jones interaction
  !! @param energy1 : return energy of protein 1 in the grid of protein 2
  !! @param energy2 : return energy of protein 2 in the grid of protein 1
  !! @param subgrid : subgrid number (only LJ now)
  !! @param type_p1, type_p2 : integer type of the solutes
  !! @param test_dist2 : distance square between the solutes
  subroutine compute_energy_sdamm_2proteins ( prot1, prot2, pos_relat, grid_number, &
                                              dummy_array, energy1, energy2, subgrid , &
                                              type_p1, type_p2 , test_dist2) !NJB: add protein type
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    interface
          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr
    end interface

    type ( protein ), intent( in ) :: prot1, prot2
    real ( kind=8 ), dimension(3), intent ( in ) :: pos_relat
    integer, intent ( in ) :: grid_number, subgrid
    real ( kind=8 ), dimension ( : ), intent ( in ) :: dummy_array
    real ( kind=8 ), intent ( out ) :: energy1, energy2

    real ( kind=8 ), dimension(:), pointer :: p_charge1
    real ( kind=8 ), dimension(:), pointer :: p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position2
    ! new added
    type ( type_grid ), pointer :: p_grid1, p_grid2

    ! tmp orientation, need inversion for forces on protein1
    real ( kind=8 ), dimension(3) :: rto, rtn
    real ( kind=8 ) :: potential1, potential2
    integer :: nq1, nq2, j

    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: relevent distance to test for cut off 
    real (kind=4) , intent (in) :: test_dist2

    !write(*,*) 'Entry compute energy for a pair sdamm, include asymetry'
    !nq1 = 0
    !nq2 = 0

    ! need to initialize ?? YES
    energy1 = 0.d0
    energy2 = 0.d0

!    ! NJB: If either protein is a surface, find the z distance
!    if ((prot1%surface).OR.(prot2%surface)) then
!       call compute_distance(prot1,prot2,test_dist2)
!       test_dist2 = test_dist2*test_dist2
!    ! If not surface find the square distance between protein centres
!    else
!       test_dist2 = real ( dot_product(pos_relat,pos_relat), 4 )
!    end if


    call select_charge2 ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                          p_charge_position1, p_charge_position2, dummy_array, &
                          p_grid1, p_grid2, subgrid )

    ! add explicit tests, easier than an array_grid_calc for each subtype
    ! two cases for making the computation
    !if ( ( nq2 > 0 ) .and. (associated ( prot1 % p_sogrid % pset_grid ( grid_number ) % p_grid ) ) ) then
    if ( (nq2 > 0) .and. ( associated ( p_grid1) ) ) then
       !write(*,*) 'grid_number ', grid_number, ' p1 grid is associated '
       !write(*,*) 'associated ( p_grid1 ) ',  associated ( p_grid1 )
       !write(*,*) 'associated ( p_charge_position2 ) ',  associated (p_charge_position2  )
       !write(*,*) 'associated ( p_charge2 ) ',  associated ( p_charge2 )
       !write(*,*) 'nq2 ', nq2

       !NJB: Check cut off distance for interaction
       if (test_dist2.LT.prot1%p_sogrid%cut2_array(grid_number,type_p2)) then
          ! energy1 is the energy of the charge of protein2 in the grid of protein1
          ! second array
  ENERG2: do j = 1,nq2

             rto = pos_relat + p_charge_position2 (:,j)
             call tr (rto, rtn, prot1 % orientation_x, prot1 % orientation_y, &
                      prot1 % orientation_z )

             call potential_atom_uhbd( rtn, potential1, p_grid1 ) !prot1 % p_sogrid % pset_grid ( grid_number) )
             !write(*,*) 'potential 1', potential1
             energy2 = energy2 + potential1 * p_charge2 ( j )
          end do ENERG2
       end if

    endif

    if ( ( nq1 > 0 ) .and. ( associated ( p_grid2) ) ) then
       !write(*,*) 'grid_number ', grid_number, ' p2 is associated '
       !write(*,*) 'associated ( p_grid2 ) ',  associated ( p_grid2 )
       !write(*,*) 'associated ( p_charge_position1 ) ',  associated (p_charge_position1  )
       !write(*,*) 'associated ( p_charge1 ) ',  associated ( p_charge1 )
       !write(*,*) 'nq1 ', nq1

       !NJB: Check cut off distance for interaction
       if (test_dist2.LT.prot2%p_sogrid%cut2_array(grid_number,type_p1)) then


  ENERG1: do j = 1,nq1

             rto = -pos_relat + p_charge_position1(:,j)
             call tr (rto, rtn, prot2 % orientation_x, prot2 % orientation_y, &
                      prot2 % orientation_z )
             call potential_atom_uhbd( rtn, potential2, p_grid2 )
             !write(*,*) 'potential2 ', potential2
             energy1 = energy1 + potential2 * p_charge1( j )

          end do ENERG1
       end if
     endif

end subroutine compute_energy_sdamm_2proteins

!!!!!!!!!!!!!!!!!!!!!  version fast     !!!!!!!!!!!!!!!!!!!!
! modify output, split over different types

!> Compute the energy for sdamm simualtions with the fast algorithm
!!
!! If couple_grid is not used, asymetry is treated. ( see limitation with 2 asymetric grids )
!!
!! Include 2 cut-offs, one between proteins ( use distance ) and one between a surface and the proteins ( use z-distance )\n
!! Same than fast for force, delete the loop over the grids and compute rotation of "charge" only once\n
!! @param tab_prot : array with all solutes
!! @param geom : instance of mod_geometry
!! @param nb_prot : total number of solutes
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param dummy_array : used for the computation of repulsive lennard-jones interaction
!! @param array_energy : return, to record energies
subroutine  energy_epedhdlj_sdamm_fast ( tab_prot, geom, nb_prot, param_force_energy, &
                                         array_energy, dummy_array )
    USE mod_array_protein
    USE mod_geometry
    USE mod_force_energy
!$  USE omp_lib
    IMPLICIT NONE

    interface

        subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
        end subroutine compute_dist_pbc

    end interface

    type ( array_protein_type ), target :: tab_prot
    type ( geometry ), intent ( in ) :: geom
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension( :, :), intent ( out ) :: array_energy
    real ( kind = 8 ), dimension ( : ), intent ( in ) :: dummy_array
    integer, intent ( in ) :: nb_prot

    type ( protein ), pointer :: pprot1, pprot2
    real ( kind=8 ) :: energy11, energy12, energy21, energy22
    real ( kind=8 ), dimension(3) :: pos_relat
    real ( kind=4 ) :: dist, test_dist2
    integer :: counter, n1, n2, nb_types, type_prot1, type_prot2, shift

    array_energy = 0.d0

    !write(*,*) 'compute energy fast cut2 cut2_surf ',Cut2, Cut2_surf
    !write(*,*) 'nb_prot ', nb_prot
    !write(*,*) 'nb_types, tab_prot nb_grids ', tab_prot % nb_grids

    nb_types = tab_prot % nb_grids

!$OMP PARALLEL DEFAULT ( NONE ) SHARED ( tab_prot, nb_prot, nb_types, param_force_energy, geom, dummy_array ) &
!$OMP PRIVATE ( pprot1, pprot2, pos_relat, dist , test_dist2) &
!$OMP PRIVATE ( counter, shift, type_prot1, type_prot2, energy11, energy12, energy21, energy22 ) &
!$OMP REDUCTION(+:array_energy) !! could try manual reduction like in force..not necessary for energy in sdamm

!$OMP DO SCHEDULE( DYNAMIC )
PROT1: &
    do n1 = 1, nb_prot
      pprot1 => tab_prot % array ( n1 )
!      write(*,*) 'prot 1: ',n1

PROT2: do  n2 = n1 + 1, nb_prot
        !write(*,*) 'fill array_energy n1 n2', n1, n2
        pprot2 => tab_prot % array ( n2 )

        call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
        test_dist2 = dist*dist

        if (test_dist2.GT.max(pprot1%p_sogrid%max_cut2,pprot2%p_sogrid%max_cut2)) then
            cycle PROT2
        end if

        ! now need the type
        type_prot1 = param_force_energy % array_prot_type ( n1 )
        type_prot2 = param_force_energy % array_prot_type ( n2 )

        ! see force for other algorithm
        counter = 1
!        write(*,*) '    prot2: ', n2

        ! always exist ?? if fast computation should be checked
        shift = 1 + ( param_force_energy % array_grid_calc ( 1, 1 ) - 1 ) * nb_types

        ! is true if : ( electrostatic and desolvation ) AND (both are symetric)
        if ( param_force_energy % couple_ElDes ) then

         !write(*,*) 'compute couple el fast'
         call compute_energy_Couple_sdamm_fast ( pprot1, pprot2, pos_relat, &
                 !param_force_energy % array_grid_calc ( 2, 1 ), param_force_energy % array_grid_calc ( 1, 1 ), &
                 !! here sure 1 and 2
                 param_force_energy % array_grid_calc ( 1, 1 ), param_force_energy % array_grid_calc ( 2, 1 ), &
                 1, dummy_array, energy11, energy12, energy21, energy22 , &
                 type_prot1 , type_prot2 , test_dist2) !NJB: Add protein types and separation squared

          ! elect first
          array_energy ( 1 + type_prot2, n1 ) = array_energy ( 1 + type_prot2, n1 ) + energy11
          array_energy ( 1 + type_prot1, n2 ) = array_energy ( 1 + type_prot1, n2 ) + energy21
          ! desol shift of nb_types * 1
          array_energy ( 1 + nb_types + type_prot2, n1 ) = array_energy ( 1 + nb_types + type_prot2, n1 ) + energy12
          array_energy ( 1 + nb_types + type_prot1, n2 ) = array_energy ( 1 + nb_types + type_prot1, n2 ) + energy22

          ! in this case next counter will be 3
          counter = 3

        ! if electro and desolv are asymetric will work for sdamm, but here crash
        ! no much difference for energy, much more ? for forces ? to test
        !
        ! should test type_of_grid: array_grid_calc ( 1, 2 ), cannot here for each pair
        ! shoud test both if 1 then test if 2, because if 1 is asymetric, couple will be false
        ! equivalent to grid number
        else if ( ( param_force_energy % array_grid_calc ( 1, 1 ) == 1 ) .or. &
                  ( param_force_energy % array_grid_calc ( 1, 1 ) == 2 ) ) then

          !write(*,*) 'compute one grid el or ed grid number ', param_force_energy % array_grid_calc ( 1, 1 )
          call compute_energy_sdamm_2proteins (  pprot1, pprot2, pos_relat, &
                              param_force_energy % array_grid_calc ( 1, 1 ), &
                              dummy_array, energy11, energy21, 1 , &
                              type_prot1 , type_prot2 , test_dist2 ) !NJB: add protein types and separation squared

          !write(*,*) 'energy11, energy21 ',energy11, energy21
          !write(*,*) 'assign to array 1 ',  shift + type_prot2
          !write(*,*) 'assign to array 2 ',  shift + type_prot1

          ! fine if asymetric, one energy will be 0
          array_energy ( shift + type_prot2, n1 ) = array_energy ( shift + type_prot2, n1 ) + energy11
          array_energy ( shift + type_prot1, n2 ) = array_energy ( shift + type_prot1, n2 ) + energy21

          ! in this case, next counter will be 2
          counter = 2
        end if

        !need to test size, before continuing ??
        if ( param_force_energy % nb_grid < counter ) then
          !write(*,*) 'in energy no more grid return '
          GOTO 11 !CYCLE PROT2 ?? yes certainly
        end if

        ! shift apply to both cases
        ! shift for hydrophobic, shifted by (grid_number-1)*nb_type, for rep-lenjones grid_number*nb_type
        shift = 1 + ( param_force_energy % array_grid_calc ( counter, 1 ) - 1 ) * nb_types

        !! test only hydrophobic and lj-rep
        if ( param_force_energy % couple_ApDesLJrep ) then

          call compute_energy_Couple_sdamm_fast ( pprot1, pprot2, pos_relat, &
                 !! here could set 3 and 4
                 param_force_energy % array_grid_calc ( counter, 1 ), param_force_energy % array_grid_calc ( counter + 1, 1 ), &
                 2, dummy_array, energy11, energy12, energy21, energy22 , &
                 type_prot1 , type_prot2 , test_dist2) !NJB: Add protein types and separation squared

          ! fill array, protein 1
          array_energy ( shift + type_prot2 , n1 ) = array_energy ( shift + type_prot2 , n1 ) + energy11
          array_energy ( shift + nb_types + type_prot2, n1 ) = &
            array_energy ( shift + nb_types + type_prot2, n1 ) + energy12
          ! fill array prot 2
          array_energy ( shift + type_prot1, n2 ) = array_energy ( shift + type_prot1, n2 ) + energy21
          array_energy ( shift + nb_types + type_prot1, n2 ) = &
            array_energy ( shift + nb_types + type_prot1, n2 ) + energy22

          !! may update if needed for more  grids
          !counter = counter + 2

        ! else only one the grid to compute, shift is still valid
        else if ( ( param_force_energy % array_grid_calc ( counter, 1 ) == 3 ) .or. &
                  ( param_force_energy % array_grid_calc ( counter, 1 ) == 4 ) ) then

          call compute_energy_sdamm_2proteins (  pprot1, pprot2, pos_relat, &
                           param_force_energy % array_grid_calc ( counter, 1 ), &
                           dummy_array, energy11, energy21, 1 , &
                           type_prot1 , type_prot2 , test_dist2) !NJB: Add protein types and separatiion squared

          !write(*,*) 'hd or lj-rep ', energy11, energy21
          !write(*,*) 'will fill array ', param_force_energy % array_grid_calc ( counter, 1 ) + 1
          !write(*,*) '       or array ', param_force_energy % array_grid_calc ( counter, 2 ) + 1


          !! shift one for total energy
          array_energy ( shift + type_prot2, n1 ) = array_energy ( shift + type_prot2, n1 ) + energy11
          array_energy ( shift + type_prot1, n2 ) = array_energy ( shift + type_prot1, n2 ) + energy21
        end if

11   CONTINUE

   end do PROT2
 end do PROT1
 !$OMP END DO
 !$OMP END PARALLEL

 !stop
 end subroutine energy_epedhdlj_sdamm_fast

 !> Compute a couple of interaction at once\n
 !!
 !! Can compute electrostatic/desolvation or hydrophobic/ repulsive lennard-jones at the same time
 !! Need all symetric interaction.
 !!
 !! @param prot1 : solute 1
 !! @param prot2 : solute 2
 !! @param pos_relat : position relative after PBC
 !! @param nb_grid1 : grid to use for interaction 1
 !! @param nb_grid2 : grid to use for interaction 2
 !! @param gridtype : type of the couple of grids (1 = Ep/Des or 2 == Hd/RepLj )
 !! @param dummy_array : : used for the computation of repulsive lennard-jones interaction
 !! @param energy11 : energy of protein 1 in the grid of protein 2 ( with type grid 1 )
 !! @param energy12 : energy of protein 1 in the grid of protein 2 ( with type grid 2 )
 !! @param energy21 : energy of protein 2 in the grid of protein 1 ( with type grid 1 )
 !! @param energy22 : energy of protein 2 in the grid of protein 1 ( with type grid 2 )
 !! @param type_p1, type_p2 : integer type of the solutes
 !! @param test_dist2 : distance square between the solutes
 subroutine compute_energy_Couple_sdamm_fast ( prot1, prot2, pos_relat, nb_grid1, nb_grid2, &
                                 gridtype, dummy_array, energy11, energy12, energy21, energy22 , &
                                 type_p1, type_p2 , test_dist2) ! NJB: Add protein types and separation squared
     USE mod_protein
     USE mod_force_energy
     IMPLICIT NONE
     type ( protein ), intent( in ) :: prot1, prot2
     real ( kind=8 ), dimension ( 3 ), intent ( in ) :: pos_relat
     real ( kind=8 ), dimension ( : ), intent ( in ), target :: dummy_array
     real ( kind=8 ), intent ( out ) :: energy11, energy12, energy21, energy22
     integer, intent ( in ) :: nb_grid1, nb_grid2, gridtype

     real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2, p2_charge1, p2_charge2
     real ( kind=8 ), dimension(:,:), pointer :: p_pos_charge1, p_pos_charge2
     real ( kind=8 ), dimension(3) :: rto, rtn,ex, ey, ez
     real ( kind=8 ) :: potential
     integer :: nq1, nq2, j

     ! NJB: Protein types
     integer , intent (in) :: type_p1, type_p2
     ! NJB: relevent distance to test for cut off 
     real (kind=4) , intent (in) :: test_dist2

     !write(*,*) 'entry compute couple energy'
     !write(*,*) 'grid number ', nb_grid1, nb_grid2
     !write(*,*) 'gridtype ', gridtype

!     ! NJB: If either protein is a surface, set surf_int and find the z distance
!     if ((prot1%surface).OR.(prot2%surface)) then
!        call compute_distance(prot1,prot2,test_dist2)
!        test_dist2 = test_dist2*test_dist2
!     ! If not surface find the square distance between protein centres
!     else
!        test_dist2 = real ( dot_product(pos_relat,pos_relat), 4 )
!     end if

     ! function select for sda_2proteins, could be used ?
     ! case El. and Desolv.
     if ( gridtype == 1 ) then
        nq1 = prot1 % nb_charge
        nq2 = prot2 % nb_charge
        p_pos_charge1 => prot1 % p_pos_charge
        p_pos_charge2 => prot2 % p_pos_charge
        p_charge1 => prot1 % p_sogrid % charge_grid
        p_charge2 => prot2 % p_sogrid % charge_grid
        p2_charge1 => prot1 % p_sogrid % charge_square_grid
        p2_charge2 => prot2 % p_sogrid % charge_square_grid
     !! case Hyd. and LJ-repulsive
     else
        nq1 = prot1 % nb_atom_surface
        nq2 = prot2 % nb_atom_surface
        p_pos_charge1 => prot1 % p_list_atom_surface
        p_pos_charge2 => prot2 % p_list_atom_surface
        p_charge1 => prot1 % p_sogrid % accessibility
        p_charge2 => prot2 % p_sogrid % accessibility
        p2_charge1 => dummy_array
        p2_charge2 => dummy_array
     end if

     energy11 = 0.d0
     energy12 = 0.d0
     energy21 = 0.d0
     energy22 = 0.d0

     ex = prot1 % orientation_x
     ey = prot1 % orientation_y
     ez = prot1 % orientation_z

     !! energy of charge of protein2 in grid of protein1
     ! NJB: If distance less than both cut offs, calc both
     if ((test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid1,type_p2)).AND. &
        (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid2,type_p2))) then

        do j = 1, nq2

           !rto = pos_relat + prot2 % p_pos_charge ( :, j )
           rto = pos_relat + p_pos_charge2 ( :, j )
           call tr (rto, rtn, ex, ey, ez ) ! get the relative position of the atom in coord.system of prot.1

           call potential_atom_uhbd ( rtn, potential, prot1 % p_sogrid % pset_grid ( nb_grid1 )  )
           energy21 = energy21 + potential * p_charge2 ( j )

           call potential_atom_uhbd ( rtn, potential, prot1 % p_sogrid % pset_grid ( nb_grid2 )  )
           energy22 = energy22 + potential * p2_charge2 ( j )
        end do

     ! NJB: If only smaller than the first calc that 
     else if (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid1,type_p2)) then

        do j = 1, nq2

           !rto = pos_relat + prot2 % p_pos_charge ( :, j )
           rto = pos_relat + p_pos_charge2 ( :, j )
           call tr (rto, rtn, ex, ey, ez ) ! get the relative position of the atom in coord.system of prot.1

           call potential_atom_uhbd ( rtn, potential, prot1 % p_sogrid % pset_grid ( nb_grid1 )  )
           energy21 = energy21 + potential * p_charge2 ( j )

        end do

     ! NJB: If only smaller than the second calc that 
     else if (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid2,type_p2)) then

        do j = 1, nq2

           !rto = pos_relat + prot2 % p_pos_charge ( :, j )
           rto = pos_relat + p_pos_charge2 ( :, j )
           call tr (rto, rtn, ex, ey, ez ) ! get the relative position of the atom in coord.system of prot.1

           call potential_atom_uhbd ( rtn, potential, prot1 % p_sogrid % pset_grid ( nb_grid2 )  )
           energy22 = energy22 + potential * p2_charge2 ( j )

        end do

     end if

     ex = prot2 % orientation_x
     ey = prot2 % orientation_y
     ez = prot2 % orientation_z
     !! energy of charge of protein1 in grid of protein2
     ! NJB: If distance less than both cut offs, calc both 
     if ((test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid1,type_p1)).AND. &
         (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid2,type_p1))) then

        do j = 1, nq1
           rto =  - pos_relat + p_pos_charge1 ( :, j )
           call tr (rto, rtn, ex, ey, ez )
           call potential_atom_uhbd ( rtn, potential, prot2 % p_sogrid % pset_grid ( nb_grid1 )  )
           energy11 = energy11 + potential * p_charge1 ( j )

           call potential_atom_uhbd ( rtn, potential, prot2 % p_sogrid % pset_grid ( nb_grid2 )  )
           energy12 = energy12 + potential * p2_charge1 ( j )
        end do

     ! NJB: If only smaller than the first calc that
     else if (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid1,type_p1)) then 

        do j = 1, nq1
 
           rto =  - pos_relat + p_pos_charge1 ( :, j )
           call tr (rto, rtn, ex, ey, ez )
           call potential_atom_uhbd ( rtn, potential, prot2 % p_sogrid % pset_grid ( nb_grid1 )  )
           energy11 = energy11 + potential * p_charge1 ( j )

        end do

     ! NJB: If only smaller than the second calc that
     else if (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid2,type_p1)) then 

        do j = 1, nq1
           rto =  - pos_relat + p_pos_charge1 ( :, j )
           call tr (rto, rtn, ex, ey, ez )
           call potential_atom_uhbd ( rtn, potential, prot2 % p_sogrid % pset_grid ( nb_grid2 )  )
           energy12 = energy12 + potential * p2_charge1 ( j )
        end do
     end if

    !write(*,*) 'energy11, energy12 ',energy11, energy12
    !write(*,*) 'energy21, energy22 ',energy21, energy22

end subroutine compute_energy_Couple_sdamm_fast


END MODULE mod_compute_energy_sdamm
