!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_force.f90
!! group functions to compute force for 2 proteins

!> Module groups computation of forces and energies of sda_2proteins and related.\n
!! simplify interface if all in one file.
MODULE mod_compute_force

contains

! MIC: pX_type could be hardcoded here
! MIC: Many tests in Couple computation now. Really improve the timing ?
!      Maybe more tests to add before setting the boolean Couple
! MIC test tsurface with compute_distance neccessary ?

!> Compute the force for sda_2proteins and derivatives
!!
!! Unoptimised algorithm. Extension with new interactions are simpler to introduce here
!!
!! Add distance as parameter, for sda_2proteins if surface is present the distance is computed differently
!!
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param dist : distance between the solutes, use a cut-off
!! @param f, t, t1 : return, store the results of force and torques
!! @param dummy_array : used for repulsive lennard-jones interactions
subroutine  force_epedhdlj_2proteins ( prot1, prot2, param_force_energy, dist, f, t, t1, dummy_array )
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    type ( protein ), intent ( in ) :: prot1, prot2
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension ( : ), intent ( in ) :: dummy_array
    real ( kind=8 ), dimension ( 3 ) :: f, t, t1 !first comput could out only
    real ( kind=4 ), intent ( in ) :: dist

    real ( kind=8 ), dimension(3) :: ff1, ff2, tt1, tt2

    ! use several time protein2 % position relative, certainly in sdamm also
    integer :: i, k, grid_number

    !NJB: protein type
    integer :: p1_type , p2_type
    !NJB: square of separation used for cut off calculations
    real ( kind=4 ) :: test_dist2

    ff2 = 0.d0
    ff1 = 0.d0
    tt2 = 0.d0
    tt1 = 0.d0
    !! should take into account surface

    ! NJB: Find square of distance
    test_dist2 = dist*dist

    ! NJB: If protein separation greater than the cut off of the largest grid of each protein, skip
    if (test_dist2.GT.max(prot1%p_sogrid%max_cut2,prot2%p_sogrid%max_cut2)) then
!write(0,*) "CUT", (dist*dist), " larger than", prot1%p_sogrid%max_cut2, " and", prot2%p_sogrid%max_cut2
       return
    end if
!write(0,*) "NOCUT", (dist*dist), " smaller than", prot1%p_sogrid%max_cut2, " or", prot2%p_sogrid%max_cut2
 
    !write(*,*) 'Compute force 2proteins not fast '
    !write(*,*) 'param_force nb_grid ', param_force_energy % nb_grid

    ! NJB: Find type of both proteins
    ! Trivial for 2 protein
    p1_type = param_force_energy%array_prot_type(prot1%id)
    p2_type = param_force_energy%array_prot_type(prot2%id)

!write (0,*) "type p1: ", p1_type, "type p2: ", p2_type !NJB

!! loop over gird, sure to be associated
GRID: do i = 1, param_force_energy % nb_grid

    grid_number =  param_force_energy % array_grid_calc ( i, 1 )

    ! test if it is a symetric interaction
    if ( param_force_energy % array_grid_calc ( i, 2 ) == symetric ) then
        !write(*,*) 'compute a symetric interaction '
        call compute_force_sda_2proteins ( prot1, prot2, grid_number, dummy_array, &
                                           ff2, tt1, tt2 , p1_type , p2_type , test_dist2) !NJB: Add protein types

        ! just copy the values now, correction done inside the function
        f = f + ff2
        t = t + tt2
        t1 = t1 + tt1

    else
        !write(*,*) 'grid i=,', i,' number ', param_force_energy % array_grid_calc ( i, 1 )
        !write(*,*) 'compute an asymetric interaction ', param_force_energy % array_grid_calc ( i, 2 )
        !write(*,*) 'number of sub-grids ', param_force_energy % array_grid_calc ( i, 3 )

        do k = 1,  param_force_energy % array_grid_calc ( i, 3 )
          !write(*,*) 'subgrid k ', k
          call force_asymetric ( prot1, prot2, grid_number, dummy_array, ff1, ff2, tt1, tt2, &
                                 ! reverse
                                 param_force_energy % array_grid_calc ( i, 2 ), &
                                 ! subgrid
                                 k , &
                                 p1_type , p2_type , test_dist2) !NJB: Add protein types
          !! case c2g1, need to sum up forces here
          if ( param_force_energy % array_grid_calc ( i, 2 ) == c2g1 ) then
              !write(*,*) 'after asym c2g1 '
              !write(*,*) 'ff1 ', ff1
              !write(*,*) 'tt1 ', tt1
              !! normal case
              f = f + ff2
              t = t + tt2
              ! for test_force_energy need to assign something
              t1 = t1 + tt1
          !! c1g2, compute f1 and t1
          else
              !write(*,*) 'after asym c1g2 '
              ! because asymetry assign -f1 to f2
              ! t1 not modified
              ! new correctly computed inside function, both tt1 and tt2 are different of 0
              f = f -ff1
              t = t + tt2
              t1 = t1 + tt1
          end if
          !! should multiply by two the result ? to reproduce prometcs ?
        end do

    end if

!        write(*,*) 'FORCE ', i
!        write (*,'(''Test new force f: '',3E14.5 )') f
!        write (*,'(''Test new torque t: '',3E14.5 )') t
!        write (*,'(''Test new torque t1: '',3E14.5 )') t1

    end do GRID

end subroutine force_epedhdlj_2proteins



!> Compute one asymetric interaction, send back ff2, tt1, tt2, caller should decide what to do ( increment ).\n
!!
!! normal version, not fast.
!! Here aplly final torque correction, in order to reproduce identical force than other methods
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param grid_number : number of the grid, as referred in \ref mod_gridtype
!! @param dummy_array : used for repulsive lennard-jones interactions
!! @param ff1, ff2, tt2, tt1 : return force and torque on solute 1 and 2
!! @param reverse : indicate the sense of asymetry
!! @param subgrid : uhbd subgrid
!! @param type_p1, type_p2 : integer type of the solutes
!! @param test_dist2 : distance square between the solutes
subroutine force_asymetric ( prot1, prot2, grid_number, dummy_array, ff1, ff2, tt1, tt2, reverse, subgrid &
                            , type_p1, type_p2 , test_dist2)  !NJB: Add protein types
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE
    type ( protein ), intent ( in ) :: prot1, prot2
    integer, intent ( in ) :: grid_number
    real ( kind=8 ), dimension ( : ), intent ( in ) :: dummy_array
    real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ff1, ff2, tt1, tt2 !first comput could be out only
    integer :: reverse, subgrid

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2
    type ( type_grid ), pointer :: p_grid1, p_grid2

    real ( kind=8 ), dimension(3) :: tmp1, tmp2, pos_relat !,tt2_copy, tmp0
    real ( kind=8 ), dimension(3) :: x, p1, p2, qf, rto, rtn, force, force_o
    real ( kind=8 ), dimension(3) :: ex, ey, ez, xe, ye, ze
    real ( kind=8 ) :: q
    integer :: j, nq1, nq2

    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: Square of protein separation
    real ( kind=4 ) , intent (in) :: test_dist2


    nullify ( p_grid1, p_grid2 )

    ff1 = (/0., 0., 0./)
    tt1 = (/0., 0., 0./)
    ff2 = (/0., 0., 0./)
    tt2 = (/0., 0., 0./)
    tmp1 = (/0., 0., 0./)
    tmp2 = (/0., 0., 0./)
    nq1 = 0
    nq2 = 0
    pos_relat = prot2 % position

    !write(*,*) 'compute force_asymetric , reverse ', reverse
    !write(*,*) 'subgrid ', subgrid

    ! apply for all grids now
    !if ( present ( opt_subgrid ) ) then
    call select_charge2 ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                          p_charge_position1, p_charge_position2, dummy_array, &
                          ! added specifically for lennard-jones need the grid
                          ! no need to make optional, not used it is all, make optional in select_charge only
                          p_grid1, p_grid2, subgrid )

!write(0,*) "ASYMMETRIC"

    !write(*,*) 'after selection '
    ! "normal " p2 charges in grid of p1, less geometrical transformation to do
    if ( reverse == c2g1 ) then
        !write(*,*) 'case c2g1 '

       !NJB: apply cut off
       if (test_dist2.LT.prot1%p_sogrid%cut2_array(grid_number,type_p2)) then

          FORCE2: do j=1,nq2

            ! for electrostatic, force and torques divided by 2 at the end
            q = p_charge2 ( j )
            !! for hydrophobic should be list_atom_surface ! before nq2p1,ns2l
            p2 = p_charge_position2 (:,j) ! !p2 = y(:,j)
            !for force
            x = pos_relat + p2 ! x = r + p2 !y(:,j)

            call force_atom_uhbd ( x, force, p_grid1 )
            !write(*,*) 'force ', force
            qf = force * q !!qf2 = f2 * q2
            !write(*,*) 'qf electro', qf
            call cross (tmp2, p2, qf ) !call cross( ti2, p2, qf2 )
            tt2 = tt2 + tmp2 !tt2 = tt2 + ti2
            ff2 = ff2 + qf !ff2 = ff2 + qf2

          end do FORCE2

       end if 
      
       !. torque on 1    (r)x(-ff2), CALL cross(tmp2,r,ff2)
       call cross ( tmp2, pos_relat, ff2 )
       ! write(*,*) 'value tt2 ', tt2
       !write(*,*) 'ff2, tt2 ', ff2, tt2
       ! correct correction reproduce other results !
       ! write(*,*) 'value torque on protein 1 ', - tt2 - tmp2
       ! tt1 = tt1 - tt2_copy - tmp2

       !! tt1 should be zero at this point
       !tt1 = tt1 - tt2 - tmp2
       tt1 = -tt2 - tmp2

    ! inverse order p1 charges in grid of p2, need more change of referential
    else
        !write(*,*) 'case c1g2 '
        !write(*,*) 'nq1 ', nq1
        !write(*,*) 'associated p_charge1 ', associated ( p_charge1 )
        !write(*,*) 'associated p_pos_charge1 ', associated ( p_charge_position1 )
        !write(*,*) 'associated p_grid2 ', associated ( p_grid2 )

        ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
        ey = prot2 % orientation_y
        ez = prot2 % orientation_z
        xe(1) = ex(1)
        xe(2) = ey(1)
        xe(3) = ez(1)
        ye(1) = ex(2)
        ye(2) = ey(2)
        ye(3) = ez(2)
        ze(1) = ex(3)
        ze(2) = ey(3)
        ze(3) = ez(3)

       !NJB: apply cut off
       if (test_dist2.LT.prot2%p_sogrid%cut2_array(grid_number,type_p1)) then

          FORCE1: do j = 1, nq1

             p1 = p_charge_position1(:,j) ! p1 = x1(:,j)
             !write(*,*) 'p1 ', p1
             rto = - pos_relat + p1 !rto = -r + x1(:,j) !!position charge protein1
             !write(*,*) 'rto ', rto
             call tr(rto, rtn, ex, ey, ez )
             call force_atom_uhbd ( rtn, force, p_grid2 )

             call tr ( force, force_o, xe, ye, ze )
             qf = force_o * p_charge1 ( j )  !!qf1 = force_o * q1

             call cross(tmp1, p1, qf) !call cross(ti1, p1, qf1)
             tt1 = tt1 + tmp1 !!tt1 = tt1 + ti1
             ff1 = ff1 + qf !!ff1 = ff1 + qf1

          end do FORCE1
       end if
        !write(*,*) 'first tt1 ', tt1
        ! here should apply corrections on protein2
        !c. torque on 2   (-r)x(-ff1)
        call cross( tmp1, pos_relat, ff1 )
        !write(*,*) 'correction ', tmp1
        !write(*,*) 'correction torque on p2 ', -tt1  + tmp1
        ! tt2 zero in this case, but tmp1 different of zero
        !tt2 = tt2 - tt1 + tmp1
        tt2 = -tt1  + tmp1

    end if ! opt_reverse

    ! moved into correct if statements, or compute here with 0
    !stop
    ! can compute correction if needed, to add to torque
    !c. torque on 2   (-r)x(-ff1)
    !call cross( tmp1, pos_relat, ff1 )
    !write(*,*) 'correction torque on p2 ', tmp1
    !c. torque on 1    (r)x(-ff2)
    !CALL cross(tmp2,r,ff2)

end subroutine force_asymetric

!> Compute only one interaction, always symetric => applies correction
!!
!! In fact could work with asymetric if 1 nb_charge == 0
!!
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param grid_number : number of the grid, as referred in \ref mod_gridtype
!! @param dummy_array : used for repulsive lennard-jones interactions
!! @param ff2, tt2, tt1 : return force and torque on solute 2, torque on solute 1
!! @param type_p1, type_p2 : integer type of the solutes
!! @param test_dist2 : distance square between the solutes
subroutine compute_force_sda_2proteins ( prot1, prot2, grid_number, dummy_array, &
                                         ff2, tt1, tt2 , type_p1, type_p2, test_dist2) !NJB: Add protein types
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    type ( protein ), intent( in ) :: prot1, prot2
    real ( kind=8 ), dimension ( : ), intent ( in ), target :: dummy_array
    real ( kind=8 ), dimension( 3 ), intent ( out ) :: ff2, tt1, tt2
    integer, intent ( in ) :: grid_number

    interface
          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr

          subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross

    end interface

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2

    real ( kind=8 ), dimension(3) :: ff1, tmp1, tmp2, pos_relat, tt2_copy
    real ( kind=8 ), dimension(3) :: x, p1, p2, qf, rto, rtn, force, force_o
    real ( kind=8 ), dimension(3) :: ex, ey, ez, xe, ye, ze
    real ( kind=8 ) :: q
    integer :: j, nq1, nq2

    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: Square of protein separation passed in
    real ( kind=4 ) , intent (in):: test_dist2

    !real(kind=8), dimension(3,3) :: tmp_mat
    !real ( kind=8 ), dimension(3) :: tmp1_n

    ff1 = (/0.d0, 0.d0, 0.d0/)
    tt1 = (/0.d0, 0.d0, 0.d0/)
    ff2 = (/0.d0, 0.d0, 0.d0/)
    tt2 = (/0.d0, 0.d0, 0.d0/)
    tmp1 = (/0.d0, 0.d0, 0.d0/)
    tmp2 = (/0.d0, 0.d0, 0.d0/)
    nq1 = 0
    nq2 = 0
    pos_relat = prot2 % position

    !write(*,*) 'pos_relat', pos_relat
    !! assign the correct pointer
    call select_charge ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                         p_charge_position1, p_charge_position2, dummy_array )

    !NJB: Check if distance is less than relevant cut off
    if (test_dist2.LT.prot1%p_sogrid%cut2_array(grid_number,type_p2)) then
!write(0,*) "P1 NOCUT: dist ", test_dist2, "  cut", prot1%p_sogrid%cut2_array(grid_number,type_p2)

    FORCE2: do j=1,nq2

            ! for electrostatic, force and torques divided by 2 at the end
            ! here q == q2
            q = p_charge2 ( j )
            !! for hydrophobic should be list_atom_surface ! before nq2p1,ns2l
            p2 = p_charge_position2 (:,j) ! !p2 = y(:,j)
            !for force
            x = pos_relat + p2 ! x = r + p2 !y(:,j)

            call force_atom_uhbd ( x, force, prot1 % p_sogrid % pset_grid ( grid_number) )
            qf = force * q !!qf2 = f2 * q2
            !write(*,*) 'qf electro', qf
            call cross (tmp2, p2, qf )
            tt2 = tt2 + tmp2 !tt2 = tt2 + ti2
            ff2 = ff2 + qf !ff2 = ff2 + qf2

            ! ok working for asymetric, but should multiply by 2 for getting sda results
            !if ( grid_number == 2 ) then
            !    write(*,*) ' j add electro desolv.',j , qf(1)
            !    write(*,*) ' j add electro desolv.',j , qf(2)
            !    write(*,*) ' j add electro desolv.',j , qf(3)
            !end if
        end do FORCE2

!    else
!write(0,*) "P1 CUT: dist ", test_dist2, "  cut", prot1%p_sogrid%cut2_array(grid_number,type_p2)
    end if

! torques and force for the first protein
        ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
        ey = prot2 % orientation_y
        ez = prot2 % orientation_z
        xe(1) = ex(1)
        xe(2) = ey(1)
        xe(3) = ez(1)
        ye(1) = ex(2)
        ye(2) = ey(2)
        ye(3) = ez(2)
        ze(1) = ex(3)
        ze(2) = ey(3)
        ze(3) = ez(3)

        !tmp_mat (1,:) = xe
        !tmp_mat (2,:) = ye
        !tmp_mat (3,:) = ze
        !write(*,*) 'determinant ', determinant( tmp_mat )

    !NJB: Check if distance is less than relevant cut off
    if (test_dist2.LT.prot2%p_sogrid%cut2_array(grid_number,type_p1)) then
!write(0,*) "P2 NOCUT: dist ", test_dist2, "  cut", prot2%p_sogrid%cut2_array(grid_number,type_p1)

FORCE1: do j=1,nq1

           p1 = p_charge_position1(:,j) ! p1 = x1(:,j)
           rto = - pos_relat + p1 !rto = -r + x1(:,j) !!position charge protein1
           call tr(rto, rtn, ex, ey, ez )
           call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid ( grid_number) )

           call tr ( force, force_o, xe, ye, ze )

           qf = force_o * p_charge1 ( j )  !!qf1 = force_o * q1
           call cross(tmp1, p1, qf) !call cross(ti1, p1, qf1)
           tt1 = tt1 + tmp1
           ff1 = ff1 + qf

        end do FORCE1
!    else
!write(0,*) "P2 CUT: dist ", test_dist2, "  cut", prot2%p_sogrid%cut2_array(grid_number,type_p1)
    end if 

        ! torque computed on ff1 and ff2, will be asymetric
!. torque on 2   (-r)x(-ff1), CALL cross(tmp1,r,ff1)
        call cross ( tmp1, pos_relat, ff1 )
!. torque on 1    (r)x(-ff2), CALL cross(tmp2,r,ff2)
        call cross ( tmp2, pos_relat, ff2 )

        !write(*,*) 'tt1 tmp1 ',tt1, tmp1
        !write(*,*) 'tt2 tmp2 ',tt2, tmp2

        ! tmp copy to not overwritte the value
        tt2_copy = tt2
        ! should not correct if asymetric, why not ??
        ff2 = ff2 - ff1
        tt2 = tt2 - tt1 + tmp1
        tt1 = tt1 - tt2_copy - tmp2

end subroutine compute_force_sda_2proteins

!> Use fast algorithm, only for sda_2proteins and derivatives
!!
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param dist : distance between the solutes, use a cut-off
!! @param f, t, t1 : return, store the results, force and torques
!! @param dummy_array : used for repulsive lennard-jones interactions
subroutine  force_epedhdlj_2proteins_fast ( prot1, prot2, param_force_energy, dist, f, t, t1, dummy_array )
      USE mod_protein
      USE mod_force_energy
      IMPLICIT NONE

      interface

          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr

          subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross

      end interface

      type ( protein ), intent( in ) :: prot1, prot2
      type ( type_force_energy ), intent ( in ) :: param_force_energy
      real ( kind=8 ), dimension( 3 ) :: f, t, t1 !first comput could out only
      real ( kind =8 ), dimension ( : ), intent ( in ) :: dummy_array
      real ( kind=4 ), intent ( in ) :: dist

      ! force, torque and correction for protein1 and 2, force replace f1 and f2
      real ( kind=8 ), dimension( 3 ) :: ff1_o, ff1_n, ff2, tt1_o, tt1_n, tt2, tmp1, tmp2
      ! tmp orientation, need inversion for forces on protein1
      real ( kind=8 ), dimension(3) :: ex, ey, ez, xe, ye, ze
      real ( kind=8 ), dimension(3) :: pos_relat
      integer :: nq1, nq2, counter

      !NJB: protein type
      integer :: p1_type , p2_type
      !NJB: square of separation used for cut off calculations
      real ( kind=4 ) :: test_dist2


      !write(*,*) 'Entry force_epedhdlj_2proteins_fast'
      !write(*,*) 'orient protein 1 ', prot1 % orientation_x

      nq1 = 1
      nq2 = 1
      pos_relat = prot2 % position

      ! NJB: Find square of separation
      test_dist2 = dist*dist

      ! NJB: If protein separation greater than the cut off of the largest grid of each protein, skip
      if (test_dist2.GT.max(prot1%p_sogrid%max_cut2,prot2%p_sogrid%max_cut2)) then
!write(0,*) "CUT", (dist*dist), " larger than", prot1%p_sogrid%max_cut2, " and", prot2%p_sogrid%max_cut2
         return
      end if
!write(0,*) "NOCUT", (dist*dist), " smaller than", prot1%p_sogrid%max_cut2, " or", prot2%p_sogrid%max_cut2
 
      ! NJB: Find type of both proteins
      ! Trivial for 2 protein
      p1_type = param_force_energy%array_prot_type(prot1%id)
      p2_type = param_force_energy%array_prot_type(prot2%id)

      ! in declaration do not work, executed only once !
      ff1_n = 0.d0
      tt1_n = 0.d0
      ff2 = 0.d0
      tt2 = 0.d0
      ff1_o = 0.d0
      tt1_o = 0.d0

    !! force and torque of protein 1 are ALWAYS computed in referential of protein 2, need a transformation only at the end
    !! if if electrostatic and el.desolvation can be computed together
    counter = 1
    if ( param_force_energy % couple_ElDes .eqv. .true. ) then
        call force_couple_fast ( prot1, prot2, param_force_energy % array_grid_calc ( 1, 1 ), &
               param_force_energy % array_grid_calc ( 2, 1 ), &
               1, dummy_array, ff1_n, ff2, tt1_n, tt2 , p1_type , p2_type , test_dist2 ) !NJB: Add protein types and distance
               counter = 3

    !! here cannot use fast with both eletrostatic and delsovation asymetric
    !! code correct, but cannot test the method fast in this case
    else if ( ( param_force_energy % array_grid_calc ( 1, 1 ) == 1 ) .or. &
              ( param_force_energy % array_grid_calc ( 1, 1 ) == 2 ) ) then

              !write(*,*) 'compute electrostatic or desolvation'
              call force_sda_2proteins_fast (  prot1, prot2, &
                                  param_force_energy % array_grid_calc ( 1, 1 ), &
                                  dummy_array, ff1_n, ff2, tt1_n, tt2 ,  p1_type , p2_type , test_dist2) !NJB: Add protein types an distance

              counter = 2
    end if

    !need to test size, before continuing
    if ( param_force_energy % nb_grid < counter ) then
        !write(*,*) 'in energy no more grid return '
        GOTO 11
    end if

    !! test only hydrophobic, need to check counter like in energy ??
    if ( param_force_energy % array_grid_calc ( counter, 1 ) == 3 ) then
        call force_sda_2proteins_fast (  prot1, prot2, &
                           param_force_energy % array_grid_calc ( counter, 1 ), &
                           dummy_array, ff1_n, ff2, tt1_n, tt2 , p1_type , p2_type , test_dist2) !NJB: Add protein types and separation
    end if

! repulsive lennard-jones not used with sda_2proteins
!    if ( param_force_energy % couple_ApDesLJrep .eqv. .true. ) then
!        call compute_force_Couple_uhbd ( prot1, prot2, param_force_energy % array_grid_calc ( counter, 1 ), &
!               param_force_energy % array_grid_calc ( counter + 1, 1 ), &
!               3, dummy_array, ff1_n, ff2, tt1_n, tt2 )
!               counter = counter + 2
!     else if ( ( param_force_energy % array_grid_calc ( counter, 1 ) == 1 ) .or. &
!               ( param_force_energy % array_grid_calc ( counter + 1, 1 ) == 2 ) ) then
!              call compute_force_2proteins_uhbd_fast (  prot1, prot2, &
!                                  param_force_energy % array_grid_calc ( 1, 1 ), &
!                                  dummy_array, ff1_n, ff2, tt1_n, tt2  )
!              counter = 2
!     end if

11  CONTINUE

    ! need to transform back total force and total torque only at the end in this version
    ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
    ey = prot2 % orientation_y
    ez = prot2 % orientation_z
    xe(1) = ex(1)
    xe(2) = ey(1)
    xe(3) = ez(1)
    ye(1) = ex(2)
    ye(2) = ey(2)
    ye(3) = ez(2)
    ze(1) = ex(3)
    ze(2) = ey(3)
    ze(3) = ez(3)

    call tr ( ff1_n, ff1_o, xe, ye, ze )
    call tr ( tt1_n, tt1_o, xe, ye, ze )

!   torque on 2   (-r)x(-ff1)
    call cross ( tmp1, pos_relat, ff1_o )
!. torque on 1    (r)x(-ff2)
    call cross ( tmp2, pos_relat, ff2 )

    ! only difference is tt1_o, tmp1 identical between fast/normal
    !write(*,*) 'tt1 tmp1 ', tt1_o, tmp1
    ! tt2 tmp2 are equal whatever the metod fast/normal
    !write(*,*) 'tt2 tmp2 ',tt2, tmp2

    f = ff2 - ff1_o
    t = tt2 - tt1_o + tmp1
    t1 = tt1_o - tt2 - tmp2

end subroutine force_epedhdlj_2proteins_fast

!> Compute 1 interactions with the fast algorithm
!!
!! Compute force and torque on the solute 1 in the referential of the solute 2
!!
!! Normally called if symetric interaction,
!! but is working if asymetric and if 1 nb_charge == 0
!!
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param grid_number : number of the grid as referred in \ref mod_gridtype
!! @param dummy_array : used for repulsive lennard-jones interactions
!! @param ff1_n, tt1_n : return force and torques on the solute 1
!! @param  ff2, tt2 : return force and torques on the solute 2
!! @param type_p1, type_p2 : integer type of the solutes
!! @param test_dist2 : distance square between the solutes
subroutine force_sda_2proteins_fast ( prot1, prot2, grid_number, dummy_array, &
                                      ff1_n, ff2, tt1_n, tt2 , type_p1, type_p2 , test_dist2) !NJB: Add types and separation squared
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    interface
          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr

          subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross
    end interface

    type ( protein ), intent ( in ) :: prot1, prot2
    real ( kind=8 ), dimension ( : ), intent ( in ) :: dummy_array
    real ( kind=8 ), dimension( 3 ) :: ff1_n, ff2, tt1_n, tt2
    integer, intent ( in ) :: grid_number

    real ( kind=8 ), dimension (3 ) :: pos_relat
    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_pos_charge1, p_pos_charge2
    integer :: nq1, nq2, j

    real ( kind=8 ), dimension ( 3 ) :: force, x, p2, tmp1_n, rtn, qf, tmp2
    real ( kind=8 ), dimension(3) :: ex, ey, ez, pos_relat_n

    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: relevent distance to test for cut off 
    real (kind=4) , intent (in) :: test_dist2

    call select_charge ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                    p_pos_charge1, p_pos_charge2, dummy_array )

    pos_relat = prot2 % position

!write (0,*) "FAST NOCOUPLE" , test_dist2 , prot1%p_sogrid%cut2_array(grid_number,type_p2) , &
!           prot2%p_sogrid%cut2_array(grid_number,type_p1)

    ! NJB:Apply cut off 
    if (test_dist2.LT.prot1%p_sogrid%cut2_array(grid_number,type_p2)) then
       !! charge of protein 2 in grid 1
       do j = 1, nq2

          p2 = p_pos_charge2 ( :, j )
          x = pos_relat + p2

          call force_atom_uhbd ( x, force, prot1 % p_sogrid % pset_grid(  grid_number )  )
          qf = force *  p_charge2 ( j )

          !! add to force and compute torque
          ff2 = ff2 + qf
          call cross (tmp2, p2, qf )
          tt2 = tt2 + tmp2
       end do

    end if

    ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
    ey = prot2 % orientation_y
    ez = prot2 % orientation_z

       !! here ok , called only once, not needed in fact
       ! need one tr for the position anyway
       call tr ( pos_relat, pos_relat_n, ex, ey, ez )
!else
!write (0,*) "CUT P1"


    ! NJB:Apply cut off 
    if (test_dist2.LT.prot2%p_sogrid%cut2_array(grid_number,type_p1)) then
       !! compute charge of protein 1 in grid of protein 2
       do j = 1, nq1

          !rto = - pos_relat + p_pos_charge1 ( :, j )
          !call tr ( rto, rtn, ex, ey, ez )
          !write(*,*) 'rtn ', rtn
          call tr ( p_pos_charge1(:,j), rtn, ex, ey, ez )

          ! keep here to remeber it is equivalent, to delete later
          !x = rtn + pos_relat_n !! here identical to p_posc_chrgee in other referential
          !write(*,*) 'x=rtn+pos_relat_n ', x

          !call tr( p_pos_charge1 ( :, j ), rtn2, ex,ey,ez )
          !write(*,*) 'rtn2 ', rtn2 !! rtn2 fully comparable to x, normal

          !call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( grid_number ) )
          call force_atom_uhbd ( rtn - pos_relat_n, force, prot2 % p_sogrid % pset_grid( grid_number ) )
          qf = force *  p_charge1 ( j )
          !! add to force and compute torque
          !call cross (tmp1_n, x, qf )
          call cross ( tmp1_n, rtn, qf )
          tt1_n = tt1_n + tmp1_n
          ff1_n = ff1_n + qf

       end do !! loop over atoms
!else
!write (0,*) "CUT P2"
    end if

    !write(*,*) 'sum tt1_n   ', tt1_n
    !write(*,*) 'summ tt1_n2 ', tmp1_n2_sum

    !! ff1 and tt1_n result will be in referential of protein2

end subroutine force_sda_2proteins_fast

!> Compute 2 interactions with the fast algorithm
!!
!! Can compute two interactions inside the same loop, because use the same "charge"
!!
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param nb_grid1, nb_grid2 : number of the grid2, as referred in \ref mod_gridtype
!! @param gridtype : 1 or 2 used to know which couple of interactions to compute
!! @param dummy_array : used for repulsive lennard-jones interactions
!! @param ff1_n, tt1_n : force and torques on the solute 1, in the referential of the solute 2
!! @param  ff2, tt2 :  force and torques on the solute 2
!! @param type_p1, type_p2 : integer type of the solutes
!! @param test_dist2 : distance square between the solutes
subroutine force_couple_fast ( prot1, prot2, nb_grid1, nb_grid2, gridtype, dummy_array, &
                               ff1_n, ff2, tt1_n, tt2 , type_p1, type_p2 , test_dist2) !NJB: Add types
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    interface
          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr

          subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross
    end interface

    type ( protein ), intent( in ) :: prot1, prot2
    integer, intent ( in ) :: nb_grid1, nb_grid2, gridtype
    real ( kind=8 ), dimension ( : ), intent ( in ), target :: dummy_array
    real ( kind=8 ), dimension( 3 ) :: ff1_n, ff2, tt1_n, tt2

    real ( kind= 8), dimension ( 3 ) :: pos_relat
    real ( kind=8 ), dimension ( 3 ) :: force, x, p2, tmp1_n, rto, rtn, qf, tmp2
    real ( kind=8 ), dimension(3) :: ex, ey, ez, pos_relat_n

    real ( kind=8 ), dimension ( : ), pointer :: p_charge1, p_charge2, p2_charge1, p2_charge2
    real ( kind=8 ), dimension (:,:), pointer :: p_pos_charge1, p_pos_charge2
    integer :: j, nq1, nq2


    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: relevent distance to test for cut off 
    real (kind=4) , intent (in) :: test_dist2

    !! case El. and Desol.
    if ( gridtype == 1 ) then
       nq1 = prot1 % nb_charge
       nq2 = prot2 % nb_charge
       p_pos_charge1 => prot1 % p_pos_charge
       p_pos_charge2 => prot2 % p_pos_charge
       p_charge1 => prot1 % p_sogrid % charge_grid
       p_charge2 => prot2 % p_sogrid % charge_grid
       p2_charge1 => prot1 % p_sogrid % charge_square_grid
       p2_charge2 => prot2 % p_sogrid % charge_square_grid
    !! case Hyd and lennardjones
    else
       nq1 = prot1 % nb_atom_surface
       nq2 = prot2 % nb_atom_surface
       p_pos_charge1 => prot1 % p_list_atom_surface
       p_pos_charge2 => prot2 % p_list_atom_surface
       p_charge1 => prot1 % p_sogrid % accessibility
       p_charge2 => prot2 % p_sogrid % accessibility
       p2_charge1 => dummy_array
       p2_charge2 => dummy_array
    end if
!write (0,*) "FAST COUPLE"

    pos_relat = prot2 % position

!write(0,*) "test" , test_dist2

   ! NJB: If distance less than both P1 cut offs, calc both
    if ((test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid1,type_p2)).AND. &
       (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid2,type_p2))) then 
       !! Compute charge of protein2 in grid of protein1
!write(0,*) "P1 NOCUT"
       do j = 1, nq2

          p2 = p_pos_charge2 ( :, j )
          x = pos_relat + p2

          call force_atom_uhbd ( x, force, prot1 % p_sogrid % pset_grid( nb_grid1 )  ) !p1_grid1
          qf = force *  p_charge2 ( j ) !!qf2 = f2 * q2

          call force_atom_uhbd ( x, force, prot1 % p_sogrid % pset_grid( nb_grid2 ) ) !p1_grid2
          qf = qf + force * p2_charge2 ( j )

          ff2 = ff2 + qf
          !! add to force and compute torque
          call cross (tmp2, p2, qf ) !call cross( ti2, p2, qf2 )
          tt2 = tt2 + tmp2 !tt2 = tt2 + ti2
       end do


    ! NJB: Else if only smaller than the first calc that
    else if (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid1,type_p2)) then
!write(0,*) "P1 G2CUT"
       !! Compute charge of protein2 in grid of protein1
       do j = 1, nq2

          p2 = p_pos_charge2 ( :, j )
          x = pos_relat + p2

          call force_atom_uhbd ( x, force, prot1 % p_sogrid % pset_grid( nb_grid1 )  ) !p1_grid1
          qf = force *  p_charge2 ( j ) !!qf2 = f2 * q2

          ff2 = ff2 + qf
          !! add to force and compute torque
          call cross (tmp2, p2, qf ) !call cross( ti2, p2, qf2 )
          tt2 = tt2 + tmp2 !tt2 = tt2 + ti2
       end do


    ! NJB: Else if only smaller than the second calc that
    else if (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid2,type_p2)) then
!write(0,*) "P1 G1CUT"
       !! Compute charge of protein2 in grid of protein1
       do j = 1, nq2

          p2 = p_pos_charge2 ( :, j )
          x = pos_relat + p2
          call force_atom_uhbd ( x, force, prot1 % p_sogrid % pset_grid( nb_grid2 ) ) !p1_grid2
          qf = force * p2_charge2 ( j )

          ff2 = ff2 + qf
          !! add to force and compute torque
          call cross (tmp2, p2, qf ) !call cross( ti2, p2, qf2 )
          tt2 = tt2 + tmp2 !tt2 = tt2 + ti2
       end do

!else
!write(0,*) "P1 ALLCUT"
    
    end if


    ! NJB: If distance less than both P2 cut offs, calc both 
    if ((test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid1,type_p1)).AND. &
       (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid2,type_p1))) then

       ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
       ey = prot2 % orientation_y
       ez = prot2 % orientation_z

       !! here ok , called only once, could be computed only once and send as argument
       call tr ( - pos_relat, pos_relat_n, ex, ey, ez )

!write(0,*) "P2 NOCUT"
       !! compute charge of protein 1 in grid of protein 2
       do j = 1, nq1

          rto = - pos_relat + p_pos_charge1 ( :, j )
          call tr ( rto, rtn, ex, ey, ez )
          x = - pos_relat_n + rtn

          call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( nb_grid1 ) )
          qf = force *  p_charge1 ( j )

          call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( nb_grid2 ) )
          qf = qf + force * p2_charge1 ( j )

          !! add to force and compute torque
          call cross (tmp1_n, x, qf )
          tt1_n = tt1_n + tmp1_n
          ff1_n = ff1_n + qf
       end do !! loop over atoms

    ! NJB: If only smaller than the first calc that
    else if (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid1,type_p1)) then 
!write(0,*) "P2 G2CUT"

       ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
       ey = prot2 % orientation_y
       ez = prot2 % orientation_z

       !! here ok , called only once, could be computed only once and send as argument
       call tr ( - pos_relat, pos_relat_n, ex, ey, ez )

       !! compute charge of protein 1 in grid of protein 2
       do j = 1, nq1

          rto = - pos_relat + p_pos_charge1 ( :, j )
          call tr ( rto, rtn, ex, ey, ez )
          x = - pos_relat_n + rtn

          call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( nb_grid1 ) )
          qf = force *  p_charge1 ( j )

          !! add to force and compute torque
          call cross (tmp1_n, x, qf )
          tt1_n = tt1_n + tmp1_n
          ff1_n = ff1_n + qf
       end do !! loop over atoms

    ! NJB: If only smaller than the second calc that
    else if (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid2,type_p1)) then 

       ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
       ey = prot2 % orientation_y
       ez = prot2 % orientation_z

       !! here ok , called only once, could be computed only once and send as argument
       call tr ( - pos_relat, pos_relat_n, ex, ey, ez )

!write(0,*) "P2 G1CUT"
       !! compute charge of protein 1 in grid of protein 2
       do j = 1, nq1

          rto = - pos_relat + p_pos_charge1 ( :, j )
          call tr ( rto, rtn, ex, ey, ez )
          x = - pos_relat_n + rtn

          call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( nb_grid2 ) )
          qf = force * p2_charge1 ( j )

          !! add to force and compute torque
          call cross (tmp1_n, x, qf )
          tt1_n = tt1_n + tmp1_n
          ff1_n = ff1_n + qf
       end do !! loop over atoms

!else
!write(0,*) "P2 ALLCUT"
    end if

end subroutine force_couple_fast


END MODULE mod_compute_force
