!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_force_sdamm.f90
!! all functions to compute force in sdamm (OpenMP loop)

!> Group all function for computing energy for sdamm.\n
MODULE mod_compute_force_sdamm

contains

! standard version, simple to understand and extend\n
! as for sda, do not divide anymore by 2 for ep\n
! set epfct at 0.5 => identical force, modify energy to be correct\n
! will compute ep, ed, hd and lj forces
!
! extend for use with asymetric forces, still need to implement OpenMP

!> Compute forces and torques for sdamm, simple version, easier to extend
!!
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param nb_prot : total number of solutes
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param all_force, all_torque : return all forces and torques
!! @param dummy_array : used for repulsive lennard-jones interactions
subroutine force_epedhdlj_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                   all_force, all_torque, dummy_array )
    USE mod_array_protein
    USE mod_geometry
    USE mod_force_energy
    IMPLICIT NONE

    interface

          subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
          end subroutine compute_dist_pbc

      end interface

      type ( array_protein_type ), target :: tab_prot
      type ( geometry ), intent ( in ) :: geom
      type ( type_force_energy ), intent ( in ) :: param_force_energy
      real ( kind=8 ), dimension ( : , : ), intent ( out ) :: all_force, all_torque
      real ( kind=8 ), dimension( : ), target :: dummy_array
      integer, intent ( in ) :: nb_prot

      type ( protein ), pointer :: pprot1, pprot2
      ! local force and torque and correction for protein1 and 2,
      real ( kind=8 ), dimension(3) :: ff1, ff2, tt1, tt2
      real ( kind=8 ), dimension(3) :: pos_relat
      real ( kind=4 ) :: dist
      ! use several time protein2 % position relative, certainly in sdamm also
      integer :: n1, n2, i, k, grid_number
      ! NJB: need to know protein type for both proteins
      integer :: p1_type, p2_type
      !NJB: square of separation used for cut off calculations
      real ( kind=4 ) :: test_dist2

      !write(*,*) 'Entry force sdamm nb_prot', nb_prot
      !write(*,*) 'param_force_energy % array_grid_calc ', param_force_energy % array_grid_calc(:,1)
      !write(*,*) 'param_force_energy % array_grid_calc ', param_force_energy % array_grid_calc(:,2)
      !write(*,*) 'param_force_energy % array_grid_calc ', param_force_energy % array_grid_calc (:,3)

!! simpler, use REDUCTION, later test and extend to manula reduction, like sdamm_fast
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED ( tab_prot, nb_prot, param_force_energy, geom, dummy_array ) & ! all_force,all_torque) &
!$OMP PRIVATE ( grid_number, ff1, ff2, tt1, tt2 ) &
!$OMP   PRIVATE ( pprot1, pprot2 ) &  
!$OMP   PRIVATE ( p1_type, p2_type ) & !NJB
!$OMP   PRIVATE ( test_dist2 ) & !NJB
!$OMP   PRIVATE ( pos_relat, dist ) & !! tid )
!! simpler, use REDUCTION, later test and extend to manula reduction, like sdamm_fast
!$OMP REDUCTION(+:all_force,all_torque)

!$OMP DO SCHEDULE( DYNAMIC )
PROT1: do n1 = 1, nb_prot
PROT2:    do  n2 = n1 + 1, nb_prot

            pprot1 => tab_prot % array ( n1 )
            pprot2 => tab_prot % array ( n2 )

            call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
            !write(*,*) 'After compute dist pbc '
            !write (*,*) 'after dist, pos relat ',  dist, pos_relat, dot_product ( pos_relat, pos_relat )
            test_dist2 = dist*dist

            if (test_dist2.GT.max(pprot1%p_sogrid%max_cut2,pprot2%p_sogrid%max_cut2)) then
                cycle PROT2
            end if

            ! NJB: find types of both proteins
            p1_type = param_force_energy%array_prot_type(pprot1%id)
            p2_type = param_force_energy%array_prot_type(pprot2%id)

GRID: do i = 1, param_force_energy % nb_grid

        grid_number =  param_force_energy % array_grid_calc ( i, 1 )
        !write(*,*) 'grid_number ', grid_number

        !write(*,*) 'force ', n1, n2
        !write(*,*) 'number subgrids ',  param_force_energy % array_grid_calc ( i, 3 )
        ! new added
        do k = 1,  param_force_energy % array_grid_calc ( i, 3 )
            call compute_force_sdamm_2proteins( pprot1, pprot2, pos_relat, grid_number, dummy_array, &
                                            ff1, ff2, tt1, tt2, k , &
                                            p1_type, p2_type, test_dist2) ! NJB: add protein types

          all_force ( :,n1 ) = all_force ( :,n1 ) +  ff1
          all_force ( :,n2 ) = all_force ( :,n2 ) +  ff2
          all_torque ( :,n1 ) = all_torque ( :,n1 ) + tt1
          all_torque ( :,n2 ) = all_torque ( :,n2 ) + tt2
        end do

        !write(*,*) 'Total FORCE ' !, n1, n2
        !write (*,*) 'Test new force f1: ', all_force (:,n1)
        !write (*,*) 'Test new force f2: ', all_force (:,n2)
        !write (*,*) 'Test new torque t1:', all_torque (:,n1)
        !write (*,*) 'Test new torque t2:', all_torque (:,n2)

    end do GRID

  end do PROT2
end do PROT1
!$OMP END DO
!$OMP END PARALLEL

end subroutine force_epedhdlj_sdamm

! version for sdamm, need to change referentiel for both proteins
! and do not need to correct for fixed center of mass, still done
! could combine both later ??

!> Compute forces and torques for one pair of solutes and one interaction.\n
!!
!! Explicitely check if there iare asymetries
!!
!! @param prot1, prot2 : instance of \ref mod_protein::protein "protein"
!! @param pos_relat : relative position between the 2 solutes
!! @param grid_number : number of the grid, refer to \ref mod_gridtype
!! @param dummy_array : used for repulsive lennard-jones interactions
!! @param ff1, ff2, tt1, tt2 : return forces and torques on solutes 1 and 2
!! @param subgrid : sub-grid number, normally one (used for LJ)
!! @param type_p1, type_p2 : Solute types for interacting solutes
!! @param type_p1, type_p2 : integer type of the solutes
!! @param test_dist2 : distance square between the solutes
subroutine compute_force_sdamm_2proteins( prot1, prot2, pos_relat, grid_number, dummy_array, &
                                          ff1, ff2, tt1, tt2, subgrid , &
                                          type_p1, type_p2, test_dist2 ) !NJB: add protein types and separation squared
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    interface
          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr

          subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross
    end interface

    type ( protein ), intent( in ) :: prot1, prot2
    real ( kind=8 ), dimension ( 3 ), intent ( in ) :: pos_relat
    real ( kind=8 ), dimension ( : ), intent ( in ), target :: dummy_array
    real ( kind=8 ), dimension( 3 ), intent ( out ) :: ff1, ff2, tt1, tt2
    integer, intent ( in ) :: grid_number, subgrid

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2
    ! new added
    type ( type_grid ), pointer :: p_grid1, p_grid2

    real ( kind=8 ), dimension(3) :: rto, rtn
    ! force, torque and correction for protein1 and 2, force replace f1 and f2
    real ( kind=8 ), dimension(3) :: tmp1, tmp2, p1, p2, force, force_o
    ! tmp orientation, need inversion for forces on protein1
    real ( kind=8 ), dimension(3) :: ex, ey, ez, xe, ye, ze, qf, tmp_ff1, tmp_tt1
    integer :: j, nq1, nq2
    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: is one of the proteins a surface 
    !logical :: surf_int !NOT NEEDED??
    ! NJB: relevent distance to test for cut off 
    real (kind=4) , intent (in) :: test_dist2


    ff1 = 0.d0
    tt1 = 0.d0
    ff2 = 0.d0
    tt2 = 0.d0
    tmp1 = 0.d0
    tmp2 = 0.d0

    !write(*,*) 'Entry compute_force_sdamm_2proteins'
    !write(*,*) 'grid number ', grid_number
    !write(*,*) 'subgrid ', subgrid

    ! assign the correct pointer, version deal with lennard-jones and multipgrids
    call select_charge2 ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                          p_charge_position1, p_charge_position2, dummy_array, &
                          p_grid1, p_grid2, subgrid )

    ! compute 2nde protein first, with correct pos_relat ( p2 - p1 )
    ! add explicit tests, easier than an array_grid_calc for each subtype
    if (  ( nq2 > 0 ) .and. (associated ( p_grid1 ) ) ) then
        !write(*,*) 'grid_number ', grid_number, ' p_grid1 is associated '
        !write(*,*) 'associated ( p_grid1 ) ',  associated ( p_grid1 )
        !write(*,*) 'associated ( p_charge_position2 ) ',  associated (p_charge_position2  )
        !write(*,*) 'associated ( p_charge2 ) ',  associated ( p_charge2 )
        !write(*,*) 'nq2 ', nq2

       !NJB: Check cut off distance for interaction
       if (test_dist2.LT.prot1%p_sogrid%cut2_array(grid_number,type_p2)) then
!write(0,*) "NOT CUT"
     
          !! compute force and torque for first protein
          !!! prepare transformation of the first protein
          ex = prot1 % orientation_x
          ey = prot1 % orientation_y
          ez = prot1 % orientation_z
          xe(1) = ex(1)
          xe(2) = ey(1)
          xe(3) = ez(1)
          ye(1) = ex(2)
          ye(2) = ey(2)
          ye(3) = ez(2)
          ze(1) = ex(3)
          ze(2) = ey(3)
          ze(3) = ez(3)

FORCE2:   do j=1,nq2

             p2 = p_charge_position2 (:,j)
             rto = p2 + pos_relat

             call tr ( rto, rtn, ex, ey, ez )
             !call force_atom_uhbd ( rtn, force, prot1 % p_sogrid % pset_grid ( grid_number) )
             call force_atom_uhbd ( rtn, force, p_grid1 )
             !write(*,*) 'force_n charge2 dans 1 ', force, force * p_charge2(j)
             call tr ( force, force_o, xe, ye, ze )
             qf = force_o * p_charge2(j)

             call cross (tmp2, p2, qf )
             !write( *,*) 'tmp2 ', (tmp2(m),m=1,3)
             tt2 = tt2 + tmp2
             ff2 = ff2 + qf

          end do FORCE2

      ! move correction terms
      !. torque on 1    (r)x(-ff2)
      !CALL cross(tmp2,r,ff2)
      !call cross ( tmp2, pos_relat, ff2 )
!       else
!write(0,*) "CUT"
       !NJB: else do nothing - delete
       end if
    endif

    if ( ( nq1 > 0 ) .and. ( associated ( p_grid2 ) ) ) then
        !write(*,*) 'grid_number ', grid_number, ' p_grid2 is associated '
        !write(*,*) 'associated ( p_grid2 ) ',  associated ( p_grid2 )
        !write(*,*) 'associated ( p_charge_position1 ) ',  associated (p_charge_position1  )
        !write(*,*) 'associated ( p_charge1 ) ',  associated ( p_charge1 )
        !write(*,*) 'nq1 ', nq1

       !NJB: Check cut off distance for interaction
       if (test_dist2.LT.prot2%p_sogrid%cut2_array(grid_number,type_p1)) then
!write(0,*) "NOT CUT"

          ! change for second protein
          ! torques and force for the first protein
          ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
          ey = prot2 % orientation_y
          ez = prot2 % orientation_z
          xe(1) = ex(1)
          xe(2) = ey(1)
          xe(3) = ez(1)
          ye(1) = ex(2)
          ye(2) = ey(2)
          ye(3) = ez(2)
          ze(1) = ex(3)
          ze(2) = ey(3)
          ze(3) = ez(3)

FORCE1:   do j=1,nq1

             p1 = p_charge_position1(:,j) ! p1 = x1(:,j)
             rto = p1 - pos_relat
             !write(*,*) 'rto', (rto(m),m=1,3)
             call tr(rto, rtn, ex, ey, ez )
             call force_atom_uhbd ( rtn, force, p_grid2 )
             !write(*,*) 'force 2,1 ', force
             !write(*,*) 'force_n charge1 dans 2 ', force, force * p_charge1(j)
             call tr ( force, force_o, xe, ye, ze )

             qf = force_o * p_charge1 ( j )
             call cross(tmp1, p1, qf)
             tt1 = tt1 + tmp1
             ff1 = ff1 + qf

          end do FORCE1


!. torque on 1    (r)x(-ff2)
!    !CALL cross(tmp2,r,ff2)
!    call cross ( tmp2, pos_relat, ff2 )
       else
!write (0,*) "CUT"
       ! NJB: Else do nothing - delete
       end if     
    end if


  ! now apply correction to all cases, symetric and asymetric
  ! works here because ff1 or ff2 will be zero in case of asymetry, or move inside if charge/grid exist

    ! torque on 2   (-r)x(-ff1)
    ! CALL cross(tmp1,r,ff1)
    call cross ( tmp1, pos_relat, ff1 )
    ! torque on 1    (r)x(-ff2)
    ! CALL cross(tmp2,r,ff2)
    call cross ( tmp2, pos_relat, ff2 )
    !write(*,*) 'correction torque ', tmp1, tmp2

    ! memorize because overwritten
    tmp_ff1 = ff1
    tmp_tt1 = tt1
    ! save diff tt2-tt1 for instance
    ! how to apply correction in case of asymetry ??
    ! update force and torques with the correction
    ff1 = ff1 - ff2
    ff2 = ff2 - tmp_ff1
    tt1 = tt1 - tt2 - tmp2
    tt2 = tt2 - tmp_tt1 + tmp1

    !write(*,*) 'all forces sdamm uhbd'
    !write(*,*) 'ff1 ', ff1
    !write(*,*) 'ff2 ', ff2
    !write(*,*) 'tt1 ', tt1
    !write(*,*) 'tt2 ', tt2
    !write (*,'(''Test new force f1: '',3E14.5 )') all_force (:,n1)
    !write (*,'(''Test new force f2: '',3E14.5 )') all_force (:,n2)
    !write (*,'(''Test new torque t1: '',3E14.5 )') all_torque (:,n1)
    !write (*,'(''Test new torque t2: '',3E14.5 )') all_torque (:,n2)

end subroutine compute_force_sdamm_2proteins

!!!!!!!!!!!!!!!!!!!!!!!      fast version  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Version fast, compute the sum of forces and torques of prot1 in referential of prot2\n
! ff1_n, ff2_n. Come back to box referential only at the end.\n
! if (ep and ed) or (hd and lj-rep), they are computed togethe,\n
! compared to test, 1/4 faster\n
! compared to initial almost 50 % with 128 proteins\n

!> Compute forces for sdamm with the fast algorithm
!!
!! If possible compute a couple of interactions at the same type
!!
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param nb_prot : total number of solutes
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param dummy_array : used for repulsive lennard-jones interactions
!! @param all_force, all_torque : return  all forces and torques
subroutine force_epedhdlj_sdamm_fast ( tab_prot, geom, nb_prot, param_force_energy, &
                                       all_force, all_torque, dummy_array )
    USE mod_array_protein
    USE mod_geometry
    USE mod_force_energy
!$  USE omp_lib
    IMPLICIT NONE

    interface

        subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
        end subroutine compute_dist_pbc

        subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
        end subroutine tr

        subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
        end subroutine cross
    end interface

    type ( array_protein_type ), target :: tab_prot
    type ( geometry ), intent ( in ) :: geom
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension ( : , : ), intent ( out ) :: all_force, all_torque
    real ( kind=8 ), dimension ( : ), intent ( in ) :: dummy_array
    integer :: nb_prot

    !temporary all_force, all torque
    real ( kind=8 ), dimension ( : , : ), allocatable :: tmp_all_force, tmp_all_torque

    type ( protein ), pointer :: pprot1, pprot2
    real ( kind=8 ), dimension(3) :: ff1_o, ff1_n, ff2_o, ff2_n, tt1_o, tt1_n, tt2_o, tt2_n
    real ( kind=8 ), dimension(3) :: ff1_sum, tt1_sum, tt2_sum
    real ( kind=8 ), dimension(3) :: tmp1, tmp2
    real (kind=4) :: dist
    ! tmp orientation, need inversion for forces on protein1
    real ( kind=8 ), dimension(3) :: ex2, ey2, ez2, xe2, ye2, ze2, ex1, ey1, ez1, xe1, ye1, ze1
    real ( kind=8 ), dimension(3) :: pos_relat
    integer :: n1, n2, counter, n3
    ! NJB: need to know protein type for both proteins
    integer :: p1_type, p2_type
    !NJB: square of separation used for cut off calculations
    real ( kind=4 ) :: test_dist2

    !integer :: tid

    all_force = 0.d0
    all_torque = 0.d0

    !nullify(pprot1)
    !nullify(pprot2)
    ff1_o = 0.d0
    tt1_o = 0.d0
    ff2_o = 0.d0
    tt2_o = 0.d0
    ff1_sum = 0.d0
    tt1_sum = 0.d0
    tt2_sum = 0.d0

    !write(*,*) 'Entry compute force fast nb_prot', nb_prot

!what the default for schedule task ? most static ?
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP SHARED (tab_prot,nb_prot,param_force_energy,geom,dummy_array,all_force,all_torque) &
!$OMP PRIVATE ( tmp_all_force, tmp_all_torque ) &
!$OMP   PRIVATE ( ff1_sum, tt1_sum, tt2_sum ) &
!$OMP   PRIVATE ( pprot1, pprot2, ex1, ey1, ez1, ex2, ey2, ez2, xe2, ye2, ze2, xe1, ye1, ze1 ) &
!$OMP   PRIVATE ( p1_type, p2_type ) & !NJB
!$OMP   PRIVATE ( test_dist2 ) & !NJB
!$OMP   PRIVATE ( ff1_o, ff1_n, ff2_o, ff2_n, tt1_o, tt1_n, tt2_o, tt2_n ) &
!$OMP   PRIVATE ( counter, tmp1, tmp2, pos_relat, dist ) !, tid )


!! here allocate local array
allocate ( tmp_all_force ( 3, nb_prot ) )
tmp_all_force = 0.d0
allocate ( tmp_all_torque ( 3, nb_prot ) )
tmp_all_torque = 0.d0

!what the default for schedule task ? most static ? Seems good because of Cutoff
! only the first loop is parrallelized.
!$OMP DO SCHEDULE( DYNAMIC )
PROT1: do n1 = 1, nb_prot

         pprot1 => tab_prot % array ( n1 )

         ex1 = pprot1 % orientation_x
         ey1 = pprot1 % orientation_y
         ez1 = pprot1 % orientation_z
         xe1(1) = ex1(1)
         xe1(2) = ey1(1)
         xe1(3) = ez1(1)
         ye1(1) = ex1(2)
         ye1(2) = ey1(2)
         ye1(3) = ez1(2)
         ze1(1) = ex1(3)
         ze1(2) = ey1(3)
         ze1(3) = ez1(3)

PROT2: do  n2 = n1 + 1, nb_prot

        !tid = omp_get_thread_num()
        !write(*,*) 'prot 1-2', n1, n2 !, ' tid ', tid

         pprot2 => tab_prot % array ( n2 )

         call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
         !write(*,*) 'After compute dist pbc '
         !write (*,*) 'after dist, pos relat ',  dist, pos_relat, dot_product ( pos_relat, pos_relat )
         test_dist2 = dist*dist

         if (test_dist2.GT.max(pprot1%p_sogrid%max_cut2,pprot2%p_sogrid%max_cut2)) then
            cycle PROT2
         end if

         ! Get protein type of each protein 
         p1_type = param_force_energy%array_prot_type(pprot1%id)
         p2_type = param_force_energy%array_prot_type(pprot2%id)
!***NJB********************!

         ex2 = pprot2 % orientation_x
         ey2 = pprot2 % orientation_y
         ez2 = pprot2 % orientation_z
         xe2(1) = ex2(1)
         xe2(2) = ey2(1)
         xe2(3) = ez2(1)
         ye2(1) = ex2(2)
         ye2(2) = ey2(2)
         ye2(3) = ez2(2)
         ze2(1) = ex2(3)
         ze2(2) = ey2(3)
         ze2(3) = ez2(3)

         !! reinitialize for each pair or set of pairs
         !! new force will be added
         ff1_n = 0.d0
         tt1_n = 0.d0
         ff2_n = 0.d0
         tt2_n = 0.d0
         tmp1 = 0.d0
         tmp2 = 0.d0

         !! compute ep and ed interactions, if they are symetric
         counter = 1
         if ( param_force_energy % couple_ElDes .eqv. .true. ) then
             !write(*,*) 'compute couple ElDes '
             call compute_force_Couple_sdamm_fast ( pprot1, pprot2, pos_relat, &
                    ! could be 1 and 2
                    param_force_energy % array_grid_calc ( 1, 1 ), param_force_energy % array_grid_calc ( 2, 1 ), &
                    1, dummy_array, ff1_n, ff2_n, tt1_n, tt2_n , &
                    p1_type, p2_type , test_dist2) !NJB: add protein type

             !! in this case, next counter will be 3
             counter = 3

         !! compute only one if they are asymetric, cannot compute both ( fast_computation is false for all )
         !! so all force uses the very(in case of force) slow algorithm
         else if ( ( param_force_energy % array_grid_calc ( 1, 1 ) == 1 ) .or. &
                   ( param_force_energy % array_grid_calc ( 1, 1 ) == 2 ) ) then

              !write(*,*) 'compute one grid el or ed'
              call compute_force_sdamm_2proteins_fast ( pprot1, pprot2, pos_relat, &
                                  param_force_energy % array_grid_calc ( 1, 1 ), &
                                  dummy_array, ff1_n, ff2_n, tt1_n, tt2_n , &
                                  p1_type , p2_type , test_dist2) !NJB: add protein type and separation squared

              !write(*,*) 'force ep/ed ', ff1_n, ff2_n, tt1_n, tt2_n
              !! in this case, next counter will be 2
              counter = 2
         end if

         !need to test size, before continuing ??, cannt return here !
         if ( param_force_energy % nb_grid < counter ) then
             !write(*,*) 'in energy no more grid return '
             GOTO 11
         end if

       !! compute hd/ lj-rep interactions
       if ( param_force_energy % couple_ApDesLJrep .eqv. .true. ) then

           !write(*,*) 'compute couple ApDESLJrep counter ', counter
           call compute_force_Couple_sdamm_fast ( pprot1, pprot2, pos_relat, &
               !! could be 3 and 4
               param_force_energy % array_grid_calc ( counter, 1 ), param_force_energy % array_grid_calc ( counter + 1, 1 ), &
               2, dummy_array, ff1_n, ff2_n, tt1_n, tt2_n , &
               p1_type , p2_type, test_dist2) !NJB: add protein type and separation squared

               !counter = counter + 2

       !! else if only one grid to compute
       else if ( ( param_force_energy % array_grid_calc ( counter, 1 ) == 3 ) .or. &
                 ( param_force_energy % array_grid_calc ( counter, 1 ) == 4 ) ) then

               !write(*,*) 'compute one grid hd or lj-rep counter ', counter
               call compute_force_sdamm_2proteins_fast ( pprot1, pprot2, pos_relat, &
                    param_force_energy % array_grid_calc ( counter, 1 ), &
                    dummy_array, ff1_n, ff2_n, tt1_n, tt2_n  , &
                    p1_type , p2_type , test_dist2) !NJB: add protein type and separation squared

               !write(*,*) 'force hd/jj-rep ', ff1_n, ff2_n, tt1_n, tt2_n
               !counter = counter + 1
       end if

11     CONTINUE

       ! end calculation, come back to original referential
       call tr ( ff2_n, ff2_o, xe1, ye1, ze1 )
       call tr ( tt2_n, tt2_o, xe1, ye1, ze1 )
       call tr ( ff1_n, ff1_o, xe2, ye2, ze2 )
       call tr ( tt1_n, tt1_o, xe2, ye2, ze2 )

! correction done in box referential in all cases
! torque on 2   (-r)x(-ff1)
       call cross ( tmp1, pos_relat, ff1_o )
! torque on 1    (r)x(-ff2)
       call cross ( tmp2, pos_relat, ff2_o )
       !write( *,*) 'tt1_o ff1_o ', tt1_o, ff1_o

       !write(*,*) 'new force and torque'
       !write(*,*) 'ff1_n ',ff1_n
       !write(*,*) 'ff2_n ',ff2_n
       !write(*,*) 'ff1_o ', ff1_o
       !write(*,*) 'ff2_o ', ff2_o
       !write(*,*) 'tt1_o ', tt1_o
       !write(*,*) 'tt2_o ', tt2_o
       !write(*,*) 'ff1 ', ff1_o - ff2_o
       !write(*,*) 'ff2 ', ff2_o - ff1_o
       !write(*,*) 'tt1 ', tt1_o - tt2_o - tmp2
       !write(*,*) 'tt2 ', tt2_o - tt1_o + tmp1
       !write(*,*) 'torque1 ', tt1_o
       !write(*,*) 'torque2 ', tt2_o
       !write(*,*) 'correction ', tmp1, tmp2

       !ff1_sum =  ff1_o - ff2_o
       !ff2_sum =  - ff1_sum
       !tt1_sum =  tt1_o - tt2_o - tmp2
       !tt2_sum =  tt2_o - tt1_o + tmp1

       ! before with reduction here, now update local
       tmp_all_force ( :,n1 ) = tmp_all_force ( :,n1 ) +  ff1_o - ff2_o
       tmp_all_force ( :,n2 ) = tmp_all_force ( :,n2 ) -  ff1_o + ff2_o
       tmp_all_torque ( :,n1 ) = tmp_all_torque ( :,n1 ) + tt1_o - tt2_o - tmp2
       tmp_all_torque ( :,n2 ) = tmp_all_torque ( :,n2 ) + tt2_o - tt1_o + tmp1


!!!!!!!!!!!!!!!!!!!!!!!!! END of force computation
    !write(*,*) 'Total AllForce prot ', n1, n2
    !write (*,*) 'force n1', all_force ( :, n1 )
    !write (*,*) 'force n2', all_force ( :, n2 )
    !write (*,*) 'torque t1', all_torque ( :, n1 )
    !write (*,*) 'torque t2', all_torque ( :, n2 )

   end do PROT2
 end do PROT1
!$OMP END DO

! here can update all_force, all_torque, each thread copy its own data
do n3= 1, nb_prot
!  do n2= n1 +1, nb_prot
!    do j= 1, 3
!! first version, faster for smaller number of cpus
!! !$OMP CRITICAL
!       all_force ( :,n3 ) =  all_force ( :,n3 ) + tmp_all_force (:,n3)
!       all_torque (:,n3 ) = all_torque (:,n3) + tmp_all_torque (:,n3)
!!!$OMP END CRITICAL
!$OMP ATOMIC
       all_force ( 1,n3 ) =  all_force ( 1,n3 ) + tmp_all_force(1,n3)
!$OMP ATOMIC
       all_force ( 2,n3 ) =  all_force ( 2,n3 ) + tmp_all_force(2,n3)
!$OMP ATOMIC
       all_force ( 3,n3 ) =  all_force ( 3,n3 ) + tmp_all_force(3,n3)
!$OMP ATOMIC
       all_torque ( 1,n3 ) = all_torque ( 1,n3 ) + tmp_all_torque(1,n3)
!$OMP ATOMIC
       all_torque ( 2,n3 ) = all_torque ( 2,n3 ) + tmp_all_torque(2,n3)
!$OMP ATOMIC
       all_torque ( 3,n3 ) = all_torque ( 3,n3 ) + tmp_all_torque(3,n3)
end do


deallocate ( tmp_all_force )
deallocate ( tmp_all_torque )
!$OMP END PARALLEL

end subroutine force_epedhdlj_sdamm_fast

!> Compute forces and torques for one pair of solutes and one interaction
!!
!! Do nopt check for asymetries in this case.
!!
!! @param prot1, prot2 : instance of \ref mod_protein::protein "protein"
!! @param pos_relat : relative position between the 2 solutes
!! @param grid_number : number of the grid, refer to \ref mod_gridtype
!! @param dummy_array : used for repulsive lennard-jones interactions
!! @param ff1_n, ff2_n, tt1_n, tt2_n : return forces and torques on solutes 1 and 2
!! @param type_p1, type_p2 : Solute types for interacting solutes
!! @param type_p1, type_p2 : integer type of the solutes
!! @param test_dist2 : distance square between the solutes
subroutine compute_force_sdamm_2proteins_fast ( prot1, prot2, pos_relat, grid_number, dummy_array, &
                                                ff1_n, ff2_n, tt1_n, tt2_n , &
                                                type_p1 , type_p2 , test_dist2) ! NJB: Add protein types and separation squared
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    interface
          subroutine tr(vo,vn,ex,ey,ez)
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent ( out ) :: vn
            real ( kind=8 ), dimension (3), intent ( in ) :: vo,ex,ey,ez
          end subroutine tr

          subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross
    end interface

    type ( protein ), intent ( in ) :: prot1, prot2
    real ( kind=8 ), dimension ( : ), intent ( in ) :: dummy_array
    real ( kind=8 ), dimension( 3 ), intent ( inout ) :: ff1_n, ff2_n, tt1_n, tt2_n
    integer, intent ( in ) :: grid_number

    real ( kind=8 ), dimension (3 ) :: pos_relat
    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_pos_charge1, p_pos_charge2

    real ( kind=8 ), dimension ( 3 ) :: force, x, tmp1_n, tmp2_n, rto, rtn, qf
    real ( kind=8 ), dimension(3) :: ex, ey, ez, pos_relat_n
    integer :: nq1, nq2, j

    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: is one of the proteins a surface
    !logical :: surf_int !NOT NEEDED??
    ! NJB: relevant distance to test for cut off 
    real (kind=4) , intent (in) :: test_dist2


    !! select correct grid
    call select_charge ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                         p_pos_charge1, p_pos_charge2, dummy_array )

    !write(*,*) 'Entry compute_force_ 2 prot fast sdamm'
    !write(*,*)'pos_relat ', pos_relat

    ! NJB: Are charges on protein 2 within grid of protein 1, apply cut off if not 
    if (test_dist2.LT.prot1%p_sogrid%cut2_array(grid_number,type_p2)) then
!write(0,*) "NOCUT  Distance :" ,test_dist2, "  cut: ", prot1%p_sogrid%cut2_array(grid_number,type_p2)

       ex = prot1 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
       ey = prot1 % orientation_y
       ez = prot1 % orientation_z

       !! here ok , called only once, trick, why ?? forgot...
       !! can try modif like for sda_2proteins, method clearer i htink
       call tr ( pos_relat, pos_relat_n, ex, ey, ez )
       !! charge of protein 2 in grid 1
       do j = 1, nq2

          rto = pos_relat + p_pos_charge2 ( :,j )
          call tr ( rto, rtn, ex, ey, ez )
          !! compute pos in new referentiel
          x = - pos_relat_n + rtn

          call force_atom_uhbd ( rtn, force, prot1 % p_sogrid % pset_grid(  grid_number )  )
          qf = force *  p_charge2 ( j )
          !write(*,*) 'qf electro', qf
          ff2_n = ff2_n + qf
          ! here in new referentiel
          call cross (tmp2_n, x, qf )
          tt2_n = tt2_n + tmp2_n
          !write(*,*) 'torque_n charge2 dans 1 ', tt2_n

       end do
    else
!write(0,*) "YESCUT  Distance :" ,test_dist2, "  cut: ", prot1%p_sogrid%cut2_array(grid_number,type_p2)
    end if 

    !NJB: Apply cut off to second protein
    if (test_dist2.LT.prot2%p_sogrid%cut2_array(grid_number,type_p1)) then
!write(0,*) "NOCUT  Distance :" ,test_dist2, "  cut: ", prot2%p_sogrid%cut2_array(grid_number,type_p1)

       ex = prot2 % orientation_x ! ex(m) = orientation_X(m)for each axis  all m !
       ey = prot2 % orientation_y
       ez = prot2 % orientation_z

       !! here ok , called only once trick, why ??
       call tr ( pos_relat, pos_relat_n, ex, ey, ez )
       !! compute charge of protein 1 in grid of protein 2
       do j = 1, nq1

          rto = - pos_relat + p_pos_charge1 ( :, j )
          call tr ( rto, rtn, ex, ey, ez )
          x = pos_relat_n + rtn

          call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( grid_number ) )
          qf = force *  p_charge1 ( j )
          !write(*,*) 'force_n charge1 dans 2 ', force, qf
          !! add to force and compute torque
          call cross (tmp1_n, x, qf ) !call cross( ti2, p2, qf2 )
          tt1_n = tt1_n + tmp1_n
          !write(*,*) 'torque_n charge2 dans 1 ', tt1_n
          ff1_n = ff1_n + qf

       end do !! loop over atoms
    else
!write(0,*) "YESCUT  Distance :" ,test_dist2, "  cut: ", prot2%p_sogrid%cut2_array(grid_number,type_p1) &
!     , "  maxcut: ", prot2%p_sogrid%max_cut
    end if

    !! ff1 and tt1_n result will be in referential of protein2
    !! sdamm same for ff2_n and tt2_n, will transform only at the end

end subroutine compute_force_sdamm_2proteins_fast

!> Compute forces and torques of two interactions, for one pair of solutes
!!
!! Compute a couple of interactions in the same loop, do not check for asymetries
!! Fastest method
!!
!! @param prot1, prot2 : instance of \ref mod_protein::protein "protein"
!! @param pos_relat : relative position between the 2 solutes
!! @param nb_grid1, nb_grid2 : number of the grids, refer to \ref mod_gridtype
!! @param gridtype : 1 or 2 used to know which couple of interactions to compute
!! @param dummy_array : used for repulsive lennard-jones interactions
!! @param ff1_n, ff2_n, tt1_n, tt2_n : return forces and torques on solutes 1 and 2
!! @param type_p1, type_p2 : integer type of the solutes
!! @param test_dist2 : distance square between the solutes
subroutine compute_force_Couple_sdamm_fast ( prot1, prot2, pos_relat, nb_grid1, nb_grid2, &
                                             gridtype, dummy_array, ff1_n, ff2_n, tt1_n, tt2_n , &
                                             type_p1 , type_p2 , test_dist2) !NJB : Add protein types
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE
    type ( protein ), intent( in ) :: prot1, prot2
    integer, intent ( in ) :: nb_grid1, nb_grid2, gridtype
    real ( kind=8 ), dimension ( : ), intent ( in ), target :: dummy_array
    real ( kind=8 ), dimension( 3 ), intent( inout ) :: pos_relat, ff1_n, ff2_n, tt1_n, tt2_n

    real ( kind=8 ), dimension(3) :: ex, ey, ez, pos_relat_n
    real ( kind=8 ), dimension ( 3 ) :: tmp1_n, tmp2_n, rto, rtn, qf, x, force
    real ( kind=8 ), dimension ( : ), pointer :: p_charge1, p_charge2, p2_charge1, p2_charge2
    real ( kind=8 ), dimension (:,:), pointer :: p_pos_charge1, p_pos_charge2
    integer :: j, nq1, nq2

    ! NJB: Protein types
    integer , intent (in) :: type_p1, type_p2
    ! NJB: is one of the proteins a surface
    !logical :: surf_int !NOT NEEDED??
    ! NJB: relevent distance to test for cut off 
    real (kind=4) , intent (in) :: test_dist2

    !write(*,*) 'entry compute couple force'
    !write(*,*) 'grid number ', nb_grid1, nb_grid2
    !write(*,*) 'gridtype ', gridtype

    ! case El. and Desolv.
    if ( gridtype == 1 ) then
       nq1 = prot1 % nb_charge
       nq2 = prot2 % nb_charge
       p_pos_charge1 => prot1 % p_pos_charge
       p_pos_charge2 => prot2 % p_pos_charge
       p_charge1 => prot1 % p_sogrid % charge_grid
       p_charge2 => prot2 % p_sogrid % charge_grid
       p2_charge1 => prot1 % p_sogrid % charge_square_grid
       p2_charge2 => prot2 % p_sogrid % charge_square_grid
    !! case Hyd. and LJ-repulsive
    else
       nq1 = prot1 % nb_atom_surface
       nq2 = prot2 % nb_atom_surface
       p_pos_charge1 => prot1 % p_list_atom_surface
       p_pos_charge2 => prot2 % p_list_atom_surface
       p_charge1 => prot1 % p_sogrid % accessibility
       p_charge2 => prot2 % p_sogrid % accessibility
       p2_charge1 => dummy_array
       p2_charge2 => dummy_array
    end if

    !! energy of charge of protein2 in grid of protein1
    ex = prot1 % orientation_x
    ey = prot1 % orientation_y
    ez = prot1 % orientation_z

    call tr ( pos_relat, pos_relat_n, ex, ey, ez )

   ! NJB: If distance less than both cut offs, calc both
    if ((test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid1,type_p2)).AND. &
       (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid2,type_p2))) then

       do j = 1, nq2

          rto = pos_relat + p_pos_charge2 ( :,j )
          call tr ( rto, rtn, ex, ey, ez )
          !! compute pos in new referentiel
          x = - pos_relat_n + rtn

          call force_atom_uhbd ( rtn, force, prot1 % p_sogrid % pset_grid ( nb_grid1 ) )
          qf = force *  p_charge2 ( j )

          call force_atom_uhbd ( rtn, force, prot1 % p_sogrid % pset_grid ( nb_grid2 ) )
          qf = qf + force * p2_charge2 ( j )

          !! add to force and compute torque
          call cross (tmp2_n, x, qf )
          tt2_n = tt2_n + tmp2_n
          ff2_n = ff2_n + qf

       end do !! loop over atoms of prot 2
    
    ! NJB: If only smaller than the first calc that 
    else if (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid1,type_p2)) then
!write (0,*) "P1SECOND CUT"

       do j = 1, nq2

          rto = pos_relat + p_pos_charge2 ( :,j )
          call tr ( rto, rtn, ex, ey, ez )

          !! compute pos in new referentiel
          x = - pos_relat_n + rtn

          call force_atom_uhbd ( rtn, force, prot1 % p_sogrid % pset_grid ( nb_grid1 ) )
          qf = force *  p_charge2 ( j )

          !! add to force and compute torque
          call cross (tmp2_n, x, qf )
          tt2_n = tt2_n + tmp2_n
          ff2_n = ff2_n + qf

       end do

    ! NJB: If only smaller than the second calc that 
    else if (test_dist2.LT.prot1%p_sogrid%cut2_array(nb_grid2,type_p2)) then
!write (0,*) "P1FIRST CUT"

       do j = 1, nq2

          rto = pos_relat + p_pos_charge2 ( :,j )
          call tr ( rto, rtn, ex, ey, ez )

          !! compute pos in new referentiel
          x = - pos_relat_n + rtn

          call force_atom_uhbd ( rtn, force, prot1 % p_sogrid % pset_grid ( nb_grid1 ) )
          qf = force *  p2_charge2 ( j )

          !! add to force and compute torque
          call cross (tmp2_n, x, qf )
          tt2_n = tt2_n + tmp2_n
          ff2_n = ff2_n + qf

       end do
!else 
!write (0,*) "P1BOTH CUT"

    end if 

    !! energy of charge of protein1 in grid of protein2
    ex = prot2 % orientation_x
    ey = prot2 % orientation_y
    ez = prot2 % orientation_z
    call tr ( pos_relat, pos_relat_n, ex, ey, ez )

    ! NJB: If distance less than both cut offs, calc both 
    if ((test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid1,type_p1)).AND. &
       (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid2,type_p1))) then

       do j = 1, nq1

           rto = - pos_relat + p_pos_charge1 ( :, j )
           call tr ( rto, rtn, ex, ey, ez )
           x = pos_relat_n + rtn

           call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( nb_grid1 ) )
           qf = force *  p_charge1 ( j )

           call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( nb_grid2 ) )
           qf = qf + force * p2_charge1 ( j )

           !! add to force and compute torque
           call cross (tmp1_n, x, qf )
           tt1_n = tt1_n + tmp1_n
           ff1_n = ff1_n + qf

       end do !! loop over atoms of prot 1
    
    ! NJB: If only smaller than the first calc that
    else if (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid1,type_p1)) then 
!write (0,*) "P2SECOND CUT"
    
       do j = 1, nq1

           rto = - pos_relat + p_pos_charge1 ( :, j )
           call tr ( rto, rtn, ex, ey, ez )
           x = pos_relat_n + rtn

           call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( nb_grid1 ) )
           qf = force *  p_charge1 ( j )

           !! add to force and compute torque
           call cross (tmp1_n, x, qf )
           tt1_n = tt1_n + tmp1_n
           ff1_n = ff1_n + qf

       end do !! loop over atoms of prot 1

    ! NJB: If only smaller than the second calc that
    else if (test_dist2.LT.prot2%p_sogrid%cut2_array(nb_grid2,type_p1)) then 
!write (0,*) "P2FIRST CUT"

       do j = 1, nq1

           rto = - pos_relat + p_pos_charge1 ( :, j )
           call tr ( rto, rtn, ex, ey, ez )
           x = pos_relat_n + rtn

           call force_atom_uhbd ( rtn, force, prot2 % p_sogrid % pset_grid( nb_grid2 ) )
           qf = force *  p2_charge1 ( j )
          
           !! add to force and compute torque
           call cross (tmp1_n, x, qf )
           tt1_n = tt1_n + tmp1_n
           ff1_n = ff1_n + qf

       end do
else
!write (0,*) "P2BOTH CUT"
    end if

end subroutine compute_force_Couple_sdamm_fast


END MODULE mod_compute_force_sdamm
