!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_image_charge.f90
!! Compute image charge for metal surface (2 proteins and sdamm)

!> Group computation of force and energy when image charge is used.\n
!! Some specific interactions are implemented for ProMetCS force field
!!
!! \todo implement a cutoff for image charge
MODULE mod_compute_image_charge

contains

!> force image charge, version for sda_2proteins.\n
!! Could hard-code for protein 2 only, no nned to loop because different version for SDAMM.\n
!! compute only the force on solute 2.\n
!!
!! @param tab_protein : instance of mod_array_protein
!! @param param_force_energy : instance of type_force_energy in mod_force_energy
!! @param f, t : sum of forces and torques to increment
subroutine force_image_charge ( tab_protein, param_force_energy, f, t )
    USE mod_array_protein
    USE mod_force_energy
    USE mod_protein
    IMPLICIT NONE

    type ( array_protein_type ), intent ( in ) :: tab_protein
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension( 3 ), intent (out) :: f, t

    ! keep same logic work with internal values, and add at the end
    real ( kind=8), dimension( 3 ) :: ff2, tt2

    real ( kind=8 ) :: zero_plane
    integer :: i, j

    !write(*,*)
    !write(*,*) 'force_image_charge '
    !write(*,*) 'surface_id ', param_force_energy % surface_id

    ff2 = 0.d0
    tt2 = 0.d0

    zero_plane = tab_protein % array ( param_force_energy % surface_id ) % p_sogrid % distance_max
    !write(*,*) 'zero_plane ', zero_plane

    ! array_img_chg contains the solute number to compute image charge, could be set_of_protein for sdamm
    do i= 1, size ( param_force_energy %  array_img_chg )

        j = param_force_energy %  array_img_chg ( i )
        !write(*,*) 'force image charge i ', i, ' for protein j ', j

        ! compute image_charge for one protein
        ! result is multilplied by two inside : for elec because 0.5 in input ( valid )
        ! now use edfct = 3.44 in input, reproduce sda6 results
        call force_image_charge_one_protein ( tab_protein % array ( j ), param_force_energy, &
                                              1, zero_plane, ff2, tt2 )

        !write(*,*) 'return from image_charge j', j
        !write(*,*) 'ff2 ', ff2
        !write(*,*) 'tt2 ', tt2
        f = f + ff2
        t = t + tt2

    end do

end subroutine force_image_charge

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!> Compute image charge for only one protein.\n
!! called by sda_2proteins and sdamm.\n
!!
!! Add electrostatic desolvation image charge, I am really not convinced..
!! to use ep_scale for setting or not the epsilon correction, will follow identical asymetry
!! in sda6 no correction of the torque is done, if implemented will influence only protein 1, the surface,
!! so no influence
!!
!! @param prot : instance of protein, solute to compute interaction
!! @param param_force_energy : instance of \ref mod_force_energy "mod_force_energy"
!! @param grid_number : give grid_number of the electrostatic grid
!! @param zero_plane : set z value for the surface
!! @param ff2, tt2 : value of force and torque on prot
subroutine force_image_charge_one_protein ( prot, param_force_energy, grid_number, zero_plane, &
                                            ff2, tt2 )
    USE mod_force_energy
    USE mod_protein
    IMPLICIT NONE

    type ( protein ), intent ( in ) :: prot
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    integer, intent ( in ) :: grid_number
    real ( kind=8 ), intent ( in ) :: zero_plane
    real ( kind=8 ), dimension( 3 ), intent ( out ) :: ff2, tt2

    type ( type_grid ), pointer :: p_grid_desolv
    real ( kind=8 ), dimension(:), pointer :: p_charge2, p_chargesq2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position2
    real ( kind=8 ), dimension(3) :: rto, rtn, pos_relat
    real ( kind=8 ) :: ych !q

    integer :: nq2, j
    real ( kind=8 ) :: ep_scale
    real ( kind=8 ), dimension(3) ::  exi, eyi, ezi, xei, yei, zei
    real ( kind=8 ), dimension(3) ::  p2, qf, force,tmp2, force_o

    ! for test, to delete
    !real ( kind=8 ), dimension(3) :: sum_image_charge, sum_image_charge_desolv, sum_epsilon

    !write(*,*) 'force image charge one protein '
    !write(*,*) 'zero plane ', zero_plane
    !write(*,*) 'include desolvation : ', param_force_energy % iimg_chg_desolv

!-----define -vectors from p2 to p1-------------------
    ff2 = (/0., 0., 0./)
    tt2 = (/0., 0., 0./)
    ! only for testing
    !sum_image_charge = 0.d0
    !sum_image_charge_desolv = 0.d0
    !sum_epsilon = 0.d0

    nullify ( p_grid_desolv, p_chargesq2 )

    pos_relat = prot % position

!-----define vectors from the p2 image to p2 real-----
    exi=prot % orientation_x
    eyi=prot % orientation_y
    ezi=prot % orientation_z
    ! now we have to re-define the z component
    exi(3)=-prot % orientation_x(3) !-y(3,na2-2)
    eyi(3)=-prot % orientation_y(3) !-y(3,na2-1)
    ezi(3)=-prot % orientation_z(3) !-y(3,na2)

    xei(1)=exi(1)
    xei(2)=eyi(1)
    xei(3)=ezi(1)
    yei(1)=exi(2)
    yei(2)=eyi(2)
    yei(3)=ezi(2)
    zei(1)=exi(3)
    zei(2)=eyi(3)
    zei(3)=ezi(3)

    p_charge_position2 =>  prot % p_pos_charge
    p_charge2 => prot % p_sogrid % charge_grid
    nq2 = prot % nb_charge

    ! if image charge desolvation is needed
    if ( param_force_energy % iimg_chg_desolv ) then
        !write(*,*) ' iimg associate grid number ', get_nb_gridtype('desolvation' )
        p_grid_desolv => prot % p_sogrid % pset_grid ( get_nb_gridtype('desolvation' ) )
        p_chargesq2 => prot % p_sogrid % charge_square_grid

        !! for test only, to comment later
        !if ( ( associated ( p_chargesq2 ) .eqv. .false. ) .or. &
        !     ( associated ( p_grid_desolv ) .eqv. .false. ) ) then
        !     write(*,*) 'Error with desolvation image charge association '
        !     stop 1
        !end if
    ! compiler complains, still complains
    !else
    !    nullify ( p_chargesq2 )
    !    nullify ( p_grid_desolv )
    end if

    ! can use epsilon correction to electrostatic image_charge
    ep_scale = param_force_energy % epsilon_scale  ! parameter showing if a non-constant dielectric perm. must be used

! very similar to c2g1 asymetric, until now surface is always protein 1
!-----------------------------------------------------
! torques and force for the second protein
FORCE2: do j=1,nq2
           !write(*,*)'================'
           !write(*,*) 'j= ', j
           p2 = p_charge_position2(:,j)  ! p2(m) = y(m,j)
           !q = p_charge2 ( j )
           !write(*,*) 'p_poscharge2 ', p2
           ych=p2(3)+(pos_relat(3)-zero_plane)! ych=y(3,j)+r(3)-ABS(xc1(3))

           ! needed for desolvation as well, computed for nothing if epsilon and not desolv
           rto=  p2
           rto(3)= p2(3) +2*(pos_relat(3)-zero_plane) !y(3,j)+2*(r(3)-ABS(xc1(3)))
           !write(*,*) 'rto ', rto
           call tr(rto,rtn,exi,eyi,ezi)

           !write(*,*) 'ych ', ych
!---------- force from the image charges in the vicinity from the surface computed explicitly with variable diel. permit.
           if ( ( ep_scale > 0.0 ) .and. ( (ych >= 3.0) .and. (ych <= 5.5))) then ! added here, not in sda6 .AND.(ych.GT.2.5)) THEN
              !write(*,*) '3 <= ych <= 5.5 ', ych

              ! return directly the equivalent of qf, not multiplied by two inside
              call force_atom_variable_epsilon( prot, ych, j, zero_plane, qf , ep_scale)
              ! here multiply by 2 to make equivalent to sda6 output ( temporary equivalence ) ??
              !write(*,*) 'force epsilon', qf
              !test
              !sum_epsilon = sum_epsilon + qf

           ! else normal image charge, called even if less than 3.
           else
              ! just a test, lots of output < 3
              !if ( ych <= 3.0 ) then
              !  write(*,*) 'Warning : force_image_charge ych <= 3.0 ', ych
              !endif

!-------force_o electrostatic potential of the image
              call force_atom_uhbd ( rtn, force, prot % p_sogrid % pset_grid ( grid_number) ) !call epfor1R2(rtn,f2i)
              ! change referential corresponding to the prot2 orientation...
              call tr(force,force_o,xei,yei,zei)
              ! force * 2 correct here, because grid factor = 0.5
              ! use factor one in sda6 for elect image, ( factor 2 for desolv image and desolv )
              qf = - 2.d0 * p_charge2 ( j ) * force_o

              !test only
              !sum_image_charge = sum_image_charge  - 2.d0* p_charge2 ( j ) * force_o
              !! here to comapre to what done in sda6
              !write(*,*) 'force', 2.d0*force
           end if

           ! if asked always compute electrostatic desolvation image charge
           ! moved factor 2 in input for prometcs
           if ( param_force_energy % iimg_chg_desolv ) then
                call force_atom_uhbd ( rtn, force, p_grid_desolv )
                call tr(force,force_o,xei,yei,zei)
                !!write(*,*) 'force desolv image charge ', 2.d0 * p_chargesq2( j ) * force_o

                ! here factor 2 specific to desolvation for getting sda6 results
                ! only for prometcs
                !qf = qf + 2.d0 * p_chargesq2( j ) * force_o
                ! test factor 2 in input, here normal, ok works like this at 30 A
                qf = qf + p_chargesq2( j ) * force_o
                ! test
                !sum_image_charge_desolv = sum_image_charge_desolv + p_chargesq2( j ) * force_o
           end if

           !write(*,*) 'total force j ', qf
           call cross(tmp2, p2, qf)
           tt2 = tt2 + tmp2
           ff2 = ff2 + qf

       end do FORCE2

       !write(*,*) 'final force '
       !write(*,*) 'ff2 ',ff2
       !write(*,*) 'tt2 ', tt2
       !write(*,*) 'sum_image_charge ', sum_image_charge
       !write(*,*) 'sum_image_charge_desolv ', sum_image_charge_desolv
       !write(*,*) 'sum_epsilon ', sum_epsilon

       ! sda6 if box, no correction in eforce
       ! if want to apply correction here, only apply to tt1 ( surface do not move )
       ! so do not have influence for aubs
       ! call cross ( tmp2, pos_relat, ff2 )
       ! tt1 = -tt2 - tmp2

end subroutine force_image_charge_one_protein

!> Compute force with the analytical correction, apply to only one solute
!!
!! @param prot : instance of \ref mod_protein "protein"
!! @param ychi :
!! @param i : protein number
!! @param zero_plane : z-value of the surface position
!! @param f : return force
!! @param ep_scale : value of the factor, defined in mod_force_energy
subroutine force_atom_variable_epsilon( prot, ychi, i, zero_plane, f , ep_scale)
    USE mod_force_energy
    USE mod_protein
    IMPLICIT NONE
    type ( protein ), intent ( in ) ::  prot
    real ( kind=8 ), intent ( in ) :: ychi, zero_plane, ep_scale
    real ( kind=8 ), dimension( 3 ), intent ( out ) :: f

    integer, intent ( in ) :: i

    real ( kind = 8 ), dimension ( : ), pointer ::  p_charge
    real ( kind=8 ) :: fx, fy, fz, f1
    real ( kind=8 ), dimension(3) :: pos_relat, pii, pj
    real ( kind=8 ) :: q, qi, ychj, distij, distij2_xy
    real ( kind=8 ) ::  alph, beth, eps
    ! here hard-coded
    real ( kind=8), parameter :: ep_scaleA = 0.8d0, ep_scaleB = 0.39d0, ep_scaleC = 10.4d0
 
    integer :: nq2, j
    
    f = (/0.d0, 0.d0, 0.d0/)

    !write(*,*) 'entry force_atom_varialble epsilon, ychi ', ychi
    !write(*,*) 'i ', i
    !write(*,*) 'zero_plane ', zero_plane

    pii = prot % p_pos_charge(:,i)
    nq2 = prot % nb_charge
    p_charge => prot % p_sogrid % charge_grid
    qi = prot % p_sogrid % charge_grid( i )
    ! works becasue always proteins 2 send in sda_2protein
    pos_relat = prot % position

    eps=4.0+ep_scaleA*ychi*ychi+exp(ychi/ep_scaleB-ep_scaleC)
    !write(*,*) 'eps ', eps

    DO j=1,nq2
       !write(*,*) 'j= ',j
       f1=0.d0
       fx=0.d0
       fy=0.d0
       fz=0.d0

       pj = prot % p_pos_charge (:,j)
       !write(*,*) 'ypj ', pj

       ! ych2 distance from the effective charge j to the surface
       ychj= pj(3)+ pos_relat(3)-ABS(zero_plane)!ych2=y(3,j)+r(3)-ABS(xc1(3))
     ! moved one test here, avoid unenessecary distance computation
     if ( ( ychj <= 5.5d0 ) .and. ( ychj >= 3.d0 ) ) then

        q = p_charge(j)

        ! disij - distance between an eff. charge i and image charge from the effective carge j
        ! distij2_xy - squared distance in the xy plane
        distij2_xy = (pj(1)-pii(1))*(pj(1)-pii(1))+(pj(2)-pii(2))*(pj(2)-pii(2))
        distij = sqrt ( distij2_xy + (pj(3)-pii(3))*(pj(3)-pii(3)) )
        !write(*,*) 'ychj ', ychj
        !write(*,*) 'disttij ', distij
        ! this if statement needs attention and probably some correction in the future!
        !IF((ychj.LE.5.5).AND.(disij.LE.5.5).AND.(ychj.GE.3)) THEN  ! IF((ych2.LE.5.5).AND.(disij.LE.5.5).AND.(ych2.GE.3)) THEN the second condition may be omitted if it would improve speed ????????
        !if ( ( ychj <= 5.5d0 ) .and. ( ychj >= 3.d0 ) .and. ( distij <= 5.5d0 ) ) then
        ! here better to use 5.5*5.5 = 30.25d0, distij never used after, avoid sqrt
        if ( distij <= 5.5d0 ) then

           f1 = -553.d0 * q * qi * ep_scale * ep_scale / &
              ( eps * ((ychi+ychj)*(ychi+ychj) + distij2_xy ))
           !write(*,*) 'f1 ',f1
           alph = atan ( sqrt ( distij2_xy ) / ( ychi+ychj ) )

           ! test about alpha, not present before
           !if ( alph <= 0. ) then
           !  write(*,*) 'alpha inferior to 0 !!'
           !  stop 1
           !end if
           fz=f1*cos(alph)
           !write(*,*) 'fz ', fz

           ! strange test, if not ? fx and fy == 0. ?? negative if below the surface !!
           ! in fact test if i != j, in this case fx, fy == 0 by default
           ! appends i==j  alpha == 0 => fx=0, fy=0, avoid divide by zero
           IF( alph > 0.d0 ) THEN
              ! here protect division by 0
              beth= atan((pii(1)-pj(1))/(pii(2)-pj(2))) !atan((y(1,i)-y(1,j))/((y(2,i)-y(2,j))))

              fx = f1 * sin(alph) * sin(beth)
              if ( pii(1) > pj(1) ) fx = -fx !IF(y(1,i).GT.y(1,j)) fx=-fx
              ! why fy =fy + ?? it zero
              !fy= fy + f1 * sin(alph) * cos(beth)
              fy = f1 * sin(alph) * cos(beth)
              if (pii(2) > pj(2)) fy = -fy !IF(y(2,i).GT.y(2,j)) fy=-fy
           END IF
           ! original + !! and fx but will be divided by two after
           !f(1)=f(1) + fx*0.5 !we have to keep normalization the same as fornormal el. potential
           !f(2)=f(2) + fy*0.5
           !f(3)=f(3) + fz*0.5
           ! but factor 2 missing
           f(1)=f(1) + fx !we have to keep normalization the same as fornormal el. potential
           f(2)=f(2) + fy
           f(3)=f(3) + fz
           !write(*,*) 'j force ', j, fx, fy, fz

        END IF ! end if distij < 5.5
      END IF ! 3 <= ychj <= 5
     END DO ! loop nq2
     !write(*,*)
     !write(*,*) 'final force epsilon ', f

end subroutine force_atom_variable_epsilon

!> Force image charge, variant for sdamm.\n
!! call identical subfunctions, just add OpenMP
!!
!! @param tab_protein : instance of \ref mod_array_protein "mod_array_protein"
!! @param param_force_energy : instance of type_force_energy in mod_force_energy
!! @param all_force, all_torque : forces and torques to increment
subroutine force_image_charge_sdamm ( tab_protein, param_force_energy, all_force, all_torque )
    USE mod_array_protein
    USE mod_force_energy
    USE mod_protein
    IMPLICIT NONE

    type ( array_protein_type ), intent ( in ) :: tab_protein
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension ( : , : ), intent ( out ) :: all_force, all_torque
    !real ( kind=4 ), intent ( in ) :: Cut2

    ! keep same logic work with internal values, and add at the end
    real ( kind=8), dimension( 3 ) :: ff2, tt2

    real ( kind=8 ) :: zero_plane
    integer :: i, j

    !write(*,*)
    !write(*,*) 'force_image_charge_sdamm Cut2 ', Cut2
    !write(*,*) 'surface_id ', param_force_energy % surface_id

    ff2 = 0.d0
    tt2 = 0.d0

    zero_plane = tab_protein % array ( param_force_energy % surface_id ) % p_sogrid % distance_max
    !write(*,*) 'zero_plane ', zero_plane

    ! OpenMP easy to parrallelize if one loop over all proteins directly
    do i= 1, size ( param_force_energy %  array_img_chg )

        j = param_force_energy %  array_img_chg ( i )
        !write(*,*) 'force image charge i ', i, ' for protein j ', j

        ! compute image_charge for one protein
        ! result is multilplied by two inside : for elec because 0.5
        ! for desolv image charge, use edfct = 3.44 ( 1.67 * 2 )
        call force_image_charge_one_protein ( tab_protein % array ( j ), param_force_energy, &
                                              1, zero_plane, ff2, tt2 )

        !write(*,*) 'return from image_charge j', j
        !write(*,*) 'ff2 ', ff2
        !write(*,*) 'tt2 ', tt2
        ! no need to protect if OpenMP, only one access at a time
        all_force ( :, j ) = all_force( :, j ) + ff2
        all_torque ( :, j ) = all_torque( :, j ) + tt2

        !! here would need to protect if OpenMP
        !! apply reciproqual force to the surface, never used but consistent with tests
        !! can be commented
        all_force (  :, param_force_energy % surface_id ) = &
            all_force ( :, param_force_energy % surface_id ) - ff2

    end do

end subroutine force_image_charge_sdamm

!!!!!!========================================================================
!!!!!!!!!!!!!!!!!! Image charge energy in a high -dielectic solvent
!!!!!!!!!!!!plus a correction for variable dielectric constant !!!!!!!!!!!!!!


!> Compute energy image charge in case of sda_2proteins
!! include electrostatic desolvation image charge if needed
!! Need to check for cutoff
subroutine energy_image_charge ( tab_protein, param_force_energy, array_energy )
    USE mod_array_protein
    USE mod_force_energy
    USE mod_gridtype
    IMPLICIT NONE

    type ( array_protein_type ), intent ( in ) :: tab_protein
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension( :,: ) :: array_energy

    real(kind=8) :: energy_prot, zero_plane
    integer :: i, j
    integer  :: grid_number

    !write(*,*)
    !write(*,*) 'Entry energy image_charge '

    ! need to check Cut2_surf
    ! position of the surface
    zero_plane = tab_protein % array ( param_force_energy % surface_id ) % p_sogrid % distance_max

    grid_number = get_nb_gridtype( 'image_charge' )
    !write(*,*) 'Grid number for image ', grid_number

     do i= 1, size ( param_force_energy %  array_img_chg )

        j = param_force_energy %  array_img_chg ( i )
        !write(*,*) 'image charge ', i,' for protein j ', j

        ! compute image_charge for one protein
        call energy_image_charge_one_protein ( tab_protein % array ( j ), param_force_energy, 1, &
                                               zero_plane, energy_prot )
        ! here consider only protein 2 interacting
        array_energy ( grid_number + 1, 2 ) = energy_prot
    end do

end subroutine energy_image_charge

!!!!!!!!!!!!!!!!!!=======================================================

! version which follow sda6 edener, very different of force
subroutine energy_image_charge_one_protein ( prot, param_force_energy, grid_number, &
                                             zero_plane, energy )
    USE mod_protein
    USE mod_force_energy
    IMPLICIT NONE

    type ( protein ), intent ( in ) :: prot
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    integer, intent ( in ) :: grid_number
    real ( kind=8 ), intent ( in ) :: zero_plane
    real ( kind=8 ), intent ( out ) :: energy

    integer :: nq2, j, m
    real ( kind=8 ) :: potential, q, ych
    real ( kind=8 ) :: tmp_en, ep_scale
    real ( kind=8 ), dimension(3) :: rto, rtn, pos_relat,ex,ey,ez, p2
    type ( type_grid ), pointer :: p_grid_desolv
    real ( kind=8 ), dimension(:), pointer :: p_charge2, p_chargesq2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position2

    !! for test only should be deleted / commented
    !real(kind=8) :: sum_epsilon, sum_image_charge sum_image_charge_desolv
    !sum_epsilon = 0.d0
    !sum_image_charge = 0.d0
    !sum_image_charge_desolv = 0.d0
    !tmp_en_total = 0.d0

    nullify ( p_grid_desolv, p_chargesq2 )
    energy = 0.d0
    potential = 0.d0
    pos_relat = prot % position
    p_charge_position2 =>  prot % p_pos_charge
    p_charge2 => prot % p_sogrid % charge_grid
    nq2 = prot % nb_charge

     if ( param_force_energy % iimg_chg_desolv ) then
        !write(*,*) 'associate grid number ', get_nb_gridtype('desolvation' )
        p_grid_desolv => prot % p_sogrid % pset_grid ( get_nb_gridtype('desolvation' ) )
        p_chargesq2 => prot % p_sogrid % charge_square_grid

        !! for test only, to comment
        !if ( ( associated ( p_chargesq2 ) .eqv. .false. ) .or. &
        !     ( associated ( p_grid_desolv ) .eqv. .false. ) ) then
        !     write(*,*) 'Error with desolvation image charge association '
        !     stop 1
        !end if
    end if

    ep_scale = param_force_energy % epsilon_scale  ! parameter showing if a non-constant dielectric perm. must be used

!--------vector from image to protein-------------
    DO m=1,2
          ex(m) = prot % orientation_x(m) ! ex(m) = orientation_X(m) for each axis  all m // ex(m)=y(m,na2-2)
          ey(m) = prot % orientation_y(m)  !ey(m)=y(m,na2-1)
          ez(m) = prot % orientation_z(m)  !ez(m)=y(m,na2)
    END DO
    ex(3)=-prot % orientation_x(3)  !ex(3)=-y(3,na2-2)
    ey(3)=-prot % orientation_y(3)  !ey(3)=-y(3,na2-1)
    ez(3)=-prot % orientation_z(3)  !ez(3)=-y(3,na2)

! energy2 is the energy of the image charge of protein2 in the  grid of protein2
ENERG2: do j = 1, nq2
             ! should not be needed
             tmp_en = 0.d0

             !write(*,*)'================'
             !write(*,*) 'j ', j
             p2 = p_charge_position2 (:,j) ! !p2 = y(:,j)
             ! should move inside if
             q = p_charge2 ( j )
	    
             !ych = z distance from an effective charge to the surface plane
             ych = p2(3) + (pos_relat(3)-zero_plane) ! ych=y(3,j)+r(3)-ABS(xc1(3))
             !write(*,*) 'ych ', ych

             ! needed for desolvation as well, computed for nothing if epsilon and not desolv
             rto=  p2
             rto(3)= p2(3) +2*(pos_relat(3)-zero_plane) !y(3,j)+2*(r(3)-ABS(xc1(3)))
             !write(*,*) 'rto ', rto
             call tr(rto,rtn,ex,ey,ez)

             ! assign energy to tmp_en
             if ( ( ep_scale > 0.0 ) .and. ( (ych > 2.5) .and. (ych <= 5.5))) then

                call potential_atom_variable_epsilon( prot, j, ych , zero_plane, tmp_en ,ep_scale)
                ! for tests...
                ! sum_epsilon = sum_epsilon + tmp_en
             ! else normal image charge calculation
             else

                call potential_atom_uhbd( rtn, potential, prot % p_sogrid % pset_grid ( grid_number) )  ! CALL epenr1R2(rtn,e1)
                ! write(*,*) 'e1 ', 2.d0*potential
                ! normal here because 0.5 in input for ep
                tmp_en = - 2.d0 * p_charge2 ( j ) * potential
                ! for tests
                ! tmp_en_total = tmp_en_total  + tmp_en
                ! write(*,*) '-e1*q ', tmp_en
                ! write(*,*) 'energy elec image_charge', - 2.d0 * q * potential

             endif ! case ep_scale

    ! test distance to surface, minimum actual 2.2
    !if (ych < 2.5 ) then
       !write(*,*) "error:  PROBLEM with IMAGE-CHARGE !!!!! ",q,potential,tmp_en,rto(3),ych
    !   write(*,*) "error:  PROBLEM with IMAGE-CHARGE !!!!! ych, rto(3) ", ych, rto(3)
    !end if

             ! if asked always compute electrostatic desolvation image charge and add directly to tmp_en
             if ( param_force_energy % iimg_chg_desolv ) then
                call potential_atom_uhbd ( rtn, potential, p_grid_desolv )
                !q2 = p_chargesq2( j )
                ! use 3.44 for edfct in input
                tmp_en = tmp_en + p_chargesq2( j ) * potential

                ! only for tests
                ! sum_image_charge_desolv = sum_image_charge_desolv +  p_chargesq2( j ) * potential
             end if
           !write(*,*) 'add energy j ',j, tmp_en
           energy = energy + tmp_en
        end do ENERG2


        !write(*,*) 'computed image charge epenergy ', energy
        !write(*,*) 'sum_epsilon ', sum_epsilon
        !write(*,*) 'sum_image_charge_desolv ', sum_image_charge_desolv

    !write(*,*) '==================='
    !write(*,*) 'final energy edener'
    !write(*,*) 'sum_epsilon ', sum_epsilon
    !write(*,*) 'sum_image_charge ', sum_image_charge
    !write(*,*) 'sum_image_charge_desolv ', sum_image_charge_desolv
    !write(*,*) 'e2nergy ', energy
    !write(*,*) 'energy_ew '
    !write(*,*) 'e2nergy_i elec desolv. '
    !write(*,*) 'e2nergy_p '
    !write(*,*) 'e1nergy '

end subroutine energy_image_charge_one_protein

!!!!!!!!!!!!!!!!!!======= electrostatic interaction of the effective charge i with all image charges at small distances from the surface====================================

subroutine potential_atom_variable_epsilon( prot, i, ychi , zero_plane, potential, ep_scale )
    USE mod_protein
    IMPLICIT NONE
    type ( protein ), intent ( in ), target ::  prot
    real ( kind=8 ), intent ( in ) :: ychi, zero_plane
    real ( kind=8 ), intent ( out ) :: potential
    integer :: i

    real ( kind=8 ) :: qi, qj, ychj, distij, distij2_xy, eps
    real ( kind=8 ) :: ychD, ep_scale
    real ( kind=8 ), dimension(:,:), pointer ::  p_charge_position
    real ( kind = 8 ), dimension ( : ), pointer ::  p_charge
    real ( kind=8 ), dimension(3) ::  pos_relat,pj, pi
    integer :: nq, j

    real ( kind=8), parameter :: ep_scaleA = 0.8d0, ep_scaleB = 0.39d0, ep_scaleC = 10.4d0


    eps=4.0+ep_scaleA*ychi*ychi+exp(ychi/ep_scaleB-ep_scaleC)

    pos_relat = prot % position
    nq = prot % nb_charge
    p_charge => prot % p_sogrid % charge_grid
    ! for energy ( but not force ? need qi...could be multiplied after
    qi = prot % p_sogrid % charge_grid( i )
    p_charge_position =>  prot % p_pos_charge
    pi = p_charge_position (:,i) ! !p2 = y(:,j)

    potential = 0.d0

!-----interaction of the effective charge with its own image and  with images of surounding effective charges
    do j = 1, nq

        pj = p_charge_position (:,j) ! !p2 = y(:,j)
        qj = p_charge(j)
        ! ych2 distance from the effective charge j to the surface
        ychj= pj(3)+ pos_relat(3)-zero_plane !ych2=y(3,j)+r(3)-ABS(xc1(3))
        !write(*,*) 'ych2 ', ychj


        ! disij - distance between an eff. charge i and image charge from the effective carge j
        ! distij2_xy - squared distance in the xy plane
        distij2_xy = (pj(1)-pi(1))*(pj(1)-pi(1))+(pj(2)-pi(2))*(pj(2)-pi(2))
        distij= sqrt((pj(3)-pi(3))*(pj(3)-pi(3))+distij2_xy)  !sqrt((y(1,i)-y(1,j))*(y(1,i)-y(1,j))+(y(2,i)-y(2,j))*(y(2,i)-y(2,j))+ (y(3,i)-y(3,j))*(y(3,i)-y(3,j)))
        !write(*,*) 'disij ', distij

        if ( (ychj < 5.5d0 ) .and. (distij < 5.5d0 ) ) then  ! the second condition may be omitted if it would improve speed ????????
            ychD=sqrt((ychi+ychj)*(ychi+ychj)+ distij2_xy)
            ! for test
            !write(*,*) 'ychD ', ychD
            !! if follow sda6, need qi as well, could multiply by 553*qi at the end
            potential = potential-553.0d0*ep_scale*ep_scale*qi*qj/(eps*ychD)
            !write(*,*) 'ychj < 5.5, incr potential ', potential
        end if
     end do

     !write(*,*) 'potential epsilon ', potential

end subroutine potential_atom_variable_epsilon

!> Energy image charge for sdamm, need to care for the different types now.\n
!! allow last protein as surface.
!!
!! @param tab_protein : instance of \ref mod_array_protein
!! @param param_force_energy : instance of type_force_energy in mod_force_energy
!! @param array_energy : array to store energy values
subroutine energy_image_charge_sdamm ( tab_protein, param_force_energy, array_energy )
    USE mod_array_protein
    USE mod_force_energy
    USE mod_gridtype
    IMPLICIT NONE

    type ( array_protein_type ), intent ( in ) :: tab_protein
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension( :,: ) :: array_energy

    real(kind=8) :: energy_prot, zero_plane
    integer :: i, j, nb_types, type_prot
    integer  :: grid_number

    !write(*,*)
    !write(*,*) 'Entry energy image_charge_sdamm '

    zero_plane = tab_protein % array ( param_force_energy % surface_id ) %  p_sogrid % distance_max
    !write(*,*) 'zero_plane ', zero_plane

    grid_number = get_nb_gridtype( 'image_charge' )
    !write(*,*) 'Grid number for image charge ', grid_number
    !write(*,*) 'array image charge ', param_force_energy % array_img_chg

    nb_types = tab_protein % nb_grids
    !write(*,*) 'nb_types ', nb_types

    do i = 1, size ( param_force_energy %  array_img_chg )

        j = param_force_energy %  array_img_chg ( i )
        !write(*,*) 'energy image charge i ', i,' for protein j ', j
        type_prot = param_force_energy % array_prot_type ( j )
        !write(*,*) 'type of prot ', type_prot

        ! compute image_chatrge for one protein
        call energy_image_charge_one_protein ( tab_protein % array ( j ), param_force_energy, 1, &
                                               zero_plane, energy_prot )

        !write(*,*) 'return from energy ', energy_prot
        !write(*,*) 'assign to array_energy :', grid_number + 1, j
        ! seems correct version
        !write(*,*) 'assign to array_energy : ', 1 + nb_types * (grid_number-1) + type_prot

        ! assign energy terms, no risk of overlap here if OpenMP
        array_energy ( 1 + nb_types * (grid_number-1) + type_prot, j ) = energy_prot

    end do

end subroutine energy_image_charge_sdamm

END MODULE mod_compute_image_charge
