!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_metaldesolv.f90
!! Compute metal desolvation (ProMetCS) for 2 proteins and sdamm

!> Used for metal desolvation computation
MODULE mod_compute_metaldesolv

contains

!> Compute energy of metal desolvation
!!
!! param_force_energy may be needed later
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param param_metal : structure for metal_desovaltion parameters
!! @param array_energy : update the array with energies
subroutine compute_energy_metaldesolv_2prot ( tab_protein, param_metal, array_energy )
    USE com_sda
    USE mod_array_protein
    USE mod_force_energy
    IMPLICIT NONE
    type ( array_protein_type ), target :: tab_protein
    type ( parameter_metaldesolv ) :: param_metal
    real ( kind=8 ), dimension(:,:) :: array_energy

    real ( kind=8 ) :: energy
    integer :: grid_number

!    write(*,*) '!! entry compute_energy_metaldsolv_2prot '
    !! access to the surface, used for sdamm, 2 proteins be consider only the first one
!    write(*,*) 'surface_id ', param_force_energy % surface_id

    ! to know which indice to fill for energy
    grid_number = get_nb_gridtype('metal_desolvation')
!    write(*,*) 'grid_number ', grid_number

    ! call function, here always prot2 solute, prot1 always surface, easy to extend
    energy = 0.d0
    call compute_energy_metaldesolv_1prot ( tab_protein % array ( 2 ), tab_protein % array( 1 ), &
                                            param_metal, energy )

    ! here apply only to protein 2 interacting
    array_energy ( grid_number + 1, 2 ) = energy

end subroutine compute_energy_metaldesolv_2prot

!> Compute energy of metal desolvation for sdamm runs
!!
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param param_metal : structure for metal_desovaltion parameters
!! @param array_energy : update the array with energies
subroutine compute_energy_metaldesolv_sdamm ( tab_protein, param_force_energy, param_metal, array_energy )
     USE com_sda
     USE mod_array_protein
     USE mod_force_energy
     IMPLICIT NONE
     type ( array_protein_type ), target :: tab_protein
     type ( type_force_energy ), intent ( in ) :: param_force_energy
     type ( parameter_metaldesolv ), intent ( in ) :: param_metal
     real ( kind=8 ), dimension(:,:), intent ( inout ) :: array_energy

     real ( kind=8 ) :: energy
     integer :: grid_number, nb_types, type_prot, i

     !write(*,*) 'entry compute_energy_metaldsolv_sdmm '
     !! access to the surface, only one possible
!     write(*,*) 'surface_id ', param_force_energy % surface_id

     ! to know which indice to fill for energy
     grid_number = get_nb_gridtype('metal_desolvation')

     nb_types = tab_protein % nb_grids
!     write(*,*) 'nb_types ', nb_types

     !! easy to add OpenMP after
     do i = 1, size ( tab_protein % array )

        !! WRONG, should use same array than image charge
        !type_prot = param_force_energy % array_prot_type ( j )
        type_prot = 2

         if ( tab_protein % array( i ) % imetaldesolv ) then
             !! here send surface first, whatever it number ( better if last solute )
             call compute_energy_metaldesolv_1prot ( tab_protein % array( i ), &
                                  tab_protein % array ( param_force_energy % surface_id ), &                                                    
                                  param_metal, energy )

             ! assign energy, correct subtype ('normally')
             array_energy ( 1 + nb_types * (grid_number-1) + type_prot, i ) = energy
         !else nothing to do
         end if
     end do


end subroutine compute_energy_metaldesolv_sdamm

!> Compute energy only for one solute.\n
!! Called from sda and sdamm
!!
!! @param prot : mobile solute
!! @param surf : surface solute
!! @param param_metal : structure for metal_desovaltion parameters
!! @param energy : return energy
subroutine compute_energy_metaldesolv_1prot ( prot, surf, param_metal, energy )
    USE com_sda
    USE mod_protein
    IMPLICIT NONE
    type ( protein ), intent ( in ) :: prot, surf
    type ( parameter_metaldesolv ) :: param_metal
    real ( kind=8 ), intent ( out ) :: energy

    real ( kind=8), dimension( 3 ) :: x, pos_relat
    real ( kind=8 ) :: zero_plane,Zadw,Eadw,Aadw,Radw,Zadwlim,x3,RSell, area
    integer  ::  maxx_S,k,l,k1,l1,i,j,jmin,jmax,imin,imax
    real ( kind=8 ), dimension(:,:), pointer :: atoms
    real ( kind=8 ), dimension(:,:), pointer :: adw

    ! access to zero_plan
!    zero_plane = prot % array ( 1 ) % p_sogrid % distance_max
      zero_plane = surf %  p_sogrid % distance_max
! --- contact grid will be filled---------------
      Zadw =  param_metal % Zadw
      Eadw =  param_metal % Eadw
      Radw =  param_metal % Radw
      Aadw =  param_metal % Aadw

      atoms => prot % p_list_atom_surface
      !desorption energy is zero if surface-p2_atom distance is GT that Zadwlim
      Zadwlim=param_metal % Zadw + 10.d0
      maxx_S = prot % size_metal
      RSell= Radw / 3.d0
      ! to check if always valid with sdamm
      pos_relat = prot % position  ! r
      adw => prot % array_mdesolv
      ! fill matrix with maximal distances
      adw = Zadwlim+1.0  ! must be larger than Zadwlim
      imin=maxx_S/2
      imax=maxx_S/2
      jmin=maxx_S/2
      jmax=maxx_S/2
!
! --- protein-metal contact grid will be filled---------------
!---------p2 is placed at the  center of the grid----------
! 
! return if :
! i) surface grid  is less than 2x2; 
!      IF((maxx_S.LE.2).OR.(maxy_S.LE.2)) RETURN
! ii) or if number of protein surface atoms is 0; 
!      IF (ns2.LT.1) RETURN
! iii) or protein does not reach layer of adsorbed water on the surface
!      IF((r(3)-dismx2).GT.(Zadwlim+ABS(xc1(3)))) RETURN

! --- contact grid will be filled---------------


      imin=maxx_S/2
      imax=maxx_S/2
      jmin=maxx_S/2
      jmax=maxx_S/2
      DO 202 j=1,prot % nb_atom_surface !DO 202 j=1,ns2
        x(1)=prot % p_list_atom_surface(1,j) !x(1)=y(1,j+nq2)
        x(2)=prot % p_list_atom_surface(2,j) !x(2)=y(2,j+nq2)
        x(3)=pos_relat(3)+prot % p_list_atom_surface(3,j) !x(3)=r(3)+y(3,j+nq2)
     ! distance from the metal surface
        x3=x(3)-zero_plane  
     ! if protein atom does not reach the surface continue
        IF(x3.GT.Zadwlim)   GOTO 202
     ! else find contact position; p2 is placed at the  center of the grid !
        k=maxx_S/2+int(x(1)/RSell)
        l=maxx_S/2+int(x(2)/RSell)
    ! check if protein is in the box (xmin-xmax)x(ymin-ymax) and count contacts
    !        IF((k.GT.1).AND.(l.GT.1)
   !     +     .AND.(k.LT.maxx_S-1).AND.(l.LT.maxx_S-1)) THEN ! .AND.(k.LT.maxx_S-1).AND.(l.LT.maxy_S-1)) THEN 
   !---define  contact area boundaries----only to save computer time in the next calculation step !!-----
   !            IF(adw(k,l).GE.Zadwlim)  iadwat=iadwat+1
   ! store min surface -atom distance for the sell (k,l) in the matrix adw(k,l)
         DO k1=k-1,k+1
            DO l1=l-1,l+1
              IF(x3.LT.adw(k1,l1)) THEN
                 adw(k1,l1)=x3   ! fill grid with min distances
                 IF(k1.LE.imin) imin=k1
                 IF(k1.GE.imax) imax=k1
                 IF(l1.LE.jmin) jmin=l1
                 IF(l1.GE.jmax) jmax=l1
!                      write(*,*) "AREA / ENERGY = ",x3,k1,l1
                 END IF
             END DO
         END DO
! else error message and stop
!        ELSE
!           WRITE(6,*) 'PROTEIN IS OUT OF the SURFACE AREA (x/y):'
!     +,x(1),x(2)
!        END IF
        
 202  CONTINUE

!----compute desorption and electrostatic energy---
!---if some water molecules are replaced by a protein----
      energy=0.d0
      area = 0.d0
      k = 0
      l = 0
!      IF(iadwat.EQ.0) RETURN  
!-- calculate water desorption energy for the 1st and 2d water layers (at 0-Zadw and Zadw-Zadwlim, respectively) 
!        write(6,*) 'LIMITS:',imin,imax,jmin,jmax,iadwat
      DO  i=imin,imax
          DO  j=jmin,jmax     
!----look for the closest p2-surface contact within the sell RadwxRadw ( that is 3x3 in our case)
             x3=adw(i,j)
             IF(x3.LE.Zadw) THEN
               energy=energy+Eadw*RSell*RSell
               area = area + 1
               k = k +1
             ELSE
               IF(x3.LT.Zadwlim) THEN   
                  energy=energy+Eadw*RSell*RSell*exp(-(x3-Zadw)/Aadw)
                  area = area + 1
                  l = l +1
!                  write(*,*) "AREA / ENERGY = ",x3,l
               END IF
             END IF
           END DO
       END DO

!=================================================================
!     write(*,*) "AREA / ENERGY = ",pos_relat(3),area,energy,k,l

end subroutine compute_energy_metaldesolv_1prot


!> Compute force of metal desolvation
!!
!! This function is commented in the MainLoop, strange dynamics.
!!
!! param_force_energy may be needed later
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param param_metal : structure for metal_desovaltion parameters
!! @param force, torque : update force and torque
subroutine compute_force_metaldesolv_2prot( tab_protein, param_metal,force,torque )
    USE com_sda
    USE mod_array_protein
    USE mod_force_energy
    IMPLICIT NONE
    type ( array_protein_type ), target :: tab_protein
    type ( parameter_metaldesolv ) :: param_metal
    real ( kind=8), dimension( 3 ) :: force, torque


    real ( kind=8), dimension( 3 ) :: f, t
    integer :: grid_number

!!    write(*,*) 'entry compute_force_metaldesolv_2prot '
!!    write(*,*) 'force_image_charge_sdamm Cut2 ', Cut2
!!    access to the surface, used for sdamm, 2 proteins be consider only the first one
!!    write(*,*) 'surface_id ', param_force_energy % surface_id

    ! easy for 2 proteins, it is fixed
    grid_number = get_nb_gridtype('metal_desolvation')
!!    write(*,*) 'grid_number ', grid_number

    ! here always prot2 solute, prot1 surface
    call compute_force_metaldesolv_1prot ( tab_protein % array ( 2 ), tab_protein % array( 1 ), &
                                            param_metal, f, t )

    ! here apply only to protein 2 interacting
    force = force + f
    torque = torque + t

end subroutine compute_force_metaldesolv_2prot

!> Compute force and torque of metal desolvation for sdamm runs
!!
!! This function is commented in the MainLoop, strange dynamics
!!
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param param_metal : structure for metal_desovaltion parameters
!! @param force, torque : increment force and torque arrays
subroutine compute_force_metaldesolv_sdamm( tab_protein, param_force_energy, param_metal, force,torque )
    USE com_sda
    USE mod_array_protein
    USE mod_force_energy
    IMPLICIT NONE
    type ( array_protein_type ), target :: tab_protein
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    type ( parameter_metaldesolv ) :: param_metal
    real ( kind=8), dimension(:,:) :: force, torque

    real ( kind=8), dimension( 3 ) :: f, t
    integer :: i

    !! write(*,*) 'Entry compute_force_metaldesolv_sdamm '
    !! access to the surface, used for sdamm, 2 proteins be consider only the first one
    !! easy for 2 proteins, it is fixed
    !! write(*,*) 'surface_id ', param_force_energy % surface_id

    !! easy to add OpenMP after
    do i = 1, size ( tab_protein % array )

        ! do not apply if this is the surface
        if ( tab_protein % array( i ) % imetaldesolv ) then
            !write(*,*) 'protein i=',i

            ! call function, here always prot2 solute, prot1 surface
            call compute_force_metaldesolv_1prot ( tab_protein % array ( i ), &
                                                   tab_protein % array( param_force_energy % surface_id ), &
                                                   param_metal, f, t )

            ! here apply only to protein 2 interacting
            force(:,i) = force(:,i) + f
            torque(:,i) = torque(:,i) + t

            ! copy form image charge, for unit-test only
            ! here would need to protect if OpenMP
            ! apply reciproqual force to the surface, never used but consistent for unit-tests
            ! can be commented
            force (  :, param_force_energy % surface_id ) = &
                force ( :, param_force_energy % surface_id ) - f
        end if
    end do

end subroutine compute_force_metaldesolv_sdamm

!> Compute force only for one solute.\n
!! Called from sda and sdamm
!!
!! @param prot : mobile solute
!! @param surf : surface solute
!! @param param_metal : structure for metal_desovaltion parameters
!! @param ff2, tt2 : return force and torque on solute 2
subroutine compute_force_metaldesolv_1prot ( prot, surf, param_metal, ff2,tt2)

    USE mod_array_protein
    USE mod_force_energy
    USE mod_protein
    USE com_sda

    IMPLICIT NONE
    type ( protein ), intent ( in ) :: prot, surf
    type ( parameter_metaldesolv ) , intent ( in ) :: param_metal

    real ( kind=8 ) :: zero_plane
    ! used only for tests
    real ( kind=4 ) ::  areaz

    ! keep same logic work with internal values, and add at the end
    real ( kind=8), dimension( 3 ) , intent ( out ):: ff2, tt2
    real ( kind=8 ), dimension(3) ::  p2, tmp2, f,x,pos_relat

    real ( kind=8 ) :: fz,x3,x1,x2 ! fx,fy
    real ( kind=8 ) :: RSell,Zadw,Eadw,Aadw,Radw,Zadwlim
    integer  ::  maxx_S,k,l,k1,l1,i,j,jmin,jmax,imin,imax
    real ( kind=8 ), dimension(:,:), pointer ::  atoms
    real ( kind=8 ), dimension(:,:), pointer :: adw

!    write(*,*)
!    write(*,*) 'force_metaldesolv_1prot '

     ff2 = 0.d0
     tt2 = 0.d0
     f = 0.d0

      zero_plane = surf %  p_sogrid % distance_max
! --- contact grid will be filled---------------
      Zadw =  param_metal % Zadw
      Eadw =  param_metal % Eadw
      Radw =  param_metal % Radw
      Aadw =  param_metal % Aadw
    
      atoms => prot % p_list_atom_surface
      Zadwlim=param_metal % Zadw + 10.d0 !desorption energy is zero if surface-p2_atom distance is GT that Zadwlim
      maxx_S = prot % size_metal
      RSell= Radw / 3.d0
      ! maybe not always valid with sdamm, if surface is not shifted to zero
      pos_relat = prot % position  ! r
      adw => prot % array_mdesolv
      ! fill matrix with maximal distances
      adw = Zadwlim + 1
      ! same value ! why different variables
      imin=maxx_S/2
      imax=maxx_S/2
      jmin=maxx_S/2
      jmax=maxx_S/2

      DO 202 j=1,prot % nb_atom_surface !DO 202 j=1,ns2
        x(1) = atoms(1,j) !x(1)=y(1,j+nq2)
        x(2) = atoms(2,j) !x(2)=y(2,j+nq2)
        x(3) = pos_relat(3)+atoms(3,j) ! x(3)=r(3)+y(3,j+nq2)
     ! distance from the metal surface
        x3=x(3)-zero_plane  
     ! if protein atom does not reach the surface continue
        IF(x3.GT.Zadwlim)   GOTO 202
     ! else find contact position; p2 is placed at the  center of the grid !
        k=maxx_S/2+int(x(1)/RSell)
        l=maxx_S/2+int(x(2)/RSell)
    ! check if protein is in the box (xmin-xmax)x(ymin-ymax) and count contacts
    !        IF((k.GT.1).AND.(l.GT.1)
   !     +     .AND.(k.LT.maxx_S-1).AND.(l.LT.maxx_S-1)) THEN ! .AND.(k.LT.maxx_S-1).AND.(l.LT.maxy_S-1)) THEN 
   !---define  contact area boundaries----only to save computer time in the next calculation step !!-----
   !            IF(adw(k,l).GE.Zadwlim)  iadwat=iadwat+1
   ! store min surface -atom distance for the sell (k,l) in the matrix adw(k,l)
         DO k1=k-1,k+1
            DO l1=l-1,l+1
              IF(x3.LT.adw(k1,l1)) THEN
                 adw(k1,l1)=x3   ! fill grid with min distances
                 IF(k1.LE.imin) imin=k1
                 IF(k1.GE.imax) imax=k1
                 IF(l1.LE.jmin) jmin=l1
                 IF(l1.GE.jmax) jmax=l1
              END IF
             END DO
         END DO
        
 202  CONTINUE
 
!----compute ddesolvation force  (only z component!)---
!---if some water molecules are replaced by a protein----
      x2 = 0.d0
      x1 = 0.d0 
      x3 = 0.d0
      !area1x = 0.0
      !area2x = 0.0
      !area1y = 0.0
      !area2y = 0.0
      ! only for debugging
      !areaz  = 0.0
!-- calculate difference of the water desorption energies 
!for the 1st and 2d water layers (at 0-Zadw and Zadw-Zadwlim, respectively) 
      DO  i=imin,imax
          DO  j=jmin,jmax     
!----look for the closest p2-surface contact within the sell RadwxRadw ( that is 3x3 in our case)
             x3 = adw(i,j)
             fz = 0.d0
! --- protein atom moves inside the 1st water shell - no forces
             IF((x3.LE.Zadwlim).AND.(x3.GT.(Zadw))) THEN
                  fz = exp(-(x3-Zadw)/Aadw)/Aadw
                  areaz = areaz +1
             END IF
! I think the following block does not make sense because the surface is homogeneous (i.e. must be fx=fy=0)
!             fx = 0.d0
!             fy = 0.d0
!             x1 = adw(i-1,j)
!             x2 = adw(i+1,j)
!             IF((x1.LE.(Zadw)).AND.(x2.LE.(Zadw))) THEN
!                fx = exp(-(x2-Zadw)/Aadw)-exp(-(x1-Zadw)/Aadw)
!                area1x = area1x +1
!             ELSE
!                IF((x1.LE.(Zadw)).AND.(x2.LE.(Zadwlim))) THEN
!                    fx = exp(-(x2-Zadw)/Aadw)-1.0
!                    area1x = area1x +1
!                ELSE
!                   IF((x2.LE.(Zadw)).AND.(x1.LE.(Zadwlim))) THEN
!                         fx = 1.0-exp(-(x1-Zadw)/Aadw)
!                         area2x = area2x +1
!                   END IF
!                END IF
!             END IF
!             x1 = adw(i,j-1)
!             x2 = adw(i,j+1)
!             IF((x1.LE.(Zadw)).AND. (x2.LE.(Zadw))) THEN
!                 fy = exp(-(x2-Zadw)/Aadw)-exp(-(x1-Zadw)/Aadw)
!                    area1y = area1y +1
!             ELSE
!                IF((x1.LE.(Zadw)).AND.(x2.LE.(Zadwlim))) THEN
!                       fy=exp(-(x2-Zadw)/Aadw)-1.0
!                    area1y = area1y +1
!                ELSE
!                  IF((x2.LE.(Zadw)).AND.(x1.LE.(Zadwlim))) THEN
!                      fy = 1.0-exp(-(x1-Zadw)/Aadw)
!                         area2y = area2y +1
!                  END IF
!                END IF
!             END IF

   ! computed force is in the system of coordinates of the surface (i.e. protein 1)
!           fx =  Eadw*RSell*RSell*fx/RSell
!           fy =  Eadw*RSell*RSell*fy/RSell
!           f(1) = fx
!           f(2) = fy
           fz =  Eadw*RSell*RSell*fz
           f(3) = fz ! is positive, should be ok
           ! michael f = - dE/dz
           !f(3) = -fz
    ! we have to restore position of the atom that contact the surface in to the cell adw(i,j)
           p2(1)=(i-maxx_S/2)*RSell
           p2(2)=(j-maxx_S/2)*RSell
           p2(3)=zero_plane+x3-pos_relat(3)
           call cross(tmp2, p2, f)
!           write(*,*) "P: ",p2,f,tmp2

           ff2 = ff2 + f
           tt2  = tt2 + tmp2
           END DO
       END DO
       ff2 = 0.0d0
       tt2 = 0.0d0
       !write(*,*) 'metaldesolv ff2 ', ff2
       !write(*,*) 'metaldesolv tt2 ', tt2
!      write(*,*) "Z/AREA/FORCE/TORQUE = ",pos_relat(3),areaz,ff2,tt2

end subroutine compute_force_metaldesolv_1prot

END MODULE
