!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_pmf.f90
!! group functions for the computation of PMF on 

MODULE mod_compute_pmf
implicit none

contains

!> Main subroutine to compute the pmf in each dZ distance above the surface.\n
!! Called by MainLoop_sda_energy.f90
subroutine main_pmf(tab_protein,param_force_energy,mentx,menty,mentth,mentfi,mentom,ent, &
                    surf_atoms,charge_points, cmx, dx,dy,dT,erotnorm,energymin,irtst, &
                    dS,dZ,wz,pz,sz,dist,dummy_array,param_metaldesolv,type_calc,o_complexes)
    
!$  USE omp_lib
    USE mod_array_protein ! include mod_setofgrid, which includes mod_grid
    USE mod_gridtype
    USE mod_geometry
    USE mod_record
    USE mod_protein
    USE mod_force_energy
    USE mod_compute_energy
    USE mod_set_parameters
    USE mod_compute_metaldesolv
    USE mod_compute_image_charge
    USE com_sda
    USE mod_type_calc
    USE AllTimer
    IMPLICIT NONE

    interface
 
        subroutine check_exclusion ( prot1, prot2, rtry, rot_try, overlap, do_rot )
            USE mod_protein
            IMPLICIT NONE
            type ( protein ), intent( in ), target :: prot1, prot2
            real ( kind=8 ), dimension(3), intent( in ) :: rtry
            real ( kind=8  ), dimension (3,3), intent( in ) :: rot_try
            logical , intent ( out ) :: overlap
            logical, intent ( in ) :: do_rot
        end subroutine check_exclusion

    end interface

    type ( record ), intent ( in ) :: o_complexes
    type ( array_protein_type ), intent ( inout ), target :: tab_protein
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    type ( type_calculation ), intent ( in ) :: type_calc
    type ( parameter_metaldesolv ) :: param_metaldesolv

    real ( kind=8 ), dimension(:,:), pointer:: array_energy
    real ( kind = 8 ), dimension ( : ), intent ( in ) :: dummy_array
    real ( kind=4 ), intent ( in ) :: dist

    integer, intent ( IN ) :: mentth, mentfi, mentom
    integer, intent ( IN ) :: mentx, menty
    integer, intent ( INOUT ) :: irtst
    real ( kind=8 ), dimension (:,:,:,:,:), intent ( INOUT ) :: ent
    real ( kind=8 ), dimension (:,:), intent ( INOUT ) :: surf_atoms,charge_points
    real ( kind=8 ), dimension (15), intent ( INOUT ) :: cmx
    real ( kind=8 ), intent ( IN ) :: dx, dy, dZ, dS
    real ( kind=8 ), intent ( OUT ) :: energymin, wz, pz, sz
    real ( kind=8 ), intent ( INOUT ) :: dT, erotnorm
    
    real ( kind=8 ), dimension ( 3,3 ) :: rot
	
    logical :: overlap
    integer :: ix, iy, ith, ifi, iom, n_av, l, ierr
    integer :: size_array_energy, max_energy_terms
    real ( kind=8 ) :: PI_C, piGr, irtst3, total_energy, energy_av
    real ( kind=8 ) :: rotentth, rotentfi, rotentom, eee
    type ( mytimer ) :: timer_thread

    !Only for OpenMP
!$  integer :: tid
!!$  integer :: merge_step, merge_step_shared, size_thread_complexe

    !! make a general pointer ptab_protein, allows to use one variable only, needed for openmp
    type ( array_protein_type ), pointer :: ptab_protein
    !! local variables, used for simplicity
    type ( protein ), pointer :: protein1, protein2

    PI_C = acos(-1.0)
    piGr = PI_C/180.
    energymin=0.0
    erotnorm = 0.0
    irtst3 = irtst*irtst*irtst

    energy_av = 0.d0
    n_av = 0

    size_array_energy = 0
    max_energy_terms = 0

    nullify ( ptab_protein )
    nullify ( protein1, protein2 )

    !!initialize the energy matrix
    !! store only zeros in the matrix!
    ent = 0.d0

    !write(*,*) 'z here is: ', tab_protein % array (2) % position(3)
!!!!!!!!!!  For debugging
!    iom = 4
!    ifi = 4
!    ith = 2
!    ix  = 2
!    iy  = 2
!    rot = 0.d0
!    call set_array_energy ( type_calc, o_complexes, size ( tab_protein % array ), &
!                            array_energy, max_energy_terms )
!    size_array_energy = get_nbterms_energy_compatible ( o_complexes, .true. )
!    rotentom = real( iom-1 ) * 2.d0 * PI_C / real(mentom)
!    rotentfi = real( ifi-1 ) * 2.d0 * PI_C / real(mentfi)
!    rotentth = real( ith-1 ) * PI_C / real(mentth)
!    call rotation_pmf(rot, rotentfi,rotentth,rotentom)
!    call update_rotation_pmf ( tab_protein % array (2), rot, surf_atoms,charge_points )
!    tab_protein % array (2) %  orientation_x = (/rot(1,1),rot(1,2),rot(1,3)/)
!    tab_protein % array (2) %  orientation_y = (/rot(2,1),rot(2,2),rot(2,3)/)
!    tab_protein % array (2) %  orientation_z = (/rot(3,1),rot(3,2),rot(3,3)/)
!
!    array_energy = 0.d0
!    tab_protein % array (2) % position(1) = real(ix-1) * dx
!    tab_protein % array (2) % position(2) = real(iy-1) * dy
!    call compute_distance ( tab_protein % array (1) , tab_protein % array (2) , dist )
!    call energy_epedhdlj_2proteins ( tab_protein % array (1),tab_protein % array (2), param_force_energy, &
!         dist, array_energy, dummy_array )
!    if ( param_force_energy % iimg_chg ) then
!        call energy_image_charge ( tab_protein, param_force_energy, array_energy )
!    endif
!    if ( param_force_energy % imetaldesolv ) then
!        call compute_energy_metaldesolv_2prot( tab_protein, param_metaldesolv, array_energy )
!    end if
!    total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
!    write (*,'(5i2, 6f8.4)') iom, ifi, ith, ix, iy, &
!                             total_energy, array_energy(7,1)+array_energy(7,2),&
!                             array_energy(3,1)+array_energy(3,2), array_energy(4,1)+array_energy(4,2), &
!                             array_energy(6,1)+array_energy(6,2), array_energy(8,1)+array_energy(8,2)
!    STOP 1
!
!!!!!!!!!!!!!

!$OMP PARALLEL  DEFAULT(NONE) &
!$OMP SHARED (tab_protein,type_calc,o_complexes, cmx, dummy_array, &
!$OMP         ent,mentth,mentfi,mentom,irtst,erotnorm,energymin, PI_C, & 
!$OMP         mentx,menty,param_force_energy,param_metaldesolv, &
!$OMP         dx,dy,dZ,dS,irtst3, ierr, surf_atoms, charge_points, dist) &
!$OMP PRIVATE(iom,ifi,ith,rotentth,rotentfi,rotentom,total_energy,rot, array_energy, &
!$OMP         ptab_protein,protein1,protein2, max_energy_terms, size_array_energy, &
!$OMP         overlap,ix,iy,dT,eee,tid,timer_thread)

!$  tid = omp_get_thread_num()
!!$  write(*,*) 'tid ', tid

    nullify ( protein1, protein2 )
    nullify ( ptab_protein )
    nullify ( array_energy )

    ! local copy for tab_protein, master points to the shared
    ! associate pointer ptab_protein to a copy of initial tab_protein if OpenMP, master points to the original
!$  if ( tid > 0 ) then
!! !$      write (*,*) 'openmp activated make copy_array_protein for tid', tid
!$      call copy_array_protein ( tab_protein, ptab_protein, type_calc, param_metaldesolv )
    !! normal if not OpenMP or master
!$  else
        ptab_protein => tab_protein
!$  end if

    protein1 => ptab_protein % array ( 1 )
    protein2 => ptab_protein % array ( 2 )

!$  if ( tid > 0 ) then
!$    protein1 % position = tab_protein % array ( 1 ) % position
!$    protein2 % position = tab_protein % array ( 2 ) % position
!$  end if

!$OMP BARRIER

    rot = 0.d0

    call set_array_energy ( type_calc, o_complexes, size ( ptab_protein % array ), &
                            array_energy, max_energy_terms )
    !! specific to sda_2proteins, need the larger one as intermediate
    size_array_energy = get_nbterms_energy_compatible ( o_complexes, .true. )

!$OMP  DO  SCHEDULE( STATIC ) REDUCTION(+:erotnorm) 
    do iom=1,mentom,irtst
        !call init_mytimer(timer_thread)
        !call start_mytimer(timer_thread)
        rotentom = real( iom-1 ) * 2.d0 * PI_C / real(mentom)
        do ifi=1,mentfi,irtst
            rotentfi = real( ifi-1 ) * 2.d0 * PI_C / real(mentfi)
            innermost: do ith=1,mentth,irtst
                rotentth = real( ith-1 ) * PI_C / real(mentth)
                total_energy=0.0
!$POMP INST BEGIN(rotation)
                !!compute the rotation matrix for the given angles
                call rotation_pmf(rot, rotentfi,rotentth,rotentom)

                !! check for overlap
                !call check_exclusion ( protein1, protein2, protein2 % position, &
                !                       rot, overlap, .true. )
                !if (overlap .eqv. .false.) cycle innermost
                
                !!rotate the system of coordinates of protein
                !!and update the orientation vectors
                call update_rotation_pmf ( protein2, rot, surf_atoms,charge_points )
                protein2 %  orientation_x = (/rot(1,1),rot(1,2),rot(1,3)/)
                protein2 %  orientation_y = (/rot(2,1),rot(2,2),rot(2,3)/)
                protein2 %  orientation_z = (/rot(3,1),rot(3,2),rot(3,3)/)
!$POMP INST END(rotation)
                do ix=1,mentx,irtst
                    do iy=1,menty,irtst

                        array_energy = 0.d0 ! array energy? look at the declaration!
                        !!update the position of the center of the protein
                        protein2 % position(1) = real(ix-1) * dx
                        protein2 % position(2) = real(iy-1) * dy
                        !tab_protein % array ( 2 ) % position(3) = 

                        !!compute interaction energy and image charge energy
!$POMP INST BEGIN(energy)
                        call energy_epedhdlj_2proteins ( protein1,protein2, param_force_energy, &
                             dist, array_energy, dummy_array )
!$POMP INST END(energy)
!$POMP INST BEGIN(image)
                        if ( param_force_energy % iimg_chg ) then
                            call energy_image_charge ( ptab_protein, param_force_energy, array_energy )
                        endif
!$POMP INST END(image)
!$POMP INST BEGIN(mdesolv)
                        if ( param_force_energy % imetaldesolv ) then
                            !write(*,*) 'Will compute metal desolvation energy'
                            !! energy updated inside
                            call compute_energy_metaldesolv_2prot( ptab_protein, param_metaldesolv, array_energy )
                        end if
!$POMP INST END(mdesolv)
                        total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
                        !write(*,*) total_energy

                        !!update the energy matrix
                        ent(ith,ifi,iom,ix,iy) = total_energy
                        !! compute normalization factors
                        dT = abs( sin( rotentth ) ) * ( 2.d0*PI_C/mentfi ) * ( PI_C/mentth ) * &
                            ( 2.d0*PI_C/mentom ) * real(irtst3)
                        eee = dT*dS* exp( -1.d0 * total_energy )
                        erotnorm = erotnorm + eee

                        !write (*,'(5i2, 6f8.4)') iom, ifi, ith, ix, iy, &
                        !         total_energy, array_energy(7,1)+array_energy(7,2),&
                        !         array_energy(3,1)+array_energy(3,2), array_energy(4,1)+array_energy(4,2), &
                        !         array_energy(6,1)+array_energy(6,2), array_energy(8,1)+array_energy(8,2)
                      
!$OMP CRITICAL
                        !!Keep record of the details of the minimum energy orientation
                        !!find min energy at each z and save conformation in the file
                        if (energymin .gt. total_energy) then
                            energymin = total_energy
                            !save record!!
                            do l=1,3
                                cmx(l)   = protein2 % position(l)
                                cmx(3+l) = protein2 % orientation_x(l)
                                cmx(6+l) = protein2 % orientation_y(l)
                            end do
                            cmx(10) = total_energy
                            cmx(11) = array_energy( 6, 1 ) + array_energy ( 6 , 2 )
                            cmx(12) = array_energy(7,1)+array_energy (7,2)+array_energy(8,1)+array_energy(8,2)
                        end if
!$OMP END CRITICAL
                    end do
                end do
            end do innermost
        end do
        !call stop_mytimer (timer_thread)
        !call print_mytimer(timer_thread)
    end do
!$OMP END DO NOWAIT

!!    if array_energy allocated explicitely inside parrallel
    if ( associated ( array_energy ) ) then
        deallocate ( array_energy, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy '
            STOP 1
        end if
        nullify ( array_energy )
    end if

    nullify ( protein1, protein2 )
!$  if ( tid > 0 ) then
!!      !option true for NOT deleting the grid
!$      call delete_array_protein ( ptab_protein, .true. )
!! now need explicit deallocation of the pointer
!$      deallocate ( ptab_protein,  stat = ierr )
!$      if ( ierr /= 0 ) then
!$          write (*,*) 'ERROR In allocating ptab_dest '
!$          STOP 1
!$      end if
!$      nullify ( ptab_protein )
!$  end if

!!    call stop_mytimer (timer_thread)
!!    call print_mytimer(timer_thread)
!$OMP END PARALLEL
!    STOP 1
    !! compute PMF and dS at current z
    !! compute parameters integrated over z: probability, lowest PMF and lowest energy
    if (energymin .lt. 0.0) then
        !write(*,*) 'energymin: ' , energymin
        !!Here compute the pmf value
        call comp_pmf(ent,mentth,mentfi,mentom,mentx,menty,dS,wz,pz, sz,erotnorm,irtst)
    end if    
end subroutine main_pmf

!> Subroutine to compute the pmf value for a given z above the surface.\n
!! rotational and translation configurational distributions c and 
!! write occupation map (translational and rotational)
!! norm to 1; dV=dT*dz- 3-dimentional space volume = zmax*4pi
!!
!! ppp*dT*dZ is the probability of  finding the protein at Y+dT, Z+dZ
!! DH- change of enthalpy, DSt - change of translational entropy; DSr- rot entropy
subroutine comp_pmf(ent,mentth,mentfi,mentom,mentx,menty,dS,wz,pz, sz,erotnorm,istp)
    IMPLICIT NONE

    integer, intent (INOUT) :: istp
    integer, intent (IN) :: mentth, mentfi, mentom, mentx, menty
    real ( kind=8 ), dimension (:,:,:,:,:), intent ( IN ) :: ent
    real ( kind=8 ), intent (IN) :: erotnorm, dS
    real ( kind=8 ), intent (INOUT) :: wz, pz, sz

    integer :: ix, iy, p, k, j
    real ( kind=8 ) :: PI_C, piGr, rotentth, rotentfi, rotentom
    real ( kind=8 ) :: dT, dS1, ppp

    PI_C = acos(-1.0)
    piGr = PI_C/180.
    
    wz = 0.0
    sz = 0.0
    pz = 0.0
	
    if(istp .lt. 1) istp = 1

    dS1 = dS*real(istp*istp)

    do p=1,mentom,istp     
        rotentom = (p-0.5)*2*PI_C/mentom
        do k=1,mentth,istp
            rotentth = (k-0.5)*PI_C/mentth
            do j=1,mentfi,istp
                rotentfi = (j-0.5)*2*PI_C/mentfi
                dT = real(istp*istp*istp)*ABS(sin(rotentth)*(2*PI_C/real(mentfi))*(PI_C/real(mentth))*(2*PI_C/real(mentom)))
                do ix=1,mentx,istp ! energy sampled over xy 
                    do iy=1,menty,istp ! energy sampled over xy 
                        !! probability factor	
                        ppp = exp(-ent(k,j,p,ix,iy)) ! energy sampled over xy 
                        !! local enthalpy 
                        wz = wz+dT*ppp*dS1*ent(k,j,p,ix,iy)/erotnorm
                        !! local entropy 
                        if (ppp.gt.1.e-20)   then ! energy sampled over xy 
                            sz=sz-dT*(ppp/erotnorm)*dS1*log(8*PI_C*PI_C*ppp*(dS*real(mentx*menty))/erotnorm)
                        end if
                        pz=pz+ppp*dT*dS1  ! energy sampled over xy 
                    end do
                end do
            end do
        end do
    end do  ! energy sampled over xy 
    return
end subroutine comp_pmf

!> compute the rotation matrix for a given set of angles.\n
subroutine rotation_pmf(rot, rotentfi,rotentth,rotentom)

    IMPLICIT NONE
    real ( kind=8 ), intent (IN) :: rotentfi, rotentth, rotentom

    real ( kind=8 ) PI_C, c1, c2, c3, s1, s2, s3
    real ( kind=8 ), dimension ( 3 ) :: om
    real ( kind=8 ), dimension ( 3,3 ), intent (INOUT) :: rot

    PI_C = acos(-1.0)
    om(1)=rotentfi
!   rotation around x		 
    om(2)=rotentth 
!   rotation around old z
    om(3)=rotentom
!   rot around new z
    c1 = dcos(om(1))
    s1 = dsin(om(1))
    c2 = dcos(om(2))
    s2 = dsin(om(2))
    c3 = dcos(om(3))
    s3 = dsin(om(3))
    rot(1,1) = c1*c3-s1*c2*s3
    rot(1,2) = c1*s3+s1*c2*c3
    rot(1,3) = s1*s2
    rot(2,1) = -s1*c3-c1*c2*s3
    rot(2,2) = -s1*s3+c1*c2*c3
    rot(2,3) = s2*c1
    rot(3,1) = s2*s3
    rot(3,2) = -s2*c3
    rot(3,3) = c2
    return
end subroutine rotation_pmf

!> get new protein axis and atoms coordinate after rotation and update.\n
subroutine update_rotation_pmf ( prot2, matrix_rot, surf_atoms, charge_points )
    USE mod_protein
    IMPLICIT NONE
    type ( protein ), intent( inout ) :: prot2
    real ( kind=8 ), dimension(3,3), intent( in ) :: matrix_rot
    real ( kind=8 ), dimension (:,:), intent ( in ) :: surf_atoms,charge_points

    real ( kind=8 ) :: xold, yold, zold
    integer :: ia

    !! update charge_pos, faster to test nb_charge ??
    if ( associated (  prot2 % p_pos_charge ) ) then
        !write(*,*) 'charge'
        do ia = 1, prot2 % nb_charge
            xold = charge_points (1, ia )
            yold = charge_points (2, ia )
            zold = charge_points (3, ia )
            prot2 % p_pos_charge (1, ia) = matrix_rot(1,1)*xold + &
                                           matrix_rot(2,1)*yold + &
                                           matrix_rot(3,1)*zold
            prot2 % p_pos_charge (2, ia) = matrix_rot(1,2)*xold + &
                                           matrix_rot(2,2)*yold + &
                                           matrix_rot(3,2)*zold
            prot2 % p_pos_charge (3, ia) = matrix_rot(1,3)*xold + &
                                           matrix_rot(2,3)*yold + &
                                           matrix_rot(3,3)*zold
            !write(*,*) prot2 % p_pos_charge (1, ia),  prot2 % p_pos_charge (2,ia), prot2 % p_pos_charge (3, ia)
        end do
    endif

    if ( associated ( prot2 % p_list_atom_surface ) ) then
        !write(*,*) 'atoms'
        do ia = 1, prot2 % nb_atom_surface
            xold = surf_atoms (1, ia)
            yold = surf_atoms (2, ia)
            zold = surf_atoms (3, ia)
            prot2 % p_list_atom_surface (1, ia) = matrix_rot(1,1)*xold + &
                                                  matrix_rot(2,1)*yold + &
                                                  matrix_rot(3,1)*zold
            prot2 % p_list_atom_surface (2, ia) = matrix_rot(1,2)*xold + &
                                                  matrix_rot(2,2)*yold + &
                                                  matrix_rot(3,2)*zold
            prot2 % p_list_atom_surface (3, ia) = matrix_rot(1,3)*xold + &
                                                  matrix_rot(2,3)*yold + &
                                                  matrix_rot(3,3)*zold
        end do
    endif

end subroutine update_rotation_pmf

END MODULE mod_compute_pmf
