!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_grid.f90
!!
!! Define 3D UHBD

! Summary uhbd format
! Seems important to transform e12.6 to e12.5, but not clear if true, only for header ??
! for resid3d, no matter of .5 or .6 but always positive, maybe error

! http://www.ks.uiuc.edu/Research/vmd/plugins/doxygen/uhbdplugin_8C-source.html
! VMD use scale, first real, if 0 ascii, if != 0 binary and applies the format
! in uhbdplugin ( VMD )
! float scale; // 0 for ascii files, otherwise signals binary and provides
!              // scale factor for written data., but wrong !!
! in fact try to read and check with fread

!> Module for dealing with 3 dimensional grids
MODULE mod_grid

    !integer, parameter :: enum_uhbd = 1, enum_dtgrid = 2

    ! derived class, uhbd type can be an array of real or logical values
    type type_uhbd_grid
        real ( kind=8 ), dimension(:,:,:), pointer :: pgrid_real
        logical, dimension(:,:,:), pointer :: pgrid_logical
    end type type_uhbd_grid

    !> Contains all data related to grid
    !!
    !! Most are common to both types : origin, size, h, etc\n
    !! Grids can be real or logical, be read/writen in binary or ascii format
    !!
    !! Only the pointer to the real array may be of type UHBD or dtgrid
    type type_grid

        !> origin of the grid
        real ( kind=8  ), dimension(3) :: origin
        !> multiplicative factor read in the input
        real ( kind=8 ) :: factor
        !> spacing of the grid, and scale read from uhbd header but not used
        !! @{
        real ( kind=8 ):: h, scale
        !> @}
        ! hp = h, hpin = 1 / hp, rp = max (rx = ( o_grid%size_x - 1.d0)*h3)/2. , rpsq =rp^2
        ! but default real ( k=4 )
        !> precomcomputed values
        !! @{
        real :: hpin, rp, rpsq
        !> @}

        !!! grid_type : uhbd or dtgrid
        !> refer to mod_gridtype
        integer :: grid_type
        !> if the format is ascii ( true ) or binary ( false )
        integer :: iform
        !> size of the grid, in number of cells
        !! @{
        integer  :: size_x, size_y, size_z
        !> @}
        !> not used, to delete ?
        !! @{
        integer :: npx, npy, npz !! use in force calculation = always equal to im3 ?? TO DELETE
        !> @}

        character*128 filename
        character*72  title

        !> pointer to the 3d array
        !! @{
        type ( type_uhbd_grid ), pointer :: p_grid
        !> @}

        !> parameter uhbd or dtgrid
        !integer :: type_of_grid
        !> scaling factot, used ?
        real ( kind=8 ) :: scfct
        !> if set to yes, store logical values, otherwise real number
        logical :: is_logical

    end type type_grid

    contains

        !> Allocate grid
        !!
        !! @param this : instance of type_grid
        !! @param xmax, ymax, zmax : size of the grid in number of cells
        !! @param opt_logical : if values stored are logical. Default real*4 in disk, real*8 in memory
        subroutine allocate_grid ( this, xmax, ymax, zmax, opt_logical )
            IMPLICIT NONE
            type ( type_grid ) ,intent( inout ) :: this
            !integer :: io_unit
            integer, intent ( in ) :: xmax, ymax, zmax

            logical, intent ( in ), optional :: opt_logical
            logical :: be_logical

            be_logical = .false.
            if ( present ( opt_logical ) ) then
                be_logical = opt_logical
            end if
            this % is_logical = be_logical
            !write(*,*) 'is logical ', this % is_logical

            ! in case called directly
            nullify ( this % p_grid )

            allocate ( this % p_grid )
            ! assign size if not done in read, for instance in make_exclusion_grid
            this % size_x = xmax
            this % size_y = ymax
            this % size_z = zmax
            call allocate_grid_uhbd ( this % p_grid, xmax, ymax, zmax, this % is_logical )

            !! should add default behaviour, for fully constructed grid

        end subroutine allocate_grid

        !> Allocate the 3d imensional uhbd grid
        !!
        !! Called by \ref allocate_grid
        subroutine allocate_grid_uhbd ( this, xmax, ymax, zmax, be_logical )
            IMPLICIT NONE
            type (type_uhbd_grid) ,intent(inout) :: this
            integer, intent ( in ) :: xmax, ymax, zmax
            logical, intent ( in ) :: be_logical

            integer :: ierr=1

            ! write(*,*) 'Entry allocate_grid_uhbd '
            nullify ( this % pgrid_real )
            nullify ( this % pgrid_logical )

            if ( be_logical .eqv. .false. ) then
                !write(*,*) 'allocate tye_uhbd_grid % pgrid_real'
                allocate( this % pgrid_real ( xmax, ymax, zmax), stat=ierr)
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR Couldn t allocate memory for grid, size=',xmax,ymax,zmax
                    stop 1
                endif
                this % pgrid_real = 0.d0

            else
                allocate( this % pgrid_logical ( xmax, ymax, zmax), stat=ierr)
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR Couldn t allocate memory for grid, size=',xmax,ymax,zmax
                    stop 1
                endif
                this % pgrid_logical = .false.
            end if

        end subroutine allocate_grid_uhbd

        !> Delete the grid
        !!
        !! deallocate 3 dimensional grid
        !! @param this : instance of type_grid
        subroutine delete_grid ( this )
            IMPLICIT NONE
            type ( type_grid ) :: this
            integer :: ierr = 1

            if ( associated ( this % p_grid ) ) then
                call delete_uhbd_grid ( this % p_grid )
                deallocate ( this % p_grid, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In delalocating p_grid'
                    STOP 1
                end if
                nullify ( this % p_grid )
            end if

        end subroutine delete_grid

        !> Delete UHBD grid
        !! Called by \ref delete_grid
        subroutine delete_uhbd_grid ( this )
            IMPLICIT NONE
            type ( type_uhbd_grid ) :: this

            integer :: ierr = 1

            if ( associated ( this % pgrid_real ) ) then
                deallocate ( this % pgrid_real , stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocate pgrid_real'
                    stop 1
                end if
                nullify ( this % pgrid_real )
            end if

            if ( associated ( this % pgrid_logical ) ) then
                deallocate ( this % pgrid_logical , stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocating pgrid_logical'
                    stop 1
                end if
                nullify ( this % pgrid_logical )
            end if

        end subroutine delete_uhbd_grid

        !> function to test if the 3d grid is correctly initiliazed
        !!
        !! Will check if the correct pointer is associated
        !! @param this : instance of type_grid
        !! @return is : true or false
        function is_associated ( this ) result ( is )
            IMPLICIT NONE
            type ( type_grid ), intent ( in ) :: this

            logical :: is

            is = .false.
            if ( associated ( this % p_grid ) )  then
                is = .true.
            end if

        end function is_associated

        !> function to test the type of grid
        !! variant return type_of_grid
        !! @param this : instance of type_grid
        !! @return is : 1 is UHBD, 2 is dtgrid
        !function is_associated_int ( this ) result ( is )
        !    IMPLICIT NONE
        !    type ( type_grid ), intent ( in ) :: this
        !    integer :: is

        !    is = 0

        !    if ( associated ( this % p_grid ) ) then
        !        is = 1
        !    else if ( associated ( this % p_dtgrid ) ) then
        !        is = 2
        !    end if

        !end function is_associated_int


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! End allocation and delete !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        !> Read grid file
        !!
        !! @param this : instance of type_grid
        !! @param filename : name of the file
        !! @param grid_type_opt : optionaly the grid number, \ref mod_gridtype "mod_gridtype"
        !! @param scfct_opt : optionaly if a multiplicative factor must be applied to all grid points. Default 1
        !! @param read_header_only_opt : optionaly just print information of the header file. Do not load the grid
        !! @param iform_opt : optionaly if the file is ascii ( 1 ) or binary ( 0 ). Not necessary for uhbd, will call \ref test_format_grid
        subroutine read_grid ( this, filename, grid_type_opt, scfct_opt, read_header_only_opt, iform_opt )
            IMPLICIT NONE
            type ( type_grid ) :: this
            character*128 filename
            integer , intent (in), optional :: iform_opt
            integer ,intent(in), optional :: grid_type_opt
            real ( kind=8 ) ,intent(in), optional :: scfct_opt
            logical, intent(in), optional  :: read_header_only_opt

            integer :: grid_type
            logical :: ex ,read_header_only
            real ( kind=8 ) :: rx, ry, rz
            integer :: status, i, j , k, l, kx, im, jm, km
            ! dangerous
            integer :: io_unit=1111

            ! ok works, but still memcpy_fast with intel compiler
            integer :: im_buf
            real ( kind=4 ), dimension(:), allocatable :: xbuf

            ! deal with default argument
            if ( present(grid_type_opt) ) then
                grid_type = grid_type_opt
            else
                grid_type = 0
            endif

            if ( present(scfct_opt) ) then
                this % scfct = scfct_opt
            else
                this % scfct = 1.
            endif

            ! The option is to be removed ?!
            if ( present(read_header_only_opt) ) then
                read_header_only = read_header_only_opt
            else
                read_header_only = .false.
            endif

            !!Check if file exist now !! whatever format
            inquire ( file=filename, exist = ex )
            if ( .NOT.ex ) then
                write(*, * ) 'ERROR Grid does not exist or grid name is missing:',trim(filename),'!!'
                stop 1
            endif

            !! assign filename
            this % filename = filename

            ! not needed anymore, test_format all the time in sda.
            ! Keep option if needed later, not fully tested
            if ( present ( iform_opt ) ) then
                if ( iform_opt == 0 ) then
                    this % iform = 0
                else if ( iform_opt == 1 ) then
                    this % iform = 1
                endif
            ! default
            else
                call test_format_grid ( this, io_unit )
            end if

            ! read_header
            call read_header_grid ( this, io_unit, status )
            if ( status /= 0 ) then
                    write(*,*) 'ERROR In reading header'
                    return
            endif
            !write(*,*) 'Read header scale:', this%scale, 'size_x', this%size_x, 'origin_x',this%origin(1)

            if ( read_header_only .eqv. .true. ) return

            !! else continue and allocate the grid
            call allocate_grid ( this, this % size_x, this % size_y, this % size_z )

            !write(*,*) 'test grid type_of_grid ', this % type_of_grid
            !! if uhbd and real, make function read_file_data

            im = this % size_x
            jm = this % size_y
            km = this % size_z

            im_buf = im*jm
            allocate ( xbuf ( im_buf ) )
            !write(*,*) 'this % scfct', this % scfct

            ! if binary
            if ( this % iform == 0 ) then
                do k = 1, km
                    read (io_unit) kx, im, jm
                    ! valgrind warning, Source and destination overlap in memcpy(0x48b0d28, 0x48b0de4, 8004)
                    ! shoud use other function, impossible in fortran ??
                    read (io_unit) (xbuf(l),l=1,im*jm)
                    do l = 1, im * jm
                        j = (l-1)/im + 1
                        i = l - (j-1)*im
                        ! the multiplicative factor has to be added
                        this % p_grid % pgrid_real ( i, j, k ) = xbuf(l) * this % scfct
                    end do
                    !write(*,*) 'xbuf(1)', xbuf(1)
                    !write(*,*) this % p_grid % pgrid_real ( 1,1,k)
                end do
            ! ascii
            else if ( this % iform == 1 ) then
                do k = 1, km
                    read (io_unit, fmt='(3i7)' ) kx, im, jm
                    read (io_unit, fmt='((6 (1x,e12.5)))') (xbuf(l),l=1,im*jm)
                    do l = 1, im * jm
                        j = (l-1)/im + 1
                        i = l - (j-1)*im
                        ! the multiplicative factor has to be added
                        this % p_grid % pgrid_real ( i, j, k ) = xbuf(l) * this % scfct
                    end do
                end do
            endif

            deallocate ( xbuf )
            close ( io_unit )


            !set up last variables
            this % factor = this % scfct
            this % grid_type = grid_type
            this % hpin = real ( 1. / this % h, 4 )
            rx = ( this % size_x - 1.d0) * this % h
            ry = ( this % size_y - 1.d0) * this % h
            rz = ( this % size_z - 1.d0) * this % h
            this % rp =  real ( MAX ( rx, ry, rz ) / 2.d0 , 4 )
            this % rpsq = ( this % rp )**2
            !write(*,*) 'rx ', rx
            !write(*,*) 'rz ', rz
            !write(*,*) 'rp ', this % rp
            !write(*,*) 'rpsq ', this % rpsq

        end subroutine read_grid

        !> Test for uhbd and dtgrid
        !!
        !! not used anymore, call directly the grid normally
        function get_value_real_uhbd ( this, i, j, k ) result ( value )
            IMPLICIT NONE
            type ( type_uhbd_grid ), intent ( in ) :: this
            integer, intent ( in ) :: i, j, k

            real ( kind=8 ) :: value
            value = this % pgrid_real ( i,j,k )

        end function get_value_real_uhbd

        !> Test function
        !!
        !! keep as intermediate, efficient call get directly the value
        function get_value ( this, i, j, k ) result ( value )
            IMPLICIT NONE
            type ( type_grid ), intent ( in ) :: this
            integer, intent ( in ) :: i, j, k

            real ( kind=8 ) :: value

            !if ( associated ( this % p_grid % pgrid_real ) ) then
            value = this % p_grid % pgrid_real (  i, j, k )
            !end if

        end function get_value

        !> Copy a grid
        !!
        !! Used by tools only \n
        !! Not a real copy, only the members are copied\n
        !! The pointer will use the original grid\n
        !! The new grid can modify some data members ( iform, origin ) and write a new version
        !!
        !! May add an option real copy, or internal % real_copy, to known if need to delete\n
        !! @param this : instance of type_grid, source grid
        !! @param outgrid : instance of the target grid
        !! @param opt_form : optionaly provides the new format. By default use the source format
        subroutine copy_grid ( this, outgrid, opt_form )
            IMPLICIT NONE
            type ( type_grid ) :: this, outgrid
            integer, optional :: opt_form

            if ( present ( opt_form ) ) then
                outgrid % iform = opt_form
            else
                outgrid % iform = this % iform
            end if

            outgrid % filename = this % filename
            outgrid % title = this % title
            !! 3dimension vector
            outgrid % origin = this % origin
            outgrid % h = this % h
            !outgrid % hp = this % hp
            outgrid % size_x = this % size_x
            outgrid % size_y = this % size_y
            outgrid % size_z = this % size_z

            outgrid % grid_type = this % grid_type
            outgrid % hpin = this % hpin
            outgrid % rp = this % rp
            outgrid % rpsq = this % rpsq

            outgrid % scale = this % scale
            outgrid % is_logical = this % is_logical
            nullify ( outgrid % p_grid )

            ! points to the same array by default, not a real copy !
            outgrid % p_grid => this % p_grid
            write(*,*) 'associated outgrid % p_grid ', associated ( outgrid % p_grid )
            write(*,*) 'associated outgrid % p_grid % pgrid_logical ', associated ( outgrid % p_grid % pgrid_logical )
            write(*,*) 'associated outgrid % p_grid % pgrid_real ', associated ( outgrid % p_grid % pgrid_real )

        end subroutine copy_grid

        !> Multiply all grid points by a scaling factor
        !!
        !! Apply only to UHBD grid
        !! @param this : instance of type_grid
        !! @param scaling : multiplicative factor
        subroutine grid_scale ( this, scaling )
            IMPLICIT NONE
            type ( type_grid ) :: this
            real ( kind=8 ) :: scaling

            if ( this % is_logical .eqv. .false. ) then
                this % p_grid % pgrid_real = this % p_grid % pgrid_real * scaling
            else
                write(*,*) 'ERROR Grid is logical, cannot apply scale factor'
                stop 1
            end if

        end subroutine grid_scale

        !> Exclude potential interior of grid solute
        !!
        !! this is a pointer to the input grid, can be overloaded in this case
        !!
        !! @param this : instance of type_grid, pointer to the input grid called by convert_grid -exclude can be overloaded in this case
        !! @param exclud_grid : grid already loaded in convert grid
        subroutine exclude( this, exclud_grid )
            IMPLICIT NONE
            type ( type_grid ) :: this, exclud_grid
            integer :: i,j,k

            !write(*,*) 'entry mod_grid exclude '
            !write(*,*) 'is logical ', exclud_grid % is_logical

            !! strange format, 1 is empty space, 0 are a few points inside the protein
            if ( this % is_logical .eqv. .false. ) then
                !this % p_grid % pgrid_real = this % p_grid % pgrid_real * scaling
                !where (a.neq.0.0)
                !a = 1.0/a
                !elsewhere
                !a = 1.1
                !end where
                !this % p_grid % pgrid_real = product( :xclud_grid % p_grid % pgrid_real )
                do i = 1, this % size_x
                    do j = 1, this % size_y
                        do k = 1, this % size_z
                           this % p_grid % pgrid_real(i,j,k) = this % p_grid % pgrid_real(i,j,k) * &
                                                               exclud_grid % p_grid % pgrid_real(i,j,k)
                        end do
                    end do
                end do

            else
                write(*,*) 'ERROR Exclude is a logical grid, cannot not apply exclude function'
                stop 1
            end if

        end subroutine exclude

        !> Set to 0. all points further than a cutoff.\n
        !! The cutoff is the maximum shpere we can include  inside the cube
        !! @param this : instance of type_grid
        subroutine make_sphere ( this )
            IMPLICIT NONE
            type ( type_grid ) :: this

            real ( kind=8 ), dimension(3) :: center
            integer :: cutoff
            integer :: i, j, k

            ! max diameter is the size/2 = cutoff
            ! cutoff based on indice is certainly easier
            ! to check if odd or even.
            cutoff = min ( this % size_x/2, this % size_y/2, this % size_z/2 )

#ifdef DEBUG
            write(*,*) 'Make_sphere'
            write(*,*) 'cutoff in cell ', cutoff
            write(*,*) 'cutoff in angstrom ', cutoff * this % h
#endif

            center(1) = this % size_x/2
            center(2) = this % size_y/2
            center(3) = this % size_z/2
            cutoff = cutoff*cutoff
#ifdef DEBUG
            write(*,*) ' center ', center
            write(*,*) 'cutoff ', cutoff !,sqrt(cutoff)
#endif

            ! basic algo
            do i = 1, this % size_x
              do j = 1, this % size_y
                do k = 1, this % size_z
                    if ( (i-center(1))**2 + (j-center(2))**2 + (k-center(3))**2 .gt. cutoff ) then
                        this % p_grid % pgrid_real (i,j,k) = 0.
                    end if
                end do
              end do
            end do

        end subroutine make_sphere

        !> Error in make ep_grid, grdflag is a good variable for binary or ascii, but works only if created by module
        !! no rules for uhbd ( 1 ) or apbs ( -1 )
        !! use title of electrostatic grid to decide
        !! by default in sda, force to read in binary, can be changed now

        !> Test the format of a grid file
        !!
        !! Test if the grid is writen in binary or ascii format
        !! Trick : I use the filed 4 ( called grdflag ) to store this information in the grid writen by this software\n
        !! Does not apply to electrostatic grids produced by UHBD or APBS. Need to use header content
        !!
        !! @param this : instance of type_grid
        !! @param io_unit : file descriptor of the file
        subroutine test_format_grid ( this, io_unit )
            IMPLICIT NONE
            type ( type_grid ) :: this
            integer :: io_unit

            integer :: ierr, io
            character*72 :: tmp_title
            real ( kind=4 ) :: tmp_real, tmp_real2
            integer :: grdflag

            !! use to test binary or ascii of electrostatic potential
            character*72, parameter :: str_uhbd='Potential Grid                                                          '
            character*72, parameter :: str_apbs='                                                        POTENTIAL (kT/e)'

            !write(*,*) 'Entry test_binary_grid'

            !! test binary format
            open (io_unit, file= this % filename , form='unformatted', status='old', iostat = ierr )
            if ( ierr /= 0 ) write(*,*) 'Cannot open file in binary format but file present, filename: ', this % filename
            read ( io_unit, iostat = io ) tmp_title, tmp_real, tmp_real2, grdflag

            if ( io /= 0 ) then
                write(*,*) 'cannot read file filename:', trim ( this % filename )
                stop 1
            end if
            !! difficult to get an error
            !write (*,*)
            !write (*,*) 'title !!!',tmp_title,'!!!!'
            !write (*,*) 'tmp_real ', tmp_real, tmp_real2
            !write (*,*) 'grdflag ', grdflag
            !write (*,*) 'io ', io

            !! test first with fixed title if coming from uhbd or apbs
            if ( ( tmp_title == str_uhbd ) .or. ( tmp_title == str_apbs ) ) then
                !write(*,*) 'title similar, it is binary format of electrostatic grid '
                this % iform = 0
                rewind ( io_unit )
                return
            end if

           
            if ( grdflag == 0 )  then
                !write(*,*) 'grdflag == 0 seems binary assign enum_uhbd ', grdflag ! or 1
                this % iform = 0
                rewind ( io_unit )
                return
            end if

            close ( io_unit )
            !write(*,*) 'read in ascii'

            ! open in formatted mode
            open (io_unit, file= this % filename , form='formatted', status='old', iostat = ierr )
            if ( ierr /= 0 ) write(*,*) 'Cannot open file in ascii format, but file present filename: ', this % filename
            read ( io_unit, fmt='(a72,/,2e12.6,i7)', iostat = io ) tmp_title, tmp_real, tmp_real2, grdflag

            if ( io /= 0 ) then
                write(*,*) 'cannot read file filename: ', trim ( this % filename )
                close ( io_unit )
                stop 1
            end if
            !write (*,*) 'ascii title !!!', tmp_title,'!!!'
            !write (*,*) 'grdflag ', grdflag

            if ( ( tmp_title == str_uhbd ) .or. ( tmp_title == str_apbs ) ) then
                write(*,*) 'title similar, it is ascii format of electrostatic grid '
                !write(*,*) 'Warning : read
                if ( tmp_title == str_apbs ) then
                    write(*,*)
                    write(*,*) 'WARNING For sda usage, it is an APBS electrostatic grid, ', &
                               'if you re running SDA you should convert the grid to kcal before ', &
                               '(convert_grid -scale 0.6 -convert )'
                    write(*,*)
                endif
                this % iform = 1
                if ( grdflag .ge. 2 ) then
                    write(*,*) 'ERROR grdflag cannot be dtgrid '
                    STOP 1
                endif
                rewind ( io_unit )
                return
            end if

            if ( grdflag == 1 ) then
                !write(*,*) 'read tmp_real in ascii grdflag', tmp_real, grdflag
                this % iform = 1
                rewind ( io_unit )
            else
                write(*,*) 'ERROR in test_format_grid'
                close ( io_unit )
                stop 1
            end if

        end subroutine test_format_grid

        !> Read header of the file
        !!
        !! Fill data member with information of the header
        !! iform should be set correctly before calling the function
        !! @param this : instance of type_grid
        !! @param io_unit : file descriptor
        !! @param status : return error code
        subroutine read_header_grid( this, io_unit, status )
            IMPLICIT NONE
            type ( type_grid ) :: this
            integer :: status
            integer, intent(in) :: io_unit

            integer :: io
            !!dummy variable for binary reading
            real (kind=4) :: dum2,dum3,dum4,dum5,dum6,dum7,dum8
            integer :: idum1,idum2,idum3,idum4,idum5,idum6,idum7

            !need tmp variable for format
            real (kind=4) :: tmph,tmp_scale, tmp_ox,tmp_oy,tmp_oz
            !character*72 tmp_title

            status = 1 !error by default
            if ( this % iform == 0) then

                read (io_unit, iostat=io ) this % title, tmp_scale, dum2,idum1, &
                     idum2,idum3,idum4,idum5, &
                     this % size_x, this % size_y, this % size_z, tmph, &
                     tmp_ox, tmp_oy, tmp_oz, &
                     dum3,dum4,dum5,dum6,dum7,dum8,idum6,idum7
                if ( io /=0 ) then
                    write(*,*) 'ERROR Format do not match header file'
                    STOP 1
                endif

                this % scale = tmp_scale
                this % h = tmph
                this % origin(1) = tmp_ox
                this % origin(2) = tmp_oy
                this % origin(3) = tmp_oz
                !write(*,*) 'read_grid_header'
                !write(*,*) 'origin ', tmp_ox, tmp_oy, tmp_oz
                !write(*,*) 'size ', this % size_x, this % size_y, this % size_z
                !write(*,*) 'h ', this % h

            !format already tested in read_grid, read ascii
            else
                read (io_unit,77, iostat = io ) this % title,this % scale, dum2, idum1, &
                     idum2,idum3,idum4,idum5, &
                     this % size_x, this % size_y, this % size_z, this % h, &
                     this % origin(1), this % origin(2), this % origin(3), &
                     dum3,dum4,dum5,dum6,dum7,dum8,idum6,idum7
                if ( io /=0 ) then
                    write(*,*) 'ERROR Format do not match header file'
                    stop 1
                endif
             endif
             !!else no error
             status = 0
   77 FORMAT (a72/2e12.6,5i7/3i7,4e12.6/4e12.6/2e12.6,2i7)

        end subroutine read_header_grid

        !> Write a grid to a file
        !!
        !! If a pointer to a real*4 3d grid is provided in option, its content will be printed into the file\n
        !! It avoids the conversion and copy from real*4 to real*8\n
        !! Apply only to UHBD grids
        !!
        !! @param this : instance of type_grid
        !! @param filename : name of the file
        !! @param opt_modif_name : optionaly if true append a suffix ".bin.grd" or ".ascii.grd"
        !! @param opt_pgrid4 : optionaly use an external pointer to 3d dimensional grid. Default use \ref type_grid::p_grid
        subroutine write_grid ( this, filename, opt_modif_name, opt_pgrid4 )
            IMPLICIT NONE

            interface
                subroutine create_filename_new ( pdb_filename, suffix )
                    IMPLICIT NONE
                    character*128 :: pdb_filename
                    character, intent( in ) :: suffix*(*)
                end subroutine create_filename_new
            end interface

            type ( type_grid ) :: this
            character*128 :: filename
            real ( kind=4 ), dimension (:, :, :), pointer, optional :: opt_pgrid4
            logical, optional :: opt_modif_name

            logical :: bool_modif_name

            !!if opt created by external, tool or else, not from a regular grid, do not modify the name
            if ( present ( opt_modif_name ) ) then
                bool_modif_name = opt_modif_name
            else
                bool_modif_name = .false.
            end if

            !! setup format for open file
            !! if binary
            if ( this % iform == 0 ) then
                !write(*,*) 'filename', filename

                if ( bool_modif_name ) then
                    call create_filename_new ( filename, 'bin.grd' )
                end if
                !string_format='unformatted'
            !! else ascii
            else
                if ( bool_modif_name ) then
                    call create_filename_new ( filename, 'ascii.grd' )
                end if
                !string_format='formatted'
            end if

            call write_uhbd_grid ( this % p_grid, this, filename, opt_pgrid4 )

        end subroutine write_grid

        !> Write UHBD grid to a file
        !! Called by \ref write_grid
        subroutine write_uhbd_grid ( this, base, filename, opt_pgrid4 )
            IMPLICIT NONE

            type ( type_uhbd_grid ), intent ( in ) :: this
            type ( type_grid ), intent ( in ) :: base
            character*128, intent ( in ) :: filename
            real ( kind=4 ), dimension (:, :, :), pointer, optional :: opt_pgrid4

            !! if no opt-grid4, need tmp copy ,for each plan on z less consumming memory,otherwise need a full copy
            real ( kind=4 ), dimension(:,:), allocatable :: tmp_plan_grid4
            character*20 :: string_format
            integer :: i, j, k, io_unit, io, ierr

            real ( kind=4 ) :: tmp_ox,tmp_oy,tmp_oz,tmph,tmp_scale
            real ( kind=4 ), dimension ( 7 ) :: dum
            integer, dimension ( 6 ) :: idum

            io_unit = 80
            string_format=''
            !!init dummy
            idum = 0
            dum = 0.
            !! tmp_variable of kind=4, only if binary output
            tmp_ox = 0.
            tmp_oy = 0.
            tmp_oz = 0.
            tmph = 0.

            !write(*,*) 'Entry write_grid ',filename
            !bug with intel, need to print this here when writing resid_3d grid, otherwise crash !!
            !bug solved
            !write(*,*) 'associated(opt_pgrid4) ', associated(opt_pgrid4)

            !! if original grid is real*4 do not need tmp_plan_grid4, necessary only for binary
            if ( ( present ( opt_pgrid4 ) ) .eqv. .false. ) then
                !! is a usual grid, will need intermediate plan to transform to real 4

                !! error call by resid3d this is not associated !!
                if ( associated ( this % pgrid_real ) ) then
                    !write(*,*) 'allocate tmp_plan_4 ', base % size_x, base % size_y
                    allocate ( tmp_plan_grid4 ( base % size_x, base % size_y ), stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR In allocation of tmp_plan_grid4'
                        stop 1
                    end if
                    ! init to zero
                    tmp_plan_grid4 = 0.
                else
                    write(*,*) 'ERROR pgrid_real not associated and no grid as optional !!'
                    stop 1
                end if
            !! add for bug ??
            else if ( associated(opt_pgrid4) .eqv. .false. ) then
                write(*,*) 'opt_pgrid4 present but not associated'
                stop 1
            end if

            !! setup format for open file
            !! if binary
            if ( base % iform == 0 ) then
                string_format='unformatted'
            !! else ascii
            else
                string_format='formatted'
            end if

            !! open the file, correct form and correct filename
            !open ( unit=io_unit, file=filename, form=string_format, status = 'new', iostat = ierr )
            open ( unit=io_unit, file=filename, form=string_format, status = 'replace', iostat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'Cannot open file filename: ', trim ( filename ) !, 'already existing ???'
                return
            end if

            !! print header and grid data
            !! if binary
            if ( base % iform == 0 ) then
                write(*,*) 'will make binary output'

                tmp_ox = real ( base % origin ( 1 ), 4 )
                tmp_oy = real ( base % origin ( 2 ), 4 )
                tmp_oz = real ( base % origin ( 3 ), 4 )
                tmp_scale = real ( base % scale, 4 )
                tmph = real ( base % h, 4 )


                write ( io_unit,iostat=io ) base % title, tmp_scale, dum(1), base % iform , &
                        !idum(1), idum(2),idum(3),idum(4), &
                        idum(1),base % size_z, idum(2), base % size_z, &
                        base % size_x, base % size_y, base % size_z, tmph, &
                        tmp_ox, tmp_oy, tmp_oz, &
                        dum(2),dum(3),dum(4),dum(5),dum(6),dum(7),idum(5),idum(6)
                if ( io /= 0 ) then
                    write(*,*) 'ERROR In writing binary header file of grid'
                    stop 1
                end if

                do k = 1, base % size_z
                    !write(*,*) 'k ', k
                    write ( io_unit ) k, base % size_x,base % size_y
                    if ( present ( opt_pgrid4 ) ) then
                        write ( io_unit ) ((opt_pgrid4 ( i,j,k ),i=1,base % size_x),j=1,base % size_y)
                    !! else need local copy to real*4
                    else
                        !tmp_plan_grid4 = base % pgrid ( k, :, : )
                        tmp_plan_grid4 = real ( this % pgrid_real ( :, :, k ), 4 )
                        write ( io_unit ) ((tmp_plan_grid4 ( i ,j ),i=1,base % size_x),j=1,base % size_y)
                    end if

                end do

            !! if ascii
            else
                !write(*,*) 'will make ascii output'
                !write ( io_unit, fmt='(a72/2e12.6,5i7/3i7,4e12.6/4e12.6/2e12.6,2i7)', iostat=io ) &
                ! correct output for ascii format
                write ( io_unit, fmt='(a72/2e12.5,5i7/3i7,4e12.5/4e12.5/2e12.5,2i7)', iostat=io ) &
                            base % title, base % scale, dum(1), base % iform, &
                            idum(1),base % size_z, idum(2), base % size_z, &
                            base % size_x, base % size_y, base % size_z, base % h, &
                            base % origin, &
                            dum(2),dum(3),dum(4),dum(5),dum(6),dum(7),idum(5),idum(6)
                if ( io /= 0 ) then
                    write(*,*) 'ERROR In writing ascii header file of grid'
                    stop 1
                end if

                do k = 1, base % size_z
                    write ( io_unit, fmt='(3i7)' ) k, base % size_x,base % size_y
                    !!use optional argument, no problem with format in ascii, no tmp_plan_grid4
                    if ( present ( opt_pgrid4 ) ) then
                        write ( io_unit, fmt='((6 (1x,e12.5)))' ) ((opt_pgrid4 ( i,j,k ),i=1,base % size_x),j=1,base % size_y)
                    else
                        !write ( io_unit, fmt='((6 (1x,e12.5)))' ) (( (this % p_grid % pgrid_real( i,j,k )),i=1,this % size_x),j=1,this % size_y)
                      write ( io_unit, fmt='((6 (1x,e12.5)))' ) ((this%pgrid_real( i,j,k ),i=1,base % size_x),j=1,base % size_y)
                    end if
                end do
            end if

            if ( allocated ( tmp_plan_grid4 ) ) then
                deallocate ( tmp_plan_grid4 )
            end if

            close ( io_unit )

        end subroutine write_uhbd_grid

        !> Specific function for reading an exclusion grid
        !!
        !! Shortcut for exclusion grid. Always logical values
        !! Can read UHBD or dtgrid format
        !! @param this : instance of type_grid
        !! @param filename : name of the file
        !! @param prot_cm : center of geometry of the pdb file. used to recenter the grid ?? needed ??
        !! @param param_probe : instance of \ref com_sda::probe_type "probe_type"
        !! @param status : for error report
        subroutine read_excl_grid ( this, filename ,prot_cm , param_probe, status )
            USE com_sda
            IMPLICIT NONE
            type (  type_grid ) :: this
            character :: filename*(*)
            real ( kind = 8 ), dimension ( 3 ) :: prot_cm
            type ( probe_type ) :: param_probe
            logical :: status

            integer :: io, ierr, io_unit, i, j, k, kx, ky, kz
            real ( kind=4 ), dimension ( :, :), allocatable :: tmp_plan4

            real ( kind=4 ), dimension ( 3 ) :: tmp_origin
            real ( kind= 4 ) :: tmp_scale, tmph
            character ( 15 ) :: format_file
            real ( kind=4 ), dimension ( 0:7 ) :: dum
            integer, dimension ( 6 ) :: idum
            integer :: tmp_iform
            character*72 :: title

            io_unit = 91
            title = ''
            kx = 0
            ky = 0
            kz = 0
            tmp_scale = 0.
            tmph = 0.

            dum = 0.
            idum = 0
            tmp_origin = 0.
            format_file = ''
            status = .false.

            ! binary
            if ( param_probe % save_excl == 1 ) then
                !write(*,*) 'open unformatted'
                format_file = 'unformatted'
            ! ascii
            else
                !write(*,*) 'open formatted'
                format_file = 'formatted'
            end if

            open (io_unit, file=filename, form= format_file, status='old', iostat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'Cannot open file: ', trim ( filename )
                stop 1
            end if

            tmp_origin = 0.
            tmp_scale = 1.0
            tmph =  0.

            ! binary
            if ( param_probe % save_excl == 1 ) then
                !write(*,*) 'read excl in binary, default'

                !! to check if reproducible, seems important here dum(0) = 1. ( scale ) not zero
                !dum(0) = 1.
                read ( io_unit,iostat=io ) title, tmp_scale, dum(1), tmp_iform, &
                        idum(1), kx, ky, kz, &
                        kx, ky, kz, tmph, &
                        tmp_origin(1), tmp_origin(2), tmp_origin(3), &
                        !tmp_origin, &
                        dum(2),dum(3), dum(4),dum(5),dum(6),dum(7),idum(5),idum(6)
                if ( io /= 0 ) then
                    write(*,*) 'ERROR In reading binary header file of grid'
                    stop 1
                end if
            ! ascii
            else
                !write(*,*) 'read excl grid ascii '
                read ( io_unit,fmt='(a72/2e12.5,5i7/3i7,4e12.5/4e12.5/2e12.5,2i7)', iostat=io ) &
                        title, tmp_scale, dum(1), tmp_iform, &
                        idum(1), kx, ky, kz, &
                        kx, ky, kz, tmph, &
                        tmp_origin(1), tmp_origin(2), tmp_origin(3), &
                        !tmp_origin, &
                        dum(2),dum(3),dum(4),dum(5),dum(6),dum(7),idum(5),idum(6)
                if ( io /= 0 ) then
                    write(*,*) 'ERROR In reading binary header file of grid'
                    stop 1
                end if

            end if

#ifdef DEBUG
            write(*,*) 'read_excl_grid rigin ', tmp_origin
            write(*,*) 'prot_cm ', prot_cm
#endif
            !write(*,*) 'tmp_iform ', tmp_iform
            !write(*,*) ' this % type_of_grid ', this % type_of_grid

            !! compare hexclusion and probep from input
            if ( ( tmph /= param_probe % hexclusion ) .or. ( dum(2) /= param_probe % probep ) ) then
                write(*,*) 'Parameters in exclusion grid are different of the input file, need to recompute the exclusion grid'
                status = .true.
                close ( io_unit )
                return
                !stop 1
            endif

            ! allocate directly the grid
            call allocate_grid ( this, kx, ky, kz, .true. )

            !! now read data in ubhd, should be in loop
            allocate ( tmp_plan4 ( this % size_x, this % size_y ), stat =ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR in allocating tmp_plan4'
                STOP 1
            end if

            this % origin = tmp_origin - prot_cm
            this % h = tmph
#ifdef DEBUG
            write(*,*) 'assign this % origin ', this % origin
#endif

            do k = 1, this % size_z

                if ( param_probe % save_excl == 1 ) then
                    read ( io_unit ) kx, ky, kz
                    read ( io_unit ) (( tmp_plan4 ( i,j ),i=1,this % size_x),j=1,this % size_y)
                !!ascii
                else
                    read ( io_unit, fmt= '(3i7)' ) kx, ky, kz
                    read ( io_unit, fmt= '((6 (1x,e12.5)))' ) (( tmp_plan4 ( i,j ),i=1,this % size_x),j=1,this % size_y)
                end if

                do i = 1, this % size_x
                    do j = 1, this % size_y
                        if ( tmp_plan4 ( i, j ) == 1. ) then
                            this % p_grid % pgrid_logical( i, j, k ) = .true.
                        else if ( tmp_plan4 ( i, j ) == 0. ) then
                            this % p_grid % pgrid_logical ( i, j, k ) = .false.
                        else
                            write(*,*) 'ERROR In reading of exlcusion grid binary'
                            stop 1
                        end if
                    end do
                end do
            ! end do loop over k
            end do

            ! deallocate tmp_plan4
            if ( allocated ( tmp_plan4 ) ) then
                deallocate ( tmp_plan4 )
            end if

            close ( io_unit )

        end subroutine read_excl_grid

        !> Write exclusion grid
        !!
        !! @param this : instance of type_grid
        !! @param filename : name of the file
        !! @param prot_cm : center of geometry of the solute
        !! @param param_probe : instance of \ref com_sda::probe_type "probe_type"
        subroutine write_excl_grid ( this, filename, prot_cm , param_probe )
            USE com_sda
            IMPLICIT NONE
            type ( type_grid ) :: this
            character :: filename*(*)
            real ( kind = 8 ), dimension ( 3 ) :: prot_cm
            type ( probe_type ) :: param_probe

            integer :: io_unit, ierr, io, i, j, k
            real ( kind=4 ), dimension ( :, :), allocatable :: tmp_plan4
            real ( kind=4 ), dimension ( 3 ) :: tmp_origin
            real ( kind= 4 ) :: tmp_scale, tmph
            character ( 15 ) :: format_file
            real ( kind=4 ), dimension ( 0:7 ) :: dum
            integer, dimension ( 6 ) :: idum
            character*72 :: title

            io_unit = 91
            title = 'Exclusion grid'

            dum = 0.
            idum = 0
            tmp_origin = 0.

            !! binary, correct name done in prepare_atom
            if ( param_probe % save_excl == 1 ) then
                format_file = 'unformatted'
            !! ascii
            else
                format_file = 'formatted'
            end if

            !open (io_unit, file=filename, form= format_file, status='new', iostat = ierr )
            open (io_unit, file=filename, form= format_file, status='replace', iostat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'Cannot open file: ', trim ( filename )
                write(*,*) 'Already exist ???'
            end if

            allocate ( tmp_plan4 ( this % size_x,  this % size_y ) )
            tmp_plan4 = 0.

            !! binary
            if ( param_probe % save_excl == 1 ) then
                write(*,*) 'write excl in binary, default'

                tmp_origin = real ( this % origin + prot_cm, 4 )
                tmp_scale = 1.0
                tmph = real ( this % h, 4 )
                !! to check if reproducible, seems important here dum(0) = 1. ( scale ) not zero
                dum(0) = 1.

                write ( io_unit,iostat=io ) title, dum(0), dum(1), param_probe % save_excl - 1, &
                        idum(1),this % size_z, idum(2), this % size_z, &
                        this % size_x, this % size_y, this % size_z, tmph, &
                        tmp_origin(1), tmp_origin(2), tmp_origin(3), &
                        !tmp_origin, &, dum(2),dum(3)
                        param_probe % probep, dum(3) ,dum(4),dum(5),dum(6),dum(7),idum(5),idum(6)
                if ( io /= 0 ) then
                    write(*,*) 'ERROR In writing binary header file of grid'
                    stop 1
                end if

                do k = 1, this % size_z
                    write ( io_unit ) k, this % size_x,this % size_y

                    do i = 1, this % size_x
                        do j = 1, this % size_y
                            if ( this % p_grid % pgrid_logical(i,j,k) .eqv. .true. ) then
                                tmp_plan4 ( i, j ) = 1.
                            else
                                tmp_plan4 ( i, j ) = 0.
                            end if
                        end do
                    end do

                    write ( io_unit ) (( tmp_plan4 ( i,j ),i=1,this % size_x),j=1,this % size_y)
                end do

            !! else ascii
            else if ( param_probe % save_excl == 2 ) then
                write(*,*) 'write excl in ascii'
                write ( io_unit, fmt='(a72/2e12.5,5i7/3i7,4e12.5/4e12.5/2e12.5,2i7)', iostat=io ) &
                            title, dum(0), dum(1), param_probe % save_excl - 1, &
                            !idum(1),idum(2),idum(3),idum(4), &
                            idum(1),this % size_z, idum(2), this % size_z, &
                            this % size_x, this % size_y, this % size_z, this % h, &
                            this % origin + prot_cm, & !dum(2)
                            param_probe % probep, dum(3), dum(4),dum(5),dum(6),dum(7),idum(5),idum(6)
                if ( io /= 0 ) then
                    write(*,*) 'ERROR In writing ascii header file of grid'
                    stop 1
                end if

                do k = 1, this % size_z
                    write ( io_unit, fmt='(3i7)' ) k, this % size_x,this % size_y

                    do i = 1, this % size_x
                        do j = 1, this % size_y
                            if ( this % p_grid % pgrid_logical(i,j,k) .eqv. .true. ) then
                                tmp_plan4 ( i, j ) = 1.
                            else
                                tmp_plan4 ( i, j ) = 0.
                            end if
                        end do
                    end do

                    write ( io_unit, fmt='((6 (1x,e12.5)))' ) (( tmp_plan4 ( i,j ),i=1,this % size_x),j=1,this % size_y)

                end do

            else
                write(*,*) 'Problem with exclusion grid'
                stop 1
            end if

            if ( allocated ( tmp_plan4 ) ) then
                deallocate ( tmp_plan4 )
            end if

            close ( io_unit )

        end subroutine write_excl_grid

        !> Write grid to dx format
        !!
        !! Used only by the tool \ref read_record "read_record"
        subroutine write_grid_dx ( this, filename, opt_modif_name )
            IMPLICIT NONE
            type ( type_grid ) :: this
            character*128 :: filename
            logical, optional :: opt_modif_name

            integer :: ierr, io_unit = 13
            integer :: k, j, i
            !logical :: bool_modif_name

            !! always ascii ??
            if ( ( present ( opt_modif_name ) ) .and. ( opt_modif_name .eqv. .true. ) ) then
                call create_filename_new ( filename, 'dx.grd' )
            end if
            !! else keep filename

            open (io_unit, file=filename, status='replace', form='formatted', iostat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'Cannot open file ', trim ( filename )
                return
            end if

!             write header:
            write(io_unit, fmt='(''object 1 class gridpositions counts '',3i4)' ) &
                  this % size_x ,this % size_y, this % size_z
            !write(io_unit,101) ox+h,oy+h,oz+h
            write ( io_unit, fmt = '(''origin '',3e14.6)' ) this % origin
            write ( io_unit, fmt= '(''delta '',3e14.6 )' ) this % h, 0.d0, 0.d0
            write ( io_unit, fmt= '(''delta '',3e14.6 )' ) 0.d0, this % h, 0.d0
            write ( io_unit, fmt= '(''delta '',3e14.6 )' ) 0.d0, 0.d0, this % h
            write ( io_unit, fmt= '(''object 2 class gridconnections counts '',3i4)' ) &
                    this % size_x ,this % size_y, this % size_z
            write ( io_unit, fmt= '(''object 3 class array type double rank 0 items '',i9,'' data follows'')' ) &
                    this % size_x * this % size_y * this % size_z
!           write_data
            write ( io_unit, fmt='(3e14.6)' ) &
               ((( this % p_grid %pgrid_real(i,j,k),k=1, this % size_z),j=1,this % size_y ),i=1,this % size_x )

            !!close file
            close( io_unit )

 !100   format('object 1 class gridpositions counts ',3i4)
 !101   format('origin ',3e14.6)
 !102   format('delta ',3e14.6)
 !103   format('object 2 class gridconnections counts ',3i4)
 !104   format('object 3 class array type double rank 0 items ',i9,' data follows')
 !105   format(3e14.6)

        end subroutine write_grid_dx

        !> Print information about the grid into a log file
        !!
        !! Write slice of potential in fort.77 as before and convenient for comparing methods
        !! @param this : instance of type_grid
        !! @param io_unit : file descriptor of the output file. Use a log file in sda
        subroutine print_grid ( this, io_unit )
            IMPLICIT NONE
            type ( type_grid ), intent(in) :: this
            integer , intent(in) :: io_unit
            integer :: nhx, nhy, nqx, nqy, k, np2x
            WRITE (io_unit,*)
            WRITE (io_unit,*) 'electrostatic potentials (scaled; in kT)'
            WRITE (io_unit,*)

            np2x = this%size_x
            nhx = np2x/2
            nhy = np2x/2
            nqx = np2x/4
            nqy = np2x/4
            do k = 1, this % size_z

                WRITE (io_unit,502) k, this % p_grid % pgrid_real (nhx,nhy,k), &
                       this % p_grid % pgrid_real (nhx-nqx,nhy,k), this % p_grid % pgrid_real (nhx+nqx,nhy,k), &
                       this % p_grid % pgrid_real (nhx,nhy-nqy,k), this % p_grid % pgrid_real (nhx,nhy+nqy,k)
            end do
  502 FORMAT (i7,5f12.6)

        end subroutine print_grid

        !> Return the grid type as defined in \ref mod_gridtype
        function get_gridtype ( this ) result ( id )
            IMPLICIT NONE
            type (type_grid) :: this
            integer:: id

            id = this % grid_type

        end function get_gridtype

        !> Shift the origin of the grid to the given position
        !!
        !! Center is the new position of the origin. UHBD uses origin at the lower left corner of the grid
        !! @param this : instance of type_grid
        !! @param center : position of the new origin
        subroutine shift_origin( this, center )
            IMPLICIT NONE
            type (type_grid), intent(inout) :: this
            real ( kind=8 ), dimension(3), intent(in) :: center

            this % origin = this % origin - center

        end subroutine shift_origin

        !> Print origin, for debug
        subroutine print_origin( this )
            IMPLICIT NONE
            type (type_grid) :: this

            write(*,fmt='(''grid origin :'',3f10.3 )') this % origin
            write (*, fmt='(''shape      :'',3i8)' ) this % size_x, this % size_y, this % size_z
            write (*, fmt='(''spacing h  :'',f8.3)') this % h
        end subroutine print_origin

        !> Print information about the grid, may extend
        subroutine info_grid( this )
            IMPLICIT NONE
            type (type_grid) :: this

            !write (*,"(5x,a40)") 'filename: ', trim (this % filename)
            write (*, fmt='(''filename   : '',a)'   ) trim (this % filename)
            write (*, fmt='(''shape      :'',3i8)' ) this % size_x, this % size_y, this % size_z
            write (*, fmt='(''scale      :'',f8.3)') this % scale
            write (*, fmt='(''spacing h  :'',f8.3)') this % h
            write (*, fmt='(''iform      :'',i8)'  ) this % iform
            write (*, fmt='(''grid_type  :'',i8)'  ) this % grid_type
            !write (*, fmt='(''t_of_grid  :'',i8)'  ) this % type_of_grid

            call print_origin( this )
        end subroutine info_grid

END MODULE mod_grid
