!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file mod_gridtype.f90
!!
!! Enumeration and function of the different grids and charges and array used in sda

!> Module grid and charges types\n
!! Writen long time ago, convenient in force, but could evolve
MODULE mod_gridtype

    integer,parameter :: electro=1, desolv=2, hydrophob=3, repulsive_lj=4, lennardjones =5, &
                         ! needed here metal_desolv, image_charge ??
                         image_charge = 6, metal_desolv = 7, &
                         exclusion=8, accessibility=9, list_surfatom=10, charge=11, charge_square=12

    integer :: enum_grid_types(12) = [electro,desolv,hydrophob,repulsive_lj,lennardjones,&
                                      image_charge, metal_desolv, &
                                      exclusion,accessibility,list_surfatom, charge, charge_square]

    !! no need of external file, can even combine with existing
    !! normal grid use 1, 2, 3, analytical grid can use -1, -2 , -3
    integer,parameter :: debyeh=-1, other=-2
    integer :: enum_analytic_types(2) = [debyeh, other]

    !> only the first enum_grid_force_calc are used for force and energies computation
    integer :: enum_grid_force_calc = 5

    contains

        !> Print the name of the grid given its number
        subroutine print_gridtype( t_grid )
            IMPLICIT NONE
            integer, intent(in) :: t_grid
            select case( t_grid )
                case ( 1 )
                   write(*,*) t_grid,' electrostatic'
                case ( 2 )
                   write(*,*) t_grid,' desolvation'
                case ( 3 )
                    write(*,*) t_grid,' hydrophobic '
                case ( 4 )
                    write(*,*) t_grid,' repulsive_lj '
                case ( 5 )
                    write(*,*) t_grid,' lennardjones '
                 case ( 6 )
                    write(*,*) t_grid, 'image_charge '
                 case ( 7 )
                    write(*,*) t_grid, 'metal_desolvation'
                 case ( 8 )
                    write(*,*) t_grid,' exclusion '
            !write(*,*) 'type of the grid', enu
                case default
                    write(*,*) t_grid,' isunknown'
                end select
        end subroutine print_gridtype

        !> Get the grid number given its name
        !!
        !! Do not enter this function with gdb (in eclipse at least), bug because of name*(*) !!
        function get_nb_gridtype(name) result (id)
            IMPLICIT NONE
            !character, dimension(:),intent(in) :: name
            character, intent(in) :: name*(*)

            integer :: id
            !select case( name )
            if ( name == 'electrostatic' ) then
                    id=electro
            else if ( name == 'desolvation' )  then
                    id=desolv
            else if ( name == 'hydrophobic' ) then
                    id=3
            else if ( name == 'repulsive_lj' ) then
                    id=4
            else if ( name == 'lennardjones' ) then
                    id=5
            else if ( name == 'image_charge' ) then
                    id = 6
            else if ( name == 'metal_desolvation' ) then
                    id = 7
            else if ( name == 'exclusion' ) then
                    id=8
            else if ( name == 'accessibility' ) then
                    id=9
            else if ( name == 'list_surface_atom' ) then
                    id=10
            else if ( name == 'charge' ) then
                    id=11
            else if ( name == 'charge_square' ) then
                    id=12
            else
                    write(*,*) name, ' is unknow in module typegrid return -1'
                    id=-1
            endif
        end function get_nb_gridtype

        !> function could be in mod_setofgrid\n
        !! not really used
        function get_sizeof_sgrid () result( sizeof_sgrid )
            IMPLICIT NONE
            integer  :: sizeof_sgrid

            sizeof_sgrid = size ( enum_grid_types )
            ! write(*,*) 'get_sizeof_sgrid ', sizeof_sgrid

        end function get_sizeof_sgrid

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Analitycal grid, certianly can combine negative / positive number !!!!!

        !> Print the name of the grid given its number
        subroutine print_analyticaltype( code_anal )
            IMPLICIT NONE
            integer, intent(in) :: code_anal

            select case( code_anal )
                case ( -1 )
                    write(*,*) code_anal, ' debye-huckel'
                case ( -2 )
                    write(*,*)  code_anal, ' other'
                case default
                    write(*,*) code_anal,' not used or unknown'
                end select

        end subroutine print_analyticaltype

        !> Get the grid number given its name
        !!
        !! Do not enter this function with gdb (in eclipse at least), bug because of name*(*) !!
        function get_nb_analyticaltype(name) result (id)
            IMPLICIT NONE
            character, intent(in) :: name*(*)

            integer :: id
            !select case( name )
            if ( name == 'debye-huckel' ) then
               id = debyeh
            else if ( name == 'other' )  then
               id = other
            else
                write(*,*) name, ' is unknow in module typegrid return -1'
                id=-1
            endif
        end function get_nb_analyticaltype

        !> function could be in mod_setofgrid\n
        !! not really used
        function get_sizeof_analytic_grid () result( sizeof_sgrid )
            IMPLICIT NONE
            integer  :: sizeof_sgrid

            sizeof_sgrid = size ( enum_analytic_types )
            ! write(*,*) 'get_sizeof_sgrid ', sizeof_sgrid

        end function get_sizeof_analytic_grid



END MODULE mod_gridtype
