!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_hydrodynamics.f90
!! function for volume fraction and rescaling of parameters
!! HI on trans. diff and rot. diff (Tokuyama for tr. diff. and Cichocki for rot. diff.)

!> Module Hydrodynamics.\n
!! Only contain function, state independent
MODULE mod_hydrodynamics

contains

! MIC: can avoid wrapper in sending a pointer to function. See energy_flex and ptrf_energy

!! JRom: the parallelization is done differently in sda_2proteins and sdamm cases,
!! so we need this wrapper to call a specific function
!> Wrapper function
subroutine scale_diffusion_coefficients ( tab_protein, geom, cutoff_phi )
    USE mod_array_protein
    USE mod_geometry
    USE com_sda
    IMPLICIT NONE

    type ( array_protein_type ) :: tab_protein
    type ( geometry ), intent ( in ) :: geom
    real (kind=8), intent ( in ) :: cutoff_phi
    integer :: nb_prot
    
    nb_prot = size( tab_protein % array )
    if ( nb_prot .eq. 2 ) then
        call scale_diffusion_coefficients_sda2prot ( tab_protein, geom, cutoff_phi )
    else
        call scale_diffusion_coefficients_sdamm ( tab_protein, geom, cutoff_phi )
    end if

end subroutine scale_diffusion_coefficients

subroutine scale_diffusion_coefficients_sdamm ( tab_protein, geom, cutoff_phi )
    USE mod_array_protein
    USE mod_geometry
    USE com_sda
    IMPLICIT NONE

    interface

        subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
        end subroutine compute_dist_pbc

    end interface

    type ( array_protein_type ) :: tab_protein
    type ( geometry ), intent ( in ) :: geom
    real (kind=8), intent ( in ) :: cutoff_phi

    real (kind=8), dimension(3) :: pos_relat
    real (kind=8) :: vol_occ, dist2
    real (kind=4) :: dist
    integer :: i, j, nb_prot

    !4.d0/3.d0*pi
    real(kind=8), parameter :: const = 4.d0/3.d0*pi

#ifdef VERBOSE
!    write(*,*) 'Entry t_diff_coeff : scale_diff_coeff '
    write(*,*) 'cutoff_phi ', cutoff_phi
#endif
    nb_prot = size( tab_protein % array )

!$OMP PARALLEL DEFAULT ( NONE ) SHARED ( tab_protein, nb_prot, geom, cutoff_phi ) &
!$OMP PRIVATE ( pos_relat, dist, dist2 ) &
!$OMP PRIVATE ( vol_occ, i, j )

!$OMP DO SCHEDULE( STATIC )
TABPROT: &
    do i = 1, nb_prot

        ! init volume occupency
        vol_occ = 0.d0

        !! loop prot2, maybe inverting loop prot 1 and 2
        do j = 1, nb_prot
            ! only with different proteins
            if ( i /=j ) then
                ! need test surface, maybe allready inside !
                !call position_pbc ( geom, tab_protein % array ( i ) % position, tab_protein % array ( j ) % position, &
                !                    pos_relat )
                !write(*,*) 'pos_relat ', pos_relat
                ! need test surface
                !dist2 = dot_product ( pos_relat, pos_relat )
                call compute_dist_pbc ( geom, tab_protein % array ( i ), tab_protein % array ( j ), dist, pos_relat )
                dist2 = dist*dist

                ! call localphi, need dist and protein j only
                call local_volume_fraction ( tab_protein % array ( j ), dist2, cutoff_phi, vol_occ)
            end if
        end do
        ! reassing same variable
        !vol_occ = vol_occ / (4.d0/3.d0*pi*(cutoff_phi**3))
        vol_occ = vol_occ / (const*(cutoff_phi**3))
        ! write(*,*) 'vol_occ ', vol_occ

        ! @todo group both functions. hydro trans and hydro rot
        call local_trans_diff_scaling( tab_protein % array ( i ) % p_sogrid % diffusion_trans , &
                                       vol_occ, &
                                       tab_protein % array ( i ) % diff_trans )
        ! default scaling for rotation as well
        call local_rot_diff_scaling ( tab_protein % array ( i ) % p_sogrid % diffusion_rotat,&
                                      vol_occ,&
                                      tab_protein % array ( i ) % diff_rotat )

#ifdef VERBOSE
        write(*,*) 'protein ',i
        write(*,*) 'vol occ ', vol_occ
        write(*,*) 'diff0, diff_s ', tab_protein % array ( i ) % p_sogrid % diffusion_trans, &
                                     tab_protein % array ( i ) % diff_trans
        write(*,*) 'rot0, rot_s ',   tab_protein % array ( i ) % p_sogrid % diffusion_rotat, &
                                     tab_protein % array ( i ) % diff_rotat
#endif

    end do TABPROT
!$OMP END DO
!$OMP END PARALLEL


    !write(*,*) 'End scale_diff coeff'

!    vpocc = 0.d0
!        do jp=1,nprot
!          if (ip.ne.jp) then
!           call dist_ij(ip,jp,r,box,xyzij,d_ij2)
!           call localphi(r,d_ij2,r_phi,stoke_r3,stoke_r,ip,jp,vpocc)
!          endif
!        enddo
!        phi_l=vpocc/(4.d0/3.d0*pi*(r_phi**3))
!        call localdmd(phi_l,dm_ip,dm_ipl)
!        dmd=dm_ipl*dt
!        rmsnow=dsqrt(dm_ipl*2.d0*dt)
!        drd=dr_ip*(1.d0 - 0.631d0*phi_l - 0.726d0*phi_l*phi_l)

end subroutine scale_diffusion_coefficients_sdamm

!! almost the same as above but now, no parallelization per each protein pair
subroutine scale_diffusion_coefficients_sda2prot( tab_protein, geom, cutoff_phi )
    USE mod_array_protein
    USE mod_geometry
    USE com_sda
    IMPLICIT NONE

    interface

        subroutine compute_distance ( prot1, prot2, dist )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
        end subroutine compute_distance

    end interface

    type ( array_protein_type ) :: tab_protein
    type ( geometry ), intent ( in ) :: geom
    real (kind=8), intent ( in ) :: cutoff_phi

    real (kind=8) :: vol_occ, dist2
    real (kind=4) :: dist

    real(kind=8), parameter :: const = 4.d0/3.d0*pi

#ifdef VERBOSE
!    write(*,*) 'Entry t_diff_coeff : scale_diff_coeff '
    write(*,*) 'cutoff_phi ', cutoff_phi
#endif

    ! init volume occupency
    vol_occ = 0.d0

    call compute_distance ( tab_protein % array ( 1 ), tab_protein % array ( 2 ), dist )
    !write(*,*) 'dist: ', dist
    dist2 = real ( dist*dist , 8 )
#ifdef VERBOSE
    write(*,*) 'scale diff coeff dist ', dist
#endif


    ! call localphi, need dist and protein 2 only
    call local_volume_fraction ( tab_protein % array ( 1 ), dist2, cutoff_phi, vol_occ)

    ! reassing same variable
    vol_occ = vol_occ / (const*(cutoff_phi**3))
    call local_trans_diff_scaling( tab_protein % array ( 2 ) % p_sogrid % diffusion_trans, &
                                    vol_occ, tab_protein % array ( 2 ) % diff_trans )
    ! default scaling for rotation as well
    call local_rot_diff_scaling ( tab_protein % array ( 2 ) % p_sogrid % diffusion_rotat, vol_occ, &
                                    tab_protein % array ( 2 ) % diff_rotat )

#ifdef VERBOSE
    write(*,*) 'protein 2'
    write(*,*) 'vol occ ', vol_occ
    write(*,*) 'diff0, diff_s ', tab_protein % array ( 2 ) % p_sogrid % diffusion_trans, tab_protein % array ( 2 ) % diff_trans
    write(*,*) 'rot0, rot_s ',   tab_protein % array ( 2 ) % p_sogrid % diffusion_rotat, &
                                    tab_protein % array ( 2 ) % diff_rotat
#endif

    vol_occ = 0.d0
    call local_volume_fraction ( tab_protein % array ( 2 ), dist2, cutoff_phi, vol_occ)
    vol_occ = vol_occ / (const*(cutoff_phi**3))
    
    call local_trans_diff_scaling( tab_protein % array ( 1 ) % p_sogrid % diffusion_trans, &
                                    vol_occ, tab_protein % array ( 1 ) % diff_trans )
    ! default scaling for rotation as well
    call local_rot_diff_scaling ( tab_protein % array ( 1 ) % p_sogrid % diffusion_rotat, vol_occ, &
                                      tab_protein % array ( 1 ) % diff_rotat )

#ifdef VERBOSE
    write(*,*) 'protein 1'
    write(*,*) 'vol occ ', vol_occ
    write(*,*) 'diff0, diff_s ', tab_protein % array ( 1 ) % p_sogrid % diffusion_trans, tab_protein % array ( 1 ) % diff_trans
    write(*,*) 'rot0, rot_s ',   tab_protein % array ( 1 ) % p_sogrid % diffusion_rotat, &
                                    tab_protein % array ( 1 ) % diff_rotat
#endif

    ! update the relative diff.coefficient in case of 2 proteins
    tab_protein % array ( 2 ) % diff_trans = tab_protein % array ( 2 ) % diff_trans + tab_protein % array ( 1 ) % diff_trans
#ifdef VERBOSE
    write(*,*)
    write(*,*) 'HI, update prot2 diff_trans = ', tab_protein % array ( 2 ) % diff_trans

    write(*,*) 'HI, update prot2 rot_trans = ', tab_protein % array ( 2 ) % diff_rotat
    write(*,*) 'HI, update prot1 rot_trans = ', tab_protein % array ( 1 ) % diff_rotat
    write(*,*)
#endif
!    write(*,*) 'End scale_diff coeff'

end subroutine scale_diffusion_coefficients_sda2prot

!> compute volume fraction
!!
!! @param prot2 : second solute
!! @param dist2 : distance square
!! @param cutoff_phi : cutoff distance
!! @param vol_occ : return fraction of occupied volume
subroutine local_volume_fraction ( prot2, dist2, cutoff_phi, vol_occ )
    USE mod_array_protein
    IMPLICIT NONE
    type ( protein ) :: prot2
    real ( kind=8 ) :: dist2, cutoff_phi
    real ( kind=8 ) :: vol_occ

    real (kind=8 ) :: dist, srj, stoke_r3, pi

!#ifdef DEBUG
!    write(*,*) 'entry local volume fraction'
!    write(*,*) 'dist2 ', dist2
!#endif

    pi = acos(-1.d0)
    ! stoke radius of protein 2
    srj = prot2 % p_sogrid % stoke_radius
    ! to optimize
    stoke_r3 = srj*srj*srj
    !dist = sqrt( dist2 )

    ! totally included, will be recomputed many times
    if ( dist2 < (cutoff_phi - prot2 % p_sogrid % stoke_radius)**2 ) then
        vol_occ = vol_occ + 4.d0/3.d0*pi* stoke_r3 !( prot2 % stoke_r3)

    ! partially included
    else if ( dist2 < (cutoff_phi + prot2 % p_sogrid % stoke_radius)**2 ) then

        dist = sqrt( dist2 )

        vol_occ = vol_occ + ( ( pi * ( cutoff_phi + srj - dist )**2 ) * &
        ( dist2 + 2.d0*dist*srj - 3.d0* srj**2 + 2.d0*dist*cutoff_phi + 6.d0*srj*cutoff_phi - 3.d0*cutoff_phi**2 ) ) / (12.d0*dist)

    end if

!    if (dsqrtdd.lt.(r_phi-stoke_r(jp))) then
        ! Totally included proteins
!        vpocc=vpocc+4.d0/3.d0*pi*(stoke_r3(jp))
        !print *, 'TI ',ip,jp,vpocc
!       elseif (dsqrtdd.lt.(r_phi+stoke_r(jp))) then
!       !  Partially included proteins (compute spherical intersection between cutoff and jp)
!         vpocc = vpocc + ( ( pi * ( r_phi + srj - dsqrtdd )**2 ) *
!     &          (dd + 2*dsqrtdd*srj - 3*srj**2 + 2*dsqrtdd*r_phi + 6*srj*r_phi - 3*r_phi**2) ) / (12*dsqrtdd)

        !print *, 'PI ',ip,jp,dsqrtdd,r_phi,srj,vpocc
!       endif
!       return
end subroutine local_volume_fraction

!> Scale the translation diffusion coefficient according to
subroutine local_trans_diff_scaling(dm_init, vol_occ, dm_scaled)
    IMPLICIT NONE
    real ( kind=8 ), intent ( in ) :: dm_init, vol_occ
    real ( kind=8 ), intent ( out ) :: dm_scaled

    real(kind=8) :: b,b2,c,H

    real ( kind=8 ), parameter :: div_9_8 = 9.d0/8.d0, div_11_16 = 11.d0/16.d0

    b=0.d0 ! sqrt(9/8*phi_l)
    b2 = 0.d0 ! 9/8*phi_l
    c=0.d0 ! 11/16*phi_l
    H=0.d0 ! local tensor

    !b=dsqrt(9.d0/8.d0*vol_occ)
    !b2=9.d0/8.d0*vol_occ
    !c=11.d0/16.d0*vol_occ
    b = dsqrt( div_9_8 * vol_occ )
    b2 = div_9_8 * vol_occ
    c = div_11_16 * vol_occ

    H=(2.d0*b2)/(1.d0-b) - c/(1.d0+2.d0*c) - (b*c*(2.d0+c)) / ((1.d0+c)*(1.d0-b+c))

    !dm_ipl = dm_ip / (1+H)
    dm_scaled = dm_init / ( 1.d0+H )

!    write(*,*) 'b ', b
!    write(*,*) 'c ', c
#ifdef VERBOSE
    write(*,*) 'H ',H
    write(*,*) 'dm_init ', dm_init
    write(*,*) 'dm_scaled ', dm_scaled
#endif

end subroutine local_trans_diff_scaling

!> Scale the rotational diffusion coefficient according to
subroutine local_rot_diff_scaling(dr_init, vol_occ, dr_scaled)
    IMPLICIT NONE
    real ( kind=8 ), intent ( in ) :: dr_init, vol_occ
    real ( kind=8 ), intent ( out ) :: dr_scaled

    !dr_ip*(1.d0 - 0.631d0*phi_l - 0.726d0*phi_l*phi_l)
    dr_scaled = dr_init * ( 1.d0 - 0.631d0*vol_occ - 0.726d0*vol_occ*vol_occ )

#ifdef VERBOSE
    write(*,*) 'dr_init ', dr_init
    write(*,*) 'dr_scaled ', dr_scaled
#endif

end subroutine local_rot_diff_scaling

END MODULE mod_hydrodynamics


