!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_init_position.f90
!! group functions for setting the initial position

!> Groups functions for intialisation of position.\n
!! Include all random_position subroutines.
!!
!! Do not define objects
MODULE mod_init_position


private :: random_position_xyz, random_position_sphere, random_position_halfsphere, random_orientation

contains

!>  Initialize the position of the protein at each run.\n
!!
!! Used only by sda_2proteins
!! Work with a restart for sda_koff\n
!! Otherwise generate a random position and orientation\n
!! Should be thread-safe ...normally...
subroutine init_position_protein ( tab_protein, type_calc, geom, dist, iflex, current_conf, &
                                   !! for sda_koff p_tmp_onecomplexe deleted
                                   p_restart, current_occur, n_repeat_occur )
    USE mod_array_protein
    USE mod_geometry
    USE mod_type_calc
    USE mod_record
    !! OMP only used if sda_koff
!$   USE omp_lib
    IMPLICIT NONE
    type ( array_protein_type ) :: tab_protein
    type ( geometry ), intent ( in ) :: geom
    type ( type_calculation ), intent ( in ) :: type_calc
    type ( record ), pointer :: p_restart
    !type ( one_complexe ), pointer :: p_tmp_onecomplexe
    real ( kind=4 ), intent ( out ) :: dist
    integer, intent ( in ) :: n_repeat_occur
    integer :: current_conf, current_occur
    logical, intent ( in ) :: iflex

    type ( protein ), pointer :: protein2
    integer :: i, ierr

    protein2 => tab_protein % array ( 2 )

    if ( type_calc %  enum_type_calc == sda_2proteins ) then
        !! choose random position and move_atoms ! no need copy_movable
        !! not efficient done again if change of conformation
        call random_pos_orient ( protein2, geom, type_calc )

        !!test return before copymovable, create bug with OpenMP
        !return

        dist = geom % start_position

        !! treatment of flex is different if sda_2proteins or sda_koff
        if ( iflex ) then
            !write( *, * ) ' Init flexibility at the beginning of each run '
            !! sda_2proteins, only 2 protein
            do i = 1, 2
                if ( tab_protein % array ( i ) % iflex ) then
                    !! in order to force a new initialisation, need to setup current_conf to -1
                    tab_protein % array ( i ) % p_flex % current_conf = -1
                    !! update p_current node and p_sogrid
                    !call init_conformation ( tab_protein % array ( i ) )
                    call initialize_conformation ( tab_protein % array( i ) )
                    !! make explicit copy if needed, moved after
                    !call copy_movable_array ( tab_protein % array ( i ) )
                    ! done in initialize now
                    !current_conf = tab_protein % array ( i ) % p_flex % current_conf

                    !! could be in init_conformation
                    call init_flexible_freq ( tab_protein % array ( i ) % p_flex )
                    !write(*,*) 'after init time flex frequence_protein ', tab_protein % array ( i ) % p_flex % time_next_flex
                    end if

                    !!certainly if protein 2 or done in init_conf ??
                    !! may need copy_movale_array if flexible ??
                    !! can group common code later
            end do
        end if !! end if iflex

    !! if sda_koff, need to reset to the correct position
    else if ( type_calc %  enum_type_calc == sda_koff ) then
        !write(*,*) 'sda_koff current_occur', current_occur

!$OMP CRITICAL ( init_sdakoff )
        !! now apply to both complexe or trajectory, current_occur and p_restart shared are shared
        !! account_occurency not needed, trajectory modified to have 1. in other_real ( 1 )

        !write(*,*) current, do not work with OpenMP, p_tmp_onecomplexe not updated

        ! need to use internal , always up to date
!! to delete !$      p_tmp_onecomplexe => p_restart % p_onecompl

        !if ( current_occur < int ( p_tmp_onecomplexe % other_real_ocompl ( 1 ) ) * n_repeat_occur ) then
        if ( current_occur < int ( p_restart % p_onecompl % other_real_ocompl ( 1 ) ) * n_repeat_occur ) then
            current_occur = current_occur + 1
            !write(*,*) 'update current_occur ', current_occur, &
            !            n_repeat_occur * int ( p_restart % p_onecompl % other_real_ocompl ( 1 ) )
        else
            !write(*,*) 'update to the next complexe'
            call getNext( p_restart, ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR in init_position:getNext()'
                stop 1
            end if
            current_occur = 1
        end if
!$OMP END CRITICAL ( init_sdakoff )

        !! need to reset at the initial position, every time
        !protein2 % position = p_tmp_onecomplexe % position
        protein2 % position = p_restart % p_onecompl  % position
        protein2 % orientation_x = p_restart % p_onecompl % orient_x
        protein2 % orientation_y = p_restart % p_onecompl % orient_y
        !write(*,*) 'prot2 pos ', protein2 % position

        !! certainly best way to normalize, not perfect
        call norm ( protein2 % orientation_x )
        call norm ( protein2 % orientation_y )
        call cross ( protein2 % orientation_z, protein2 % orientation_x, protein2 % orientation_y )

        !! here read conformation
        if ( iflex ) then
            do i = 1, 2
                if ( tab_protein % array ( i ) % iflex ) then
                    !! can change at any moment if OpenMP
                    current_conf = p_restart % p_onecompl % int_ocompl ( get_indice_nconf ( p_restart ) )

                    ! both regrouped in initilaize conformation
                    !tab_protein % array ( i ) % p_flex % current_conf = current_conf
                    !! p_flex % current_conf need to be correctly set up before
                    !call init_conformation ( tab_protein % array ( i ) )
                    call initialize_conformation ( tab_protein % array ( i ), current_conf )
                    !! could be in init_conformation, why commented ??
                    call init_flexible_freq ( tab_protein % array ( i ) % p_flex )
                    write(*,*) 'after init time flex frequence_protein ', tab_protein % array ( i ) % p_flex % time_next_flex
                    write(*,*) 'current_conf ', current_conf
                end if
            end do
        !! end if iflex
        end if

        !! need to check if flexible
        call compute_distance ( tab_protein % array ( 1 ), protein2, dist )
        !write(*,*) 'distance ', dist

        !! already done if flexible ??
        !write(*,*) 'copy movable array in initialisation'
        !! why 1 ?? in case of flexibility
        call copy_movable_array ( tab_protein % array ( 1 ) )
        call copy_movable_array ( tab_protein % array ( 2 ) )

    end if !end sdamm/sda_koff

end subroutine init_position_protein

!> Set random position and orientation for one protein
!!
!! Only this function should be called\n
!! Take into account the geometry and the type of calculation\n
!! If a surface is present with a sphere geometry, the position is chosen on the positive half-sphere
!! ( with hardcoded parameter see random_position_halfsphere )
!! @param prot : instance of \ref mod_protein::protein "protein"
!! @param geom : instance of \ref mod_geometry::geometry
!! @param t_calc : instance pf \ref mod_type_calc::type_calculation "type_calculation"
subroutine random_pos_orient ( prot, geom, t_calc )
    USE mod_geometry
    USE mod_protein
    USE mod_type_calc
    IMPLICIT NONE

    type ( protein ) :: prot
    type ( geometry ), intent( in ) :: geom
    type ( type_calculation ), intent ( in ) :: t_calc

    !! should make better function in the module for instance
    real ( kind=8 ), dimension(3) :: rtry
    real ( kind=8 ), dimension(3,3) :: rot_try

    if ( geom % type == enum_sphere ) then
        !! if sda_2proteins, sdamm in a sphere ??

        if ( geom % surface .eqv. .true. ) then
            write(*,*) "generate halfshpere initial position, like Helms article"
            call random_position_halfsphere ( rtry, geom % start_position )
        else
            call random_position_sphere ( rtry, geom % start_position )
        end if
        !write(*,*) 'rtry random ', rtry

    !! if box
    else
        if ( t_calc % enum_type_calc == sda_2proteins ) then
             call random_position_xyz ( geom, rtry, geom % start_position )
        !!never used in sdamm, until now use restart
        !else if ( t_calc % enum_type_calc == sdamm ) then
        !     call random_position_xyz ( geom, rtry )
        else
            write (*,*) 'random_pos_orient, not implemented'
        end if
    endif

    !! random orientation
    call random_orientation ( rot_try )
    !write(*,*) 'rtry ',rtry
    !write(*,*) 'rot_try ', rot_try

    !! update position and orientation
    !! sdamm, could make a loop to try to insert a new portein in the box
    prot % position = rtry

    !! new version with copy_movable_atom
    !! update the orientation
    call update_rot_orientation ( prot, rot_try )

    !! update position of atom, copy_movable not adapted in case of 2proteins
    !! to see with flex, should be ok
    call update_rotation_protein ( prot, rot_try )

end subroutine random_pos_orient

! for sda z is fixed by start_position, sdamm could be a random x,y,z position
!> Called by \ref random_pos_orient, depending on the geometry
subroutine random_position_xyz ( geom, rtry,  opt_fixed_z )
    USE mod_geometry
    IMPLICIT NONE
    type ( geometry ), intent ( in ) :: geom
    real ( kind=8 ), dimension ( 3 ), intent ( out ) :: rtry
    real ( kind=4 ), optional :: opt_fixed_z

    real ( kind=4 ) :: xmin, xmax, ymin, ymax, zmin, zmax, fixed_z
    real ( kind=4 ), dimension(6) :: params
    real ( kind=8 ) :: unif
    logical :: d3

    if ( present ( opt_fixed_z ) ) then
        d3 = .false.
        fixed_z = opt_fixed_z
    else
        d3 = .true.
        !! to think later, redondant with d3 ??
        fixed_z = 0
    endif

    call get_parameters ( geom, params )
    xmin = params ( 1 )
    xmax = params ( 2 )
    ymin = params ( 3 )
    ymax = params ( 4 )
    zmin = params ( 5 )
    zmax = params ( 6 )


    call ggubs_thread_safe ( unif )
    !write(*,*) 'unif ', unif
    rtry ( 1 ) = xmin + ( xmax - xmin ) * unif
    call ggubs_thread_safe ( unif )
    !write(*,*) 'unif ', unif
    rtry ( 2 ) = ymin + ( ymax - ymin ) * unif
    if ( d3 ) then
        call ggubs_thread_safe ( unif )
        !write(*,*) 'unif d3', unif
        rtry ( 3 ) = zmin + ( zmax - zmin ) * unif
    else
        ! dangerous here, sould set rtry( 3 ) = fixed_z ??
        rtry ( 3 ) = zmin + fixed_z
    end if

end subroutine random_position_xyz

!> Called by \ref random_pos_orient
subroutine random_position_sphere ( rtry, start_pos )
    !USE mod_geometry
    IMPLICIT NONE
    real ( kind=8 ), dimension ( 3 ), intent ( out ) :: rtry
    real ( kind=4 ), intent ( in ) :: start_pos
    !real ( kind=4 ), optional :: start_pos

    real ( kind=8 ), dimension ( 3 ) :: gaus
    real ( kind=8 ) :: sum_inv, b_sum

    call ggnml_thread_safe( gaus )
    rtry = gaus
    sum_inv= 1.d0 / dsqrt(rtry(1)**2+rtry(2)**2+rtry(3)**2)
    b_sum = start_pos * sum_inv
    rtry = rtry * b_sum

end subroutine random_position_sphere

!> Called by \ref random_pos_orient
subroutine random_position_halfsphere ( rtry, start_pos )
    IMPLICIT NONE
    real ( kind=8 ), dimension ( 3 ), intent ( out ) :: rtry
    real ( kind=4 ), intent ( in ) :: start_pos

    rtry = 0.

    do while ( rtry(3) < 20. )
        call random_position_sphere ( rtry, start_pos )
    end do

end subroutine random_position_halfsphere

!> Called by \ref random_pos_orient
!!
!! Here independent of the geometry
subroutine random_orientation ( rot_try )
    IMPLICIT NONE
    real ( kind=8 ), dimension ( 3, 3 ), intent ( out ) :: rot_try

    real ( kind=8 ), dimension ( 3 ) :: om
    real ( kind=8 ) :: unif, unif2
    real ( kind=8 ) :: rmsang
    integer :: i

    rmsang = 3.1415926535d0

    do i = 1,3
        call ggubs_thread_safe ( unif )
        call ggubs_thread_safe ( unif2 )
        if ( unif2 > 0.499999) unif = -unif
        om(i) = rmsang * unif
    end do
    call matrix( om, rot_try )

end subroutine random_orientation

END MODULE mod_init_position
