!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file mod_input_parameter.f90
!!
!! List all parameters in use for sda and some of the tools

!> Contain all input parameter in an unique type\n
!! Used only for reading and preparation, not anymore in the code after
MODULE mod_input_parameter

       USE mod_type_calc
       USE com_sda

       !> Regroup all input parameter
       !! Used only for reading sda_input file and preparation stage, not anymore in the code after
       type param_list

           !> Variables not associated to a group
           !! @{
           real ( kind=8 ) :: dseed, timemax
           integer :: nrun, novers
           integer :: napt1, napt2, nrec, save_exclusion
           real ( kind=4 ) :: probep, probew, hexcl, thresold
           real ( kind=4 ) :: rboost !wbmax, ebmin
           logical :: save_access
           !! only for sda_koff but general variable
           logical :: account_occurency
           ! for image_charge specificity
           integer ::  desolv_img_chg
           real ( kind=8 ) :: epsilon_scale
           !> @}
 
           !> assign to the type type_computation: association/dissociation rate or docking or electron transfert
           character ( 128 ) :: computation

           !> For group type calculation
           !! @{
           type ( type_calculation ) :: t_calc
           integer :: total_solutes, total_grids
           !! @}

           !> For rate claculation
           !! @{
           real ( kind=4 ) :: win0, dwin
           integer :: nwin, nb_contact
           logical :: ibootstrap, ifpt
           !! @}

           !> For complexes and trajectories
           !! @{
           real ( kind=8 ) :: rmsd_min
           integer :: ncmx, ntraj_rec, max_array_traj, freq_print
           logical :: ionerun
           character ( 128 ) :: fcomplexes, ftrajectories
           logical :: format_ascii_complexe, format_ascii_trajectory, traj_sum_nrj, compl_sum_nrj
           logical :: rst_complex
           integer :: merge_step, size_thread
           !> @}

           !> For reaction criteria
           !! @{
           character ( 128 ) :: rxna12f, et1f, et2f
           real ( kind=4 ) :: dind
           integer :: nnnons, nwrec
           !> @}

           !> For timestep
           !! @{
           logical :: itimestep
           real ( kind=8 ) :: dt1, swd1, dt2, swd2
           !> @}

           !> For geometry
           !! @{
           real ( kind=4 ) :: c, xmin, xmax, ymin, ymax, zmin, zmax, start_pos
           logical :: escape, pbc, surf_geom
           !! enum in mod_geometry
           integer :: type_geom
           !> @}

           !> global parameter, apply to all solutes
           !! @{
           real ( kind=8 ) :: epfct, edfct, hdfct, ljrepfct, ljfct !, ionic
           ! moved to a group, after global parameter, before solute_grid
           !logical :: debyeh
           !! @}

           !> analytic group, only HI and DH now, crowder later
           !! @{
           real (kind=4) :: ionic, h_analytic, lvol_rcut
           logical :: ianalytic, debyeh, hydro
           !! @}

           !> For solute group
           !! @{
           character ( 128 ) :: pdb_filename, qef, epf, edf, hdf, lj_repf
           character ( 128 ), dimension(:),allocatable :: ljf
           character ( 2 ), dimension(:),allocatable :: atom_type_lj
           real ( kind=8 ) :: dm, dr, stoke_rad
           integer :: nb_solute, nb_lenjones
           !! new added, for proteins, image_charge to be computed
           logical :: irot, isurf, iflex, iimg_chg
           !> @}

           !> For flexibility
           !! @{
           integer :: total_conf, initial_conf
           logical :: nearest_ngb
           !integer :: init_nb_conf, max_conf, first_conf
           real ( kind=4 ) :: frequence, std_freq
           character ( 128 ) :: list_conf_name, flex_method
           !! @}

           !> For restart
           !! @{
           character ( 128 ) :: restart_filename
           logical :: irestart
           !> @}

           !> For residence_time
           !! @{
           integer :: fixe_bin_size
           character ( 128 ) :: enum_str_residtime
           real ( kind=4) :: size_cell_x, size_cell_y, size_cell_z
           real ( kind=4) :: shift_rt_x, shift_rt_y, shift_rt_z
           integer :: nb_cell_x, nb_cell_y, nb_cell_z, format_rt3d
           logical :: pair_proteins
           !> @}

           !> Special group for MetalDesolvation,  distance_to_surface, energy_per_water, radius_patch, radius_water
           !! @{
           !integer :: int_md_param
           !real ( kind=4 ) :: real_md_param
           logical :: imetaldesolv
           real(kind=4) :: Zadw
           real(kind=4) :: Eadw
           real(kind=4) :: Radw
           real(kind=4) :: Aadw
           !> @}

           !> Group for PMF calculations
           !! @{
           logical :: ipmf
           integer :: mentth
           integer :: mentfi
           integer :: mentom
           integer :: mentx
           integer :: menty
           real(kind=8) :: pstart
           real(kind=8) :: pfinish
           real(kind=8) :: dZ
           !> @}


       end type param_list

       !> Define input parameter for tools only, NEVER used in sda
       type param_list_tools

            !!common input for ed and hd
            character*128:: grid_name, pdb_name
            integer :: iform
            integer, dimension ( 3 ) :: dimension
            real ( kind=8 ) :: hscale
            !! specific ed
            real ( kind=8 ) :: iostr,epssol,rion
            !! specific hd
            real ( kind=8 ) :: a,b, grid_value
            !!specific ljrep
            real ( kind=8 ) :: fraction !factor as well but used in other place
            integer :: nexp

            !! not read from input, but computed only once, before main
            !! just convenient for tool make_edhd_grid, to pass constant
            real ( kind=8 ) :: deb, factor, rfios, radius_average

        end type param_list_tools

       contains

      !> Assign default values for all input variables
      !!
      !! Could create a set default values for Protein_Grid group, allow use default or specific one for flex
      !! @param this : instance of param_list
      subroutine set_default_values ( this )
        IMPLICIT NONE
        type ( param_list ) :: this

        this % t_calc % enum_type_calc = sda_2proteins
        !!default value for sda_2proteins
        this % total_solutes = 2
        this % total_grids = 2

        this % dseed = 256.d0
        !! default values for sdamm
        this % nrun = 1
        this % timemax = 0.d0
        !!only sda_koff
        this % account_occurency = .false.

        this % probep = 1.7
        this % probew = 1.4
        this % hexcl = 0.5
        this % thresold = 0.
        this % save_access = .false.
        this % save_exclusion = 0

        !! sdaflex, but not in module
        this % restart_filename = ''
        this % irestart = .false.

        !! timestep
        this % itimestep = .false.
        this % dt1 = 1.
        this % swd1 = 50.
        this % dt2 = 20.
        this % swd2 = 90.
        !this % rswd = 90.
        !! geometry
        this % type_geom = 1
        !this % b = 100.
        this % c = 150.
        this % xmin = 0.
        this % xmax = 0.
        this % ymin = 0.
        this % ymax = 0.
        this % zmin = 0.
        this % zmax = 0.
        this % start_pos = 100.
        this % escape = .false.
        this % pbc    = .false.

        this % rboost = 1.
        this % novers = 150

        !! for complexes and trajectories
        this % fcomplexes = ''
        this % ftrajectories = ''
        this % ncmx = 500
        this % rmsd_min = 2.d0
        this % ionerun = .true.
        this % ntraj_rec = -1
        this % max_array_traj = 0
        this % format_ascii_complexe = .true.
        this % format_ascii_trajectory = .true.
        ! default as initial
        this % traj_sum_nrj = .true.
        this % compl_sum_nrj = .false.
        this % rst_complex = .false.
        this % freq_print = 100

        !!only for OpenMP
        this % merge_step = 0
        this % size_thread = 0

        !! for ReactionCriteria
        this % rxna12f = 'p12.rxna'
        this % et1f =''
        this % et2f =''
        this % nnnons = 2
        this % dind = 6.0
        this % nwrec = 0

        !! for rate_calculation
        this % win0 = 3.0
        this % dwin = 0.5
        this % nwin = 0
        this % nb_contact = 4
        this % ibootstrap = .false.
        this % ifpt = .false.

        ! for proteins and grids, apply to all
        this % epfct = 0.d0
        this % edfct = 0.d0
        this % hdfct = 0.d0
        this % ljrepfct = 0.d0
        this % ljfct = 0.d0

        ! moved to its own group
        this % ianalytic = .false.
        this % h_analytic = 0.01
        this % debyeh = .false.
        this % hydro = .false.
        this % lvol_rcut = 0.0
        ! ionic strength, apply only to debye-huckel, real*4
        this % ionic = 0.150

        ! done in reset_default_solutegrid
        this % nb_solute = 1
        this % pdb_filename = ''
        this % epf = ''
        this % edf = ''
        this % hdf = ''
        this % qef = ''
        this % lj_repf = ''
        !this % ljf = '' is allocatable now like atom_type_lj
        this % dm = 0.0123
        this % dr = 1.36e-4
        this % stoke_rad = 0.d0
        this % irot = .true.
        this % isurf = .false.
        this % iflex = .false.
        this % iimg_chg = .false.
        this % nb_lenjones = 0
        ! moved to param_probe
        !this % exclusion_dtgrid = .false.

        !! flexibility ie the same block
        this % flex_method = ''
        this % nearest_ngb = .false.
        !this % first_conf = 1
        this % initial_conf = 1
        !this % init_nb_conf = 0
        this % total_conf = 0
        !this % max_conf = 0
        this % frequence = 100.
        this % std_freq = 0.
        !this % list_conf_name = 'list_conf_name.dat'
        this % list_conf_name = ''

        !! here to write for residence_time / escape_time
        !this % fixe_bin_size = .true.
        this % fixe_bin_size = 0
        this % enum_str_residtime = ''
        !! default bin size is fixed, and use 1 Ang
        this % size_cell_x = 1.0
        this % size_cell_y = 1.0
        this % size_cell_z = 1.0
        !## default or wanted to impose the number, large membrane
        ! if fixe_bin_size false, will read nb_cell, and recompute bin_size
        this % nb_cell_x = 20
        this % nb_cell_y = 20
        this % nb_cell_z = 20
        !!if want to apply a shift to the default behaviour
        this % shift_rt_x = 0.
        this % shift_rt_y = 0.
        this % shift_rt_z = 0.
        !! output format binary by default
        this % format_rt3d = 0
        ! for sdamm want rdf for all pairs of solutes
        this % pair_proteins = .false.

        !! to really add biasing ? not implemented
        !this % wbmax = 0.
        !this % ebmin = 0.

        !image_charge option, associated to param_force_energy
        this % epsilon_scale = 0.d0
        this % desolv_img_chg = 0

        ! metal desolvation
        this % imetaldesolv = .false.
        this % Zadw = 6.5
        this % Eadw = 0.13
        this % Radw = 6.0
        this % Aadw = 1.5

        ! PMF
        this % ipmf = .false.
        this % mentth = 60
        this % mentfi = 120
        this % mentom = 60
        this % mentx  = 6
        this % menty  = 6
        this % pstart = 0.d0
        this % pfinish = 60.0
        this % dZ = 0.2

      end subroutine set_default_values

      !! needed for asymetry in default solutegrid
      !! could include flexibility as well
      subroutine reset_default_solutegrid( this )
        IMPLICIT NONE
        type ( param_list ) :: this

        this % nb_solute = 1
        this % pdb_filename = ''
        this % epf = ''
        this % edf = ''
        this % hdf = ''
        this % qef = ''
        this % lj_repf = ''
        !this % ljf = ''
        !this % ljf = '' is allocatable now like atom_type_lj
        if ( allocated ( this % ljf ) ) then
            deallocate ( this % ljf )
        end if
        this % dm = 0.0123
        this % dr = 1.36e-4
        this % stoke_rad = 0.d0
        this % irot = .true.
        this % isurf = .false.
        this % iflex = .false.
        this % iimg_chg = .false.
        this % nb_lenjones = 0

        !> to add flexibility
        !! \todo add flexibility to reset_default_values ??

        !> to add analytic ? a pririo not

      end subroutine reset_default_solutegrid

      ! could add set_default_values_tool


END MODULE
