!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file mod_rate_calculation.f90
!!
!! Module for rates calculation, electron transfer and first passage time


! add array for first time passage, extend bootstrap option to both association, FTP, and dissociation

! try to be as general as possible\n
! association rate, dissociation sda-koff, absorbance\n
! define all tabular needed for rate calculation\n


!> Define type association_rate
!!
!! Used only with sda_2proteins
MODULE mod_rate_calculation

    !> General type used storing statistics about the runs
    type association_rate

        !> global array for assocation rate and electron transfert. Keep the average in a 2d array ( nb_window, nb_contact )
        !! need double accuracy when used for electron transfer
        real( kind=8 ), dimension(:,:), allocatable :: escape_prob ! sumrresc sda_2proteins / sumtime for sda_koff
        !> for association, factor of conversion
        real( kind=8 ), dimension(:,:), allocatable :: beta
        !> Sum of time, sumtime by default for all calculation
        !! can use this array with sdamm case ( Xiaofeng CYP ), always escpae_prob, beta and sumtime allocated/deallocated together
        real ( kind=4 ), dimension(:,:), allocatable :: sumtime
        !> array for firt passage time, contains average of fpt
        real( kind=4 ), dimension(:,:), allocatable :: array_mfpt

        !> window definition, used for monitoring reaction pair
        !! @{
        real ( kind=4 ), dimension(:), allocatable :: window
        real ( kind=4 ), dimension(:), allocatable :: window2
        !! @}

        !> corrective factor dependent of the b and surface
        !! @{
        real ( kind=8) :: brate, crate
        !! @}
        !> Used for windows definition, in Angstrom
        !! @{
        real ( kind=4 ) :: first_window, width_window
        !! @}
        !previous lmx, 4 by default, 1 for icomm = 1,4,5

        !> define number of windows and number of contact to record
        !! @{
        integer :: nb_window, nb_contact
        !> @}

        !> keep file descriptor for a continuous update
        !!for later use of bootstrap and mean first time passage
        !! @{
        integer :: io_bootstp, io_boot_fpt
        !! @}

        !> if bootstrap for association rate, firt passage time or bootstrap for electron transfert was required
        !! @{
        logical :: bootstrap, fpt, boot_et
        !! @}
        !> Format of theses files. Only ascii implemented at the moment
        !! @{
        character*64 :: format_boot, format_boot_fpt, format_boot_real
        !! @}
        !> Array for storing bootstrap for association rate, only windows number need to be save data. size ( nrun, nb_contacts )
        integer, dimension (:,:), allocatable :: array_bootstrap
        !> Real*8 array for electron_transfer, only 1 contact. Size (nrun, nb_window )
        real ( kind=8 ), dimension(:,:), allocatable :: array_bootstrap_real
        !> Array for storing bootstrap of first passage time size (nrun, nb_win, nb_contact ), very large array
        real ( kind=4 ), dimension (:,:,:), allocatable :: array_boot_fpt
        !> And logical if need to trnasform 4 integer
        logical :: transform_escape

    end type association_rate

    contains

    !> Initialize association_rate and allocate arrays
    !!
    !! dm need to be the sum of the diffusion coeff.\n
    !! @param this : instance of association_rate
    !! @param first_window : distance of the first window, in Angstrom
    !! @param nb_window : number of windows
    !! @param width_window : width of each window
    !! @param nb_contact : number of contacts to record, from 1 to nb_contact included
    !! @param bsurf, csurf : values of the b and c surface
    !! @param dm : use the sum of the diffusion terms, absolute trans. diff. coeff.
    !! @param ibootsp, ifpt, iboot_et : file descriptor for the bootstrap output files
    subroutine allocate_rate_association ( this, first_window, nb_window, &
                               width_window, nb_contact, bsurf, csurf, dm, &
                               ibootsp, ifpt, iboot_et )
        USE com_sda
        IMPLICIT NONE
        type ( association_rate ) :: this
        real ( kind=4 ), intent(in) :: first_window, width_window
        integer, intent(in) :: nb_window, nb_contact
        real ( kind=4 ), intent ( in ) :: bsurf, csurf
        real ( kind=8 ), intent ( in ) :: dm
        logical :: ibootsp, ifpt, iboot_et

        ! error with ifort, cannot use datan in expression
        real ( kind=8 ) :: fopi
        real (kind=8 ), parameter :: rfc = 6.022d+8
        character*64 :: filename_boot, filename_boot_assoc, filename_boot_fpt, filename_boot_et

        integer :: iw, ierr

        !write(*,*) 'entry allocate rate_association ', dm
        !write(*,*) 'Entry rate_association bsurf csurf dm', bsurf, csurf, dm
        !write(*,*) 'nb_window, nb_contact ', nb_window, nb_contact

        ! initilaize variables
        fopi = datan(1.d0)*16.d0
        ierr = 0
        filename_boot = ''
        filename_boot_assoc = 'assoc_bootstrap.log'
        filename_boot_et =    'electron_transfer.log'
        filename_boot_fpt =   'fpt_bootstrap.log'

        this % first_window = first_window
        this % width_window = width_window
        this % nb_window = nb_window
        this % nb_contact = nb_contact
        ! constant used at the end for rate calculation
        ! here dm should be the relative translational diffusion coeff.
        this % brate = fopi * rfc * bsurf * dm
        this % crate = fopi * rfc * csurf * dm

        !! get io number for common sda
        this % io_bootstp = io_bootstrp
        this % io_boot_fpt = io_boot_fpt
        this % bootstrap = ibootsp
        this % fpt = ifpt
        this % boot_et = iboot_et
        this % format_boot = ''
        this % format_boot_real = ''
        this % format_boot_fpt = ''
        ! default use 4 integer, then need to transform for storing data in escape_prob
        ! set to false in case of electron transfert
        this % transform_escape = .true.

        !!make like sda, order may be changed , TO CHECK
        allocate ( this% escape_prob ( nb_window, nb_contact), &
                   this% beta ( nb_window, nb_contact ), this % sumtime( nb_window, nb_contact ), stat=ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR Couldn t allocate memory for rate_calculation '
            stop 1
        endif

        ! create windows array
        allocate ( this % window ( nb_window ), this % window2 ( nb_window ), stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR Couldn t allocate memory for window, size ', nb_window
            stop 1
        endif

        do iw = 1, nb_window
            this % window ( iw ) = first_window + ( iw - 1.0 ) * width_window
            !! elec_tranfert windows2 = windows.... in sda 6 why, to check ??
            this % window2 ( iw ) = ( this % window(iw) )**2
        end do
        !! difference for icomm == 1 and 4 ?? or just trick ?? seems bad trick...

        !! if fpt allocate array for average at least
        if ( this % fpt ) then
            allocate ( this % array_mfpt ( nb_window, nb_contact ), stat = ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR Couldn t allocate memory for array_mfpt in rate_calculation '
                stop 1
            endif
            this % array_mfpt = 0.
        end if

        !! initialize array
        this % escape_prob = 0.
        this % beta = 0.
        this % sumtime = 0.

        !! if bootstrap open a file for output, create the format for the bootstrap
        if ( this % bootstrap ) then
            !! default bootstrap for association rate

            ! adjust name file, assoc or et
            if ( iboot_et .eqv. .false. ) then
                filename_boot = filename_boot_assoc
            else
                filename_boot = filename_boot_et
            end if
            ! open file
            open (this % io_bootstp, file=filename_boot, form= 'formatted', status='replace', iostat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'Cannot open file: ', trim ( filename_boot )
            endif
            ! write header, title depend if assoc_rate or et, rest is similar
            if ( iboot_et .eqv. .false. ) then
                write( this % io_bootstp,'("# association_rate")' )
            else
                write( this % io_bootstp,'("# electron_transfer")' )
                 this % transform_escape = .false.
            end if
            ! common to both
            write( this % io_bootstp,'("#nb_contact:",i6)' ) this % nb_contact
            write( this % io_bootstp,'("#nb_wind.  :",i6)' ) this % nb_window
            write( this % io_bootstp,'("#first win :",f9.2)' ) this % first_window
            write( this % io_bootstp,'("#width win :",f9.2)' ) this % width_window
            write( this % io_bootstp,'("#b_rate    :",e14.6)' ) this % brate
            write( this % io_bootstp,'("#c_rate    :",e14.6)' ) this % crate

            ! write format for bootstrap, can be for association
            write( this % format_boot,'("(",i3,"i4)")' ) this % nb_contact
            ! or for electron_transfer
            write( this % format_boot_real,'("(",i3,"e17.10)")' ) this % nb_window
            !write(*,*) 'format_boot ', this % format_boot

            !! if fpt, apply bootstrap also
            if ( this % fpt ) then
                open (this % io_boot_fpt, file=filename_boot_fpt, form= 'formatted', status='replace', iostat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'Cannot open file: ', trim ( filename_boot_fpt )
                endif
                !! write header
                write( this % io_boot_fpt,'("# first_passage_time")' )
                write( this % io_boot_fpt,'("#nb_contact:",i6)' ) this % nb_contact
                write( this % io_boot_fpt,'("#nb_wind.  :",i6)' ) this % nb_window
                write( this % io_boot_fpt,'("#first win :",f9.2)' ) this % first_window
                write( this % io_boot_fpt,'("#width win :",f9.2)' ) this % width_window
                ! write format boot
                write( this % format_boot_fpt,'("(",i3,"e12.4)")' ) this % nb_contact * this % nb_window
                write(*,*) 'format_boot_fpt ', this % format_boot_fpt
            end if
        end if

    end subroutine allocate_rate_association

    !> Delete arrays and reinitialize data members
    !! @param this : instance of association_rate
    subroutine deallocate_rate_association ( this )
        IMPLICIT NONE
        type ( association_rate ) :: this

        integer :: ierr

        this % first_window = 0
        this % nb_window = 0
        this % width_window = 0
        this % nb_contact = 0

        if ( allocated ( this % escape_prob ) ) then
            deallocate ( this % escape_prob, this % beta, this % sumtime, stat = ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR Couldn t deallocate memory for escape_prob in rate_calculation '
                STOP 1
            endif
        end if

        if ( allocated ( this % window ) ) then
            deallocate ( this % window , this % window2, stat = ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR Couldn t deallocate memory for windows in rate_calculation '
                STOP 1
            endif
        end if

        !! deallocate mean firt_time array, not necessary for valgrind
        if ( allocated ( this % array_mfpt ) ) then
            deallocate ( this % array_mfpt, stat = ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR Couldn t deallocate memory for array_mfpt in rate_calculation '
                stop 1
            endif
        end if

        !! close file if bootstrap, array created and destroyed in MainLoop
        if ( this % bootstrap ) then
            close ( this % io_bootstp )
            !! check if boot_fpt activated
            if ( this % fpt ) then
                close ( this % io_boot_fpt )
            end if
        end if

    end subroutine deallocate_rate_association

!    subroutine set_rate_coeff ( this, dm, bsurf, csurf )
!        IMPLICIT NONE
!        type ( association_rate ) :: this
!        real (kind=8), intent(in) :: dm
!        real (kind=4) :: bsurf, csurf
!
!        real ( kind=8 ) :: fopi
!        real (kind=8 ), parameter :: rfc = 6.022d+8
!
!
!        fopi = datan(1.d0)*16.d0
!
!        !! TO CHECK, dm may be the sum of both if sda_2proteins !!
!        !write(*,*) 'check value of dm in allocate_rate !!! should be the sum of both if sda_2proteins', dm
!        this % brate = fopi * rfc * real(bsurf,8) * dm
!        this % crate = fopi * rfc * real(csurf,8) * dm
!
!    end subroutine set_rate_coeff

    ! need to add argument local_escape, == escape ?? done
    ! nb traj loose its accurracy with openMP, maybe better to have another function for recomputing beta ?? kind of checkpoint

    !> Update the arrays after each trajectory
    !!
    !! Update all arrays at the end of trajectories\n
    !! Function is thread-safe, with the use of ATOMIC sections minimum overhead.\n
    !!
    !! @param this : instance of association_rate
    !! @param nb_traj : the total number of trajectories
    !! @param visited_windows : array for association rate ( version with 4 integers )
    !! @param escape : array for electron transfert, cannot anymore share the format with association rate
    !! @param sum_time, first_time : external arrays belonging to each thread for updating global sumtime and firstime
    subroutine update_escape ( this, nb_traj, visited_windows, escape, sum_time, first_time )
!$      USE omp_lib
        IMPLICIT NONE
        type ( association_rate ) :: this
        integer, intent(in) :: nb_traj
        integer, dimension(:), intent(in) :: visited_windows
        real ( kind=8 ), dimension(:,:), intent( in ) :: escape
        real ( kind=4 ), dimension (:,:), intent( in ) :: sum_time, first_time

        integer :: iw, ict

        ! need temporary array if we need to transform from 4 integer to large array
        ! could be optional if only bootstrap is activated, keep for comparing results
        real ( kind=8 ), dimension( this % nb_window, this % nb_contact ) :: escape_windows

        !write(*,*) 'entry update_escape '
        !write(*,*) 'need transform escape ',  this % transform_escape
        !write(*,*) 'visited windows ', visited_windows

        if ( this % transform_escape ) then
          ! new version, only for associaition rate need to pass from int(4) to escape(:,:)
          ! escape default all 1, so first 0 gives min contact
LOOP_CT:  do ict = 1, this % nb_contact
             ! set all to 1, not visited
             escape_windows ( :, ict ) = 1.d0

             ! check it has been visited !
             if ( visited_windows( ict ) .ne. -1 ) then
                do iw = visited_windows( ict ), this % nb_window
                    escape_windows ( iw, ict ) = 0.d0
                    !write(*,*) 'assign min_windows for contact ', min_windows ( ict ), ict
                    !cycle LOOP_CT
                end do
             end if
          end do LOOP_CT

        else
          !write(*,*) 'no transformation, just copy the array'
          escape_windows = escape
       end if

       !! output to check
       !write(*,*) 'check new_version '
       !write(*,*) escape_windows


        ! reinitiliasation escape
        ! apply for both association rate and electron transfert, keep sum in memory for averaging
        do ict = 1, this % nb_contact
            do iw = 1, this % nb_window
                !write(*,*) 'escape ', this % escape (iw, ict)

                ! update escape_prob, used for association and electron_transfer
                ! more efficient with ATOMIC than CRITICAL
                !$OMP ATOMIC
                this % escape_prob ( iw, ict ) = this % escape_prob ( iw, ict ) + escape_windows ( iw, ict )
                !this % escape_prob ( iw, ict ) = this % escape_prob ( iw, ict ) + escape ( iw, ict )

                !$OMP ATOMIC
                this % sumtime ( iw, ict ) = this % sumtime ( iw, ict ) + sum_time ( iw, ict )

                ! here could add assoc_bootstrap...

                !! update fpt... how to account for mean time if not happens
                if ( ( this % fpt ) .and. ( escape ( iw, ict ) == 0 ) ) then
                    !write(*,*) 'update first_time'

                    !$OMP ATOMIC
                    this % array_mfpt ( iw, ict ) = this % array_mfpt ( iw, ict ) + first_time ( iw, ict )

                    !write(*,*) 'array_mfpt ', this % array_mfpt ( iw, ict )
                end if

                ! for OpenMP, compute beta only at the end of the full run
            end do
        end do

        !write(*,*) 'escape ', escape
        ! write(*,*) 'escape_prob ', this % escape_prob
        !do iw = 1, this % nb_window
        !    write(*,*) 'prob ', iw, this % escape_prob(iw, 1)
        !end do

        ! could make explictit loop if fpt...
        if (  this % bootstrap ) then
            !write(*,*) 'bootstrap on in update_escape '

            !write(*,*) 'allocated this % array_bootstrap ', allocated ( this % array_bootstrap )
            !write(*,*) 'allocated this % array_bootstrap_real ', allocated ( this % array_bootstrap_real )

            ! check bootstrap and association rate
            if ( allocated ( this % array_bootstrap ) ) then
                ! no need of critical section, nb_traj always different for each thread
                this % array_bootstrap ( nb_traj, : ) = visited_windows

            ! case of electron transfert
            ! always one contact for electron transfert
            else if ( allocated ( this % array_bootstrap_real ) ) then
                this % array_bootstrap_real ( nb_traj, : ) = escape ( :, 1)
            end if

            ! test fpt apart
            if ( this % fpt ) then
                this % array_boot_fpt ( nb_traj , :, : ) = first_time
                !write(*,*) 'test bootstrap fpt ', this % array_boot_fpt ( nb_traj , :, : )
            end if

        end if ! end if bootstrap

    end subroutine update_escape

    ! cannot make log, need to save header, to do
    ! could send current_run and get checkpoints

    !> Write the bootstrap output files
    !!
    !! Call only at the end of the computation,
    !! @param this : instance of association_rate
    !! @param total_run : the total number of runs really done (case of intercepted signal)
    subroutine write_boot_file ( this, total_run )
        IMPLICIT NONE
        type ( association_rate ) :: this
        integer :: total_run

        integer :: i, nwin
        ! D for using double precision !
        real ( kind=8 ), parameter :: CUTOFF = 1.0D-50

#ifdef DEBUG
        write(*,*) 'Entry write_boot_file total_run ', total_run
#endif

        ! should be similar with writing header of bootstrap
        write ( this % io_bootstp, '("#nrun      :",i6)' ) total_run

        ! same file, can print only one or the other !!
        ! case of association rate
        if ( allocated ( this % array_bootstrap ) ) then
            do i = 1, total_run
                write ( this % io_bootstp, fmt= this % format_boot ) this % array_bootstrap ( i, : )
            end do

        ! case of electron transfert
        else if ( allocated ( this % array_bootstrap_real ) ) then
            !write(*,*) 'write boot real'

            ! problem with printing XXE-100, set to zero
            do i = 1, total_run
                do  nwin = 1, this % nb_window
                    if ( this % array_bootstrap_real ( i, nwin ) < CUTOFF ) then
                        this % array_bootstrap_real ( i, nwin ) = 0.d0
                    end if
                end do
            end do

            do i = 1, total_run
                    write ( this % io_bootstp, fmt= this % format_boot_real ) this % array_bootstrap_real ( i, : )
            end do
        end if

        if ( this % fpt ) then
            write ( this % io_boot_fpt, '("#nrun      :",i6)' ) total_run
            do i = 1, total_run
                write ( this % io_boot_fpt, fmt= this % format_boot_fpt ) this % array_boot_fpt ( i, : , :)
                !write(*,*) 'boot nb ',i
                !write(*,*) this % array_boot_fpt ( i, : , :)
            end do
        end if

    end subroutine write_boot_file

    !> Print final results, could be used for restart or checkpoint
    !!
    !! Print data in the main output file\n
    !! opt_koff, will not print the rate, does not make sense\n
    !! if sda_2proteins, print all\n
    !! @param this : instance of association_rate
    !! @param total_run : total number of runs
    !! @param opt_koff : optionaly print association rates
    subroutine print_table_rate ( this, total_run, opt_koff )
        IMPLICIT NONE
        type ( association_rate ) :: this
        integer :: total_run
        logical, optional :: opt_koff

        real ( kind=8 ) :: rateiw, timeiw
        real (kind=4) :: ftotal, fovertotal_run
        integer :: ict, iw
        logical :: koff

        koff = .false.
        if ( present ( opt_koff ) ) then
            !write(*,*) 'opt_koff'
            koff = opt_koff
        end if

        ftotal = float ( total_run )
        fovertotal_run = 1./ ftotal

#ifdef DEBUG
        write(*,*) 'print table'
        write(*,*) 'brate, crate ', this % brate, this % crate
        write(*,*) 'total_run ', total_run
#endif

        write (*,*)
        if ( koff .eqv. .false. ) then
            write (*,*) 'Final table of rates:'
            write (*,*) '    rxn       M^-1s^-1      beta      time'
        else
            write (*,*) 'Final table of times:'
            write (*,*) '    rxn        time'
        end if

        ! update beta, the probability of contacts
        do ict = 1, this % nb_contact
            do iw = 1, this % nb_window
                this % beta (iw, ict ) = ( ftotal - this % escape_prob ( iw, ict ) ) &
                                         * fovertotal_run
            end do
        end do

        do ict = 1, this % nb_contact
            write(*,fmt='(''#'',i5)') ict
            do iw = 1, this % nb_window
                ! print assocition rate and beta and sumtime ( still meaning full )
                if ( koff .eqv. .false. ) then
                    rateiw = this % beta( iw,ict ) * this % brate /  &
                             ( 1.d0 - (1.d0 -  this % beta( iw, ict ) ) * this % brate/ this % crate )

                    write(*,fmt='(4e12.4)') this % window( iw ), rateiw, this % beta( iw,ict ), this % sumtime( iw, ict ) / ftotal

                ! if koff do not print association rate, computed but no meaning !
                else
                    write(*,fmt='(2e12.4)') this % window( iw ), this % sumtime( iw, ict ) / ftotal
                end if
            end do
        end do

        ! should avoid koff and fpt, when reading input
        if ( this % fpt ) then
            write (*,*)
            write (*,*) 'Final table of first passage time:'
            write (*,*) '    rxn       ps'

            do ict = 1, this % nb_contact
                write(*,fmt='(''#'',i5)') ict
                do iw = 1, this % nb_window

                    !!       distance never reached, avoid div by 0                            distance always reached == 0
                    if ( this % escape_prob ( iw, ict ) == total_run ) then !.or. ( this % escape_prob ( iw, ict ) == 0 ) ) then
                        write(*,fmt='(1e12.4,11X,a)') this % window( iw ),"0" !"-"
                        !write(*,*) 'never reached'
                        !write(*,fmt='(1e12.4)') this % window( iw )

                    else
                        !write(*,*) 'first_time ', this % array_mfpt ( iw, ict )
                        !write(*,*) 'number recorded ', this % escape_prob ( iw, ict ), ' divided by ', &
                        !                               ftotal - real(this % escape_prob ( iw, ict ),4)

                        fovertotal_run = ftotal - real(this % escape_prob ( iw, ict ),4)
                        timeiw = this % array_mfpt ( iw, ict ) / fovertotal_run
                        write(*,fmt='(2e12.4)') this % window( iw ), timeiw
                    end if
                    !write(*,fmt='(2e12.4)') this % window( iw ), timeiw
                end do ! nb_window
            end do ! nb_contact
        end if

    end subroutine print_table_rate

    !> Print only the array_sumtime, implemented for sdamm and association_rate
    !! @param this : instance of association_rate
    !! @param norm : normalisation factor
    subroutine print_sumtime ( this, norm )
        implicit none
        type ( association_rate ) :: this
        real ( kind=4 ) :: norm

        integer :: ict, iw

        write(*,*) 'print_sumtime norm ', norm

        write (*,*) 'Final table of times:'
        write (*,*) '    rxn        time'

        do ict = 1, this % nb_contact
            write(*,fmt='(''#'',i5)') ict
            do iw = 1, this % nb_window

                !write(*,fmt='(2e12.4)') this % window( iw ), this % sumtime( iw, ict )
                !write(*,fmt='(2e12.4)') this % window( iw ), this % sumtime( iw, ict ) / nstepsf
                write(*,fmt='(2es12.4)') this % window( iw ), this % sumtime( iw, ict ) / norm

            end do
        end do

    end subroutine print_sumtime

END MODULE mod_rate_calculation
