!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_record.f90
!! Module for reading and writing trajectories and complexes files
!!
!! most general type is record, subtypes are complex and trajectory

! 23/2/12 add getNext(), can use both list of complexe or trajectory
!    easier for restart ( sdakoff sda-energy )

! Add copy record % pcomplexe, for use with OpenMP

! 3 other real for complexe ( occurence, average total energy, sum energy square for computing standart deviation of energy )\n
! 1 other real for trajectory: time\n
!
! type record is the base class, ListComplexe and Trajetory derived one\n
! All use mod_onecomplexe to store data\n

! Add fill_listcomplexe_norecursive
! Large recursive function apporx 9 000 (work) - 12 000(home) crash !!
! go to non recursive function

!> MODULE mod_record\n
!! Define type record and subtypes Trajectory and ListComplexe
!!
!! Used for dealing with trajectories and complexes\n
!! They both store mod_onecomplexe objects, and then share the same format and rules for output\n
!! Trajectory is a fixed array of one_complexe, the size can be fixed for output optimisation\n
!! ListComplexe is a list implemenation of one_necomplexe ( fast insertion and deletion whatever the position ), but has a max_nb_complexe to limit the size
!!
!! All records can read/writen in ascii or binary format\n
!! Output are variable:\n
!! Maximum 4 integer : ntraj(nrun), nstep, nprotein, nconf\n
!! Maximum 20 real for energy: total energy, electrostatic,...\n
!! But the content can be modified at run-time ( do not store or print what is not used )
MODULE mod_record

    USE mod_onecomplexe
    USE mod_array_protein
    USE mod_type_calc

    !> Versioning
    integer, parameter :: LAST_VERSION_REC = 2

    !> enumeration for the state of the node
    integer, parameter :: intern = 0, first = 1, last = 2

    !> enumeration for type_record
    integer, parameter :: trajectory_enum=1, complexe_enum=2, off_record = 3

    !> names of the energy terms for the output
    character ( 13 ), dimension ( 8 ) :: ener_colnames = &
        (/ character(13):: 'TotEn', 'El', 'ED', 'HD', 'rLJ', 'LJ', 'ImgChg', 'MetDes' /)
    character ( 13 ), dimension ( 8 ) :: var_colnames = &
        (/ character(13):: 'varTotEn', 'varEl', 'varED', 'varHD', 'varRLJ', 'varLJ', 'varImgChg', 'varMetDes' /)

    !> Define one Node for the ListComplexe\n
    type NodeComplexe

       !> pointer to instance of mod_onecomplexe
       type ( one_complexe ), pointer :: p_data
       !> information about the state of the node: intern, first or last
       integer :: state
       !> pointer to the next node
       type ( NodeComplexe ), pointer :: p_next
       !> pointer to the previous node
       type ( NodeComplexe ), pointer :: p_previous

    end type NodeComplexe

    !> Complexes implemented as a list, fast insertion and deletion\n
    !! To define a list, we use type NodeComplexe which contains one object\n
    !! Can extend the description here
    type ListComplexe

       !> pointer on first and last node, last necessary for reaching the end
       type ( NodeComplexe ), pointer :: p_first, p_last
       !> only pointer, associated to current node, never allocated
       type ( NodeComplexe ), pointer :: p_current

       !! previous method better ! because check each time
       ! member of ListComplexe
       real ( kind=8 ) :: max_energy, rmsd_min
       !> actual number of complexes, maximum nb of complexes in record
       integer :: nb_complexe
       !> option for rmsd_min
       logical :: one_run, one_conf

    end type ListComplexe

    !> Implemented as an array, fixed size and append data only at the end
    type Trajectory

       !> 1 dimensional array of instances of mod_onecomplexe
       type ( one_complexe ), dimension(:), allocatable :: array_trajectory
       !> member of Trajectory
       integer :: size_array, current_pos

    end type Trajectory


    !> Main "base" class\n
    !! All derived can be created from this one
    type record

       ! convenient to add a name ?? for restart particularly convenient ( sdamm et sda_ener )
       ! fixe fixe name for the moment for restart, and io_restart

       !> Type of the record : complexe or trajectory
       integer :: type_record
       !> pointer to a subtype as ListComplexe
       type ( ListComplexe ), pointer :: pcomplexe
       !> pointer to a subtype as Trajectory
       type ( Trajectory ), pointer :: ptrajectory
       ! add a pointer directly to the data, works whatever complexe or trajectory
       ! updated by getNext
       !> internal pointer to an instance of mod_onecomplexe, foor internal usage and convenience mainly
       type ( one_complexe ), pointer :: p_onecompl
       !character ( 128 ) :: filename_out

       ! nb_max_complexe value used for maximum of complexes,
       ! or maximum of record before writing the trajectory
       ! add nb_protein as variable. For tools, not association to tab_prot

       ! for grouping http://www.star.bnl.gov/public/comp/sofi/doxygen/grouping.html
       ! in fortan
       !> general member variable of records
       !! @{
       integer :: nb_max_complexe, ntraj_rec, freq_print, nb_protein
       logical :: opt_sum, box_info
       !> @}
       !> file descriptor of the record
       integer :: io_compl
       !> Local information about the size of mod_onecomplexe array\
       !! new nb_real_to_sum, needed to indicate to set_onecomplexe how many fields for each interaction
       !! @{
       integer :: nb_int, nb_real, nb_other_real, nb_types
       !> @}
       !> Internal use, indicates the  content of the integer and energy of mod_onecomplexe
       !> @{
       integer ( kind=2 ) :: bit_energy, bit_integer
       !> @}
       !> new, maximum number of fields to add, for the moment hardcoded, should be set up by one complexe ?? but only once
       integer :: max_fields
       !> Trick here, keep the indice for accessing ntraj, nrun,..\n
       !! because not known by onecomplexe, and needed for test one_run one_conf
       !> @{
       integer :: indice_ntraj, indice_nstep, indice_nprot, indice_nconf
       !> @}
       !> pointer to tab of protein, use for writing the header
       type ( array_protein_type ), pointer :: ptab_prot
       ! split format and header
       !> format string ( only ascii ) of mod_onecomplexe, computed during initialisation, should not change later
       character ( 128 ) :: format_string_ascii
       !> format of the header in ascii, format is constructed during initialization
       character ( 128 ) :: format_header_ascii
       !> full string for header in ascii constructed only once. Do not know how to make equivalent for binary.
       character( 128 ):: header_str
       !> user readable description of the column, should be with header_str, but do not manage to deal deal with new line
       character ( 512 ) :: description_column_str
       !> if true format is ascii, if false input or output format is binary
       !! Input type can be tested with subroutine test_binary
       logical :: format_ascii
       !> for header, version not yet used, fixed to "version0.1"
       character(10) :: version
       !> Now used, more convenient to store a number
       integer :: version_number
       !> store the type of calculation : sda_2proteins, sdamm, sda_koff...
       type ( type_calculation ) :: record_type_calc

    end type record

    ! not used in other modules, but here may clarify the use of the functions intented to be used normally
    ! They all apply to both complexes and trajectories
    !> for creation, copy and deletion
    public :: new_record, init_record, initialize_record, delete_record, copy_record
    public :: reset_position, getNext
    public :: get_nb_terms_energy_compatible, get_nbterms_integer_compatible, return_array_wrap, set_onecomplexe_wrap
    public :: compute_all_occurencies, add_record, read_file_record, write_record, read_header_record
    ! all others functions should be considered as private


!!!!!!!!!!!!! Base Class Record   !!!!!!!!!!!!!!!!!!!!!!!!!!!!
    contains

        !> Allocate a object record : complexe or trajectory \n
        !! allocate a list_complexe or an array for trajectory and open the output file with the correct format\n
        !! The obect is not fully ready to use at this stage\n
        !! It needs correct "bit_integer" and  "bit_energy" and a call to \ref initialize_trajectory
        !!
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param type_record_string : "complexe" or "trajectory"
        !! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
        !! @param nb_max_complexe : fixe the size of array for trajectory, or the maximum size of a list
        !! @param filename_out : filename of output
        !! @param io_compl : file descriptor
        !! @param opt_sum : bool indicates if the energy terms should be added ( ep1_2+ep2_1 ) or separated in the output. Internally always splitted, influences only the output
        !! @param box_info : bool if box information is present ( used only for sdamm )
        !! @param ntraj_rec : the trajectory number to record
        !! @param freq_print : the frequence in steps (not in time ) with which the trajectory is written
        !! @param rmsd_min : value used for complexes output
        !! @param one_run : value used for complexes output
        !! @param one_conf : value used for complexes output
        !! @param format_ascii : bool, if true ascii output, else bunary output
        !! @param nb_types : number of different types of proteins
        subroutine new_record ( this, type_record_string, type_calc, &
                                nb_max_complexe, filename_out, io_compl, opt_sum, box_info, &
                                ntraj_rec, freq_print, &
                                rmsd_min, one_run, one_conf,&
                                format_ascii, nb_types )
            IMPLICIT NONE
            type ( record ) :: this
            character type_record_string*(*)
            type ( type_calculation ) :: type_calc
            integer :: nb_max_complexe
            character ( 128 ) :: filename_out
            real ( kind=8 ) :: rmsd_min
            integer :: io_compl , ntraj_rec, freq_print
            logical :: one_run, one_conf, opt_sum, format_ascii, box_info
            ! new added
            integer, optional :: nb_types

            character ( 15 ) :: format_file !< formatted or unformatted
            integer :: ierr

            ! include versioning, if new with new_record apply the last one
            this % version_number = LAST_VERSION_REC
            this % version = create_version_str( LAST_VERSION_REC )

            this % nb_max_complexe = nb_max_complexe
            this % ntraj_rec = ntraj_rec
            this % freq_print = freq_print
            this % io_compl = io_compl
            this % opt_sum = opt_sum
            this % box_info = box_info
            this % nb_types = nb_types

            ! setup new type_calc and initialize bits to 0
            this % record_type_calc = type_calc
            this % bit_energy = O'0'
            this % bit_integer = O'0'
            this % nb_protein = 0

            ! indices
            this % indice_ntraj = 0
            this % indice_nstep = 0
            this % indice_nprot = 0
            this % indice_nconf = 0

            this % format_ascii = format_ascii
            this % header_str = ''
            this % description_column_str = ''
            this % format_string_ascii = ''
            this % format_header_ascii = ''

            ! init all pointers to null
            nullify ( this % ptab_prot )
            nullify ( this% pcomplexe, this % ptrajectory )
            nullify ( this % p_onecompl )

            ! set nb_other real, fixe to 1 for traj ( time ), 3 for complexe
            ! and allocate correct complexe or trajectory
            if ( type_record_string == 'complexe' ) then
                 this % type_record = complexe_enum
                 this % nb_other_real = 3
                 allocate ( this % pcomplexe, stat = ierr )
                 if ( ierr /= 0 ) then
                     write ( *, * ) 'Problem in allocating pcomplexe'
                     stop 1
                 end if
                 !! creaes the list, no data yet
                 call new_listcomplexe ( this % pcomplexe, rmsd_min, one_run, one_conf )

            else if ( type_record_string == 'trajectory' ) then
                 this % type_record = trajectory_enum
                 this % nb_other_real = 1
                 allocate ( this % ptrajectory, stat = ierr )
                 if ( ierr /= 0 ) then
                     write ( *, * ) 'Problem in allocating ptrajectory '
                     stop 1
                 end if
                 !! creates the array, size fixed by nb_max_complexe
                 call new_trajectory ( this % ptrajectory, this % nb_max_complexe )
            else
                write(*,*) 'ERROR Type in record:', type_record_string, ' is an unknow value'
                stop 1
            end if

            ! file opening, if open here, no need to save the name
            ! depends on ascii or binary
            if ( len_trim ( filename_out ) > 0 ) then
                ! ascii, formatted
                if ( this % format_ascii ) then
                    format_file = 'formatted'
                ! binary, unformatted
                else
                    format_file = 'unformatted'

                end if

                open (io_compl, file=filename_out , form= format_file, status='replace', iostat = ierr )
                if ( ierr /= 0 ) then
                       write(*,*) 'Cannot open file: ', trim ( filename_out )
                endif
            else
                write(*,*) 'filename_out is empty, module used by a tool ??'
            end if

        end subroutine new_record

        !> A simpler initialization, used when reading an existing file
        !! nullify the pointers mainly
        !! new add opt_iocmx if want to force
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param opt_iocmx : optional, gives a file descriptor, usefull if more than 1 record opened at the same time
        subroutine init_record ( this, opt_iocmx )
            IMPLICIT NONE
            type (record ) :: this
            integer, optional :: opt_iocmx

            !write(*,*) 'Entry init_record '
            this % bit_energy = O'0'
            this % bit_integer = O'0'
            this % nb_protein = 0
            ! indices
            this % indice_ntraj = 0
            this % indice_nstep = 0
            this % indice_nprot = 0
            this % indice_nconf = 0

            this % format_ascii = .true. !format_ascii
            ! init string to len = 0 for further test
            this % header_str = ''
            this % description_column_str = ''
            this % format_string_ascii = ''
            this % format_header_ascii = ''

            if ( present ( opt_iocmx ) ) then
                this % io_compl = opt_iocmx
            end if

            ! init all pointers to null, it is the important point
            nullify ( this % ptab_prot )
            nullify ( this% pcomplexe, this % ptrajectory )
            nullify ( this % p_onecompl )

        end subroutine init_record

        !> Finalize the creation of a record.\n
        !! Setup the bits for variable fields, assign tab_prot and writes the header, except if opt_read is true.\n
        !! This function has been improved and allocate basic trajectory / list_complexe automatically if not done before
        !! Usefull for tools, shoud make tab_prot an option also ( want to use easy construction without tab_prot )
        !!
        !! the \ref mod_record::record::bit_energy "bit_integer" and \ref mod_record::record::bit_energy "bit_energy" should have been set before the call\n
        !! need to be separate, because need to set value of bits before
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param tab_prot : optional, an instance of \ref mod_array_protein::array_protein_type "array_protein_type"
        !! @param opt_read : optional, open the record for reading, will not write the header file
        !! @param opt_nb_protein : optional, indicates the number of the number of proteins, by default use the size of array_protein
        !! @param opt_large_traj : optional, boolean: if the trajectory file is large, it will be read one by one, instead of allocating the entire trajectory at once
        subroutine initialize_record ( this, tab_prot, opt_read, opt_nb_protein, opt_large_traj )
            IMPLICIT NONE
            type (record ) :: this
            type ( array_protein_type ) , target, optional :: tab_prot
            logical, optional, intent ( in ) :: opt_read, opt_large_traj
            integer, optional, intent ( in ) :: opt_nb_protein

            logical :: b_write_header, large_traj

            !write(*,*) 'Entry initialize_record '

            ! by default opended in write mode, print the header at the end of the function
            b_write_header = .true.

            if ( present ( opt_read ) ) then
                if ( opt_read .eqv. .true. )  then
                    b_write_header = .false.
                end if
            end if

            large_traj = .false.
            if ( present( opt_large_traj ) ) then
                if( opt_large_traj ) then
                    large_traj = .true.
                end if
            end if
            
            ! this association could be put much after, very usefull for tools
            if ( present ( tab_prot ) ) then
                this % ptab_prot => tab_prot
                this % nb_protein = size ( tab_prot % array )
            else ! could not pass last argument assum for reading
                b_write_header = .false.
            end if

            ! overwrittes the number of proteins, implemented for complexes and sdamm
            if ( present ( opt_nb_protein ) ) then
                this % nb_protein = opt_nb_protein
                write(*,*) 'overwritte nb_proteins ', this % nb_protein
            end if

            ! setup nb_int and nb_real when bits are known (nb_other is done during new_record)
            call set_variable_field ( this )
            ! local variable
            this % indice_ntraj = get_indice_ntraj ( this )
            this % indice_nconf = get_indice_nconf ( this )
            !write(*,*) 'indice_nconf initialize_record ', this % indice_nconf

            ! initialize the array if not done already, need correct bits
            if ( this % type_record == trajectory_enum ) then
                ! check if not associated, allocate a default ( set up by init_record mandatory )
                if ( associated ( this % ptrajectory ) .eqv. .false. )  then
                    !write(*,*) 'allocate a ptrajectory '
                    allocate ( this % ptrajectory )
                end if

                ! write(*,*) 'will call initialize trajectory '
                !! create the fixed array of one_complexe
                call initialize_trajectory ( this % ptrajectory, this, opt_large_traj = large_traj )
                !! new added
                call reset_position ( this )

            !! else complexes
            else if ( this % type_record == complexe_enum ) then

                !write(*,*) 'initialize_record, test if associated complexe already '
                !write(*,*) 'asoociated pcomplexe ', &
                !    associated ( this % pcomplexe )

                ! check if not associated, allocate a default ( set up by init_record mandatory )
                if ( associated ( this % pcomplexe ) .eqv. .false. ) then
                    ! if not yet associated ( no previous call to new_record, make defualt here
                    ! mainly for tool
                    !write(*,*) 'allocate a pcomplexe '
                    allocate ( this % pcomplexe )
                    call new_listcomplexe ( this % pcomplexe, 0.d0, .false. , .false. )
                                                           !rmsd_min, one_run, one_conf )
                end if

                ! Check for indice of ntraj and nconf in integer of onecomplexe
                ! not sure still necessary
                if ( this % indice_ntraj == 0 ) this % pcomplexe % one_run = .false.
                if ( this % indice_nconf == 0 ) this % pcomplexe % one_conf = .false.

            else
                write ( *, * ) 'ERROR In intialize record, type calculation not known'
                stop 1
            end if

            ! make the format string once, identical for read and write ( only ascii )
            call make_format_string_ascii ( this )

            !! only if open for writing
            if ( b_write_header .eqv. .true. ) then
                call write_header_record ( this )
            end if

        end subroutine initialize_record

        !> Delete record : complexe or trajectory\n
        !! Only nullify the pointer to ptab_protein, do not delete the object.
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param opt_donot_check_array : by default write on disk the last configurations in trajectory before deleting it. Set to true to deactivate ( tool or restart )
        subroutine delete_record ( this, opt_donot_check_array )
            IMPLICIT NONE
            type ( record ) :: this
            logical, optional :: opt_donot_check_array

            logical :: donot_check_array, lopen

            !write(*,*) 'Entry delete record type ', this % type_record

            !! not allocated complexe or trajectory, just return ( because not a pointer on record )
            if ( this % type_record == off_record ) return

            donot_check_array = .false.

            if ( present ( opt_donot_check_array ) ) then
                if ( opt_donot_check_array .eqv. .true. ) donot_check_array = .true.
            end if

            if ( associated ( this % pcomplexe ) ) then
                !write(*,*) 'will delete listcomplex'
                call delete_listcomplexe ( this % pcomplexe )
                deallocate ( this % pcomplexe )
                !write(*,*) 'deallocated pcomplexe'
            end if

            if ( associated ( this % ptrajectory ) ) then
                !write(*,*) 'will delete trajectory'
                call delete_trajectory ( this % ptrajectory, this, donot_check_array )
                deallocate ( this % ptrajectory )
            end if
            nullify ( this % pcomplexe, this % ptrajectory )

            ! nullify pointer to protein
            if ( associated ( this % ptab_prot ) ) then
                !write(*,*) 'ptab_prot associated '
                nullify ( this % ptab_prot )
            !else
                !write(*,*) 'ptab_prot not associated'
            end if
            this % nb_protein = 0

            ! finally close the file
            inquire( unit= this % io_compl,opened=lopen)
            if ( lopen ) then
                close ( this % io_compl )
            end if

            this % io_compl = 0

        end subroutine delete_record

        !> Make a copy of a record
        !! Note : it is a copy of all data members, the content of the array or list is not copied.\n
        !! Use by tools, and by parallelisation to assign a complexe to each thread\n
        !! test if p_copy is already associated ( tool ), or create one ( OpenMP )
        !! @param this :  an instance of \ref mod_record::record "record"
        !! @param p_copy : a pointer to the destination record
        !! @param opt_header : optional copy the header informations or initialize them
        !! @param opt_newsizemax : optional fixe the value of \ref mod_record::record::nb_max_complexe "nb_max_complexe" of the destination record to opt_newsizemax or copy the value from the source
        subroutine copy_record ( this, p_copy, opt_header, opt_newsizemax )
            IMPLICIT NONE
            type ( record ) :: this
            type ( record ), pointer :: p_copy
            logical, optional :: opt_header
            integer, optional :: opt_newsizemax

            integer :: ierr, nb_max_complexe
            logical :: header

            if ( present ( opt_header ) ) then
                header = opt_header
            else
                header = .false.
            end if

            if ( present ( opt_newsizemax ) ) then
                nb_max_complexe = opt_newsizemax
            else
                nb_max_complexe = this % nb_max_complexe
            end if

            if ( associated ( p_copy ) .eqv. .false. ) then
                !write(*,*) 'p_copy is not associated allocate a protein'
                allocate ( p_copy, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In allocating p_copy of record'
                    stop 1
                end if
                nullify ( p_copy % pcomplexe, p_copy % ptrajectory )
            !else
            !    write(*,*) 'p_copy is already associated '
            end if

            ! common data of record
            ! adjust to a different size if needed
            p_copy % nb_max_complexe = nb_max_complexe

            ! will test if 1 is opened
            p_copy % io_compl = 1

            ! setup new type_calc and bits
            p_copy % type_record = this % type_record
            p_copy % record_type_calc = this % record_type_calc
            p_copy % bit_integer = this % bit_integer
            p_copy % bit_energy = this % bit_energy
            p_copy % nb_protein = this % nb_protein
            p_copy % opt_sum = this % opt_sum
            p_copy % nb_types  =  this % nb_types
            p_copy % nb_int        = this % nb_int
            p_copy % nb_real       = this % nb_real
            p_copy % nb_other_real = this % nb_other_real

            p_copy % indice_ntraj = this % indice_ntraj
            p_copy % indice_nconf = this % indice_nconf

            p_copy % box_info = this % box_info

            ! need association, dangerous here for OpenMP, not used anymore
            ! reactivated for tool read_record need cm
            ! to check if OpenMP problem
            if ( associated ( this % ptab_prot ) ) then
                p_copy % ptab_prot => this % ptab_prot
            else
                nullify (p_copy % ptab_prot )
            end if
            ! by default same format, same version
            p_copy % format_ascii = this % format_ascii
            p_copy % version = this % version

            ! if opt_header copy header and string format
            ! may be not necessary, should be recomputed ( with correct change ) when write
            if ( header ) then
                p_copy % format_string_ascii = this % format_string_ascii
                p_copy % format_header_ascii = this % format_header_ascii
                p_copy % header_str =  this % header_str
                p_copy % description_column_str = this % description_column_str
            else
                p_copy % format_string_ascii = ''
                p_copy % format_header_ascii = ''
                p_copy % header_str =  ''
                p_copy % description_column_str = ''
            end if

            ! allocate complexe and traj
            if ( associated ( this % pcomplexe ) ) then
                !write(*,*) 'want a copy of  complexe'
                allocate ( p_copy % pcomplexe, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In allocating p_copy % pcomplexe of record'
                    stop 1
                end if

                call new_listcomplexe ( p_copy % pcomplexe, this % pcomplexe % rmsd_min, &
                                        this % pcomplexe % one_run, this % pcomplexe % one_conf )

            else if ( associated ( this % ptrajectory ) ) then
                write(*,*) 'copy of ptrajectory not yet implemented '
            end if

        end subroutine copy_record

        !> Get version from the header string
        !! @param version_str : string to parse
        !! @param int_version : the version number as integer
        subroutine extract_version_number ( version_str, int_version )
            IMPLICIT NONE
            character, intent(in) :: version_str*(*)
            integer :: int_version

            character*1 :: version_char
            integer :: pos1

            !write(*,*) 'Extract version number', version_str

            pos1 = INDEX ( version_str, ".", .true. )
            ! could check size
            version_char = version_str ( pos1+1:)
            !write(*,*) 'version_char ', version_char
            read ( version_char, fmt='(i1)' ) int_version
            !write(*,*) 'Extracted integer :', int_version
        end subroutine extract_version_number

        !> Create the string for the versino number
        !! @param int_version : version as integer
        !! @return str_version : string with version number, formated for header
        function create_version_str ( int_version ) result ( str_version )
            IMPLICIT NONE
            integer, intent ( in ) :: int_version
            character*10 :: str_version

            character*1 :: char_version

            write ( char_version, fmt='(i1)' ) int_version
            str_version = 'version0.' // char_version

        end function create_version_str

        !> Set bit of \ref mod_record::record::bit_energy "bit_energy" or \ref mod_record::record::bit_integer "bit_integer" to 1\n
        !! Better for internal usage,  or Should consider than the order will not change in futur.
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param position : bit position to set
        !! {@
        subroutine set_bit_energy ( this, position )
            IMPLICIT NONE
            type ( record ) :: this
            integer :: position

            this % bit_energy = ibset ( this % bit_energy, position )

        end subroutine set_bit_energy

        !! @param this : an instance of \ref mod_record::record "record"
        !! @param position : bit position to set
        subroutine set_bit_integer ( this, position )
            IMPLICIT NONE
            type ( record ) :: this
            integer :: position

            this % bit_integer = ibset ( this % bit_integer, position )

        end subroutine set_bit_integer
        !> @}

        !> Set the size of the variable field ( energy and integer ) \ref mod_record::record::nb_int "nb_int" and \ref mod_record::record::nb_real "nb_real"
        !! Convenient function to call set_nb_int and set_nb_real at once. Need the bits to be set correclty before the call \n
        !! Adviced to use this function from sda and external ( tools ).\n
        !! @param this : an instance of \ref mod_record::record "record"
        subroutine set_variable_field ( this )
            IMPLICIT NONE
            type ( record ) :: this

            call set_nb_int ( this )
            call set_nb_real ( this )

        end subroutine set_variable_field

        !> For internal use. Call set_variable_field adviced
        subroutine set_nb_int ( this )
            IMPLICIT NONE
            type ( record ) :: this
            integer :: counter, i

            counter = 0
            do i = 0, BIT_SIZE( this % bit_integer ) - 1
                if ( BTEST( this % bit_integer, i ) ) counter = counter + 1
            !end if
            end do
            this % nb_int = counter
            !write(*,* ) 'set this nb int', this % nb_int
        end subroutine set_nb_int

        !> For internal use. Call set_variable_field adviced
        subroutine set_nb_real ( this )
            IMPLICIT NONE
            type ( record ) :: this

            integer :: counter, i

            counter = 0
            do i = 0, BIT_SIZE( this % bit_energy ) - 1
                if ( BTEST( this % bit_energy, i ) ) then
                    !! electrostatic, only one term at this time
                    !if ( i == 0 ) then
                    !    counter = counter + 1
                    !else
                    if ( this % opt_sum ) then
                        counter = counter + 1
                    else
                        !counter = counter + 2 here error
                        counter = counter + this % nb_types
                    end if

                end if
            end do
            !! add one because of total energy
            this % nb_real = counter + 1
            !write(*,* ) 'set this nb real', this % nb_real
        end subroutine set_nb_real

        !> Function wrapper to mod_onecomplexe::return_compatible_array
        !!
        !! Easier to use, in combination with getNext(). Most of parameters come from record.\n
        !! by default use use internal p_onecompl pointer to one_complexe, but can pass an other pointer as an option.
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param array_int, array_real, array_other : array to fill with data
        !! @param opt_p_onecomplexe : optional, use this pointer to retrieve the data
        subroutine return_compatible_array_wrap ( this, array_int, array_real, array_other, opt_p_onecomplexe  )
            IMPLICIT NONE
            type ( record ) :: this
            real( kind=8 ), dimension(:) :: array_real, array_other
            integer, dimension(:) :: array_int
            type ( one_complexe ), optional, pointer :: opt_p_onecomplexe

            if ( present ( opt_p_onecomplexe ) ) then
                if ( associated ( opt_p_onecomplexe ) ) then
                    call return_compatible_array ( opt_p_onecomplexe, array_int, array_real, array_other, &
                                           this % nb_other_real, &
                                           this % bit_integer, this % bit_energy, this % opt_sum, &
                                           this % nb_types )
                else
                    write(*,*) 'pointer is not associated'
                end if
             ! normal case, use internal pointer
             else
                 call return_compatible_array ( this % p_onecompl, array_int, array_real, array_other, this % nb_other_real, &
                                                this % bit_integer, this % bit_energy, this % opt_sum, &
                                                this % nb_types )
             end if

        end subroutine return_compatible_array_wrap

        !> Function wrapper to mod_onecomplexe::set_onecomplexe
        !!
        !! same idea than return_compatible_array, easy to use with getNext() function.
        !! for sdamm must be called in a loop, maybe to extend later ( in the wrapper for simplicity )
        !! add optional opt_p_onecomplexe, to set a new onecomplexe before it is inserted inside the record
        subroutine set_onecomplexe_wrap ( this, position, box, orient_x, orient_y, &
                                          array_int_compl, array_real_compl, other_real, opt_p_onecomplexe )
            implicit none
            type ( record ) :: this
            ! box could be optional
            integer, dimension ( 3 ), intent ( in ) :: box
            real ( kind=8 ), dimension( 3 ), intent ( in ) :: position, orient_x, orient_y
            real ( kind=8 ), dimension( : ), intent( in ) :: array_real_compl
            real ( kind=8 ), dimension( : ), intent( in ) :: other_real
            integer, dimension ( : ), intent ( in ) :: array_int_compl
            type ( one_complexe ), optional, pointer :: opt_p_onecomplexe

            if ( present ( opt_p_onecomplexe ) ) then
                if ( associated ( opt_p_onecomplexe ) ) then
                    call set_onecomplexe ( opt_p_onecomplexe, position, box, orient_x, orient_y, &
                                           array_int_compl, array_real_compl, other_real, this % nb_other_real, &
                                           this % bit_integer, this % bit_energy, this % opt_sum, this % nb_types )
                else
                    write(*,*) 'pointer is not associated'
                end if
             ! normal case, use internal pointer
            else

                call set_onecomplexe ( this % p_onecompl, position, box, orient_x, orient_y, &
                                       array_int_compl, array_real_compl, other_real, &
                                       this % nb_other_real, this % bit_integer, this % bit_energy, &
                                       this % opt_sum, this % nb_types )
            end if

        end subroutine set_onecomplexe_wrap

        !> Return the maximum indice used for energy.\n
        !! used to set size array_energy of sda_2proteins correctly.
        !! Order of array_energy is fixed.
        !! \todo combine with get_nbterms_X, easier to use.\n
        !!
        !! @param this : an instance of \ref mod_record::record "record"
        !! @return nb_bits : integer, return the maximum used indice
        function get_max_bit_energy ( this ) result ( nb_bits )
            IMPLICIT NONE
            type ( record ) :: this
            integer :: nb_bits

            integer :: i

            nb_bits = 0
            do i = 0, BIT_SIZE( this % bit_energy ) - 1
                if ( BTEST( this % bit_energy, i ) ) then
                    nb_bits = i
                end if
            end do
            !! for the shift from 0
            nb_bits = nb_bits + 1

        end function get_max_bit_energy

        !> Get the minimum size of the array to contain the energy terms.
        !!
        !! Normally for reading, but if force_split get the minimum size of array_1d ( sda_2proteins )
        !! or for sdamm if opt_sum of record is one
        !! @param this :  an instance of \ref mod_record::record "record"
        !! @param opt_force_split : force the size to take account than the energy terms are splitted
        !! @return nb_terms : size
        function get_nbterms_energy_compatible ( this, opt_force_split ) result ( nb_terms )
            IMPLICIT NONE
            type ( record ) :: this
            logical, optional :: opt_force_split

            integer :: i, index_max
            integer :: nb_terms
            logical :: force_split

            !write(*,*) 'Entry get_nbterms_energy_compatible from record '
            ! always total energy
            nb_terms = 1
            index_max = -1
            ! default false
            force_split = .false.

            ! if true do not use this % opt_sum, get the maximum size
            if ( present ( opt_force_split ) ) then
                force_split = opt_force_split
            endif

            do i = 0, BIT_SIZE( this % bit_energy ) - 1
                if ( BTEST( this % bit_energy, i ) ) then
                    index_max = i
                end if
            end do
            !write(*,*) 'index_max ', index_max

            do i = 0, index_max ! included index_max

                !! need to account for terms which are missing
                if ( ( force_split .eqv. .false. ) .and. ( this % opt_sum )  ) then
                    nb_terms = nb_terms + 1
                else
                    nb_terms = nb_terms + this % nb_types
                end if
            end do
            !write(*,*) 'return size comaptible array energy ', nb_terms
        end function get_nbterms_energy_compatible

        !> Similar for integer terms.\n
        !! @param this :  an instance of \ref mod_record::record "record"
        !! @return nb_terms : size
        function get_nbterms_integer_compatible ( this ) result ( nb_terms )
            IMPLICIT NONE
            type ( record ) :: this

            integer :: i, nb_terms

            !write(*,*) 'Entry get_nbterms_integer_compatible '
            nb_terms = 0

            do i = 0, BIT_SIZE( this % bit_integer ) - 1
                if ( BTEST( this % bit_integer, i ) ) then
                    nb_terms = i
                end if
            end do
            ! take into account the shift
            nb_terms = nb_terms + 1

            !write(*,*) 'nb_terms ', nb_terms

        end function get_nbterms_integer_compatible

        !> Reset the current pointer of trajectory or complexe to the first element\n
        !! Kind of rewind.
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param opt_status : optional, return 1 if the list is empty
        subroutine reset_position ( this, opt_status )
            IMPLICIT NONE
            type ( record ) :: this
            integer, optional, intent ( out ) :: opt_status

            !! listcomplexe, points to the first node
            if ( this % type_record == complexe_enum ) then
                this % pcomplexe % p_current => this % pcomplexe % p_first % p_next
                this % p_onecompl => this % pcomplexe % p_current % p_data

                 !! check if already at the end of the complex
                if ( present ( opt_status ) ) then
                    if ( associated ( this % pcomplexe % p_current, this % pcomplexe % p_last ) ) then
                        write(*,*) 'reset_position of an empty file'
                        opt_status = 1
                    else
                        opt_status = 0
                    end if
                end if

            !! new for trajectory, access to the first data, modify access to 0 : nothing
            !! error because use getNext first
            else if ( this % type_record == trajectory_enum ) then
                 this % ptrajectory % current_pos = 1
                 this % p_onecompl => this % ptrajectory % array_trajectory ( 1 )
                 ! correct for error when add_record(), should start from 0, but not logic and problem in trajectory2dcd
                 !this % ptrajectory % current_pos = 0
                 !nullify ( this % p_onecompl )
                 if ( present ( opt_status ) ) then
                    if ( associated ( this % p_onecompl ) ) then
                        opt_status = 0
                    else
                        opt_status = 1
                    end if
                 end if
            end if

        end subroutine reset_position

        !> Update the current pointers to the next position.\n
        !! Update the current node p_current and
        !! and the pointer on a \ref mod_onecomplexe::one_complexe "one_complexe" to the next element of the list or array.
        !! Usefull for traversing complexe or trajectory records.
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param status : return 1 if no data are present, otherwise 0
        !! @param opt_step : optional, specify how many complexes should the jump omit (default: 1)
        !! @param opt_p_onecomplexe : optional, will point to the same than adress than p_onecompl, used for shortcut to p_onecompl
        subroutine getNext ( this, status, opt_step, opt_p_onecomplexe )
            IMPLICIT NONE
            type ( record ) :: this
            integer :: status
            integer, intent ( in ), optional :: opt_step
            type ( one_complexe ), optional, pointer :: opt_p_onecomplexe

            integer :: step, i
            
            step = 1
            if ( present( opt_step ) ) then
                if ( opt_step .gt. 0 ) then
                    step = opt_step
                else
                    write(*,*) 'ERROR (function getNext) opt_step should be greater than 0!'
                    stop 1
                end if
            end if

            ! default no error
            status = 0

            if ( this % type_record == complexe_enum ) then
                !! check "size"
                do i = 1, step
                  if ( this % pcomplexe % p_current % p_next % state /= last ) then
                    this % pcomplexe % p_current => this % pcomplexe % p_current % p_next
                  else
                    status = 1
                    exit
                  end if
                end do
                if ( status .ne. 1 ) then
                    this % p_onecompl => this % pcomplexe % p_current % p_data
                end if

            else if ( this % type_record == trajectory_enum ) then
                ! check size
!                write(*,*) "curr pos: ", this % ptrajectory % current_pos, "; next: ", this % ptrajectory % current_pos + step
                if ( ( this % ptrajectory % current_pos + step ) .le. this % ptrajectory % size_array ) then
                    this % ptrajectory % current_pos = this % ptrajectory % current_pos + step
                    this % p_onecompl => this % ptrajectory % array_trajectory ( this % ptrajectory % current_pos )
                    !write(*,*) 'getNext() current_pos ', this % ptrajectory % current_pos
                else
!                    write(*,*) "ERROR: (function getNext) next pos. (", this % ptrajectory % current_pos + step,") > size (", &
!                                this % ptrajectory % size_array,")!"
                    status = 1
                end if
            else
                write(*,*) 'ERROR In record::getNext()'
                stop 1
            end if

            if ( present ( opt_p_onecomplexe ) ) then
                opt_p_onecomplexe => this % p_onecompl
            end if

        end subroutine getNext

        ! could be used in read_record as well
        ! by default occurencies not written in trajectory, trick for sda_koff, set to 1.

        !> Compute the sum of all occurencies in the complex file\n
        !! Used mainly by sda_2proteins and sda_koff.
        !! If trajectory return the size of the array.
        !! Occurrencies are always stored at the first index of \ref mod_onecomplexe::one_complexe::other_real_ocompl "other_real_ocompl"
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param total_occur : sum of occurencies
        subroutine compute_all_occurencies ( this, total_occur )
                IMPLICIT NONE
                type ( record ) :: this
                integer, intent ( out ) :: total_occur

                integer :: ierr

                total_occur = 0
                ierr = 0
#ifdef DEBUG
                write(*,*) 'Entry compute_all_occurencies'
#endif
                ! not necessary but clean
                if ( this % type_record == trajectory_enum ) then
                     total_occur = this % nb_max_complexe
                     return
                end if

                ! new method
                call reset_position ( this, ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'No data in record for compute_all_occurencies'
                    return
                end if

                do while ( ierr == 0 )
                    !total_occur = total_occur + int ( p_tmp_onecomplexe % other_real_ocompl ( 1 ) )
                    !call getNext ( this, p_tmp_onecomplexe, ierr )
                    total_occur = total_occur + int ( this % p_onecompl %  other_real_ocompl ( 1 ) )
                    call getNext ( this, ierr )
                    !write(*,*) 'ierr ', ierr
                end do

        end subroutine compute_all_occurencies

        !>  Function to write only the header to the output file\n
        !! Called by write_record
        !! @param this : an instance of \ref mod_record::record "record"
        subroutine write_header_record ( this )
            IMPLICIT NONE
            type ( record ) :: this

            if ( this % format_ascii ) then
                call write_header_ascii ( this )
            else
                !! no format with binary, cannot pre_calculte the output
                call write_header_bin ( this )
            end if

        end subroutine write_header_record


        ! keep separate make_format and make_header, until reading is not written
        ! may combine, make_format_header only check version and makes the appropiate format
        ! only makes the most general format, used by read_header

        !> Make the format string for the header, for reading and writing the header
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param to_write : if true make the format for writing, otherwise for reading
        !! @param opt_version : optional, force the version number
        subroutine make_format_header_ascii ( this, to_write, opt_version )
            IMPLICIT NONE
            type ( record ) :: this
            logical :: to_write
            integer, optional :: opt_version

            integer :: last_version

            ! by default use the last one, 2
            last_version = LAST_VERSION_REC
            ! write(*,*) 'last version ', last_version
            if ( present(opt_version) ) then
                last_version = opt_version
            end if

            if ( to_write ) then
                ! add nb_types after nb_proteins( 2i4 ) in version 2
                if ( last_version == 1 ) then
                    this % format_header_ascii =  '(''##'',a10,'' '',a15,'' '',a10,i4,3i3,2b3)'
                else if ( last_version == 2 ) then
                    this % format_header_ascii =  '(''##'',a10,'' '',a15,'' '',a10,2i4,3i3,2b3)'
                else
                    write(*,*) 'problem with version ', last_version
                    stop 1
                end if
            ! cannot use string character ## when reading ??
            else
                if ( last_version == 1 ) then
                    this % format_header_ascii =  '(2x,a10,1x,a15,1x,a10,i4,3i3,2b3)'
                else if  ( last_version == 2 ) then
                    this % format_header_ascii =  '(2x,a10,1x,a15,1x,a10,2i4,3i3,2b3)'
                else
                    write(*,*) 'problem with version ', last_version
                    stop 1
                end if
            end if ! to_write or to_read

        end subroutine make_format_header_ascii

        ! Extended for human readable explanation of the column, description_column_str.\n
        !! @param this : an instance of \ref mod_record::record "record"
        subroutine make_header_ascii ( this )
            IMPLICIT NONE
            type ( record ) :: this

            character(10) :: type_record_str
            integer :: i
            !! for setting the beginning of the string '##'
            logical :: first

            first = .true.
            !write(*,*) 'entry make_header_ascii'

            !initialize the string, already done ??
            this % description_column_str = ''

            do i = 0, BIT_SIZE( this % bit_integer ) - 1
               if ( BTEST( this % bit_integer, i ) ) then
                    if ( i == 0 ) then
                        this % description_column_str = '## RunNb'
                        first = .false.
                    else if ( i == 1 ) then
                        if ( first .eqv. .true. ) then
                            this % description_column_str = trim ( this % description_column_str ) // '##StepNb'
                        else
                            this % description_column_str = trim ( this % description_column_str ) // '  StepNb'
                        end if
                    else if ( i == 2 ) then
                        this % description_column_str = trim ( this % description_column_str ) // ' SolutNb'
                    else if ( i == 3 ) then
                        this % description_column_str = trim ( this % description_column_str) // '  ConfNb'
                    else
                        write(*,*) 'need to extend definition of bit_integer in make_header_ascii '
                        stop 1
                    end if
               end if
            end do

            this % description_column_str = trim ( this % description_column_str) // &
                 '   TransX   TransY   TransZ    Rot1X    Rot1Y    Rot1Z    Rot2X    Rot2Y    Rot2Z'

            ! function to create energy terms, option for variance to write
            call make_header_energy_ascii ( this, .false. )

            if ( associated ( this % pcomplexe ) ) then
                this % description_column_str = trim( this % description_column_str ) // &
                    '       Occur    AvEnergy StdAvEnergy'
            else if ( associated ( this % ptrajectory ) ) then
                this % description_column_str = trim( this % description_column_str ) // &
                    '        Time'
            end if

            if ( this % box_info ) then
                this % description_column_str = trim( this % description_column_str ) // &
                    '     Info Cells'
            end if

            !write(*,*) 'description column ', this % description_column_str
            !write(*,*) 'size of description line ', len ( trim( this % description_column_str ) )

            ! create the most general format, should depend on version, true option to write
            call make_format_header_ascii ( this, .true. )

            ! if fixed to 10 character exactly seems ok
            if (  this % type_record == complexe_enum) then
                type_record_str='complexe  '

            else if ( this % type_record == trajectory_enum ) then
                type_record_str='trajectory'
            end if

            !write(*,*) 'test box_info is wrong ', this % box_info corrected is some cases
            !write(*,*) 'test format_header_ascii ', trim( this % format_header_ascii )
            write ( this % header_str, fmt = this % format_header_ascii ) &
                    this % version, this % record_type_calc % type_calc_str, type_record_str, this % nb_protein, &
                    this % nb_types, &
                    this % bit_integer, this % bit_energy, this % nb_other_real, this % opt_sum, this % box_info

        end subroutine make_header_ascii

        !> header with variance, used by tools
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param if_var : true will add column for variance of the energy terms
        subroutine make_header_energy_ascii ( this, if_var )
            IMPLICIT NONE
            type ( record ) :: this
            ! whether to include variance columns or not
            logical :: if_var

            integer :: i, j
            character ( 48 ) :: shortString
            ! 1 should be enought ! 9 possible different types
            character ( 1 ) :: char_type
            ! like this force the size
            character ( 12 ) :: good_format_string

            !write(*,*) 'Entry make_header_energy_ascii opt_sum ', this % opt_sum
            !write(*,*) 'test ener_colnames(1):!!',ener_colnames(1),'!!'

            ! copy TotEnergy in all cases
            good_format_string = trim(ener_colnames(1))
            this % description_column_str = trim( this % description_column_str )//adjustr(good_format_string)

            do i = 0, BIT_SIZE( this % bit_energy ) - 1
              if ( BTEST( this % bit_energy, i ) ) then

                if ( ( i + 2 ).gt. size( ener_colnames ) ) then
                  write(*,*) 'need to extend definition in make_header_ascii '
                  stop 1
                end if

                ! could check than size ener_colnames+2 <= 12
                if ( ( len(trim(ener_colnames(i+2))) + 2 ) > 12 ) then
                    write(*,*) 'name in ener_colnames is too long ', ener_colnames(i+2)
                    stop 1
                endif

                if ( this % opt_sum ) then

                    ! size is 12 correct for output
                    good_format_string = trim( ener_colnames( i + 2 ) )
                    ! adjustr align on the right and conserv or add blank before
                    this % description_column_str = trim( this % description_column_str )// adjustr(good_format_string)
                else ! not opt_sum, use nb_types
                    do j = 1, this % nb_types

                       write ( char_type, fmt='(i1)' ) j
                       !write(*,*) 'char_type!!',char_type,'!!'
                       good_format_string = trim(ener_colnames( i + 2 ))//'-'//char_type
                       !! here ok, keep 12
                       !write(*,*) 'good_format_string:!!',good_format_string,'!!'
                       ! trick adjustr
                       this % description_column_str = &
                            trim( this % description_column_str)//adjustr(good_format_string)
                    end do

                end if

              end if ! end loop activated bit
            end do
            !write(*,*) 'final_descp'
            !write(*,*) 'desp:!!',trim( this % description_column_str ),'!!'

            ! add the variance header
            if ( if_var ) then
              do i = 0, BIT_SIZE( this % bit_energy ) - 1
                if ( BTEST( this % bit_energy, i ) ) then
                  if ( ( i + 2 ).gt. size( var_colnames ) ) then
                    write(*,*) 'need to extend definition in make_header_ascii '
                    stop 1
                  end if

                  shortString = ' '
                  if ( this % opt_sum ) then
                      this % description_column_str = trim( this % description_column_str )// var_colnames(i + 2)
                  else if ( this % type_record .eq. trajectory_enum ) then
                      shortString = trim( var_colnames(i + 2) )//'1-1  '//trim( var_colnames(i + 2) )//'1-2  '
                      this % description_column_str = trim( this % description_column_str ) // '  '// shortString
                  else
                      shortString = trim( var_colnames(i + 2) )//'1-2  '//trim( var_colnames(i + 2) )//'2-1  '
                      this % description_column_str = trim( this % description_column_str ) // '  '// shortString
                 end if
                end if
              end do
            end if

        end subroutine make_header_energy_ascii

        !> ??
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param if_var : true will add column for variance of the energy terms
        subroutine make_header_int_pairs_ascii( this, if_var )
            IMPLICIT NONE
            type ( record ), intent (inout) :: this
            logical, intent (in) :: if_var

            integer :: i, j, k
            ! 1 should be enought ! 9 possible different types
            character ( 3 ) :: char_type
            ! like this force the size
            character ( 12 ) :: good_format_string

            if ( this % opt_sum ) then
                call make_header_energy_ascii( this, if_var )
                stop 1
            end if

            good_format_string = trim( ener_colnames( 1 ) )
            this % description_column_str = trim( this % description_column_str )//adjustr( good_format_string )

            do i = 0, BIT_SIZE( this % bit_energy ) - 1
              if ( BTEST( this % bit_energy, i ) ) then

                if ( ( i + 2 ).gt. size( ener_colnames ) ) then
                  write(*,*) 'need to extend definition in make_header_ascii '
                  stop 1
                end if

                ! could check than size ener_colnames+2 <= 12
                if ( ( len(trim(ener_colnames(i+2))) + 2 ) > 12 ) then
                    write(*,*) 'name in ener_colnames is too long ', ener_colnames(i+2)
                    stop 1
                endif

                do j = 1, this % nb_types
                   do k = j, this % nb_types
                      write ( char_type, fmt='(i1,a,i1)' ) j, "-", k
                      !write(*,*) 'char_type!!',char_type,'!!'
                      good_format_string = trim(ener_colnames( i + 2 ))//char_type
                      !! here ok, keep 12
                      !write(*,*) 'good_format_string:!!',good_format_string,'!!'
                      ! trick adjustr
                      this % description_column_str = &
                           trim( this % description_column_str)//adjustr(good_format_string)
                   end do ! k
                end do ! j
              end if ! end loop activated bit
            end do
            
            if ( if_var ) then
              do i = 0, BIT_SIZE( this % bit_energy ) - 1
                if ( BTEST( this % bit_energy, i ) ) then
                  if ( ( i + 2 ).gt. size( var_colnames ) ) then
                    write(*,*) 'need to extend definition in make_header_ascii '
                    stop 1
                  end if

                  if ( this % type_record .eq. trajectory_enum ) then
                    do j = 1, this % nb_types
                       do k = j, this % nb_types
                          write ( char_type, fmt='(i1,a,i1)' ) j, "-", k
                          !write(*,*) 'char_type!!',char_type,'!!'
                          good_format_string = trim( var_colnames( i + 2 ) )//char_type
                          ! here ok, keep 12
                          !write(*,*) 'good_format_string:!!',good_format_string,'!!'
                          ! trick adjustr
                          this % description_column_str = &
                               trim( this % description_column_str)//adjustr(good_format_string)
                       end do ! k
                    end do ! j
                  else
                      good_format_string = trim( var_colnames(i + 2) )//'1-2'
                      this % description_column_str = &
                             trim( this % description_column_str ) // adjustr(good_format_string)
                      good_format_string = trim( var_colnames(i + 2) )//'2-1'
                      this % description_column_str = &
                             trim( this % description_column_str ) // adjustr(good_format_string)
                 end if
                end if
              end do
            end if
            
        end subroutine make_header_int_pairs_ascii
        
        !> Make format for all \ref mod_onecomplexe::one_complexe "one_complexe" instances, for read and write.\n
        !! Common to trajectory and complexes. For internal use normally.
        !! to check if used correctly, should be called only once ! here works for read and write
        !! @param this : an instance of \ref mod_record::record "record"
        subroutine make_format_string_ascii ( this )
            IMPLICIT NONE
            type ( record ) :: this

            !write(*,*) 'Entry make_format_string_ascii, Should be called only once in a normal case '
            ! create the string format, only for ascii, if ibox add 3 colomns at the end
            if ( this % box_info ) then
                write( this % format_string_ascii, '("(", i1, "i8,9f9.3,", i2, "es12.4,", i1, "es12.4,3i5)")') &
                                                      this % nb_int, this % nb_real,this % nb_other_real
            else
                write( this % format_string_ascii, '("(", i1, "i8,9f9.3,", i2, "es12.4,", i1, "es12.4)")') &
                                                      this % nb_int, this % nb_real,this % nb_other_real
            end if
            !write(*,*) 'format string_ascii done ', trim( this % format_string_ascii )
        end subroutine make_format_string_ascii

        !> Write header in ascii format
        !! @param this : an instance of \ref mod_record::record "record"
        subroutine write_header_ascii ( this )
            IMPLICIT NONE
            type ( record ) :: this

            !! "on the fly" header should be called only once and saved, later just printed
            if ( len_trim (  this % header_str ) == 0 ) then
                 call make_header_ascii ( this )
            end if
            !write(*,*) 'header:', this % header_str

            !! add description columns
            write ( this % io_compl, fmt='(a)' ) trim ( this % description_column_str )
            write ( this % io_compl, fmt='(a)' ) trim ( this % header_str )
            !! if different of sdamm and ascii, print cm proteins in different lines
            !if ( this % record_type_calc % enum_type_calc /= sdamm ) then
            !                        for sdamm and complexes want the center of mass of protein1
            !if ( ( this % record_type_calc % enum_type_calc /= sdamm ) .or. ( associated ( this % pcomplexe ) ) ) then

            ! all complexes use 2 proteins mass centers(even sdamm complexe), sda_2proteins need for trajectory as well
            if ( associated ( this % pcomplexe ) .or. ( this % record_type_calc % enum_type_calc /= sdamm ) ) then

                !write ( this % io_compl, fmt='("#",3f8.3)' ) this % ptab_prot % array ( 1 ) % center_protein
                !write ( this % io_compl, fmt='("#",3f8.3)' ) this % ptab_prot % array ( 2 ) % center_protein

                ! first set always defined
                write ( this % io_compl, fmt='("#",3f8.3)' ) this % ptab_prot % array ( 1 ) % center_protein
                ! second set not defined if only homodimer ( in sdamm )
                !write(*,*) 'test size set of proteins ', size ( this % ptab_prot % set_of_protein )
                if ( size ( this % ptab_prot % set_of_protein ) > 2 ) then
                    write ( this % io_compl, fmt='("#",3f8.3)' ) &
                            this % ptab_prot % array ( this % ptab_prot % set_of_protein ( 2,1 ) ) % center_protein
                ! else only one type, use same cm
                else
                    write ( this % io_compl, fmt='("#",3f8.3)' ) this % ptab_prot % array ( 1 ) % center_protein
                end if

            end if

        end subroutine write_header_ascii

        !> Write header in binary format, always print cm
        !! @param this : an instance of \ref mod_record::record "record"
        subroutine write_header_bin ( this )
            IMPLICIT NONE
            type ( record ) :: this

            !write(*,*) 'Entry write_header_bin '
            !write(*,*) 'version: ', this % version
            !write(*,*) 'record_type_calc: ', this % record_type_calc % enum_type_calc

            write ( this % io_compl ) this % version, this % record_type_calc % enum_type_calc, this % type_record, &
                                      this % nb_protein, this % nb_types, this % bit_integer, this % bit_energy, &
                                      this % nb_other_real, &
                                      this % opt_sum, this % box_info, this % ptab_prot % array ( 1 ) % center_protein, &
                                      this % ptab_prot % array ( 2 ) % center_protein

        end subroutine write_header_bin

        !> Read the header of the record, call appropriate ascii or binary sub-functions\n
        !! Setup internal variables with inforamtion from the header.
        !! Adviced to use this function directly.\n
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param filename : file to read
        !! @param opt_cm : optionaly read the center of proteins ( only for sda_2proteins ). Safe with sdamm : do nothing
        !! @param opt_nb_record : optionaly as input, count the number of lines ( set opt_nb_record for return and \ref mod_record::record::nb_max_complexe "nb_max_complexe" )
        subroutine read_header_record ( this, filename, opt_cm, opt_nb_record )
            IMPLICIT NONE
            type ( record ) :: this
            character, intent ( in ) :: filename*(*)
            integer, intent ( out ), optional :: opt_nb_record
            real ( kind = 8 ), dimension ( 2, 3 ), intent(out), optional :: opt_cm

            type ( type_calculation ) :: type_calc
            integer :: nb_prot, nb_type, nb_other_real, int_type_record
            integer ( kind=2 ) :: bit_integer, bit_real
            real ( kind = 8 ), dimension ( 2, 3 ) :: cm
            logical :: opt_sum, box_info
            integer :: io_err
            ! fixed size 10 character ( for binary )
            character*10 version
            integer :: int_version

            !write(*,*) 'Entry read_header_record '
            cm = 0.d0
            ! with fortran and valgrind, not initilaized
            io_err = 0

            ! depends on version, need initialization
            nb_type = 0

            ! test if file exists, setup format_ascii and version
            ! finish on rewind, to read data from file
            call test_binary ( this, filename, version )

            ! extract the number verion
            call extract_version_number ( version, int_version )
            !write(*,*) 'extract version integer ', int_version

            ! should test the version, function called by restart or tool, object not created yet
            if ( int_version /= LAST_VERSION_REC ) then
                write(*,*)
                write(*,*) 'WARNING You are reading an older version of the record, nb_types is not accessible '
                write(*,*) 'WARNING Your version is ', int_version
                write(*,*) 'WARNING The current is ', LAST_VERSION_REC
                write(*,*)
            end if

            ! read header, do not assign the variable to record
            if ( this % format_ascii ) then

                write( *, *) 'Ascii file detected ', this % io_compl, trim ( filename )
                open ( this % io_compl, file=filename, form='formatted', action='read', iostat = io_err )
                if ( io_err /= 0 ) then
                     write(*,*) 'ERROR problem in opening ascii file ', filename
                     write(*,*) 'ERROR code ', io_err
                     STOP 1
                end if

                ! add int_version for version number
                call read_header_ascii ( this, type_calc, int_type_record, &
                                         nb_prot, nb_type, bit_integer, bit_real, nb_other_real, opt_sum, box_info, cm, &
                                         int_version )

            else
                write( *, *) 'Binary file detected '
                open ( this % io_compl, file=filename, form='unformatted', action='read', iostat = io_err )
                if ( io_err /= 0 ) then
                     write(*,*) 'problem in opening binary file ', filename
                end if
                ! add a version number
                call read_header_binary ( this, type_calc, int_type_record, nb_prot, nb_type, &
                         bit_integer, bit_real, nb_other_real, opt_sum, box_info, cm, int_version )
            end if

            ! assign variable to record
            this % record_type_calc = type_calc
            ! assign complexe or trajectory enumeration
            this % type_record = int_type_record
            this % nb_protein = nb_prot

            !write(*,*) 'read nb_type from header ', nb_type
            !write(*,*) 'actual this % version_number ', this % version_number
            if ( int_version /= LAST_VERSION_REC ) then
                !write(*,*) 'need to set up nb_types '
                if ( type_calc % enum_type_calc /= sdamm ) then
                    write(*,*) 'It is a sda_2proteins type of record, setup nb_types to 2 '
                    this % nb_types = 2
                else if ( this % nb_types .eq. 0 ) then
                    write(*,*) 'WARNING Something is wrong: sdamm record with nb_types = 0; will set default: 1'
                    this % nb_types = 1

                    ! apply only from v1 to v2
                    if ( .not.opt_sum ) then
                        write(*,*)
                        write(*,*) 'ERROR opt_sum is false, you may have problems to read the file'
                        write(*,*)
                        stop 1
                    end if
                end if
            else
                write(*,*) 'read actual version nb_type from the header=', nb_type
                this % nb_types = nb_type
            end if

            ! need to assign version_number at a point
            this % version_number = int_version
            !write(*,*) 'assign int_version to version number, need for reading file ', this % version_number
            this % version = create_version_str( int_version )

            this % bit_integer = bit_integer
            this % bit_energy = bit_real
            this % nb_other_real = nb_other_real
            this % opt_sum = opt_sum
            this % box_info = box_info

            write(*,*) '========='
            write(*,*) 'Summary header '
            write(*,*) 'version            ', this % version
            write(*,*) 'version number     ', this % version_number
            write(*,*) 'type calc          ', type_calc % type_calc_str
            write(*,*) 'int type calc      ', type_calc % enum_type_calc
            write(*,*) 'int type record    ', int_type_record
            write(*,*) 'nb protein         ', this % nb_protein
            write(*,*) 'nb_types           ', this % nb_types
            write(*,*) 'bit_integer        ', this % bit_integer
            write(*,*) 'bit_real           ', this % bit_energy
            write(*,*) 'nb other real      ', this % nb_other_real
            write(*,*) 'opt_sum            ', this % opt_sum
            write(*,*) 'box_info           ', this % box_info

            this % indice_nconf = get_indice_nconf ( this )
            this % indice_ntraj = get_indice_ntraj ( this )

            write(*,*) 'indice conf        ', this % indice_nconf
            write(*,*) 'indice  traj       ', this % indice_ntraj

            if ( type_calc % enum_type_calc /= sdamm ) then
                 write(*,*) 'cm 1    ', cm ( 1,:)
                 write(*,*) 'cm 2    ', cm ( 2,:)

                ! test for tool, maybe dangerous if wrong
                if ( associated ( this % ptab_prot ) ) then
                    write(*, fmt = '("cm prot1   ",3f8.3)' ) this % ptab_prot % array ( 1 ) % center_protein
                    write(*, fmt = '("cm prot2   ",3f8.3)' ) this % ptab_prot % array ( 2 ) % center_protein
                end if
                if ( present ( opt_cm ) ) then
                    !write(*, fmt = '("cm 1, not associated to prot  ",3f8.3)' ) cm ( 1, : )
                    !write(*, fmt = '("cm 2, not associated to prot  ",3f8.3)' ) cm ( 2, : )
                    !write (*,*) 'assign to opt_cm'
                    opt_cm = cm
                end if
            end if

            ! can read the number of records
            if ( present ( opt_nb_record ) ) then
                opt_nb_record = 0
                call count_records ( this, opt_nb_record )
                !write(*,*) 'nb_max_complexe    ', opt_nb_record
                this % nb_max_complexe = opt_nb_record
            else
                write(*,*) 'size record has not been determined'
                this % nb_max_complexe = 0
            end if
            ! can setup fully variable fields now
            call set_variable_field ( this )
            call print_bits ( this )
            !write(*,*) 'nb_int', this % nb_int
            !write(*,*) 'nb_real', this % nb_real
            !write(*,*) 'nb_other_real', this % nb_other_real

        end subroutine read_header_record

        !> Internal use.\n
        subroutine read_header_binary ( this, type_calc, int_type_record, nb_prot, nb_type, &
                                        bit_integer, bit_real, nb_other_real, opt_sum, box_info, cm, int_version )
            IMPLICIT NONE
            type ( record ) :: this
            type ( type_calculation ) :: type_calc
            integer :: int_type_record
            integer, intent ( inout ) :: nb_prot, nb_type, nb_other_real
            integer ( kind=2 ), intent ( inout ) :: bit_integer, bit_real
            logical, intent ( out ) :: opt_sum, box_info
            real ( kind = 8 ), dimension ( 2, 3 ) :: cm
            integer :: int_version

            integer :: io_err

            cm = 0.d0

            !write(*,*) 'Entry read_header_binary '
            !write(*,*) 'io_compl ', this % io_compl

            !! type_calc is an integer in binary format, cm1 and cm2 always printed in binary, but not used later
            if ( int_version == LAST_VERSION_REC ) then
                read ( this % io_compl, iostat = io_err ) this % version, type_calc % enum_type_calc, int_type_record, &
                   nb_prot, nb_type, bit_integer, bit_real, nb_other_real, opt_sum, box_info, cm(1,:), cm(2,:)
            else if ( int_version == 1 ) then
                read ( this % io_compl, iostat = io_err ) this % version, type_calc % enum_type_calc, int_type_record, &
                   nb_prot, bit_integer, bit_real, nb_other_real, opt_sum, box_info, cm(1,:), cm(2,:)
            ! not yet implemented or error
            else
                write(*,*) 'mod_record::read_header_binary does not know this version: ', int_version
                stop 1
            end if

            if ( associated ( this % ptab_prot ) ) then
                this % ptab_prot % array ( 1 ) % center_protein = cm(1,:)
                this % ptab_prot % array ( 2 ) % center_protein = cm(2,:)
            !! to add else make cm1, cm2 as argument
            end if

            if ( io_err /= 0 ) then
                write(*,*) 'Problem in reading header binary '
            end if

            !! make string for type_calc
            type_calc % type_calc_str = get_string ( type_calc )

        end subroutine read_header_binary

        !> Internal use.\n
        subroutine read_header_ascii ( this, type_calc, int_type_record, &
                                       nb_prot, nb_type, bit_integer, bit_real, nb_other_real, opt_sum, box_info, cm, &
                                       int_version )
            IMPLICIT NONE
            type ( record ) :: this
            type ( type_calculation ) :: type_calc
            integer :: int_type_record
            integer, intent ( inout ) :: nb_prot, nb_type, nb_other_real
            integer ( kind=2 ), intent ( inout ) :: bit_integer, bit_real
            real ( kind = 8 ), dimension ( 2, 3 ) :: cm
            logical, intent ( out ) :: opt_sum, box_info
            integer  :: int_version

            character(10) :: type_record_str
            !! version already set during test_binary, may add later
            character(10) :: version_tmp
            integer :: io_err, i

            !! get general format_header, option to_write is false
            call make_format_header_ascii ( this, .false., int_version  )
            !write(*,*) 'format header:',this % format_header_ascii

            !! Read description colonnes
            read ( this % io_compl, fmt='(a)' ) this % description_column_str

            ! new version, read nb_type
            !if ( present ( opt_version ) .and. ( opt_version == LAST_VERSION_REC ) ) then
            if ( int_version == LAST_VERSION_REC ) then
                read ( this % io_compl, fmt = this % format_header_ascii, iostat = io_err ) &
                version_tmp, type_calc % type_calc_str, type_record_str, nb_prot, nb_type ,bit_integer, bit_real, &
                nb_other_real, opt_sum, box_info

            ! read old format, nb_types is absent
            !else if (  present ( opt_version ) .and. ( opt_version == 1 ) ) then
            else if ( int_version == 1 ) then
                read ( this % io_compl, fmt = this % format_header_ascii, iostat = io_err ) &
                version_tmp, type_calc % type_calc_str, type_record_str, nb_prot, bit_integer, bit_real, &
                nb_other_real, opt_sum, box_info
            ! not yet implemented or error
            else
                write(*,*) 'mod_record::read_header_ascii does not know this version: ', int_version
                stop 1
            end if

            if ( io_err /= 0 ) then
                write(*,*) 'problem in reading header '
                stop 1
            end if

            ! make enumeration conversion
            type_calc % enum_type_calc = get_type_enum ( type_calc )

            ! read cm in ascii for all sda_2proteins types
            if ( type_calc %  enum_type_calc /= sdamm ) then
                !write(*,*) 'read cm prot1 and 2 '

                ! added test for tool read_record,
                ! when read_header first time array_prot not created yet
                if ( associated ( this % ptab_prot ) ) then
                    do i = 1, 2
                        read ( this % io_compl, fmt='(1x,3f8.3)' ) this % ptab_prot % array ( i ) % center_protein
                    end do
                ! need to read for nothing, important for count line
                ! add otpion cm, optional in calling read_header
                else
                    do i = 1, 2
                        read ( this % io_compl, fmt='(1x,3f8.3)' ) cm(  i, : )
                    end do
                    !read ( this % io_compl, fmt='(1x,3f8.3)' ) cm ( 2, : )
                end if
            end if
            !write(*,*) 'read header ascii'
            !write(*,*) 'nb_protein ', nb_prot
            !write(*,*) 'type_calc_str ', type_calc % type_calc_str
            !write(*,*) 'type_record_str ', type_record_str

            if ( type_record_str == 'complexe' ) then
                !write(*,*) 'record is complexe'
                int_type_record = complexe_enum
            else if ( type_record_str == 'trajectory' ) then
                !write(*,*) 'record is trajectory'
                int_type_record = trajectory_enum
            else
                write(*,*)
                write(*,*) 'ERROR type_record_str in read_header_ascii is unknown: ',type_record_str
                write(*,*) 'Check the header of the record file '
                write(*,*)
                stop 1
             end if

        end subroutine read_header_ascii

        ! add tab_protein as entry, need to use for sdamm and sda_2proteins if OpenMP
        ! make changes, now expects a real fixed format, no more test sdamm or sda_2proteins

        !> Add a new entry into the record\n
        !! Creates a new \ref mod_onecomplexe::one_complexe "one_complexe" if needed ( in case of complex )\n
        !! This function is not thread-safe, critical sections are needed
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param array_prot : an instance of \ref mod_array_protein::array_protein_type "array_protein_type"
        !! @param ntraj : the current trajectory number
        !! @param nstep : the current step number
        !! @param array_energy : input array which contains energies terms
        !! @param time : the current time
        !! @param opt_nconf : optionaly the current conformation
        subroutine add_record ( this, array_prot, ntraj, nstep, &
                                array_energy, time , opt_nconf )
            IMPLICIT NONE
            type ( record ) :: this
            type ( protein ), dimension ( : ), intent ( in ) :: array_prot
            real ( kind=8 ), dimension ( :, : ), intent ( in ) :: array_energy
            real ( kind=4 ), intent ( in ) :: time
            integer, intent ( in ) :: ntraj, nstep
            integer, optional, intent ( in ) :: opt_nconf

            ! bad should not include constant here
            real ( kind=8 ), dimension ( 20 ) :: array_energy_1d
            integer :: nconf

            ! only for complexes
            real ( kind=8 ) ::dist_max

            !! fix the size of other at the maximum, can receive any data
            real ( kind=8 ), dimension ( 3 ) :: other_real
            integer, dimension ( 4 ) :: array_integer

            integer :: counter, i, ierr
            
            !write(*,*) 'entry add record time ', time
            !write(*,*) 'current position ', this % ptrajectory % current_pos
            !write(*,*)
            !write(*,*) 'nb_real', this % nb_real

            ! use opt_conf for flex with sda_2protein, nconf can be solute 1 or 2, if both problems
            ! argument current_conf, only works for 1 flexible protein
            ! cannot make like for sdamm, because only one protein send
            if ( present ( opt_nconf ) ) then
                nconf = opt_nconf
            else
                nconf = 1
            end if

            array_integer ( 1 ) = ntraj
            array_integer ( 2 ) = nstep
            counter = 0
            i = 0
            array_energy_1d = 0.d0

            ! default value for both complexe and trajectory. for sdamm will overwritte for every proteins
            ! if both flexible, same need for sda_2proteins...
            array_integer ( 4 ) = nconf

            ! if trajectory
            if ( this % type_record == trajectory_enum ) then

                !write(*,*) 'add record trajectory nb_types ', this % nb_types
                !write(*,*) 'current_pos ', this % ptrajectory % current_pos
                ! trajectory other real contains only the time
                other_real ( 1 ) = time

                ! sda_2proteins, should be 1, how can it work ?? sda_2proteins send only prot 2 ! trick !
                ! could call only once getNext, and update with array with i if sure array size is compatible
                ! here works even if array_size is smaller, not a modulo of nb_proteins
                do i = 1, size ( array_prot )

                    !set up nprotein, printed only if sdamm
                    array_integer ( 3 ) = array_prot ( i ) % id

                    ! get the conformation
                    if  ( associated ( array_prot ( i ) % p_flex ) ) then
                         !write(*,*) 'sdamm use current_conf', array_prot ( i ) % p_flex % current_conf
                         array_integer ( 4 ) = array_prot ( i ) % p_flex % current_conf
                    end if

                    array_energy_1d(1: size(array_energy(:,i)) ) = array_energy(:,i)

                    call set_onecomplexe_wrap ( this, array_prot ( i ) % position, array_prot ( i ) % box, &
                                                array_prot ( i ) % orientation_x, array_prot ( i ) % orientation_y, &
                                                array_integer, array_energy_1d, other_real )

                    ! moved to next after insertion
                    call getNext ( this, ierr )
                    ! check if array is full, write to disk and return at the beginning of the array
                    ! ierr == 1 if getNext outside array, force to write and reset at first
                    if ( ierr /= 0 ) then
                        ! write to disk
                        call write_array ( this % ptrajectory, this )

                        ! reset to position 1
                        call reset_position ( this )
                    end if

                end do ! loop protein

            ! at the moment look only for protein2, for sda, to extend to set of protein
            ! but easy to process for any good formatted input, could receive a set of protein from sdamm
            else if ( this % type_record == complexe_enum ) then

                    !write(*,*) 'add record complexe'
                    array_energy_1d(1: size(array_energy(:,1)) ) = array_energy(:,1)
                    ! do nothing if total_energy is higher than the max energy
                    ! sda_2proteins send only the relevant 1D array in 2D array
                    ! modified, energy_1d now total energy in 1,1
                    if ( ( array_energy( 1, 1 ) > this % pcomplexe % max_energy ) .and. &
                         (  this % pcomplexe % nb_complexe == this % nb_max_complexe ) ) then
                        !write(*,*) 'energy is higher than max_energy, nothing to do energy max_energy', &
                        !            array_energy ( 1, 1 ), this % pcomplexe % max_energy
                        return
                    endif

                    ! needed for srms, added a if for read_record
                    ! to check if still needed, old code
                    if ( associated ( array_prot ( 1 ) % p_sogrid ) ) then
                        dist_max = array_prot ( 1 ) % p_sogrid % distance_max
                        !write(*,*) 'distmax2 ', dist_max
                    else
                        ! at home need 1., with 0. numerical error ? should disable check if explicetely 0
                        ! and do not care
                        ! see loop overlist, to do at work
                        dist_max = 1.
                    end if

                    ! transform to usefull type for the listcomplexe
                    ! nb protein, here default for sda,
                    ! could be this % nb_types
                    array_integer ( 3 ) = 2
                    ! other should be set directly by complexes
                    !other_real = 0.

                    ! write(*,*) 'Call fill_listcomplexe '
                    ! here send all info for creating the one_complexe
                    call fill_listcomplexe ( this % pcomplexe, this,  array_prot ( 1 ) % position, &
                                             array_prot ( 1 ) % orientation_x, array_prot ( 1 ) % orientation_y, &
                                             !array_energy4, array_integer, dist_max )
                                             array_energy_1d, array_integer, dist_max )

            end if

        end subroutine add_record

        !! just for tests, extend complexes to sdamm
        !! extend add_record
        !! want to record potential complexes ( start with only 2 ),
        !! time should allow to distinguish the complexes
        !!
        !! add size_array_energy in input, to avoid fixed size (20), maybe to add in
        !! always receive prot1-prot2, only good array from sda_2proteins energies
!        subroutine add_complexes_sdamm ( this, array_prot, nb_prot1, nb_prot2, &
!                                         ntraj, nstep, &
!                                         array_energy, size_array_energy, time )
!            implicit none
!            type ( record ) :: this
!            type ( protein ), dimension ( : ), intent ( in ) :: array_prot
!            real ( kind=8 ), dimension ( :, : ), intent ( in ) :: array_energy
!            real ( kind=4 ), intent ( in ) :: time
!            integer, intent ( in ) :: nb_prot1, nb_prot2, size_array_energy, ntraj, nstep
!
!
!            ! dimension ( nb_prot==2, nb_energy_terms )
!            real ( kind=8 ), dimension ( :, : ), allocatable :: position4, orient4_x, orient4_y
!            !real ( kind=8 ), dimension ( 20 ) :: array_energy4
!            real ( kind=8 ), dimension(:,:), allocatable :: array_energy4
!            ! need to extend as well, easy use basecl % nb_protein
!            integer, dimension ( 2, 4 ) :: array_integer
!            !real(kind=8) :: other_real
!            integer :: ierr
!
!            write(*,*) 'Entry add_complexes_sdamm '
!            write(*,*) 'associated pcomplexes ', associated ( this % pcomplexe )
!            write(*,*) 'size_array_energy ', size_array_energy
!
!            !! will contain the total energy, whatever the other terms
!            write(*,*) 'energy (1,1)', array_energy(1,1)
!
!            ! use basecl % nb_protein for extension
!            allocate ( position4 (2,3), orient4_x(2,3), orient4_y(2,3), stat = ierr )
!            if ( ierr /=0 ) then
!                write(*,*) 'error in allocating pos4, orient4'
!            end if
!
!            allocate ( array_energy4( size_array_energy, 2), stat = ierr )
!            if ( ierr /=0 ) then
!                write(*,*) 'error in allocating array_energy4 size 2 and ,', size_array_energy
!            end if
!
!            ! common
!            array_integer = 0
!            array_integer ( :,1 ) = ntraj
!            array_integer ( :,2 ) = nstep
!            !counter = 0
!            !i = 0
!            array_energy4 = 0.d0
!
!            ! default value for both complexe and trajectory. for sdamm need to check for every proteins
!            ! if both flexible, same need for sda_2proteins...
!            !array_integer ( 4 ) = nconf
!
!            ! fill array, could be a loop
!            !do i = 1, 2
!            position4(1,:) =  array_prot ( nb_prot1 ) % position
!            orient4_x(1,:) =  array_prot ( nb_prot1 ) % orientation_x
!            orient4_y(1,:) =  array_prot ( nb_prot1 ) % orientation_y
!            ! fill protein number
!            array_integer(1,3) = nb_prot1
!            ! write conf number
!            !array_integer(1,4)
!            !write(*,*) 'size array_energy nb_prot1 ', size( array_energy(:,nb_prot1) )
!            array_energy4(:,1) = array_energy(:,1) !array_energy(:,nb_prot1)
!
!            position4(2,:) = array_prot ( nb_prot2 ) % position
!            orient4_x(2,:) = array_prot ( nb_prot2 ) % orientation_x
!            orient4_y(2,:) = array_prot ( nb_prot2 ) % orientation_y
!            array_integer(2,3) = nb_prot2
!            !array_integer(2,4) =
!            !array_energy4(:,2) = array_energy(:,nb_prot2)
!            array_energy4(:,2) = array_energy(:,2)
!
!            call fill_listcomplexe_sdamm ( this % pcomplexe, this, &
!                                           position4, orient4_x, orient4_y, &
!                                           array_energy4, array_integer, &
!                                           ! dist_max used ?? to check
!                                           1.0d0 )
!
!        end subroutine add_complexes_sdamm

        ! general function for getting indices
        ! Record will store this value, allow to find the indice of int_ocompl for ListComplex
        ! called only once, during initialisation

        !> Function to get the indice of the trajectory number.\n
        !!
        !! Function to use in collaboration with \ref mod_onecomplexe::return_compatible_array "return_compatible_array", \ref mod_onecomplexe::return_array_int "return_array_int" and
        !! \ref mod_onecomplexe::return_array_real "return_array_real"
        !! Safer to use than directly access the data, easier to maintain when there will be modification.
        !! @param this : an instance of \ref mod_record::record "record"
        !! @return indice : indice where to get the trajectory number from onecomplexe::array_int, return 0 if not present
        function get_indice_ntraj ( this ) result ( indice )
            IMPLICIT NONE
            type ( record ) :: this

            integer :: indice
            ! ntraj first bit, easy to test
            if ( BTEST( this % bit_integer, 1 ) .eqv. .false. ) then
                indice = 0
            else
                indice = 1
            endif

        end function get_indice_ntraj

        !> Same for conformation number
        !! @param this : an instance of \ref mod_record::record "record"
        !! @return indice : indice where to get the conformation number from onecomplexe::array_int, return 0 if not present
        function get_indice_nconf ( this ) result ( indice )
            IMPLICIT NONE
            type ( record ) :: this
            integer :: indice

            integer :: i

            indice = 0
            ! nconf is a 4, test if activated, then loop to find its indice,
            ! why this complication if always at 4 ! maybe not always internal
            if ( BTEST( this % bit_integer, 3 ) .eqv. .false. ) then
                indice = 0
            ! search until the last one
            else
                do i = 0,  BIT_SIZE( this % bit_integer ) - 1
                   if ( BTEST( this % bit_integer, i ) ) then
                        indice = indice + 1
                   end if
               end do
            endif
            !write(*,*) 'get_indice_nconf, ', indice
        end function get_indice_nconf

        !> Print info about bits
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param opt_for_choice : bool if true less output used in tool when a choice is expected, default false
        subroutine print_bits ( this, opt_for_choice )
            USE mod_gridtype
            IMPLICIT NONE
            type ( record ) :: this
            logical, optional :: opt_for_choice

            integer :: i
            logical :: bool_choice

            bool_choice = .false.

            if ( present ( opt_for_choice ) ) then
                bool_choice = opt_for_choice
            end if

            if ( bool_choice .eqv. .false. ) then
                write(*,fmt= '(''Value of bits for integer:'', I10, B10,'' Size: '', I5 )' ) &
                              this % bit_integer, this % bit_integer, BIT_SIZE( this % bit_integer )

                do i = 0, BIT_SIZE( this % bit_integer ) - 1
                    if ( BTEST( this % bit_integer, i ) ) then
                        write( *, fmt = '(''refer to : '')', ADVANCE = "NO" )
                        if ( i == 0 ) then
                            write(*,*) 'run number'
                        else if ( i == 1 ) then
                            write(*,*) 'step number'
                        else if ( i == 2 ) then
                            write(*,*) 'protein number'
                        else if ( i == 3 ) then
                            write(*,*) 'conformation number'
                        else
                            write(*,*) 'not implemented'
                        end if
                    end if
                end do

                write(*,fmt= '(''Value of bits for energy :'', I10, B10,'' Size: '', I5 )' ) &
                                this % bit_energy, this % bit_energy, BIT_SIZE( this % bit_energy )

            !! end if bool_choice
            end if

            !! in tool read_record only this output is needed
            do i = 0, BIT_SIZE( this % bit_energy ) - 1
                if ( BTEST( this % bit_energy, i ) ) then
                    write( *, fmt = '(''term : '')', ADVANCE = "NO" )
                    call print_gridtype ( i + 1 )
                end if
            end do

        end subroutine print_bits

        !> Print infos about the record.\n
        !! for complexes can ask to recompute the dispersion when printing.
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param opt_recompute_dispersion : optionaly recompute the dispersion ( only for complexes ). Default true.
        subroutine print_record ( this, opt_recompute_dispersion )
            IMPLICIT NONE
            type ( record ) :: this
            logical, optional :: opt_recompute_dispersion

            logical :: recompute_dispersion

            recompute_dispersion = .true.
            if ( ( present ( opt_recompute_dispersion ) ) .and. ( opt_recompute_dispersion .eqv. .false. ) ) then
                recompute_dispersion = .false.
            end if

            if ( this % type_record == complexe_enum ) then
                call print_listcomplexe ( this % pcomplexe, recompute_dispersion )
            else if ( this % type_record == trajectory_enum ) then
                call print_array ( this % ptrajectory, this )
            end if

        end subroutine print_record


        !> Main function to write the record.\n
        !! Make internal call to write header.\n
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param opt_recompute_dispersion : optionaly recompute the dispersion ( only for complexes ). Default true.
        subroutine write_record ( this, opt_recompute_dispersion )
            IMPLICIT NONE
            type ( record ) :: this
            logical, optional :: opt_recompute_dispersion

            logical :: recompute_dispersion

            ! see for print_record, same reason
            recompute_dispersion = .true.
            if ( present ( opt_recompute_dispersion ) ) then
                if ( opt_recompute_dispersion .eqv. .false. ) then
                  recompute_dispersion = .false.
                endif
            end if

            ! rewind file for complexe and rewrite the header
            if ( this % type_record == complexe_enum  ) then
                 !write(*,*) 'entry write_record for complexe, rewind the file '
                 ! for the moment, rewind, later do not but depends of ascii or binary
                 rewind ( this % io_compl )
                 call write_header_record ( this )
                 !write(*,*) 'header written'
            end if
            ! if traj append by default

            ! different function for printing, will take care of the format
            if ( this % type_record == complexe_enum ) then
                 ! write data of the list, ascii binary see later
                 call write_listcomplexe ( this % pcomplexe, this, recompute_dispersion )

            ! if trajectory
            else if ( this % type_record == trajectory_enum ) then
                 ! header printed only once at the beginning
                 call write_array ( this % ptrajectory, this )
            end if

        end subroutine write_record

        !> Read the record and fill the array or list with the data
        !!
        !! The file is first "rewinded"
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param opt_rebuild_sum_square : optional, if the square energy must be recomputed (restart complexe)
        subroutine read_file_record ( this, opt_rebuild_sum_square )
            IMPLICIT NONE
            type ( record ) :: this
            logical, optional :: opt_rebuild_sum_square

            character*256 string
            integer :: io_err
            logical :: rebuild_sum_square

            ! do not know how to avoid to reread the data in header
            ! in binary the size is fixed, could shift
            type ( type_calculation ) :: type_calc_tmp
            integer :: int_type_record_tmp
            integer :: nb_prot_tmp, nb_other_real_tmp, nb_type_tmp
            integer ( kind=2 ) :: bit_integer_tmp, bit_real_tmp
            real ( kind = 8 ), dimension ( 2,3 ) :: cm_tmp
            logical :: opt_sum_tmp, box_info

            io_err = 0
            opt_sum_tmp = .false.
            box_info = .false.
            cm_tmp = 0.d0
            !! default false, only used for restarting complex
            rebuild_sum_square = .false.

            !write(*,*) 'entry read_file_record '
            !write(*,*) 'check version ', this % version_number
            !write(*,*) 'filename io_compl ', this % io_compl

            rewind ( this % io_compl )

            if ( present ( opt_rebuild_sum_square ) ) then
                rebuild_sum_square = opt_rebuild_sum_square
            endif
            !write(*,*) 'rebuild sum_square ', rebuild_sum_square

            ! create format for reading ascii
            if ( this % format_ascii ) then
                !! pass description column and header, already read
                read ( this % io_compl, fmt='(a)' ) string
                read ( this % io_compl, fmt='(a)' ) string
                !write(*,*) 'string ', trim(string)

                ! complexes and sdamm possible now
                if ( ( this % record_type_calc % enum_type_calc /= sdamm ) .or. &
                     ( associated ( this % pcomplexe ) ) ) then
                    !! read 2 extra lines for cm first line
                    read ( this % io_compl, fmt='(a)' ) string
                    read ( this % io_compl, fmt='(a)' ) string
                    !write(*,*) 'string ', trim(string)
                end if

                ! prepare format for reading onecomplexe, maybe already done
                call make_format_string_ascii ( this )
                !write ( *, * ) 'format_string_ascii ', this % format_string_ascii
            else
                ! binary, need to reread the header, cm always printed in binary
                call read_header_binary ( this, type_calc_tmp, int_type_record_tmp, &
                     nb_prot_tmp, nb_type_tmp, bit_integer_tmp, bit_real_tmp, nb_other_real_tmp, opt_sum_tmp, box_info, cm_tmp, &
                     this % version_number )
            end if

            !write(*,*) 'io_err first ', io_err
            ! read the data
            do while ( io_err == 0 )
                if ( this % type_record == complexe_enum ) then
                    ! read_string_complexe, need to create a new one-complexe for each record data
                    call read_string_complexe ( this % pcomplexe, this, rebuild_sum_square, io_err )
                else
                    ! trajectory, send to trajectory, no use of rebuilt_sum_square here
                    call read_string_trajectory ( this % ptrajectory, this, io_err )
                end if
                !write(*,*) 'io_err ', io_err

            end do

        end subroutine read_file_record

        !> Function to count the number of lines.\n
        !! Tricky for binary : need to read data without saving, read in temporary variables\n
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param nb_line : return the number of lines
        subroutine count_records ( this, nb_line )
            IMPLICIT NONE
            type ( record ) :: this
            integer, intent ( out ) :: nb_line

            character*128 string
            type ( one_complexe ), pointer :: ptmp_onecomplexe
            integer :: io_err

            ! write(*,*) 'Entry count_records ', this % box_info
            nb_line = 0
            string=''
            allocate ( ptmp_onecomplexe )

            ! test, works for trajectorytodcd and read_record, very strange !!
            call new_onecomplexe ( ptmp_onecomplexe, 0, 0, 0, .false. ) !this % box_info )
            !write(*,*) 'ok nb_int nb_other ', this % nb_int, this % nb_real, this % nb_other_real

            if ( this % format_ascii .eqv. .false. ) then
                call read_binary_string ( ptmp_onecomplexe, this % io_compl, .false., io_err )!this % box_info, io_err )
                !write(*,*) 'io_err ', io_err
                do while ( io_err == 0 )
                    nb_line = nb_line + 1
                    call read_binary_string ( ptmp_onecomplexe, this % io_compl, .false., io_err )
                    !write(*,*) 'io_err ', io_err
                end do
            ! read ascii
            else
                read ( this % io_compl, fmt = '(a)', iostat=io_err ) string
                !write(*,*) 'string countline', string
                if ( io_err /= 0 ) then
                    write(*,*) 'No complexes present in the file '
                end if

                do while ( io_err == 0 )
                    nb_line = nb_line + 1
                    read ( this % io_compl, fmt = '(a)', iostat=io_err ) string
                end do
            end if

            !write(*,*) 'Count line ', nb_line
            call delete_onecomplexe ( ptmp_onecomplexe )
            deallocate ( ptmp_onecomplexe )
            !! here only count the line, rewind for a new reading of data
            rewind ( this % io_compl )
            !close ( this % io_compl )

        end subroutine count_records

        !> Test if a file is in binary or ascii format.\n
        !! Called internally\n
        !! test version not used yet
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param filename : name of the file
        !! @param version : not used yet
        subroutine test_binary ( this, filename, version )
            IMPLICIT NONE
            type ( record ) :: this
            character*10 version

            character *1 one_character
            character *64 filename
            integer :: io_err

            !write(*,*) 'Entry test binary '
            !! try to open as unformatted and read the first character
            open ( this % io_compl, file=filename, form='unformatted', action='read', iostat = io_err )
            if ( io_err /= 0 ) then
                 write(*,*) 'problem in opening file as binary', filename
            end if

            !write (*,*) 'read first character '
            read ( this % io_compl, iostat = io_err ) one_character
            !if ( io_err /= 0 ) then
            !    write(*,*) 'one ', one_character
            !    write(*,*) 'problem in reading one character ', one_character
            !end if

            !write(*,*) 'one_char ', one_character
            if ( ( io_err == 0 ) .and. ( one_character == 'v' ) ) then
                this % format_ascii = .false.
                !write ( *, * ) 'Binary file detected '

                !! read the version
                rewind ( this % io_compl )
                read ( this % io_compl, iostat = io_err ) version
                if ( io_err /=0 ) then
                    !write(*,*) 'version:',version,'!!!'
                    !write(*,*) 'Problem to read version in binary'
                    stop 1
                !else
                    !write(*,*) 'version:',version,'!!!'
                end if

                rewind ( this % io_compl )
                !! no need to test ascii
                return
            end if
            !! close before opening in formatted mode
            close ( this % io_compl )

            !! try to open ascii
            open ( this % io_compl, file=filename, form='formatted', action='read', iostat = io_err )
            if ( io_err /= 0 ) then
                 write(*,*) 'problem in opening as ascii file ', filename
            end if
            read ( this % io_compl, fmt = '(a)', iostat = io_err ) one_character
            if ( io_err /= 0 ) then
                 write(*,*) 'two ', one_character
                 write(*,*) 'problem in reading one character ', one_character
            end if
            if ( ( io_err == 0 ) .and. ( one_character == '#' ) ) then
                this % format_ascii = .true.
                !write ( *, * ) 'Ascii file detected '

                !! Read the version
                rewind ( this % io_compl )
                !read ( this % io_compl, fmt = '(''##'',a10)', iostat = io_err ) version

                !! read descriptor line
                read ( this % io_compl, fmt='(a)' ) this % description_column_str
                read ( this % io_compl, fmt = '(2x,a10)', iostat = io_err ) version
                if ( io_err /=0 ) then
                    write(*,*) 'Problem to read version in ascii'
                    stop 1
                !else
                    !write(*,*) 'version:',version,'!!!'
                end if
            ! cannot read properly in ascii => it is binary
            else
                write( *, * ) 'Problem in test_binary '
                write(*,*) 'one_char ', one_character
                stop
            end if

            rewind ( this % io_compl )

        end subroutine test_binary


!!!!!!!!!!!!!!!!!!!!!!!!!!!!  Class trajectory    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        !> Allocate a trajectory
        !!
        !! Allocate the array with instances of empty \ref mod_onecomplexe::one_complexe "one_complexe"\n
        !! Need to give nb_max_complexe for allocating the array
        !! @param this : an instance of "Trajectory"
        !! @param nb_max_complexe : size of the array to allocate
        subroutine new_trajectory ( this, nb_max_complexe )
            IMPLICIT NONE
            type ( Trajectory ) :: this
            integer :: nb_max_complexe

            this % size_array = nb_max_complexe
            ! first position
            this % current_pos = 0

        end subroutine new_trajectory

        !> Finalize the creation of a trajectory
        !!
        !! Called by \ref mod_record::initialize_record "initialize_record"\n
        !! To call after initialisation of \ref mod_record::record::bit_energy "bit energy" and \ref mod_record::record::bit_integer "bit_integer" of type record\n
        !! by default use nb_max_complexe of basecl, could add optional size ( for large file )
        !! @param this : an instance of \ref mod_record::trajectory "Trajectory"
        !! @param basecl : an instance of \ref mod_record::record "record"
        !! @param opt_large_traj : optional, boolean: if the trajectory is large, it will be read one by one
        subroutine initialize_trajectory ( this, basecl, opt_large_traj )
            IMPLICIT NONE
            type ( Trajectory ) :: this
            type ( record ) :: basecl
            logical, optional, intent (in) :: opt_large_traj

            integer :: i, ierr
            logical :: large_traj
            
            large_traj = .false.
            if( present( opt_large_traj ) ) then
                if ( opt_large_traj ) then
                    large_traj = .true.
                end if
            end if

            !write(*,*) 'entry initilize trajectory box_info=', basecl % box_info
            !write(*,*) 'basecl % nb_int, basecl % nb_real, basecl % nb_other_real', &
            !    basecl % nb_int, basecl % nb_real, basecl % nb_other_real
            !write(*,*) 'basecl % nb_types ', basecl % nb_types

            this % size_array = basecl % nb_max_complexe
            !write(*,*) 'this size_array ', this % size_array

            if ( large_traj ) then
                allocate ( this % array_trajectory ( 1 ), stat = ierr )
                if ( ierr /= 0 ) then
                    write ( *, * ) 'Cannot allocate array_trajectory'
                end if

                call new_onecomplexe ( this % array_trajectory ( 1 ), &
                                       basecl % nb_int, basecl % nb_real, basecl % nb_other_real, basecl % box_info ) !, basecl % nb_types )
            else
                allocate ( this % array_trajectory ( this % size_array ), stat = ierr )
                if ( ierr /= 0 ) then
                    write ( *, * ) 'Cannot allocate array_trajectory'
                end if

                !! initialize only one all complexe, no need to recreate
                do i = 1, this % size_array
                    !write(*,*) 'new one compl size ', basecl % nb_int, basecl % nb_real, basecl % nb_other_real
                    ! here multiply by nb_types in entry if 1 fine, here nb_real should be correct
                    call new_onecomplexe ( this % array_trajectory ( i ), &
                                            basecl % nb_int, basecl % nb_real, basecl % nb_other_real, &
                                            basecl % box_info )!, basecl % nb_types )
                end do
            end if
            ! keep same logic, should be 1
            !this % current_pos = 0
            this % current_pos = 1
            !basecl % p_onecompl => this % array_trajectory ( 1 )


        end subroutine initialize_trajectory

        !> Delete trajectory
        !!
        !! For some tools we want to desable automatic writing of array when finishing the program
        !! @param this : an instance of \ref mod_record::trajectory "Trajectory"
        !! @param basecl : an instance of \ref mod_record::record "record"
        !! @param donot_check_array : bool if true do not write the last in the array into the disk
        subroutine delete_trajectory ( this, basecl, donot_check_array )
            IMPLICIT NONE
            type ( trajectory ) :: this
            type ( record ) :: basecl
            logical :: donot_check_array

            integer :: i, ierr

            !write(*,*) 'mod_record::delete trajectory'
            !write(*,*) 'do not check for array ', donot_check_array
            !write(*,*) 'current pos ', this % current_pos

            !! check if something to write, except for tools: read_record
            if ( ( donot_check_array .eqv. .false. ) .and. ( this % current_pos > 1 ) ) then
                !write(*,*) 'write resting traj to file'
                call write_array ( this, basecl )
            end if

            !! delete one_complexe
            do i = 1, this % size_array
                call delete_onecomplexe ( this % array_trajectory ( i ) )
            end do
            !! deallocate array_trajectory
            deallocate ( this % array_trajectory, stat = ierr )
            if ( ierr /= 0 ) then
                write ( *, * ) 'Cannot deallocate trajectory array '
            end if

        end subroutine delete_trajectory

        !> Write trajectory files.\n
        !!
        !! Called when the array is full and for writing the restart file ( last conformation ).
        !! Size array can be of any size for the trajectory, but for restart need to be larger than nb_proteins.
        !!
        !! @param this : an instance of \ref mod_record::trajectory "trajectory"
        !! @param basecl : an instance of \ref mod_record::record "record"
        !! @param opt_last_conf : optionaly write only the last configuration ( for write_restart )
        subroutine write_array ( this, basecl, opt_last_conf )
            IMPLICIT NONE
            type ( trajectory ) :: this
            type ( record ) :: basecl
            logical, optional :: opt_last_conf

            logical :: last_conf
            integer :: first, last

            !write(*,*) 'entry write_array '
            !write(*,*) 'current_pos ', this % current_pos
            !write(*,*) 'number protein ', basecl % nb_protein

            !! default
            last_conf = .false.
            first = 1
            !last = this % current_pos
            ! modify here here or in write_array, last - 1
            ! not valid in all cases
            !last = this % current_pos - 1

            ! works only if full when write_array, if end of array current_pos not updated to + 1
            if ( this % current_pos == this % size_array ) then
                last = this % size_array
            ! else ask to print when array is not filled, current_pos updated to the next
            else
                last = this % current_pos - 1
            end if

            if ( present ( opt_last_conf ) ) then
                last_conf = opt_last_conf
            end if

            !write(*,*) 'opt_last_conf ', last_conf

            !! if something to write
            !! code modified in reset_position and add_record, current_pos ALWAYS > 0 now
            !if ( this % current_pos > 0 ) then

                if ( last_conf .eqv. .true. ) then
                    !write(*,*) 'Only last conf'
                    ! original code
                    !last = this % current_pos
                    !first = last - basecl % nb_protein + 1
                    ! new code, possibility than current_pos already returned to 1
                    if ( this % current_pos == 1 ) then
                        !write(*,*) 'current_pos is 1 '
                        last = this % size_array
                        first = last - basecl % nb_protein + 1
                        !first = 1
                    else
                        !write(*,*) 'current pos > 1' ! e.g. 1101 ok works,
                        ! but if random get first < 0, cannot write the restart. trajectory is fine works in all cases
                        ! may have a test if current_pos
                        last = this % current_pos - 1
                        first = last - basecl % nb_protein + 1

                        ! for tests only
                        if ( first < 1 ) then
                            write(*,*) 'error first < 1, should not happen ??'
                            !stop 1
                            return
                        end if
                    endif

                    !write(*,*) 'io_compl ', basecl % io_compl
                end if
            !! should not happen anymore, never at 0
            !! case where last conf has been printed in trajectory, but still need last for restart
            !else if ( ( this % current_pos == 0 ) .and. ( last_conf .eqv. .true. ) ) then
                !write(*,*) 'case of restart and traj is empty'
            !    last = this % size_array
            !    first = last - basecl % nb_protein + 1
            !else
            !    write(*,*) 'should not happen, nothing to write return'
            !    return
            !end if

            !write (*,*) 'first, last ', first, last
            call write_part_array ( this, basecl, first, last )
            
        end subroutine write_array

        !> this is the real writing to disk part
        !! other
        subroutine write_part_array ( this, basecl, first, last )
            IMPLICIT NONE
            type ( trajectory ) :: this
            type ( record ) :: basecl
            integer, intent( in ) :: first, last

            character*512 :: string
            integer :: i, io_err

            !write(*,*) 'entry write_part_array '
            !write(*,*) 'first, last ', first, last

            do i = first, last

                if ( basecl % format_ascii ) then
                    !call make_ascii_string( this % array_trajectory ( i ), basecl % bit_energy, &
                    !                        string, basecl % opt_sum )
                    !call make_ascii_string( this % array_trajectory ( i ), basecl % bit_energy, &
                    !                        string, basecl % format_string_ascii )
                    call make_ascii_string( this % array_trajectory ( i ), string, basecl % format_string_ascii )
                    write ( basecl % io_compl, fmt='(a)' ) trim(string)
                !! else write binary
                else
                    call write_binary_string ( this % array_trajectory ( i ), basecl % io_compl, &
                                               io_err )
                end if

            end do

            !! assure than the file is written to disk, works on gigantes, not sure on cluster
            flush( basecl % io_compl )

        end subroutine write_part_array

        ! ok for ascii or binary
        subroutine read_string_trajectory ( this, basecl, io_err )
            IMPLICIT NONE
            type ( trajectory ) :: this
            type ( record ) :: basecl
            integer :: io_err

            character*512 string
            !write(*,*) 'current_pos, max ', this % current_pos, this % size_array

            if ( basecl % format_ascii ) then
                !! read ascii line
                read ( basecl % io_compl , fmt = '(a)', iostat = io_err ) string
                !write(*,*) 'string', string
                if ( io_err == 0 ) &
                    !call read_ascii_string ( this % array_trajectory ( this % current_pos + 1 ), &
                    !                         string, basecl % format_string_ascii )
                    call read_ascii_string ( this % array_trajectory ( this % current_pos ), &
                                             string, basecl % format_string_ascii )
                    !write(*,*) 'read ascii string ok:',string
                    !write(*,*) 'data ', this % array_trajectory ( this % current_pos + 1 ) % position
                    !write(*,*) 'data real8 ', real ( this % array_trajectory ( this % current_pos + 1 ) % position, 8 )
            else
                !! read the binary line
                !call read_binary_string ( this % array_trajectory ( this % current_pos + 1 ), &
                !                          basecl % io_compl, basecl % box_info, io_err )
                call read_binary_string ( this % array_trajectory ( this % current_pos ), &
                                          basecl % io_compl, basecl % box_info, io_err )
            end if
            !write(*,*) 'update current pos before ', this % current_pos
            if ( io_err == 0 ) then
                this % current_pos =  this % current_pos + 1
            end if

        end subroutine read_string_trajectory

        ! only for tests
        subroutine print_array ( this, basecl )
            IMPLICIT NONE
            type ( trajectory ) :: this
            type ( record ) :: basecl

            character*512 :: string
            integer :: i

            !write(*,*) 'current_pos ', this % current_pos
            if ( this % current_pos > 0 ) then
                do i = 1, this % current_pos
                    !call make_ascii_string( this % array_trajectory ( i ), basecl % bit_energy, &
                    !          string, basecl % opt_sum )
                    !call make_ascii_string( this % array_trajectory ( i ), basecl % bit_energy, &
                    !          string, basecl % format_string_ascii )
                    call make_ascii_string( this % array_trajectory ( i ), string, basecl % format_string_ascii )
                    write ( *, * ) trim(string)
                end do
            end if

           end subroutine print_array



!!!!!!!!!!!!!!!!!!!!!!!!!!   Class Complexe  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        !> Allocate the \ref mod_record::listcomplexe "ListComplexe"
        !!
        !! Create the nodes necessary for the ListComplexe management
        !! @param this : an instance of "ListComplexe"
        !! @param rmsd_min : parameter for the recording of complexes
        !! @param one_run : parameter for the recording of complexes
        !! @param one_conf : parameter for the recording of complexes
        subroutine new_listcomplexe ( this , rmsd_min, one_run, one_conf )
            IMPLICIT NONE
            type ( ListComplexe ) :: this
            real ( kind=8 ) :: rmsd_min
            !integer :: nb_max_complexe
            logical :: one_run, one_conf

            this % rmsd_min = rmsd_min
            !this % nb_max_complexe = nb_max_complexe
            this % max_energy = +1000.0
            this % nb_complexe = 0
            this % one_run = one_run
            this % one_conf = one_conf

            !! create first and last node, first point to the last
            !! no data associated to theses nodes
            allocate ( this % p_first )
            this % p_first % state = first
            nullify ( this % p_first % p_data )
            allocate ( this % p_last  )
            this % p_last % state = last
            nullify ( this % p_last % p_data )

            this % p_first % p_next => this % p_last
            this % p_last % p_previous => this % p_first
            !! init p_current to p_first, no data here !
            this % p_current => this % p_first !! no need of previous, next, just memo
            !! to be clean
            nullify ( this % p_first % p_previous )
            nullify ( this % p_last  % p_next )

        end subroutine new_listcomplexe

        !> Delete the ListComplex
        !!
        !! @param this : an instance of \ref mod_record::listcomplexe "ListComplexe"
        !! @param opt_clean : optionaly reset member data nb_complexe ?? to check
        subroutine delete_listcomplexe ( this, opt_clean )
            IMPLICIT NONE
            type ( ListComplexe ) :: this
            logical, optional :: opt_clean

            integer :: status
            logical :: head_reached

            !write(*,*) 'entry delete_listcomplexe'

            head_reached = .false.

            if (  associated ( this % p_last % p_previous, this % p_first ) ) &
                  head_reached = .true.

            do while ( head_reached .eqv. .false. )
                call delete_last_data ( this, status )
                !! check status

                !! check if still data to delete
                if ( associated ( this % p_last % p_previous, this % p_first ) ) &
                     head_reached = .true.
                !write(*,*) ' nb_complexe ', this % nb_complexe
            end do

            !! if opt_clean is true, exit before deleting the first and last node
            if ( present ( opt_clean ) ) then 
                if ( opt_clean .eqv. .true. ) then
                  !write(*,*) 'call delete list with option clean'
                  this % nb_complexe = 0
                  return
                end if
            end if

            deallocate ( this % p_first, this % p_last )
            nullify ( this % p_first, this % p_last, this % p_current )

        end subroutine delete_listcomplexe

        !> Merge 2 records into 1 unique\n
        !! Implemented only for "ListComplexe" ( DO NOT USE with trajectory ! )
        !!
        !! Respect the energetic order when doing the merging\n
        !! The total size will not exceed the size of this % nb_max_complexe, the other data will be deleted\n
        !! Used with OpenMP for merging the complexes of different threads
        !! @param this : an instance of \ref mod_record::record "record"
        !! @param p_copy : pointer to the record to be merged into this
        subroutine merge_complexe ( this, p_copy )
            IMPLICIT NONE
            type ( record ) :: this
            type ( record ) :: p_copy

            type ( NodeComplexe ), pointer :: tmp_node_complexe
            type ( one_complexe ), pointer :: tmp_onecomplexe
            !!real ( kind=4 ), dimension ( 3 ) :: position4, orient4_x, orient4_y, orient4_z

            !! other real extendedto contain sum_energy_square, will no be used for any printing
            real ( kind=8 ), dimension ( 4 ) :: other_real
            real ( kind=8 ), dimension( 20 ) :: array_energy
            integer ( kind=4 ), dimension ( 4 ) :: array_integer

            integer :: i

            nullify ( tmp_node_complexe, tmp_onecomplexe )

            !write(*,*) 'Entry merge complexe nb_complexe ', p_copy % pcomplexe % nb_complexe
            !!write(*,*) 'copy nb_complexe', p_copy % pcomplexe % nb_complexe


            tmp_node_complexe => p_copy % pcomplexe % p_first

LOOP_COMP:  do i = 1, p_copy % pcomplexe % nb_complexe

                tmp_node_complexe => tmp_node_complexe % p_next
                tmp_onecomplexe => tmp_node_complexe % p_data

                !call print_onecomplexe ( tmp_onecomplexe )
                call return_compatible_array ( tmp_onecomplexe, array_integer, array_energy, other_real, this % nb_other_real,&
                                               this % bit_integer, this % bit_energy, this % opt_sum, this % nb_types )

                !write(*,*) 'array integer ', array_integer
                !write(*,*) 'array_energy', array_energy
                !write(*,*) 'array_other real ', other_real
                !call extend_1Darray_float ( 4 )
                other_real ( 4 ) = tmp_onecomplexe % sum_square_energy
                !write(*,*) 'extended array_other real ', other_real

                !! anyway need real update of other real, one defualt one with occurency average...provided
                !! add optional other_real
                call fill_listcomplexe ( this % pcomplexe, this, tmp_onecomplexe % position, &
                                         tmp_onecomplexe % orient_x, tmp_onecomplexe % orient_y, &
                                         array_energy, array_integer, &
                                         this % ptab_prot % array ( 1 ) % p_sogrid % distance_max, other_real )

            end do LOOP_COMP

            !! should clean the copy_complexe, otherwise introduce 2 times the same complexe !
            !! option clean is true, just delete the complexes, not the first and last node
            !! can be reused later
            call delete_listcomplexe ( p_copy % pcomplexe, .true. )

        end subroutine merge_complexe


        !> Delete the last data of the list
        !!
        !! Used when the size of complexes exceeds the maximum and for deleting all the list
        !! @param this : an instance of \ref mod_record::listcomplexe "ListComplexe"
        !! @param status : return 0 if the data have correctly deleted
        subroutine delete_last_data ( this, status )
            IMPLICIT NONE
            type ( ListComplexe ) :: this
            integer, intent ( out ) :: status

            type ( NodeComplexe ), pointer :: tmp_node
            integer :: ierr

            !! error by default
            status = 1

            !write(*,*) 'entry delete_last_data'

            tmp_node => this % p_last % p_previous % p_previous

            !! delete the onecomplexe and deallocate the node
            call delete_onecomplexe ( this % p_last % p_previous % p_data )
            deallocate ( this % p_last % p_previous % p_data )
            !nullify ( this % p_last % p_previous % p_data, this % p_last % p_previous % p_previous, &
            !          this % p_last % p_previous % p_next )
            deallocate ( this % p_last % p_previous, stat = ierr )
            if ( ierr /=0 ) then
                write(*,*) 'Cannot deallocate the node in delete_last_data'
            end if
            status = 0

            !!assign correct pointer
            tmp_node % p_next => this % p_last
            this % p_last % p_previous => tmp_node

            !! update nb_complexe
            this % nb_complexe = this % nb_complexe - 1

            !write(*,*) 'deleted last data nb_complexe ', this % nb_complexe

        end subroutine delete_last_data

        ! avoid the use of recursive function, merge fill_listcomplexe and listoverloop
        ! add option other_real, if a complexe is send for merging for instance and other tools ( merge complexe )
        ! but need to transform before complexe to array, only for merging
        ! opt_other_real or ( 0. 0. 0. ) by default, may contain additional sum_square_energy

        !> Create and insert a new \ref mod_onecomplexe::one_complexe "one_complexe" into the list\n
        !! Normally called by \ref mod_record::add_record "add_record", but can be filled manually
        !!
        !! The instance one_complexe is inserted according to its total energy\n
        !! A check is performed if a previous one is closer than rmsd_min. See doc for the complete story
        !! @todo Check the logic, tested but could make simpler ( maybe faster )\n
        !!
        !! @param this : an instance of \ref mod_record::listcomplexe "ListComplexe"
        !! @param basecl : an instance of \ref mod_record::record "record"
        !! @param position : position of the solute
        !! @param orient_x, orient_y : orientation of the axis
        !! @param array_energy : store energy terms in the expected order
        !! @param array_integer : correspond to ntraj, nstep,...
        !! @param dist_max : maximal radius of the protein 2, \ref mod_setofgrid::sogrid::distance_max "sogrid % distance_max"
        !! @param opt_other_real : optionaly the array \ref mod_onecomplexe::one_complexe::other_real_ocompl "other_real_ocompl", default (0.,0.,0.)
        subroutine fill_listcomplexe ( this, basecl, position, orient_x, orient_y, &
                                       array_energy, array_integer, dist_max, opt_other_real )
            IMPLICIT NONE
            type ( ListComplexe ) :: this
            type ( record ) :: basecl
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: position, orient_x, orient_y
            real ( kind=8 ), dimension ( : ), intent ( in ) :: array_energy
            integer, dimension ( 4 ), intent ( in ) :: array_integer
            real ( kind=8 ), intent ( in ) :: dist_max
            real ( kind=8 ), dimension ( : ), optional :: opt_other_real

            type ( one_complexe ), pointer :: pnew_onecomplexe
            type ( NodeComplexe ), pointer :: p_node_energy, p_node_minrmsd, p_node_tmp
            real ( kind=8 ), dimension ( 3 ) :: other_real
            real ( kind=8 ) :: rmsd2_min_input, min_rmsd2, rm2
            integer :: ierr, status, counter
            integer :: nb_other_real
            ! temporary box, not used with complexe, but still added into set_onecomplexe
            integer, dimension ( 3 ) :: tmp_box

            other_real = 0.
            tmp_box = 0

            !! nullify the pointers
            nullify ( p_node_energy, p_node_minrmsd, p_node_tmp )
            nullify ( pnew_onecomplexe )

            status = 0
            counter = 0
            ! if complexes need additional info
            nb_other_real = 3

            ! init large value
            min_rmsd2 = 1000.0
            ! input data rmsd_min
            rmsd2_min_input =  this % rmsd_min * this % rmsd_min

            p_node_tmp => this % p_first % p_next
            !write(*,*) 'init p_node_tmp state ', p_node_tmp % state
            !write(*,*) 'energy to insert', array_energy( 1 )
            !write(*,*) 'nb_complexe ', this % nb_complexe

            !!Main loop over the complexe
 OVERNODES: do while ( p_node_tmp % state /= last )
                !write(*,*) 'new counter ', counter
                !!check for energy if pointer not yet associated
                if ( associated ( p_node_energy ) .eqv. .false. ) then
                    !write(*,*) ' node energy not associated , compare ', array_energy( 1 )
                    ! if total energy is lower assign the pointer, done only once
                    if ( array_energy ( 1 ) <= p_node_tmp % p_data % real_ocompl ( 1 ) ) then
                        !write(*,*) 'new energy lower, associate here'
                        p_node_energy => p_node_tmp
                    end if
               end if

               ! check if rmsd computation is not needed
               ! add, but not fully tested, if rmsd_min=0 do not calculate
               ! easier a real if, else
               if (  ( rmsd2_min_input == 0. ) .or. &
                     (  ( this % one_run) .and. &
                     ( array_integer ( 1 ) /= &
                       p_node_tmp % p_data % int_ocompl ( basecl % indice_ntraj ) ) ) .or. &
                     ( ( this % one_conf) .and. &
                     ( array_integer ( 4 ) /= &
                       p_node_tmp % p_data % int_ocompl ( basecl % indice_nconf ) ) ) ) then

                    ! no need to continue the loop
                    ! can avoid the copy and the cycle, but need to invert the previous condition
                    p_node_tmp => p_node_tmp % p_next
                    counter = counter + 1

                    CYCLE OVERNODES

               end if
               ! else default, only one update of the pointer but need to invert the previous condition
               call rms_complexe ( rm2, dist_max, &
                        position, orient_x, orient_y, &
                        p_node_tmp % p_data % position, &
                        p_node_tmp % p_data % orient_x, &
                        p_node_tmp % p_data % orient_y )

                    ! check if assign min rmsd
                    ! could be simple, but should be equal ??
                    if ( rm2 <= min_rmsd2 ) then
                          min_rmsd2 = rm2
                          ! associate this node to p min rmsd
                          if ( rm2 <= rmsd2_min_input ) then
                            !write(*,*) 'associate p_node_rmsmin rm2 min_rmsd2', rm2, min_rmsd2
                            p_node_minrmsd => p_node_tmp

                          end if
                    end if

                ! update pointer to the next, seems to work
                p_node_tmp => p_node_tmp % p_next
                ! counter not used, to delete
                counter = counter + 1

            end do OVERNODES


            ! Case for only updating data, p_node_minrmsd is allocated
            ! node energy not associated but min rmsd found=> should increment and return
            if ( associated ( p_node_minrmsd ) .eqv. .true. ) then

                ! no minimum energy OR an energy minimum but higher than min_rmsd => update min_rmsd
                ! should work as well, important for update energy of min_rmsd is smaller than array_energy
                ! test at some points from Julia, can delete
                !if  ( array_energy ( 1 ) .ge. p_node_minrmsd % p_data % real_ocompl ( 1 ) )  then
		!original version
                if  ( array_energy ( 1 ) > p_node_minrmsd % p_data % real_ocompl ( 1 ) )  then

                    ! need statistic with complexes, average, std deviation in one complexe
                    if ( present ( opt_other_real ) ) then
                        !write(*,*) 'opt_other_real should implement update in this case'
                        ! in this p_new_onecomplexe is null at this point
                        call update_statistics ( p_node_minrmsd % p_data, pnew_onecomplexe, opt_other_real )

                    ! here need option, increment in update_statics
                    else
                        call update_statistics ( p_node_minrmsd % p_data, pnew_onecomplexe, array_energy ( 1:1 ) )

                    ! let this if needed later, should be in update statistic
                    ! p_node_minrmsd % p_data % other_real_ocompl ( 1 ) = p_node_minrmsd % p_data % other_real_ocompl ( 1 ) + 1
                    ! update average energy, can be done directly. occurence is minimum 2 at this point
                    ! p_node_minrmsd % p_data % other_real_ocompl ( 2 ) = ( p_node_minrmsd % p_data % other_real_ocompl ( 2 ) * &
                    !                                    ( p_node_minrmsd % p_data % other_real_ocompl ( 1 ) - 1 )  &
                    !                        + array_energy ( 1 ) ) / p_node_minrmsd % p_data % other_real_ocompl ( 1 )
                    ! only update sum of energy square, dispersion computed only when printing results
                    ! p_node_minrmsd % p_data % other_real_ocompl ( 3 ) =  p_node_minrmsd % p_data % other_real_ocompl ( 3 ) + &
                    !                                                    ( array_energy ( 1 ) )**2

                    !p_node_minrmsd % p_data % sum_square_energy = p_node_minrmsd % p_data % sum_square_energy + &
                    !                                              ( array_energy ( 1 ) )**2
                    end if

                    ! write(*,*) 'increment done with energy ', array_energy ( 1 )
                    ! call print_onecomplexe ( p_node_minrmsd % p_data )
                    return
                end if

            end if

            ! if here should create a new complexe ?? only if node_energy associated

            ! how to avoid this test ?? If max size not reached add automatically the complexe at the end
            ! behavioir could be changed, energy criteria or ...
            if ( ( this % nb_complexe < basecl % nb_max_complexe ) .and. &
                 ( associated ( p_node_energy ) .eqv. .false. ) ) then
                   ! write(*,*) 'node_energy not associated, but max size not reached, assign to p_last '
                   p_node_energy => this % p_last
            end if

            if ( associated ( p_node_energy ) ) then
                ! write(*,*) 'create a new one_complexe'
                allocate ( pnew_onecomplexe, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'Cannot allocate one_complexe'
                end if
                ! call new_onecomplexe ( pnew_onecomplexe, bit_integer, bit_energy, nb_other_real, opt_sum )
                ! default opt_ibox false
                call new_onecomplexe ( pnew_onecomplexe , &
                                       basecl % nb_int, basecl % nb_real, basecl % nb_other_real )

                !write(*,*) 'after record new_onecomplexe basecl % integer ', &
                !            basecl % nb_int, basecl % nb_real, basecl % nb_other_real

                ! can fill the one_complexe
                !write(*,*) 'set one complexe basecl % nb_types ', basecl % nb_types
                call set_onecomplexe ( pnew_onecomplexe, &
                                       position, tmp_box, orient_x, orient_y, &
                                       ! other real will be known and updated only when the insertion is done
                                       array_integer, array_energy, other_real, basecl % nb_other_real, &
                                       basecl % bit_integer, basecl % bit_energy, basecl % opt_sum, &
                                       basecl % nb_types )

                ! call print_onecomplexe ( pnew_onecomplexe )
                ! write(*,*) 'insert one_complexe'

                ! opt_other_real come from a previous complexe, merge_complexe
                if ( present ( opt_other_real ) ) then
                    pnew_onecomplexe % other_real_ocompl ( 1 ) = opt_other_real ( 1 )
                    pnew_onecomplexe % other_real_ocompl ( 2 ) = opt_other_real ( 2 )
                    pnew_onecomplexe % other_real_ocompl ( 3 ) = opt_other_real ( 3 )
                    pnew_onecomplexe % sum_square_energy = opt_other_real ( 4 )
                    !write(*,*) 'copy other_real to new complexe', pnew_onecomplexe % other_real_ocompl ( 1 ), &
                    !    pnew_onecomplexe % other_real_ocompl ( 2 ), pnew_onecomplexe % other_real_ocompl ( 3 ), &
                    !    pnew_onecomplexe % other_real_ocompl ( 4 )

                    !call update_statisttics ( pnew_complexe,
                    !pnew_onecomplexe % sum_square_energy = ( array_energy ( 1 ) )**2

                ! new complexe means default data, sum_square_energy
                else
                    pnew_onecomplexe % other_real_ocompl ( 1 ) = 1.0
                    pnew_onecomplexe % other_real_ocompl ( 2 ) = array_energy ( 1 )
                    ! 3 is energy dispersion == <E>2 - <E>, store only Sum E2
                    ! done when created
                    !pnew_onecomplexe % other_real_ocompl ( 3 ) = 0. !( array_energy ( 1 ) )**2
                    pnew_onecomplexe % sum_square_energy = ( array_energy ( 1 ) )**2
                end if

                ! in all case insert the new_onecomplexe, size has been checked before
                call insert_new_node ( p_node_energy, pnew_onecomplexe )
                this % nb_complexe = this % nb_complexe + 1

                !! test if min_rmsd need to be deleted: case node_minrmsd true and array_energy higher was tested in update
                if ( associated ( p_node_minrmsd ) .eqv. .true. ) then
                    !write(*,*) 'replace existing, need to update statistics and delete p_node_minrmsd '
                    !call print_onecomplexe ( p_node_minrmsd % p_data )
                    call update_statistics ( pnew_onecomplexe, p_node_minrmsd % p_data )

                    ! if here, energy should be lower, update statistic
                    ! pnew_onecomplexe % other_real_ocompl ( 1 ) =   p_node_minrmsd % p_data % other_real_ocompl ( 1 ) + 1
                    ! average energy = ( prev_av * occurence-1 ) + energy / occurence
                    ! recompute sum of energy of p_min_rmsd, add new value
                    ! pnew_onecomplexe % other_real_ocompl ( 2 ) = ( p_node_minrmsd % p_data % other_real_ocompl ( 2 ) * &
                    !                                               p_node_minrmsd % p_data % other_real_ocompl ( 1 )  &
                                                                !! add new data and divide by new occurence
                    !                     + pnew_onecomplexe % real_ocompl ( 1 ) ) / pnew_onecomplexe % other_real_ocompl ( 1 )
                    !  average energy, total square energy ??
                    !  first version, but need to recompute standart deviation when printing
                    ! pnew_onecomplexe % other_real_ocompl ( 3 ) = p_node_minrmsd % p_data % other_real_ocompl ( 3 ) + &
                    !                                          ( pnew_onecomplexe % real_ocompl ( 1 ) )**2

                    ! update sum_square_energy, std deviation computed when printing
                    ! pnew_onecomplexe % sum_square_energy = p_node_minrmsd % p_data % sum_square_energy + &
                    !                                   ( pnew_onecomplexe % real_ocompl ( 1 ) )**2

                    ! write(*,*) 'delete one onecomplexe'
                    call delete_onecomplexe_fromlist ( p_node_minrmsd )
                    deallocate ( p_node_minrmsd, stat = ierr )
                        if ( ierr /= 0 ) then
                            write (*,*) 'Cannot deallocate node_minrmsd'
                        end if
                    this % nb_complexe = this % nb_complexe - 1
                    nullify ( p_node_minrmsd )

                endif ! end if p_node_minrmsd associated
            endif ! end if p_node_energy associated

            ! check if max size is reached
            if ( this % nb_complexe > basecl % nb_max_complexe ) then
                    !write(*,*) 'too many complexe delete the last one '
                    !call print_onecomplexe ( this % p_last % p_previous % p_data )
                    call delete_last_data ( this, status )
            end if

            ! update energy_max, tricky, last one may be deleted
            ! write ( *, * ) ' energy max ', this % p_last % p_previous % p_data % real_ocompl ( 1 )
            this % max_energy = this % p_last % p_previous % p_data % real_ocompl ( 1 )

            !write(*,*) 'exit fill listcomplexe new max_energy', this % max_energy
            !write(*,*) 'exit fill listcomplexe nb_complexe ', this % nb_complexe

        end subroutine fill_listcomplexe

        !> Read a binary or ascii line from a complexe file
        !!
        !! Create a new \ref mod_onecomplexe::one_complexe "one_complexe" and insert in the list
        !! @todo combine with read_string_trajectory if possible, check why io_err !\n
        !! Check if line is not empty, and really read some data\n
        !!
        !! @param this : an instance of "ListComplexe"
        !! @param basecl : an instance of \ref mod_record::record "record"
        !! @param rebuild_sum_square : logicalm if the square of energy needs to be recomputed ( complex as restart )
        !! @param io_err : used as status ??
        subroutine read_string_complexe ( this, basecl, rebuild_sum_square, io_err )
            IMPLICIT NONE
            type ( ListComplexe ) :: this
            type ( record ) :: basecl
            integer :: io_err
            logical :: rebuild_sum_square

            type ( one_complexe ), pointer :: pnew_onecomplexe
            character*512 string
            integer :: ierr

            !write(*,*) 'Entry read_string_complexe create one_complexe and insert it'

            allocate ( pnew_onecomplexe, stat = ierr )
            if ( ierr /= 0 ) then
                 write(*,*) 'Cannot allocate one_complexe'
            end if

            ! make a new onecomplexe
            call new_onecomplexe ( pnew_onecomplexe , &
                                   basecl % nb_int, basecl % nb_real, basecl % nb_other_real )

            !! make the reading of string by onecomplexe, depends on format
            if ( basecl % format_ascii ) then

                read ( basecl % io_compl , fmt='(a)', iostat = io_err ) string
                !write(*,*) 'string ', trim(string)
                !write(*,*) 'io_err ', io_err

                if ( io_err == 0 ) &
                    call read_ascii_string ( pnew_onecomplexe, string, basecl % format_string_ascii )
            else
                call read_binary_string ( pnew_onecomplexe, basecl % io_compl, basecl % box_info, io_err )
            end if

            !! recompute sum_square_energy from the new onecomplexe ( value not saved on disk )
            if ( ( rebuild_sum_square ) .and. ( io_err == 0 ) ) then
                !write(*,*) 'rebuild sum_square'
                !write(*,*) 'init other_real ', pnew_onecomplexe % other_real_ocompl
                !write(*,*) 'sum_square ', pnew_onecomplexe % sum_square_energy

                !! ok ,works for all even if occurence == 1, because std_dev == 0
                pnew_onecomplexe % sum_square_energy = pnew_onecomplexe % other_real_ocompl(1) * &
                   ( (pnew_onecomplexe % other_real_ocompl(3))**2 + (pnew_onecomplexe % other_real_ocompl(2))**2 )
                !write(*,*) 'sum_square_energy ', pnew_onecomplexe % sum_square_energy

            end if


            !! create a new node and insert BEFORE this, here at the end of the list
            if ( io_err == 0 ) then
                call insert_new_node ( this % p_last, pnew_onecomplexe )
                !! otherwise delete the  last created onecomplexe, because string is empty
                !! increment nb_complexe
                this % nb_complexe = this % nb_complexe + 1
            else
                call delete_onecomplexe ( pnew_onecomplexe )
                deallocate ( pnew_onecomplexe )
            end if

        end subroutine read_string_complexe

        !> Allocate and insert a new \ref mod_record::nodecomplexe "NodeComplexe" always BEFORE this one.
        !!
        !! Associate the pointer to the one_complexe to the new created node
        !! @param this : an instance of \ref mod_record::nodecomplexe "NodeComplexe"
        !! @param pnew_onecomplexe : pointer to a rencently created \ref mod_onecomplexe::one_complexe "one_complexe"
        subroutine insert_new_node ( this, pnew_onecomplexe )
            IMPLICIT NONE
            type ( NodeComplexe ), target :: this
            type ( one_complexe ), pointer :: pnew_onecomplexe

            type ( NodeComplexe ), pointer :: pnew_node
!            type ( NodeComplexe ), pointer :: tmp_node
            integer :: ierr

            !write(*,*) 'Entry insert_new_node '
            allocate ( pnew_node, stat = ierr )
            if ( ierr /=0 ) then
                write( *,*) 'Cannot allocate new node in complexe'
            end if

            ! internal node
            pnew_node % state = intern
            ! point to the correct data
            pnew_node % p_data => pnew_onecomplexe

            call insert_node( this, pnew_node )

        end subroutine insert_new_node
        
        !> Insert an existing node before this one
        !!
        !! @param this : an instance of \ref mod_record::nodecomplexe "NodeComplexe"
        !! @param pnew_node : pointer \ref mod_onecomplexe::one_complexe "one_complexe"
        subroutine insert_node( this, pnew_node )
            IMPLICIT NONE
            type ( NodeComplexe ), target :: this
            type ( NodeComplexe ), pointer :: pnew_node, tmp_node
            
            tmp_node => this % p_previous

            pnew_node % p_next => this
            pnew_node % p_previous => tmp_node
            
            tmp_node % p_next => pnew_node
            this % p_previous => pnew_node
        end subroutine insert_node

        ! delete this node, should not be first or last
        ! in fact extract from list, cannot delete itself !! !!

        !> Delete one_complexe from the list
        !!
        !! Delete the content ( one_complexe ) and rearrange the pointers of the list\n
        !! The deallocation of the node itself must be done by an other function ( cannot delete itself )
        !! @param this : an instance of \ref mod_record::nodecomplexe "NodeComplexe"
        subroutine delete_onecomplexe_fromlist ( this )
            IMPLICIT NONE
            type ( NodeComplexe ), target :: this

            !write(*,*) 'state next', this % p_next % state
            !write(*,*) 'state previous', this % p_next % state
            this % p_previous % p_next => this % p_next
            this % p_next % p_previous => this % p_previous
            nullify ( this % p_next, this % p_previous )
            ! delete the onecomplexe attached to this node
            call delete_onecomplexe ( this % p_data )
            ! deallocate the pointer associated to the onecomplexe
            deallocate ( this % p_data )
            nullify ( this % p_data )

        end subroutine delete_onecomplexe_fromlist

        ! Use non recursive function, merge write_listcomplexe and write node_list
        ! otherwise crash if > 10000 lines

        !> Write the entire list of complexes to the disk
        !!
        !! Called by \ref mod_record::write_record "write_record"
        !! Only the sum of the square of the total energy is stored during the computation\n
        !! The dispersion is computed only before writing to disk
        !!
        !! @param this : an instance of mod_record::listcomplexe "ListComplexe"
        !! @param basecl : an instance of \ref mod_record::record "record"
        !! @param recompute_dispersion : bool if true compute the dispersion
        subroutine write_listcomplexe ( this, basecl, recompute_dispersion )
            IMPLICIT NONE
            type ( ListComplexe ) :: this
            type ( record ) :: basecl
            logical, intent ( in ) :: recompute_dispersion

            type ( NodeComplexe ), pointer :: p_tmp_node
            character ( 512 ) :: string
            integer :: io_err !, status
            real ( kind=8 ) :: tmp_positive

            tmp_positive = 0.d0

            p_tmp_node => this % p_first % p_next

            do while ( p_tmp_node % state /= this % p_last % state )

                if ( recompute_dispersion .eqv. .true. ) then
                    !write(*,*) 'recompute dispersion yes'
                    ! Avoid NaN !!
                    if ( p_tmp_node % p_data % other_real_ocompl ( 1 ) == 1 ) then
                       p_tmp_node % p_data % other_real_ocompl ( 3 ) = 0
                       !tmp_dispersion = 0.
                       !write(*,*) '1 sum_square ', p_tmp_node % p_data % sum_square_energy

                    else
                        !write ( *, * ) 'sum average square ', p_tmp_node % p_data % other_real_ocompl ( 3 )
                        !write ( *, * ) 'sum average square ', p_tmp_node % p_data % sum_square_energy
                        !write(*,*) 'before sqrt ', &
                        !         p_tmp_node % p_data % other_real_ocompl ( 3 ) / p_tmp_node % p_data % other_real_ocompl ( 1 ) - &
                        !         ( p_tmp_node % p_data % other_real_ocompl ( 2 ) )**2
                        tmp_positive = p_tmp_node % p_data % sum_square_energy / p_tmp_node % p_data % other_real_ocompl ( 1 ) - &
                                       ( p_tmp_node % p_data % other_real_ocompl ( 2 ) )**2

                        p_tmp_node % p_data % other_real_ocompl ( 3 ) = sqrt ( tmp_positive )
                    end if

                end if

                ! if ascii, creates the string with appropriate format
                if ( basecl % format_ascii ) then
                    !call make_ascii_string( p_tmp_node % p_data , basecl % bit_energy, &
                    !                        string, basecl % opt_sum )
                    !call make_ascii_string( p_tmp_node % p_data , basecl % bit_energy, &
                    !                        string, basecl % format_string_ascii )
                    call make_ascii_string( p_tmp_node % p_data , string, basecl % format_string_ascii )
                    !write(*,*) 'string2:',string,'!!!'

                    ! write the ascii string
                    write ( basecl % io_compl, fmt='(a)' ) trim(string)
                ! if binary
                else
                    call write_binary_string ( p_tmp_node % p_data, basecl % io_compl, io_err )
                end if
                ! increment pointer
                p_tmp_node => p_tmp_node % p_next
            end do

        end subroutine write_listcomplexe

        !> Only for test\n
        !! keep as example of recursive function for dealing with a list
        !!
        !! Function crash if recursivity > 10 000 approx.
        subroutine print_listcomplexe( this, recompute_dispersion)
            IMPLICIT NONE
            type ( ListComplexe ) :: this
            logical :: recompute_dispersion

            integer :: status
            ! skip header if p_first
            status = print_node_list ( this % p_first % p_next, recompute_dispersion )
            ! write(*,*) ' print list ok '
        end subroutine print_listcomplexe

        !> Only for test
        recursive function print_node_list ( this, recompute_dispersion ) result (status)
           IMPLICIT NONE
           type ( NodeComplexe ) :: this
           logical :: recompute_dispersion

           real ( kind=8 ) :: tmp_real
           integer :: status

           ! not used, could be for error or other
           status=0

           if ( this % state == first ) then
                !write(*,*) 'rewind file and write header'
                status = print_node_list ( this % p_next, recompute_dispersion )
                return
           else if ( this % state == last ) then
                !write(*,*) 'do nothing or close the file'
                return
           else
                ! only called for complexes
                ! here modify edisp, easy than updating and do not modify
                ! print_onecomplexe
                ! make as in sda, save only Sum E2
                ! print sqrt ( Sum E2 / occurence  - < E >2  == sqrt ( <E2> - <E>2)
                ! problem modify array, solution, save and recopy after. Try later update
                if ( recompute_dispersion .eqv. .true. ) then
                    tmp_real = this % p_data % other_real_ocompl ( 3 )
                    !write(*,*) 'other 3', this % p_data % other_real_ocompl ( 3 )
                    !write(*,*) 'other 2', this % p_data % other_real_ocompl ( 2 )
                    !write(*,*) 'other 1', this % p_data % other_real_ocompl ( 1 )
                    this % p_data % other_real_ocompl ( 3 ) = &
                    sqrt ( this % p_data % other_real_ocompl ( 3 ) / this % p_data % other_real_ocompl ( 1 ) - &
                             ( this % p_data % other_real_ocompl ( 2 ) )**2 )
                    call print_onecomplexe ( this % p_data )
                    !! reassign the correct data
                    this % p_data % other_real_ocompl ( 3 ) = tmp_real
                else
                    call print_onecomplexe ( this % p_data )
                end if
                !! continue the loop
                status = print_node_list ( this % p_next, recompute_dispersion )
                return
           endif
        end function print_node_list


!!!!!!!!!!!!!!!!!!!!! other function

        !> Compute an approximation of the root mean square difference between 2 complexes.\n
        !! If rmsd2 < "rmsd_min", the complexe with the minimum energy will conserved. The other deleted\n
        !! Only an approximation for rigid body protein, we use only the orientation axis for the computation\n
        !! Typical values should be rmsd_min <= 2 when a large number of binding sites\n
        !! want to be recorded and <= 1 for more accuracy close to a binding site\n
        !! @param rmsd2 : return mean square difference between 2 complexes
        !! @param dist_max :  maximal radius of the protein 2, \ref mod_setofgrid::sogrid::distance_max "sogrid % distance_max"
        !! @param r1 : position of complex 1
        !! @param y11, y12 : orientation of complex 1
        !! @param r2 : position of complex 2
        !! @param y21, y22 : orientation of complex 2
        subroutine rms_complexe ( rmsd2, dist_max, r1,y11,y12,r2,y21,y22 )
            IMPLICIT NONE
            real ( kind=8 ), intent ( out ) :: rmsd2
            real ( kind=8 ), intent ( in ) :: dist_max
            real ( kind=8 ), dimension ( 3 ) :: r1, y11, y12, y13
            real ( kind=8 ), dimension ( 3 ) :: r2, y21, y22, y23
            real ( kind=8 ), dimension ( 3 ) :: yy11, yy12, yy13
            real ( kind=8 ), dimension ( 3 ) :: yy21, yy22, yy23
            real ( kind=8 ), dimension ( 3 ) :: yy1m1, yy1m2, yy1m3
            real ( kind=8 ), dimension ( 3 ) :: yy2m1, yy2m2, yy2m3

            rmsd2 = 0.d0
            call cross( y13, y11, y12 )
            call cross( y23, y21, y22 )

            ! write(*,*) 'dismx ',dismx
            ! write(*,*) 'r1', (r1(m),m=1,3)
            ! write(*,*) 'y11', (y11(m),m=1,3)
            ! write(*,*) 'y12', (y12(m),m=1,3)
            ! write(*,*) 'y13', (y13(m),m=1,3)
            ! write(*,*) 'r2', (r2(m),m=1,3)
            ! write(*,*) 'y21', (y21(m),m=1,3)
            ! write(*,*) 'y13 ', (y13(m),m=1,3 )
            ! write(*,*) 'y23 ', (y23(m),m=1,3)

            yy11 = y11 * dist_max + r1
            yy12 = y12 * dist_max + r1
            yy13 = y13 * dist_max + r1
            yy21 = y21 * dist_max + r2
            yy22 = y22 * dist_max + r2
            yy23 = y23 * dist_max + r2

            yy1m1 = -y11 * dist_max + r1
            yy1m2 = -y12 * dist_max + r1
            yy1m3 = -y13 * dist_max + r1
            yy2m1 = -y21 * dist_max + r2
            yy2m2 = -y22 * dist_max + r2
            yy2m3 = -y23 * dist_max + r2

            rmsd2 = dot_product ( r1 - r2, r1 - r2 ) + &
                   2. * dot_product ( yy11 - yy21, yy11 - yy21 ) + &
                   2. * dot_product ( yy12 - yy22, yy12 - yy22 ) + &
                   2. * dot_product ( yy13 - yy23, yy13 - yy23 ) + &
                   dot_product ( yy1m1 - yy2m1, yy1m1 - yy2m1 ) + &
                   dot_product ( yy1m2 - yy2m2, yy1m2 - yy2m2 ) + &
                   dot_product ( yy1m3 - yy2m3, yy1m3 - yy2m3 )

            rmsd2 = rmsd2 / 7.

        end subroutine rms_complexe

        !with fortran faster, to test again
!        subroutine rms_complexe_original ( rmsd2, idist_max, r1,y11,y12,r2,y21,y22 )
!            IMPLICIT REAL*4 (a-h,t-z)
!            DIMENSION r1(3),y11(3),y12(3),y13(3)
!            DIMENSION r2(3),y21(3),y22(3),y23(3)
!            DIMENSION yy11(3),yy12(3),yy13(3)
!            DIMENSION yy21(3),yy22(3),yy23(3)
!            DIMENSION yy1m1(3),yy1m2(3),yy1m3(3)
!            DIMENSION yy2m1(3),yy2m2(3),yy2m3(3)
!            real*4 rmsd2,idist_max
!
!
!            call cross4 ( y13, y11, y12 )
!            call cross4 ( y23, y21, y22 )
!
!            DO m = 1,3
!                yy11(m) = y11(m)*idist_max + r1(m)
!                yy12(m) = y12(m)*idist_max + r1(m)
!                yy13(m) = y13(m)*idist_max + r1(m)
!                yy21(m) = y21(m)*idist_max + r2(m)
!                yy22(m) = y22(m)*idist_max + r2(m)
!                yy23(m) = y23(m)*idist_max + r2(m)
!                yy1m1(m) = -y11(m)*idist_max + r1(m)
!                yy1m2(m) = -y12(m)*idist_max + r1(m)
!                yy1m3(m) = -y13(m)*idist_max + r1(m)
!                yy2m1(m) = -y21(m)*idist_max + r2(m)
!                yy2m2(m) = -y22(m)*idist_max + r2(m)
!                yy2m3(m) = -y23(m)*idist_max + r2(m)
!            END DO
!            rmsd2 = 0.
!            DO m = 1,3
!                rmsd2 = rmsd2 + (r1(m)-r2(m))**2 + &
!                     2.* (yy11(m)-yy21(m))**2 + &
!                    2.* (yy12(m)-yy22(m))**2 + &
!                    2.* (yy13(m)-yy23(m))**2 + (yy1m1(m)-yy2m1(m))**2 + &
!                    (yy1m2(m)-yy2m2(m))**2 + (yy1m3(m)-yy2m3(m))**2
!            END DO
!            rmsd2 = rmsd2/7.
!
!        end subroutine rms_complexe_original

END MODULE mod_record


