!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_sda_new_common.f90
!! Define general types and global variables

!> Module with general small structures
MODULE com_sda

    !> Define type for timestep, can be variable in sda
    type parameter_timestep
        !> values read in input, rswd to delete
        !! @{
        real ( kind=8 ) :: dt1, swd1, dt2, swd2, rswd
        !> @}
        !> need to be computed only once, slope
        real ( kind=8 ) :: dt2f
        !> true if timestep is variable, sda. false if constant sdamm
        logical :: variable
        !> new added, distance between the max timestep and the exit, to check with cutoff if interaction
        real ( kind=8 ) :: optimize_exit
    end type parameter_timestep

    !> Group probe variables
    type probe_type
        real ( kind=4 ) :: probep, probew, thresold, hexclusion
        integer :: save_excl
        logical :: save_access !, exclusion_dtgrid
    end type probe_type

    !> To group OpenMP option, only used for the merging of complexes
    type type_option_omp
        integer :: merge_complexe, maxsize_thread_complexe
    end type type_option_omp

    !> Analytic parameters.\n
    !! To extend for all analytic interactions, include hydrodynamics input
    !! @{
    type parameter_analytic
        logical :: ianalytic, debyeh
        ! maybe later for crowders..
        !logical :: crowder
        real (kind=4) :: ionic, h_analytic, lvol_rcut
        logical :: hydro
    end type parameter_analytic
    !! @}

    !> Group MetalDesolvation parameters
    type parameter_metaldesolv
        logical :: imetaldesolv
        !real(kind=8) :: a_real
        real(kind=8) :: Zadw
        real(kind=8) :: Eadw
        real(kind=8) :: Radw
        real(kind=8) :: Aadw
    end type parameter_metaldesolv

    !> Group PMF parameters
    type parameter_pmf
        logical :: ipmf
        integer :: mentth
        integer :: mentfi
        integer :: mentom
        integer :: mentx
        integer :: menty
        real(kind=8) :: pstart
        real(kind=8) :: pfinish
        real(kind=8) :: dZ
    end type parameter_pmf

    !> physical constant
    !! temperature,diel constant of solvent, etc
    !! @{
    !real ( kind=8 ) :: temper = 300.
    !real ( kind=8 ) :: epssol = 78.
    !> @}
    !> if R = 1,987 cal/(mol.K), from k = R/ Na find correct value (wikipedia) 3.289.10^(-24) cal/K
    !! here expressed in kcal/mol
    !real ( kind=8 ), parameter :: rcnst = 1.987e-3
    ! real ( kind=8 ), parameter :: rfc = 6.022d+8

    !! added for debye length, copy from sdamm, paolo
    !
    !  Physical constants and other constant variables
    !
    !  epsi         soft-core rescaling value for sphere interactions

    !  rcnst        Gas constant [kcal mol^-1 K^-1 ]
    ! if R = 1,987 cal/(mol.K), from k = R/ Na find correct value (wikipedia) 3.289.10^(-24) cal/K
    ! here expressed in kcal/(mol.K)
    real( kind=8 ), parameter :: rcnst = 1.987e-3
    !  rcnstj       Gas constant [joule mol^-1 K^-1]
    real( kind=8 ), parameter :: rcnstj = 8.314472d0
    !  kcnst        Boltzmann konstant [kg m^2 s^-2 K^-1]
    real( kind=8 ), parameter :: kcnst = 1.3806503e-23
    !  e_q          Elementary charge [A s]
    real( kind=8 ), parameter :: e_q = 1.602176487e-19
    !  temper       Temperature [K]
    real( kind=8 ), parameter :: temper = 300.d0
    !  epssol       Solvent relative permittivity
    real( kind=8 ), parameter :: epssol = 78.5d0
    !  epsi_o       Vacuum permittivity [A^2 s^4 Kg^-1 m^-3]
    real( kind=8 ), parameter :: epsi_o = 8.854187817e-12
    !  navo         Avogadro number [mol^-1]
    real( kind=8 ), parameter :: navo = 6.0221415e23
    !  pi           Pi
    real( kind=8 ), parameter :: pi = 3.14159265358979d0



    !> maximum size for the variable array, not used by sda, but by tool read_record
    !! if here, will be easier to modify, should always be less then 16 (integer(kind=2))
    integer, parameter :: max_nb_integer_bit = 8, max_nb_real_bit = 15

    !> Input / Output, not all used..\n
    !! Used for output, has a name which can be modified
    !! @{
    integer, parameter :: inputf=10, io6=6, iocmx=55, iocmxn=56, io_restart = 57, &
                    iopoti=77, iurxna=52, &
                    io_bootstrp = 200, io_boot_fpt = 201, &
                    io_steps = 300, io_rdf = 400, &
                    iocmxp=88, iopmf=89
    !> @}

    logical :: EXIT_PROG = .false.

!    contains

!    subroutine close_log_file ()
!        IMPLICIT NONE
!
!        !close ( iopoti )
!        !close ( io_access )
!        close ( iurxna )
!
!    end subroutine close_log_file

END MODULE com_sda
