!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_set_parameters.f90
!! Group the setting of cutoff, rhit ( distance for overlap testing), array of energies and dummy_array (used by all simulations)

!> Group functions to compute the cutoff, allocate dummy_array and energy array.\n
!! used by all types of simulations
MODULE mod_set_parameters

contains

! Compute a cutoff distance for saving computation time
!
! todo Check if used, now use array_cut, to delete ?
!
! Apply to sdamm and sda_2protein\n
! If one solute is a surface, we compute 2 cutoffs : one for solute-solute interaction, an other for solute-surface interaction\n
! The second cutoff is only used by sdamm at the moment\n
! param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
! param geom : instance of \ref mod_geometry::geometry "geometry"
! param Cut2 : square of the cutoff distance
! param opt_Cut2_surf : optionaly the square of the surface-solute cutoff distance
!subroutine set_cutoff ( tab_protein, geom, Cut2, opt_Cut2_surf )
!    USE mod_array_protein
!    USE mod_geometry
!    USE com_sda ! only for io6
!    IMPLICIT NONE
!    type ( array_protein_type ), intent ( in ) :: tab_protein
!    type ( geometry ), intent ( in ) :: geom
!    real ( kind=4 ) :: Cut2
!    real ( kind=4 ), optional :: opt_Cut2_surf
!
!    !! to check should be sqrt(3), not 2 in 3d
!    real ( kind=4 ) :: sq_root3, max_dist_atom !,sq_root2
!    real ( kind=4 ) :: min_sizebox, Cut2_surf
!    integer :: i, j, prot_nb
!
!    ! for test
!    real ( kind = 4 ) :: hp
!    real ( kind = 8 ) :: origin_z
!
!#ifdef DEBUG
!    write(*,*) '=== Entry set_cutoff ==='
!    write(*,*) 'geom % surface ', geom % surface
!    write(*,*) 'nb_grids ', tab_protein % nb_grids
!#endif
!    ! square root of 2 and 3
!    !sq_root2 = 1.41421356
!    sq_root3 = 1.732050808
!
!    Cut2 = 0.
!    ! added in case of surface
!    Cut2_surf = 0.
!    max_dist_atom = 0.
!
!    do i = 1, tab_protein % nb_grids
!        prot_nb = tab_protein % set_of_protein ( i, 1 )
!        !write(*,*) 'i ', i, ' prot_nb ', prot_nb
!        !write(*,*) 'prot ',prot_nb,' is a surface' , tab_protein % array ( prot_nb ) % surface
!
!        !! look over all uhbd grid, in fact consider all grids for multiple conf are identical
!        do j = 1, size ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid )
!
!            ! check if grid is associated
!            if ( is_associated ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( j ) ) ) then
!                !!this%rp =  MAX( rx,ry,rz )/ 2.d0
!                !write(*,*) 'grid j', j, 'associated'
!
!                if ( tab_protein % array ( prot_nb ) % surface ) then
!                    !write(*,*) 'loop over uhbd grid prot ',prot_nb,' is a surface'
!
!                    !write(*,*) ' grid ',j,' rp', tab_protein % array ( i ) % p_sogrid % pset_grid ( j ) % rp
!                    !write(*,*) ' grid ',j,' rpsq', tab_protein % array ( i ) % p_sogrid % pset_grid ( j ) % rpsq
!                    !write(*,*) ' origin ', tab_protein % array ( i ) % p_sogrid % pset_grid ( j ) % origin
!                    !write(*,*) ' grid ',j,' size_z ', tab_protein % array ( i ) % p_sogrid % pset_grid ( j ) % size_z
!                    !write(*,*) ' h ', tab_protein % array ( i ) % p_sogrid % pset_grid ( j ) % h
!                    !write(*,*) ' hp ', tab_protein % array ( i ) % p_sogrid % pset_grid ( j ) % hp
!
!                    hp = real ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( j ) % h, 4 )
!                    origin_z = tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( j ) % origin ( 3 )
!
!                    !! not need of shift by hp, size * h == real distance ( but no grid between 0 and hp ! ) to check
!            if ( ( ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( j ) % size_z + origin_z ) * hp ) > Cut2_surf )&
!                   Cut2_surf = real ( ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( j ) % size_z + origin_z ) * hp, 4 )
!                    !write(*,*) 'Cut2_surf ',Cut2_surf
!
!                ! not a surface, to check sq(3) in 3 dimension !!
!                else
!                    if ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( j ) % rp * sq_root3 > Cut2 ) &
!                         Cut2 = real ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( j ) % rp * sq_root3, 4 )
!                endif ! end surface
!
!            end if ! end if associated
!
!        end do !! end do j over grid
!
!
!        ! same needed for lennard-jones, or use tab_protein
!        if ( associated ( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid ) ) then
!
!           if ( associated( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % array_ljgrid ) ) then
!
!            do j = 1, size( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % array_ljgrid )
!                !write(*,*) 'j ', j
!                !write(*,*) 'test associated ', &
!                !    associated( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % array_ljgrid (j) % p_grid )
!
!                if ( tab_protein % array ( prot_nb ) % surface ) then
!                    !write(*,*) 'I am a surface'
!                    hp = real ( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % array_ljgrid ( j ) % h, 4 )
!                    origin_z = tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % array_ljgrid ( j ) % origin ( 3 )
!
!  if ( ( ( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % array_ljgrid ( j ) % size_z + origin_z ) * hp ) > Cut2_surf )&
!             Cut2_surf = real ( ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( j ) % size_z + origin_z ) * hp, 4 )
!
!                !! not a surface, to check sq(3) in 3 dimension !!
!                else
!                    !write(*,*) 'I am not a surface'
!                    if ( tab_protein % array ( prot_nb ) % p_sogrid %  p_ljgrid % array_ljgrid ( j ) % rp * sq_root3 > Cut2 ) &
!                      Cut2 = real ( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % array_ljgrid ( j ) % rp * sq_root3, 4 )
!                endif ! end surface
!
!            end do
!          end if ! if p_grid associated
!
!        end if ! if lj
!
!        !! distance_max is already correct if surface or flex
!        if ( tab_protein % array ( prot_nb ) % iflex ) then
!            if ( tab_protein % array ( prot_nb ) % p_flex % plist_flex % distance_max > max_dist_atom ) then
!                max_dist_atom = real ( tab_protein % array ( prot_nb ) % p_flex % plist_flex % distance_max, 4 )
!            end if
!        else if ( tab_protein % array ( prot_nb ) % p_sogrid % distance_max > max_dist_atom ) then
!            max_dist_atom = real ( tab_protein % array ( prot_nb ) % p_sogrid % distance_max, 4 )
!        end if
!        !write(*,*) 'distance_max prot ',i,' max_dist_atom', max_dist_atom
!
!    end do
!
!    write ( io6, * ) 'max size for the grid ', Cut2, 'A'
!    write ( io6, * ) 'max size for the grid with surface ', Cut2_surf, 'A'
!    write ( io6, * ) 'max distance for proteins ', max_dist_atom, 'A'
!    Cut2 = Cut2 + max_dist_atom
!    Cut2 = Cut2 * Cut2
!
!    ! spend too much time because forgot to setup this value, make a test
!    if ( ( Cut2_surf > 0 ) .and. ( geom % surface .eqv. .false. ) ) then
!        write(*,*)
!        write(*,*) 'ERROR Cut2_surf > 0, flag surface should be in set geometry !'
!        write(*,*) 'ERROR, previous bug due to this missing tests, to check if valid...'
!        stop 1
!    end if
!    !! by default apply on Cut2_surf as well
!    Cut2_surf = Cut2_surf + max_dist_atom
!    Cut2_surf = Cut2_surf * Cut2_surf
!
!    !! Check cutoff is smaller than half the box if pbc or box
!    !! assume it is a box by default, could check explicitely
!    !! add case if surface, no pbc in z direction
!    if ( ( geom % pbc ) .and. ( geom % type == enum_box ) ) then
!
!        ! if no surface look at minimum value
!        if ( geom % surface .eqv. .false. ) then
!            min_sizebox = MINVAL ( geom % pbox % sizebox )
!        ! else take min between x and y  only, no pbc on z
!        else
!            min_sizebox = geom % pbox % sizebox ( 1 )
!            if ( geom % pbox % sizebox ( 2 ) < min_sizebox ) &
!                min_sizebox = geom % pbox % sizebox ( 2 )
!        end if
!        !write ( *, * ) 'min box size', min_sizebox
!
!        ! check only with Cut2, Cut2_surf can be much larger
!        if ( Cut2 > (min_sizebox/2.0)**2  ) then
!            write( io6, * ) 'WARNING One grid (ed,hd,lj.ep) (excluding the surface) is larger ', &
!                            ' than 1/2 of the simulation box size: ', (min_sizebox/2.0), ' Cut ', sqrt(Cut2)
!            write( io6, * ) 'WARNING It is fine ONLY if it is a grid of a surface'
!            !! should be fine if it corresponds to a surface...
!        endif
!    endif
!
!    ! will be present only for sdamm, and if surface, sda_2proteins should not change
!    if ( present ( opt_Cut2_surf ) .and. ( geom % surface .eqv. .true. ) ) then
!        opt_Cut2_surf = Cut2_surf
!
!        !comment for the release, should be fine
!        !but certainly full test needed
!        !if ( opt_Cut2_surf < Cut2 ) then
!        !    write(*,*)
!        !    write(*,*) 'WARNING Cutoff for the surface is smaller than cutoff for protein-protein interaction'
!        !    write(*,*) 'Possible error in force and energy to check !!'
!        !    write(*,*)
!        !end if
!    end if
!
!end subroutine set_cutoff

!> General function to setup the dummy_array and logical variable
!! lennard-jones is missing here
!!
!! Set only shared variable, can be called in serial part.
!! Apply to both sda_2proteins, and sdamm.
!! certainly param_force_energy would be needed later.
!!
!! @param tab_protein : : instance of \ref mod_array_protein
!! @param dummy_array : array of dummy atoms to allocate
!! @param iflex : indicates if additional tests must be done for flexibility
subroutine set_dummy_array_and_logical ( tab_protein, dummy_array, iflex )
    USE mod_array_protein
    implicit none
    type ( array_protein_type ), intent ( in ) :: tab_protein
    real ( kind=8 ), dimension( : ), pointer :: dummy_array
    logical, intent ( inout ) :: iflex

    integer :: size_dummy_array, prot_nb, i, ierr
    ! default to zero, only needed if repulsive lennard-jones AND new lennard-jones
    size_dummy_array = 0

#ifdef DEBUG
    write(*,*) '=== Entry set_dummy_array ==='
#endif

    do i = 1, tab_protein % nb_grids
        ! first protein of the group
        prot_nb = tab_protein % set_of_protein ( i, 1 )
        !write(*,*) 'prot_nb ', prot_nb

        ! test if flexible, why ? here only need to setup iflex and isurf
        if ( tab_protein % array ( prot_nb ) % iflex ) then
            !! if at least one, will check over all protein, no groups defined
            iflex = .true.
       end if

        ! check explicitly for repulsive-lj, need a dummy array
        ! will need as weel if apolar !!
        if ( is_associated ( tab_protein % array ( prot_nb ) % p_sogrid % pset_grid ( get_nb_gridtype ( 'repulsive_lj') ) ) ) then
            !write(*,*) 'wiill allocate later a dummy_array for repulsive lj, size=', size_dummy_array

            ! if flexible, get max number from flexible module
            if ( tab_protein % array ( prot_nb ) % iflex ) then
                !iflex = .true.
                if ( tab_protein % array ( prot_nb ) % p_flex % plist_flex % max_size_movable_atom > size_dummy_array ) then
                   size_dummy_array = tab_protein % array ( prot_nb ) % p_flex % plist_flex % max_size_movable_atom
                end if
            ! not flexible
            else
                !write(*,*) 'repulsive lennard-jones associated and not flexible, look at max atom surface '
                if ( tab_protein % array ( prot_nb ) % nb_atom_surface > size_dummy_array ) then
                    size_dummy_array = tab_protein % array ( prot_nb ) % nb_atom_surface
                end if
            end if
        ! not associated
        !else
        !    write(*,*) 'repulsive lennard-jones is not associated size_dummy_array ', size_dummy_array
        end if

        if ( associated ( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid ) ) then
            !write(*,*) 'lennard jones grid, check size_dummy_array '
            if ( tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % max_pos_atom > size_dummy_array ) then
                size_dummy_array = tab_protein % array ( prot_nb ) % p_sogrid % p_ljgrid % max_pos_atom
            end if
        !else
        !    write(*,*) 'lennard-jones is not associated size_dummy_array ', size_dummy_array
        end if


    end do
    !write(*,*) 'iflex ', iflex
    !write(*,*) 'Size dummy array', size_dummy_array

    ! initialize to the maximum size
    if ( size_dummy_array > 0 ) then
        ! need to check for both proteins
        !!write(*,*) 'Allocate dummy_array, size ', size_dummy_array
        ! or repulsive-lj for sdamm
        allocate ( dummy_array ( size_dummy_array ),  stat = ierr )
        if ( ierr /=0 ) then
            write( *, *) 'Problem in allocating dummy array size, size_dummy_array '
            stop 1
#ifdef DEBUG
        else
            write (*,*) 'Dummy_array allocated, size ', size_dummy_array
#endif
        end if
    else
        ! need to allocate to send to a subroutine, valgrind complains !
#ifdef DEBUG
        write(*,*) 'Allocate dummy array with size of 1'
#endif
        allocate ( dummy_array ( 1 ),  stat = ierr )
        if ( ierr /=0 ) then
            write( *, *) 'Problem in allocating dummy array size size_dummy_array 1 '
            stop 1
        end if
    end if
    ! dummy_array contains only 1.d0 values
    dummy_array = 1.d0

end subroutine set_dummy_array_and_logical

!!> Allocate minimum size for array energy.\n
!! Unique function, ssed by sda_2proteins and sdamm.\n
!!
!! @param type_calc : instance of \ref type_calculation
!! @param o_record : instance pf \ref mod_record
!! @param nb_prot : number of proteins
!! @param array_energy : pointer to allocate the array
!! @param opt_max_energy_terms : optional, if present contains the maximum energy terms ( maximum grid used in \ref mod_grid_type )
!! @param opt_size_array_energy : optional,  if present contains the size of the array
!! @param opt_size_array_energy : optional,  if present force the allocation of a maximum size ( used by unit-tests)
subroutine set_array_energy ( type_calc, o_record, nb_prot, array_energy, &
                              opt_max_energy_terms, opt_size_array_energy, opt_force_large_array )
    use mod_record
    use mod_type_calc
    implicit none
    type ( type_calculation ), intent ( in ) :: type_calc
    type ( record ), intent ( in ) :: o_record
    integer, intent ( in ) :: nb_prot
    real ( kind=8 ), dimension(:,:), pointer :: array_energy
    integer, optional, intent( out ) :: opt_max_energy_terms, opt_size_array_energy
    logical, intent ( in ), optional :: opt_force_large_array

    integer :: ierr, size_array_energy, size_array
    integer :: max_energy_terms

    max_energy_terms = 0
    size_array_energy = 0
    size_array = 0

#ifdef DEBUG
    write(*,*) '=== Entry set_array_energy ==='
#endif
    ! setup energy terms
    ! new setup of energy_terms for output, becomes general to many cases and tests...
    ! here small difference between sda_2proteins and sdamm

    ! if no complex / no trajectory, not need to compute and store energy, could use param_force_energy

    ! check for complexes first,
    !  - if sda trajectory will have same format
    if ( ( associated ( o_record % pcomplexe ) ) .or. ( associated( o_record % ptrajectory ) ) ) then

        max_energy_terms = get_max_bit_energy( o_record )
        size_array_energy = get_nbterms_energy_compatible ( o_record, .true. )
#ifdef DEBUG
        write(*,*) 'max_energy_terms ', max_energy_terms
        write(*,*) 'size_array_energy ', size_array_energy
#endif
        ! complex in all cases sda_2proteins and sdamm, no subtypes but adds total energy
        if ( associated ( o_record % pcomplexe ) .or. ( type_calc % enum_type_calc /= sdamm ) ) then
            size_array = max_energy_terms + 1
        ! else trajectory and sdamm
        else
            size_array = size_array_energy
        end if

        ! force to allocate larger array
        if ( present ( opt_force_large_array ) ) then
            if ( opt_force_large_array .eqv. .true. ) then
                !write(*,*) 'force large array, will use size_array_energy value ', size_array_energy
                size_array = size_array_energy
            end if
        end if
#ifdef DEBUG
        write(*,*) 'Allocate array_energy ', size_array, nb_prot
#endif
        allocate ( array_energy ( size_array, nb_prot ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In the allocation of array_energy size ', size_array_energy
            stop 1
        endif
        array_energy = 0.d0
    else
        write(*,*) 'ERROR Problem with setting array_energy'
        stop 1
    endif

    ! needed for sda_2proteins, max_energy without total energy
    if ( present ( opt_max_energy_terms ) ) then
        opt_max_energy_terms = max_energy_terms !+ 1
    end if

    ! added for sdamm and complexes to create, maybe independent later, first use same terms
    if ( present ( opt_size_array_energy ) ) then
        opt_size_array_energy = size_array_energy
    end if

end subroutine set_array_energy

!! Apply only for sda_2proteins type ( not sdamm )\n
!! if surface, dist_max correctly updated when reading input, no need to modification\n
!! limit to protein1 is a surface, but protein2 can be flexible\n

!> To compute rhit, maximum distance where we need to check for overlap.\n
!!
!! Apply only for sda_2proteins type ( not sdamm )\n
!! if surface, dist_max correctly updated when reading input, no need to modification\n
!! limit to protein1 is a surface, but protein2 can be flexible\n
!! @param tab_prot : instance of \ref mod_array_protein
!! @param param_probe : instance of sda_com::param_probe
!! @param rhit : distance to set for chaecking overlap
subroutine set_rhit ( tab_prot, param_probe, rhit )
    USE mod_array_protein
    USE com_sda
    IMPLICIT NONE
    type ( array_protein_type ) :: tab_prot
    real ( kind=4 ) :: rhit
    type ( probe_type ) :: param_probe

    real ( kind=4 ), dimension ( 2 ) :: dist_max
    integer :: i

#ifdef DEBUG
    write(*,*) 'Entry set_rhit'
#endif
    dist_max = 0.

    ! surface has been correctly taken into account in prepare_atom_protein
    ! \todo check if valid for sdamm, why only 1,2
    do i = 1, 2
        if ( tab_prot % array ( i ) % iflex ) then
            dist_max ( i ) = real ( tab_prot % array ( i ) % p_flex % plist_flex % distance_max, 4 )
        else
            dist_max ( i ) = real ( tab_prot % array ( i ) % p_sogrid % distance_max, 4 )
        endif

#ifdef DEBUG
        write(*,*) 'prot i=',i, dist_max(i)
#endif
        if ( tab_prot % array ( i ) % surface ) then
            write(*,*)
            write(*,*) 'INFO 1 solute is a surface, check the rhit value in the output'
            write(*,*)
        end if
    end do

    !! 1.9 armax before but fixed
    !rhit = protein1 % distance_max + protein2 % distance_max + param_probe % probep + 1.9
    write(*,*)
    write(*,*) 'Initialisation of rhit distance, above this value no check for overlap is done'
    write(*,*) 'rhit = sum of the distance max of each solute + a constant = ', param_probe % probep + 1.9
    !write(*,*) 'dist max1 ', dist_max(1), ' dist max 2 ', dist_max(2), ' rhit ', &
    !            dist_max(1) + dist_max(2)  + param_probe % probep + 1.9
    write(*,*) 'dist max1 ', dist_max(1), ' dist max 2 ', dist_max(2)
    rhit = dist_max ( 1 ) + dist_max ( 2 )  + param_probe % probep + 1.9
    write(*,*) 'Init rhit ', rhit

end subroutine set_rhit

!> Routine to generate an interaction cut off specific to each grid type.\n
!!
!! Using cut offs specific to each grid saves time by removing uneccessary calculations of zero interactions with small grids.\n
!! For each solute type, it checks to see if the solute has multiple conformations.\n
!!   -If it does it calls /ref mod_set_parameters::alloc_fill_cut_array for each conformation.\n
!!   -Else it calls once.\n
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type 
!! @param geom : instance of \ref mod_geometry::geometry
!! @param cut2_max : maximum cut-off of all interactions
subroutine set_cut_array( tab_protein, geom, cut2_max )
    USE mod_array_protein
    USE mod_gridtype
    USE mod_geometry
    USE com_sda ! only for io6
    IMPLICIT NONE
    ! Passed variables
    type ( array_protein_type ), intent ( in ) :: tab_protein
    type ( geometry ), intent ( in ) :: geom
    real ( kind=4 ), intent (out) :: cut2_max

    ! Subroutine variables
    integer :: numsogrid, numinter, id_ljgrid !,numljgrid
    integer :: prot_t1 !, prot_t2
    integer nconf
    real (kind=4) :: min_sizebox, test_cut
    ! Counters
    integer :: i,j,k
    ! Pointer to flexible data
    type (NodeFlex), pointer :: tmp_node
    ! Pointer to sogrid in which to generate cut_array
    type (sogrid), pointer :: p_sogrid

    !Set initial data
    numsogrid = tab_protein%nb_grids !Number of different sets of grids (different solutes)
    numinter = enum_grid_force_calc  !Number of interactions calculated per grid (defined in mod_gridtype)
    id_ljgrid = get_nb_gridtype('lennardjones') ! Function defined in mod_gridtype
    cut2_max = 0.

    nullify(tmp_node,p_sogrid)

    ! Print label for summary
    write (*,'("Cut off summary")')

#ifdef DEBUG
    write(*,*) 'numsogrid ', numsogrid
    write(*,*) 'numinter ', numinter
    write(*,*) 'id_ljgrid ', id_ljgrid
#endif

    ! Loop over members of protein array
    do i=1,numsogrid

       ! Find first protein of next type
       prot_t1=tab_protein%set_of_protein(i,1)
       ! Print label for summary
       write (*,'(" Protein: ",i3)'),i

       ! If flexible, allocate and fill array for all conformations
       if (tab_protein%array(prot_t1)%iflex) then

          ! Number of protein conformations
          nconf = tab_protein%array(prot_t1)%p_flex%plist_flex%total_conf

          ! Point tmp_node to first in flexible list
          tmp_node => tab_protein%array(prot_t1) % p_flex % plist_flex % p_first 
          ! First is empty, point to next for first conformation
          tmp_node => tmp_node % p_next

          ! Counter for loop
          j = 1
          ! Loop over protein conformations. Stop if tmp_node points to last (which is empty)
          do while (.not.associated(tmp_node,tab_protein%array(prot_t1)%p_flex%plist_flex%p_last))

             ! Point to sogrid for current conformation
             p_sogrid => tmp_node % p_data

             ! Print conformation number for output summary
             write (*,'("  Conformation: ",I4)') j
             ! Call subroutine to generate cut_array for current sogrid
             call alloc_fill_cut_array(tab_protein,p_sogrid,prot_t1)

             if ( p_sogrid % max_cut2 > cut2_max ) then
                cut2_max = p_sogrid % max_cut2
             endif
             tmp_node => tmp_node % p_next
             ! Update counter
             j = j + 1 
          end do

       else 
          ! In non-flexible case, point in sogrid within protein
          p_sogrid => tab_protein%array(prot_t1)%p_sogrid
          ! Call subroutine to generate cut_array for current sogrid
          call alloc_fill_cut_array(tab_protein,p_sogrid,prot_t1)

          if ( p_sogrid % max_cut2 > cut2_max ) then
                cut2_max = p_sogrid % max_cut2
          endif

       end if
    end do

    ! From old cut routine: Check no cut off bigger than half smallest box dimension
    ! Modified to loop over all non-surface cut offs
    if ( ( geom % pbc ) .and. ( geom % type == enum_box ) ) then 

        ! if no surface look at minimum value
        if ( geom % surface .eqv. .false. ) then
            min_sizebox = MINVAL ( geom % pbox % sizebox )

        ! else take min between x and y  only, no pbc on z
        else
            min_sizebox = geom % pbox % sizebox ( 1 )
            if ( geom % pbox % sizebox ( 2 ) < min_sizebox ) &
                min_sizebox = geom % pbox % sizebox ( 2 )
        end if

        ! Loop over all protein types
        do i=1,numsogrid

           ! Get number of first protein of type
           prot_t1=tab_protein%set_of_protein(i,1)

           ! If Protein is a surface, skip test
           ! Cut off can be greater than half box dimension in that case
           if (tab_protein%array(prot_t1)%surface) then
              cycle
           end if

           ! Loop over interaction grid types
           do j=1,numinter
              ! Loop over interacting proteins
              do k=1,numsogrid   
                 ! Get value of cut off
                 test_cut = real( tab_protein%array(prot_t1)%p_sogrid%cut2_array(j,k), 4)

                 ! Test if cut off is too large
                 if ( test_cut > (min_sizebox/2.0)**2  ) then
                    write( io6, * ) 'WARNING One grid (ed,hd,lj.ep) (excluding the surface) is larger ', &
                                    ' than 1/2 of the simulation box size: ', (min_sizebox/2.0), ' Cut ', sqrt(test_cut)
                    write( io6, * ) 'WARNING It is fine ONLY if it is a grid of a surface'
                    ! should be fine if it corresponds to a surface...
                 endif
              end do ! End loop over interacting protein
           end do ! End loop over interaction grid types
        end do ! End loop over protein types
    endif ! End if PBC

#ifdef DEBUG
    write(*,*) 'set cut2_max ', cut2_max
#endif

end subroutine set_cut_array

!> Routine to allocate and fill a grid specific cut off array for each solute/conformation
!!
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type 
!! @param cur_sogrid : instance of \ref mod_setofgrid::sogrid
!! @param cur_prot : integer used as index to get protein data from mod_array_protein::array_protein_type::array
subroutine alloc_fill_cut_array(tab_protein,cur_sogrid,cur_prot)

   USE mod_array_protein
   USE mod_gridtype
   USE mod_geometry
   IMPLICIT NONE

   ! Passed in variables
   type(array_protein_type), intent(in) :: tab_protein
   type(sogrid), pointer, intent(in) :: cur_sogrid
   integer , intent(in) :: cur_prot
   ! Subroutine variables
   integer :: numsogrid, numinter
   integer :: gsize_z
   integer :: prot_t2
   integer :: id_ljgrid,numljgrid
   real (kind=4) :: sq_root3, dcut
   real (kind=4) :: hp,max_dist_atom
   real (kind=8) :: origin_z
   ! Counters
   integer :: i,j,k,l
   ! Status flags
   integer :: ierr
 

   ! Set initial variables
   sq_root3 = 1.732050808
   numsogrid = tab_protein%nb_grids !Number of different sets of grids (different solutes)
   numinter = enum_grid_force_calc  !Number of interactions calculated per grid (defined in mod_gridtype)
   id_ljgrid = get_nb_gridtype('lennardjones') ! Function defined in mod_gridtype 

   ! Allocate current cut_array
   allocate (cur_sogrid%cut2_array(numinter,numsogrid), stat=ierr)
   if (ierr.NE.0) then
      write (*,*) "Error allocating cut off array for protein",i
      stop 1
   end if

   ! Initialise cut off array to zero
   cur_sogrid%cut2_array = 0.
   ! Initialise max_cut2 to zero
   cur_sogrid%max_cut2 = 0.
   ! Initialize local variable
   hp = 0.
   origin_z = 0.
   gsize_z = 0

   ! Loop over interaction types
   do i=1,numinter

          !write(*,*) 'i from numinter ', i

          ! MIC: Cannot work with LJ, because is_associated will return false when i=5
          ! Check grid is associated
          if (is_associated(cur_sogrid%pset_grid(i))) then
             ! If surface, set variables relating to z axis
             if (tab_protein%array(cur_prot)%surface) then
                ! Extent of grid in z direction is grid spacing * distance from origin
                hp=real(cur_sogrid%pset_grid(i)%h,4)
                origin_z=cur_sogrid%pset_grid(i)%origin(3)
                gsize_z=cur_sogrid%pset_grid(i)%size_z
             end if

             ! Loop over interacting proteins
             do j=1,numsogrid
                ! Find first protein of interacting type
                prot_t2=tab_protein%set_of_protein(j,1)

                ! Find maximum distance of an atom in interacting protein from its centre
                if (tab_protein%array(prot_t2)%iflex) then
                   max_dist_atom = real(tab_protein%array(prot_t2)%p_flex%plist_flex%distance_max,4)
                else
                   max_dist_atom = real(tab_protein%array(prot_t2)%p_sogrid%distance_max,4)
                end if
                !write(*,*) 'max_dist_atom ', max_dist_atom

                ! Cut off for surfaces
                if (tab_protein%array(cur_prot)%surface) then
                  ! Cut off is (size_z+origin_z)*hp+distance_max
                  cur_sogrid%cut2_array(i,j) = real( ((origin_z+gsize_z)*hp) + max_dist_atom ,4)
                ! Cuf off for non-surfaces
                else
                  ! Cutoff is sqrt3*biggest grid dimension plus maximum atom distance of interacting protein
                  cur_sogrid%cut2_array(i,j) = real(cur_sogrid%pset_grid(i)%rp*sq_root3 + max_dist_atom,4)
                end if
                !write(*,*) 'cut2_array i,j,value ', i,j, cur_sogrid%cut2_array(i,j)

                ! Square cut off
                cur_sogrid%cut2_array(i,j) = &
                   cur_sogrid%cut2_array(i,j) * cur_sogrid%cut2_array(i,j)

                ! If larger than max_cut2, update
                if (cur_sogrid%cut2_array(i,j) .GT. cur_sogrid%max_cut2) then
                    cur_sogrid%max_cut2 = cur_sogrid%cut2_array(i,j)
                end if

             end do ! End loop over interacting proteins, case pset_grid

          ! if is_associated is false, test LJ
          else
            !write(*,*) 'pset_grid not associated, check LJ with i=', i
            ! to be sure the correct i is used
            if (i.EQ.id_ljgrid) then
                !write(*,*) 'i == id_ljgrid, test LJ if associated'
                !write(*,*) 'associated(cur_sogrid%p_ljgrid) ', associated(cur_sogrid%p_ljgrid)
                ! Check if LJ grid array is allocated
                if (associated(cur_sogrid%p_ljgrid)) then
                    !write(*,*) 'test cur_sogrid%p_ljgrid associated'
                    if (associated(cur_sogrid%p_ljgrid%array_ljgrid)) then

                       !write(*,*) 'cur_sogrid%p_ljgrid is associated'
                       ! Find size of LJ array
                       numljgrid=size(cur_sogrid%p_ljgrid%array_ljgrid)
                       !write(*,*) 'numljgrid ', numljgrid

                    ! Loop over interacting proteins
                    do j=1,numsogrid
                      ! Find first protein of interacting type
                      prot_t2=tab_protein%set_of_protein(j,1)

                      ! Find maximum distance of an atom in interacting protein from its centre
                      if (tab_protein%array(prot_t2)%iflex) then
                         max_dist_atom = real(tab_protein%array(prot_t2)%p_flex%plist_flex%distance_max,4)
                      else
                         max_dist_atom = real(tab_protein%array(prot_t2)%p_sogrid%distance_max,4)
                      end if

                      ! Loop through LJ array
                      do l=1,numljgrid
                        if (tab_protein%array(cur_prot)%surface) then
                           ! Get values for cut calculation
                           hp=real(cur_sogrid%p_ljgrid%array_ljgrid(l)%h,4)
                           origin_z=cur_sogrid%p_ljgrid%array_ljgrid(l)%origin(3)
                           gsize_z=cur_sogrid%p_ljgrid%array_ljgrid(l)%size_z

                           ! Calculate required cut off
                           dcut = real(((origin_z+gsize_z)*hp)+max_dist_atom,4)
                        else
                           ! Calculate required cut off
                           dcut = real(cur_sogrid%p_ljgrid%array_ljgrid(j)%rp*sq_root3 + max_dist_atom,4)
                        end if

                        ! Square cut off
                        dcut = dcut*dcut
                        !write(*,*) 'dcut, dcut2', sqrt(dcut), dcut

                        ! Is calculated cut off greater than stored cut off
                        if (dcut.GT.cur_sogrid%cut2_array(i,j)) then
                           cur_sogrid%cut2_array(i,j)=dcut
                        end if

                        ! check for the maximum
                        if (cur_sogrid%cut2_array(i,j) .GT. cur_sogrid%max_cut2) then
                            cur_sogrid%max_cut2 = cur_sogrid%cut2_array(i,j)
                        end if

                      end do ! End loop over LJ array

                    end do ! loop over proteins

                    end if !end array_ljgrid associated
                end if ! end p_ljgrid associated
            end if ! id.EQ. id_ljgrid
          endif ! is_associated(pset_grid) or ljgrid

   ! End loop over grid types
   end do
 
   ! Print summary
   do j=1,numsogrid
      write (*,'(i3)',advance='no'),j
      do k=1,numinter
         write (*,'(f12.3)',advance='no'), cur_sogrid%cut2_array(k,j)
      end do
      write (*,*)
   end do

end subroutine alloc_fill_cut_array

!> Routine to find the Rcut value.\n
!!  -- This is the radius used to define local volume in mean-field hydrodynamic approach.\n
!!  -- At present, it finds the largest Stokes radius of any solute, and multiplies this by a factor.
!!
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type 
!! @param geom : instance of \ref mod_geometry::geometry
!! @param hi_rad : radius Rcut used in hydrodynamic calculations
subroutine set_hi_rad (tab_protein, geom, input_rad, hi_rad)

    USE mod_array_protein
    USE mod_gridtype
    USE mod_geometry
    USE com_sda 
    IMPLICIT NONE
    ! Passed variables
    type ( array_protein_type ), intent ( in ) :: tab_protein
    type ( geometry ), intent ( in ) :: geom
    real ( kind=4 ) , intent (in) :: input_rad
    real ( kind=4 ), intent (out) :: hi_rad
    ! Subroutine variables
    ! Counters
    integer i, j
    ! System constants
    integer numsogrid ! total number of setofgrid objects, ie number of solute types
    integer prot_type ! Index to first solute of current type
    integer nconf ! Number of conformations of current flexible solute
    ! Pointers to objects
    type(sogrid), pointer :: p_sogrid ! pointer to current setofgrid object
    type(NodeFlex), pointer :: tmp_node   
    ! General
    real (kind=4) :: stokes_max

    ! Initial data
    hi_rad = 0.d0 
    stokes_max = 0.d0
    numsogrid = tab_protein%nb_grids !Number of different sets of grids (different solutes)
    

    ! Loop over solutes
    do i=1,numsogrid

       ! Find first protein of next type
       prot_type=tab_protein%set_of_protein(i,1)

       ! If current solute is flexible, also need to loop over conformations
       ! --Technically this is not needed, as only a single Stokes radius is defined
       !   for each solute. Makes sense to add the code now in case conformation 
       !   specific radii and diffusion coefficients are added later.
       if (tab_protein%array(prot_type)%iflex) then

          ! Number of protein conformations
          nconf = tab_protein%array(prot_type)%p_flex%plist_flex%total_conf

          ! Point tmp_node to first in flexible list
          tmp_node => tab_protein%array(prot_type) % p_flex % plist_flex % p_first
          ! First is empty, point to next for first conformation
          tmp_node => tmp_node % p_next

          ! Counter for loop
          j = 1
          ! Loop over protein conformations. Stop if tmp_node points to last (which is empty)
          do while (.not.associated(tmp_node,tab_protein%array(prot_type)%p_flex%plist_flex%p_last))

             ! Point to sogrid for current conformation
             p_sogrid => tmp_node % p_data

             ! Check if current Stokes radius is gives larger radius
             if (p_sogrid%stoke_radius.GT.stokes_max) then
                stokes_max = real(p_sogrid%stoke_radius,kind=4)
             end if 

             tmp_node => tmp_node % p_next
             ! Update counter
             j = j + 1
          end do

       else
          ! In non-flexible case
          p_sogrid => tab_protein%array(prot_type)%p_sogrid

          ! Check if current Stokes radius is gives larger radius
          if (p_sogrid%stoke_radius.GT.stokes_max) then
             stokes_max = real(p_sogrid%stoke_radius,kind=4)
          end if

       end if
    end do

    ! Check that input radius is larger than largest solute Stokes radius
    if (stokes_max.GT.input_rad) then
       write (*,'("Error: Largest solute Stokes radius =",F9.3,". Radius of local volume =",F9.3)'), stokes_max, input_rad
       write (*,*) "Error: Radius for local volume must be larger than largest solute Stokes radius"
       stop 1
    else
       hi_rad = input_rad
    end if


  

end subroutine set_hi_rad

END MODULE
