!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file mod_type_calculation.f90
!!
!! Define the type of the calculation : sda_2proteins, sda_energy, sdamm,...

! Make a module apart, convenient for input/output of complexes and trajectories
! Lots of test before, now generic function for conversion,
! but do not set the value, has to be done apart

!> Define the type of calculation, and "false parralel" option
MODULE mod_type_calc

    !> enumeration for the type of the calculation
    !! @{
    integer, parameter :: sda_2proteins = 1, &
                          sdamm = 2, &
                          sda_koff =3, &
                          sda_energy = 4, &
                          sda_pmf = 5
    !> @}

    !> Define an enumeration for the type of the calculation and parameters for the false parralel options
    type type_calculation
            !> type of the calculation : \ref sda_2proteins, \ref sdamm, \ref sda_koff, \ref sda_energy or \ref sda_pmf ( not implemented )
            integer :: enum_type_calc
            !> temporary string for the conversion to \ref enum_type_calc\n
            !! size fixed to 15, important for reading header of record
            character(15) :: type_calc_str

            !> For false parrallel options
            !! @{
            logical :: bool_par
            integer :: nb_par
            !> @}
    end type type_calculation

    contains

    !> Get the string format of the type_calculation
    !! @param this : instance of type_calculation
    !! @return string : string format ( "sda_2proteins" / "sdamm" )
    function get_string ( this ) result ( string )
        IMPLICIT NONE
        type ( type_calculation ) :: this
        character(15) :: string

        if ( this % enum_type_calc == sda_2proteins ) then
            string = 'sda_2proteins  '
        else if ( this % enum_type_calc == sdamm ) then
            string = 'sdamm          '
        else if ( this % enum_type_calc == sda_koff ) then
            string = 'sda_koff       '
        else if ( this % enum_type_calc == sda_energy ) then
            string = 'sda_energy     '
        else if ( this % enum_type_calc == sda_pmf ) then
            string = 'sda_pmf        '
        end if

    end function get_string

    !> Get the integer value from the type_calculation
    !! @param this : instance of type_calculation
    !! @return enum_calc : integer value of the enumeration
    function get_type_enum ( this ) result ( enum_calc )
        IMPLICIT NONE
        type ( type_calculation ) :: this
        integer :: enum_calc

        enum_calc = 0
        if ( this % type_calc_str ==      'sda_2proteins  ' ) then
            enum_calc = sda_2proteins
        else if ( this % type_calc_str == 'sdamm          ' ) then
            enum_calc = sdamm
        else if ( this % type_calc_str == 'sda_koff       ' ) then
            enum_calc = sda_koff
        else if ( this % type_calc_str == 'sda_energy     ' ) then
            enum_calc = sda_energy
        else if ( this % type_calc_str == 'sda_pmf        ' ) then
            enum_calc = sda_pmf
        end if


    end function get_type_enum

END MODULE mod_type_calc
