!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_van.f90
!! for Van der Walls and test carges setup

!> group VdW and test charges(for ECM)
!! define predefined values which can be overwritten by the user (add_atoms file)
MODULE mod_van

    !> For VdW values
    type atom_van
        character(2) :: atom_name
        character(3) :: residue_name
        real(kind=8) :: vdw
    end type

    !> Make a separate group for test_charge values
    type atom_tcharge
        character(3) :: atom_name, residue_name
        real ( kind=4 ) :: tcharge
    end type

    ! to delete and put com_sda
    integer, parameter :: io_ext_file = 12345

    integer, parameter :: NB_DEFAULT_VDW = 6
    character(2), dimension(NB_DEFAULT_VDW), parameter :: default_name = &
       (/ 'C*' , 'N*' , 'O*' , 'H*' , 'P*' , 'S*'/)
    character(3), dimension(NB_DEFAULT_VDW), parameter ::  default_resid = &
       (/ '   ', '   ','   ', '   '  ,'   ', '   '/)
    real(kind=8), dimension(NB_DEFAULT_VDW), parameter :: default_vdw = &
       (/ 1.9d0  , 1.625d0, 1.48d0 , 1.2d0 ,  1.9d0  , 1.85d0 /)

    integer, parameter :: NB_DEFAULT_TCHARGE = 23

    !> Regroup VdW and test charges in one object
    type type_van

        logical :: extf_present

        type ( atom_van ), dimension(:), allocatable :: array_atom
        type ( atom_tcharge ), dimension(:), allocatable :: array_tcharge

    end type type_van

    contains

    subroutine new_atom ( this, name, resid, vdw )
        IMPLICIT NONE
        type ( atom_van ) :: this
        character, intent(in) :: name*2, resid*3
        real(kind=8) :: vdw

        !write(*,*) 'Entry new_atom '

        this % atom_name = name
        this % residue_name = resid
        this % vdw = vdw
    end subroutine new_atom

    subroutine new_tcharge ( this, name, resid, tcharge )
        IMPLICIT NONE
        type ( atom_tcharge ) :: this
        character, intent(in) :: name*3, resid*3
        real(kind=4) :: tcharge

        !write(*,*) 'Entry new_tcharge '

        this % atom_name = name
        this % residue_name = resid
        this % tcharge = tcharge

    end subroutine new_tcharge


    subroutine set_default_testcharges ( this, nb_line_tcharge )
        IMPLICIT NONE
        type ( type_van ) :: this
        integer :: nb_line_tcharge

        integer :: index_ch

        ! write(*,*) 'Entry default_testcharges nb_line_tcharge ', nb_line_tcharge

        ! 18 + 5 = 23 up to now

        index_ch = 1 + nb_line_tcharge
        ! RNA/DNA old format
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'A  ' , -1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'T  ' , -1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'G  ' , -1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'C  ' , -1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'U  ' , -1.0 )
        index_ch = index_ch + 1
        ! RNA/DNA new format amber
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'DA ' , -1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'DT ' , -1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'DG ' , -1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'DC ' , -1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'P  ', 'DU ' , -1.0 )
        index_ch = index_ch + 1


        ! amino-acid
        ! histidine, different between whatif, pdb2pqr...
        call new_tcharge ( this % array_tcharge ( index_ch ), 'N  ', 'NTR' ,  1.0 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'ND1', 'HI2' ,  0.5 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'NE2', 'HI2' ,  0.5 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'ND1', 'HIP' ,  0.5 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'NE2', 'HIP' ,  0.5 )
        index_ch = index_ch + 1
        ! other amino-acid, well defined normally
        call new_tcharge ( this % array_tcharge ( index_ch ), 'NZ ', 'LYS' ,  1.0 )
        index_ch = index_ch + 1
        ! could use NH*
        call new_tcharge ( this % array_tcharge ( index_ch ), 'NH1', 'ARG' ,  0.5 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'NH2', 'ARG' ,  0.5 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'OD1', 'ASP' , -0.5 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'OD2', 'ASP' , -0.5 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'OE1', 'GLU' , -0.5 )
        index_ch = index_ch + 1
        call new_tcharge ( this % array_tcharge ( index_ch ), 'OE2', 'GLU' , -0.5 )
        index_ch = index_ch + 1

        ! C-terminal, a lot of variation in sda6
        call new_tcharge ( this % array_tcharge ( index_ch ), 'O* ', 'CTR' , -0.5 )
        index_ch = index_ch + 1

    end subroutine set_default_testcharges

    subroutine init_van( this )
        IMPLICIT NONE
        type ( type_van ) :: this

        integer :: ierr, i, nb_line_vdw, nb_line_tcharge
        logical :: ex

#ifdef DEBUG
        write(*,*) 'Entry init_van '
#endif

        this % extf_present = .false.
        ex = .false.

        inquire (file='add_atoms', exist = ex )
        ! allocate the array_atom and read the first types
        ! works if add_atoms file is empty as well
        if ( ex .eqv. .true. ) then
            write(*,*)
            write(*,*) 'Found "add_atoms" file'
            call read_add_file ( this, nb_line_vdw, nb_line_tcharge )

        ! just use the default size
        else
            write(*,*)
            write(*,*) 'No "add_atoms" file provided '
            allocate ( this % array_atom ( NB_DEFAULT_VDW ), stat = ierr )
            if ( ierr /=0 ) then
                write(*,*) 'ERROR In allocating array_atom '
                stop 1
            end if
            nb_line_vdw = 0

            allocate ( this % array_tcharge ( NB_DEFAULT_TCHARGE ), stat = ierr )
            if ( ierr /=0 ) then
                write(*,*) 'ERROR In allocating array_tcharge '
                stop 1
            end if
            nb_line_tcharge = 0
        endif

        ! for VDW, copy the common data AFTER the ones given by external files, allow to override the values
        do i = 1, NB_DEFAULT_VDW
            call new_atom ( this % array_atom ( i + nb_line_vdw ),default_name(i),default_resid(i),default_vdw(i) )
        end do

        ! for test_charge call external function
        call set_default_testcharges ( this, nb_line_tcharge )

#ifdef DEBUG
        ! print all values in array
        call print_array_atom( this )
#endif

    end subroutine init_van

    subroutine delete_van ( this )
        IMPLICIT NONE
        type ( type_van ) :: this

        integer :: ierr

        if ( allocated ( this % array_atom ) ) then
            deallocate ( this % array_atom, stat =ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocation of array_atom '
                stop 1
            end if
        end if
        if ( allocated ( this % array_tcharge ) ) then
            deallocate ( this % array_tcharge, stat =ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocation of array_tcharge '
                stop 1
            end if
        end if

        !write(*,*) 'after delete van ', allocated ( this % array_atom ), allocated ( this % array_tcharge )
    end subroutine delete_van

     !> Read add_atoms file.\n
     !! copy from read_input_file style
     !! @param this : instance of \ref mod_van
     !! @param nb_line_vdw, nb_line_tcharge : number of input read from add_atoms
    subroutine read_add_file ( this, nb_line_vdw, nb_line_tcharge )
        USE mod_interface
        IMPLICIT NONE
        type ( type_van ) :: this
        integer, intent ( out ) :: nb_line_vdw, nb_line_tcharge


        character(200) :: string
        logical :: vdw_on, tcharge_on

        character(2) :: name2
        character(3) :: name3, resid
        real ( kind=4 ) :: fvalue
        real ( kind=8 ) :: dvalue

        character( 25 ) :: key
        character(128) :: cvalue
        integer :: io_ext_file, io_status

        !write(*,*) 'Entry read_add_file add_atoms '

        vdw_on = .false.
        tcharge_on = .false.
        io_status = 0
        io_ext_file = 111

        open ( io_ext_file, file='add_atoms', form='formatted', status='old' )
        nb_line_vdw = 0
        nb_line_tcharge = 0

        read ( io_ext_file, '(a)', iostat=io_status ) string
        if ( io_status < 0 ) then
            write( *,* ) 'add_atoms file is empty'
        endif
#ifdef DEBUG
        write(*,*) 'io_status ', io_status
#endif


COUNT_LINE : do while ( io_status == 0 )
           !write(*,*) 'new string:!!' , trim(string),'!!'

            if ( (string(1:16) == '# VdW Parameters') .or. (string(1:14) == '# Test Charges') ) then
                write(*,*)
                write(*,*) 'Depreceated format of add_atom file, will stop'
                write(*,*)
                ! working, send to the bash echo $?
                stop 1
            endif

            if ( ( string(1:1) == '#' ) .or. (string(1:1) == '!' ) .or. &
                  len_trim ( string ) < 2) then
                 !write(*,*) 'read a comment, pass'
                 read ( io_ext_file, '(a)', iostat=io_status ) string
                 CYCLE COUNT_LINE
            end if

            ! with GROUP put to false both, order should not matter
            if (string == 'END GROUP') then
                !write (*,*) 'found END GROUP'
                vdw_on = .false.
                tcharge_on = .false.
                read ( io_ext_file, '(a)', iostat=io_status ) string
                CYCLE COUNT_LINE
            end if

            call extract_key ( key, string )
            !write(*,*) 'key ', key, ' ',string

            if ( key == 'GROUP' ) then

                call read_value ( string, cvalue )
                !write(*,*) 'read GROUP cvalue=', trim(cvalue)
                !call read_group ( Inp_Par_List, cvalue, inputf, been_read )

                if ( cvalue == "VdW" ) then
                     !write (*,*) 'found Vdw'
                     vdw_on = .true.
                     tcharge_on = .false.
                     ! add error in count
                     read ( io_ext_file, '(a)', iostat=io_status ) string
                     CYCLE COUNT_LINE
                ! endif

                else if ( cvalue == "TestCharges" ) then
                    vdw_on = .false.
                    tcharge_on = .true.
                    ! add error in count
                    read ( io_ext_file, '(a)', iostat=io_status ) string
                    CYCLE COUNT_LINE
                end if
            end if

            ! general allow
            !write(*,*) 'check vdw or tcharge'
            !write(*,*) 'Something to read'

            if ( vdw_on .eqv. .true. ) then
                !write(*,*) 'VdW on'
                nb_line_vdw = nb_line_vdw + 1

            endif

            if ( tcharge_on .eqv. .true. ) then
                !write(*,*) 'test charge on'
                nb_line_tcharge = nb_line_tcharge + 1

            endif

            read ( io_ext_file, '(a)', iostat=io_status ) string
        enddo COUNT_LINE

#ifdef DEBUG
        write(*,*) 'nb_line_vdw ', nb_line_vdw
        write(*,*) 'nb_line_tcharge ', nb_line_tcharge
#endif
        !write(*,*)

        ! know the final size
        allocate ( this % array_atom ( nb_line_vdw + NB_DEFAULT_VDW ), stat = io_status )
        if ( io_status /= 0 ) then
            write(*,*) 'ERROR In allocating array_atom '
            stop 1
        endif

        allocate ( this % array_tcharge ( nb_line_tcharge + NB_DEFAULT_TCHARGE ), stat = io_status )
        if ( io_status /= 0 ) then
            write(*,*) 'ERROR in allocating array_tcharge '
            stop 1
        endif

        nb_line_vdw = 0
        nb_line_tcharge = 0
        vdw_on = .false.
        tcharge_on = .false.

        ! now redread, but register the new atoms
        rewind ( io_ext_file )
        !write(*,*) 'rewind file'

READ_LINE : do while ( io_status == 0 )
            !write(*,*) 'load data'
            !write(*,*) 'new string load ' , trim(string),'!!'

            if ( ( string(1:1) == '#' ) .or. (string(1:1) == '!' ) .or. &
                   len_trim ( string ) < 2) then
                 !write(*,*) 'read a comment, pass load data'
                 read ( io_ext_file, '(a)', iostat=io_status ) string
                 CYCLE READ_LINE
            end if

            ! with GROUP put to false both, order should not matter
            if (string == 'END GROUP') then
                !write (*,*) 'found END GROUP'
                vdw_on = .false.
                tcharge_on = .false.
                read ( io_ext_file, '(a)', iostat=io_status ) string
                CYCLE READ_LINE
            end if

            !write(*,*) 'check vdw or tcharge'

            !! new version with groups

            call extract_key ( key, string )
            !write(*,*) 'key ', key, ' ',string

            if ( key == 'GROUP' ) then

                call read_value ( string, cvalue )
                !write(*,*) 'read GROUP cvalue=', trim(cvalue)
                !call read_group ( Inp_Par_List, cvalue, inputf, been_read )

                if ( cvalue == "VdW" ) then
                     !write (*,*) 'found Vdw'
                     vdw_on = .true.
                     tcharge_on = .false.
                     ! add error in count
                     read ( io_ext_file, '(a)', iostat=io_status ) string
                     CYCLE READ_LINE

                else if ( cvalue == "TestCharges" ) then
                    !write (*,*) 'found TestCharges'
                    vdw_on = .false.
                    tcharge_on = .true.
                    ! add error in count
                    read ( io_ext_file, '(a)', iostat=io_status ) string
                    CYCLE READ_LINE
                end if
            end if


            if ( vdw_on .eqv. .true. ) then
                !write(*,*) 'read line for vdw '
                read ( string(1:15), fmt='( a2, 3x, a3, 1x, f6.2)' ) name2, resid, dvalue
                !write(*,*) 'name :', name2
                !write(*,*) 'resid :', resid
                !write(*,*) 'value :', fvalue
                nb_line_vdw = nb_line_vdw + 1
                call new_atom ( this % array_atom ( nb_line_vdw ), name2, resid , dvalue )
            endif

            if ( tcharge_on .eqv. .true. ) then
                !write(*,*) 'read line for tcharge '
                read ( string(1:15), fmt='( a3, 2x, a3, 1x, f6.2)' ) name3, resid, fvalue
                !write(*,*) 'name :', name3
                !write(*,*) 'resid :', resid
                !write(*,*) 'value :', fvalue
                nb_line_tcharge = nb_line_tcharge + 1
                call new_tcharge ( this % array_tcharge ( nb_line_tcharge ), name3, resid , fvalue )
            endif

            read ( io_ext_file, '(a)', iostat=io_status ) string
        enddo READ_LINE

        ! avoids problem if new reading
        close ( io_ext_file )

    end subroutine read_add_file

    ! intital sda only 1 character, use intermediate for the moment because 1 character read
    ! expects real*8
    subroutine get_VdW ( this, atom_name, vdw_value ) ! can add residue name, but does not fit for the moment
        IMPLICIT NONE
        type ( type_van ) :: this
        character(2), intent ( in ) :: atom_name
        !character*3, intent ( in ) :: residue_name
        real ( kind=8 ) :: vdw_value

        integer :: i, pos !, pos_res
        logical :: accepted

        !write(*,*) 'Entry get_VdW atom_name!',trim(atom_name),'!!'
        accepted = .false.

LOOP_VDW: do i = 1, size ( this % array_atom )
            ! not needed, always false here
            !accepted = .false.

            ! check first if * is present in the string
            pos = INDEX ( this % array_atom ( i ) % atom_name, "*" )

            ! comapre atom_name first
            if ( pos /= 0 ) then
                if ( this % array_atom ( i ) % atom_name(1:pos-1) == atom_name(1:pos-1) ) then
                    accepted = .true.
                endif
            else if ( trim ( this % array_atom ( i ) % atom_name ) == atom_name ) then
                ! set accepted to true
                accepted = .true.
            endif

            ! could CYCLE here, no match with name no need to continue
            if ( accepted .eqv. .false. ) then
                CYCLE LOOP_VDW
            ! else only becasue residue not implemented in VdW
            else
                vdw_value = this % array_atom ( i ) % vdw
                return
            end if

        end do LOOP_VDW

        write(*,*) 'WARNING Atom type :',trim(atom_name), ' has not been found, use a default VdW of 1.9'
        vdw_value = 1.9

    end subroutine get_VdW

    !subroutine get_test_charge ( this, atom_name, residue_name, test_charge, charge )
    ! use 0 return value, used when we want to override an atom type
    ! maybe test test /= 0 is dangerous with real*4, to see on test
    subroutine get_test_charge ( this, atom_name, residue_name, charge )
        IMPLICIT NONE
        type ( type_van ) :: this
        character(3), intent ( in ) :: atom_name, residue_name
        real ( kind = 4 ), intent ( out ) :: charge
        !logical, intent ( out ) :: test_charge

        integer :: i, pos, pos_res
        logical :: accepted

        ! default false and 0.
        ! test_charge = .false.
        accepted = .false.
        charge = 0.

LOOP_TCH : do i = 1, size ( this % array_tcharge )

            ! check first if '*' is present in the name string of the atom
            pos = INDEX ( this % array_tcharge ( i ) % atom_name, "*" )

            ! check atom name first
            if (  pos /= 0 ) then
                if ( this % array_tcharge ( i ) % atom_name(1:pos-1) == atom_name(1:pos-1) )then
                    accepted = .true.
                end if
            else if ( atom_name == this % array_tcharge ( i ) % atom_name ) then
                accepted = .true.
            end if

            ! could CYCLE here, no match with name no need to continue
            if ( accepted .eqv. .false. ) then
                CYCLE LOOP_TCH
            end if

            accepted = .false.
            pos_res = INDEX ( this % array_tcharge ( i ) % residue_name, "*" )

            if ( pos_res /= 0 ) then

                ! if only wildchar, always accept, use adjustl, not tested
                !if ( trim(this % array_tcharge ( i ) % residue_name) == "  *" ) then
                if ( trim(adjustl(this % array_tcharge ( i ) % residue_name)) == "*" ) then
                    accepted = .true.
                ! called in case of HI*
                else if ( residue_name(1:pos_res-1) == this % array_tcharge ( i ) % residue_name(1:pos_res-1) ) then
                    accepted = .true.
                end if
            ! check fullname, use of trim & adjustl to be a bit more flexible
            !else if ( residue_name == this % array_tcharge ( i ) % residue_name ) then
            else if ( trim(Adjustl(residue_name)) == trim(adjustl(this % array_tcharge ( i ) % residue_name)) ) then
                    accepted = .true.
                    !write(*,*) 'residue name2!',trim(residue_name),'!!'
                    if ( ( trim(residue_name) == 'CTR') .or. &
                         ( trim(residue_name) == 'NTR') ) then
                        write(*,*)
                        write(*,*) 'INFO CTR or NTR is present, a default value has been associated '
                        write(*,*) 'INFO Check the *.tcha file and/or overwrites values in add_atoms '
        endif
            end if

            ! if still accepted a second time, match is found, can return
            if ( accepted .eqv. .true. ) then
                !test_charge = .true.
                charge = this % array_tcharge ( i ) % tcharge
                return
            endif

        end do LOOP_TCH
        ! if here no match, return value 0.
        !write(*,*) 'test charge, no match for ', atom_name,' ', residue_name
        if ( trim(atom_name) == 'OXT') then
             write(*,*)
             write(*,*) 'INFO OXT atom is present, but no default value is associated. Add a value in add_atoms.'
             write(*,*) 'INFO Certainly charged termini are present, check as well for the N-terminus.'
        endif

        return

    end subroutine get_test_charge

    subroutine print_array_atom ( this )
        IMPLICIT NONE
        type ( type_van ) :: this

        integer :: i

        write(*,*)
        write(*,*) 'Entry print_array_atom Vdw and TestCahrges'
        write(*,*) 'Default atom definitions: '
        write(*,*) 'Size VdW ', size( this % array_atom )
        write(*,*) 'Size Test Charges ', size( this % array_tcharge )

        do i = 1, size ( this % array_atom )
            write(*,*) this % array_atom( i ) % atom_name,' ',  &
                             this % array_atom( i ) % residue_name,' ', &
                 this % array_atom( i ) % vdw !,' ',this % array_atom( i ) % tcharge
        end do
         write(*,*) 'Array atoms Test charges ', size( this % array_tcharge )
        do i = 1, size ( this % array_tcharge )
            write(*,*) this % array_tcharge( i ) % atom_name,' ',  &
                             this % array_tcharge( i ) % residue_name,' ', &
                 this % array_tcharge( i ) % tcharge !,' ',this % array_atom( i ) % tcharge
        end do


    end subroutine print_array_atom

END MODULE mod_van
