!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file prepare_grid.f90
!! Load the grids and call \ref prepare_atom_protein::prepare_atom_protein "prepare_atom_protein"

! Group all allocation for set of grids
! initialisation: allocate_sogrid is done by the calling function
! and initialize protein is done before
! this way can be used for sda and flex module


!> Allocate one set of grid and return an associated pointer on it
!!
!! Group all allocation for set of grids, call \ref prepare_atom_protein "prepare_atom_protein" to load the pdbs and charges\n
!! The initialisation of sogrid must be done before\n
!! Correctly read position of reaction coordinate ( in case of flex )\n
!! Add iflex, if THIS Solute_Grid group is not flexible, assign criteria pos from rxna ( default counter for id 1 or 2 )
!! If Flex, read new coordinate in tmp_atom_rcriteria, to not erase pos from rxna
!! @param pdb_filename : name of the pdb
!! @param q_filename, ep_filename, ed_filename, hd_filename, lj_rep_filename, lj_filename : name of the charge and grids
!! @param atom_type_name : atom to use for lennard-jones interaction
!! @param epfct, edfct, hdfct, lj_rep_fct, ljfct : multiplicative factors for the grids
!! @param param_analytic : structure \ref com_sda::parameter_analytic "parameter_analytic"
!! @param dm, dr : translational and rotational coefficient diffusion
!! @param stoke_rad : stoke_radius for hydrodynamic interactions
!! @param param_probe : contains the different probe in \ref com_sda::probe_type "param_probe"
!! @param bool_excl_grid : if the exclusion grid must be computed ( or read from the disk )
!! @param psgrid : pointer to \ref mod_setofgrid::sogrid "sogrid" to fill
!! @param center : center of the protein
!! @param counter_grid, counter_conf : internal counters
!! @param rcriter : instance of \ref mod_reactioncriteria::react_criter "react_criter"
!! @param iflex : if the solute is flexible
!! @param isurface : if the surface is a surface
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param nb_lj_opt : optional number of lennard-jones grid or atom type to read
subroutine prepare_grid ( pdb_filename, q_filename, ep_filename, ed_filename, &
                          hd_filename, lj_rep_filename, lj_filename, atom_type_name, &
                          epfct, edfct, hdfct, lj_rep_fct, ljfct, param_analytic, dm, dr, stoke_rad, &
                          param_probe, bool_excl_grid, psgrid, center, &
                          counter_grid, counter_conf, rcriter, iflex, isurface, &
                          type_calc, nb_lj_opt )
    USE mod_setofgrid
    USE mod_ReactionCriteria
    USE com_sda
    USE mod_type_calc
    IMPLICIT NONE

    interface
        subroutine prepare_atom_protein ( sgrid, rcriter, pdb_filename, param_probe, &
                                  center, counter, bool_excl_grid, iflex, isurface, type_calc )
            USE mod_setofgrid
            USE mod_ReactionCriteria
            USE mod_type_calc
            USE com_sda ! needed for log output of access atoms
            IMPLICIT NONE
            type ( sogrid ) :: sgrid
            type ( react_criter ) :: rcriter
            type ( type_calculation ) :: type_calc
            character ( 128 ), intent ( in ) :: pdb_filename
            type ( probe_type ), intent ( in ) :: param_probe
            logical, intent ( in ) :: iflex, isurface
            logical, intent ( in ) :: bool_excl_grid
            integer :: counter
            real ( kind=8 ), dimension(3), intent ( out ) :: center
        end subroutine prepare_atom_protein
    end interface

    !! will return a new allocated psgrid
    type ( sogrid ), pointer :: psgrid
    type ( react_criter ) :: rcriter
    type ( type_calculation ) :: type_calc
    type ( parameter_analytic ) :: param_analytic
    character*128, intent( in ) :: pdb_filename, q_filename, ep_filename, ed_filename, hd_filename, lj_rep_filename
    character*128, dimension(:), allocatable, intent( in ):: lj_filename
    character*2, dimension(:), allocatable, intent( in ):: atom_type_name
    real ( kind=8 ), intent( in ) :: epfct, edfct, hdfct, lj_rep_fct, ljfct
    real ( kind=8 ), intent( in ) :: dm, dr, stoke_rad
    type ( probe_type ) :: param_probe
    real ( kind=8 ), dimension( 3 ) :: center
    integer, intent ( in ) :: counter_grid, counter_conf
    logical :: bool_excl_grid, iflex, isurface

    ! new added optional
    integer, optional :: nb_lj_opt

    integer :: nb_lj
    real ( kind=8 ) :: rt_1
    integer :: ierr, i, j
    character*128 :: empty_string


    ierr = 1
    rt_1 = 1.d0 / ( temper*rcnst )
    !write(*,*) 'rt_1', rt_1
    ! needed for correct compilation with intel
    empty_string=""

    if ( present ( nb_lj_opt ) ) then
        !write(*,*) 'nb_lj_opt is present ', nb_lj_opt
        nb_lj = nb_lj_opt
    else
        nb_lj = 0
    endif

    write(*,*)
    write(*,*) '=========================================================================================='
    write(*,*) '     Preparation of Set of Grid ', counter_grid, 'Conformation number ', counter_conf
    write(*,*)
    !write(*,*) 'isurface ', isurface
    !write(*,*) 'hdfct ', hdfct, hdfct * rt_1
    !write(*,*) 'lj_rep_fct', lj_rep_fct

    ! allocate the set of grid
    allocate ( psgrid, stat = ierr )
    if ( ierr /= 0 ) then
       write ( *,* ) 'ERROR Cannot allocate the set of grid for flexibility'
       !return
       STOP 1
    end if
    ! and intialize it
    call allocate_sogrid ( psgrid )

    if ( epfct /= 0. ) then
        !write(*,*) 'load grid ep'
        if ( len(trim(ep_filename)) == 0 ) then
            write(*,*)
            write(*,*) 'INFO epfct != 0 but ep_filename is empty for solute-grid ', counter_grid
            write(*,*)
        else
            call load_grid ( psgrid, ep_filename, &
                             get_nb_gridtype('electrostatic') , epfct * rt_1 )
        end if
    endif

    if ( edfct /= 0. ) then
        !write(*,*) 'want to load ed_filename ', trim(ed_filename),'!!'
        if ( len(trim(ed_filename)) == 0 ) then
            write(*,*)
            write(*,*) 'INFO edfct != 0 but ed_filename is empty for solute-grid ', counter_grid
            write(*,*)
        else
            call load_grid ( psgrid, ed_filename, &
                             get_nb_gridtype('desolvation'), edfct * rt_1 )
        end if
        ! can set asymetry here also
    endif

    if ( hdfct /= 0. ) then
        !write(*,*) 'want to load hd_filename ', trim(hd_filename),'!!'
        if ( len(trim(hd_filename)) == 0 ) then
            write(*,*)
            write(*,*) 'INFO hdfct != 0 but hd_filename is empty for solute-grid ', counter_grid
            write(*,*)
        else
            call load_grid ( psgrid, hd_filename, &
                             get_nb_gridtype('hydrophobic'), hdfct * rt_1 )
        end if
    endif

    if ( lj_rep_fct /=0. ) then
        if ( len(trim(lj_rep_filename) ) == 0 ) then
            write(*,*)
            write(*,*) 'INFO lj_rep_fct != 0 but lj_rep_filename is empty for solute-grid ', counter_grid
            write(*,*)
        else
            call load_grid ( psgrid, lj_rep_filename, &
                             get_nb_gridtype('repulsive_lj'), lj_rep_fct * rt_1 )
        end if
    end if

    !write(*,*) 'Before lennardjones get_nb_gridtype(lennardjones) ', get_nb_gridtype('lennardjones')
    if ( ljfct /= 0. ) then

        !write(*,*) 'ljfct \= 0 lj_filename ', trim ( lj_filename )
        if ( allocated ( lj_filename ) ) then
            !write(*,*) 'lj_filename allocated with size :', size ( lj_filename )
            call load_ljgrid ( psgrid, 1, lj_filename, pdb_filename, rt_1 )
        end if

        if ( allocated ( atom_type_name ) ) then
            !write(*,*) 'Atom type is allocated with size : ', size ( atom_type_name )
            call load_ljgrid ( psgrid, 2, atom_type_name, pdb_filename )
        end if
    endif

    if ( ( len_trim ( q_filename ) > 0 ) ) then
         !write(*,*) 'read effective charge! ', trim ( q_filename )
         call load_grid ( psgrid, q_filename, get_nb_gridtype('charge'), 1.d0 )
    end if

    ! Debye-huckel, analytical grid, no need of filename
    ! if not called, can be filled later ? in fact do nothing here ? to check
    if ( param_analytic % ianalytic ) then

#ifdef DEBUG
        write(*,*)
        write(*,*) 'prepare_grid for analytcal interactions'
        write(*,*) 'bin size ', param_analytic % h_analytic
        write(*,*) 'debyeh ', param_analytic % debyeh
        write(*,*) 'hydrodynamic ', param_analytic % hydro
        write(*,*)
#endif

        if ( param_analytic % debyeh ) then

            ! indicate sogrid must include debye-huckel
            psgrid % idebye = .true.

#ifdef DEBUG
            write(*,*) 'preload debye-huckel interaction type integer =', get_nb_analyticaltype('debye-huckel')
            write(*,*) 'get_nb_analytical debye-huckcel ',get_nb_analyticaltype('debye-huckel')
            write(*,*) 'ionic strength= ', param_analytic % ionic
#endif

            ! test if electrostatic has been correctly loaded, number 1 normally
            if ( is_associated ( psgrid % pset_grid( get_nb_gridtype('electrostatic') ) ) .eqv. .false. ) then
                write(*,*) 'ERROR For using Debye-Huckel interaction'
                write(*,*) 'ERROR An electrostatic potential grid must be present (set a correct epf)'
                STOP 1
            end if

            ! always for ep grid and DH, asumme UHBD grid already loaded
            write(*,*)
            write(*,*) 'INFO Make Electrostatic grid spherical for use with Debye-Huckel'
            call make_sphere ( psgrid % pset_grid ( get_nb_gridtype('electrostatic') ) )
        end if
    ! only for tests
#ifdef DEBUG
    else
        write(*,*)
        write(*,*) 'No analytical interactions (prepare_grids.f90)'
        write(*,*)
#endif
    end if

    ! assign dm and dr, to make in one allocation or prep later
    ! could be different for each conformation, use of default value possible
    psgrid % diffusion_trans = dm
    psgrid % diffusion_rotat = dr
    ! check debyeh or crowder stock_radius mandatory !
    if ( param_analytic % ianalytic ) then
        if ( stoke_rad == 0.d0 ) then
            write(*,*) 'ERROR Grid number ', counter_grid
            write(*,*) 'ERROR stoke_radius must be set correctly for using Debye-Huckel or Crowders interactions'
            STOP 1
        end if
    end if
    ! set to 0.d0 by default
    psgrid % stoke_radius = stoke_rad

    ! will read pdb and assign effective charge, atom_surface and maybe exclusion grid
    ! will compute center, maybe read new rcriteria position
    call prepare_atom_protein ( psgrid, rcriter, pdb_filename, param_probe, center, &
                                counter_grid, bool_excl_grid, iflex, isurface, type_calc )

    write(*,*)
    write(*,*) "=== 3D Grids ==="
    ! here group all shift to center of mass, atom_surface could be done in read_pdb or prepare_atom_protein
    ! rcriteria shift done in prepare_atom_protein
    ! Here the sogrid is filled, move to the center of mass of all the grids and list of atom for protein
    do i =1,4
        if ( is_associated ( psgrid % pset_grid( i ) ) ) then
            ! write name of grid
            call print_gridtype( i )
            !call print_origin ( psgrid % pset_grid( i ) )
            call shift_origin ( psgrid % pset_grid ( i ), center )
            call print_origin ( psgrid % pset_grid( i ) )
        end if
    end do

    ! specific to exclusion grid
    ! in a separate module, should combine later...
    if ( associated ( psgrid % p_excl_grid ) ) then
        !write(*,*) 'exclusion grid origin', psgrid % p_excl_grid % pgrid % origin
        call print_gridtype( get_nb_gridtype('exclusion') )
        ! not stored there
        !call print_origin ( psgrid % pset_grid( get_nb_gridtype('exclusion') ) )
        call print_origin ( psgrid % p_excl_grid % pgrid )
#ifdef DEBUG
    else
        write(*,*) 'No exclusion grid'
#endif
    end if

    ! make shift for the charges
    if ( allocated ( psgrid % position_charge ) ) then
        do i = 1, psgrid % nb_charge
            psgrid % position_charge(:,i) = psgrid % position_charge(:,i) - center
       end do
    end if

    ! for atom_surface
    if ( allocated ( psgrid % list_atom_surface ) ) then
        do i = 1, psgrid % nb_atom_surface
            psgrid % list_atom_surface (:,i) = psgrid % list_atom_surface(:,i) - center
        end do
    end if

    ! need same for atom_type
#ifdef DEBUG
    write(*,*)
    write(*,*) 'Before center atoms center ', center
    write(*,*) 'first charge position ', psgrid % position_charge (:,1)
    write(*,*) 'first atom list_atom_surface ', psgrid % list_atom_surface (:,1)
#endif

    if ( associated ( psgrid % p_ljgrid ) ) then
      !write(*,*) 'prepare grid psgrid % p_ljgrid is associated'
      ! move array pos if loaded
      if ( allocated ( psgrid % p_ljgrid % array_position_atom_lj ) ) then
        ! loop over the grids
        do i = 1, psgrid % p_ljgrid % nb_ljgrid
          !write(*,*) 'i '
          ! solve problem, to make better, but can imagine some "blank entry"
          if ( associated ( psgrid % p_ljgrid % array_position_atom_lj ( i ) % p_position_atom ) ) then

            !write(*,*) 'psgrid % p_ljgrid % array_position_atom_lj ( i ) % nb_pos_atom', &
            !    psgrid % p_ljgrid % array_position_atom_lj ( i ) % nb_pos_atom
            !write(*,*) 'move atoms to center : ', center

            ! loop over the atom
            do j = 1, psgrid % p_ljgrid % array_position_atom_lj ( i ) % nb_pos_atom
                psgrid % p_ljgrid % array_position_atom_lj ( i ) % p_position_atom (:,j) = &
                  psgrid % p_ljgrid % array_position_atom_lj ( i ) % p_position_atom (:,j) - center
            end do
          end if
        end do
      end if ! if atom_pos

      ! shift the grid
      if ( associated ( psgrid % p_ljgrid % array_ljgrid ) ) then
          !write(*,*) 'Will shift the lj grid with center ', center
          ! 5 associated to lennard-jones
          !call print_gridtype( 5 )
          call print_gridtype( get_nb_gridtype('lennardjones') )

          do i = 1, psgrid % p_ljgrid % nb_ljgrid
              !call print_gridtype( i )
              !call print_origin ( psgrid % p_ljgrid % array_ljgrid ( i ) )
              call shift_origin ( psgrid % p_ljgrid % array_ljgrid ( i ), center )
              call print_origin ( psgrid % p_ljgrid % array_ljgrid ( i ) )
          end do
      end if ! array_ljgrid associated

    end if ! if lenard_jones associated

    ! reaction criteria are done apart, not always wanted

    ! print log_file
    ! could be in read_input_file, put in prepare_grid
    !if ( bool_log ) then
    !    call print_excl_grid_log ( sgrid % p_excl_grid, iopoti )
    !end if

    !write(*,*)
    !write(*,*) ' Exit Preparation of Set of Grid ', counter_grid, 'Conformation number ', counter_conf
    !write(*,*)

end subroutine prepare_grid
