!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file read_eff_charges.f90
!!
!! Read the effective charges

! could be in mod_sogrid
! Fill array \ref mod_setofgrid::sogrid::position_charge "position_charge" and \ref mod_setofgrid::sogrid::charge_grid "charge_grid" of \ref mod_setofgrid::sogrid "sogrid"

!> Read effective charges from file
!!
!! Fill arrays \ref mod_setofgrid::sogrid::position_charge "position_charge" and \ref mod_setofgrid::sogrid::charge_grid "charge_grid" of \ref mod_setofgrid::sogrid "sogrid"
!! @param o_sogrid : instance of \ref mod_setofgrid::sogrid "sogrid"
!! @param filename : name of the file
!! @param qtot : total charge read from the file
subroutine read_eff_charges( o_sogrid , filename, qtot )
    USE mod_setofgrid
    USE com_sda

    IMPLICIT NONE
    type ( sogrid ), intent(inout) :: o_sogrid
    character*128, intent ( in ) :: filename
    real ( kind=8 ), intent(inout) :: qtot

    character*128 :: string
    integer :: nat, io, i
    integer :: iueff = 32
    logical :: ex

    !write(*,*) 'read_eff_charges ', trim(filename)
    inquire ( file=filename, EXIST=ex )
    if ( .not.ex ) then
       write ( 6, *) "ERROR file with charges does not exist or filename is missing ", &
                     trim ( filename )
       !return
       stop 1
    end if

    open ( iueff, file=filename, form='formatted', status='old' )
    nat = 0
    qtot = 0.d0
    !count the number of line
    read ( iueff, '(a)', iostat=io ) string
    if ( io < 0 ) then
      write( *,* ) 'file is empty', filename
      close ( iueff )
      return
    endif

    do while ( io == 0 )
      if ( (string(1:4)=='ATOM').or.(string(1:6)=='HETATM') ) then
          nat =nat+1
      endif
      read ( iueff, '(a)', iostat=io ) string
    enddo

    ! commented for test, if empty return before
    if ( nat == 0 ) then
       write(*,*) 'INFO Number of charge is 0, return'
       ! could print summary here
       return
    endif

    call allocate_charge ( o_sogrid, nat )
    !write(*,*) 'number of charges ', nat

    rewind( iueff )
    io=0
    nat = 0
    ! read first line, otherwise make one turn more
    read ( iueff, '(a)', iostat=io ) string
    do while ( io == 0 )

      if ( (string(1:4)=='ATOM').or.(string(1:6)=='HETATM') ) then
          nat = nat + 1
          read (string,501) ( o_sogrid % position_charge(i,nat),i=1,3 ), &
                             o_sogrid % charge_grid(nat)

          qtot = qtot +  o_sogrid % charge_grid( nat )
      endif
      read ( iueff, '(a)', iostat=io ) string
    enddo

    close ( iueff )

    !write(*,*) 'tot charge ', qtot
    write (io6, fmt = '("Read effective charges : ",a)' ) trim ( filename )
    write (io6,601) nat, qtot

  501 FORMAT (30x,3f8.3,6x,f7.3)
  601 FORMAT ('No of charges read for grid : ',i8,/, &
              'Net charge :                ',f10.3 )

end subroutine read_eff_charges
