!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file read_group_file.f90
!!
!! Read all parameters of one specific GROUP from the sda input file

! try to put this usefull function, want to use by tool in a separate program
! need to use mod_interface for correct usage of overloading

!> Separate cases for reading each GROUP
!! @param p_list : instance of \ref mod_input_parameter::param_list "param_list"
!! @param keyword : value of the GROUP
!! @param io_unit : file descriptor of the input file
!! @param been_read : used to detect misspelling
subroutine read_group ( p_list, keyword, io_unit, been_read )
    USE com_sda
    USE mod_interface
    USE mod_input_parameter

    USE mod_geometry
    USE mod_protein

    IMPLICIT NONE
    type ( param_list ) :: p_list
    character ( 128 ) :: keyword
    integer :: io_unit, io_status
    logical :: been_read

    logical :: stop_read, local_been_read, local_been_read2
    character( 200 ) :: Line_Buffer
    character( 25 ) :: key
    real ( kind=8 ) :: dvalue
    real ( kind=4 ) :: fvalue
    integer :: ivalue
    character(128) :: cvalue
    character(128), dimension(:), allocatable :: array_cvalue


    logical :: lvalue

    !! particular values for groups
    integer :: nb_solute

    stop_read = .false.
    local_been_read = .true.
    local_been_read2 = .true.
    !! default values for variable only in groups
    nb_solute = 1

    !keep it if problem reading
    !write(*,*) 'Entry Read GROUP ',trim(keyword)

NEXT: &
    do while ( stop_read .eqv. .false. )

        !can check if error, only print for first time
        if ( ( local_been_read .eqv. .false. ) .and. ( local_been_read2 .eqv. .true. ) ) then
            !write(*,*) 'we got an error in reading'
            local_been_read2 = .false.
        end if

        read ( io_unit, '(a)', iostat=io_status ) Line_Buffer
        if  ( Line_Buffer == 'END GROUP' ) then
            stop_read = .true.
            CYCLE NEXT
        endif

        if ( ( Line_Buffer(1:1) == '#' ) .or. ( Line_Buffer(1:1) == '!' ) .or. &
           ( len_trim ( Line_Buffer ) == 0)  ) CYCLE NEXT

            !write(*,*) 'length ', len_trim(Line_Buffer)
            call extract_key ( key, Line_Buffer )

!!!!!!!! Read GROUP Type_Calculation !!!!!!!!!!!!!!!!!!!!!!!!!!!
          if ( keyword == 'Type_Calculation' ) then

            if ( key == 'type' ) then
                call read_value ( Line_Buffer, cvalue )
                !write(*,*) 'type calculation : ', cvalue

                if ( cvalue == 'sda_2proteins' ) then
                    p_list % t_calc % enum_type_calc = sda_2proteins
                else if ( cvalue == 'sdamm' ) then
                    p_list % t_calc % enum_type_calc = sdamm
                else if ( cvalue == 'sda_koff' ) then
                    p_list % t_calc % enum_type_calc = sda_koff
                else if ( cvalue == 'sda_energy' ) then
                    p_list % t_calc % enum_type_calc = sda_energy
                else if ( cvalue == 'sda_pmf' ) then
                    p_list % t_calc % enum_type_calc = sda_pmf
                else
                    write( *, * ) 'not a valid name for type ',cvalue
                    stop 1
                end if
                CYCLE NEXT
            endif

            if ( key == "total_solutes" ) then
                call read_value ( Line_Buffer, ivalue)
                p_list % total_solutes = ivalue
                CYCLE NEXT
            end if

            if ( key == "total_grids" ) then
                call read_value ( Line_Buffer, ivalue)
                p_list % total_grids = ivalue
                CYCLE NEXT
            end if

            write(*,*)
            write(*,*) 'WARNING Miss reading in Type_Calculation of the key ', key
            write(*,*)
            local_been_read = .false.

        end if

!!!!!!!! Read GROUP Solute_Grid !!!!!!!!!!!!!!!!!!!!!!!!!!!
        if ( keyword == 'Solute_Grid' ) then
            !write(*,*) 'key ', key
            if ( key == 'nb_solute' ) then
                call read_value ( Line_Buffer, ivalue )
                p_list % nb_solute = ivalue
                CYCLE NEXT
            endif

            if ( key == 'pdb_filename' ) then
                call read_value ( Line_Buffer, cvalue )
                p_list % pdb_filename  = cvalue
                CYCLE NEXT
            endif

            if ( key == 'rotate' ) then
                call read_value ( Line_Buffer, lvalue )
                p_list % irot  = lvalue
                CYCLE NEXT
            endif

            if ( key == 'surface' ) then
                call read_value ( Line_Buffer, lvalue )
                p_list % isurf  = lvalue
                !! if surface do not move
                !p_list % imove = .false.
                CYCLE NEXT
            endif

            if ( key == 'flex' ) then
                call read_value ( Line_Buffer, lvalue )
                p_list % iflex  = lvalue
                CYCLE NEXT
            endif

            if ( key == 'image_charge' ) then
                call read_value ( Line_Buffer, lvalue )
                p_list % iimg_chg  = lvalue
                CYCLE NEXT
            endif

            if ( key == 'epf' ) then
                call read_value ( Line_Buffer, cvalue )
                p_list % epf  = cvalue
                CYCLE NEXT
            endif

            if ( key == 'qef' ) then
                call read_value ( Line_Buffer, cvalue )
                p_list % qef  = cvalue
                CYCLE NEXT
            endif

            if ( key == 'edf' ) then
                call read_value ( Line_Buffer, cvalue )
                p_list % edf  = cvalue
                CYCLE NEXT
            endif

            if ( key == 'hdf' ) then
                call read_value ( Line_Buffer, cvalue )
                p_list % hdf  = cvalue
                CYCLE NEXT
            end if

            if ( key == 'lj_repf' ) then
                call read_value ( Line_Buffer, cvalue )
                p_list % lj_repf  = cvalue
                CYCLE NEXT
            end if

            if ( key == 'nb_lj' ) then
                call read_value ( Line_Buffer, ivalue )
                p_list % nb_lenjones = ivalue
                CYCLE NEXT
            endif

            if ( key == 'ljf' ) then
                !write(*,*) 'new lennard-jones'

                call read_value ( Line_Buffer, array_cvalue )
                ! array cvalue allocated
                !write(*,*) 'allocated p_list % ljf ', allocated ( p_list % ljf )
                !write(*,*) 'size array_cvalue ', size(array_cvalue)

                if ( allocated ( p_list % ljf ) .eqv. .false. ) then
                    allocate ( p_list % ljf( size(array_cvalue) ) )
                end if
                ! could check the size

                !write(*,*) 'make copy size ljf ', size ( p_list % ljf )
                p_list % ljf  = array_cvalue
                !write(*,*) 'plist %ljf ', p_list % ljf
                CYCLE NEXT
            end if

            if ( key == 'atom_lj' ) then
                !write(*,*) 'new lennard-jones'
                call read_value ( Line_Buffer, array_cvalue )

                ! assur it is allocated the first time
                if ( allocated ( p_list % atom_type_lj ) .eqv. .false. ) then
                    allocate ( p_list % atom_type_lj( size(array_cvalue) ) )
                end if

                ! here warning 128 to 2, should check the size
                p_list % atom_type_lj  = array_cvalue
                CYCLE NEXT
            end if

!!!!!!!!!! Diffusion Coefficient
            if ( key ==  'diffusion_trans' ) then
                call read_value ( Line_Buffer, dvalue )
                p_list % dm = dvalue
                CYCLE NEXT
            endif

            if ( key == 'diffusion_rotat' ) then
                call read_value ( Line_Buffer, dvalue )
                p_list % dr = dvalue
                CYCLE NEXT
            endif

            if ( key == 'stoke_radius' ) then
                call read_value ( Line_Buffer, dvalue )
                p_list % stoke_rad = dvalue
                CYCLE NEXT
            endif

            !! all following only used by sda_flex, otherwise default value
            if ( key == 'total_conf' ) then
                call read_value ( Line_Buffer, ivalue )
                p_list % total_conf = ivalue
                CYCLE NEXT
            endif

            if ( key == 'initial_conf' ) then
                call read_value ( Line_Buffer, ivalue )
                p_list % initial_conf = ivalue
                CYCLE NEXT
            endif

            if ( key == 'nearest' ) then
                call read_value ( Line_Buffer, lvalue )
                p_list % nearest_ngb =lvalue
                CYCLE NEXT
            endif

            if ( key == 'frequency' ) then
                call read_value ( Line_Buffer, fvalue )
                !p_list % first_conf = ivalue
                p_list % frequence = fvalue
                CYCLE NEXT
            endif

            if ( key == 'std_frequency' ) then
                call read_value ( Line_Buffer, fvalue )
                !p_list % first_conf = ivalue
                p_list % std_freq = fvalue
                CYCLE NEXT
            endif

            if ( key == 'method' ) then
                call read_value ( Line_Buffer, cvalue )
                p_list % flex_method = cvalue
                CYCLE NEXT
            endif

            if ( key == 'list_conf' ) then
                call read_value ( Line_Buffer, cvalue )
                p_list % list_conf_name  = cvalue
                CYCLE NEXT
            endif

            write(*,*)
            write(*,*) 'WARNING Miss reading in Solute_Grid with the key ', key
            write(*,*)
            local_been_read = .false.

        end if !! end if protein_grid

!!!!!!!!!!!!!!!!! Read geometry !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if ( keyword == 'Geometry' ) then

            if ( key == 'type' ) then
                call read_char ( Line_Buffer, cvalue )
                if ( cvalue == 'sphere' ) then
                    p_list % type_geom = enum_sphere
                else if ( cvalue == 'box' ) then
                    p_list % type_geom = enum_box
                else
                write ( *,*) 'ERROR Type is unknown in geometry:',cvalue,'!'
                stop 1
            endif
            CYCLE NEXT
        endif

        if ( key == 'start_pos' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % start_pos = fvalue
            CYCLE NEXT
        endif

        if ( key == 'pbc' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % pbc = lvalue
            CYCLE NEXT
        endif

        if ( key == 'escape' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % escape = lvalue
            CYCLE NEXT
        endif

        if ( key == 'surface' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % surf_geom = lvalue
            CYCLE NEXT
        endif

        if ( key == 'c' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % c = fvalue
            CYCLE NEXT
        endif

        if ( key == 'xmin' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % xmin = fvalue
            CYCLE NEXT
        endif

        if ( key == 'xmax' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % xmax = fvalue
            CYCLE NEXT
        endif

        if ( key == 'ymin' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % ymin = fvalue
            CYCLE NEXT
        endif

        if ( key == 'ymax' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % ymax = fvalue
            CYCLE NEXT
        endif

        if ( key == 'zmin' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % zmin = fvalue
            CYCLE NEXT
        endif

        if ( key == 'zmax' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % zmax = fvalue
            CYCLE NEXT
        endif

        write(*,*)
        write(*,*) 'WARNING Miss reading in Geometry with the key ', key
        write(*,*)
        local_been_read = .false.

    end if !! end if group geometry

!!!!!!!!!!!!!!!!! Read ResidenceTime !!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    if ( keyword == 'ResidenceTime' ) then

        if ( key == 'type_residence' ) then
            call read_char ( Line_Buffer, cvalue )
            p_list % enum_str_residtime = cvalue
            CYCLE NEXT
        endif

        if ( key == 'fixed_bin_size' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % fixe_bin_size = ivalue
            CYCLE NEXT
        endif

        if ( key == 'size_x' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % size_cell_x = fvalue
            CYCLE NEXT
        endif

        if ( key == 'size_y' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % size_cell_y = fvalue
            CYCLE NEXT
        endif

        if ( key == 'size_z' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % size_cell_z = fvalue
            CYCLE NEXT
        endif

        if ( key == 'cell_x' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % nb_cell_x = ivalue
            CYCLE NEXT
        endif

        if ( key == 'cell_y' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % nb_cell_y = ivalue
            CYCLE NEXT
        endif

        if ( key == 'cell_z' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % nb_cell_z = ivalue
            CYCLE NEXT
        endif

        if ( key == 'shift_rt_x' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % shift_rt_x = fvalue
            CYCLE NEXT
        endif

        if ( key == 'shift_rt_y' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % shift_rt_y = fvalue
            CYCLE NEXT
        endif

        if ( key == 'shift_rt_z' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % shift_rt_z = fvalue
            CYCLE NEXT
        endif

        if ( key == 'format_rt3d' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % format_rt3d = ivalue
            CYCLE NEXT
        endif

        if ( key == 'pair_solutes' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % pair_proteins = lvalue
            CYCLE NEXT
        endif

        write(*,*)
        write(*,*) 'WARNING Miss reading in Residence with the key ', key
        write(*,*)
        local_been_read = .false.

    end if ! end ResidenceTime

!!!!!!!!!!!!!!!!!!! Timestep  !!!!!!!!!!!
    if ( keyword == 'Timestep' ) then

        if ( key == 'variable' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % itimestep = lvalue
            CYCLE NEXT
        endif

        if ( key == 'dt1' ) then
            call read_value ( Line_Buffer, dvalue )
            p_list % dt1 = dvalue
            CYCLE NEXT
        endif

        if ( key == 'swd1' ) then
            call read_value ( Line_Buffer, dvalue )
            p_list % swd1 = dvalue
            CYCLE NEXT
        endif

        if ( key == 'dt2' ) then
            call read_value ( Line_Buffer, dvalue )
            p_list % dt2 = dvalue
            CYCLE NEXT
        endif

        if ( key == 'swd2' ) then
            call read_value ( Line_Buffer, dvalue )
            p_list % swd2 = dvalue
            CYCLE NEXT
        endif

        ! never used in the code
        !if ( key == 'rswd' ) then
        !    call read_value ( Line_Buffer, dvalue )
        !    p_list % rswd = dvalue
        !    CYCLE NEXT
        !endif

        write(*,*)
        write(*,*) 'WARNING Miss reading in Timestep with the key ', key
        write(*,*)
        local_been_read = .false.

    end if !! end if group timestep

!!!!!!!!!!!!!!!!! RateCalculation  !!!!!!!!!!!!!!!!!!!
    if ( keyword == 'RateCalculation' ) then

        if ( key == 'nwin' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % nwin = ivalue
            CYCLE NEXT
        endif

        if ( key == 'nb_contact' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % nb_contact = ivalue
            CYCLE NEXT
        endif

        if ( key == 'win0' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % win0 = fvalue
            CYCLE NEXT
        endif

        if ( key == 'dwin' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % dwin = fvalue
            CYCLE NEXT
        endif

        if ( key == 'bootstrap' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % ibootstrap = lvalue
            CYCLE NEXT
        endif

        if ( key == 'fpt' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % ifpt = lvalue
            CYCLE NEXT
        endif

        write(*,*)
        write(*,*) 'WARNING Miss reading in RateCalculation with the key ', key
        write(*,*)
        local_been_read = .false.

    end if !! end if group rate calculation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!ReactionCriteria  !!!!!!!!!!!!!!!!!!!!!
    if ( keyword == 'ReactionCriteria' ) then

        if ( key == 'rxna12f' ) then
            call read_value ( Line_Buffer, cvalue )
            p_list % rxna12f = cvalue
            CYCLE NEXT
        endif

        if ( key == 'et_sol1' ) then
            call read_value ( Line_Buffer, cvalue )
            p_list % et1f = cvalue
            CYCLE NEXT
        endif

        if ( key == 'et_sol2' ) then
            call read_value ( Line_Buffer, cvalue )
            p_list % et2f = cvalue
            CYCLE NEXT
        endif

        if ( key == 'nnnons' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % nnnons = ivalue
            CYCLE NEXT
        endif

        if ( key == 'nwrec' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % nwrec = ivalue
            CYCLE NEXT
        endif

        if ( key == 'dind' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % dind = fvalue
            CYCLE NEXT
        endif

        if ( key == 'computation' ) then
            call read_value ( Line_Buffer, cvalue )
            p_list % computation = cvalue
            CYCLE NEXT
        endif

        write(*,*)
        write(*,*) 'WARNING Miss reading in ReactionCriteria with the key ', key
        write(*,*)
        local_been_read = .false.

    end if !! end if group reaction criteria
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!! Record:  Complexes and trajectories !!!!!!!!!!!!!!
    if ( keyword == 'Complexes' ) then

        if ( key == 'fcomplexes' ) then
            call read_value ( Line_Buffer, cvalue )
            p_list % fcomplexes = cvalue
            CYCLE NEXT
        endif

        if ( key == 'ftrajectories' ) then
            call read_value ( Line_Buffer, cvalue )
            p_list % ftrajectories = cvalue
            CYCLE NEXT
        endif

        if ( key == 'nb_complexes' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % ncmx = ivalue
            CYCLE NEXT
        endif

        if ( key == 'max_array_traj' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % max_array_traj = ivalue
            CYCLE NEXT
        endif

        if ( key == 'ntraj_rec' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % ntraj_rec = ivalue
            CYCLE NEXT
        endif

        if ( key == 'rmsd_min' ) then
            call read_value ( Line_Buffer, dvalue )
            p_list % rmsd_min = dvalue
            CYCLE NEXT
        endif

        if ( key == 'ionerun' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % ionerun = lvalue
            CYCLE NEXT
        endif

! ***Printout frequency
        if ( key == 'freq_print' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % freq_print = ivalue
            CYCLE NEXT
        endif

        if ( key == 'binary_complex' ) then
            call read_value ( Line_Buffer, lvalue )
            if ( lvalue .eqv. .true. ) then
                p_list % format_ascii_complexe = .false.
            end if
            CYCLE NEXT
        endif

        if ( key == 'binary_trajectory' ) then
            call read_value ( Line_Buffer, lvalue )
            if ( lvalue .eqv. .true. ) then
                p_list % format_ascii_trajectory = .false.
            end if
            CYCLE NEXT
        endif

        if ( key == 'merge_step' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % merge_step = ivalue
            CYCLE NEXT
        endif

        if ( key == 'size_thread' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % size_thread = ivalue
            CYCLE NEXT
        endif

        if ( key == "complex_sum_energy" ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % compl_sum_nrj = lvalue
            CYCLE NEXT
        end if

        if ( key == "trajectory_sum_energy" ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % traj_sum_nrj = lvalue
            CYCLE NEXT
        end if

        !! new added,
        if ( key == "restart_complex" ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % rst_complex = lvalue
            CYCLE NEXT
        end if


        write(*,*)
        write(*,*) 'WARNING Miss reading in Complexes with the key ', key
        write(*,*)
        local_been_read = .false.

    end if !! end if group complexes

!!!!!!!!!!!!!!!!! for analytical interactions
    if ( keyword == 'Analytical' ) then

        ! always set to true if the group is found
        p_list % ianalytic = .true.

        if ( key == 'debyeh' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % debyeh = lvalue
            CYCLE NEXT
        endif

        if ( key == 'ionic_strength' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % ionic = fvalue
            CYCLE NEXT
        end if

        if ( key == 'hydrodynamic' ) then
            call read_value ( Line_Buffer, lvalue )
            p_list % hydro = lvalue
            CYCLE NEXT
        end if

        if ( key == 'h_analytic' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % h_analytic = fvalue
            CYCLE NEXT
        end if

        if ( key == 'lvol_rcut' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % lvol_rcut = fvalue
            CYCLE NEXT
        end if

        write(*,*)
        write(*,*) 'WARNING Miss reading in Analytical with the key ', key
        write(*,*)
        local_been_read = .false.
    end if ! group analytical

!!!!!!!!!!!!!!!!!  For MetalDesolvation

    if ( keyword == 'MetalDesolvation' ) then

        if ( key == 'distance_to_surface' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % Zadw = fvalue
            CYCLE NEXT
        endif

        if ( key == 'energy_per_water' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % Eadw = fvalue
            CYCLE NEXT
        endif

        if ( key == 'radius_patch' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % Radw = fvalue
            CYCLE NEXT
        endif

        if ( key == 'radius_water' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % Aadw = fvalue
            CYCLE NEXT
        endif

    end if !! end if group MetalDesolvation

!!!!!!!!!!!!!!!!!  For PMF

    if ( keyword == 'PMF' ) then

        if ( key == 'mentth' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % mentth = ivalue
            CYCLE NEXT
        endif

        if ( key == 'mentfi' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % mentfi = ivalue
            CYCLE NEXT
        endif

        if ( key == 'mentom' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % mentom = ivalue
            CYCLE NEXT
        endif

        if ( key == 'mentx' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % mentx  = ivalue
            CYCLE NEXT
        endif

        if ( key == 'menty' ) then
            call read_value ( Line_Buffer, ivalue )
            p_list % menty  = ivalue
            CYCLE NEXT
        endif

        if ( key == 'pstart' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % pstart = fvalue
            CYCLE NEXT
        endif

        if ( key == 'pfinish' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % pfinish = fvalue
            CYCLE NEXT
        endif

        if ( key == 'dz' ) then
            call read_value ( Line_Buffer, fvalue )
            p_list % dZ = fvalue
            CYCLE NEXT
        endif

    end if !! end if group PMF

    if  ( Line_Buffer == 'END GROUP' ) stop_read = .true.

    end do NEXT
    !write(*,*) 'loop finished'

    ! should be true if never miss read
    been_read = local_been_read2

end subroutine read_group

!> Extract the keyword from a line
!!
!! Split every line X = Y\n
!! Xould extend to X = Y1 Y2 Y3
!! @param string : the full line
!! @param key : retrun the extracted keyword
subroutine extract_key ( key, string )
    implicit none
    character( 25 ), intent ( out ) :: key
    character( 200 ), intent ( in ) :: string

    integer :: pos1, pos2

    pos1 = 1
    pos2 = INDEX ( string ( pos1: ), "=" )
    key = trim ( adjustl ( string ( pos1:pos1+pos2-2) ) )

end subroutine extract_key

!> Read a single parameter
!!
!! Group in one function the reading of a single parameter, not in a group\n
!! do not actually read anything, just assign the correct value
!! @param key : keyword of the parameter
!! @param string : full line
!! @param inp_p_list : instance of \ref mod_input_parameter::param_list "param_list"
!! @param param_probe : instance of \ref com_sda::probe_type "probe_type"
!! @param been_read : used for detecting misspelling
subroutine read_single_parameter ( key, string, inp_p_list, param_probe, been_read )
    USE com_sda
    USE mod_interface
    USE mod_input_parameter
    implicit none
    character( 25 ), intent(in) :: key
    character( 200 ), intent(in) :: string
    type ( param_list ) :: inp_p_list
    type ( probe_type ) :: param_probe
    logical :: been_read

    character ( 128 ) :: cvalue
    real ( kind=8 ) :: dvalue
    real ( kind=4 ) :: fvalue
    integer :: ivalue
    logical :: lvalue

    !write(*,*) 'Entry read single key ', trim(key)
    !write(*,*) 'string ',trim(string)

    if ( key ==  'dseed' ) then
        !write(*,*) 'will read dvalue in dseed ', trim(string)
        call read_value ( string, dvalue )
        inp_p_list % dseed = dvalue
        ! now can initialze the random generator, moved in calling function
        been_read = .true.
    endif

! ***Number of steps
    if ( key == 'nrun' ) then
        call read_value ( string, ivalue )
        inp_p_list % nrun = ivalue
        been_read = .true.
    endif

! *** timemax to discuss for sda ( timehit )
    if ( key == 'timemax' ) then
        call read_value ( string, dvalue )
        inp_p_list % timemax = dvalue
        been_read = .true.
    endif

! *** only for sda_koff
    if ( key == 'account_occurency' ) then
        call read_value ( string, lvalue )
        !write (*,*) 'read param acc occureny' , lvalue
        inp_p_list % account_occurency = lvalue
        been_read = .true.
    endif

! **** Probes, assign directly because need for prepare proteins
    if ( key == 'probep') then
        call read_value ( string, fvalue )
        inp_p_list % probep = fvalue
        param_probe % probep = fvalue
        been_read = .true.
    endif

    if ( key == 'probew' ) then
        call read_value ( string, fvalue )
        inp_p_list % probew = fvalue
        param_probe % probew = fvalue
        been_read = .true.
    endif

    if ( key == 'hexcl' ) then
        call read_value ( string, fvalue )
        inp_p_list % hexcl = fvalue
        !! could belong to sogrid
        param_probe % hexclusion = fvalue
        been_read = .true.
    endif

!***Thresold for accessibility. calculation
    if ( key == 'threshold' ) then
        call read_value ( string, fvalue )
        inp_p_list % thresold = fvalue
        param_probe % thresold = fvalue
        been_read = .true.
    endif

!***Option save_access
    if ( key == 'save_access' ) then
        call read_value ( string, lvalue )
        inp_p_list % save_access = lvalue
        param_probe % save_access = lvalue
        been_read = .true.
    endif

    if ( key == 'save_exclusion' ) then
        call read_value ( string, ivalue )
        inp_p_list % save_exclusion = ivalue
        param_probe % save_excl = ivalue
        been_read = .true.
    endif

!!!!!!! Boost and overlap
    if ( key == 'rboost' ) then
        call read_value ( string, fvalue )
        inp_p_list % rboost = fvalue
        been_read = .true.
    endif

    if ( key == 'novers' ) then
        call read_value ( string, ivalue )
        inp_p_list % novers = ivalue
        been_read = .true.
    endif

!!!!!!!!!!!! Related to grid, multiplicative factor for all grids
    if ( key == 'epfct' ) then
        call read_value ( string, dvalue )
        inp_p_list % epfct = dvalue
        been_read = .true.
    endif

    if ( key == 'edfct' ) then
        call read_value ( string, dvalue )
        inp_p_list % edfct = dvalue
        been_read = .true.
    endif

    if ( key == 'hdfct' ) then
        call read_value ( string, dvalue )
        inp_p_list % hdfct = dvalue
        been_read = .true.
    endif

    if ( key == 'lj_rep_fct' ) then
        call read_value ( string, dvalue )
        inp_p_list % ljrepfct = dvalue
        been_read = .true.
    endif

    if ( key == 'ljfct' ) then
        call read_value ( string, dvalue )
        inp_p_list % ljfct = dvalue
        been_read = .true.
    endif

    if ( key == 'restart' ) then
        call read_value ( string, cvalue )
        inp_p_list % restart_filename = cvalue
        inp_p_list % irestart = .true.
        been_read = .true.
    endif

    ! for image charge, want to desactivate prometcs default, to associate to param_force_energy
    if ( key == 'correction_image_charge' ) then
        call read_value ( string, fvalue )
        inp_p_list % epsilon_scale = fvalue
        been_read = .true.
    end if

    if ( key == 'desolv_image_charge' ) then
        call read_value ( string, ivalue )
        inp_p_list % desolv_img_chg = ivalue
        been_read = .true.
    end if

    !do not check been_reaad
    if ( been_read .eqv. .false. ) then
        write(*,*)
        write(*,*) 'WARNING been read is false with key ', trim(key)
        write(*,*)
    end if

end subroutine read_single_parameter

!!> functions to read input, may be called indirectly by interface read_value in mod_interface.f90
subroutine read_float ( buffer, fvalue )
    implicit none
    CHARACTER( 200 ) :: buffer
    character( 128) :: value
    real ( kind = 4 ) :: fvalue
    integer pos1,pos2

    pos1 = 1
    pos2 = INDEX(buffer(pos1:), "=")
    value = trim(adjustl(buffer(pos2+1:)))
    read ( value, * ) fvalue

end subroutine read_float

! >Read double value in inputfile
subroutine read_double(buffer,dvalue)
    implicit none
    CHARACTER( 200 ) :: buffer
    character( 128) :: value
    real*8 dvalue
    integer pos1,pos2

    pos1 = 1
    pos2 = INDEX(buffer(pos1:), "=")
    value = trim(adjustl(buffer(pos2+1:)))
    read( value, * ) dvalue
end subroutine read_double

! >Read integer value in inputfile
subroutine read_int(buffer,ivalue)
    implicit none
    CHARACTER( 200 ) :: buffer
    character( 128 ) :: value
    integer ivalue
    integer pos1,pos2
    pos1 = 1
    pos2 = INDEX(buffer(pos1:), "=")
    value = trim(adjustl(buffer(pos2+1:)))
    read( value, * ) ivalue

end subroutine read_int

!> Read character value in inputfile
subroutine read_char(buffer,cvalue)
    implicit none
    CHARACTER( 200 ) :: buffer
    character( 128 ) :: cvalue,value
    integer pos1,pos2

    character*10 :: str_format

    pos1 = 1
    pos2 = INDEX(buffer(pos1:), "=")
    value = trim(adjustl(buffer(pos2+1:)))
    !write(*,*) 'pos1 pos2 ', pos1, pos2
    !write(*,*) 'value ', value

    ! problem here to read ../toto will read only .., solved with str_format
    write(str_format, '("(",i2,"a)")' ) len(trim(value))
    !write(*,*) 'str_format ', str_format
    read( value,str_format ) cvalue
    !write(*,*) 'cvalue ', cvalue

end subroutine read_char

subroutine read_logical ( buffer, lvalue )
    implicit none
    CHARACTER( 200 ) :: buffer
    character( 128 ) :: value
    logical :: lvalue
    integer pos1,pos2

    lvalue = .false.
    pos1 = 1
    pos2 = INDEX(buffer(pos1:), "=")
    value = trim(adjustl(buffer(pos2+1:)))

    if ( value .eq. '1' ) then
        lvalue = .true.
    else if ( value .eq. '0' ) then
        lvalue = .false.
    else
        write(*,*) 'probelms reading boolean value in input file'
        stop 1
    end if

end subroutine read_logical

!> Read a variable number of strings.\n
!! used only by lennard-jones at hte moment, could extend to box...
!!
!! old bug : crash with portland compiler pgfortran corrected, interface was missing "allocatable"
subroutine read_array_char( buffer,cvalue )
    implicit none
    CHARACTER( 200 ) :: buffer
    !character( 128 ) :: cvalue,value
    character, dimension(:), allocatable :: cvalue*(*)

    integer :: ierr, pos1, pos2, tmp_fields, counter
    character*200 :: tmp_value

    !temporary variable
    logical :: to_allocate, do_not_insert

    to_allocate = .false.
    do_not_insert = .false.

    !write(*,*)
    !write(*,*) 'Entry read_array_char'

    call count_fields( buffer, tmp_fields )

    if ( allocated ( cvalue ) ) then

        !write(*,*) 'cvalue already allocated'
        !write(*,*) 'size of cvalue entry', size(cvalue)

        ! check consistency
        if ( tmp_fields /= size(cvalue) ) then
             !write(*,*) ' Your input is different of the allocated size, need to reallocate !!'
             if ( tmp_fields > size(cvalue) ) then
                 write(*,*) 'read_group_file, line 1076 '
                 write(*,*) 'Cannot reallocate a non allocatable array, check your code !'
                 stop 1

             else
                write(*,*) 'WARNING In read_array_char, the array is provided with a larger size '
                write(*,*) 'WARNING data will fit, but unexpected results may appear '
             end if
             to_allocate = .false.

        else
            !write(*,*) 'allocated to the correct size '
            to_allocate = .false.
        end if

    ! else need to allocate for sure
    else
        !write(*,*) 'not allocated will allocate on the fly '
        to_allocate = .true.
   end if

   if ( to_allocate .eqv. .true. ) then
        !write(*,*) ' Whatever before allocate with size tmp_fields', tmp_fields
        allocate ( cvalue ( tmp_fields ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR in allocating cvalue '
            stop 1
        end if
   end if

   ! extract line after the equal sign
   pos1 = 1
   pos2 = INDEX(buffer(pos1:), "=")
   tmp_value = trim(adjustl(buffer(pos2+1:)))
   !write(*,*) 'tmp_value:!!',trim(tmp_value),'!!'

    ! now read, know the maximum
    pos1 = 1
    !! here should have at least one number
    pos2 = SCAN( trim(tmp_value(pos1:)) ," " )
    !write(*,*) 'first pos2 ', pos2

    ! case of only one value
    if ( pos2 == 0 ) then
        !write(*,*) 'first pos2 counter == 0'
        cvalue( 1 ) = trim(adjustl( tmp_value(pos1:len(tmp_value)) ) )
    end if
    counter = 1

LOOP: do while ( pos2 /= 0 )

        !write(*,*) 'pos1 pos2 ', pos1, pos2
        !if ( pos1 + 1 /= pos1+pos2 - 1 ) then

        cvalue( counter ) = trim(adjustl(tmp_value(pos1:pos1+pos2-1) ) )
        !write(*,*) 'register !',trim(cvalue(counter)),'!!'

        pos2 = SCAN( trim(tmp_value(pos1:)) ," " )
        !write(*,*) 'pos1 pos2 ', pos1, pos2

        ! test if two blanks
        if ( pos2 == 1 ) then
            write(*,*) 'WARNING double blank characters in the input line : ', buffer
            stop 1
            !pos1 = pos1+1
            !pos2 = 1
            !pos2 = SCAN( trim(tmp_value(pos1:)) ," " )
            !CYCLE LOOP
        end if

        pos1 = pos1+pos2
        counter = counter + 1


    end do LOOP


end subroutine read_array_char

!! test variable array
subroutine count_fields( buffer, counter )
    IMPLICIT NONE
    CHARACTER( 200 ), intent ( in ) :: buffer
    integer, intent ( out ) :: counter

    character*200 :: tmp_value
    integer :: pos1, pos2

    ! extract first to equal sign
    pos1 = 1
    pos2 = INDEX(buffer(pos1:), "=")
    tmp_value = trim(adjustl(buffer(pos2+1:)))
    !write(*,*) 'tmp_value:!!',trim(tmp_value),'!!'

    if ( len( trim(tmp_value) ) == 0 ) then
       write(*,*) 'ERROR In input line :', trim ( buffer )
       stop 1
   endif

    !! just count, do not assign values
    counter = 0
    pos1 = 1
    !! here should have at least one number
    pos2 = SCAN( trim(tmp_value(pos1:)), " " )
    !write(*,*) 'pos2 ', pos2
    ! needed if only one value
    if ( pos2 == 0 ) then
        counter = 1
    end if

LOOP: do while ( pos2 /= 0 )

        !write(*,*) 'pos2 ', pos2

        pos2 = SCAN( trim(tmp_value(pos1:)) ," " )
        !write(*,*) 'pos2 ', pos2

        if ( pos2 == 1 ) then
            !write(*,*) 'Warning : double blank characters in the input line : ', buffer
            !stop 1
            pos1 = pos1+1
            CYCLE LOOP
        end if

        pos1 = pos1+pos2
        counter = counter + 1

    end do LOOP

    !write(*,*) 'count_fields exit counter ', counter

end subroutine count_fields


