!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file read_input_file.f90
!! Main function to read sda input file

!> Function to read the input file and allocate all proteins, grids, flexible and all other variables
!!
!! Print summary informations about parameters
!!
!! Simple variable read by \ref read_single_parameter :\n
!!      - number of runs, timemax, probe, scale factor for grids...
!!
!! Use of groups/block for complexe modules, read by \ref read_group :
!!      - No maximum predefined number of Solute_Grid groups
!!      - No maximum predefined Flexible list and proteins
!!      - Could be extended to more than one reaction criteria ??
!!
!! Inconvenient:
!!      - Order of some GROUPs are important
!!
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param rcriteria : instance of \ref mod_reactioncriteria::react_criter "react_criter"
!! @param resid_time : instance of \ref mod_residencetime::residence_time "residence_time"
!! @param o_complexes, trajectories, p_restart : instance of \ref mod_record::record "record"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param param_timestep : instance of \ref com_sda::parameter_timestep "parameter_timestep"
!! @param param_probe : instance of \ref com_sda::probe_type "probe_type"
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param param_analytic : definition of analytical interaction, instance of \ref com_sda::parameter_analytic "parameter_analytic"
!! @param param_metaldesolv : instance of \ref com_sda::parameter_metaldesolv "parameter_metaldesolv"
!! @param option_omp : used for OpenMP
!! @param timemax : maximum time for each trajectory, 0 is infinite
!! @param rboost : distance when boost
!! @param nrun : number of trajectories
!! @param nmax_overlap : number of overlap before applying a boost
!! @param account_occurency : if occurences must be accounted
subroutine read_input_file ( tab_protein, rcriteria, resid_time, o_complexes, trajectories, geom, p_restart, &
                             type_calc, param_timestep, param_probe, param_force_energy, param_analytic, param_metaldesolv, &
                             param_pmf, option_omp, timemax, rboost, nrun, nmax_overlap, &
                             account_occurency )

    USE mod_array_protein
    USE mod_ReactionCriteria
    USE mod_ResidenceTime
    USE mod_type_calc
    USE mod_geometry
    USE mod_record
    !! Include input parameters, the module contains all variables to be read.
    !! Theses variables are then copied to the variables or modules,
    !! and then never used directly except in read_input_file and preration of proteins and grids.
    !! much easier to change the input file independently of the modules and ohter functions.
    USE mod_input_parameter
    USE mod_interface
    USE mod_flexibility
    USE mod_force_energy
    USE com_sda

    IMPLICIT NONE
      interface
          subroutine prepare_grid ( pdb_filename, q_filename, ep_filename, ed_filename, hd_filename, &
                                    lj_rep_filename, lj_filename, atom_type_name, &
                                    epfct, edfct, hdfct, lj_rep_fct, ljfct, param_analytic, dm, dr, stoke_rad, &
                                    param_probe, bool_excl_grid, psgrid, center,  &
                                    counter_grid, counter_conf, rcriter, iflex, isurface, &
                                    type_calc, nb_lj_opt )
              USE mod_setofgrid
              USE mod_ReactionCriteria
              USE com_sda
              USE mod_type_calc
              IMPLICIT NONE
              ! will return an allocated psgrid
              type ( sogrid ), pointer :: psgrid
              type ( react_criter ) :: rcriter
              type ( type_calculation ) :: type_calc
              type ( parameter_analytic ) :: param_analytic
              character ( 128 ), intent ( in ) :: pdb_filename, q_filename, ep_filename, ed_filename, hd_filename
              character ( 128 ), intent ( in ) :: lj_rep_filename
              character*128, dimension(:), allocatable, intent( in ):: lj_filename
              character*2, dimension(:), allocatable, intent( in ):: atom_type_name
              real ( kind=8 ), intent ( in ) :: epfct, edfct, hdfct, lj_rep_fct, ljfct
              real ( kind=8 ), intent ( in ) :: dm, dr, stoke_rad
              type ( probe_type ) :: param_probe
              real ( kind=8 ), dimension(3) :: center
              integer, intent ( in ) :: counter_grid, counter_conf
              logical :: bool_excl_grid, iflex, isurface

              integer, optional :: nb_lj_opt
          end subroutine prepare_grid

          subroutine read_list_conf ( plist_flex, input_param, counter_grid, filename, param_probe, &
                                      param_analytic, tmp_exclud, center, rcriteria, type_calc, been_read )
              USE mod_input_parameter
              USE mod_ReactionCriteria
              USE mod_flexibility
              USE mod_type_calc
              USE com_sda
              IMPLICIT NONE
              type ( list_flex_type ) :: plist_flex
              type ( param_list ) :: input_param
              type ( type_calculation ) :: type_calc
              integer :: counter_grid
              character ( 128 ) :: filename
              type ( probe_type ) :: param_probe
              type ( parameter_analytic ) :: param_analytic
              logical :: tmp_exclud, been_read
              real ( kind=8 ), dimension(3) :: center
              type ( react_criter ) :: rcriteria
          end subroutine read_list_conf

          subroutine read_restart ( p_record, tab_prot, type_calc, filename, io_restart, account_occur, &
                                    opt_nb_types )
              USE mod_record
              USE mod_type_calc
              IMPLICIT NONE
              type ( record ), pointer :: p_record
              type ( type_calculation ) :: type_calc
              type ( array_protein_type ), target :: tab_prot
              character (128) :: filename
              integer :: io_restart
              logical :: account_occur
              integer, optional, intent ( in ) :: opt_nb_types
          end subroutine read_restart

          subroutine restart_origin_conf ( p_record, tab_prot )
              USE mod_record
              USE mod_type_calc
              IMPLICIT NONE
              type ( record ), pointer :: p_record
              type ( array_protein_type ), target :: tab_prot
          end subroutine restart_origin_conf

          subroutine initialize_bits ( o_record, tab_prot )
              USE mod_array_protein
              USE mod_record
              USE com_sda
              IMPLICIT NONE
              type ( record ) :: o_record
              type ( array_protein_type ), target :: tab_prot
          end subroutine initialize_bits

          subroutine initialize_random_generator ( type_calc, dseed )
              USE mod_type_calc
              !$  USE omp_lib
              !$  USE Par_Zig_mod
              IMPLICIT NONE
              type ( type_calculation ) :: type_calc
              real ( kind=8 ) :: dseed
          end subroutine initialize_random_generator

          subroutine check_timestep( geom, param_timestep, tab_protein )
              USE mod_array_protein
              USE mod_geometry
              USE com_sda
              IMPLICIT NONE
              type ( geometry ) :: geom
              type ( parameter_timestep ) :: param_timestep
              type ( array_protein_type ) :: tab_protein
          end subroutine check_timestep

    end interface

    type ( array_protein_type ) :: tab_protein
    ! now list_flex points to rcriteria, made make a pointer to avoid to pass as arguments
    ! could be mod_array_protein, more general ?, but good for sdamm if not flexible
    ! or more general use, 2 reaction criteria..
    ! change make rcriteria independent
    type ( react_criter ), target :: rcriteria
    type ( residence_time ) :: resid_time

    type ( record ) :: o_complexes, trajectories

    type ( geometry ) :: geom
    ! pointer on record for restart, for sdamm could be only in read restart function,
    !            for sda_energy ?? read once or read during the main loop ?? need to send back. To adapt later
    type ( record ), pointer :: p_restart
    type ( type_calculation ) :: type_calc
    type ( parameter_timestep ) :: param_timestep
    type ( probe_type ) :: param_probe
    type ( type_force_energy ) :: param_force_energy
    type ( parameter_analytic ) :: param_analytic
    type ( parameter_metaldesolv ) :: param_metaldesolv
    type ( parameter_pmf ) :: param_pmf
    type ( type_option_omp ) :: option_omp
    real ( kind = 4 ) :: rboost
    real ( kind=8 ), intent ( out ) :: timemax
    integer :: nrun, nmax_overlap
    logical :: account_occurency, opt_box

    ! defined in module mod_input_parameter
    type ( param_list ) :: Inp_Par_List

    ! temporary variable, they can be used multiple times. interest of pointers !!
    real ( kind=8 ), dimension(3) :: center
    ! need temporary pointer on grid, should be associated by a protein
    type ( sogrid ), pointer :: psgrid
    ! pointer on list_flex, associated by a pflexible. Allocated only if necessary.
    type ( list_flex_type ), pointer :: plist_flex
    ! temp pointer to flexible, associated by each flexible solute
    type ( flexible ), pointer :: pflexible
    ! need to know the relative diffusion coefficient before calling allocate_rate
    real ( kind=8 ) :: sum_diffusion

    ! used for reading input file
    integer :: Line_Read_Count, nreadin, io_status, ierr
    character( 200 ) :: Line_Buffer
    character( 25 ) :: key
    character(128) :: cvalue
    logical :: tmp_exclud, been_read

    ! count the number of created proteins and grid
    ! id is assigned to each protein, not used yet
    integer :: count_id, count_grid, i, count_mol_types !, k
    ! warning if write directly string
    character ( 128 ) :: title_complexe, title_trajectory
    integer :: status
    ! to clean for restart complexe
    character ( 15 ) :: format_file

    ! default value, could be in default ?? it is in default witth ''
    title_complexe = 'complexe'
    title_trajectory = 'trajectory'
    ! error by default
    status = 1
    ! nullify the pointer, safer if tested
    nullify ( psgrid, pflexible, plist_flex )
    ! to setup additional box data in trajectory with sdamm
    opt_box = .false.
    sum_diffusion = 0.d0
    ! init to off by default
    ! could be done in main
    o_complexes % type_record = off_record
    trajectories % type_record = off_record
    ! init should, may be done by caller
    rcriteria % type_reaction = off_rc
    nullify ( o_complexes % pcomplexe, o_complexes % ptrajectory )
    nullify ( trajectories % pcomplexe, trajectories % ptrajectory )

    ! Initialise counters
    Line_Read_Count = 0
    nreadin = 0
    io_status = 0
    ! counter for protein id, grid , and grid_flex ( ? )
    count_id = 0
    count_grid = 0
    count_mol_types = 0

    ! comput grid exclusion, false by default
    tmp_exclud = .false.
    key = ''

    ! initialize with default values, cannot do in a type declaration
    call set_default_values ( Inp_Par_List )

    ! assign default value to param_probe, needed for prepare proteins and grid
    ! SHOULD BE overwritten before GROUP Solute_Grid is read, to have effect
    param_probe % probep = Inp_Par_List % probep
    param_probe % probew = Inp_Par_List % probew
    param_probe % thresold = Inp_Par_List % thresold
    param_probe % hexclusion = Inp_Par_List % hexcl
    param_probe % save_access = Inp_Par_List % save_access
    param_probe % save_excl = Inp_Par_List % save_exclusion
    !param_probe % exclusion_dtgrid = .false. ! no need copy here ??

    ! load default value, false and 0.150 Molar, space 0.01 A
    param_analytic % ianalytic = Inp_Par_List % ianalytic
    param_analytic % h_analytic = Inp_Par_List % h_analytic
    param_analytic % debyeh = Inp_Par_List % debyeh
    param_analytic % ionic = Inp_Par_List % ionic
    param_analytic % hydro = Inp_Par_List % hydro
    param_analytic % lvol_rcut = Inp_Par_List % lvol_rcut

    ! same for metaldesolv ?? no not needed !
    ! default value, only for Inp_Par_List, here needed
    param_metaldesolv % imetaldesolv = .false.
    param_pmf % ipmf = .false.

    ! to have default value, false even if no group Timestep
    param_timestep % variable = Inp_Par_List % itimestep

    Line_Buffer = ''
    read ( inputf, '(a)', iostat=io_status ) Line_Buffer
    if ( io_status < 0 ) then
        write( io6, * ) 'input file is missing or empty'
    endif

! read all lines from the input file
BEGIN:  &
    do while ( io_status == 0 )
        ! -- Increment line read count
        Line_Read_Count = Line_Read_Count + 1

        if ( ( Line_Buffer(1:1) == '#' ) .or. ( Line_Buffer(1:1) == '!' ) .or. &
             ( len_trim ( Line_Buffer ) == 0)  ) then
            read ( inputf, '(a)', iostat=io_status ) Line_Buffer
            CYCLE BEGIN
        end if

        ! to check that a keyword has been effectively read
        been_read = .false.
        nreadin = nreadin + 1
        call extract_key ( key, Line_Buffer )

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! ***   Check first for GROUP
        if ( key == 'GROUP' ) then

          call read_value ( Line_Buffer, cvalue )
          ! write(*,*) 'read GROUP ', trim(cvalue)
          call read_group ( Inp_Par_List, cvalue, inputf, been_read )

!!!!!!!!!! make the group initialization !!!!!!!!

          if ( cvalue == "Type_Calculation" ) then
              !write(*,*) 'Group Type_Calculation ', Inp_Par_List % total_solutes, Inp_Par_List % total_grids
              type_calc % enum_type_calc = Inp_Par_List % t_calc % enum_type_calc
              type_calc % type_calc_str = get_string ( type_calc )

              ! allocate array_protein, with new set_of_protein
              call new_array_protein ( tab_protein, Inp_Par_List % total_solutes, Inp_Par_List % total_grids )

          ! Solute Grid need to be initialized during the reading, not sure for others, may be in a separate function
          else if ( cvalue == 'Solute_Grid' ) then
              ! counter for grids, count 1 for sogrid or list_flex, whatever flex or not
              count_grid = count_grid + 1

#ifdef DEBUG
              write(*,*) 'Solute_Grid GROUP'
              write(*,*) 'count_grid ', count_grid
              write(*,*) 'Inp stoke ', Inp_Par_List % stoke_rad
#endif

              ! by default only the first protein will have an exclusion grid. count_id incremented at the end
              ! sda_2proteins, need the sum of diffusion trans. to allocate rate_calculation
              if ( type_calc % enum_type_calc /= sdamm ) then
                  !write(*,*) 'count_id  == 0,  set tmp_exclud to true '
                  if ( count_id == 0 ) tmp_exclud = .true.
                  sum_diffusion = sum_diffusion + Inp_Par_List % dm
                  !write(*,*) 'sda_2proteins update sum_diffusion ', sum_diffusion
#ifdef DEBUG
                  write(*,*) 'tmp_exclud ', tmp_exclud
                  write(*,*) 'sda_2proteins update sum_diffusion ', sum_diffusion
#endif
              end if

              ! Update set_of_protein
              ! first grid begin at 1
              if ( count_grid == 1 ) then
                  tab_protein % set_of_protein ( count_grid , 1 ) = 1
              else
                  tab_protein % set_of_protein ( count_grid , 1 ) = tab_protein % set_of_protein ( count_grid - 1 , 2 ) + 1
              end if

              tab_protein % set_of_protein ( count_grid , 2 ) = &
                                  tab_protein % set_of_protein ( count_grid , 1 ) + Inp_Par_list % nb_solute - 1
              !write(*,*) 'update set_of_protein count_grid ', count_grid, &
              !            tab_protein % set_of_protein ( count_grid, 1 ), tab_protein % set_of_protein ( count_grid, 2 )

              ! test fpar and list_conf, should read the correct grid, like no flex but get correct input filename
              if ( ( Inp_Par_List % list_conf_name /= '' ) .and. ( type_calc % bool_par ) ) then
                  !write (*,*) 'list_conf_name ', Inp_Par_List % list_conf_name
                  write (*,*)
                  write (*,*) 'fparallel and list conf given, flex ', Inp_Par_List % iflex, trim ( Inp_Par_List % list_conf_name )
                  if ( Inp_Par_List % iflex .eqv. .true. ) then
                      write(*,*) 'Cannot use fparallel option with flex = 1'
                      stop 1
                  end if

                  ! here plist_flex is not created, nor associated, and not used in read_list_conf
                  ! keep as argument for format
                  call read_list_conf ( plist_flex, Inp_Par_List, count_grid, Inp_Par_List % list_conf_name, param_probe, &
                                        param_analytic, tmp_exclud, center, rcriteria, type_calc, been_read )
                  !! test dynamically allocated
                  !if ( allocated ( Inp_Par_List % ljf ) ) then
                  !  write(*,*) 'ljf allocated '
                    !write(*,*) 'after read_list_conf', Inp_Par_List % ljf
                  !end if
              end if

              ! allocate proteins and grid, after having read the GROUP Solute_Grid
              ! if not flexible
              if ( Inp_Par_List % iflex .eqv. .false. ) then
                  !write(*,*) 'iflex false, before prepare_grid '
                  !write(*,*) 'input surface ', Inp_Par_List % isurf
                  !write(*,*) 'Inp % stoke_rad ', Inp_Par_List % stoke_rad

                  if ( ( Inp_Par_List % list_conf_name /= '' ) .and. ( type_calc % bool_par .eqv. .false. ) ) then
                      write(*,*)
                      write(*,*) 'ERROR Given list conf in input, but flex = 0 => incompatible '
                      write(*,*) 'ERROR Include filenames directly in sda.in, or try with option -fparrallel 1'
                      write(*,*)
                      stop 1
                  endif

                  ! only one set of grid to prepare by group, exclusion grid set correctly before
                  ! load grids and pdb, effective charges, compute atom surface ...
                  ! added rcriteria, if sda_2proteins default
                  call prepare_grid ( Inp_Par_List % pdb_filename, Inp_Par_List % qef, Inp_Par_List % epf, &
                          Inp_Par_List % edf, Inp_Par_List % hdf, Inp_Par_List % lj_repf, Inp_Par_List % ljf, &
                          Inp_Par_List % atom_type_lj, &
                          Inp_Par_List % epfct, Inp_Par_List % edfct, Inp_Par_List % hdfct, &
                          Inp_Par_List % ljrepfct, Inp_Par_List % ljfct, param_analytic, & !Inp_Par_List % param_analytic, &
                          Inp_Par_List % dm, Inp_Par_List % dr, Inp_Par_List % stoke_rad, &
                          param_probe, tmp_exclud , psgrid, center, count_grid, 1, &
                          rcriteria, Inp_Par_List % iflex, Inp_Par_List % isurf, type_calc, Inp_Par_List % nb_lenjones )

                  ! should send rcriteira to preparation grid, needed when reading pdb and saving position of reaction criteria

                  ! test should be associated to a set of grid
                  if ( associated ( psgrid ) .eqv. .false. ) then
                      write (*,*) 'ERROR psgrid not associated '
                      stop 1
                  !else
                  !    write (*,*) 'prepare_grid ok'
                  end if

                  ! many identical proteins, create them and set theirs pointers ( psogrid ) of each to the same set of grid
                  ! use count_id equivalent to position in array, may change later. Id could be the type..
                  !write(*,*) 'will create not flexible ', Inp_Par_List % nb_solute, ' solutes '
                  !write(*,*) 'with param_metaldesolv % imetaldesolv ', param_metaldesolv % imetaldesolv
                  do i = 1, Inp_Par_List % nb_solute
                      call new_protein ( tab_protein % array ( count_id + i ), &
                                         Inp_Par_List % irot, Inp_Par_List % isurf, Inp_Par_List % iflex, &
                                         Inp_Par_List % iimg_chg, count_id + i, param_metaldesolv % imetaldesolv )

                      ! makes association to the grid
                      !write(*,*) 'create new protein cound_id+i= ', count_id+i
                      tab_protein % array ( count_id + i ) % p_sogrid => psgrid
                      tab_protein % array ( count_id + i ) % center_protein = center
                      !write(*,*) 'test associated psgrid ', associated ( tab_protein % array ( count_id + i ) % p_sogrid )

                  end do

                  ! reset psgrid before creating new sogrid, should be done if flexible also
                  nullify ( psgrid )
                  ! no need to reset allocatable arrays also
                  if ( allocated ( Inp_Par_List % ljf) )  deallocate ( Inp_Par_List % ljf )
                  if ( allocated ( Inp_Par_List % atom_type_lj ) ) deallocate ( Inp_Par_List % atom_type_lj )

              ! else if iflex == true
              else
                  !write(*,*) 'iflex true new_list_flex and read_list_conf tmp_exclud=', tmp_exclud
                  !if ( Inp_Par_List % debyeh .eqv. .true. ) then
                  !    write(*,*) 'INFO Flex and Debye-H, not really tested'
                  !end if

                  ! Create only one list_flex by GROUP if solutes are flexible
                  ! All solutes of this GROUP will point to it, through an independent flexible class
                  allocate ( plist_flex, stat = ierr )
                  if ( ierr /= 0 ) then
                      write(*,*) 'Cannot allocate plist_flex'
                      stop 1
                  end if

                  call new_list_flex ( plist_flex, Inp_Par_List % flex_method, Inp_Par_List % list_conf_name, &
                                       Inp_Par_List % total_conf, &
                                       Inp_Par_List % initial_conf, Inp_Par_List % nearest_ngb, &
                                       Inp_Par_List % frequence, Inp_Par_List % std_freq ) ! &
                                       !Inp_Par_List % init_nb_conf, !Inp_Par_List % max_conf )

                  ! read the additional input file for flexible in an external file, e.g., list_conf.dat
                  ! needed to load the grids, for each conf allocate the sogrid and insert in the list_flexible
                  !write(*,*) 'will read list conf ', Inp_Par_List % list_conf_name
                  call read_list_conf ( plist_flex, Inp_Par_List, count_grid, Inp_Par_List % list_conf_name, param_probe, &
                                        param_analytic, tmp_exclud, center, rcriteria, type_calc, been_read )

                  ! need to initalize the proteins, could be common to flex or not, before testing iflex : more general
                  do i = 1, Inp_Par_List % nb_solute
                      call new_protein ( tab_protein % array ( count_id + i ), &
                                         Inp_Par_List % irot, Inp_Par_List % isurf, Inp_Par_List % iflex, &
                                         Inp_Par_List % iimg_chg, count_id + i, param_metaldesolv % imetaldesolv )

                      tab_protein % array ( count_id + i ) % center_protein = center

                      ! allocate pflexible and  by default points to the first node of list_flexible in fact : list_flex % p_first
                      call set_flexible ( tab_protein % array ( count_id + i ), plist_flex )
                      ! assign p_sogrid to the first list_flex, needed during initilisation of complexes
                      ! could assign to p_first, no data, use now for initialize_bits, could change

                      ! not logical here, or points to flexible the first position,
                      ! but necessary for init_movable, seen in tool
                      tab_protein % array ( count_id + i ) % p_sogrid => plist_flex % p_first % p_next % p_data
                  end do

              endif ! end if flex

              !write(*,*) 'reset solute param whatever exclud, flex or fpar'

              ! reset exclusion grid if actived
              tmp_exclud = .false.
              ! Reset default values for reading of next Solute grid, why not in reset function ?
              Inp_Par_List % iflex = .false.
              !Inp_Par_List % init_nb_conf  = 0
              Inp_Par_List % total_conf  = 0
              !Inp_Par_List % max_conf  = 0
              ! new case of parallel reset list_conf_name
              Inp_Par_List % list_conf_name = ''
              Inp_Par_List % nearest_ngb = .false.

              ! increment counter protein, whatever flex or not
              count_id = count_id + Inp_Par_List % nb_solute
              ! increase the counter for different types of solutes - used for SDAMM
              count_mol_types = count_mol_types + 1

              ! new added for all input filename, could group all reset with flex into this function
              call reset_default_solutegrid( Inp_Par_List )

          ! end if protein_grid
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          else if ( cvalue == 'Analytical' ) then
            !write(*,*) 'Read analytical group ionic ', Inp_Par_List % ionic
            !write(*,*) 'Read analytical group hydro ', Inp_Par_List % hydro

            ! activates analytic boolean for debye-H / later for crowder in solute-grids
            ! add hydrodynamics as well, apply to all
            param_analytic % ianalytic = .true.
            param_analytic % h_analytic = Inp_Par_List % h_analytic
            param_analytic % debyeh = Inp_Par_List % debyeh
            param_analytic % ionic = Inp_Par_List % ionic
            param_analytic % hydro = Inp_Par_List % hydro
            param_analytic % lvol_rcut = Inp_Par_List % lvol_rcut
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

          else if ( cvalue == 'Geometry' ) then
              ! need to setup the type sphere/ box first, before calling the initialisation
              geom % type = Inp_Par_List % type_geom

              call init_geometry( geom, Inp_Par_List % pbc, Inp_Par_List % escape, Inp_Par_List % surf_geom, &
                                  Inp_Par_List % c, Inp_Par_List % xmin, Inp_Par_List % xmax, &
                                  Inp_Par_List % ymin, Inp_Par_List % ymax, Inp_Par_List % zmin, Inp_Par_List % zmax, &
                                  Inp_Par_List % start_pos )

          ! end if geometry
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

          else if ( cvalue == 'Timestep' ) then

              ! initialize param_timestep, variable initailised to false (default) if no group
              param_timestep % variable = Inp_Par_List % itimestep
              param_timestep % dt1 = Inp_Par_List % dt1
              param_timestep % swd1 = Inp_Par_List % swd1
              param_timestep % dt2 = Inp_Par_List % dt2
              param_timestep % swd2 = Inp_Par_List % swd2
              ! finalize intialisation of parameter_timestep
              param_timestep % dt2f = ( param_timestep % dt2 - &
                                        param_timestep % dt1) / &
                 ( param_timestep % swd2 - param_timestep % swd1 )

             if ( param_timestep % swd1 > param_timestep % swd2 ) then
                write(*,*) 'ERROR swd1 > swd2 '
                flush(6)
                stop 1
            endif

            if ( param_timestep % dt1 > param_timestep % dt2 ) then
                write(*,*) 'ERROR dt1 > dt2 '
                flush(6)
                stop 1
            endif

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

          else if ( cvalue == 'RateCalculation' ) then

              ! Should check than reaction criteria has already been allocated, order is important in iput file
              ! Valid if electron_trabsfer as well ?
              if ( (rcriteria % type_reaction /= assoc_rate ) .and. (rcriteria % type_reaction /= elec_transfer ) ) then
                  write(*,*)
                  write(*,*)'ReactionCriteria type is incompatible with rate calculation', rcriteria % type_reaction
                  write(*,*)'Or order of GROUP is wrong in input file', &
                            'GROUP ReactionCriteria should ALWAYS be before GROUP RateCalculation '
                  write(*,*)'Rate_Calculation not allocated'

              !else
              !    write(*,*) 'will allocate rate_calculation after initialization of geometry'
              end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

          else if ( cvalue == 'ReactionCriteria' ) then

              ! here added variable computation
              ! easier != rc_off
              if ( ( Inp_Par_List % computation == 'association' ) .or. ( Inp_Par_List % computation == 'docking' ) .or. &
                   ( Inp_Par_List % computation == 'electron_transfer' ) .or. ( Inp_Par_List % computation == 'all' ) ) then

                  ! second version, make a real allocation, easy to change if bound to mod_array or so_grid
                  call allocate_react_criter ( rcriteria, Inp_Par_List % computation, Inp_Par_List % rxna12f, &
                                               Inp_Par_List % et1f, Inp_Par_List % et2f, &
                                               Inp_Par_List % nnnons, Inp_Par_List % nwrec )

                  ! simple test for electron_transfer and dind
                  if ( ( rcriteria % type_reaction == elec_transfer ) .and. ( Inp_Par_List % dind > 0. ) ) then
                    write(*,*) 'ERROR Electron_transfer cannot be done with dind > 0.'
                    stop 1
                  end if

                  if ( ( Inp_Par_List % dind > 0. ) .and. ( rcriteria % type_reaction /= all ) ) then
                      call make_independent( rcriteria, Inp_Par_List % dind )
                  end if

                  ! write into log file, for all not necessary
                  if ( ( type_calc % nb_par == 1 ) .and. ( rcriteria % type_reaction /= all ) ) then
                      !write(*,*) 'test write log nb_par ', type_calc % nb_par
                      !write(*,*) 'test write log type_reaction ', rcriteria % type_reaction
                      call write_to_logfile ( rcriteria, iurxna )
                  end if
              ! only for test, test could be in allocate, and stop there
              else
                  write(*,*) 'this reaction criteria not yet implemented: ', trim ( Inp_Par_List % computation )
                  !      rcriteria % type_reaction = off_rc
                  stop 1
              endif


          !endif ! end if reaction criteria
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          else if ( cvalue == 'ResidenceTime' ) then
              !write(*,*) 'read group ResidenceTime'
              !write(*,*) 'format of resid_3d', Inp_Par_List % format_rt3d

              call init_resid_time ( resid_time, Inp_Par_List % fixe_bin_size, &
                                     Inp_Par_List % enum_str_residtime, Inp_Par_List % format_rt3d, &
                                     Inp_Par_List % pair_proteins,  Inp_Par_List % total_grids )

              ! best is to split in two
              if ( resid_time % enum_rt /= rt_off ) then

                  ! general for sda_2proteins
                  if ( type_calc % enum_type_calc /= sdamm ) then
                     call set_array_rt ( resid_time, geom, &
                          Inp_Par_List % nb_cell_x, Inp_Par_List % nb_cell_y, Inp_Par_List % nb_cell_z, &
                          Inp_Par_List %  size_cell_x, Inp_Par_List %  size_cell_y, Inp_Par_List %  size_cell_z )

                  ! only for sdamm, only rdf implemented
                  else
                     call set_array_rt_sdamm ( resid_time, geom, Inp_Par_List % total_grids, &
                          Inp_Par_List % nb_cell_x, Inp_Par_List % nb_cell_y, Inp_Par_List % nb_cell_z, &
                          Inp_Par_List %  size_cell_x, Inp_Par_List %  size_cell_y, Inp_Par_List %  size_cell_z )
                  end if
              end if

              !! then apply shift to the origin
              if ( ( Inp_Par_List % shift_rt_x /= 0. ) .or. ( Inp_Par_List % shift_rt_y /= 0. ) .or. &
                   ( Inp_Par_List % shift_rt_z /= 0. ) ) then
                  !write(*,*) 'Aplly shift for residence time '
                  call apply_shift_to_rt ( resid_time, Inp_Par_List % shift_rt_x, &
                                           Inp_Par_List % shift_rt_y, Inp_Par_List % shift_rt_z )
             ! else
             !   write(*,*) 'Do not apply shift for residence time'

              end if

          ! complexes and trajectotries, could be in separate GROUP = Record
          else if ( cvalue == 'Complexes' ) then

              ! allocate one complexe
              ! if fparralel modify name
              ! in fact no test needed if str_nb_par = '' change nothing !!
              if ( len_trim ( Inp_Par_List %  fcomplexes ) > 0 ) then

                  !write(*,*) 'allocate complexe'
                  if ( type_calc % bool_par ) then
                  !    write(*,*) 'parrallel run nb', type_calc % nb_par
                  !    if ( type_calc % nb_par > 1 ) then
                  !        write (*,*) 'nb_par > 1, modify name with str_nb_par', str_nb_par
                          !title_complexe = title_complexe // adjustl ( str_nb_par )
                          call add_nbpar ( Inp_Par_List % fcomplexes, type_calc % nb_par )
                          write (*,*) 'new filename for complexe ', trim ( Inp_Par_List % fcomplexes )
                  !  endif
                  end if


                  ! no restart, create a new empty complex
                  if ( Inp_Par_List % rst_complex .eqv. .false. ) then

                    ! default value for complexes: first false = opt_sum (replaced compl_sum_nrj )
                    ! second true = one_conf
                    call new_record ( o_complexes, title_complexe, type_calc, Inp_Par_List % ncmx, &
                                    Inp_Par_List % fcomplexes, iocmx, Inp_Par_List % compl_sum_nrj, opt_box, &
                                    Inp_Par_List % ntraj_rec, Inp_Par_List % freq_print, &
                                    Inp_Par_List % rmsd_min, Inp_Par_List % ionerun, .true., &
                                    Inp_Par_List % format_ascii_complexe, Inp_Par_List % total_grids )

                    ! initialize bit_energy and integer, function defined only in this file
                    call initialize_bits ( o_complexes, tab_protein )

                    ! could be initialized after. If change nb_max_complexe, for sda_energy e.g., introduce no limit
                    ! finally initialize elements, not so much to do for complexes because it is a list
                    ! sda_2proteins
                    if ( type_calc % enum_type_calc /= sdamm ) then
                        call initialize_record ( o_complexes, tab_protein )
                    ! sdamm limited to record between 2 types of proteins
                    else
                        call initialize_record ( o_complexes, tab_protein, opt_nb_protein = 2 )
                    end if

                  ! else get all information from the restart file
                  else
                      write(*,*) 'INFO Ask for restart_complexe '

                      ! similar to read_record
                      call init_record ( o_complexes, opt_iocmx = iocmx )

                      call read_header_record ( o_complexes, Inp_Par_List % fcomplexes ) ! optional, cm, nb_line )

                      ! initialize bit_energy and integer, function defined only in this file
                      call initialize_bits ( o_complexes, tab_protein )
                      ! finish to initialize to check if can mix
                      call initialize_record ( o_complexes, tab_protein, opt_read = .true. )
                      ! still need some minimum input, other taken from the input
                      o_complexes % nb_max_complexe = Inp_Par_List % ncmx
                      o_complexes % pcomplexe % rmsd_min = Inp_Par_List % rmsd_min
                      o_complexes % pcomplexe % one_run = Inp_Par_List % ionerun

                      ! fill the list complexe,
                      ! need to retransform back to get temporary values <En^2>
                      call read_file_record ( o_complexes, opt_rebuild_sum_square = .true. )
                      !write(*,*) 'after read_file_record'
                      !write(*,*) 'nb_max_complexes ', o_complexes % nb_max_complexe
                      !write(*,*) 'nb_complexes ', o_complexes % pcomplexe % nb_complexe

                      ! close file and reopen in write mode ??
                      close ( o_complexes % io_compl )

                      ! ascii, formatted
                      if ( o_complexes % format_ascii ) then
                        format_file = 'formatted'
                      ! binary, unformatted
                      else
                        format_file = 'unformatted'
                      end if

                      !write(*,*) 'format_file:',format_file,'!!!'
                      open ( o_complexes % io_compl, file= Inp_Par_List % fcomplexes, form= format_file, &
                             status='replace', iostat = ierr )
                      if ( ierr /= 0 ) then
                        write(*,*) 'WARNING Cannot open file: ', trim ( Inp_Par_List % fcomplexes )
                      endif

                  endif ! restart_complex

                  ! save option for OpenMP, should give default value if not given
                  option_omp % merge_complexe = Inp_Par_List % merge_step
                  option_omp % maxsize_thread_complexe = Inp_Par_List % size_thread

                  ! if option_omp not given, make default values
!$                if ( option_omp % merge_complexe == 0 ) then
!$                     option_omp % merge_complexe = 50
!$                end if
!$                ! 500 complexe, not too long compute can be a default value
!$                if ( option_omp % maxsize_thread_complexe == 0 ) then
!$                    if ( Inp_Par_List % ncmx > 5000 ) then
!$                        option_omp % maxsize_thread_complexe = Inp_Par_List % ncmx / 10
!$                    else if ( Inp_Par_List % ncmx <= 500 ) then
!$                        option_omp % maxsize_thread_complexe = 200
!$                    else
!$                        option_omp % maxsize_thread_complexe = 500
!$                    end if
!$                end if
              end if

              ! allocate one trajectory
              if ( len_trim ( Inp_Par_List % ftrajectories ) > 0 ) then
                  !write(*,*) 'allocate trajectories'
                  if ( type_calc % bool_par ) then
                      call add_nbpar ( Inp_Par_List % ftrajectories, type_calc % nb_par )
                      write (*,*) 'new filename for trajectory ', trim ( Inp_Par_List % ftrajectories )
                  end if

                  ! if not modified, use size of tab_protein by default
                  if ( Inp_Par_List % max_array_traj == 0 ) &
                       Inp_Par_List % max_array_traj = size ( tab_protein % array )

                  ! if sdamm, add box information by default, may extend for sda with surface ??
                  if ( type_calc % enum_type_calc == sdamm ) then
                      opt_box = .true.
                      !write(*,*) 'read input set opt_box to true'
                  end if

                  ! opt_sum is true by default for trajectory, one_conf true
                  call new_record ( trajectories, title_trajectory, type_calc, Inp_Par_List % max_array_traj, &
                                    Inp_Par_List % ftrajectories, iocmxn, Inp_Par_List % traj_sum_nrj, opt_box, &
                                    Inp_Par_List % ntraj_rec, Inp_Par_List % freq_print, &
                                    Inp_Par_List % rmsd_min, Inp_Par_List % ionerun, .true., &
                                    Inp_Par_List % format_ascii_trajectory, Inp_Par_List % total_grids )

                  ! initialize bit_energy and integer, similar to complexes
                  call initialize_bits ( trajectories, tab_protein )
                  ! finalize in initialising the elements, create array and init onecomplexe
                  call initialize_record ( trajectories, tab_protein )
              end if

          ! Metal desolvation
          else if ( cvalue == 'MetalDesolvation' ) then

                write(*,*)
                write(*,*) 'WARNING MetalDesolvation is activated '
                write(*,*) 'WARNING It has only been recently reimplemented in SDA 7 and is still under test'
                write(*,*) 'WARNING Particularly the addition of the force for this term is desactivated'
                write(*,*) 'WARNING You can acitivate the force, uncomment MainLoop_sda_2proteins_omp.f90:817'
                !write(*,*) 'Inp List ', Inp_Par_List % Zadw
                write(*,*)

                param_metaldesolv % imetaldesolv = .true.
                param_metaldesolv %  Zadw =  Inp_Par_List % Zadw
                param_metaldesolv %  Eadw =  Inp_Par_List % Eadw
                param_metaldesolv %  Radw =  Inp_Par_List % Radw
                param_metaldesolv %  Aadw =  Inp_Par_List % Aadw
          ! PMF
          else if ( cvalue == 'PMF' ) then

                write(*,*)
                write(*,*) ''
                write(*,*) ''
                write(*,*) ''
                write(*,*) ''
                write(*,*)

                param_pmf % ipmf = .true.
                param_pmf %  mentth =  Inp_Par_List % mentth
                param_pmf %  mentfi =  Inp_Par_List % mentfi
                param_pmf %  mentom =  Inp_Par_List % mentom
                param_pmf %  mentx =  Inp_Par_List % mentx
                param_pmf %  menty =  Inp_Par_List % menty
                param_pmf %  pstart =  Inp_Par_List % pstart
                param_pmf %  pfinish =  Inp_Par_List % pfinish
                param_pmf %  dZ =  Inp_Par_List % dZ

          ! if error in input
          else
              write(*,*) 'ERROR Name of the GROUP ', trim ( cvalue ), ' is unknow,  abort'
              stop 1
          end if ! cvalue

        end if !!! end if GROUP
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        ! if still false, try to read a single parameter
        if ( ( been_read .eqv. .false. ) .and. ( key /= "GROUP" ) ) then
            call read_single_parameter( key, Line_Buffer, Inp_Par_List, param_probe, been_read )
        end if

        ! check if something has been read
        ! except for GROUP because warning in read group, a possible mistake in group is checked before
        if ( ( been_read .eqv. .false. ) .and. ( key /= "GROUP" ) ) then
            write(*,*)
            write(*,*) 'WARNING line ', Line_Read_Count,' was not read, check your spelling'
            write(*,*) 'WARNING Line in question ', trim(Line_Buffer)
            write(*,*)
        end if

        if ( key == 'dseed' ) then
            if (  type_calc % enum_type_calc == 0 ) then
                write(*,*) 'ERROR Type Calculation has not yet been read when reading dseed '
                stop 1
            end if
            !write(*,*) 'will call random generator dseed ', Inp_Par_List % dseed
            call initialize_random_generator ( type_calc, Inp_Par_List % dseed )
        end if

        read ( inputf, '(a)', iostat=io_status ) Line_Buffer

    end do BEGIN ! end input file loop

    ! Here needed for default values not used by other groups
    timemax = Inp_Par_List % timemax
    account_occurency = Inp_Par_List % account_occurency
    rboost  = Inp_Par_List % rboost
    nrun    = Inp_Par_List % nrun
    nmax_overlap = Inp_Par_List % novers

    ! assign initial values to param_force_energy
    param_force_energy % epsilon_scale = Inp_Par_List % epsilon_scale
    ! here use the value of the input or the default which is zero
    param_force_energy % img_chg_desolv_input = Inp_Par_List % desolv_img_chg

    if  ( ( type_calc % enum_type_calc == sdamm ) .and. ( Inp_Par_List % irestart .eqv. .false. ) ) then
         write (*,*)
         write (*,*) 'ERROR You forgot a restart file with sdamm !'
         stop 1
    end if

    if ( Inp_Par_List % irestart ) then
        !write (*,*)
        !write (*,*) 'Read restart file ', trim ( Inp_Par_List % restart_filename )
        call read_restart ( p_restart, tab_protein, type_calc, Inp_Par_List % restart_filename, &
                            io_restart,  Inp_Par_List % account_occurency, opt_nb_types = count_mol_types )
        !write(*,*) 'after read_restart'
        !do i = 1, Inp_Par_List % total_proteins
        !    write(*,*) 'read rest pos ', (tab_protein % array ( i ) % position_relative( k ),k=1,3)
        !end do

    else if ( ( type_calc % enum_type_calc == sda_koff ) .or. &
              ( type_calc % enum_type_calc == sda_energy ) ) then
         write (*,*)
         write (*,*) 'INFO No restart file provided, will use the pdb configuration '
         !!in restart.f90, only for sda_koff or energy
         call restart_origin_conf ( p_restart, tab_protein )
    end if

    ! Initialize movable_atoms after restart,
    ! function in this file, will make correct initialisation of movable atoms depending of type % calculation
    call initialize_movable_atoms ( type_calc, tab_protein )

    ! all should be fully initialized at this point, protein associated to grid
    ! can do better
    if ( param_metaldesolv % imetaldesolv .eqv. .true. ) then
        do i = 1, size ( tab_protein % array )
            call allocate_array_mdesolv ( tab_protein % array(i), param_metaldesolv)
        end do
    end if

!!!!!!!!!!!!!!! Print infos from the input file !!!!!!!!!!!!!!!!!!!!!!!!
    write (*,*)
    write(*,*) '=========================================================================================='
    write (*,*) 'Type of calculation : ', type_calc % enum_type_calc, type_calc % type_calc_str

    if ( trim(Inp_Par_List % flex_method)=='') then
        write(*,*) 'No flexibility'
    else
        write(*,*) 'Flexibility method  : ', trim(Inp_Par_List % flex_method)
    endif

    if ( ( type_calc % enum_type_calc == sdamm ) .and. ( nrun /= 1 ) ) then
        write(*,*) 'INFO sdamm => nrun can only be set up to 1'
        nrun = 1
    end if

    ! never exclusion in sdamm
    if ( ( type_calc % enum_type_calc == sdamm ) .and. ( Inp_Par_List % save_exclusion > 0 ) )then
        write(*,*) 'INFO sdamm never uses exclusion grid, desactivates save_exclusion'
        param_probe % save_excl = 0
        Inp_Par_List % save_exclusion  = 0
        !param_probe % exclusion_dtgrid = .false.
    end if

    write ( io6, 601 ) Inp_Par_List % dseed, timemax, nrun, &
                       param_probe % probep, param_probe % probew, param_probe % thresold, &
                       param_probe % save_excl
  601 FORMAT (/,'INPUT DATA:',/,/,'random number generator:',f20.3,/, &
              'timemax                 ',f20.3,/, &
              'number of runs:         ',i20,/, &
              'probe radius protein:   ',f20.3,/, &
              'probe radius water:     ',f20.3,/, &
              'accessibility threshold:',f20.3,/, &
              'save_exclusion         :',i20 )

    ! Print analytical group
    if ( param_analytic % ianalytic .eqv. .true. ) then
        write(*,*)
        write(*,*) ' Analytic'
        write( io6, fmt='("bin size of precomputed arrays (A)",f8.3)' ), Inp_Par_List % h_analytic

        if ( param_analytic % debyeh .eqv. .true. ) then
            ! if debye, test if electrostatic grid
            if ( Inp_Par_List % epfct == 0. ) then
                write(*,*) 'ERROR Debye-Huckel activated but electrostatic grid not present(epfct = 0)'
                STOP 1
            end if

            write(*,fmt='("Debye-Huckel activated, will apply to all solutes")' )
            write(*,fmt='("Ionic Strength (Molar) :",f8.3)' ), Inp_Par_List % ionic
        else
             write(*,fmt='("No Debye-Huckel")')
        end if

        if ( param_analytic % hydro .eqv. .true. ) then
            write(*,fmt='("Hydrodynamics interaction are activated, apply only to sdamm")' )
           
            write (*,'("Radius lvol_rcut to define local volume for hydrodynamic interactions = ",f8.3)') , Inp_Par_List%lvol_rcut

            ! Check that a multiplicative factor has been defined for local volume radius
            if (param_analytic%lvol_rcut.LE.0.d0) then
               write (*,'("ERROR: lvol_rcut must be larger than zero with hydrodynamics calculation")')
               stop 1
            end if
            
        else
            write(*,fmt='("No hydrodynamics")' )
        end if
    endif

    ! Print timestep info, uninitialized if sda_pmf (valgrind error)
    if ( ( type_calc % enum_type_calc /= sda_pmf ) .and. ( type_calc % enum_type_calc /= sda_energy ) ) then
        write( io6, *)
        if ( type_calc % enum_type_calc /= sdamm ) then
            write ( io6, 6011 ) rboost ,nmax_overlap, &
                     param_timestep % dt1 , param_timestep % swd1 , &
                     param_timestep % dt2 , param_timestep % swd2
        else
            write( io6, fmt= '("timestep                 :",f8.3)') param_timestep % dt1
        end if
    endif

 6011 FORMAT('boost value:               ',f8.3,/, &
             'no of tries before boost:  ',i8,/, &
             'smallest timestep:         ',f8.3,' ps',/, &
             '  before c-t-c separation: ',f8.3,' A',/, &
             'largest timestep:          ',f8.3,' ps',/, &
             '  at c-t-c separation:     ',f8.3,' A' )


! need test if param_metaldesolv activated
6012 FORMAT(/,'       Metal Desolvation activated',/,&
            'distance_to_surface:        ',f8.3,' A',/, &
            'energy_per_water:           ',f8.3,/, &
            'radius_patch:               ',f8.3,' A',/, &
            'radius_water:               ',f8.3,' A')

    if ( param_metaldesolv % imetaldesolv ) then
        write(io6, 6012) param_metaldesolv % Zadw, param_metaldesolv % Eadw, &
                         param_metaldesolv % Radw, param_metaldesolv % Aadw
        write(io6,*)
    end if

    ! only sda with 2 proteins normally
    if ( param_timestep % variable ) then
        call check_timestep( geom, param_timestep, tab_protein )
    end if

    if ( param_pmf % ipmf .eqv. .true. ) then
        write(*,*)
        write(*,*) "  PMF input parameters "
        write ( *, 700 ) param_pmf %  mentth, param_pmf %  mentfi, param_pmf %  mentom, &
                         param_pmf %  mentx, param_pmf %  menty, param_pmf %  pstart, &
                         param_pmf %  pfinish, param_pmf %  dZ
    endif

  700 FORMAT ('mentth          ',i10,/, &
              'mentfi          ',i10,/, &
              'mentom          ',i10,/, &
              'mentx           ',i10,/, &
              'menty           ',i10,/, &
              'pstart          ',f10.3,/, &
              'pfinish         ',f10.3,/,&
              'dZ              ',f10.3)

    if ( associated ( trajectories % ptrajectory ) ) then
        write ( *, * )
        write ( *, * ) '  Trajectory '
        write ( *, 702 ) trajectories % ntraj_rec, trajectories % nb_max_complexe, trajectories % opt_sum, &
                           trajectories % freq_print, trajectories % format_ascii
    end if

    !! Info Complexe
    if ( associated ( o_complexes % pcomplexe ) ) then
        write ( *, * )
        write ( *, * ) '  Complexes '

        if ( Inp_Par_List % rst_complex .eqv. .true. ) then
            write(*,*)
            write(*,*)'Use a restart complex, filename :', Inp_Par_List % fcomplexes
            write(*,*)'with nb_complexes               :', o_complexes % pcomplexe % nb_complexe
            write(*,*)
            write(*,*)'be sure to use a different seed number (check test gaus e.g.)'
            write(*,*)
        end if

        write ( *, 701 ) o_complexes % nb_max_complexe, o_complexes % pcomplexe % one_run, &
              o_complexes % pcomplexe % one_conf, o_complexes % pcomplexe % rmsd_min, &
              o_complexes % opt_sum, &
              option_omp % merge_complexe, option_omp % maxsize_thread_complexe, &
              o_complexes % format_ascii
    end if

    write ( *, * )
    if ( associated ( o_complexes % pcomplexe ) ) then
        call print_bits ( o_complexes )
    else if ( associated ( trajectories % ptrajectory ) ) then
        call print_bits ( trajectories )
    end if

  701 FORMAT ('nb_complexes max            ',i10,/, &
              'rmsds checked for each run  ',b10,/,&
              'rmsds checked for each conf ',b10,/,&
              'rmsd_min                    ',d10.3,/,&
              'opt sum                     ',b10,/,&
              'merge_step for thread (Only OpenMP) ',i10,/, &
              'max size for thread   (Only OpenMP) ',i10,/, &
              'format ascii complexe       ',b10)

  702 FORMAT ('record trajectory number    ',i10,/, &
              'max size array              ',i10,/, &
              'opt sum                     ',b10,/,&
              'frequency output            ',i10,/,&
              'format ascii trajectory     ',b10)

    !write ( io6, * )
    if ( associated ( geom % pshere ) ) then
        write ( io6, *) 'Geometry:    sphere'
        write ( io6, 602 ) geom % pshere % b_surface, geom % pshere % c_surface

    else if ( associated ( geom % pbox ) ) then
        write ( io6, *) 'Geometry:       box'
        write ( io6, 603 ) geom % pbox % x_min, geom % pbox % x_max, &
                           geom % pbox % y_min, geom % pbox % y_max, &
                           geom % pbox % z_min, geom % pbox % z_max
    endif
    ! common to geometry
    write ( io6, fmt='("start position:          ",f8.3)' ) geom % start_position
    write ( io6, fmt='("escape position:         ",f8.3)' ) geom % stop_position
    write ( io6, fmt='("PBC:                     ",b1)' ) geom % pbc
    write ( io6, fmt='("escape                   ",b1)' ) geom % escape
    write ( io6, fmt='("contain surface:         ",b1)' ) geom % surface

  602 FORMAT ('b-surface:               ',f8.3,/, &
              'c-surface:               ',f8.3)
  603 FORMAT ('xmin:                    ',f8.3,/, &
              'xmax:                    ',f8.3,/,&
              'ymin:                    ',f8.3,/, &
              'ymax:                    ',f8.3,/,&
              'zmin:                    ',f8.3,/, &
              'zmax:                    ',f8.3)


    ! test DH and hydrodynamic are not implemented yet in the case of surface
    if ( (geom % surface .eqv. .true.) .and. (param_analytic % debyeh .or. param_analytic % hydro ) ) then
        write(*,*) 'ERROR Cannot (yet) use hydrodynamics interaction or debye-huckel with a surface'
        stop 1
    end if

!!!!!!!! Info Residence time
    if ( resid_time % enum_rt /= rt_off ) then
        write (*,*)
        write (*,*) '   ResidenceTime'
        write ( io6, 803 ) resid_time % enum_rt, trim ( Inp_Par_List % enum_str_residtime ), resid_time % fixed_bin_size, &
                           resid_time % xcell, resid_time % ycell, resid_time % zcell, &
                           resid_time % nb_xcell, resid_time % nb_ycell, resid_time % nb_zcell, &
                           Inp_Par_List % shift_rt_x, Inp_Par_List % shift_rt_y, Inp_Par_List % shift_rt_z, &
                           resid_time % x_origin, resid_time % y_origin, resid_time % z_origin
    end if
  803 FORMAT ('residence time     :                 ',i8,/,&
              'rt string          :                 ',(a),/,&
              'fixed_bin_size     :                 ',i8,/,&
              'cell size ( A )    :                 ',3f8.3,/,&
              'cell number        :                 ',3i8,/,&
              'opt. shift applied to the origin :   ',3f10.3,/,&
              'origin of the grid (after shift) :   ',3f10.3 )

    if ( rcriteria % type_reaction /= off_rc ) then
        write ( io6, *)
        write ( io6, 801 ) rcriteria % type_reaction, trim ( Inp_Par_List % computation ), &
                     rcriteria % ntotal, rcriteria % nspec, rcriteria % naspec, rcriteria % nnonspec, &
                     rcriteria % nnons_min, rcriteria % nwrec, rcriteria % dind

    end if ! end if rcriteria
  801 FORMAT ('reaction criteria:                   ',i8,/,&
              '                 :                   ',(a),/,&
              'total criteria   :                   ',i8,/,&
              'specific (only use in docking):      ',i8,/,&
              'aspecific (only use in docking):     ',i8,/,&
              'non specific     :                   ',i8,/,&
              'min. nonspecific :                   ',i8,/,&
              'window record (only use in assoc_rate):',i5,/,&
              'dind             :                    ',f8.3 )

    ! allocate array for rate or electron_transfer
    if ( ( rcriteria % type_reaction == assoc_rate ) .or. ( rcriteria % type_reaction == elec_transfer ) ) then

        ! elecron_transfer, always 1 contact
        if ( ( rcriteria % type_reaction == elec_transfer ) .and. ( Inp_Par_List % nb_contact /= 1 ) ) then
            write(*,*)
            write(*,*) 'electron transfer always uses 1 contact, force nb_contact to 1'
            write(*,*)
            Inp_Par_List % nb_contact = 1
        end if

        ! allocate here, becasue we know for sure about the geometry
        ! problem to set brate and crate with dm
#ifdef DEBUG
        write(*,*) 'call allocate rate with sum_diffusion ', sum_diffusion
#endif
        call allocate_rate ( rcriteria, Inp_Par_List % win0, Inp_Par_List % nwin, &
                             Inp_Par_List % dwin, Inp_Par_List % nb_contact, &
                             geom % start_position, geom % stop_position, sum_diffusion, & !Inp_Par_List % dm, &
                             !!bootsrap
                             Inp_Par_List % ibootstrap, Inp_Par_List % ifpt )

        write ( *, 802 ) rcriteria % p_assoc_rate % first_window, rcriteria % p_assoc_rate % width_window, &
                         rcriteria % p_assoc_rate % nb_window, rcriteria % p_assoc_rate % nb_contact, &
                         rcriteria % p_assoc_rate % bootstrap, rcriteria % p_assoc_rate % fpt, &
                         rcriteria % p_assoc_rate % brate, rcriteria % p_assoc_rate % crate

    end if
  802 FORMAT (/,'association rate :',/, &
              'rxn windows start at:    ',f10.3,' A',/, &
              'rxn winndow increment:   ',f10.3,' A',/, &
              'no of rxn windows:       ',i10,/, &
              'no of contact:           ',i10,/, &
              'bootstrap:               ',b1,/, &
              'first passage time:      ',b1,/, &
              'b_rate:                  ',es10.3,/, &
              'c_rate:                  ',es10.3)

    ! print summary about solutes in a nice format
    call print_info_array_protein ( tab_protein )

  end subroutine read_input_file

! could add a parameter for the number of the first grid 0, 1 or -7 ?? would avoid to change the numbering, done !
! add new feature, if not flex return correct name of grid for a not flexible protein ( used by fparallel )

!> Read a list of conformations for flexible solutes
!!
!! Read an external input file\n
!! Two methods implemented :
!!  - EXPLICIT : give the names of all grids explicitely
!!  - IMPLICIT : use an automatic numerotation ( eg : pbd1.pdb, pdb2.pdb,... ), need additional parameters
!!
!! If not flexible return the correct name of the grid ( used if fparrallel )
!! @param plist_flex : instance of \ref mod_flexibility::list_flex_type "list_flex_type"
!! @param input_param : instance of \ref mod_input_parameter::param_list "param_list"
!! @param counter_grid : use as identifier for grids
!! @param filename : name of external file with the list of conformations
!! @param param_probe : instance of \ref com_sda::probe_type "probe_type"
!! @param param_analytic : structure \ref com_sda::parameter_analytic "parameter_analytic"
!! @param tmp_exclud : if exclusion must be computed
!! @param center : center of solutes ?
!! @param rcriteria : instance of \ref mod_reactioncriteria::react_criter "react_criter"
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param been_read : used for checking misspelling
subroutine read_list_conf ( plist_flex, input_param, counter_grid, filename, param_probe, param_analytic, &
                            tmp_exclud, center, rcriteria, type_calc, been_read )
    USE mod_input_parameter
    USE mod_flexibility
    USE mod_ReactionCriteria
    USE mod_type_calc
    USE com_sda
    USE mod_interface
    IMPLICIT NONE

    interface
        subroutine prepare_grid ( pdb_filename, q_filename, ep_filename, ed_filename, hd_filename, &
                                  lj_rep_filename, lj_filename, atom_type_name, &
                                  epfct, edfct, hdfct, lj_rep_fct, ljfct, param_analytic, dm, dr, stoke_rad, &
                                  param_probe, bool_excl_grid, psgrid, center, &
                                  counter_grid, counter_conf, rcriter, iflex, isurface, type_calc, nb_lj_opt )
            USE mod_setofgrid
            USE mod_ReactionCriteria
            USE com_sda
            USE mod_type_calc
            IMPLICIT NONE
            type ( sogrid ), pointer :: psgrid
            type ( react_criter ) :: rcriter
            type ( type_calculation ) :: type_calc
            type ( parameter_analytic ) :: param_analytic
            character*128, intent(in) :: pdb_filename, q_filename, ep_filename, ed_filename, hd_filename
            character*128, intent(in) :: lj_rep_filename
            character*128, dimension(:), allocatable, intent ( in ) :: lj_filename
            character*2, dimension(:), allocatable, intent ( in ) :: atom_type_name
            real ( kind=8 ), intent( in ) :: epfct, edfct, hdfct, lj_rep_fct, ljfct
            real ( kind=8 ), intent( in ) :: dm, dr, stoke_rad
            type ( probe_type ) :: param_probe
            real ( kind=8 ), dimension(3) :: center
            integer, intent ( in ) :: counter_grid, counter_conf
            logical :: bool_excl_grid, iflex, isurface
            integer, optional :: nb_lj_opt
        end subroutine

        !> write output into input, may change later
        subroutine create_filename_flex ( input_filename, counter )
            IMPLICIT NONE
            character :: input_filename*(*)
            integer :: counter
        end subroutine create_filename_flex

    end interface

    type ( list_flex_type ) :: plist_flex
    type ( param_list ) :: input_param
    type ( type_calculation ) :: type_calc
    integer :: counter_grid
    character ( 128 ) :: filename
    type ( probe_type ) :: param_probe
    type ( parameter_analytic ) :: param_analytic
    logical :: tmp_exclud, been_read
    real ( kind=8 ), dimension(3) :: center
    type ( react_criter ) :: rcriteria

    ! allocated only here
    type ( sogrid ), pointer :: psgrid

    character ( 200 ) :: string
    character ( 128 ) :: cvalue
    character ( 25 ) :: key
    integer :: io_unit = 2222, current_conf, shift_nb_par, io, ierr, status, i, loop_prep
    ! for explicit shift the first number, default 1
    logical :: ex, explicit, bool_load_flex_grid

    ! temporary p_list, do not interfer with general input
    ! use an full param_list for implicit case, avoid to copy all files names
    ! could extend later if common data with operator = , for use with default, no need yet
    type ( param_list ) :: tmp_p_list

    ! maximum distance over all confs
    real ( kind=8 ) :: distance_max

    ! for movable atom
    integer :: max_move_charge, max_move_atom

    !!because of name
    character ( 128 ) :: copy_original_pdbname
    character ( 128 ), dimension ( 6 ) :: copy_original_gridname

    nullify ( psgrid )

    copy_original_pdbname = ''
    copy_original_gridname = ''

    !! loop prep, default for explicit == 1
    loop_prep = 1
    i = 0

    !! can delete first conf, just use current_conf
    current_conf = 0
    !! use only if fparallel nb_par
    shift_nb_par = 0
    max_move_charge = 0
    max_move_atom = 0
    distance_max = 0.
    explicit = .true.

    ! new parameter for fparallel, just generates the correct name of file
    ! need a parameter for really loading grid in plist_flex
    bool_load_flex_grid = .true.

#ifdef DEBUG
    write( *, * ) 'Entry read list_conf ,exclusion', tmp_exclud
#endif

    ! default false
    status = 1

    ! check for filename
    inquire ( file=filename, exist = ex )
    !write(*,*) ' check filename ', trim ( filename ), ' ex:',ex
    if ( .not.ex ) then
        write( *,* ) 'ERROR input file is missing:', trim( filename )
        stop 1
    endif

    ! trick to use the same group than Solute_Grid when reading the group
    cvalue='Solute_Grid'

    ! initialize the default values, if the keyword is not found, use the default
    call set_default_values ( tmp_p_list )

    ! if iflex is false, do not create listflex just return the correct filename
    if ( input_param % iflex .eqv. .false. ) then
        bool_load_flex_grid = .false.
        ! assign default shift_nb_par, overwritten if not.explict by first_mode
        ! was missing if explicit
        shift_nb_par = type_calc % nb_par
#ifdef DEBUG
        write(*,*) 'in read_list_conf do not load list flexible'
        write(*,*) 'bool par ', type_calc % bool_par, ' nb_par ', type_calc %  nb_par
        write(*,*) 'shift_nb_par ', shift_nb_par
#endif
    end if

    ! reinit key for valgrind
    key=''

    ! read the file
    open (io_unit, file=filename , form= 'formatted', status='old', iostat = ierr )
    read ( io_unit, '(a)', iostat=io ) string
    if ( io < 0 ) then
        write( *,* ) 'WARNING File is empty', trim ( filename )
        return
    endif

BEGIN: &
    do while ( io == 0 )

        if ( ( string(1:1) == '#' ) .or. ( string(1:1) == '!' ) .or. &
             ( len_trim ( string ) == 0)  ) then
              read ( io_unit, '(a)', iostat=io ) string
              CYCLE BEGIN
        end if

        if ( string == 'EXPLICIT' ) then
#ifdef DEBUG
            write(*,*)
            write(*,*) 'EXPLICIT reading of list conf'
#endif
            explicit = .true.
            read ( io_unit, '(a)', iostat=io ) string
            CYCLE BEGIN
        else if  ( string == 'IMPLICIT' ) then
#ifdef DEBUG
            write(*,*)
            write(*,*) 'IMPLICIT reading of list conf'
#endif
            explicit = .false.
            read ( io_unit, '(a)', iostat=io ) string
            CYCLE BEGIN
        endif

        call extract_key ( key, string )
        !write( *,* ) 'key ', trim ( key )

        ! special case not explicit use first_mode
        if ( ( .not.explicit ) .and. ( key == 'first_mode' ) ) then
            !write (*,*) 'first_mode string', trim ( string )
            !call read_value ( string, first_mode )
            call read_value ( string, current_conf )
            !delete first_mode internal variable, use current conf, only used for printing info about grid !!
            !first_mode = first_mode - 1
            current_conf = current_conf - 1
            !write (*,*) 'assign current_conf ', current_conf

            if ( bool_load_flex_grid .eqv. .false. ) then
                !write(*,*) 'nb_par ', type_calc % nb_par
                shift_nb_par = type_calc % nb_par + current_conf
                write(*,*) 'fparallel update shift_conf', shift_nb_par
            end if

            read ( io_unit, '(a)', iostat=io ) string
            cycle begin
        end if

        if ( key == 'GROUP' ) then

            !write (*,*) 'read GROUP'

            ! first reading of group read whatever explicit implicit, fparallel
            ! forced before cvalue = SoluteGrid
            call read_group ( tmp_p_list, cvalue, io_unit, been_read )

            ! if implicit always need to rewrite tmp_p_list with original name
            ! entry only once at this point, name will be modify in the loop_prep
            if ( explicit .eqv. .false. ) then

                loop_prep = input_param % total_conf
                write(*,*) 'explicit false'
                write(*,*) 'loop prep = total_conf', loop_prep

                copy_original_pdbname = tmp_p_list % pdb_filename
                copy_original_gridname ( 1 ) = tmp_p_list % qef
                copy_original_gridname ( 2 ) = tmp_p_list % epf
                copy_original_gridname ( 3 ) = tmp_p_list % edf
                copy_original_gridname ( 4 ) = tmp_p_list % hdf
                copy_original_gridname ( 5 ) = tmp_p_list % lj_repf
                if ( allocated ( tmp_p_list % ljf ) ) then
                    write(*,*)
                    write(*,*) 'NOT IMPLEMENTED for ljf , to do later'
                    write(*,*)
                end if
                if ( allocated ( tmp_p_list % atom_type_lj ) ) then
                    write(*,*)
                    write(*,*) 'NOT IMPLEMENTED for ljf , to do later'
                    write(*,*)
                end if
                !copy_original_gridname ( 6 ) = tmp_p_list % ljf
                !write(*,*) 'assign copy origin pdbname ', copy_original_pdbname
                !write(*,*) 'assign copy origin epf ', copy_original_gridname ( 2 )

                !! could be in utils
                !write(*,*) 'implicit name change name in tmp_list i', i
                !call modify_fname_implicit ( tmp_p_list, current_conf )

            end if

            !write(*,*) 'loop_prep', loop_prep

            ! why sending current_conf ?? do not work whith rcriteria
            ! should split current_grid, current_conf
            do i = 1, loop_prep

                !! update current conf mode whatever explicit or none
                !write ( *,* ) 'explicit update in loop_prep tmp_exclud ', tmp_exclud

                ! if implicit will come back in loop, if explicit will come back every group
                current_conf = current_conf + 1
                !write ( *,* ) 'update in loop_prep current_conf, first mode', current_conf, first_mode
                !write ( *,* ) 'update in loop_prep current_conf ', current_conf, ' tmp_exclud ', tmp_exclud

                !! if implicit will no use the other initial increment, could make better
                if (  explicit .eqv. .false. ) then

                    !write (*,*) 're copy original name copy_origin ', copy_original_pdbname
                    tmp_p_list % pdb_filename = copy_original_pdbname
                    tmp_p_list % qef = copy_original_gridname ( 1 )
                    tmp_p_list % epf = copy_original_gridname ( 2 )
                    tmp_p_list % edf = copy_original_gridname ( 3 )
                    tmp_p_list % hdf = copy_original_gridname ( 4 )
                    tmp_p_list % lj_repf = copy_original_gridname ( 5 )
                    !tmp_p_list % ljf = copy_original_gridname ( 6 )
                    if ( allocated ( tmp_p_list% ljf  ) ) then
                        write(*,*)
                        write(*,*) 'NOT IMPLEMENTED for ljf , to do later'
                        write(*,*)
                    end if
                    if ( allocated ( tmp_p_list % atom_type_lj ) ) then
                        write(*,*)
                        write(*,*) 'NOT IMPLEMENTED for ljf , to do later'
                        write(*,*)
                    end if
                    !write (*,*) 'copy done ', tmp_p_list % pdb_filename
                    !write(*,*) 'before modify_fname_implicit current conf ', current_conf

                    call modify_fname_implicit ( tmp_p_list, current_conf )

                end if

                !write(*, *) 'new pdb filename ', tmp_p_list % pdb_filename
                !write(*, *) 'new ep filename ', tmp_p_list % epf

                ! load if really flex, true by default
                if ( bool_load_flex_grid ) then
                    !write(*,*) 'bool_load_flex is true prepare grid tmp_exclud ', tmp_exclud

                    !! current_conf only used for printing info
                    call prepare_grid ( tmp_p_list % pdb_filename, tmp_p_list % qef, &
                        tmp_p_list % epf, tmp_p_list % edf, tmp_p_list % hdf, tmp_p_list % lj_repf, tmp_p_list % ljf, &
                        input_param % atom_type_lj, input_param % epfct, input_param % edfct, input_param % hdfct, &
                        input_param % ljrepfct, input_param % ljfct, param_analytic, &
                        input_param % dm, input_param % dr, input_param % stoke_rad, &
                        param_probe, tmp_exclud , psgrid, center, &
                        counter_grid, current_conf, rcriteria, input_param % iflex, input_param % isurf, type_calc )

                    ! allocate a new setofgrid and stock in the list lflex, false default insert at the end
                    call insert_sogrid ( plist_flex, psgrid, .false., status )
                    ! should check status

                    !write(*,*) 'before test max ',max_move_charge,max_move_atom
                    if ( psgrid % nb_charge > max_move_charge ) then
                        max_move_charge = psgrid % nb_charge
                        !write(*,*) 'max_move_charge ', max_move_charge
                    end if

                    if ( psgrid % nb_atom_surface > max_move_atom ) then
                        max_move_atom = psgrid % nb_atom_surface
                        !write(*,*) 'max_move_atom ', max_move_atom
                    end if
                    !write(*,*) 'maximum nb of charge and atom_surface ',max_move_charge,max_move_atom

                    !write(*,*) 'before distance_max ', psgrid % distance_max, distance_max
                    if ( psgrid % distance_max > distance_max ) then
                        distance_max = psgrid % distance_max
                    end if

                ! end if bool_load_flex_grid
                end if

                ! else...
                ! here test for fparallel no flex but list_conf.dat, fparallel
                if ( bool_load_flex_grid .eqv. .false. ) then
                    !write(*,*) 'bool_load_flex_grid is false current_conf and shift_nb_par ' , current_conf, shift_nb_par
                    !if ( current_conf == type_calc % nb_par ) then
                    if ( current_conf == shift_nb_par ) then
                        write(*,*) 'found correct conformation ', tmp_p_list % pdb_filename
                        input_param % pdb_filename = tmp_p_list % pdb_filename
                        input_param % qef = tmp_p_list % qef
                        input_param % epf = tmp_p_list % epf
                        input_param % edf = tmp_p_list % edf
                        input_param % hdf = tmp_p_list % hdf
                        input_param % lj_repf = tmp_p_list % lj_repf
                        !input_param % ljf = tmp_p_list % ljf
                        if ( allocated ( tmp_p_list % ljf ) ) then
                            write(*,*)
                            write(*,*) 'NOT IMPLEMENTED for ljf , to do later'
                            write(*,*)
                        end if
                        !!factor are global no need to copy
                        !! close file before quitting
                        close ( io_unit )
                        write(*,*) 'close file because nb_par and return directly'
                        return
                    end if
                    ! need error message, cannot here loop prep == 1
                    !if ( current_conf == loop_prep ) then
                    !     write(*,*) 'Current conf is the last one, and the conformation number has not been found '
                    !     stop 1
                    !endif
                end if

              end do !! end do loop_prep

        end if ! GROUP

        read ( io_unit, '(a)', iostat=io ) string

    end do BEGIN

    ! not really flexible, so fparrallel
    !if ( bool_load_flex_grid .eqv. .false. ) then
    if ( type_calc % bool_par ) then
        write(*,*) 'ERROR You asked for fparrallel, the confomration number is too high '
        stop 1
    end if

#ifdef DEBUG
    write(*,*) 'final max_charge, max_atom ', max_move_charge, max_move_atom
#endif
    plist_flex % max_size_movable_charge = max_move_charge
    plist_flex % max_size_movable_atom = max_move_atom
    plist_flex % distance_max = distance_max

    close ( io_unit )

end subroutine read_list_conf

!> Initialize bits
!!
!! Could be in mod_record  array_prot is included now.\n
!! \todo to move to mod_record ?
!!
!! Could use param_force_energy, but bits info needed for creating array for trajectory
!! to check if possible to setup param_force_energy before any record is initialized
!!
!! Initialize bits, depending on the loaded grids ( bit_energy ) and type type calculation ( bit_integer )
!! @param o_record : instance of \ref mod_record::record
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
subroutine initialize_bits ( o_record, tab_prot )
    USE mod_array_protein
    USE mod_record
    USE com_sda

    IMPLICIT NONE
    type ( record ) :: o_record
    type ( array_protein_type ), target :: tab_prot

    type ( sogrid ), pointer :: sgrid_tmp
    integer :: i, j, prot_nb

    !write(*,*) 'Entry initialize_bits '
    !write(*,*) 'tab_prot % nb_grids ', tab_prot % nb_grids

    ! for asymetries, need param_force_energy with the correct setup
    ! now loop over all proteins
    do j = 1, tab_prot % nb_grids
        !write(*,*) 'j= ',j

        prot_nb = tab_prot % set_of_protein ( j, 1 )
        sgrid_tmp => tab_prot % array ( prot_nb ) % p_sogrid

    ! here we konw the type from the number, not sure !!
    if ( is_associated ( sgrid_tmp % pset_grid( 1 ) ) ) then
        ! electrostatic
        !write( * , * ) 'will record electrostatic energy term'
        call set_bit_energy ( o_record,get_nb_gridtype('electrostatic') - 1 )
    endif

    if ( is_associated ( sgrid_tmp % pset_grid ( 2 ) ) ) then
        ! desolvation
        ! write( *, * ) 'will record desolvation energy terms'
        call set_bit_energy ( o_record, get_nb_gridtype('desolvation') - 1 )
    endif

    if ( is_associated ( sgrid_tmp % pset_grid ( 3 ) ) ) then
        ! hydrophobic
        !write(*,*) 'will record hydrophobic energy terms'
        call set_bit_energy(o_record,get_nb_gridtype('hydrophobic') - 1 )
    endif

    if ( is_associated ( sgrid_tmp % pset_grid ( 4 ) ) ) then
        ! lennard jones
        !write(*,*) 'will record repulsive lennardjones energy terms'
        call set_bit_energy(o_record,get_nb_gridtype('repulsive_lj') - 1 )
    end if

    ! cannot anymore
    if ( associated ( sgrid_tmp % p_ljgrid ) ) then
        !! lennard jones
        !write(*,*) 'will record lennardjones energy terms'
        call set_bit_energy(o_record,get_nb_gridtype('lennardjones') - 1 )
    endif

    ! test if image_charge
    if ( tab_prot % array ( prot_nb ) % iimg_chg .eqv. .true. ) then
         !write(*,*) 'found at least one image charge, activate for output nb i ',i
         !write(*,*) 'will record image charge energy terms for solute ', i
         call set_bit_energy(o_record,get_nb_gridtype('image_charge') - 1 )
    end if

    ! test if metal_desolv, if any is true will record, like image charge
    if ( tab_prot % array ( prot_nb ) % imetaldesolv .eqv. .true. ) then
        call set_bit_energy(o_record, get_nb_gridtype('metal_desolvation') - 1 )
        !write(*,*) 'set_bit_energy for metal_desolv'
    end if

    end do

    ! bit integer
    ! always nstep
    call set_bit_integer ( o_record, 1 )

    ! nprot only for sdamm
    if ( o_record % record_type_calc % enum_type_calc == sdamm ) then
        ! activate nprotein
        call set_bit_integer ( o_record, 2 )
    end if

    ! sda_pmf or sdamm or no need of nrun, sda_energy ?? sda_energy could depend of sdamm
    if ( ( o_record % record_type_calc % enum_type_calc == sda_2proteins ) .or. &
         ( o_record % record_type_calc % enum_type_calc == sda_koff      ) .or. &
         ( o_record % record_type_calc % enum_type_calc == sda_energy    ) ) then
        ! activate ntrajectory by default
        call set_bit_integer ( o_record, 0 )
    end if

    ! both may have flexibility
    ! activate nconf if at least one protein is flexible
    do i = 1, size ( tab_prot % array )
        if ( tab_prot % array ( i ) % iflex ) then
            call set_bit_integer ( o_record, 3 )
        end if
    end do

    ! if debug
    ! call print_bits ( o_record )

end subroutine initialize_bits

!> Initialize arrays for movable atoms
!!
!! Initialize the pointers of array of movable atoms for proteins\n
!! Check if they can be common and allocated into \ref mod_setofgrid::sogrid, or if a local copy is needed into each protein\n
!!
!! By default, points to sogrid. If needed, make a local copy for each protein (sdamm, sda_flex protein 2). If possible delete the copy in sogrid ( sdamm )\n
!! Do not account for OpenMP here, will be done in copy_array_protein
!! @param type_calc : instance of mod_type_calc::type_calculation "type_calculation"
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
subroutine initialize_movable_atoms ( type_calc, tab_prot )
    USE mod_array_protein
    USE mod_type_calc
    USE com_sda
    IMPLICIT NONE

    type ( type_calculation ) :: type_calc
    type ( array_protein_type ) :: tab_prot

    integer :: igrid, iprot, iprot1, iprot2, ierr

#ifdef DEBUG
    write(*,*) 'Initialize_movable_atoms '
    !write(*,*) 'nb_grids ', tab_prot % nb_grids
#endif

SPROT : &
    do igrid = 1, tab_prot % nb_grids

        iprot1 = tab_prot % set_of_protein ( igrid, 1 )
        iprot2 = tab_prot % set_of_protein ( igrid, 2 )


PROT:   do iprot  = iprot1, iprot2

            !write(*,*) 'iprot ', iprot
            ! flex or not, OpenMP or not !!
            ! only sdamm moves protein 1, all other will never move, so points to the original grid and position
            ! if protein1 flexible, conformation not initilaized here
            if ( ( type_calc % enum_type_calc /= sdamm ) .and. ( iprot == 1 ) ) then
#ifdef DEBUG
                write(*,*) 'protein 1 init local_copy is false '
#endif
                ! already associated, certainly use a default in constructing
                !write(*,*) 'test p_sogrid ', associated ( tab_prot % array ( iprot ) % p_sogrid )

                ! redondant prot AND associated grid
                call init_movable_array ( tab_prot % array ( iprot ), tab_prot % array ( iprot ) % p_sogrid, .false. )

            ! if a restart is provided, need local_copy and true rotation
            ! sdamm copy_movable only called once, should delete original ?, but read restart so need explicit rotation
            !  other sda called in loop, so need to keep original

            ! else if ( sda_type and flexible and protein 2 ) or ( flex and sdamm ) => copy atoms
            else if ( tab_prot % array ( iprot ) % iflex ) then

                !write( *, * ) ' initialize flexibility protein ', iprot
                ! should be correctly assigned before
                !write(*,*) 'test p_sogrid ', associated ( tab_prot % array ( iprot ) % p_sogrid )
                !write(*,*) 'test p_flex ', associated ( tab_prot % array ( iprot ) % p_flex )
                !write(*,*) 'current_conformation ', tab_prot % array ( iprot ) % p_flex % current_conf
                !write(*,*) 'associated p_current ', associated ( tab_prot % array ( iprot ) % p_flex % p_current_node )
                ! is false to it is the p_first certainly
                !write(*,*) 'associated p_data ', associated ( tab_prot % array ( iprot ) % p_flex % p_current_node % p_data )
                !write(*,*) 'test time init ',

                ! this part already done, to delete
                ! associate p_sogrid to the initial conf, or random
                !call init_conformation ( tab_prot % array ( iprot ) )
                ! new version in module, no argument will use the default in input file
                ! here needed ? not sure..., already from restart
                !call initialize_conformation ( tab_prot % array ( iprot ) )

                ! was missing to setup  obvious for sdamm. sda_2prot redefined before each run.
                call init_flexible_freq ( tab_prot % array ( iprot ) % p_flex )

                !write(*,*) 'after init_conformation'
                if ( associated ( tab_prot % array ( iprot ) % p_sogrid ) .eqv. .false. ) then
                    write(*,*) 'ERROR p_sogrid is not associated in initialize_movable_atoms'
                    stop 1
                end if
                ! add new test should associated here
                if ( associated ( tab_prot % array ( iprot ) % p_flex ) .eqv. .false. ) then
                    write(*,*) 'ERROR p_flex is not associated in initialize_movable_atoms'
                    stop 1
                end if

                call init_movable_array ( tab_prot % array ( iprot ), tab_prot % array ( iprot ) % p_sogrid, .true. )

            ! else if ( not flexible sdamm ) or ( sda_type not flexible )
            else

                !write( *, * ) ' initialize not flexible protein ', iprot

                ! apply only for sda_2protein and protein 2, prot1 done by the default
                ! in this case no need of original copy, no restart, no need of local copy
                if ( type_calc % enum_type_calc == sda_2proteins ) then
                    !write(*,*) 'protein 2 init movable false '
                    call init_movable_array ( tab_prot % array ( iprot ), tab_prot % array ( iprot ) % p_sogrid, .false. )

                ! else ( sdamm and not flexible , sda_koff, sda_energy, sda_pmf need restart position )
                else
#ifdef DEBUG
                    write(*,*) 'case sdamm not flexible, sda_koff, sda_energy, sda_pmf need restart position, local copy true'
#endif
                    !write(*,*) 'sdamm not flexible call init_movable_array prot ', iprot
                    !write(*,*) 'test associated ', associated ( tab_prot % array ( iprot ) % p_sogrid )
                    call init_movable_array ( tab_prot % array ( iprot ), tab_prot % array ( iprot ) % p_sogrid, .true. )
                end if

            end if

            ! copy could be applied to all now, effective copy only if necessary, known from prot % local_copy
            call copy_movable_array ( tab_prot % array ( iprot ) )

            ! check if sogrid ( data exists ) and last protein of the group:  data can be deleted
            ! apply only for sdamm not flexible
            ! tricky this part ! necessary to make this economy ?? only complexify the things...
            if ( iprot == iprot2 ) then

                ! for the moment only one case, maybe tricky with OpenMP true parrallel but not necessary.
                if ( ( type_calc % enum_type_calc == sdamm ) .and. ( tab_prot % array ( iprot ) % iflex .eqv. .false. ) ) then

                    !! delete sogrid movable_atoms
                    !write(*,*) 'will delete array in sogrid, because not needed'
                    if ( allocated ( tab_prot % array ( iprot ) % p_sogrid % position_charge ) ) then
                        deallocate ( tab_prot % array ( iprot ) % p_sogrid % position_charge, stat = ierr )
                        if ( ierr /= 0 ) then
                            write(*,*) 'ERROR In deallocation of position_charge'
                            stop 1
                        end if
                        !nullify ( tab_prot % array ( iprot ) % p_sogrid % position_charge )
                    end if

                    if ( allocated ( tab_prot % array ( iprot ) % p_sogrid % list_atom_surface ) ) then
                        deallocate ( tab_prot % array ( iprot ) % p_sogrid % list_atom_surface, stat = ierr )
                        if ( ierr /= 0 ) then
                            write(*,*) 'ERROR In deallocation of list_atom_surface'
                            stop 1
                        end if
                        !nullify ( tab_prot % array ( iprot ) % p_sogrid % list_atom_surface )
                    end if

                    if ( allocated ( tab_prot % array ( iprot ) % p_sogrid % reaction_criteria_pos ) ) then
                        deallocate ( tab_prot % array ( iprot ) % p_sogrid % reaction_criteria_pos, stat = ierr )
                        if ( ierr /= 0 ) then
                            write(*,*) 'ERROR In deallocation of reaction_criteria_pos'
                            stop 1
                        end if
                        !nullify ( tab_prot % array ( iprot ) % p_sogrid % reaction_criteria_pos )
                    end if

                    !! should deallocate p_ljgrid also...


                    !! what's for lennardjones p_ljgrid, same ?? only for atom position, not for the grid

                end if !! end if sdamm and not flexible, can delete original

            end if

        end do PROT

    end do SPROT
    ! write(*,*) 'end init'

end subroutine initialize_movable_atoms

!> Initialize the random number generators
!!
!! If dseed is equal to zero, the time is used as a seed, otherwise use the provided dseed
!!
!! Initialze both generator ( see \ref RNG ):
!!  - without OpenMP : only one seed is used
!!  - with OpenMP : each thread N uses the seed dseed+N
!!
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param dseed : input seed for RNG, if 0 use the time
subroutine initialize_random_generator ( type_calc, dseed )
    USE mod_type_calc
!$  USE omp_lib
!$  USE Par_Zig_mod

    IMPLICIT NONE
    type ( type_calculation ) :: type_calc
    real ( kind=8 ) :: dseed

    real ( kind=8 ), dimension(3) :: gaus
    integer :: iseed

    !! for zig method
!$  integer, dimension ( : ), allocatable :: tab_seed
!$  integer :: nb_thread, i, grainsize = 32

#ifdef DEBUG
    write (*,*) 'Entry initialize_random_generator dseed ', dseed
#endif

    gaus = 0.d0

    if ( ( type_calc % bool_par ) .and. ( type_calc % nb_par > 1 ) ) then
        dseed = dseed + type_calc % nb_par - 1
    end if

    ! slighlty faster the new RNG
    write(*,*)
    write(*,*) 'Random Number Generator, dseed: ', dseed
    iseed = int(dseed)

!!  only if OpenMP
!$  if ( iseed == 0 ) then
!$      call generate_seed_from_clock ( iseed )
!$      write(*,*) 'seed generated from system clock ', iseed
!$  end if
    !write(*,*) 'initialize RNG, new ziggurat thread_safe if OpenMP, otherwise original methods iseed=', iseed

    !! sdamm, need only one random generator for the moment
!$  if ( OMP_get_max_threads() > 0 ) then
!$
!$      nb_thread = OMP_get_max_threads()
!$      allocate ( tab_seed ( nb_thread  ) )
!$      do i = 0, nb_thread - 1
!! !$          write(*,*) 'new seed ', iseed + 1
!$          tab_seed ( i + 1 ) = iseed + i
!$      end do
!$
!$      !! initialize module
!$      call par_zigset( nb_thread , tab_seed, grainsize )
!$      write(*,*) 'Initialize RNG with ziggurat method'
!!!$      write(*,*) 'initialize for parrallel nb ', nb_thread, tab_seed
!$      write(*,*) 'tab_seed : ', tab_seed

!$  else
        CALL ZBQLINI(iseed)
        write(*,*) 'Initialize original method for RNG'
!$  end if

     ! call ggubs_thread_safe ( gaus(1) )
     ! write(*,*) 'test uniform ', gaus(1)
     call ggnml_thread_safe( gaus )
     write(*,*) 'test gaus ', gaus

end subroutine initialize_random_generator
