!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file read_pdb.f90
!!
!! Subroutine read_pdb()

! could compute cm here, even when counting number of line

!> Read the PDB file
!!
!! Read position of atoms, they are only temporary data. Only the accessible atoms are later used\n
!! Add array tmp_rcriter_pos,  if associated will be used for updating position of reaction criteria if the protein is flexible\n
!! Read hexadecimal number if present (VMD and more than 100000 atoms)\n
!! @param pdb_filename : name of the file
!! @param rcriter : instance of \ref mod_reactioncriteria::react_criter "react_criter"
!! @param tmp_atom_letter used later for Van der Walls
!! @param tmp_atom_position position of all atoms in pdb
!! @param tmp_rcriter_pos : position of the reaction criteria
!! @param counter_prot : protein id, used for reaction criteria
subroutine read_pdb ( pdb_filename, rcriter, tmp_atom_letter, tmp_atom_position, tmp_rcriter_pos, counter_prot )
      USE mod_ReactionCriteria
      IMPLICIT NONE
      character*128 pdb_filename
      type ( react_criter ) :: rcriter
      real ( kind=8 ), dimension(:,:), intent (inout), pointer :: tmp_atom_position, tmp_rcriter_pos
      character*2, dimension(:), intent(inout), pointer :: tmp_atom_letter
      integer, intent ( in ) :: counter_prot

      ! for reading lines of pdb
      character*2 :: a2n
      character*128 string
      character*5 :: atom_nb_string
      character*6 :: hex_set
      integer :: atom_nb
      real ( kind=8 ), dimension(3) :: xv
      !! general
      integer :: ierr, nat, number_line, io
      logical :: ex, check_criteria, bool_read_hex
      integer :: iusas = 555, pos

      !!to check if reaction coordiante are correct if protein is flexible
      logical :: status_include
      integer :: number_include

      xv = (/0.d0, 0.d0, 0.d0/)
      check_criteria = .false.
      !! default false, will be set to true, as sson the first hexadecimal number is met, ok with 100 000 (contains a )
      bool_read_hex = .false.
      atom_nb = 0
      pos = -2

      number_include = 0
      status_include = .false.

      !write(*,*) 'Entry read_pdb'
      atom_nb_string=''
      ! should deal if hexadecimal character, if more than 100000 atoms
      hex_set ='abcdef'

      inquire ( file=pdb_filename, EXIST=ex )
      if ( .NOT.ex ) then
          write (*,*) "ERROR Pdb file does not exist or file name is missing ", pdb_filename
          stop 1
      end if

      write(*,*) 'pdb filename: ', trim(pdb_filename)

      open (iusas,file=pdb_filename,form='formatted',status='old')
      nat = 0
      number_line = 0
      !count number of line
      read ( iusas, '(a)', iostat=io ) string
      if ( io < 0 ) then
        write( *,* ) 'INFO file is empty', pdb_filename
        return
      endif

      do while ( io == 0 )
        if ( (string(1:4)=='ATOM').or.(string(1:6)=='HETATM') ) then
            nat = nat+1
        endif
        number_line = number_line + 1
        read ( iusas, '(a)', iostat=io ) string
      enddo

      if ( nat == 0 ) then
         write(*,*) 'Number of atom is 0 !'
         return
      endif
      write( *, fmt='("Number of atoms in pdb :      ",i8)' ) nat

      ! allocate necessary size
      ! do not enter here with debugger
      allocate ( tmp_atom_position (3,nat), tmp_atom_letter (nat), stat=ierr )
      if ( ierr /= 0 ) then
        write (*,*) 'ERROR Cannot allocate tmp_atom_position with ', nat, 'atom'
      endif

      !! Valgrind asks for initialisation,
      !! only because error in filling !
      tmp_atom_letter(:) = '  '
      tmp_atom_position(:,:) = 0.
      !tmp_rcriter_pos(:,:) = 0., done when allocated in prepare_atom_protein

      if ( associated ( tmp_rcriter_pos ) ) then
        !write(*,*) 'Need to check position of atoms for reaction '
        check_criteria = .true.
      !else
      !  write(*,*) 'No Need to check position of atoms for reaction '
      end if


      rewind( iusas )
      io=0
      nat=1
      ! read first line, otherwise make one turn more
      read ( iusas, '(a)', iostat=io ) string

      do while ( io == 0 )

        if ( (string(1:4)=='ATOM').or.(string(1:6)=='HETATM') ) then

            ! read (string,504) atom_nb_string, an, an2, xv
            !read ( string, fmt='(4X,A7,1X,2A1,16X,3F8.3)' ) atom_nb_string, an, an2, xv
            ! was working for all except surface graphite from paolo
            !read ( string, fmt='(4X,A7,2X,A2,16X,3F8.3)' ) atom_nb_string, a2n, xv
            ! try small change, work for paolo file ! to see with other
            read ( string, fmt='(6X,A5,2X,A2,15X,3F8.3)' ) atom_nb_string, a2n, xv
            !write(*,*) 'test new atom name a2n!',trim(a2n),'!'
            !write(*,*) 'atom_nb_string!',trim(atom_nb_string),'!!'

!SCAN(STRING, SET, back)    Returns the position of the first occurrence
!                           of any character in the string SET in the string
!                           STRING. If BACK is true, you will get
!                           the rightmost such character.
            pos = scan ( atom_nb_string, hex_set )

            ! read hex , set up read_hex to true for the first hex number met in the file
            if (  ( bool_read_hex .eqv. .true. ) .or. ( pos /= 0 ) ) then
                 bool_read_hex = .true.
                read ( atom_nb_string, fmt='(Z5)' ) atom_nb
            ! else read decimal format
            else
                ! before HETATM i7
                read ( atom_nb_string, fmt='(i5)' ) atom_nb
            end if
            ! write(*,*) 'pos nat atom_nb',pos,nat, atom_nb

            ! trick for different format, and read only one character for atom
            !if (an.EQ.' ') an = an2

            tmp_atom_position( :,nat ) = xv
            tmp_atom_letter ( nat ) = trim(a2n)

            ! test if needed as rcriteria, function in rcriteria module
            if ( check_criteria .eqv. .true. ) then
                call check_add_criteria ( rcriter, counter_prot, atom_nb, xv, tmp_rcriter_pos, status_include  )
                if ( status_include ) then
                    number_include = number_include + 1
                end if
            end if

            nat = nat + 1
        endif
        read ( iusas, '(a)', iostat=io ) string
      enddo

      close (iusas)

      !! print info, for tests, may be use for warning to improve
      if ( check_criteria ) then
        write(*,*) 'number of reaction criteria read from PDB: ', number_include
        if ( number_include == 0 ) then
            write(*,*) 'ERROR no reaction read from the PDB file'
            stop 1
        end if
      end if

      ! for test
#ifdef DEBUG
      !  do i = 1 , rcriter % ntotal
      !      write(*,*) 'tmp_rcriter_pos ', tmp_rcriter_pos(:,i)
      !  end do
#endif

      end subroutine read_pdb
