!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file measure_time_force_energy.f90
!! benchmarking function


!> Program to evaluate the time of computation of different methods.\n
PROGRAM measure_time
    !! make all variables accessible from here, and functions for reading input
    USE setup_sda
    USE mod_set_parameters
    USE AllTimer

    implicit none

    interface
        subroutine measure_time_forces ( tab_prot, param_force_energy, param_metal, &
                                         dummy_array, dist, geom, &
                                         nb_loop, store_time_force, size_forces, force_fast )
            USE mod_array_protein
            USE mod_force_energy
            USE mod_compute_force
            USE mod_compute_force_sdamm
            USE mod_compute_image_charge
            USE mod_compute_metaldesolv
            USE mod_compute_analytic_2proteins
            USE mod_compute_analytic_sdamm
            USE mod_geometry
            USE com_sda
            USE AllTimer
            IMPLICIT NONE
            type ( array_protein_type ), intent(in) :: tab_prot
            type ( type_force_energy ), intent ( in ) :: param_force_energy
            type ( parameter_metaldesolv ) :: param_metal
            real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
            real ( kind=4 ), intent ( in ) :: dist
            type ( geometry ) :: geom
            real ( kind=4), dimension(:), allocatable :: store_time_force
            integer, intent ( out ) :: size_forces
            integer, intent(in) :: nb_loop
            logical :: force_fast
        end subroutine measure_time_forces

        subroutine measure_time_energies ( tab_prot, param_force_energy, param_metal, &
                                           dummy_array, dist, geom, &
                                           nb_loop, store_time_energy, size_energies, array_energy, force_fast )
            USE mod_array_protein
            USE mod_force_energy
            ! new module contains all forces function
            USE mod_compute_energy
            USE mod_compute_energy_flex
            USE mod_compute_energy_sdamm
            USE mod_compute_image_charge
            USE mod_compute_metaldesolv
            USE mod_compute_analytic_2proteins
            USE mod_compute_analytic_sdamm
            USE mod_geometry
            USE com_sda
            USE AllTimer
            IMPLICIT NONE
            type ( array_protein_type ), intent(in) :: tab_prot
            type ( type_force_energy ), intent ( in ) :: param_force_energy
            type ( parameter_metaldesolv ) :: param_metal
            real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
            real ( kind=4 ), intent ( in ) :: dist
            type ( geometry ) :: geom
            real ( kind=4), dimension(:), allocatable :: store_time_energy
            integer, intent ( out ) :: size_energies
            integer, intent(in) :: nb_loop
            real ( kind=8 ), dimension( :,: ) :: array_energy
            logical :: force_fast
        end subroutine measure_time_energies

        subroutine compare_all_time_forces ( store_time_force, size_forces )
            implicit none
            real(kind=4), dimension(:) :: store_time_force
            integer :: size_forces
        end subroutine compare_all_time_forces

        subroutine compare_all_time_energies ( store_time_energy, size_energies )
            implicit none
            real(kind=4), dimension(:) :: store_time_energy
            integer :: size_energies
        end subroutine compare_all_time_energies

    end interface


    real ( kind=4 ) :: Cut2 !, Cut2_surf !, dist
    real ( kind=4 ) :: dist
    real (kind=8), dimension(3) :: pos_relat

    !! specific for this test, for the loop, for reading restart file
    !! maybe later to average over configuration
    integer :: nb_conf, nconf, nb_loop, size_forces, size_energies, ierr
    real(kind=4), dimension(:), allocatable :: store_time_force, store_time_energy
    type ( NodeComplexe ), pointer :: tmp_nodecomplexe

    !> for array energy
    !! @{
    real ( kind=8 ), dimension(:,:), pointer:: array_energy
    real ( kind=8 ), dimension( : ), pointer :: dummy_array
    !! @}

    logical :: iflex, force_fast
    !! declare timer
    !type ( mytimer ) :: timer_global

    ! could add nb_argument
    call setup_main !( nb_argument )
    nullify ( dummy_array )
    nullify ( tmp_nodecomplexe )

    ! default false
    force_fast = .false.
    iflex = .false.
    !nb_loop = 100000
    nb_loop = 1000
    !nb_loop = 5


    write(*,*) 'Type of calculation ', type_calc % enum_type_calc
    write(*,*)'=========================='

    ! reproduce all the setup of a simulation

    ! same initialization
    call init_force_energy ( param_force_energy, tab_protein_unique, type_calc, param_metaldesolv % imetaldesolv )

    !call set_cutoff ( tab_protein_unique, geom, Cut2, Cut2_surf ) !optional Cut2_surf
    !write(*,*) 'Interaction cutoff square, Cut2 ', Cut2, 'Cut2_surf ', Cut2_surf

    !NJB: Setup cut array
    call set_cut_array ( tab_protein_unique, geom, Cut2 )
    write(*,*) 'Interaction cutoff square, Cut2 ', Cut2

    if ( param_analytic % ianalytic ) then
        write(*,*)
        write(*,*) 'call init_analytic from Main loop'
        write(*,*) 'with cutoff ', sqrt(Cut2)
        write(*,*) 'use geometry % c-surf ', geom % stop_position
        ! sda 2proteins, use the c-surf as limit
        !call init_analytic ( param_force_energy, tab_protein_unique, sqrt(Cut2), param_analytic % ionic )
        call init_analytic ( param_force_energy, tab_protein_unique, geom % stop_position, param_analytic % ionic )
    endif

    ! summary data
    call print_info_param_fe ( param_force_energy )


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! main program, loop over the restart and ask for force calculation
    write(*,*) 'p_restart associated ', associated ( p_restart )

    if ( associated ( p_restart ) ) then

        ! check if complexe or trajectory
        if ( associated ( p_restart % pcomplexe ) ) then
            !! error in nb_complexe !! corrected fine now
            nb_conf = p_restart % pcomplexe % nb_complexe
            !! associate tmp_nodecomplexe to the first node
            tmp_nodecomplexe =>  p_restart % pcomplexe % p_first

            write(*,*) 'sda restart from a complex : ',nb_conf, 'complexes read from restart complexes'
        ! if trajectory
        else if ( associated ( p_restart % ptrajectory ) ) then
            nb_conf = p_restart % ptrajectory % size_array
            write(*,*) 'sda restart from a trajectory : ', nb_conf, 'complexes read from restart trajectory'
        else
            write (*,*) 'ERROR p_restart is not associated, problem !!!'
            stop 1
        end if
    end if


    call set_dummy_array_and_logical ( tab_protein_unique, dummy_array, iflex )

    !! add all creation in a function, force to create larger, can use sda, sdamm

    ! case of sda_2proteins input
    if ( associated( o_complexes % pcomplexe ) ) then
        call set_array_energy ( type_calc, o_complexes, size ( tab_protein_unique % array ), &
                                array_energy, opt_force_large_array = .true. )
    ! case sdamm
    else if ( associated( trajectories % ptrajectory ) ) then
        call set_array_energy ( type_calc, trajectories, size ( tab_protein_unique % array ), &
                                array_energy, opt_force_large_array = .true. )
    else
        write(*,*) 'ERROR Problem with record output, should be specified'
        STOP 1
    end if

   ! maybe later, average over configuration
   allocate ( store_time_force ( 4 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Error allocation store_forces'
        stop 1
   end if
   store_time_force = 0.d0
   allocate ( store_time_energy ( 6 ), stat = ierr )
   if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating store_energies'
        stop 1
   end if
   store_time_energy = 0.d0

   !!!!!!!!!!! make real code
   write(*,*)
   write(*,*) 'Begin loop'

! dependent if 2 proteins (sda_koff) or sdamm ( read one snapshot only)
if ( type_calc % enum_type_calc /= sdamm ) then

CONFLOOP : &
    do nconf = 1, nb_conf

        write(*,*)
        write(*,*) 'nconf ',nconf

        ! if a complexe, read from listcomplexe
        if ( associated ( p_restart % pcomplexe ) ) then

            tmp_nodecomplexe => tmp_nodecomplexe % p_next
            !! just a check, coan delete
            !if ( associated ( tmp_nodecomplexe ) ) then
                !write(*,*) 'ok to the next complexe'
            if ( associated ( tmp_nodecomplexe ) .eqv. .false. ) then
                write(*,*) 'ERROR tmp_nodecomplexe not associated'
                STOP 1
            end if

            !write(*,*) 'load data from list_complexe '
            tab_protein_unique % array ( 2 ) %  position = tmp_nodecomplexe % p_data %  position
            tab_protein_unique % array ( 2 ) %  orientation_x = tmp_nodecomplexe % p_data % orient_x
            tab_protein_unique % array ( 2 ) %  orientation_y = tmp_nodecomplexe % p_data % orient_y
            !! should read conformation_nb here

        !! else if a trajectory, read from array_trajectory
        else if ( p_restart % type_record == trajectory_enum ) then
            tab_protein_unique % array ( 2 ) %  position = p_restart % ptrajectory % array_trajectory ( nconf ) % position
            tab_protein_unique % array ( 2 ) %  orientation_x = p_restart % ptrajectory % array_trajectory ( nconf ) % orient_x
            tab_protein_unique % array ( 2 ) %  orientation_y = p_restart % ptrajectory % array_trajectory ( nconf ) % orient_y
            !! shoud read conformation_nb here
        end if

        call norm ( tab_protein_unique % array ( 2 ) % orientation_x )
        call norm ( tab_protein_unique % array ( 2 ) % orientation_y )

        ! compute z as the cross product
        call cross ( tab_protein_unique % array ( 2 ) % orientation_z, &
                     tab_protein_unique % array ( 2 ) % orientation_x, tab_protein_unique % array ( 2 ) %  orientation_y )


        write(*,*) 'read position of prot2'
        write(*,*) 'pos ', tab_protein_unique % array ( 2 ) %  position
        write(*,*) 'orientx ', tab_protein_unique % array ( 2 ) % orientation_x
        write(*,*) 'orienty ', tab_protein_unique % array ( 2 ) % orientation_y
        write(*,*) 'orientz ', tab_protein_unique % array ( 2 ) % orientation_z
        pos_relat =  tab_protein_unique % array ( 2 ) %  position
        write(*,*) 'pos_relat ', pos_relat
        write(*,*) 'dist ', sqrt( dot_product ( pos_relat, pos_relat ) )

        ! make copy of movable atom, only needed if flexible protein2 or sdamm
        call copy_movable_array ( tab_protein_unique % array ( 2 ) )

        !! compute distance
        call compute_distance ( tab_protein_unique % array ( 1 ) , tab_protein_unique % array ( 2 ) , dist )
        !write(*,*) 'dist', dist

        call measure_time_forces ( tab_protein_unique, param_force_energy, param_metaldesolv, dummy_array, dist, geom, &
                                   nb_loop, store_time_force, size_forces, force_fast )


        call measure_time_energies ( tab_protein_unique, param_force_energy, param_metaldesolv, dummy_array, dist, geom, &
                                     nb_loop, store_time_energy, size_energies, array_energy, force_fast )

        call compare_all_time_forces ( store_time_force, size_forces )
        call compare_all_time_energies ( store_time_energy, size_energies )

    end do CONFLOOP

! case sdamm, use only the restart position, but call same function, to combine ?
  else if ( type_calc % enum_type_calc == sdamm ) then
      !write(*,*) 'sdamm run'
      call measure_time_forces ( tab_protein_unique, param_force_energy, param_metaldesolv, dummy_array, dist, geom, &
                                 nb_loop, store_time_force, size_forces, force_fast )
                                   !store_time_force, size_forces, force_fast )

      call measure_time_energies ( tab_protein_unique, param_force_energy, param_metaldesolv, dummy_array, dist, geom, &
                                   nb_loop, store_time_energy, size_energies, array_energy, force_fast )

      call compare_all_time_forces ( store_time_force, size_forces )
      call compare_all_time_energies ( store_time_energy, size_energies )


    endif

  call delete_force_energy ( param_force_energy )

  if ( associated ( array_energy ) ) then
        deallocate ( array_energy )
  end if
  if ( associated ( dummy_array ) ) then
        deallocate ( dummy_array )
  end if

    ! here common for all runs
    call clean_main

END PROGRAM

subroutine measure_time_forces ( tab_prot, param_force_energy, param_metal, dummy_array, dist, geom, &
                                 nb_loop, store_time_force, size_forces, force_fast )
    USE mod_array_protein
    USE mod_force_energy
    ! new module contains all forces function
    USE mod_compute_force
    USE mod_compute_force_sdamm
    USE mod_compute_image_charge
    USE mod_compute_metaldesolv
    USE mod_compute_analytic_2proteins
    USE mod_compute_analytic_sdamm
    USE mod_geometry
    USE com_sda
    USE AllTimer
    IMPLICIT NONE
    type ( array_protein_type ), intent(in) :: tab_prot
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    type ( parameter_metaldesolv ) :: param_metal
    real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
    real ( kind=4 ), intent ( in ) :: dist
    type ( geometry ) :: geom
    real ( kind=4), dimension(:), allocatable :: store_time_force
    integer, intent ( out ) :: size_forces
    integer, intent(in) :: nb_loop
    logical :: force_fast

    type ( mytimer ) :: timer_force
    real (kind=8 ), dimension(3) :: f, t, t1 !, ff1, ff2, tt1, tt2
    real (kind=8 ), dimension(3) :: pos_relat
    !! dummy values
    real(kind=8), dimension(10,3,3) :: store_dummy
    real(kind=8), dimension(3) :: df,dt,dt1
    integer :: nb_prot, ierr, i

    logical :: sdamm_only
    ! need for sdamm
    real ( kind=8 ), dimension ( :, :), allocatable :: force, torque

    write(*,*) 'Entry measure time force nb_loop ', nb_loop

    nb_prot = size ( tab_prot % array )
    sdamm_only = .false.
    if ( nb_prot > 2 ) sdamm_only = .true.

    size_forces = 0

    df = (/ 0.d0, 0.d0, 0.d0 /)
    dt = (/ 0.d0, 0.d0, 0.d0 /)
    dt1= (/ 0.d0, 0.d0, 0.d0 /)
    store_dummy = 0.d0

    !! for sda_2protein
if ( sdamm_only .eqv. .false. ) then

    call init_mytimer ( timer_force )
    call start_mytimer ( timer_force )
    do i = 1, nb_loop
        call force_epedhdlj_2proteins ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), param_force_energy, dist, &
                                        f, t, t1, dummy_array )

        ! test image charge
        if ( param_force_energy % iimg_chg ) then
            call force_image_charge ( tab_prot, param_force_energy, f, t )
        end if
        ! test metal desolvation
        if ( param_force_energy % imetaldesolv ) then
            !write(*,*) 'call force 2proteins metaldesolv '
            !write(*,*) 'f,t before ', f,t
            call compute_force_metaldesolv_2prot( tab_prot, param_metal, f,t )
        end if

        !write(*,*) 'param_force_energy % ianalytic ', param_force_energy % ianalytic
        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                ! compute and add the force inside the function
                call force_analytic_dh ( tab_prot, param_force_energy, dist, f, t, t1 )
            endif
        endif
        df = df + f
        dt = dt + t
        dt1 = dt1 + t1
    end do
    call stop_mytimer ( timer_force )
    write(*,*) 'timer 2 prot: ', get_time( timer_force )

    size_forces = size_forces + 1
    store_dummy( size_forces,1,:) = df
    store_dummy( size_forces,2,:) = dt
    store_dummy( size_forces,3,:) = dt1
    store_time_force( size_forces ) = get_time( timer_force )

   !! version fast
   if ( ( param_force_energy % fast_computation ) .or. ( force_fast .eqv. .true. ) ) then
      df = (/ 0.d0, 0.d0, 0.d0 /)
      dt = (/ 0.d0, 0.d0, 0.d0 /)
      dt1= (/ 0.d0, 0.d0, 0.d0 /)
      call init_mytimer ( timer_force )
      call start_mytimer ( timer_force )
      do i = 1, nb_loop
        call force_epedhdlj_2proteins_fast ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), param_force_energy, dist, &
                                        f, t, t1, dummy_array )
        ! they are normal, non-fast version
        if ( param_force_energy % iimg_chg ) then
            !write(*,*) 'Will compute image_charge in fast'
            call force_image_charge ( tab_prot, param_force_energy, f, t )
        end if

        if ( param_force_energy % imetaldesolv ) then
            !write(*,*) 'call force 2proteins metaldesolv '
            !write(*,*) 'f,t before ', f,t
            call compute_force_metaldesolv_2prot( tab_prot, param_metal, f,t )
            !write(*,*) 'f,t after ', f,t
        end if

        ! version fast identical to normal
        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                ! compute and add the force inside the function
                call force_analytic_dh ( tab_prot, param_force_energy, dist, f, t, t1 )
            endif
        endif

        df = df + f
        dt = dt + t
        dt1 = dt1 + t1

      end do
      call stop_mytimer ( timer_force )
      write(*,*) 'timer 2prot fast : ', get_time( timer_force )
      size_forces = size_forces + 1
      store_dummy( size_forces,1,:) = df
      store_dummy( size_forces,2,:) = dt
      store_dummy( size_forces,3,:) = dt1
      store_time_force( size_forces ) = get_time( timer_force )
    end if
endif

    !! sdamm method
    pos_relat =  tab_prot % array(2) % position - tab_prot % array(1) % position

    allocate ( force ( 3, nb_prot ), torque ( 3, nb_prot ), stat = ierr )
    if ( ierr /= 0 ) then
        write (*,*) 'Cannot allocate force and torque'
        stop 1
    end if
    force = 0.d0
    torque = 0.d0
    df = (/ 0.d0, 0.d0, 0.d0 /)
    dt = (/ 0.d0, 0.d0, 0.d0 /)
    dt1= (/ 0.d0, 0.d0, 0.d0 /)
    call init_mytimer ( timer_force )
    call start_mytimer ( timer_force )
    do i = 1, nb_loop
        call force_epedhdlj_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                    force, torque, dummy_array ) 

        if ( param_force_energy % iimg_chg ) then
            !write(*,*) 'Will compute image_charge'
            call force_image_charge_sdamm ( tab_prot, param_force_energy, force, torque )
        end if

        if ( param_force_energy % imetaldesolv ) then
             !write(*,*) 'call force 2proteins metaldesolv '
             !write(*,*) 'f,t before ', f,t
             call compute_force_metaldesolv_sdamm( tab_prot, param_force_energy, param_metal, force,torque )
        end if

        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                 call force_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, force, torque )
            end if
        end if
        df = df + force(:,1) + force(:,2)
        dt = dt + torque(:,1) + torque(:,2)
    end do
    call stop_mytimer ( timer_force )
    write(*,*) 'timer sdamm: ', get_time( timer_force )
    size_forces = size_forces + 1
    store_dummy( size_forces,1,:) = df
    store_dummy( size_forces,2,:) = dt
    store_dummy( size_forces,3,:) = dt1
    store_time_force( size_forces ) = get_time( timer_force )

    force = 0.d0
    torque = 0.d0
    df = (/ 0.d0, 0.d0, 0.d0 /)
    dt = (/ 0.d0, 0.d0, 0.d0 /)
    dt1= (/ 0.d0, 0.d0, 0.d0 /)
   if ( ( param_force_energy % fast_computation ) .or. ( force_fast .eqv. .true. ) ) then
     call init_mytimer ( timer_force )
     call start_mytimer ( timer_force )
     do i = 1, nb_loop
       call force_epedhdlj_sdamm_fast ( tab_prot, geom, nb_prot, param_force_energy, &
                                        force, torque, dummy_array )

       if ( param_force_energy % iimg_chg ) then
            !write(*,*) 'Will compute image_charge'
            call force_image_charge_sdamm ( tab_prot, param_force_energy, force, torque )
        end if

        if ( param_force_energy % imetaldesolv ) then
             !write(*,*) 'call force 2proteins metaldesolv '
             !write(*,*) 'f,t before ', f,t
             call compute_force_metaldesolv_sdamm( tab_prot, param_force_energy, param_metal, force,torque )
        end if

        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                 call force_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, force, torque )
            end if
        end if
        df = df + force(:,1) + force(:,2)
        dt = dt + torque(:,1) + torque(:,2)
     end do
     call stop_mytimer ( timer_force )
     write(*,*) 'timer sdamm fast : ', get_time( timer_force )
     size_forces = size_forces + 1
     store_dummy( size_forces,1,:) = df
     store_dummy( size_forces,2,:) = dt
     store_dummy( size_forces,3,:) = dt1
     store_time_force( size_forces ) = get_time( timer_force )
   end if

   deallocate ( force, torque )

end subroutine measure_time_forces

subroutine compare_all_time_forces ( store_time_force, size_forces )
    implicit none
    real(kind=4), dimension(:) :: store_time_force
    integer :: size_forces

    write(*,*) 'Summary forces  (sec.)'
    write(*,*) store_time_force(1:size_forces)


end subroutine compare_all_time_forces

subroutine measure_time_energies ( tab_prot, param_force_energy, param_metal, dummy_array, dist, geom, &
                                   nb_loop, store_time_energy, size_energies, array_energy, force_fast )
    USE mod_array_protein
    USE mod_force_energy
    ! new module contains all forces function
    USE mod_compute_energy
    USE mod_compute_energy_flex
    USE mod_compute_energy_sdamm
    USE mod_compute_image_charge
    USE mod_compute_metaldesolv
    USE mod_compute_analytic_2proteins
    USE mod_compute_analytic_sdamm
    USE mod_geometry
    USE com_sda
    USE AllTimer
    IMPLICIT NONE

    type ( array_protein_type ), intent(in) :: tab_prot
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    type ( parameter_metaldesolv ) :: param_metal
    real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
    real ( kind=4 ), intent ( in ) :: dist
    type ( geometry ) :: geom
    real ( kind=4), dimension(:), allocatable :: store_time_energy
    integer, intent ( out ) :: size_energies
    integer, intent(in) :: nb_loop
    real ( kind=8 ), dimension( :,: ) :: array_energy
    logical :: force_fast

    type ( mytimer ) :: timer_energy
    !real (kind=8 ), dimension(3) :: f, t, t1 !, ff1, ff2, tt1, tt2
    real (kind=8 ), dimension(3) :: pos_relat
    real (kind=8 ) :: sum_energies, energy_img_chg
    integer :: nb_prot, i

    real(kind=8), dimension(10) :: store_dummy_energies

    logical :: sdamm_only
    ! need for sdamm
    !real ( kind=8 ), dimension ( :, :), allocatable :: force, torque
    !real ( kind=8 ), dimension ( :, :), allocatable :: array_energy

    write(*,*) 'Entry measure time energy nb_loop ', nb_loop

    nb_prot = size ( tab_prot % array )
    sdamm_only = .false.
    ! if more than 2 proteins cannot use sda functions
    if ( nb_prot > 2 ) sdamm_only = .true.
    energy_img_chg = 0.d0
    size_energies = 0
    ! add dummy sum to avoid compielr to guess it is dummy
    ! did not seem to change a lot results
    sum_energies = 0.d0
    store_dummy_energies = 0.d0

! methods energies only for sda_2proteins, cannot compute more than 2 solutes
if ( sdamm_only .eqv. .false. ) then

    sum_energies = 0.d0
    call init_mytimer ( timer_energy )
    call start_mytimer ( timer_energy )
    do i = 1, nb_loop
        call energy_epedhdlj_2proteins ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), param_force_energy, dist, &
                                         array_energy, dummy_array )

        if ( param_force_energy % iimg_chg ) then
            !write(*,*) 'compute energy image charge not fast'
            call energy_image_charge ( tab_prot, param_force_energy, array_energy )
        endif

        if ( param_force_energy % imetaldesolv ) then
            !write(*,*) 'compute energy metaldesolv'
            call compute_energy_metaldesolv_2prot( tab_prot, param_metal, array_energy )
        end if
        !write(*,*) 'param_force_energy % ianalytic ', param_force_energy % ianalytic
        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                call energy_analytic_dh ( tab_prot, param_force_energy, dist, array_energy )
            endif
        endif
        ! add dummy sum of energies
        sum_energies = sum_energies + sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
    end do
    call stop_mytimer ( timer_energy )
    !total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
    !write(*,*) 'sum_energies ', sum_energies
    !write(*,*) 'timer 2 prot: ', get_time( timer_energy )
    size_energies = size_energies + 1
    store_dummy_energies(size_energies) = sum_energies
    store_time_energy( size_energies ) = get_time( timer_energy )

    !! version fast
    if ( ( param_force_energy % fast_computation ) .or. ( force_fast .eqv. .true. ) ) then
      sum_energies = 0.d0
      call init_mytimer ( timer_energy )
      call start_mytimer ( timer_energy )
      do i = 1, nb_loop
        call energy_epedhdlj_2proteins_fast ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), param_force_energy, &
                                              array_energy, dummy_array, dist )

        if ( param_force_energy % iimg_chg ) then
            call energy_image_charge ( tab_prot, param_force_energy, array_energy )
        endif

        if ( param_force_energy % imetaldesolv ) then
            call compute_energy_metaldesolv_2prot( tab_prot, param_metal, array_energy )
        end if

        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                call energy_analytic_dh ( tab_prot, param_force_energy, dist, array_energy )
            endif
        endif
        sum_energies = sum_energies + sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
      end do
      call stop_mytimer ( timer_energy )
      !write(*,*) 'sum_energies ', sum_energies
      !write(*,*) 'timer 2prot fast : ', get_time( timer_energy )
      size_energies = size_energies + 1
      store_dummy_energies(size_energies) = sum_energies
      store_time_energy( size_energies ) = get_time( timer_energy )
    end if
endif

!!!!!! Energy flex methods
    sum_energies = 0.d0
    call init_mytimer ( timer_energy )
    call start_mytimer ( timer_energy )
    do i = 1, nb_loop
        call energy_epedhdlj_flex ( tab_prot % array ( 1 ), tab_prot, geom, nb_prot, &
                                    param_force_energy, param_metal, &
                                    dummy_array, array_energy(1,1), energy_img_chg )
        sum_energies = sum_energies + array_energy(1,1)
    end do
    call stop_mytimer ( timer_energy )
    !write(*,*) 'timer flex : ', get_time( timer_energy )
    size_energies = size_energies + 1
    store_dummy_energies(size_energies) = sum_energies
    store_time_energy( size_energies ) = get_time( timer_energy )

    sum_energies = 0.d0
    call init_mytimer ( timer_energy )
    call start_mytimer ( timer_energy )
    do i = 1, nb_loop
        call energy_epedhdlj_flex_sdamm ( tab_prot % array ( 1 ), tab_prot, geom, nb_prot, &
                                          param_force_energy, param_metal, &
                                          dummy_array, array_energy(1,1), energy_img_chg )
        sum_energies = sum_energies + array_energy(1,1)
    end do
    call stop_mytimer ( timer_energy )
    !write(*,*) 'timer flex openmp : ', get_time( timer_energy )
    size_energies = size_energies + 1
    store_dummy_energies(size_energies) = sum_energies
    store_time_energy( size_energies ) = get_time( timer_energy )

    !! sdamm method
    pos_relat =  tab_prot % array(2) % position - tab_prot % array(1) % position

    sum_energies = 0.d0
    call init_mytimer ( timer_energy )
    call start_mytimer ( timer_energy )
    do i = 1, nb_loop
        call energy_epedhdlj_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                     array_energy, dummy_array ) 
        if ( param_force_energy % iimg_chg ) then
            call energy_image_charge_sdamm ( tab_prot, param_force_energy, array_energy )
        endif
        if ( param_force_energy % imetaldesolv ) then
            write(*,*) 'compute energy metaldesolv sdamm '
            call compute_energy_metaldesolv_sdamm( tab_prot, param_force_energy, param_metal, array_energy )
            !write(*,*) 'after energy metaldesolv sdamm '
        end if

        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                call energy_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                                array_energy )
            endif
        endif
        sum_energies = sum_energies + sum ( array_energy(:,:) )
    end do
    call stop_mytimer ( timer_energy )
    !write(*,*) 'timer sdamm: ', get_time( timer_energy )
    size_energies = size_energies + 1
    store_dummy_energies(size_energies) = sum_energies
    store_time_energy( size_energies ) = get_time( timer_energy )

   if ( ( param_force_energy % fast_computation ) .or. ( force_fast .eqv. .true. ) ) then
     sum_energies = 0.d0
     call init_mytimer ( timer_energy )
     call start_mytimer ( timer_energy )
     do i = 1, nb_loop
       call energy_epedhdlj_sdamm_fast ( tab_prot, geom, nb_prot, param_force_energy, &
                                        array_energy, dummy_array )

       if ( param_force_energy % iimg_chg ) then
            call energy_image_charge_sdamm ( tab_prot, param_force_energy, array_energy )
        endif
        if ( param_force_energy % imetaldesolv ) then
            write(*,*) 'compute energy metaldesolv sdamm '
            call compute_energy_metaldesolv_sdamm( tab_prot, param_force_energy, param_metal, array_energy )
            !write(*,*) 'after energy metaldesolv sdamm '
        end if

        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                call energy_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                                array_energy )
            endif
        endif
       sum_energies = sum_energies + sum ( array_energy(:,:) )
     end do
     call stop_mytimer ( timer_energy )
     !write(*,*) 'timer sdamm fast: ', get_time( timer_energy )
     size_energies = size_energies + 1
     store_dummy_energies(size_energies) = sum_energies
     store_time_energy( size_energies ) = get_time( timer_energy )
   end if


end subroutine measure_time_energies

subroutine compare_all_time_energies ( store_time_energy, size_energies )
    implicit none
    real(kind=4), dimension(:) :: store_time_energy
    integer :: size_energies

    write(*,*) 'Summary energies  (sec.)'
    write(*,*) store_time_energy(1:size_energies)


end subroutine compare_all_time_energies

