!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file setup_sda.f90
!! group all setup, including read input, used only by test_module (copy of main.f90)


!> Module to simulate the setup of a sda run
!!
!! All variable accessible by a test program
!! In fact really good idea !! even for the normal program
!! all setup, global variables and clean up, can be used by test programs
!!
!! Should be used in main program ??
!!
!! used by test_module test_newflex
MODULE setup_sda

    USE mod_array_protein ! include mod_protein which includes mod_setofgrid, which includes mod_grid...
    USE mod_gridtype
    USE mod_ReactionCriteria
    USE mod_geometry
    USE mod_record
    USE mod_ResidenceTime
    USE mod_type_calc
    USE mod_force_energy
    USE com_sda

    interface

    subroutine read_input_file ( tab_protein, &
             rcriteria, resid_time, o_complexes, trajectories, geom, p_restart, &
             type_calc, param_timestep, param_probe, param_force_energy, param_analytic, &
             param_metaldesolv, param_pmf, option_omp, timemax, rboost, nrun, nmax_overlap, &
             account_occurency )
            USE mod_array_protein
            USE mod_ReactionCriteria
            USE mod_rate_calculation
            USE mod_type_calc
            USE mod_geometry
            USE mod_record
            USE mod_input_parameter
            USE mod_interface
            USE mod_flexibility
            USE mod_ResidenceTime
            USE mod_force_energy
            USE com_sda
            type ( array_protein_type ) :: tab_protein
            type ( react_criter ), target :: rcriteria
            type ( residence_time ) :: resid_time
            type ( record ) :: o_complexes, trajectories
            type ( record ), pointer :: p_restart
            type ( geometry ) :: geom
            type ( type_calculation ) :: type_calc
            type ( parameter_timestep ) :: param_timestep
            type ( probe_type ) :: param_probe
            type ( type_force_energy ) :: param_force_energy
            type ( parameter_analytic ) :: param_analytic
            type ( parameter_metaldesolv ) :: param_metaldesolv
            type ( parameter_pmf ) :: param_pmf
            type ( type_option_omp ) :: option_omp
            real ( kind = 4 ) :: rboost
            real ( kind=8 ), intent ( out ) :: timemax
            integer :: nrun, nmax_overlap
            logical :: account_occurency
    end subroutine read_input_file

    end interface


    type ( array_protein_type ) :: tab_protein_unique
    !! all related to reaction criteria, will create rate_calculation object if needed
    type ( react_criter ) :: rcriteria
    !! for recording residence time, distance and surface in the same module
    !! could be in geometry or other ?? could be used for rdf, resid_time 1D, 2D...
    type ( residence_time ) :: resid_time
    !! for geometry
    type ( geometry ) :: geom
    !! for record : complexes and trajectories
    type ( record ) :: o_complexes, trajectories
    !! if use restart for sda_koff, sda_energy. for sdamm update coordinates in read_input_file
    type ( record ), pointer :: p_restart

    ! type_calculation: sda_2proteins, sda_koff, sdamm,...
    type ( type_calculation ) :: type_calc
    ! simple type defined in sda_com
    type ( parameter_timestep ) :: param_timestep
    type ( probe_type ) :: param_probe
    type ( type_force_energy ) :: param_force_energy
    type ( parameter_analytic ) :: param_analytic
    type ( parameter_metaldesolv ) :: param_metaldesolv
    type ( parameter_pmf ) :: param_pmf
    !! general structure to group all omp option, only complexe for now
    type ( type_option_omp ) :: option_omp

    real ( kind=8 ) :: timemax
    real ( kind=4 ) :: rboost
    integer :: nrun, nmax_overlap
    logical :: account_occurency

contains

subroutine setup_main
    implicit none

    ! new arguments, to see if really needed, otherwise to delete
    !integer :: nb_argument, count_arg,
    integer :: io_err
    character*128 :: input_filename !, str_arg
    logical :: ex

    write(*,*) 'Entry setup_main'

    nullify ( p_restart )
    resid_time % enum_rt = rt_off


    ! for arguments
    !nb_argument = 0
    !count_arg = 0
    !input_filename = ''
    !str_arg = ''
    !ex = .false.

    !! init type calc for false parrallel or other
    type_calc % bool_par = .false.
    type_calc % nb_par = 1
    !! this code does not exist, like this sure to be read before testing
    type_calc % enum_type_calc = 0


    call getarg ( 1, input_filename )
    write(*,*) 'read input file ',trim ( input_filename )

     !write(*,*) 'length of input file ', len(trim(input_filename))
    if ( (trim(input_filename) == "") .or. (len(trim(input_filename))== 0) ) then
        write(*,*) "No input file will stop "
        flush(io6)
        stop 1
    end if

    inquire ( file=input_filename, exist=ex )
    if ( .not.ex ) then
        write (*,*) "input file does not exist: ", trim ( input_filename )
        flush(io6)
        stop 1
    end if

    !! open the file, the new io is inputf in new_sda_com
    open ( inputf, file=input_filename, form='formatted', action = 'read', status = 'old', iostat = io_err )
    if ( io_err /= 0 ) then
        write (*,*) 'ERROR in opening input file, but exists ??'
        stop 1
    end if

    call read_input_file ( tab_protein_unique, rcriteria, resid_time, &
        o_complexes, trajectories, geom, p_restart, &
        type_calc, param_timestep, param_probe, param_force_energy, param_analytic, param_metaldesolv, &
        param_pmf, option_omp, timemax, rboost, nrun, nmax_overlap, account_occurency )

     write(*,*) 'after read input from setup_sda, only unit tests'
     write(*,*) 'param_analytic dh', param_analytic % debyeh

end subroutine setup_main

subroutine clean_main
    implicit none

    call delete_geometry ( geom )
    call delete_resid_time ( resid_time )

    ! if sda_energy, sda_koff, p_restart has been allocated in read_input_file
    if ( associated ( p_restart ) ) then
        !write(*,*) 'p_restart associated delete it'
        ! true, do not want to save data in memory before deleting
        call delete_record ( p_restart, .true. )
        deallocate ( p_restart )
        nullify ( p_restart )
    end if

    ! only one function for all deletion of the array and the elements
    !write(*,*) 'will delete complexe'
    call delete_record ( o_complexes )
    call delete_record ( trajectories )
    call delete_react_criter ( rcriteria )

    ! delete all proteins and associated sogrid and flexible
    call delete_array_protein ( tab_protein_unique )

end subroutine clean_main


END MODULE setup_sda
