!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file test_force_energy2.f90
!! for unit-test


!> unit_test for force and energy.\n
!!
!! test_force_energy2, follow the 'normal' rules for fast version or not.\n
!! will avoid valgrind reporting error and all tests should pass.\n
!! need argument to force this option.\n
!!
!! sdamm version of force and energies are always valid.\n
!! sda_2proteins version valid only with 2 proteins, without rep_lenjones.\n
!! energy_flex/sdamm always valid if we compute the sum of all proteins and divide by 2.\n
!!
!! Due to the error in setting a correct orientation ( det. == 1 and all axes perpendicular ),
!! The criteria to pass the test on forces is large 0.02, because of the torques.
!! Could set up a different criteria for forces and torques
!!
!! Use local module setup_sda to simplify the setup of the test
PROGRAM test_force_energy
    ! make all variables accessible from here, and functions for reading input
    USE setup_sda
    USE mod_set_parameters
    implicit none

    interface
        subroutine compute_all_forces ( tab_prot, param_force_energy, param_metal, &
                                        dummy_array, dist, geom, store_forces, size_forces, force_fast )
            USE mod_array_protein
            USE mod_force_energy
            ! new modules contain all forces/energies function
            USE mod_compute_force
            USE mod_compute_force_sdamm
            USE mod_compute_image_charge
            USE mod_compute_metaldesolv
            USE mod_compute_analytic_2proteins
            USE mod_compute_analytic_sdamm
            USE mod_geometry
            USE com_sda
            IMPLICIT NONE
            type ( array_protein_type ), intent(in) :: tab_prot
            type ( type_force_energy ), intent ( in ) :: param_force_energy
            type ( parameter_metaldesolv ) :: param_metal
            real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
            real ( kind=4 ), intent ( in ) :: dist
            type ( geometry ) :: geom
            real ( kind=8), dimension(:,:,:), allocatable :: store_forces
            integer, intent ( out ) :: size_forces
            logical :: force_fast
        end subroutine compute_all_forces

        subroutine compare_all_forces ( store_forces, size_forces, pass_assert, max_diff_force, opt_print )
            implicit none
            real ( kind=8 ), dimension(:,:,:) :: store_forces
            integer, intent ( in ) :: size_forces
            logical :: pass_assert
            real ( kind=8 ) :: max_diff_force
            logical, optional :: opt_print
        end subroutine compare_all_forces

        subroutine compute_all_energies ( tab_prot, param_force_energy, param_metal, array_energy, &
                                        dummy_array, dist, geom, store_energies, size_energies, force_fast )
            USE mod_array_protein
            USE mod_force_energy
            ! new module contains all forces function
            USE mod_compute_energy
            USE mod_compute_energy_flex
            USE mod_compute_energy_sdamm
            USE mod_compute_image_charge
            USE mod_compute_metaldesolv
            USE mod_compute_analytic_2proteins
            USE mod_compute_analytic_sdamm
            USE mod_geometry
            USE com_sda
            IMPLICIT NONE
            type ( array_protein_type ), intent(in) :: tab_prot
            type ( type_force_energy ), intent ( in ) :: param_force_energy
            type ( parameter_metaldesolv ) :: param_metal
            real ( kind=8 ), dimension( :,: ) :: array_energy
            real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
            real ( kind=4 ), intent ( in ) :: dist
            type ( geometry ) :: geom
            real ( kind=8 ), dimension(:), allocatable :: store_energies
            integer, intent ( out ) :: size_energies
            logical :: force_fast
        end subroutine compute_all_energies

        subroutine compare_all_energies ( store_energies, size_energies, pass_assert, max_diff_energy, opt_print )
            implicit none
            real ( kind=8 ), dimension(:) :: store_energies
            integer, intent ( in ) :: size_energies
            logical :: pass_assert
            real ( kind=8 ) :: max_diff_energy
            logical, optional :: opt_print
        end subroutine compare_all_energies

        ! maths function used for testing accurracy
        function determinant ( mat ) result ( det )
            IMPLICIT NONE
            real ( kind=8 ), dimension(3,3), intent ( in ) :: mat
            real ( kind=8 ) :: det
        end function determinant

        subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross

       subroutine dot(ab,a,b)
           implicit none
           real(kind=8), dimension(3), intent(in) :: a,b
           real(kind=8), intent(out) :: ab
       end subroutine dot

    end interface

    !> new argument, need to have force_fast as an option.
    logical :: force_fast

    !> Cutoff arguments, not used
    !! @{
    real ( kind=4 ) :: Cut2 !, Cut2_surf ! dist
    !> @}

    !> specific for this test, for reading restart file of 2 proteins
    !! @{
    integer :: nb_conf !, ierr
    type ( NodeComplexe ), pointer :: tmp_nodecomplexe
    !> @}

    !> for array energy
    !! @{
    real ( kind=8 ), dimension(:,:), pointer:: array_energy
    real ( kind=8 ), dimension( : ), pointer :: dummy_array
    !> @}

    integer ::  nconf, ierr
    logical :: iflex
    real ( kind=4 ) :: dist
    real (kind=8), dimension(3) :: pos_relat

    !> storing results
    !! @{
    real ( kind = 8 ), dimension(:,:,:), allocatable :: store_forces
    real ( kind = 8 ), dimension(:), allocatable :: store_energies
    integer :: size_forces, size_energies
    logical :: pass_assert
    real ( kind=8 ) :: max_diff_force, max_diff_energy
    !> @}

    ! for testing determinant of rotation matrix
    !real(kind=8), dimension(3,3) :: tmp_mat
    !real(kind=8), dimension(3) :: res_cross
    !real(kind=8) :: res_dot

    ! could add nb_argument
    call setup_main !( nb_argument )
    nullify ( dummy_array )
    nullify ( tmp_nodecomplexe )

    ! default false
    force_fast = .false.
    iflex = .false.

    pass_assert = .true.

    max_diff_force = 0.d0
    max_diff_energy = 0.d0
    Cut2 = 0.
    !Cut2_surf = 0.
    pos_relat = 0.d0

    write(*,*) 'Type of calculation ', type_calc % enum_type_calc
    write(*,*)'=========================='

    ! reproduce all the setup of a simulation

    ! same initialization
    call init_force_energy ( param_force_energy, tab_protein_unique, type_calc, param_metaldesolv % imetaldesolv )

    !call set_cutoff ( tab_protein_unique, geom, Cut2, Cut2_surf ) !optional Cut2_surf
    !write(*,*) 'Interaction cutoff square, Cut2 ', Cut2, 'Cut2_surf ', Cut2_surf

    call set_cut_array (tab_protein_unique, geom, Cut2)
    write(*,*) 'Interaction cutoff square, Cut2 ', Cut2

    if ( param_analytic % ianalytic ) then
        write(*,*)
        write(*,*) 'call init_analytic from Main loop'
        write(*,*) 'with cutoff ', sqrt(Cut2)
        write(*,*) 'use geometry % c-surf ', geom % stop_position
        ! sda 2proteins, use the c-surf as limit
        !call init_analytic ( param_force_energy, tab_protein_unique, sqrt(Cut2), param_analytic % ionic )
        call init_analytic ( param_force_energy, tab_protein_unique, geom % stop_position, param_analytic % ionic )
    endif

    ! summary data
    call print_info_param_fe ( param_force_energy )


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! main program, loop over the restart and ask for force calculation
    write(*,*) 'p_restart associated ', associated ( p_restart )

    if ( associated ( p_restart ) ) then

        ! check if complexe or trajectory
        if ( associated ( p_restart % pcomplexe ) ) then
            !! error in nb_complexe !! corrected fine now
            nb_conf = p_restart % pcomplexe % nb_complexe
            !! associate tmp_nodecomplexe to the first node
            tmp_nodecomplexe =>  p_restart % pcomplexe % p_first

            write(*,*) 'sda restart from a complex : ',nb_conf, 'complexes read from restart complexes'
        ! if trajectory
        else if ( associated ( p_restart % ptrajectory ) ) then
            nb_conf = p_restart % ptrajectory % size_array
            write(*,*) 'sda restart from a trajectory : ', nb_conf, 'complexes read from restart trajectory'
        else
            write (*,*) 'ERROR p_restart is not associated, problem !!!'
            stop 1
        end if
    end if


    call set_dummy_array_and_logical ( tab_protein_unique, dummy_array, iflex )

    !! add all creation in a function, force to create larger, can use sda, sdamm

    ! case of sda_2proteins input
    if ( associated( o_complexes % pcomplexe ) ) then
        call set_array_energy ( type_calc, o_complexes, size ( tab_protein_unique % array ), &
                                array_energy, opt_force_large_array = .true. )
    ! case sdamm
    else if ( associated( trajectories % ptrajectory ) ) then
        call set_array_energy ( type_calc, trajectories, size ( tab_protein_unique % array ), &
                                array_energy, opt_force_large_array = .true. )
    else
        write(*,*) 'ERROR Problem with record output, should be specified'
        STOP 1
    end if

    write(*,*)
    write(*,*) 'Begin loop'

    ! here fixe the size, we have 4 implementations of force calculation
    ! allocate maximum size, will fill only the first ones
    allocate ( store_forces ( 4, 4, 3 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating store_forces'
        stop 1
    end if
    store_forces = 0.d0
    ! 6 implementation of energy calculation
    allocate ( store_energies ( 6 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating store_energies'
        stop 1
    end if
    store_energies = 0.d0

! dependent if 2 proteins (sda_koff) or sdamm ( read one snapshot only)
if ( type_calc % enum_type_calc /= sdamm ) then

CONFLOOP : &
    do nconf = 1, nb_conf

        write(*,*)
        write(*,*) 'nconf ',nconf

        store_forces = 0.d0
        store_energies = 0.d0

         ! if a complexe, read from listcomplexe
        if ( associated ( p_restart % pcomplexe ) ) then

            tmp_nodecomplexe => tmp_nodecomplexe % p_next
            ! just a check, coan delete
            !if ( associated ( tmp_nodecomplexe ) ) then
                !write(*,*) 'ok to the next complexe'
            if ( associated ( tmp_nodecomplexe ) .eqv. .false. ) then
                write(*,*) 'ERROR tmp_nodecomplexe not associated'
                STOP 1
            end if

            !write(*,*) 'load data from list_complexe '
            tab_protein_unique % array ( 2 ) %  position = tmp_nodecomplexe % p_data %  position
            tab_protein_unique % array ( 2 ) %  orientation_x = tmp_nodecomplexe % p_data % orient_x
            tab_protein_unique % array ( 2 ) %  orientation_y = tmp_nodecomplexe % p_data % orient_y
            !! should read conformation_nb here

        !! else if a trajectory, read from array_trajectory
        else if ( p_restart % type_record == trajectory_enum ) then
            tab_protein_unique % array ( 2 ) %  position = p_restart % ptrajectory % array_trajectory ( nconf ) % position
            tab_protein_unique % array ( 2 ) %  orientation_x = p_restart % ptrajectory % array_trajectory ( nconf ) % orient_x
            tab_protein_unique % array ( 2 ) %  orientation_y = p_restart % ptrajectory % array_trajectory ( nconf ) % orient_y
            !! shoud read conformation_nb here
        end if


        !tab_protein_unique % array (2) % orientation_x = (/ sqrt(2.d0)/2.d0, sqrt(2.d0)/2.d0, 0.d0 /)
        !tab_protein_unique % array (2) % orientation_y = (/ sqrt(2.d0)/2.d0, -sqrt(2.d0)/2.d0, 0.d0 /)
         !tab_protein_unique % array (2) % orientation_x = (/ 0.d0, sqrt(2.d0)/2.d0, sqrt(2.d0)/2.d0 /)
        !tab_protein_unique % array (2) % orientation_y = (/ 0.d0, sqrt(2.d0)/2.d0, -sqrt(2.d0)/2.d0 /)
        !tab_protein_unique % array (2) % orientation_x = (/ sqrt(2.d0)/2.d0, sqrt(2.d0)/2.d0, 0.d0 /)
        !tab_protein_unique % array (2) % orientation_y = (/ sqrt(2.d0)/2.d0, -sqrt(2.d0)/2.d0, 0.d0 /)

        ! improve error by one order: 10-2 to 10-3 in a general case
        ! in case of tests, exact on x and sqrt2 or 0.707 , perfect
        ! assure correct normalisation and determinant of full matrix close to 1 ( 10-8)
        ! but only 10-4 10-5 on dot product between x/y
        call norm ( tab_protein_unique % array ( 2 ) % orientation_x )
        call norm ( tab_protein_unique % array ( 2 ) % orientation_y )

        ! compute z as the cross product
        call cross ( tab_protein_unique % array ( 2 ) % orientation_z, &
                     tab_protein_unique % array ( 2 ) % orientation_x, tab_protein_unique % array ( 2 ) %  orientation_y )
        ! do not improve for the error
        !call norm ( tab_protein_unique % array ( 2 ) % orientation_z )

        ! for testing
        !tmp_mat (1,:) = tab_protein_unique % array (2) % orientation_x
        !tmp_mat (2,:) = tab_protein_unique % array (2) % orientation_y
        !tmp_mat (3,:) = tab_protein_unique % array (2) % orientation_z

        ! for testing, seems to be the problem of difference between fast and normal version
        !write(*,*) 'determinant ', determinant( tmp_mat )
        !call cross ( res_cross, tmp_mat (1,:), tmp_mat (2,:) )
        !write(*,*) 'cross x/y ', res_cross
        !call cross ( res_cross, tmp_mat (2,:), tmp_mat (3,:) )
        !write(*,*) 'cross y/z ', res_cross
        ! dot product should be 0
        !call dot ( res_dot, tmp_mat (1,:), tmp_mat (2,:) )
        !write(*,*) 'dot x/y ', res_dot
        !call dot ( res_dot, tmp_mat (1,:), tmp_mat (3,:) )
        !write(*,*) 'dot x/z ', res_dot
        !call dot ( res_dot, tmp_mat (2,:), tmp_mat (3,:) )
        !write(*,*) 'dot y/z ', res_dot

        write(*,*) 'read position of prot2'
        write(*,*) 'pos ', tab_protein_unique % array ( 2 ) %  position
        write(*,*) 'orientx ', tab_protein_unique % array ( 2 ) % orientation_x
        write(*,*) 'orienty ', tab_protein_unique % array ( 2 ) % orientation_y
        write(*,*) 'orientz ', tab_protein_unique % array ( 2 ) % orientation_z
        pos_relat =  tab_protein_unique % array ( 2 ) %  position
        write(*,*) 'pos_relat ', pos_relat
        write(*,*) 'dist ', sqrt( dot_product ( pos_relat, pos_relat ) )

        ! make copy of movable atom, only needed if flexible protein2 or sdamm
        call copy_movable_array ( tab_protein_unique % array ( 2 ) )

        !! compute distance
        call compute_distance ( tab_protein_unique % array ( 1 ) , tab_protein_unique % array ( 2 ) , dist )
        !write(*,*) 'dist', dist


        call compute_all_forces( tab_protein_unique, param_force_energy, param_metaldesolv, &
                                 dummy_array, dist, geom, store_forces, size_forces, force_fast ) !, cut2_surf )

        call compute_all_energies( tab_protein_unique, param_force_energy, param_metaldesolv, &
                                 array_energy, dummy_array, dist, geom, store_energies, size_energies, force_fast )

        !write(*,*) 'size_forces ', size_forces
        call compare_all_forces ( store_forces, size_forces, pass_assert, max_diff_force, .true. )
        ! reinitialize
        store_forces = 0.d0
        !write(*,*) 'size_energies ', size_energies

         call compare_all_energies ( store_energies, size_energies, pass_assert, max_diff_energy, .true. )
        ! reinitialize
        store_energies = 0.d0

    end do CONFLOOP

! case sdamm, use only the restart position, but call same function, to combine ?
else if ( type_calc % enum_type_calc == sdamm ) then

        call compute_all_forces( tab_protein_unique, param_force_energy, param_metaldesolv, &
                                 dummy_array, dist, geom, store_forces, size_forces, force_fast ) !, cut2_surf )

        call compute_all_energies( tab_protein_unique, param_force_energy, param_metaldesolv, &
                                 array_energy, dummy_array, dist, geom, store_energies, size_energies, force_fast )

        !write(*,*) 'size_forces ', size_forces
        call compare_all_forces ( store_forces, size_forces, pass_assert, max_diff_force, .true. )
        ! reinitialize
        store_forces = 0.d0
        !write(*,*) 'size_energies ', size_energies

         call compare_all_energies ( store_energies, size_energies, pass_assert, max_diff_energy, .true. )
        ! reinitialize
        store_energies = 0.d0

end if


    write(*,*) '#REPORT max_diff_force : ', max_diff_force
    write(*,*) '#REPORT max_diff_energy: ', max_diff_energy

    if ( pass_assert .eqv. .true. ) then
        write(*,*) '#REPORT PASS ALL TEST: SUCCESS'
    else
        write(*,*) '#REPORT PASS ALL TEST: FAIL'
    endif


    ! not in the main, not needed before, certainly an error, or because global ??
    ! to check, certianly necessary, main certainly not a pointer, sllocated on stack
    call delete_force_energy ( param_force_energy )
    !if ( associated ( array_energy_1d ) ) then
    !    deallocate ( array_energy_1d )
    !    nullify ( array_energy_1d )
    !end if
    if ( associated ( array_energy ) ) then
        deallocate ( array_energy )
    end if

    if ( allocated ( store_forces ) ) then
        deallocate ( store_forces )
    end if

    if ( allocated ( store_energies ) ) &
        deallocate ( store_energies )

    if ( associated ( dummy_array ) ) then
        deallocate ( dummy_array )
    end if

    ! here common for all runs
    call clean_main

    if ( pass_assert .eqv. .false. ) then
        STOP 1
    endif

END PROGRAM

!> Compute forces with all possible functions.\n
!! add option force_fast : if true, compute with normal and fast, even if param_force_energy % fast_computation is false
!! default should be false to pass all the tests
subroutine compute_all_forces ( tab_prot, param_force_energy, param_metal, dummy_array, dist, geom, &
                                store_forces, size_forces, force_fast )
    USE mod_array_protein
    USE mod_force_energy
    ! new modules contain all forces/energies function
    USE mod_compute_force
    USE mod_compute_force_sdamm
    USE mod_compute_image_charge
    USE mod_compute_metaldesolv
    USE mod_compute_analytic_2proteins
    USE mod_compute_analytic_sdamm
    USE mod_geometry
    USE com_sda
    IMPLICIT NONE
    type ( array_protein_type ), intent(in) :: tab_prot
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    type ( parameter_metaldesolv ) :: param_metal
    real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
    real ( kind=4 ), intent ( in ) :: dist
    type ( geometry ) :: geom
    ! need 3 dimension
    real ( kind=8), dimension(:,:,:), allocatable :: store_forces
    integer, intent ( out ) :: size_forces
    logical :: force_fast

    integer :: ierr, nb_prot
    ! should be false to pass all tests
    logical :: sdamm_only

    ! used for sda_2proteins only
    real (kind=8 ), dimension(3) :: f, t, t1
    real (kind=8 ), dimension(3) :: pos_relat

    ! needed for sdamm interface only
    real ( kind=8 ), dimension ( :, :), allocatable :: force, torque

    ! There is no point to include sda_2proteins if more than 2 proteins
    ! default true, can use sdamm functions if 2 proteins, inverse cannot work
    nb_prot = size ( tab_prot % array )

    ! certainly optional argument needed here
    sdamm_only = .false.
    if ( nb_prot > 2 ) sdamm_only = .true.

    size_forces = 0

    f =  (/ 0.d0, 0.d0, 0.d0 /)
    t =  (/ 0.d0, 0.d0, 0.d0 /)
    t1 =  (/ 0.d0, 0.d0, 0.d0 /)

    write(*,*) 'Entry compute all forces, sdamm_only=', sdamm_only
    !write(*,*) 'param_force_energy % iimg_chg ', param_force_energy % iimg_chg
    !write(*,*) 'prot2 orientation_x ', tab_prot % array (2) % orientation_x
    !write(*,*) 'allocated dummy_array ', allocated ( dummy_array )
    !write(*,*) 'Size dummy_array ', size ( dummy_array )

if ( sdamm_only .eqv. .false. ) then
    ! should always works, fast or not
    write(*,*) '========='
    write(*,*) 'call force 2proteins not fast'
    call force_epedhdlj_2proteins( tab_prot % array (1), tab_prot % array(2), &
                                   param_force_energy, dist, f, t ,t1, dummy_array )

    !write(*,*) 'after uhbd forces'
    !write(*,*) 'f ', f
    !write(*,*) 't ', t
    !write(*,*) 't1 ', t1

    ! test image charge
    if ( param_force_energy % iimg_chg ) then
        call force_image_charge ( tab_prot, param_force_energy, f, t )
    end if
    ! test metal desolvation
    if ( param_force_energy % imetaldesolv ) then
        !write(*,*) 'call force 2proteins metaldesolv '
        !write(*,*) 'f,t before ', f,t
        call compute_force_metaldesolv_2prot( tab_prot, param_metal, f,t )
    end if

    write(*,*) 'param_force_energy % ianalytic ', param_force_energy % ianalytic
    if ( param_force_energy % ianalytic ) then
        if ( param_force_energy % idebye ) then
            ! compute and add the force inside the function
            call force_analytic_dh ( tab_prot, param_force_energy, dist, f, t, t1 )
        endif
    endif

    !write(*,*) 'after all forces'
    !write(*,*) 'f ', f
    !write(*,*) 't ', t
    !write(*,*) 't1 ', t1


    size_forces = 1
    ! follow sdamm order, apply a dummy force for ff1, to compare with sdamm
    store_forces ( 1, 1, : ) = -f ! ff1
    store_forces ( 1, 2, : ) = f  ! ff2
    store_forces ( 1, 3, : ) = t1 ! tt1
    store_forces ( 1, 4, : ) = t  ! tt2
    ! reset variables ( should be done in called the first force computation anyway
    f = (/ 0.d0, 0.d0, 0.d0 /)
    t = (/ 0.d0, 0.d0, 0.d0 /)
    t1 = (/ 0.d0, 0.d0, 0.d0 /)
    ! should work if fast
    if ( ( param_force_energy % fast_computation ) .or. ( force_fast .eqv. .true. ) ) then

        write(*,*) '========='
        write(*,*) 'call force 2proteins fast'
        call force_epedhdlj_2proteins_fast( tab_prot % array (1), tab_prot % array(2), &
                                            param_force_energy, dist, f, t ,t1, dummy_array )

        !write(*,*) 'after uhbd forces'
        !write(*,*) 'f ', f
        !write(*,*) 't ', t
        !write(*,*) 't1 ', t1

        ! they are normal, non-fast version
        if ( param_force_energy % iimg_chg ) then
            !write(*,*) 'Will compute image_charge in fast'
            call force_image_charge ( tab_prot, param_force_energy, f, t )
        end if

        if ( param_force_energy % imetaldesolv ) then
            !write(*,*) 'call force 2proteins metaldesolv '
            !write(*,*) 'f,t before ', f,t
            call compute_force_metaldesolv_2prot( tab_prot, param_metal, f,t )
            !write(*,*) 'f,t after ', f,t
        end if

        ! version fast identical to normal
        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                ! compute and add the force inside the function
                call force_analytic_dh ( tab_prot, param_force_energy, dist, f, t, t1 )
            endif
        endif

        size_forces =size_forces + 1
        !write(*,*) 'FORCE 2 proteins FAST version'
        !write(*,*) 'f ',f
        !write(*,*) 't ',t
        !write(*,*) 't1 ',t1
        store_forces ( size_forces, 1, : ) = -f ! ff1
        store_forces ( size_forces, 2, : ) = f  ! ff2
        store_forces ( size_forces, 3, : ) = t1 ! tt1
        store_forces ( size_forces, 4, : ) = t  ! tt2
        f = (/ 0.d0, 0.d0, 0.d0 /)
        t = (/ 0.d0, 0.d0, 0.d0 /)
        t1 = (/ 0.d0, 0.d0, 0.d0 /)

    end if
end if ! sdamm_only is fast


!!!!!!! sdamm version of force

    !! cannot compute lennard-jones in sdamm
    !if ( associated ( tab_prot % array(1) % p_sogrid % p_ljgrid ) ) then
    !    write(*,*) 'Lennard-jones present cannot compute with sdamm force/energy'
    !    return
    !endif


    pos_relat =  tab_prot % array(2) % position - tab_prot % array(1) % position
    !write(*,*) 'pos_relat ', pos_relat

    allocate ( force ( 3, nb_prot ), torque ( 3, nb_prot ), stat = ierr )
    if ( ierr /= 0 ) then
        write (*,*) 'Cannot allocate force and torque'
        stop 1
    end if
    force = 0.d0
    torque = 0.d0

    write(*,*) 'force sdamm not fast'
    ! should always work, print only force on solute 1 and 2 for comparison
    ! cannot compute lennard-jones
    call force_epedhdlj_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                force, torque, dummy_array  ) 
    !write(*,*) 'after uhbd forces'
    !write(*,*) 'f1 ', force(:,1)
    !write(*,*) 'f2 ',force(:,2)
    !write(*,*) 't1 ', torque(:,1)
    !write(*,*) 't2 ', torque(:,2)

    ! test image charge
    if ( param_force_energy % iimg_chg ) then
            !write(*,*) 'Will compute image_charge'
            call force_image_charge_sdamm ( tab_prot, param_force_energy, force, torque )
            !! apply only the force and torque on solute, not on the surface,
            !! cannot reproduce exactly sda_2proteins, do not apply the force on protein1
    end if

    if ( param_force_energy % imetaldesolv ) then
         !write(*,*) 'call force 2proteins metaldesolv '
         !write(*,*) 'f,t before ', f,t
         call compute_force_metaldesolv_sdamm( tab_prot, param_force_energy, param_metal, force,torque )
    end if

    if ( param_force_energy % ianalytic ) then
        if ( param_force_energy % idebye ) then
             call force_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, force, torque )
        end if
    end if

    size_forces = size_forces + 1
    !write(*,*) 'FORCE sdamm not fast '
    !write(*,*) 'force1 ',force(:,1)
    !write(*,*) 'force2 ',force(:,2)
    !write(*,*) 'torque1 ',torque(:,1)
    !write(*,*) 'torque2 ',torque(:,2)
    store_forces ( size_forces, 1, : ) = force(:,1)
    store_forces ( size_forces, 2, : ) = force(:,2)
    store_forces ( size_forces, 3, : ) = torque(:,1)
    store_forces ( size_forces, 4, : ) = torque(:,2)
    force = 0.d0
    torque = 0.d0

    if ( ( param_force_energy % fast_computation ) .or. ( force_fast .eqv. .true. ) ) then

      write(*,*) 'force sdamm fast'
      call force_epedhdlj_sdamm_fast ( tab_prot, geom, nb_prot, param_force_energy, &
                                     force, torque, dummy_array )
      ! test image charge
      if ( param_force_energy % iimg_chg ) then
            !write(*,*) 'Will compute image_charge'
            call force_image_charge_sdamm ( tab_prot, param_force_energy, force, torque )
      end if

    if ( param_force_energy % imetaldesolv ) then
         call compute_force_metaldesolv_sdamm( tab_prot, param_force_energy, param_metal, force, torque )
    end if

    if ( param_force_energy % ianalytic ) then
        if ( param_force_energy % idebye ) then
             call force_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, force, torque )
        end if
    end if

    size_forces = size_forces + 1
    !write(*,*) 'FORCE sdamm fast '
    !write(*,*) 'force1 ',force(:,1)
    !write(*,*) 'force2 ',force(:,2)
    !write(*,*) 'torque1 ',torque(:,1)
    !write(*,*) 'torque2 ',torque(:,2)
    store_forces ( size_forces, 1, : ) = force(:,1)
    store_forces ( size_forces, 2, : ) = force(:,2)
    store_forces ( size_forces, 3, : ) = torque(:,1)
    store_forces ( size_forces, 4, : ) = torque(:,2)
    force = 0.d0
    torque = 0.d0
    end if

    !write(*,*) 'size_forces ', size_forces
    deallocate ( force, torque )

end subroutine compute_all_forces

subroutine compare_all_forces ( store_forces, size_forces, pass_assert, max_diff_force, opt_print )
    implicit none
    real ( kind=8 ), dimension(:,:,:) :: store_forces
    integer, intent ( in ) :: size_forces
    logical :: pass_assert
    real ( kind=8 ) :: max_diff_force
    logical, optional :: opt_print

    real ( kind=8 ) :: diff
    logical :: print_tab, pass
    integer :: i,j,k


    !write(*,*) 'Entry compare all forces '
    !write(*,*) 'size ', size ( store_forces )
    !write(*,*) 'size ', size ( store_forces(:,1,1 ))
    print_tab = .false.
    diff = 0.d0
    pass = .true.

    if ( present ( opt_print ) ) then
        print_tab = opt_print
    end if

    if ( print_tab ) then

      write(*,*)
write(*,*) ' FORCE 1                                   FORCE 2                                  TORQUE 1                 TORQUE 2'
      !do i = 1, size ( store_forces(:, 1, 1) )
      do i = 1, size_forces
        write( *, fmt = '(12(ES12.5,1X))' ) ((store_forces( i, j, k ), k=1,3), j=1,4)
      end do
    end if

    ! use first one as reference sda_ 2protein not fast, secnod may have torque slightly different
    pass = .true.

    do i = 2, size_forces
        do j = 1, 4 ! loop over ff1, ff2,,
            do k = 1, 3
            ! here report all difference in torques between fast and not fast
            !if ( abs( store_forces(i,j,k) - store_forces(1,j,k) ) > 0.0001d0 ) then
            ! 0.01 find differences. 0.05 report nothing
            diff = abs( store_forces(i,j,k) - store_forces(1,j,k) )

            if ( diff > max_diff_force ) max_diff_force = diff
            ! with normalisation, wins one order 0.005 in bnbs/unit_test
            if ( diff > 0.02d0 ) then
                write(*,*) 'differences ', store_forces(i,j,k), store_forces(1,j,k )
                pass = .false.

            end if
            end do
        end do

    end do

    if ( pass .eqv. .false. ) then
        pass_assert = .false.
    endif

end subroutine

!> compute energies with all the methods.\n
!! Like for compute_all_forces
subroutine compute_all_energies( tab_prot, param_force_energy, param_metal, array_energy, dummy_array, dist, geom, &
                                 store_energies, size_energies, force_fast )
    USE mod_array_protein
    USE mod_force_energy
    ! new module contains all forces function
    USE mod_compute_energy
    USE mod_compute_energy_flex
    USE mod_compute_energy_sdamm
    USE mod_compute_image_charge
    USE mod_compute_metaldesolv
    USE mod_compute_analytic_2proteins
    USE mod_compute_analytic_sdamm
    USE mod_geometry
    USE com_sda
    IMPLICIT NONE

    type ( array_protein_type ), intent ( in ) :: tab_prot
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    type ( parameter_metaldesolv ) :: param_metal
    real ( kind=8 ), dimension( :,: ) :: array_energy
    real ( kind=8 ), dimension( : ), intent ( in ) :: dummy_array
    real ( kind=4 ), intent ( in ) :: dist
    type ( geometry ) :: geom
    real ( kind=8 ), dimension(:), allocatable :: store_energies
    integer, intent ( out ) :: size_energies
    logical :: force_fast

    real ( kind=8 ) :: total_energy, energy_img_chg
    integer :: nb_prot, j
    logical :: sdamm_only

    nb_prot = size ( tab_prot % array )
    sdamm_only = .false.
    ! if more than 2 proteins cannot use sda functions
    if ( nb_prot > 2 ) sdamm_only = .true.

    write(*,*)
    write(*,*) '=========='
    write(*,*) 'Entry compute all energies sdamm only=', sdamm_only

    size_energies = 0
    energy_img_chg = 0.d0

! methods energies only for sda_2proteins, cannot compute more than 2 solutes
if ( sdamm_only .eqv. .false. ) then
    ! should always works
    call energy_epedhdlj_2proteins ( tab_prot % array(1), tab_prot % array(2), param_force_energy, &
                                     dist, array_energy, dummy_array )

    write(*,*)
    write(*,*) 'energy_2proteins not fast'
    write(*,*) 'after uhbd energies'
    write(*,*) 'energy 1', array_energy(:,1)
    write(*,*) 'energy 2', array_energy(:,2)

    if ( param_force_energy % iimg_chg ) then
        write(*,*) 'compute energy image charge not fast'
        call energy_image_charge ( tab_prot, param_force_energy, array_energy )
    endif

    if ( param_force_energy % imetaldesolv ) then
        !write(*,*) 'compute energy metaldesolv'
        call compute_energy_metaldesolv_2prot( tab_prot, param_metal, array_energy )
    end if

    write(*,*) 'param_force_energy % ianalytic ', param_force_energy % ianalytic
    if ( param_force_energy % ianalytic ) then
        if ( param_force_energy % idebye ) then
            call energy_analytic_dh ( tab_prot, param_force_energy, dist, array_energy )
        endif
    endif

    ! same definition than sdamm, now sum up half of potential energy, to use with epfct = 0.5
    total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
    store_energies (1) = total_energy
    write(*,*) 'Energy 2 proteins not fast total energy ', total_energy
    size_energies = 1

    if ( ( param_force_energy % fast_computation ) .or. ( force_fast .eqv. .true. ) ) then
        call  energy_epedhdlj_2proteins_fast ( tab_prot % array(1), tab_prot % array(2), param_force_energy, &
                                               array_energy, dummy_array, dist )

        write(*,*)
        write(*,*) '============='
        write(*,*) 'energy_2proteins fast'
        write(*,*) 'after uhbd energies'
        write(*,*) 'energy 1', array_energy(:,1)
        write(*,*) 'energy 2', array_energy(:,2)

        if ( param_force_energy % iimg_chg ) then
            write(*,*) 'compute energy image charge fast computation'
            call energy_image_charge ( tab_prot, param_force_energy, array_energy )
        endif

        if ( param_force_energy % imetaldesolv ) then
            write(*,*) 'compute energy metaldesolv inside fast computation'
            call compute_energy_metaldesolv_2prot( tab_prot, param_metal, array_energy )
        end if

        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                call energy_analytic_dh ( tab_prot, param_force_energy, dist, array_energy )
            endif
        endif

        size_energies =size_energies + 1
        total_energy = sum ( array_energy(:,2) ) + sum ( array_energy ( :, 1 ) )
        write(*,*) 'Energy 2 proteins version FAST total energy', total_energy
        store_energies (size_energies) = total_energy
        array_energy = 0.d0
    end if

end if ! sdamm_only is false

    !! cannot compute lennard-jones in sdamm
    !if ( associated ( tab_prot % array(1) % p_sogrid % p_ljgrid ) ) then
    !    write(*,*) 'Lennard-jones present cannot compute with sdamm sda_flex force/energy'
    !    return
    !endif

!!!! Test energy flex 2 proteins

    ! with the trick of the half-sum can test for both sda_2proteins and sdamm, only difference is OpenMP
    write(*,*)
    write(*,*) '======='
    write(*,*) 'Test energy flex 2 proteins '
    total_energy = 0.d0
    do j = 1, nb_prot
        array_energy = 0.d0

        ! image charge included in energy_flex
        call energy_epedhdlj_flex ( tab_prot % array(j), tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                                    dummy_array, array_energy(1,1), energy_img_chg )
        ! add value an other time, then can divide by 2
        total_energy = total_energy + array_energy(1,1) + energy_img_chg
    end do
    ! need to divide by 2 there
    write(*,*) 'total_energy flex 2 proteins', total_energy, total_energy / 2.d0

    total_energy = total_energy / 2.d0
    size_energies = size_energies + 1
    store_energies (size_energies) = total_energy
    array_energy = 0.d0

!!!!! Test energy_flex_sdamm
    write(*,*)
    write(*,*) '==========='
    write(*,*) 'Test energy flex SDAMM'
    total_energy = 0.d0
    do j = 1, nb_prot
        array_energy = 0.d0

        ! image charge included in energy_flex,
        ! but cannot use the same trick with / 2, image-charge should not be / 2
        ! could add a parameter, but used only for this test !
        call energy_epedhdlj_flex_sdamm ( tab_prot % array(j), tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                                          dummy_array, array_energy(1,1), energy_img_chg )

        ! total_energy = total_energy + array_energy(1,1)
        ! add value of image charge a second time, then can divide by 2
        total_energy = total_energy + array_energy(1,1) + energy_img_chg
    end do
    write(*,*) 'total_energy flex sdamm', total_energy, total_energy / 2.d0
    total_energy = total_energy / 2.d0
    size_energies = size_energies + 1
    store_energies (size_energies) = total_energy
    array_energy = 0.d0

!!!!! Test SDAMM methods
    ! sdamm methods, simple should always work
    write(*,*)
    write(*,*) '==================='
    write(*,*) 'Test SDAMM not fast'

    call energy_epedhdlj_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                 array_energy, dummy_array ) 
    if ( param_force_energy % iimg_chg ) then
            call energy_image_charge_sdamm ( tab_prot, param_force_energy, array_energy )
    endif
    if ( param_force_energy % imetaldesolv ) then
        write(*,*) 'compute energy metaldesolv sdamm '
        call compute_energy_metaldesolv_sdamm( tab_prot, param_force_energy, param_metal, array_energy )
        !write(*,*) 'after energy metaldesolv sdamm '
    end if

    if ( param_force_energy % ianalytic ) then
        if ( param_force_energy % idebye ) then
            call energy_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                array_energy )
        endif
    endif

    total_energy = sum ( array_energy(:,:) )

    size_energies = size_energies + 1
    write(*,*) 'Energy sdamm not fast total ', total_energy
    store_energies (size_energies) = total_energy
    !write(*,*) 'array_energy ', array_energy
    array_energy = 0.d0

    if ( ( param_force_energy % fast_computation ) .or. ( force_fast .eqv. .true. ) ) then
        call energy_epedhdlj_sdamm_fast ( tab_prot, geom, nb_prot, param_force_energy, &
                                          array_energy, dummy_array ) 
        if ( param_force_energy % iimg_chg ) then
            call energy_image_charge_sdamm ( tab_prot, param_force_energy, array_energy )
         endif
        if ( param_force_energy % imetaldesolv ) then
            write(*,*) 'compute energy metaldesolv sdamm inside fast '
            call compute_energy_metaldesolv_sdamm( tab_prot, param_force_energy, param_metal, array_energy )
        end if

        if ( param_force_energy % ianalytic ) then
            if ( param_force_energy % idebye ) then
                call energy_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                                array_energy )
            endif
        endif

        size_energies = size_energies + 1
        total_energy = sum ( array_energy(:,:) )
        write(*,*) 'Energy sdamm fast total ', total_energy
        store_energies (size_energies) = total_energy
        !write(*,*) 'array_energy ', array_energy
        array_energy = 0.d0
    end if
    !write(*,*) 'size_energies ', size_energies

end subroutine compute_all_energies

subroutine compare_all_energies ( store_energies, size_energies, pass_assert, max_diff_energy, opt_print )
    implicit none
    real ( kind=8 ), dimension(:) :: store_energies
    integer, intent ( in ) :: size_energies
    logical :: pass_assert
    real ( kind=8 ) :: max_diff_energy
    logical, optional :: opt_print

    real ( kind=8 ) :: diff
    logical :: print_tab, pass
    integer :: k

    !write(*,*) 'Entry compare all energies '
    !write(*,*) 'size ', size ( store_energies )
    print_tab = .false.
    diff = 0.d0
    pass = .true.

    if ( present ( opt_print ) ) then
        print_tab = opt_print
    end if

    if ( print_tab ) then

      write(*,*)
      write(*,*) 'Total energy'
      !write( *, * ) ((i,j,k, k=1,3), j=1,4)
      !write( *, fmt = '(12(ES12.5,1X))' ) ((store_forces( i, j, k ), k=1,3), j=1,4)
      !write( *, fmt = '(12(F10.6,1X))' ) ((store_forces( i, j, k ), k=1,3), j=1,4)
      !write( *, fmt = '(5(ES15.7,1X))' ) store_energies
      write(*,*) store_energies(1:size_energies)

    end if

     do k = 2, size_energies

            diff = abs( store_energies(k) - store_energies(1) )

            if ( diff > max_diff_energy ) max_diff_energy = diff

            if ( diff > 0.0001d0 ) then

                write(*,*) 'LARGE ERROR'
                write(*,*) 'differences ', store_energies(k), store_energies(1)
                pass = .false.

            end if
    end do

    if ( pass .eqv. .false. ) then
        pass_assert = .false.
    endif

end subroutine compare_all_energies
