!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<


!> \file ecm_all.f90
!!
!! Program ecm_all

!> \defgroup ECM Effective Charge Method
!!
!! Calculation of effective charges
!!
!! This additional information are writen in \ref ecm_all.f90
!!
!! A step by step construction of the effective charges require 4 steps :
!!  - \ref ecm_mksites : make the test charges ( in a *.tcha file ) from the pdb. Hardcoded approximative values are assigned to some atom ( eg : oxygens of lysines have -0.5 e )
!!  - \ref ecm_expand :
!!  - \ref ecm_regularize :
!!  - \ref ecm_mkreglev : given a criteria, will compute the effective charges at a specific iteration
!!
!! The advantage is a better control at the different steps. It allows to choose precisely the level of iteration to stop
!!
!! This can be done in one step with :
!! - \ref ecm_all : call the 4 previous steps and generate directly the effective charges
!!
!! The advantage : easier to use
!!
!! To be reusable, the 4 steps are writen in 4 main subroutines ( sub_mksites, sub_expand,.. ), implemented in \ref ecm_mainsubroutine.f90
!!
!! If your system contains unknown atom or residue type, you should use an external file: add_atoms (see examples)

!> Create effective charges in one step
!!
!! \ingroup ECM
!!
!! Make all steps for computing the effective charges\n
!! Call subfunctions of the different steps
program ecm_all
    USE mod_parameters_ecm
    IMPLICIT NONE

    integer :: count_arg, nb_argument
    character*64 str_arg

    ! input parameters
    type ( param_ecm ) :: list_param_ecm

    nb_argument = 0
    count_arg = 1

    nb_argument = command_argument_count()
    if ( nb_argument == 0 ) then
        write(*,*)
        write(*,*) 'HELP'
        write(*,*)
        write(*,*) 'ecm_all -fp p1_noh.pdb -fg ep1.bin.grd -is 150. -rc 1.0 -fe p1_noh.echa '
        write(*,*)
        write(*,*) 'Generate effective charge in one ( or two ) steps'
        write(*,*) 'parameters'
        write(*,*) '-fp pdb_name'
        write(*,*) '-fg electrostatic grid'
        write(*,*) '-fe effective charge, note test_charge is created with the same name but suffix .tcha'
        write(*,*) '-is ionic strength ( mM ) : default 150.'
        write(*,*) '-pr probe radius : default 4.'
        write(*,*) '-sk skin : default 3.'
        write(*,*) '-es solvent dielectric constant : default 78.'
        write(*,*) '-rl regularization level : default -9999 not taken into account'
        write(*,*) '-rc regularization charge : default -9999.0 not taken into account'
        write(*,*)
        write(*,*) 'If -rc or -rl is given, effective charges are directly generated'
        write(*,*) 'otherwise the matrix file is written, and charges must be generated with ecm_mkreglev -reg_level 35'
        write(*,*)
        write(*,*) 'To generate effective charges with intermediate steps:'
        write(*,*) 'ecm_mksites p1_noh.pdb'
        write(*,*) 'ecm_expand ecm.in'
        write(*,*) 'optional ecm_mkreglev -reg_level 35'
        write(*,*)
        write(*,*) 'In one step:'
        write(*,*) 'ecm_all -fp p1_noh.pdb -fg ep1.bin.grd -is 150. -rc 1.0 -fe p1_noh.echa'
        write(*,*) 'optional ecm_mkreglev -reg_level 35'
        write(*,*)
        stop
    end if

    !! default value
    list_param_ecm % probes = 4.0
    list_param_ecm % skin = 3.0
    list_param_ecm % iostr= 150.
    list_param_ecm % epssol = 78.
    !! default, do not search for a level or a charge regularization
    list_param_ecm % reg_level = -9999
    !! default -1, echa not printed only print, will need ecm_mkreglev
    list_param_ecm % reg_charge = -9999.0

    do while ( count_arg <= nb_argument )

        call getarg( count_arg, str_arg )
        if ( str_arg(1:3) == "-fp" ) then
            call getarg ( count_arg + 1, str_arg )
            list_param_ecm % pdb_name = trim ( str_arg )
        !else if ( str_arg(1:3) == "-ft" ) then
        !    call getarg ( count_arg + 1, str_arg )
        !    list_param_ecm % test_charge_name = trim ( str_arg )
            !count_arg = count_arg + 1
        else if ( str_arg(1:3) == "-fe" ) then
            call getarg ( count_arg + 1, str_arg )
            list_param_ecm % eff_charge_name = trim ( str_arg )
            !count_arg = count_arg + 1
        else if ( str_arg(1:3) == "-fg" ) then
            call getarg ( count_arg + 1, str_arg )
            list_param_ecm % grid_name = trim ( str_arg )
        else if ( str_arg(1:3) == "-pr" ) then
            call getarg ( count_arg + 1, str_arg )
            read ( str_arg,* ) list_param_ecm % probes
        else if ( str_arg(1:3) == "-sk" ) then
            call getarg ( count_arg + 1, str_arg )
            read ( str_arg,* ) list_param_ecm % skin
        else if ( str_arg(1:3) == "-is" ) then
            call getarg ( count_arg + 1, str_arg )
            read ( str_arg,* ) list_param_ecm % iostr
        else if ( str_arg(1:3) == "-es" ) then
            call getarg ( count_arg + 1, str_arg )
            read ( str_arg,* ) list_param_ecm % epssol
        else if ( str_arg(1:3) == "-rl" ) then
            call getarg ( count_arg + 1, str_arg )
            read ( str_arg,* ) list_param_ecm % reg_level
        else if ( str_arg(1:3) == "-rc" ) then
            call getarg ( count_arg + 1, str_arg )
            read ( str_arg,* ) list_param_ecm % reg_charge
        end if

        count_arg = count_arg + 1
    end do

    !! should be optional
    list_param_ecm % test_charge_name =  list_param_ecm % eff_charge_name
    call create_filename_new ( list_param_ecm % test_charge_name, "tcha" )

    write( *,fmt='("pdb coordinates to be read from : ",a40)') list_param_ecm % pdb_name
    write( *,fmt='("test charges will be read from  : ",a40)') list_param_ecm % test_charge_name
    write( *,fmt='("potential grid will be read from: ",a40)') list_param_ecm % grid_name
    write( *,fmt='("exclusion probe radius and inclusion layer: ",2f8.3)') list_param_ecm % probes, list_param_ecm % skin

    write( *,fmt='("ionic strength in mM and debye-h parameter : ",f8.3)' ) list_param_ecm % iostr
    write( *,fmt='("                       dielectric constant : ",f8.3)' ) list_param_ecm % epssol
    write( *,fmt='("regularization level   : ",i5)' ) list_param_ecm % reg_level
    write( *,fmt='("regularization charge  : ",f10.3)' ) list_param_ecm % reg_charge

    if ( ( list_param_ecm % reg_level /= -9999 ) .and. ( list_param_ecm % reg_charge /= -9999. ) ) then
        write(*,*) 'Cannot specify rc and rl at the same time '
        stop 1
    end if

    !! ecm_mksites, produces *.tcha
    call sub_mksites ( list_param_ecm % pdb_name, list_param_ecm % test_charge_name )

    !! run ecm_expand and regularize
    call sub_expand ( list_param_ecm )

    if ( ( list_param_ecm % reg_level == -9999 ) .and. ( list_param_ecm % reg_charge == -9999. ) ) then
        write(*,*)
        write(*,*) 'ecm_all calculation successfull. To generate the effective charge (.echa) file you need to run ecm_mkreglev.'
        write(*,*)
    end if

end program ecm_all
