!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file ecm_expand.f90
!!
!! Program ecm_expand

!! \addtogroup ECM
!! Specific for ecm_expand

!> Make the regularization step
!!
!! \ingroup ECM
!!
!! make ecm_regularize in ecm_expand, anyway always read the same matrix, no option possible\n
!! make ecm_mkreglev in  ecm_regularize, need option\n
!! should add option if we want to write the intermediate matrix\n
!! will sub_expand will call sub_regularize if need to split after\n
program ecm_expand

    USE mod_parameters_ecm
    IMPLICIT NONE

    character*128 :: filename
    integer :: nb_argument

    type ( param_ecm ) :: list_param_ecm

    !nb_argument = iargc ()
    nb_argument = command_argument_count()

    if ( nb_argument == 0 ) then
        write(*,*)
        write(*,*) 'HELP'
        write(*,*)
        write(*,*) 'ecm_expand ecm.in '
        write(*,*)
        write(*,*) 'Generate sets of effective charges'
        write(*,*)
        write(*,*) 'input parameter in file ecm.in ( format is fixed, do not add or remove line )'
        write(*,*) 'NOTE: the format is very similar to sda 6, but NOT exactly'
        write(*,*)
        write(*,*) 'If reg_charge is equal to 0, effective charges are not written'
        write(*,*) "You'll need to use ecm_mkreglev to select the iteration number: ecm_mkreglev -reg_level 35"
        write(*,*)
        write(*,*) 'If reg_charge is different of 0, the first iteration where the absolute valute of RM1D < reg_charge is chosen'
        write(*,*) 'to generate the effective charges'
        write(*,*)
        write(*,*) 'To generate effective charges with intermediate steps:'
        write(*,*) 'ecm_mksites p1_noh.pdb'
        write(*,*) 'ecm_expand ecm.in'
        write(*,*) 'optional ecm_mkreglev -reg_level 35'
        write(*,*)
        write(*,*) 'In one step:'
        write(*,*) 'ecm_all -fp p1_noh.pdb -fg ep1.bin.grd -is 150. -rc 1.0 -fe p1_noh.echa'
        write(*,*) 'optional ecm_mkreglev -reg_level 35'
        write(*,*)
        stop
    endif

    call getarg ( 1, filename )

    write(*,*)
    write(*,*) 'input filename ', trim ( filename )

    !! alone version read input file
    call read_input_expand ( filename, list_param_ecm )

    !! call subroutine in ecm_mainsubroutines
    call sub_expand ( list_param_ecm )

end program ecm_expand

!> Read fixed format ecm.in, similar to sda6 input
!!
!! only needed for this program
subroutine read_input_expand ( filename, list_param_ecm )
    USE mod_parameters_ecm
    IMPLICIT NONE
    character :: filename*(*)
    type ( param_ecm ) :: list_param_ecm

    integer :: ierr, io_unit = 50

    open (io_unit, file=filename , form= 'formatted', status='old', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open file ', trim ( filename )
        stop 1
    end if

    read(50,*)
    read(50,*) list_param_ecm % pdb_name
    read(50,*)
    read(50,*) list_param_ecm % test_charge_name
    read(50,*)
    read(50,*) list_param_ecm % grid_name
    read(50,*)
    read(50,*) list_param_ecm % probes, list_param_ecm % skin
    read(50,*)
    read(50,*) list_param_ecm % iostr, list_param_ecm % epssol
    read(50,*)
    read(50,*) list_param_ecm % eff_charge_name
    read(50,*)
    read(50,*) list_param_ecm % reg_charge

    if ( list_param_ecm % reg_charge == 0. ) then
        list_param_ecm % reg_level = -9999
        list_param_ecm % reg_charge = -9999.
    else
        list_param_ecm % reg_level = -9999
    end if

    write(*,*)
    write(*,*) 'Input parameters '
    write(*,*) 'pdb name ', trim ( list_param_ecm % pdb_name )
    write(*,*) 'test charge file ', trim ( list_param_ecm % test_charge_name )
    write(*,*) 'grid name ', trim ( list_param_ecm % grid_name )
    write(*,*) 'probes ', list_param_ecm % probes
    write(*,*) 'skin ', list_param_ecm % skin
    write(*,*) 'ionic strength ', list_param_ecm % iostr
    write(*,*) 'dielec. solvent ', list_param_ecm % epssol
    write(*,*) 'reg_charge ', list_param_ecm % reg_charge
    write(*,*) 'reg_level ', list_param_ecm % reg_level
    write(*,*) 'output file ', list_param_ecm % eff_charge_name


end subroutine read_input_expand


