!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file ecm_mainsubroutine.f90
!!
!! Group the main subroutines for ECM calculations
!!
!! Called by \ref ecm_all, or step by step by programs \ref ecm_mksites, \ref ecm_expand, \ref ecm_regularize

!!!!!!!!!!!!!!!!    For ecm_mksites   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! make independent routine, can be called by ecm_all

!> sub_mksites, make the test charges
!!
!! Make the test charges
!! @param filename : pdb file to read atom types
!! @param out_fname : file containing the test charges ( *.tcha )
subroutine sub_mksites ( filename, out_fname )
    USE mod_van
    IMPLICIT NONE
    character, intent( in ) :: filename*(*), out_fname*(*)

    type ( type_van ) :: tvan

    integer :: nat, ncharge
    character*54 :: string
    character*3 :: atom_name, residue_name
    ! maybe be necessary if want to test a value read for a txt file
    !logical :: test_charge
    integer :: ierr, io_unit = 50, io_unit2 = 51
    real ( kind=4 ) :: charge, tot_charge

    !for reading lines of pdb, name of atoms
    character*3 an3

    nat = 0
    ncharge = 0
    tot_charge = 0

    ! initialize the new module
    call init_van ( tvan )
    call print_array_atom ( tvan )

    ! open pdb input file
    open ( io_unit, file=filename, form='formatted', status='old', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open file ', trim ( filename ) !, but present ???'
        return
    end if

    ! open output
    open ( io_unit2, file=out_fname, form='formatted', status='replace', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open output file ', trim ( out_fname )
        return
    end if

    ! no need to store, can print directly to out_fname
    read ( io_unit, '(a)', iostat = ierr ) string
    if ( ierr /= 0 ) then
        write( *,* ) 'error input file ', filename, ' is empty'
        return
    endif

    do while ( ierr == 0 )

        if ( (string(1:4).EQ.'ATOM').or.(string(1:6).EQ.'HETATM') ) then
            nat = nat+1
	        !write(*,*) nat , string
            read ( string, fmt='(13X,A3,1X,A3)' ) an3,residue_name
            atom_name = trim ( an3 )
            !write(*,*) 'an3 !',an3,'!! atom name!',atom_name,'!!'
            !write(*,*) 'residue_name !',residue_name,'!!'

            !call charge_def ( atom_name, residue_name, test_charge, charge )
            call get_test_charge ( tvan,  atom_name, residue_name, charge )

            !if ( test_charge ) then
            if ( charge /= 0. ) then
                ncharge = ncharge + 1
                tot_charge = tot_charge + charge
                write ( io_unit2, fmt='(a54,6x,f7.3)') string, charge
            end if
        endif

        read ( io_unit, '(a)', iostat=ierr ) string
    end do

    write(*,*) 'Number of atoms ', nat
    write(*,*) 'Number of test charges ', ncharge
    write(*,*) 'total charge           ', tot_charge
    write(*,*)

    write( io_unit2, fmt='("REMARK Charges for DNA/RNA are only placed on P")' )
    write( io_unit2, fmt='("REMARK Please modify tools/ECM/ecm_mainsubroutine.f90 if this")' )
    write( io_unit2, fmt='("REMARK is not sufficient")' )
    write( io_unit2, fmt='("REMARK read from pdb file           : ",a )' ) trim ( filename )
    write( io_unit2, fmt='("REMARK # of test charges found there: ",i8 )' ) ncharge
    write( io_unit2, fmt='("REMARK total test charge            : ",f8.3)' ) tot_charge

    close ( io_unit )
    close ( io_unit2 )

    call delete_van ( tvan )

end subroutine sub_mksites

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!   subroutine for ecm_expand ( include ecm_regularize )

!> sub_expand
!!
!! Use for effective charge determination
!! @param list_param_ecm : parameters for ecm, defined in \ref mod_parameters_ecm::param_ecm "param_ecm"
subroutine sub_expand ( list_param_ecm )
    USE mod_grid
    USE mod_parameters_ecm
    IMPLICIT NONE
    type ( param_ecm ) :: list_param_ecm

    interface
        subroutine read_sites ( pos_site, qT, list_param_ecm )
            USE mod_parameters_ecm
            IMPLICIT NONE
            real ( kind=8 ), dimension ( :,: ), allocatable :: pos_site
            real ( kind=8 ), dimension ( : ), allocatable :: qT
            type ( param_ecm ) :: list_param_ecm
        end subroutine read_sites

        subroutine write_sites ( qE, list_param_ecm, opt_error )
            USE mod_parameters_ecm
            IMPLICIT NONE
            real ( kind=8 ), dimension (:) :: qE
            type ( param_ecm ) :: list_param_ecm
            real ( kind=8 ), optional :: opt_error
        end subroutine write_sites

        subroutine write_matrix ( list_param_ecm, a, qT, q0, qE, norm_grid, error )
            USE mod_parameters_ecm
            IMPLICIT NONE
            real ( kind=8 ), dimension ( :,: ), intent ( in ) :: a
            real ( kind=8 ), dimension ( : ), intent ( in ) :: qT, q0, qE
            type ( param_ecm ), intent ( in ) :: list_param_ecm
            real ( kind=8 ), intent ( in ) :: norm_grid, error
        end subroutine write_matrix

        subroutine make_skin ( grid, excl_grid, skin_grid, list_param_ecm )
            USE mod_grid
            USE mod_parameters_ecm
            IMPLICIT NONE
            type ( type_grid ), intent ( in ) :: grid
            integer ( kind=1 ), dimension ( :,:,: ), allocatable :: excl_grid, skin_grid
            type ( param_ecm ) :: list_param_ecm
        end subroutine make_skin

        subroutine rearra ( grid, skin_grid, grid_linear, iar, jar, kar )
            USE mod_grid
            IMPLICIT NONE
            type ( type_grid ), intent ( in ) :: grid
            integer ( kind=1 ), dimension (:,:,:), intent ( in ) :: skin_grid
            real ( kind=8 ), dimension ( : ), intent ( out ) :: grid_linear
            integer, dimension ( : ), intent ( out ) :: iar, jar, kar
        end subroutine rearra

        subroutine  overlapping(grida, gridb, over )
            IMPLICIT NONE
            real ( kind=8 ), dimension ( : ), intent ( in ) :: grida, gridb
            real ( kind= 8 ), intent ( out ) :: over
        end subroutine overlapping

        subroutine coulomb ( coul, tmp_pos, iar, jar, kar, grid, nb_points, fact, deb )
            USE mod_grid
            IMPLICIT NONE
            type ( type_grid ) , intent ( in ) :: grid
            real ( kind=8 ), dimension ( nb_points ) :: coul
            integer, dimension ( nb_points ), intent ( in ) :: iar, jar, kar
            real ( kind=8 ), dimension ( 3 ) :: tmp_pos
            integer, intent ( in ) :: nb_points
            real ( kind=8 ), intent ( in ) :: fact, deb
        end subroutine coulomb

        subroutine sub_regularize ( a0, qT, q0, list_param_ecm, norm_grid )
            USE mod_parameters_ecm
            IMPLICIT NONE
            real ( kind=8 ),dimension (:,:), intent ( in ) :: a0
            real ( kind=8 ), dimension (:), intent ( in ) :: qT, q0
            real ( kind=8 ), intent ( in ) :: norm_grid
            type ( param_ecm ), intent ( in ) :: list_param_ecm
        end subroutine sub_regularize

        subroutine lsolve ( a, nsite, q_tmp )
            IMPLICIT NONE
            real ( kind=8 ), dimension ( nsite, nsite ) :: a
            real ( kind=8 ), dimension ( nsite ), intent ( out ) :: q_tmp
            integer, intent ( in ) :: nsite
        end subroutine lsolve

    end interface

    real ( kind = 8 ), parameter :: fact = 332.12793, ais2db = 1.0807858E-04
    real ( kind = 8 ) :: deb

    ! electrostatic grid
    type ( type_grid ) :: grid
    ! position of the sites
    real ( kind=8 ), dimension ( :,: ), allocatable :: pos_site
    ! test charges, eff. charges
    real ( kind=8 ), dimension ( : ), allocatable :: qT, qE
    ! for solving equation, charge to iterate
    real ( kind=8 ), dimension ( : ), allocatable :: q0
    ! matrix for equation
    real ( kind=8 ), dimension ( :, : ), allocatable :: a, a0

    ! grid for skin
    integer ( kind=1 ), dimension ( :,:,: ), allocatable :: excl_grid, skin_grid
    ! linear electro grid, with component in Xar
    real ( kind=8 ), dimension ( : ), allocatable :: grid_linear
    integer, dimension ( : ), allocatable :: iar, jar, kar
    ! tmp_array, could be in function
    real ( kind=8 ), dimension ( 3 ) :: tmp_pos
    ! coulomb array
    real ( kind=8 ), dimension ( : ), allocatable :: coul1
    real ( kind=8 ), dimension ( : ), allocatable :: coul2
    ! store..
    real ( kind=8 ), dimension ( : ), allocatable :: stor

    real ( kind=8 ) :: overlap, error, norm_grid
    integer :: i, i2, j, icoul

    ! check
    if ( ( list_param_ecm % reg_level /= -9999 ) .and. ( list_param_ecm % reg_charge /= -9999. ) ) then
        write(*,*) 'Cannot specify reg_charge and reg_level at the same time '
        stop 1
    end if

    ! assign constant
    deb = sqrt ( list_param_ecm % iostr * ais2db )
    list_param_ecm % deb = deb

    ! read sites in utils_ecm
    call read_sites( pos_site, qT, list_param_ecm )

    ! read electrostatic grid
    write(*,*)
    write(*,*) 'Read electrostatic grid '
    call read_grid ( grid, list_param_ecm % grid_name )
    call info_grid ( grid )
    write(*,*)

    !! allocate exclusion and skin_grid
    allocate ( excl_grid ( grid % size_x, grid % size_y, grid % size_z ) )
    allocate ( skin_grid ( grid % size_x, grid % size_y, grid % size_z ) )
    excl_grid = 0
    skin_grid = 0

    ! make skin
    call  make_skin( grid, excl_grid, skin_grid, list_param_ecm )

    ! angular restriction option, in last December 2007
    !IF (arestri) CALL sangrr(h,xo,yo,zo,im,jm,km)
    !CALL rearra(stor,grids,im,jm,km,h,xo,yo,zo,iar,jar,kar,npoi)

    ! before npoimx
    allocate ( grid_linear ( list_param_ecm % nb_points_skin ), iar ( list_param_ecm % nb_points_skin ) )
    allocate ( jar ( list_param_ecm %  nb_points_skin ), kar ( list_param_ecm % nb_points_skin ) )
    grid_linear = 0.d0
    iar = 0
    jar = 0
    kar = 0

    !nb_points_skin = list_param_ecm %  nb_points_skin
    !! linearize grid and keep only necessary data
    call rearra ( grid, skin_grid, grid_linear, iar, jar, kar )

    !!grids = grid_linear
    call overlapping( grid_linear, grid_linear, overlap )
    norm_grid = overlap
    write(6,*) 'Norm of grid on skin is ', norm_grid

    write(6,*)
    write(6,*) 'Coulomb-Debye potentials tabulating now... wait... nsite ', list_param_ecm % nsite

    allocate ( coul1 ( list_param_ecm % nb_points_skin ) )
    allocate ( coul2 ( list_param_ecm % nb_points_skin ) )
    coul1 = 0.d0
    coul2 = 0.d0

    !new code, when grid keep grid, here creates new stor
    allocate ( stor ( list_param_ecm % nsite * list_param_ecm % nb_points_skin ) )
    stor = 0.d0

! calculate the Coulomb grids
    do i = 1, list_param_ecm % nsite
        tmp_pos = pos_site ( :, i )

        call coulomb ( coul1, tmp_pos, iar, jar, kar, grid, list_param_ecm % nb_points_skin, fact, deb )

        call overlapping ( coul1, coul1, overlap )
        write(6,*) 'Norm of coulomb grid # ,is ', overlap
        !! what's that ??
        !! here use stor, original code was the electro grid, linearized
        !! new code, when grid keep grid, here creates new stor
        do j = 1, list_param_ecm % nb_points_skin
            icoul = ( i-1 ) * list_param_ecm % nb_points_skin + j
            stor ( icoul ) = coul1 ( j )
        end do
    end do

    write(6,*)
    write(6,*) 'Constructing overlap matrix... wait...'
    write(6,*) 'Rows done:'

    allocate ( q0 ( list_param_ecm % nsite ), qE ( list_param_ecm % nsite ) )
    allocate ( a0 ( list_param_ecm % nsite, list_param_ecm % nsite ), a ( list_param_ecm % nsite, list_param_ecm % nsite ) )
    q0= 0.d0
    qE = 0.d0
    a0= 0.d0
    a = 0.d0

    do i  = 1, list_param_ecm % nsite

        if ( mod ( i, 26 ) == 0 ) then
            write(*, fmt='(i3)' ) i
        else
            write(*, advance='no', fmt='(i3)' ) i
        end if

        do j = 1, list_param_ecm % nb_points_skin
            icoul = ( i -1 ) * list_param_ecm % nb_points_skin + j
            coul1 ( j ) = stor ( icoul )
        end do

        call overlapping ( coul1, grid_linear, overlap )
        qE ( i ) = overlap
        q0 ( i ) = overlap

        do i2 = 1, list_param_ecm % nsite
            do  j = 1, list_param_ecm % nb_points_skin
                icoul = ( i2 -1 ) *  list_param_ecm % nb_points_skin + j
                coul2( j ) = stor ( icoul )
            end do
            call overlapping ( coul1, coul2, overlap )
            a ( i, i2 ) = overlap
            a0 ( i, i2 ) = overlap
        end do

    end do

    write(*,*)
    write(*,*)
    write(6,*) 'Solving linear equation Aq=b...'

    !! a is modified is modified in this function, a0 is the copy of the original matrix
    call lsolve ( a, list_param_ecm % nsite, qE )
    deallocate ( a )

    ! Calculate error
    error = 0.d0
    do i = 1, list_param_ecm % nsite
        error = error - 2.d0 * q0 ( i ) * qE ( i )
        do i2 = 1, list_param_ecm % nsite
            error = error + qE ( i ) * qE ( i2 ) * a0 ( i, i2 )
        end do
    end do

    error = 1.d0 + error / norm_grid
    write(*,*) 'Estimated relative error: ',error
    write(*,*)

    !! write matrix, only if reg_charge is -9999 or 0.
    if ( ( list_param_ecm % reg_charge == -9999. ) .and. ( list_param_ecm % reg_level == -9999 ) ) then
        write(*,*)
        write(*,*) 'Write matrix for recovering effective charges with ecm_mkreglev'
        write(*,*) 'param % effective charge ', list_param_ecm % eff_charge_name
        call write_matrix ( list_param_ecm, a0, qT, q0, qE, norm_grid, error )
    end if

!!!!!!!!!!!!!!!!!!!!!!!!!! End of ecm_expand, begin ecm_regularize

    call sub_regularize ( a0, qT, q0, list_param_ecm, norm_grid )

    !! delete and deallocate
    deallocate ( excl_grid, skin_grid )
    deallocate ( pos_site, qT, qE )
    deallocate ( iar, jar, kar )

    call delete_grid ( grid )

end subroutine sub_expand

!> Make the regularization procedure
subroutine sub_regularize ( a0, qT, q0, list_param_ecm, norm_grid )
    USE mod_parameters_ecm
    IMPLICIT NONE

    interface

        subroutine compute_Echarge_at_level ( iomit, nsite, eigenvectors, a0, qn, q0, ban, bt, norm_grid, epssol, qT, disp1 )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:), intent ( in ) :: eigenvectors, a0
            real ( kind=8 ), dimension (:), intent ( out ) :: qn
            real ( kind=8 ), dimension (:), intent ( in ) :: q0, ban, bt, qT
            real ( kind=8 ), intent ( in ) :: norm_grid, epssol
            real ( kind=8 ), intent ( out ) :: disp1
            integer, intent ( in ) :: iomit, nsite
        end subroutine compute_Echarge_at_level

        subroutine write_sites ( qE, list_param_ecm, opt_error )
            USE mod_parameters_ecm
            IMPLICIT NONE
            real ( kind=8 ), dimension (:) :: qE
            type ( param_ecm ) :: list_param_ecm
            real ( kind=8 ), optional :: opt_error
        end subroutine write_sites
    end interface

    real ( kind=8 ),dimension (:,:), intent ( in ) :: a0
    real ( kind=8 ), dimension (:), intent ( in ) :: qT, q0
    type ( param_ecm ), intent ( in ) :: list_param_ecm
    real ( kind=8 ), intent ( in ) :: norm_grid

    ! local array
    ! for regularization, diagonalisation of the matrix
    real ( kind=8 ), dimension ( : ), allocatable :: eigenvalues, btt, qan, bt, ban, qn
    real ( kind=8 ), dimension (:,:), allocatable :: eigenvectors

    real ( kind=8 ) :: disp1
    integer :: i, i2, nsite, iomit, reg_level
    logical :: write_ch, seek_lev

    nsite = list_param_ecm % nsite
    write_ch = .true. ! always true for the first time, then false
    seek_lev = .true. ! default true, check for regch

    reg_level = list_param_ecm % reg_level

    if ( reg_level == -9999 ) then
        seek_lev = .false.
    end if

    ! write(*,*) 'sub_regularize seek_lev ', seek_lev, ' reg_level ', reg_level, ' lparam %reg_charge ', list_param_ecm % reg_charge
    write(*,*)

    allocate ( eigenvalues ( nsite ), eigenvectors ( nsite, nsite ) )

    call diagonalize ( nsite, a0, eigenvalues, eigenvectors )
    !!!!!!!!!!!!!!!! just for test
    !do i= 1, list_param_ecm % nsite
    !    do i2 = 1, list_param_ecm % nsite
    !        if ( a(i,i2) /= a0(i,i2) ) then
    !            write ( *,* ) ' matrix a modified !! '
    !            CYCLE
    !        end if
    !    end do
    !end do
    !!!!!!!!!!!!!!!!
    allocate ( btt ( nsite ), bt ( nsite ), qan ( nsite ), ban ( nsite ), qn ( nsite ) )

    do i = 1, nsite
        btt ( i ) = 0.d0
        qan ( i ) = 0.d0
        do i2 = 1, nsite
            btt ( i ) = btt ( i ) + q0 ( i2 ) * eigenvectors ( i2, i )
            qan ( i ) = qan ( i ) + qT ( i2 ) * eigenvectors ( i2, i )
        end do
    end do

    ! omit sequentially lowest eigenvalue components
    do i = 1, nsite
        bt ( i ) = btt ( i ) / eigenvalues ( i )
        ban ( i ) = qan ( i ) / list_param_ecm % epssol
    end do

    write(*,*) '#  N     charge    error     charge    sqrt of      net'
    write(*,*) '#  N      RMSD                RM1D      error     charge'

!!! compute and print values for all steps
    do iomit = 1, nsite

        call compute_Echarge_at_level ( iomit, nsite, eigenvectors, &
            a0, qn, q0, ban, bt, norm_grid, list_param_ecm % epssol, qT, disp1 )

        ! give reg_lev as input
        if ( seek_lev ) then
            if ( reg_level == iomit ) then
                if ( write_ch ) then
                    write(*,*)
                    write(*,*) 'write iteration ', iomit, ' reg_level given ', reg_level
                    write(*,*)
                    call write_sites ( qn, list_param_ecm )
                    write_ch = .false.
                end if
            end if
        ! need an other option, if not want to print
        ! else use regch
        else
            if ( abs ( disp1 ) < list_param_ecm % reg_charge ) then
                if ( write_ch ) then
                    Write(*,*)
                    write(*,*) 'write iteration ', iomit, ' reg_charge given ', list_param_ecm % reg_charge
                    write(*,*)
                    call write_sites ( qn, list_param_ecm )
                    write_ch = .false.
                end if
            end if
        end if

    ! end do iomit
    end do

    deallocate ( eigenvalues, qn, btt, qan, bt, ban, eigenvectors )

 end subroutine sub_regularize

!> Create skin around the solute
!! Called 2 times \ref make_one_skin with different parameters to define a skin around the solute
subroutine make_skin ( grid, excl_grid, skin_grid, list_param_ecm )
    USE mod_grid
    USE mod_parameters_ecm
    IMPLICIT NONE
    type ( type_grid ), intent ( in ) :: grid
    integer ( kind=1 ), dimension ( :,:,: ), allocatable :: excl_grid, skin_grid
    type ( param_ecm ) :: list_param_ecm

    interface
        subroutine make_one_skin ( grid, list_param_ecm, natoms, nb_points, excl_grid, skin_grid, bool_first )
            USE mod_grid
            USE mod_parameters_ecm
            IMPLICIT NONE
            type ( type_grid ), intent ( in ) :: grid
            type ( param_ecm ), intent ( in ) :: list_param_ecm
            integer ( kind=1 ), dimension ( :,:,: ), allocatable :: excl_grid, skin_grid
            integer :: nb_points, natoms
            logical :: bool_first
        end subroutine make_one_skin
    end interface

    integer :: tmp_natoms, tmp_nb_points

    write(*,*) 'Entry make_skin'
    write(*,*) 'Constructing fitting region '


    write(*,*) 'Call first skin true'
    ! if first allocate exclusion grid
    call make_one_skin ( grid, list_param_ecm, tmp_natoms, tmp_nb_points, excl_grid, skin_grid, .true. )
    write(*, fmt='("Number of atoms used for excl grid :",i8)' ) tmp_natoms
    write(*, fmt='("Number of points assigned value 1 :",i8)' ) tmp_nb_points
    list_param_ecm % natoms = tmp_natoms
    list_param_ecm % nb_points = tmp_nb_points

    write(*,*) 'Call second skin false'
    call make_one_skin ( grid, list_param_ecm, tmp_natoms, tmp_nb_points, excl_grid, skin_grid, .false. )
    write(*, fmt='("Number of atoms used for excl+skin grid :",i8)' ) tmp_natoms
    write(*, fmt='("Number of points assigned value 1 :",i8)' ) tmp_nb_points

    list_param_ecm % nb_points_skin = tmp_nb_points
    !write(*,*) 'End Constructing fitting region '
end subroutine make_skin

!> make_one_skin
!! need to re read pdb, because van argument
!! Second call probsk is increased by skin
subroutine make_one_skin ( grid, list_param_ecm, natoms, nb_points, excl_grid, skin_grid, bool_first )
    USE mod_grid
    USE mod_parameters_ecm
    USE mod_van
    IMPLICIT NONE
    type ( type_grid ), intent ( in ) :: grid
    type ( param_ecm ), intent ( in ) :: list_param_ecm
    integer ( kind=1 ), dimension ( :,:,: ), allocatable :: excl_grid, skin_grid
    integer :: nb_points, natoms
    logical :: bool_first

    ! same as other, could use as a parameter to avoid to reload each time
    type ( type_van ) :: tvan

    character*80 :: string
    character*3 :: residue_name
    character*2 :: a2n
    ! correspond to previous. to x1, x2, x3 and x1a, x2a and x3a
    real ( kind= 8 ), dimension ( 3 ) :: pos_atom, pos_atom_shift
    real ( kind=8 ) :: dexcl, hdexcl, probsk, rexmax,rexcl, rexcl2
    real ( kind=8 ) :: zk, ztemp, yj, ytemp, xi, xtemp, dist2
    integer :: k, kh, kk, j, jh, jj, ii, i, ih
    integer :: io_unit, ierr, klim

    io_unit = 51
    natoms = 0
    nb_points = 0

    ! common for both skins
    dexcl = grid % h
    hdexcl = dexcl / 2.d0
    rexcl = 0.d0
    rexcl2 = 0.d0

    pos_atom_shift = 0.

    !write(*,*) 'Entry make_one_skin'
    !write(*,*) 'option bool_first ', bool_first

    ! first skin
    if ( bool_first .eqv. .true. ) then
        probsk = list_param_ecm % probes ! needed for van
    ! second skin
    else
        probsk = list_param_ecm % probes + list_param_ecm % skin
    end if
    !write(*,*) 'probsk ',probsk

    ! init the new module
    if ( allocated ( tvan % array_atom ) .eqv. .false. ) then
        call init_van ( tvan )
        !write(*,*) 'ok done init'
        call print_array_atom ( tvan )
    !else
    !    write(*,*) 'van already initialized'
    end if

    !! common to both grids
    rexmax = 1.9d0 + probsk
    !! klim integer here !
    klim = int ( ( rexmax + dexcl * 1.5d0 ) / dexcl )

    !write(*,*) 'probsk ', probsk
    !write(*,*) 'klim ', klim,' rexmax ', rexmax,' dexcl ', dexcl, 'hdexcl ', hdexcl

    open (io_unit, file= list_param_ecm % pdb_name, form= 'formatted', status='old', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open file ', trim ( list_param_ecm % pdb_name )
        stop 1
    end if

    read ( io_unit, fmt='(a)', iostat = ierr ) string
    if ( ierr /= 0 ) then
        write(*,*) 'pdb file is empty ???'
        return
    end if

LOOPPDB : &
    do while ( ierr == 0 )

        if ( (string(1:4)/='ATOM').and.(string(1:6)/='HETATM') ) then
            read ( io_unit, fmt='(a)', iostat = ierr ) string
            CYCLE LOOPPDB
        end if

        natoms = natoms + 1
        ! original
        !read ( string(1:54), fmt='(13X,A1,16X,3f8.3)' ) an, pos_atom
        ! residue name not yet used for VdW
        read ( string, fmt='(13X,A2,2X,A3,10X,3f8.3)' ) a2n, residue_name, pos_atom
        !write(*,*) 'test new atom name a2n!',trim(a2n),'!'
        !write(*,*) 'test new residue_name!',trim(residue_name),'!'
        !write(*,*) 'test new pos_atom ', pos_atom
        
        ! modified version in expand all
        ! call van1 ( an, rexcl2 )
        call get_VdW( tvan, a2n, rexcl2 )
        !write(*,*) 'rexcl2 ', rexcl2

        rexcl2 = ( rexcl2 + probsk )**2 ! done in van
        rexcl = sqrt ( rexcl2 )
        !write(*,*) 'rexcl2 ',rexcl2, ' rexcl ', rexcl

        ! make pos_shift
        pos_atom_shift = pos_atom + hdexcl

        ! Find home cell ih, jh, kh
        ih = int ( ( pos_atom_shift ( 1 ) - grid % origin ( 1 ) ) / dexcl )
        jh = int ( ( pos_atom_shift ( 2 ) - grid % origin ( 2 ) ) / dexcl )
        kh = int ( ( pos_atom_shift ( 3 ) - grid % origin ( 3 ) ) / dexcl )
        ! write(*,*) 'home cell ', ih, jh, kh
        ! write(*,*) 'klim ', klim

LOOP_K: do kk = -klim, klim
            k = kk + kh
            if ( ( k <= 0 ) .or. ( k > grid % size_z ) ) CYCLE LOOP_K
            ! strange comment in expand.f
            zk = dfloat ( k ) * dexcl + grid % origin ( 3 )
            ztemp = zk - pos_atom ( 3 )
            if ( ztemp > rexcl ) CYCLE LOOP_K

    LOOP_J: do jj = -klim, klim
                j = jj +jh
                if ( ( j <= 0 ) .or. ( j > grid % size_y ) ) CYCLE LOOP_J
                yj = dfloat ( j ) * dexcl + grid % origin ( 2 )
                ytemp = yj - pos_atom ( 2 )
                if ( ytemp > rexcl ) CYCLE LOOP_J

       LOOP_I:  do ii = -klim, klim
                    i = ii + ih
                    if ( ( i <= 0 ) .or. ( j > grid % size_x ) ) CYCLE LOOP_I

                    !!!!!! test if already set  to 1
                    if ( excl_grid ( i,j,k ) == 1 ) CYCLE LOOP_I
                    !! first call, skin grid is fully false, so change nothing
                    if ( skin_grid ( i,j,k ) == 1 )  CYCLE LOOP_I

                    xi = dfloat ( i ) * dexcl + grid % origin ( 1 )
                    xtemp = xi - pos_atom ( 1 )

                    dist2 = xtemp*xtemp + ytemp*ytemp + ztemp*ztemp
                    if ( dist2 > rexcl2 ) CYCLE LOOP_I

                    !!!!  else fill with one
                    !! first call
                    if ( bool_first ) then
                        excl_grid ( i,j, k) = 1
                    else
                    !! second call
                        skin_grid ( i,j,k ) = 1
                    end if
                    nb_points = nb_points + 1
                end do LOOP_I

            end do LOOP_J

        end do LOOP_K

        read ( io_unit, fmt='(a)', iostat = ierr ) string

    end do LOOPPDB

    close ( io_unit )

    if ( allocated ( tvan % array_atom ) ) then
        call delete_van ( tvan )
    end if

end subroutine make_one_skin

!> Rearrange.\n
!! keep only points where skin_grid is 1, put it in a unidimensional array
subroutine rearra ( grid, skin_grid, grid_linear, iar, jar, kar )
    USE mod_grid
    IMPLICIT NONE
    type ( type_grid ), intent ( in ) :: grid
    integer ( kind=1 ), dimension (:,:,:), intent ( in ) :: skin_grid
    real ( kind=8 ), dimension ( : ), intent ( out ) :: grid_linear
    integer, dimension ( : ), intent ( out ) :: iar, jar, kar

    integer :: nb_points

    real ( kind=8 ) :: gmax, gmin
    integer :: i,j,k !, imjm

    gmax=-1.e+20
    gmin= 1.e+20

! original comment
! make a spherical grid
!
! ijk transform to n as n=i+(j-1)*im+(k-1)*im*jm
! so could keep 3 d grids until

    nb_points = 0
    do i = 2, grid % size_x -1
        do j = 2, grid % size_y -1
            do k = 2, grid % size_z -1

                if ( skin_grid ( i, j, k ) == 1 ) then
                    nb_points = nb_points + 1
                    iar ( nb_points ) = i
                    jar ( nb_points ) = j
                    kar ( nb_points ) = k
                    grid_linear ( nb_points ) = grid % p_grid % pgrid_real ( i,j,k )

                    if ( grid_linear ( nb_points ) > gmax ) gmax = grid_linear ( nb_points )
                    if ( grid_linear ( nb_points ) < gmin ) gmin = grid_linear ( nb_points )
                end if

            end do
        end do
    end do

    write(*,fmt='("rearrangement routine, grid min/max :",2e12.3)' ) gmin, gmax
    !write(*,*) 'nb_points ', nb_points

end subroutine rearra

!> Compute the overlapping
subroutine  overlapping ( grida, gridb, over )
    IMPLICIT NONE
    real ( kind=8 ), dimension ( : ), intent ( in ) :: grida, gridb
    real ( kind= 8 ), intent ( out ) :: over

    integer :: i

    over = 0.d0
    !write(*,*) 'size grid', size ( grida )
    do i = 1, size ( grida )
        over = over + grida ( i ) * gridb ( i )
    end do

end subroutine overlapping

!> Coulomb terms
subroutine coulomb ( coul, tmp_pos, iar, jar, kar, grid, nb_points, fact, deb )
    USE mod_grid
    IMPLICIT NONE
    type ( type_grid ) , intent ( in ) :: grid
    real ( kind=8 ), dimension ( nb_points ) :: coul
    integer, dimension ( nb_points ), intent ( in ) :: iar, jar, kar
    real ( kind=8 ), dimension ( 3 ) :: tmp_pos
    integer, intent ( in ) :: nb_points
    real ( kind=8 ), intent ( in ) :: fact, deb

    real ( kind=8 ) :: xx, yy, zz, dist
    integer :: i

    do i = 1, nb_points
        xx = iar ( i ) * grid % h + grid % origin ( 1 )
        yy = jar ( i ) * grid % h + grid % origin ( 2 )
        zz = kar ( i ) * grid % h + grid % origin ( 3 )

        dist = sqrt ( ( tmp_pos ( 1 ) - xx )**2 + ( tmp_pos ( 2 ) - yy )**2 + ( tmp_pos ( 3 ) - zz )**2 )
        coul ( i ) = fact / dist * exp ( -deb * dist )
    end do

end subroutine coulomb

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
