!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file ecm_mkreglev.f90
!!
!! Program ecm_mkreglev

!> Create the effective charges
!!
!! \ingroup ECM
!! Last step in the process
!!
!! Recreate effective charge from the computed matrix\n
!! Similar to \ref ecm_regularize, but stop at a particular iteration in the regularization
program ecm_mkreglev
    USE mod_parameters_ecm
    IMPLICIT NONE

    interface
        subroutine diagonalize ( nsite, a0, eigenvalues, eigenvectors )
            IMPLICIT NONE
            real ( kind=8 ), dimension ( nsite, nsite) :: a0, eigenvectors
            real ( kind=8 ), dimension ( nsite ) :: eigenvalues
            integer :: nsite
        end subroutine diagonalize

        subroutine write_sites ( qE, list_param_ecm, opt_error )
            USE mod_parameters_ecm
            IMPLICIT NONE
            real ( kind=8 ), dimension (:) :: qE
            type ( param_ecm ) :: list_param_ecm
            real ( kind=8 ), optional :: opt_error
        end subroutine write_sites

        subroutine read_matrix ( list_param_ecm, a, qT, q0, norm_grid )
            USE mod_parameters_ecm
            IMPLICIT NONE
            real ( kind=8 ), dimension ( :,: ), allocatable :: a
            real ( kind=8 ), dimension ( : ), allocatable :: qT, q0
            type ( param_ecm ), intent ( out ) :: list_param_ecm
            real ( kind=8 ), intent ( out ) :: norm_grid
        end subroutine read_matrix

        subroutine sub_regularize ( a0, qT, q0, list_param_ecm, norm_grid )
            USE mod_parameters_ecm
            IMPLICIT NONE
            real ( kind=8 ),dimension (:,:), intent ( in ) :: a0
            real ( kind=8 ), dimension (:), intent ( in ) :: qT, q0
            real ( kind=8 ), intent ( in ) :: norm_grid
            type ( param_ecm ), intent ( in ) :: list_param_ecm
        end subroutine sub_regularize

        subroutine compute_Echarge_at_level ( iomit, nsite, eigenvectors, a0, qn, q0, ban, bt, norm_grid, epssol, qT, disp1 )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:), intent ( in ) :: eigenvectors, a0
            real ( kind=8 ), dimension (:), intent ( out ) :: qn
            real ( kind=8 ), dimension (:), intent ( in ) :: q0, ban, bt, qT
            real ( kind=8 ), intent ( in ) :: norm_grid, epssol
            real ( kind=8 ), intent ( out ) :: disp1
            integer, intent ( in ) :: iomit, nsite
        end subroutine compute_Echarge_at_level
    end interface

    real ( kind=8 ), dimension (:,:), allocatable :: a0, eigenvectors
    real ( kind=8 ), dimension (:), allocatable :: q0, qT, eigenvalues,  btt, qan, bt, ban, qn

    character*54 :: argument
    real ( kind=8 ) :: norm_grid, disp1
    integer :: nb_argument, i, i2, nsite
    integer :: iu_matrix
    logical :: bool_level, bool_charge

    ! input parameters
    type ( param_ecm ) :: list_param_ecm

    bool_level = .false.
    bool_charge = .false.
    iu_matrix = 3

    nb_argument = command_argument_count()

    if ( nb_argument == 0 ) then
        write(*,*)
        write(*,*) 'HELP'
        write(*,*)
        write(*,*) 'ecm_mkreglev [-reg_level 25] [-reg_charge 1.0] '
        write(*,*)
        write(*,*) 'To use for generating effective charges '
        write(*,*) 'at a given iteration ( reg_level ) OR at a charge value ( reg_charge )'
        write(*,*) 'Use only one of both options'
        write(*,*)
        write(*,*) 'To use when reg_charge = 0 in ecm.in or -rc or -rl option missing in ecm_all'
        write(*,*) 'Need matrix file generated by ecm_expand or ecm_all'
        write(*,*)
        write(*,*) 'To generate effective charges with intermediate steps:'
        write(*,*) 'ecm_mksites p1_noh.pdb'
        write(*,*) 'ecm_expand ecm.in'
        write(*,*) 'optional ecm_mkreglev -reg_level 35'
        write(*,*)
        write(*,*) 'In one step:'
        write(*,*) 'ecm_all -fp p1_noh.pdb -fg ep1.bin.grd -is 150. -rc 1.0 -fe p1_noh.echa'
        write(*,*) 'optional ecm_mkreglev -reg_level 35'
        write(*,*)
        stop
    endif

    call getarg ( 1, argument )
    if ( argument(1:8) == "-reg_lev" ) then
        call getarg( 2, argument )
        read ( argument, fmt='(i4)' ) list_param_ecm % reg_level
        list_param_ecm % reg_charge = -9999.
        bool_level = .true.
    else if ( argument(1:11) == '-reg_charge' ) then
        call getarg( 2, argument )
        read ( argument, * ) list_param_ecm % reg_charge
        list_param_ecm % reg_level = -9999
        bool_charge = .true.
    else
        write(*,*) 'first argument is wrong'
        stop 1
    end if

    write(*,fmt='("regularize at level  :", i5)' ) list_param_ecm % reg_level
    write(*,fmt='("regularize at charge :", f10.3)' ) list_param_ecm % reg_charge
    !! read the matrix file, allocate array and read echa filename
    call read_matrix ( list_param_ecm, a0, qT, q0, norm_grid )
    nsite = list_param_ecm % nsite

    !! should be optional
    list_param_ecm % test_charge_name =  list_param_ecm % eff_charge_name
    !write(*,*) 'before create_filename_new ', list_param_ecm % test_charge_name
    call create_filename_new ( list_param_ecm % test_charge_name, "tcha" )
    !write(*,*) 'after ', list_param_ecm % test_charge_name

    !write(*,*) 'regularisation level ', list_param_ecm % reg_level
    write(*,*) 'output file ', list_param_ecm % eff_charge_name
    write(*,*) 'test charge file ', list_param_ecm % test_charge_name

    !if reg_level given only computation at one level
    if ( bool_level ) then
        allocate ( eigenvalues ( nsite ), eigenvectors ( nsite, nsite ) )
        !! function in diagonalisation_ecm.f90
        !! diagonalisation do not modify input a0
        call diagonalize ( nsite, a0, eigenvalues, eigenvectors )

        !! initialization, common to ecm_mkregularize
        allocate ( btt ( nsite ), bt ( nsite ), qan ( nsite ), ban ( nsite ), qn ( nsite ) )

        do i = 1, nsite
            btt ( i ) = 0.d0
            qan ( i ) = 0.d0
            do i2 = 1, nsite
                btt ( i ) = btt ( i ) + q0 ( i2 ) * eigenvectors ( i2, i )
                qan ( i ) = qan ( i ) + qT ( i2 ) * eigenvectors ( i2, i )
            end do
        end do

       !omit sequentially lowest eigenvalue components
        do i = 1, nsite
            bt ( i ) = btt ( i ) / eigenvalues ( i )
            ban ( i ) = qan ( i ) / list_param_ecm % epssol
        end do

        !! iomit is fixed now to reg_level
        write(*,*) '#  N     charge    error     charge    sqrt of      net'
        write(*,*) '#  N      RMSD                RM1D      error     charge'

        call compute_Echarge_at_level ( list_param_ecm % reg_level, nsite, eigenvectors, a0, qn, q0, ban, bt, &
                                        norm_grid, list_param_ecm % epssol, qT, disp1 )
        call write_sites ( qn, list_param_ecm )
    end if

    ! else need full iteration
    if ( bool_charge ) then
        call sub_regularize ( a0, qT, q0, list_param_ecm, norm_grid )
    end if

end program ecm_mkreglev

