!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file ecm_mksites.f90
!!
!! Program ecm_mksites

!> Make the test charges
!! \ingroup ECM

!! Create an intermediate file (*.tcha ) with test charges from the pdb
!!
!! Reads in pdb file, finds out OE. OD. NZ NE. atoms of
!! GLU, ASP, LYS, ARG, respectively, and saves their coordinates,
!! assign test charges to them, and count their number
!! \todo add external file
program ecm_mksites
    IMPLICIT NONE

    character*128 :: filename, out_fname
    integer :: nb_argument

    !nb_argument = iargc ()
    nb_argument = command_argument_count()
    if ( nb_argument == 0 ) then
        write(*,*)
        write(*,*) 'HELP'
        write(*,*)
        write(*,*) 'ecm_mksites file.pdb [out_file.tcha] '
        write(*,*)
        write(*,*) 'Generate the initial set of test charges necessary for computing the effective charges '
        write(*,*)
        write(*,*) 'input : file.pdb '
        write(*,*) 'optional output: out_file.tcha, if not provided will generate file.tcha'
        write(*,*)
        write(*,*) 'To add new test charge : ions, charged ligand...'
        write(*,*) 'edit subroutine charge_def in ecm_mainsubroutine,f90 and recompile with make tool'
        write(*,*)
        write(*,*) 'To generate effective charges with intermediate steps:'
        write(*,*) 'ecm_mksites p1_noh.pdb'
        write(*,*) 'ecm_expand ecm.in'
        write(*,*) 'optional ecm_mkreglev -reg_level 35'
        write(*,*)
        write(*,*) 'In one step:'
        write(*,*) 'ecm_all -fp p1_noh.pdb -fg ep1.bin.grd -is 150. -rc 1.0 -fe p1_noh.echa'
        write(*,*) 'optional ecm_mkreglev -reg_level 35'
        write(*,*)
        stop
    endif

    call getarg ( 1, filename )
    if ( nb_argument > 1 ) then
        call getarg ( 2, out_fname )
    else
        out_fname = filename
        call create_filename_new ( out_fname, "tcha" )
    endif

    write(*,*)
    write(*,*) 'input filename ', trim ( filename )
    write(*,*) 'output filename ', trim ( out_fname )

    call sub_mksites ( filename, out_fname )

end program ecm_mksites




