!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file utils_ecm.f90
!!
!! Regroup utilities functions for ecm tools

!> Read test_charges (*.tcha)
subroutine read_sites ( pos_site, qT, list_param_ecm )
    USE mod_parameters_ecm
    IMPLICIT NONE
    real ( kind=8 ), dimension ( :,: ), allocatable :: pos_site
    real ( kind=8 ), dimension ( : ), allocatable :: qT
    type ( param_ecm ) :: list_param_ecm

    character*66 :: string
    real ( kind = 8 ) :: qnet
    integer :: iutch = 9, ierr
    integer :: nsite

    write(*,*)
    write(*,*) 'Read test charge'

    qnet = 0.d0

    open ( iutch, file = list_param_ecm % test_charge_name, form='formatted', status='old', iostat = ierr )
    if ( ierr /= 0 ) then
        write( *,* ) 'Cannot open file ', trim ( list_param_ecm % test_charge_name )
        stop 1
    endif

    read ( iutch, '(a)', iostat = ierr ) string
    if ( ierr /= 0 ) then
        write( *,* ) 'error input file ', trim ( list_param_ecm % test_charge_name ), ' is empty'
        stop 1
    endif

    nsite = 0
    !! count the n umber of sites
    do while ( ierr == 0 )
        if ( (string(1:4)=='ATOM').or.(string(1:6)=='HETATM') ) then
            nsite = nsite + 1
        end if
        read ( iutch, '(a)', iostat = ierr ) string
    end do

    rewind ( iutch )
    write(*,*) 'number site ', nsite

    !! allocate array
    allocate ( pos_site ( 3, nsite ), qT ( nsite ) )
    pos_site = 0.d0
    qT = 0.d0

    !! read the data
    nsite = 0
    read ( iutch, '(a)', iostat = ierr ) string
    do while ( ierr == 0 )
        if ( (string(1:4)=='ATOM').or.(string(1:6)=='HETATM') ) then
            nsite = nsite + 1
            read ( string (31:66), fmt='(3f8.3,6X,f7.3)' ) pos_site(:,nsite), qT ( nsite )
            !if ( nsite == 1 ) then
            !    write(*,*) 'test 1', pos_site( :,1 ),qT( 1 )
            !end if
            qnet = qnet + qT ( nsite )
        end if
        read ( iutch, '(a)', iostat = ierr ) string
    end do

    close ( iutch )

    list_param_ecm % nsite = nsite

    write(*, fmt='(i6," sites read from: ",a,/,"net charge: ",f8.3)' ) &
        nsite, trim ( list_param_ecm % test_charge_name ), qnet
    !write(*, fmt='(i6," sites read from: ",a)' ) nsite, trim ( list_param_ecm % test_charge_name )
    !write(*,*) 'qnet ', qnet

end subroutine read_sites

!> Write test_charges
subroutine write_sites ( qE, list_param_ecm, opt_error )
    USE mod_parameters_ecm
    IMPLICIT NONE
    real ( kind=8 ), dimension (:) :: qE
    type ( param_ecm ) :: list_param_ecm
    real ( kind=8 ), optional :: opt_error

    character*67 :: string
    real ( kind=8 ) :: error
    integer :: iutch = 9, iuech = 10, ierr, nsite

    if ( present ( opt_error ) ) then
        error = opt_error
    else
        error = 0.d0
    endif

    !write(*,*) 'Entry write_sites '
    !write(*,*) 'test ', trim(list_param_ecm % test_charge_name)
    !write(*,*) 'eff ', trim(list_param_ecm % eff_charge_name)

    !! open test charge to get the strings

    !already open no !
    open ( iutch, file = list_param_ecm % test_charge_name, form='formatted', status='old', iostat = ierr )
    if ( ierr /= 0 ) then
        write( *,* ) 'Cannot open test charge file ', trim ( list_param_ecm % test_charge_name )
        return
    endif

    !! open new file
    open ( iuech, file = list_param_ecm % eff_charge_name, form='formatted', status='new', iostat = ierr )
    !open ( iuech, file = list_param_ecm % eff_charge_name, form='formatted', status='replace', iostat = ierr )
    if ( ierr /= 0 ) then
        write( *,* ) 'Cannot open effective charge file', trim ( list_param_ecm % eff_charge_name )
        return
    endif

    write (iuech, fmt='("REMARK   Debye-Hueckel parameter (A^-1) :",e25.6)' )  list_param_ecm % deb
    write (iuech, fmt='("REMARK   Solvent dielectric permittivity:",e25.6)' )  list_param_ecm % epssol

    nsite = 0
    !qnet = 0.d0
    read ( iutch, '(a)', iostat = ierr ) string
    if ( ierr /= 0 ) then
        write( *,* ) 'error input file ', trim ( list_param_ecm % test_charge_name ), ' is empty'
        return
    endif

    do while ( ierr == 0 )
        if ( (string(1:4)=='ATOM').or.(string(1:6)=='HETATM') ) then
            nsite = nsite + 1
            write ( string(55:67),fmt='(6x,f7.3)' ) qE ( nsite ) * list_param_ecm % epssol
            write ( iuech, fmt='(a)' ) string
        end if
        read ( iutch, '(a)', iostat = ierr ) string
    end do
    ! write total charge
    write( iuech, fmt='("REMARK Total Effective Charge  :",e25.6)' ) sum( qE) * list_param_ecm % epssol

    if ( error /= 0.d0 ) then
        write( iuech, fmt='("REMARK   Relative fit error of norm     :",e25.6)' ) error
        write( iuech, fmt='("REMARK   Relative fit error of magnitude:",e25.6)' ) sqrt ( error )
    end if

    close ( iuech )
    close ( iutch )

end subroutine write_sites

!> Read matrix
subroutine read_matrix ( list_param_ecm, a, qT, q0, norm_grid )
    USE mod_parameters_ecm
    IMPLICIT NONE
    real ( kind=8 ), dimension ( :,: ), allocatable :: a
    real ( kind=8 ), dimension ( : ), allocatable :: qT, q0
    type ( param_ecm ), intent ( out ) :: list_param_ecm
    real ( kind=8 ), intent ( out ) :: norm_grid

    integer :: iu_matrix = 3, ierr

    open ( iu_matrix, file='matrix', form='formatted', status='old', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open matrix file in read_matrix, file is missing ??'
        write(*,*) 'Rerun ecm_expand or ecm_all with reg_charge = 0'
        stop 1
    end if


    read (iu_matrix,fmt='(i5)') list_param_ecm % nsite

    ! allocate arrays
    allocate ( a ( list_param_ecm % nsite, list_param_ecm % nsite ) )
    allocate ( q0 ( list_param_ecm % nsite ), qT ( list_param_ecm % nsite ) )

    read (iu_matrix,*)
    read (iu_matrix,fmt='(5e16.8)') a
    read (iu_matrix,*)
    read (iu_matrix,fmt='(5e16.8)' ) q0
    read (iu_matrix,*)
    read (iu_matrix,fmt='(e16.8)' ) norm_grid
    read (iu_matrix,*)
    read (iu_matrix,fmt= '(5e16.8)' ) qT
    read (iu_matrix,*)
    read (iu_matrix,fmt='(2e16.8)' ) list_param_ecm % deb, list_param_ecm % epssol
    read (iu_matrix,*)
    read (iu_matrix,fmt= '(a)' ) list_param_ecm % eff_charge_name

    close ( iu_matrix )

end subroutine read_matrix

!> Write matrix
subroutine write_matrix ( list_param_ecm, a, qT, q0, qE, norm_grid, error )
    USE mod_parameters_ecm
    IMPLICIT NONE
    real ( kind=8 ), dimension ( :,: ), intent ( in ) :: a
    real ( kind=8 ), dimension ( : ), intent ( in ) :: qT, q0, qE
    type ( param_ecm ), intent ( in ) :: list_param_ecm
    real ( kind=8 ), intent ( in ) :: norm_grid, error

    integer :: iu_matrix = 3, ierr, i, i2


    open ( iu_matrix, file='matrix', form='formatted', status='replace', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open matrix file in write_matrix'
    end if

    write( iu_matrix,  fmt='( i5 )' ) list_param_ecm % nsite
    write( iu_matrix, * )
    write( iu_matrix,  fmt='(5e16.8)' ) ( ( (a(i,i2)),i=1,list_param_ecm % nsite ),i2=1,list_param_ecm % nsite )
    write( iu_matrix, * )
    write( iu_matrix,  fmt='(5e16.8)' ) q0
    write( iu_matrix, * )
    write( iu_matrix, fmt= '(e16.8)' ) norm_grid
    write( iu_matrix, * )
    write( iu_matrix, fmt= '(5e16.8)' ) qT
    write( iu_matrix, * )
    write( iu_matrix, fmt= '(2e16.8)' ) list_param_ecm % deb, list_param_ecm % epssol
    write( iu_matrix, * )
    write( iu_matrix, fmt='(a)' ) trim ( list_param_ecm % eff_charge_name )
    write( iu_matrix,* )
    write( iu_matrix, fmt='(5e16.8)' ) qE
    write( iu_matrix,* )
    write( iu_matrix, fmt='(5e16.8)' ) error, sqrt ( error) * 100.d0

    close ( iu_matrix )

end subroutine write_matrix

!> Compute the effective charge at a given iteration
!!
!! disp1 only used for checking reg_charge, not in reg_level
subroutine compute_Echarge_at_level ( iomit, nsite, eigenvectors, a0, qn, b, ban, bt, norm_grid, epssol, qT, disp1 )
    !USE mod_parameters_ecm
    IMPLICIT NONE
    real ( kind=8 ), dimension (:,:), intent ( in ) :: eigenvectors, a0
    real ( kind=8 ), dimension (:), intent ( out ) :: qn
    real ( kind=8 ), dimension (:), intent ( in ) :: b, ban, bt, qT
    real ( kind=8 ), intent ( in ) :: norm_grid, epssol
    real ( kind=8 ), intent ( out ) :: disp1
    integer, intent ( in ) :: iomit, nsite

    real ( kind= 8 ) :: disp, qnet, error, diff
    integer :: i, i2

    do i = 1, nsite
        qn ( i ) = 0.d0
        do i2 = 1, nsite
            if ( i2 <= iomit ) then
                qn ( i ) = qn ( i ) + ban ( i2 ) * eigenvectors ( i, i2 )
            else
                qn ( i ) = qn ( i ) + bt ( i2 ) * eigenvectors ( i, i2 )
            end if
        end do
    end do

    disp = 0.d0
    disp1 = 0.d0
    qnet = 0.d0
    do i = 1, nsite
        diff = qn ( i ) * epssol - qT ( i )
        disp = disp + diff**2

        if ( abs ( diff ) > abs ( disp1 ) ) disp1 = diff
        qnet = qnet + qn ( i )
    end do

    !! moved here, will be in function
    disp = dsqrt ( disp / dfloat ( nsite ) )
    qnet = qnet * epssol

!! compute error for each regularized charges
    error = 0.d0
    do i = 1, nsite
        error = error - 2.d0 * qn ( i ) * b ( i )
        do i2 = 1, nsite
            error = error + qn ( i ) * qn ( i2 ) * a0 ( i, i2 )
        end do
    end do

    error = 1.d0 + error / norm_grid
    write (*, fmt='(i5,f10.3,f10.5,f10.4,f10.5,f10.3)' ) iomit, disp, error, disp1, sqrt ( error ), qnet


end subroutine compute_Echarge_at_level
