!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> @file analyze_subroutines.f90
!! module for tools

!> Module containing several subroutines that are used in other tools
MODULE analyze_subroutines
    use mod_record
    use genbox_subroutines
    
    contains
    
    !> Small subroutine that helps in preparing the data for reading in a trajectory, and reading the header
    ! @param traj : input & output, type ( record )
    ! @param traj_file_i : number of the stream which will point to the trajectory file
    ! @param trajf : name of the trajectory file
    ! @param tot_nlines : output, total number of lines available in the trajecotry file
    subroutine prepare_read_traj ( traj, traj_file_i, trajf, tot_nlines )
      IMPLICIT NONE
      type ( record ), intent ( inout ) :: traj
      character ( 128 ), intent ( in ) :: trajf
      integer, intent ( in ) :: traj_file_i
      integer, intent ( out ) :: tot_nlines
      
      logical :: one_conf, one_run
      integer :: ierr
        
      one_conf = .true.
      one_run = .true.
      
      nullify ( traj % pcomplexe, traj % ptrajectory, traj % ptab_prot )
      traj % io_compl = traj_file_i
      traj % header_str = ''
     
      ! tot_nlines is output
      call init_record ( traj )
      call read_header_record ( traj, trajf, opt_nb_record = tot_nlines )

      !---------------------------------------------------------------
      ! initialize 'record' (had to copy from 'mod_record.f90')
      ! TODO: maybe a more robust function can be created there
      if ( traj % type_record == trajectory_enum ) then
        allocate ( traj % ptrajectory, stat = ierr )
        if ( ierr /= 0 ) then
            write ( *, * ) 'Problem in allocating ptrajectory '
            stop 1
        end if

        !! create the fixe array of one_complexe
        call initialize_trajectory ( traj % ptrajectory, traj )

      !! else complexes
      else if ( traj % type_record == complexe_enum ) then
        allocate ( traj % pcomplexe, stat = ierr )
        if ( ierr /= 0 ) then
            write ( *, * ) 'Problem in allocating pcomplexe'
            stop
        end if
        !! creates the list, no data yet
        if ( BTEST( traj % bit_integer, 0 ) ) then
            one_run = .false.
        end if
        if ( BTEST( traj % bit_integer, 3 ) ) then
            one_conf = .false.
        end if
        call new_listcomplexe ( traj % pcomplexe, 0.0_8, one_run, one_conf )
        !! Check for indice of ntraj and nconf in integer of onecomplexe
        if ( traj % indice_ntraj == 0 ) traj % pcomplexe % one_run = .false.
        if ( traj % indice_nconf == 0 ) traj % pcomplexe % one_conf = .false.
      else ! problem
        write ( *, * ) 'ERROR in intialize record, type of calculation not known'
        stop 1
      end if
        
    end subroutine prepare_read_traj
    
    !> Creating clusters of complexes, based on RMSD
    ! @param this : type ( ListComplexe ), output
    ! @param basecl : type ( record ), the base class for 'this', used for getting several constants
    ! @param prot_type : which protein type is chosen for analysis
    ! @param position : the x-y-z coordinates of COM of the protein that has to be added to the clusters
    ! @param orient_x : X axis of the inner coordinate system of the new protein
    ! @param orient_y : Y axis of the inner coordinate system of the new protein
    ! @param array_energy : the energy values for the new protein
    ! @param array_integer : which info is present (trajectory number, protein number, etc.)
    ! @param list_atom_surf_fixed : pointer to the original coordinates
    subroutine cluster_listcomplexe( this, basecl, prot_type, position, orient_x, orient_y, &
                     array_energy, array_integer, list_atom_surf_fixed )
       use mod_onecomplexe
       use mod_record
       IMPLICIT NONE
       
        type ( ListComplexe ) :: this
        type ( record ) :: basecl
        real ( kind=8 ), dimension ( 3 ), intent ( in ) :: position, orient_x, orient_y
        real ( kind=8 ), dimension ( 3 ) :: other_real
        real ( kind=8 ), dimension ( : ), intent ( in ) :: array_energy
        integer, intent( in ) :: prot_type
        integer, dimension ( 4 ), intent ( in ) :: array_integer
        integer, dimension ( 3 ) :: tmp_box
        real ( kind = 8 ), dimension ( :,: ), pointer :: list_atom_surf_fixed

        type ( one_complexe ), pointer :: pnew_onecomplexe
        type ( NodeComplexe ), pointer :: p_node_minrmsd, p_node_tmp
        real ( kind=8 ) :: rmsd2_min_input, min_rmsd2, rm2
        integer :: ierr

        nullify ( p_node_minrmsd, p_node_tmp, pnew_onecomplexe )
        
        min_rmsd2 = 1000.0
        other_real (:) = 0.0
        tmp_box (:) = 0
        rmsd2_min_input =  this % rmsd_min * this % rmsd_min
        
        p_node_tmp => this % p_first % p_next
        
        ! find the node in the list that has the lowest RMSD to the current structure
 OVERNODES: do while ( p_node_tmp % state /= last )
            rm2 = 0.0
            call rmsd ( rm2, basecl % ptab_prot % array( basecl % ptab_prot % set_of_protein( prot_type,1 ) ), &
                   list_atom_surf_fixed, &
                   p_node_tmp % p_data % position, &
                   p_node_tmp % p_data % orient_x, &
                   p_node_tmp % p_data % orient_y )

             if ( rm2 .le. min_rmsd2 ) then
                min_rmsd2 = rm2
                ! associate this node to p min rmsd
                if ( rm2 .le. rmsd2_min_input ) then
                  !write(*,*) 'associate p_node_rmsmin rm2 min_rmsd2', rm2, min_rmsd2
                  p_node_minrmsd => p_node_tmp

                end if
             end if
                    
            p_node_tmp => p_node_tmp % p_next
        end do OVERNODES
        
        if ( associated ( p_node_minrmsd ) ) then
            call update_statistics ( p_node_minrmsd % p_data, pnew_onecomplexe, array_energy ( 1:1 ) )
            return
        end if
        
        p_node_tmp => this % p_last
        allocate ( pnew_onecomplexe, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'Cannot allocate one_complexe'
        end if
        ! call new_onecomplexe ( pnew_onecomplexe, bit_integer, bit_energy, nb_other_real, opt_sum )
        ! default opt_ibox false
        call new_onecomplexe ( pnew_onecomplexe , &
                               basecl % nb_int, basecl % nb_real, basecl % nb_other_real )

        ! can fill the one_complexe
        !call set_onecomplexe ( pnew_onecomplexe, &
        !                       position, tmp_box, orient_x, orient_y, &
                               ! other real will be known and updated only when the insertion is done
        !                       array_integer, array_energy, other_real, basecl % nb_other_real, &
        !                       basecl % bit_integer, basecl % bit_energy, basecl % opt_sum, basecl % nb_types )
        call set_onecomplexe_wrap ( basecl, position, tmp_box, orient_x, orient_y, &
                                    ! other real will be known and updated only when the insertion is done
                                    array_integer, array_energy, other_real, opt_p_onecomplexe = pnew_onecomplexe )
        
        pnew_onecomplexe % other_real_ocompl ( 1 ) = 1.0
        pnew_onecomplexe % other_real_ocompl ( 2 ) = array_energy ( 1 )
        ! 3 is energy dispersion == <E>2 - <E>, store only Sum E2
        ! done when created
        !pnew_onecomplexe % other_real_ocompl ( 3 ) = 0. !( array_energy ( 1 ) )**2
        pnew_onecomplexe % sum_square_energy = ( array_energy ( 1 ) )**2
        
        call insert_new_node ( p_node_tmp, pnew_onecomplexe )
        this % nb_complexe = this % nb_complexe + 1
    end subroutine cluster_listcomplexe
    
    !> Calculate real, all-atom RMSD between two solutes (here, for simplicity, only the solvent accessible atoms are used)
    ! @param rm2 : the resulting RMSD (in Angstrom)
    ! @param prot : protein class, needed to get the list of solvent accessible atoms
    ! @param coords1 : an array with fixed coordinates
    ! @param com2 : center of mass of the 'movable' solute
    ! @param orientX2 : X axis of the inner coordinate system of the 'movable' solute
    ! @param orientY2 : Y axis of the inner coordinate system of the 'movable' solute
    subroutine rmsd ( rm2, prot, coords1, com2, orientX2, orientY2 )
        use mod_protein
        IMPLICIT NONE

        type( protein ), intent( in ) :: prot
        real ( kind = 8 ), dimension ( 3 ), intent( in ) :: com2, orientX2, orientY2
        real ( kind = 8 ), intent (out) :: rm2
        real ( kind = 8 ), dimension ( :,: ), intent ( in ) :: coords1
        real ( kind = 8 ), dimension ( :,: ), pointer :: orig_coords
        real ( kind = 8 ), dimension ( 3 ) :: orientZ2, tr_orientX2, tr_orientY2, tr_orientZ2, coords2
        integer :: i, nb_atoms

        rm2 = 0d0
        nb_atoms = size( coords1( 1,: ) )
!            write(*,*) '    nb_atoms: ', nb_atoms, '; com: ', com2

        orig_coords => prot % p_sogrid % list_atom_surface
        call cross( orientZ2, orientX2, orientY2 )
        call norm( orientZ2 )
        ! transpose the rotation matrix:
        tr_orientX2 = (/ orientX2(1), orientY2(1), orientZ2(1) /)
        tr_orientY2 = (/ orientX2(2), orientY2(2), orientZ2(2) /)
        tr_orientZ2 = (/ orientX2(3), orientY2(3), orientZ2(3) /)

        coords2( : ) = 0d0

        ! compute distance btwn each pair of atoms
        do i = 1, nb_atoms
        ! transform the mobile coords: rotate and translate
            coords2( 1 ) = dot_product( orig_coords( :,i ), tr_orientX2 ) + com2( 1 )
            coords2( 2 ) = dot_product( orig_coords( :,i ), tr_orientY2 ) + com2( 2 )
            coords2( 3 ) = dot_product( orig_coords( :,i ), tr_orientZ2 ) + com2( 3 )

            rm2 = rm2 + dot_product( coords1( :,i ) - coords2, coords1( :,i ) - coords2 )
        end do
        rm2 = rm2 / dble( nb_atoms )

    end subroutine rmsd

END MODULE analyze_subroutines
