!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> @file cluster_sdamm.f90
!! make cluster from a trajectory

!> Program cluster_sdamm
!! \ingroup tools
!! Program for clustering the configurations of complexes
!! formed between either two proteins or a protein and a surface.
!! Used with SDAMM-type of trajectory.
program cluster_sdamm
    USE mod_record
    USE mod_sda_input
    USE mod_input_parameter
    USE analyze_subroutines
    USE genbox_subroutines
    USE mod_type_calc
    ! added michael pi is in common
    USE com_sda
    IMPLICIT NONE
    
    ! sda input file - this contains all the information
    character ( 128 ) :: input_file, output_file, traj_file
    character ( 64 ) :: tmp_arg
    character ( 64 ), dimension(:,:), allocatable :: list_pdb_f
    
    integer :: nargs, tot_nlines, nprot_close, nclusters
    integer :: trajf_p, outputf_p, ierr
    integer :: nb_prot_types, nb_surf_atoms, nb_mol
    integer :: surf_mol_nb, surf_type_nb, chosen_prot_type
    ! counters
    integer :: i, j
    ! vector with step, mol_no etc., from 'onecomplexe'
    integer, dimension( 4 ) :: int_array
    
    real ( kind=8 ), dimension( 3 ) :: cur_pos, center, tmp_pos, cur_rotX, cur_rotY, cur_rotZ, tmp_rot
    real ( kind=8 ), dimension( 3,3 ) :: rot_mat
    real ( kind=8 ) :: max_dist, max_rmsd
    real ( kind=8 ) :: rot_angle, sin_rot_angle, vec_len
    ! vector with energies from 'onecomplexe'
    real ( kind = 8 ), dimension( : ), allocatable :: real_array, other_real_array, tmp_real_array
    real ( kind = 8 ), dimension ( :,: ), pointer :: list_atom_surf, orig_coords
    real ( kind = 8 ), dimension ( 3 ) :: tr_orientX, tr_orientY, tr_orientZ
    
    logical :: iflex, one_run, one_conf, new_node, isurf
    
    type ( record ) :: record_traj
    type ( record ) :: record_new
    type ( array_protein_type ), target :: tab_prot
    type ( protein ), pointer :: surf_prot
    type ( Trajectory ), pointer :: traj
    type ( param_list ) :: list_param
    type ( ListComplexe ), target :: close_proteins
    type ( NodeComplexe ), pointer :: p_first_ordered, p_last_ordered, p_curr_tmp, p_next_tmp
    
    one_run = .true.
    one_conf = .true.
    nprot_close = 0
    nclusters = 0
    rot_mat(:,:) = 0d0
    
    nargs = 4
    !if ( iargc() .lt. nargs )  then
    if ( command_argument_count() .lt. nargs ) then
        ! wrong number of arguments given - print info and exit
        write (*,*) 'cluster_sdamm - cluster configurations of molecules from SDAMM-type trajectory'
        write (*,*) ''
        write (*,*) 'USAGE:'
        write (*,*) '   ./cluster_sdamm   sda_input_file   prot_type   max_dist   max_rmsd  [traj]'
        write (*,*) '        prot_type  -  number of the type, in the order given in the input file (integer)'
        write (*,*) '        max_dist   -  for clustering of the orientations on the surface, maximum distance'
        write (*,*) '                      from the surface (float)'
        write (*,*) '        max_rmsd   -  extent of each cluster: how large can be the rmsd between an orientation'
        write (*,*) '                      and the cluster representative to include it in the same cluster'
        write (*,*) '        traj       -  trajectory file name; if no given, the trajectory from the sda_input_file'
        write (*,*) '                      will be used'
        call exit
    end if
    
    call getarg( 1, input_file )
    write(*,*) 'input argument - input file: ', input_file

    call read_input_sda( tab_prot, input_file, list_param, list_pdb_f, iflex, .true., .false., .true. )
    nb_prot_types = size( tab_prot % set_of_protein (:,1) )
    write(*,*) 'detected number of protein types: ',nb_prot_types

    call getarg( 2, tmp_arg )
    read (tmp_arg, '(i2)') chosen_prot_type
    write(*,*) 'input argument - chosen protein type: ', chosen_prot_type
    
    if ( chosen_prot_type .gt. nb_prot_types ) then
        write(*,*) 'ERROR The chosen protein type is larger than the maximal number of types!'
        STOP 1
    end if

    call getarg( 3, tmp_arg )
    read (tmp_arg, '(f8.3)') max_dist
    write(*,*) 'input argument - max distance: ', max_dist

    call getarg( 4, tmp_arg )
    read (tmp_arg, '(f8.3)') max_rmsd
    write(*,*) 'input argument - max rmsd: ', max_rmsd
    
    ! 1. read the trajectory
    !if( iargc() .eq. 5 ) then
    if ( command_argument_count() == 5 ) then
        call getarg( 5, traj_file )
        write(*,*) 'input argument - trajectory file: ', traj_file
    else
        traj_file = list_param % ftrajectories
    end if
    
    trajf_p = 111
    outputf_p = trajf_p + 1
    
    call prepare_read_traj( record_traj, trajf_p, traj_file, tot_nlines )

    if ( record_traj % type_record .ne. trajectory_enum ) then
        write(*,*) 'ERROR The file ', traj_file, ' does not contain trajectory, check the input and try again'
        stop 1
    end if
    
    call read_file_record( record_traj )
!    call print_bits( record_traj )

    call set_variable_field ( record_traj )

    traj => record_traj % ptrajectory
    
    call new_listcomplexe( close_proteins, max_rmsd, one_run, one_conf )
    write(*,*) 'new list complexe created; rmsd_min: ', close_proteins % rmsd_min
    write(*,*) 'dist_max (prot): ',tab_prot % array( tab_prot % set_of_protein( chosen_prot_type,1 ) ) % p_sogrid % distance_max
    
    write( output_file, '("clustered_complexes_type",i2.2)' ) chosen_prot_type
    call new_record( record_new, "complexe", record_traj % record_type_calc, record_traj % nb_max_complexe, &
         output_file, outputf_p, record_traj % opt_sum, record_traj % box_info, &
         record_traj % ntraj_rec, record_traj % freq_print, max_rmsd, one_run, one_conf, .true., record_traj % nb_types )
    call delete_listcomplexe( record_new % pcomplexe )
    deallocate( record_new % pcomplexe )
    nullify( record_new % pcomplexe )
    
    record_new % pcomplexe => close_proteins
    ! copy 'bits' from the old record:
    record_new % bit_integer = record_traj % bit_integer
    record_new % bit_energy = record_traj % bit_energy
    record_new % nb_int = record_traj % nb_int
    record_new % nb_real = record_traj % nb_real
    
!    call print_bits( record_new )
    call initialize_record ( record_new, tab_prot, .true. )

!    write(*,*) 'created new record, nb_real: ',record_new % nb_real,'; nb_other_real: ',record_new % nb_other_real
    
    allocate( real_array ( record_new % nb_real ), tmp_real_array ( record_new % nb_real ), stat=ierr )
    if ( ierr /= 0 ) then
      write (*,*) 'ERROR Could not allocate memory for real_array and tmp_real_array ', record_new % nb_real
      STOP 1
    endif

    allocate( other_real_array ( record_new % nb_other_real ), stat=ierr )
    if ( ierr /= 0 ) then
      write (*,*) 'ERROR Could not allocate memory for other_real_array ', record_new % nb_other_real
      STOP 1
    endif
    other_real_array (:) = 0d0
    
    orig_coords => tab_prot % array( tab_prot % set_of_protein( chosen_prot_type,1 ) ) % p_sogrid % list_atom_surface
    nb_surf_atoms = size( orig_coords (1,:) )
    write(*,*) ' nb_surf_atoms: ', nb_surf_atoms
    
    nullify ( list_atom_surf )
    allocate( list_atom_surf( 3, nb_surf_atoms ), stat=ierr )
    if ( ierr /= 0 ) then
      write (*,*) 'ERROR Could not allocate memory for list_atom_surf ', nb_surf_atoms
      STOP 1
    endif
    list_atom_surf(:,:) = 0d0

    ! 2. get only the proteins that are needed (e.g., only within a certain limit away from the surface)
    ierr = 0
    call reset_position( record_traj, ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Record contains no data'
        stop 1
    end if

    tmp_pos(:) = 0d0
    
    nullify ( surf_prot )
    isurf = .false.
    nb_mol = size( tab_prot % array )
    surf_mol_nb = -1
    surf_type_nb = -1
    do i = 1, nb_mol
        if ( tab_prot % array (i) % surface ) then
            isurf = .true.
            surf_mol_nb = i
            surf_prot => tab_prot % array (i)
            surf_type_nb = get_type( tab_prot, i )
            
            write(*,*) 'SURFACE FOUND! surface is mol.nb. ',surf_mol_nb,'; of type: ',surf_type_nb
            flush(6)
            EXIT
        end if
    end do
    
    if ( isurf ) then
        ! get the translation vector as the center of the surface, but set z-coord to 0.0
        center = surf_prot % center_protein
        center(3) = 0d0
        tmp_pos(:) = 0d0
        write(*,*) 'center: ', center
        
        do i = 1, tot_nlines
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR Could not update pointer! current pos: ', traj % current_pos, ', size: ', traj % size_array
                STOP 1
            endif
            
            !call return_compatible_array ( record_traj % p_onecompl, int_array, real_array, other_real_array, &
            !                           record_traj % bit_integer, record_traj % bit_energy, record_traj % opt_sum, nb_prot_types )
            call return_compatible_array_wrap ( record_traj, int_array, real_array, other_real_array ) !record_traj % p_onecompl

           ! if the current molecule is not the surface:
            if ( ( int_array( 3 ).ne.surf_mol_nb ) .and. ( get_type( tab_prot, int_array( 3 ) ).eq.chosen_prot_type ) ) then
              ! other_real_array: [ occurrency, avg_energy, stdev_energy ]
              ! have to change the 'total energy' here:
              if ( .not. record_traj % opt_sum ) then
                  tmp_real_array( : ) = real_array( : )
                  ! take only the energy of interaction with the surface
                  real_array( : ) = 0d0
                  do j = surf_type_nb + 1, size( real_array ), nb_prot_types
                      real_array( j ) = tmp_real_array( j )
                      real_array( 1 ) = real_array( 1 ) + tmp_real_array( j )
                  end do
!                  write(*,*) ' real_array: ', real_array(:)
              end if
              
              cur_pos = record_traj % p_onecompl % position
!              write(*,*) '    cur_pos: ',cur_pos
            ! look only at the z-coord
              if ( cur_pos( 3 ).lt.max_dist ) then
!                  write(*,*) ' protein ', int_array( 3 ), ' is close to the surface!'
            ! 3. align the proteins in x-y plane (do not change the z-coord)
                nprot_close = nprot_close + 1
                tmp_pos (3) = cur_pos (3)
!                write(*,*) 'new position: ', cur_pos
                cur_rotX = record_traj % p_onecompl % orient_x
                cur_rotY = record_traj % p_onecompl % orient_y
                vec_len = sqrt( cur_rotX(1)**2 + cur_rotX(2)**2 )
                rot_angle = acos( cur_rotX(2) / vec_len )
                sin_rot_angle = cur_rotX(1) / vec_len
                
!                write(*,*) '    cur_rotX: ', cur_rotX,'; cur_rotY: ',cur_rotY
!                write(*,*) '   rot_angle: ',rot_angle,'; sin_rot_angle: ',sin_rot_angle
                
                if ( sin_rot_angle .lt. 0d0 ) then
                    rot_angle = -rot_angle
                end if
                ! rotation vector
                tmp_rot(:) = 0d0
                tmp_rot(1) = rot_angle / (2 * PI)
                
                call rotate( tmp_rot, cur_rotX, .true. )
                call rotate( tmp_rot, cur_rotY, .true. )
!                write(*,*) '    new_rotX: ', cur_rotX,'; new_rotY: ',cur_rotY
                ! normalize the rotated vectors
                call norm( cur_rotX )
                call norm( cur_rotY )
!                write(*,*) '    new_rotX: ', cur_rotX,'; new_rotY: ',cur_rotY
!                write(*,*) ''
                
!              write(*,*) '   !!! z is less than ', max_dist
!              write(*,*) '  mol no: ', int_array( 3 ), '; position: ',cur_pos
!              write(*,*) '    int_array: ',int_array
!              write(*,*) '    real_array: ',real_array
                
                ! rotate the coords of atoms of the fixed prot:
                call cross( cur_rotZ, cur_rotX, cur_rotY )
                call norm( cur_rotZ )
                tr_orientX = (/ real ( kind=8 ) :: cur_rotX(1), cur_rotY(1), cur_rotZ(1) /)
                tr_orientY = (/ real ( kind=8 ) :: cur_rotX(2), cur_rotY(2), cur_rotZ(2) /)
                tr_orientZ = (/ real ( kind=8 ) :: cur_rotX(3), cur_rotY(3), cur_rotZ(3) /)

!                write(*,*) '    new_rotX: ', cur_rotX
!                write(*,*) '    new_rotY: ',cur_rotY
                list_atom_surf(:,:) = 0d0
                do j = 1, nb_surf_atoms
            ! transform the fixed coords: rotate and translate
                    list_atom_surf( 1,j ) = dot_product( orig_coords( :,j ), tr_orientX )
                    list_atom_surf( 2,j ) = dot_product( orig_coords( :,j ), tr_orientY )
                    list_atom_surf( 3,j ) = dot_product( orig_coords( :,j ), tr_orientZ )

                    list_atom_surf( :,j ) = list_atom_surf( :,j ) + tmp_pos
                end do
              
!                write(*,*) 'add the complex to the list'
            ! 4. cluster by true RMSD
                call cluster_listcomplexe( close_proteins, record_new, chosen_prot_type, tmp_pos, cur_rotX, &
                     cur_rotY, real_array, int_array, list_atom_surf )
              end if
            end if
!            write(*,*) 'get next...'
            call getNext ( record_traj, ierr )
        end do
    else
        write(*,*) 'ERROR clustering of SDAMM trajectories without surface not implemented yet'
        stop 1
    end if
    nclusters = close_proteins % nb_complexe
    
    write(*,*) 'found: ', nprot_close, ' proteins closer than ',max_dist,' from the surface; no. of clusters: ',nclusters
    flush(6)
    if ( nprot_close.gt.1 .and. nclusters.gt.1 ) then
        ! order clusters by occurency
        p_curr_tmp => close_proteins % p_first % p_next % p_next
        p_last_ordered => close_proteins % p_first % p_next % p_next

        write(*,*) ' re-ordering complexes...'
        flush(6)
        do while ( p_curr_tmp % state .ne. last )
            new_node = .false.
            p_first_ordered => p_curr_tmp
    !        p_curr_tmp => p_last_ordered % p_next

            do while ( p_curr_tmp % p_previous % state .ne. first .and. &
                ( p_first_ordered % p_data % other_real_ocompl(1) .ge. p_curr_tmp % p_previous % p_data % other_real_ocompl(1) ) )
                p_curr_tmp => p_curr_tmp % p_previous
                new_node = .true.
            end do

            p_next_tmp => p_first_ordered % p_next
            if ( new_node ) then
                p_first_ordered % p_next % p_previous => p_first_ordered % p_previous
                p_first_ordered % p_previous % p_next => p_first_ordered % p_next
                ! insert "p_first_ordered" BEFORE "p_curr_tmp"
                call insert_node( p_curr_tmp, p_first_ordered )
            else
                p_last_ordered => p_first_ordered
            end if

            p_curr_tmp => p_next_tmp
        end do
    end if

    write (*,*) '   Clustering finished successfully, writing output file: ',output_file
    flush(6)
    record_new % record_type_calc % enum_type_calc = sda_2proteins
    record_new % record_type_calc % type_calc_str = get_string ( record_new % record_type_calc )
    record_new % header_str = ''
    record_new % description_column_str = ''
    record_new % format_header_ascii = ''
    
    do i = 1, nb_prot_types
        record_new % ptab_prot % array (i) % center_protein(:) = center
    end do
    call write_header_record ( record_new )
    call write_listcomplexe( close_proteins, record_new, .true. )
    
    close( outputf_p )
    close( trajf_p )
    
    call delete_listcomplexe( close_proteins )
    nullify( record_new % pcomplexe )
    call delete_record( record_new, .true. )
    call delete_record( record_traj, .true. )
    call delete_array_protein ( tab_prot )
    deallocate( real_array, other_real_array, list_atom_surf )
end program cluster_sdamm
