!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> @file convert_grid.f90
!!
!! Conversion tool for UHBD grids

!> Extend formerprogram ascci2bin and bin2ascii, apbs2uhbd
!!
!! \ingroup tools
!! Convert grid UHBD type between ascci and binary format can scale value (ex from kT to kcal/mol )\n
!! can conver to dx format, for pymol\n
!! note : do not make a real copy of the grid ( save memory ), copy is false copy by default, need to add an option
program convert_grid
    USE mod_grid
    IMPLICIT NONE

    interface
        subroutine create_filename ( pdb_filename, suffix, out_filename, opt_shift )
            character*128 :: pdb_filename, out_filename
            character, intent( in ) :: suffix*(*)
            integer, optional :: opt_shift
        end subroutine create_filename
    end interface

    type ( type_grid ) :: ingrid, outgrid, excl_grid
    character*128:: infname, outfname, exclude_grid
    character *128 :: argum
    integer :: nb_argument, count_arg
    logical :: bool_fname, bool_convert, bool_scale, bool_one_input
    logical :: bool_dx, bool_header_only, bool_excl, bool_sphere
    real ( kind=8 ) :: scale_grid

    bool_fname = .true.
    bool_convert = .false.
    bool_scale = .false.
    bool_excl = .false.
    bool_sphere = .false.
    bool_dx = .false.
    bool_one_input = .false.
    bool_header_only = .false.
    scale_grid = 1.d0
    !cutoff = 0.d0

    infname =''
    outfname =''
    exclude_grid = ''

    nullify ( ingrid % p_grid, outgrid % p_grid )
    nullify ( excl_grid % p_grid )

    !nb_argument = iargc ()
    nb_argument = command_argument_count()

    if ( nb_argument == 0 ) then
        write(*,*)
        write(*,*) 'HELP'
        write(*,*) 'convert_grid grid_input [ grid_output ], [-convert] / [ -scale factor ] / [ -dx ] / [ -header_only ] '
        write(*,*)
        write(*,*) ' Print informations or modify grids'
        write(*,*) ' apply to all uhbd grids ep, ed..., resid3d, to extend to exlusion_grid '
        write(*,*)
        write(*,*) ' Most of the options cannot be freely combine, check in running steps by steps '
        write(*,*)
        write(*,*) ' output file is an option, if not given will add a suffix automaticaly ', &
                   '( eg: *.ascii.grd )'
        write(*,*) ' -convert will convert ascii to binary or binary to ascii '
        write(*,*) ' -scale is an optional multiplicative factor, apply to all grid points'
        write(*,*) '        for an equivalent of apbs2uhbd do:'
        write(*,*) '        convert_grid apbs_grid uhbd_grid -scale 0.6 -convert'
        write(*,*) ' -exclude_molsurf filename : will read grid_input and filename ( a grid file representing the surface'
        write(*,*) '                          of the solute ( binary grid generated by apbs write smol )'
        write(*,*) ' -make_spherical : make a spherical grid of the maximum size, setting to 0. outside'
        write(*,*) ' - dx is to produce dx output file, seems to be always ascii ??, read not yet implemented'
        write(*,*)
        write(*,*) ' -header_only : just print info about the grid, do not write output file'
        write(*,*)
        stop
    end if

    count_arg = 1

    do while ( count_arg <= nb_argument )

        call getarg( count_arg, argum )

        !write(*,*) 'argum:',trim(argum),'!!'

        if ( ( argum(1:1) == '-' ) .and. (  bool_fname .eqv. .true. ) ) then
            bool_fname = .false.
            !write(*,*) 'make false bool_fname'
        else if ( count_arg == 1 ) then
            infname = trim ( argum )
        else if ( count_arg == 2 ) then
            outfname = trim ( argum )
        endif

        if ( argum == '-convert' ) then
            bool_convert = .true.

        else if ( argum == '-scale' ) then
            bool_scale = .true.
            call getarg( count_arg + 1 , argum )
            read ( argum, fmt='(f8.3)' ) scale_grid
            count_arg = count_arg + 1

        else if ( argum == '-exclude_molsurf' ) then
            write(*,*) 'bool_excl set to true'
            bool_excl = .true.
            call getarg( count_arg + 1 , argum )
            read ( argum, fmt='(a)' ) exclude_grid

        else if ( argum == '-make_spherical' ) then
            write(*,*) 'make spherical'
            bool_sphere = .true.

        else if ( argum == '-dx' ) then
            bool_dx = .true.

        else if ( argum == '-header_only' ) then
            bool_header_only = .true.

        end if

        count_arg = count_arg + 1
    end do

    !! check if missing input filename
    if ( len ( trim( infname )  ) == 0 ) then
        write (*,*) 'Need at least one input filename !!'
        stop 1
    end if
    !! check if -argument provided
    if ( bool_fname ) then
        write (*,*)
        write (*,*) 'No argument ?? works but will just make a copy of the grid !!!!'
        write (*,*)
    end if

    !always print
    write (*,*)
    write (*,*) 'input filename  ', trim ( infname )

    if ( len ( trim ( outfname ) ) == 0 ) then
        !write(*,*) 'will create automatic filename one_input TRUE'
        write(*,*) 'will generate an output filename'
        bool_one_input = .true.
        outfname = infname
    else
        write (*,*) 'output filename ', trim ( outfname )
    end if
    write (*,*)

    !! read input grid_type is 0 by default
    call read_grid ( ingrid, infname )
    !write ( *,*) 'ingrid type ', ingrid % grid_type

    call info_grid( ingrid )
    !!just print data member and exit
    if ( bool_header_only ) then
        call delete_grid ( ingrid )
        stop 1
    end if

    !! create new grid, DO NOT copy the grid, by default points to the original grid
    call copy_grid ( ingrid, outgrid )

    !! do what needed
    if ( bool_convert ) then
        if ( ingrid % iform == 0 ) then
            outgrid % iform = 1
            write (*,*) 'will convert from binary to ascii'
        else if ( ingrid % iform == 1 ) then
            outgrid % iform = 0
            write (*,*) 'will convert from ascii to binary'
        end if

        !write(*,*) 'iform new grid', outgrid % iform
    end if

    if ( bool_scale ) then
        write (*,*) 'will scale the grid of ingrid because outgrid points to it'
        write(*,*) 'scale_grid ', scale_grid

        !! scaling very simple do not need a function
        !outgrid % pgrid = outgrid % pgrid * scale_grid
        call grid_scale ( outgrid, scale_grid )
    end if

    if ( bool_excl ) then

        write(*,*)
        !write(*,*) 'Will exclude grid with the surface of protein'
        call read_grid ( excl_grid, exclude_grid )
        write(*,*) 'Info from the exclusion grid'
        call info_grid ( excl_grid )

        !! can test here if grids (headers) are similar before continuing
        !! outgrid points to the original grid, so can be overwritten in this case
        call exclude( outgrid, excl_grid )
        !write(*,*) 'call after bool_excl'

        !! clean memory
        call delete_grid ( excl_grid )
    endif

    if ( bool_sphere ) then

        write(*,*)
        write(*,*) 'Will make a spherical grid'
        call make_sphere( ingrid )

    end if

    !! last test if dx
    if ( bool_dx ) then
        !! always ascii file ??
        outgrid % iform = 1
        write(*,*) 'write dx format, always ascii '
        !write(*,*) ' outfname ', trim ( outfname )

        !! if bool_one_input will ask for automatic generation of filename
        call write_grid_dx ( outgrid, outfname, bool_one_input )

    !! default write uhbd type
    else
        !! last argument, opt_pgrid4 if available
        !write(*,*) 'outfname ', trim(outfname)
       call write_grid ( outgrid, outfname, bool_one_input )
    end if

    if ( bool_one_input ) then
        write (*,*) 'generated filename : ', outfname
    end if

    !! delete first allocated grid
    call delete_grid ( ingrid )

    !! error because not real copy of the grid by default, pgrid points to the original grid
    !! may add real copy but not needed until now..., cannot check if associated in this case
    !! ok no error missing memory
    !call delete_grid ( outgrid )

end program

