!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file dcc.f90
!! Tool to estimate translational and rotational diffusion coefficient


PROGRAM dcc

    use mod_setofgrid
    ! define load_pdb
    use mod_sda_input

    IMPLICIT NONE

    interface
        subroutine compute_for_bin ( p_sogrid, bin, rad, dtrans, drot )
            use mod_setofgrid
            IMPLICIT NONE
            type ( sogrid ), pointer :: p_sogrid
            real(kind=8) :: bin
            real(kind=8), intent(out) :: rad, dtrans, drot
        end subroutine compute_for_bin
    end interface


    !! will contain the pdb in list_atom_surface
    !! need to compile with trajectory2dcd who defines load_pdb
    type ( sogrid ), pointer :: p_sogrid

    real (kind=8) ::  bin1, bin2, bin_width, dbin, dbin_copy
    character *128 :: argum, infname
    integer :: nb_argument, count_arg, ierr
    logical :: bool_binw, bool_bin1, bool_bin2, bplot, bscan
    real ( kind=8 ) :: dmax, rad, dtrans, drot

    nullify ( p_sogrid )
    infname =''
    bool_binw = .false.
    bool_bin1 = .false.
    bool_bin2 = .false.
    bplot = .false.
    bscan = .false.
    !! default value
    dbin = 0.d0
    bin1 = 0.d0
    bin2 = 0.d0

    nb_argument = command_argument_count()
    if ( nb_argument == 0 ) then
        write(*,*)
        write(*,*) 'HELP'
        write(*,*) 'dcc pdb_filename -bin_width size -bplot'
        write(*,*)
        write(*,*) 'Estimate translational and rotational diffusion coefficient'
        write(*,*)
        write(*,*) 'pdb_filename : name of the pdb file'
        write(*,*) '-bin_width : bin size (default 6A) or delta bin between bin1 and bin2'
        write(*,*) '-bin1 : first bin_width'
        write(*,*) '-bin2 : last bin_width'
        write(*,*) '-plot : store results in the plot_coeff.dat file, default false'
        write(*,*)
        write(*,*) 'ex: to compute at one bin_width'
        write(*,*) '    dcc p1_noh.pdb [-bin_width 6.0]'
        write(*,*) 'ex: to scan between 0 A and 20 A by step of 1.0 A'
        write(*,*) '     dcc p1_noh.pdb -bin_width 1.0 -bin1 1. -bin2 20 -plot'
        write(*,*) '     then plot a choose a converged value'
        stop
    end if

    count_arg = 1

    do while ( count_arg <= nb_argument )

        call getarg( count_arg, argum )

        if ( count_arg == 1 ) then
            infname = trim ( argum )
            !write(*,*) 'read filename ', trim(argum)
            !write(*,*) 'infname ', infname

        else if ( argum(1:10) == "-bin_width" ) then
            bool_binw = .true.
            call getarg( count_arg + 1 , argum )
            read ( argum, * ) bin_width

        else if ( argum(1:5) == "-bin1" ) then
            bool_bin1 = .true.
            call getarg( count_arg + 1 , argum )
            read ( argum, * ) bin1

        else if ( argum(1:5) == "-bin2" ) then
            bool_bin2 = .true.
            call getarg( count_arg + 1 , argum )
            read ( argum, * ) bin2

        else if ( argum(1:5) == "-plot" ) then
            bplot = .true.
        end if

        count_arg = count_arg + 1

    end do

    write(*,*) 'entry:'
    write(*,*) 'infname ', trim(infname)

    if ( ( bool_bin1 .eqv. .true. ) .and. ( bool_bin2 .eqv. .true. ) ) then
        bscan = .true.
        write(*,*) 'Scan value between ', bool_bin1, bool_bin2

        if ( bool_binw .eqv. .false. ) then
            ! default for scan
            bin_width = 1.0d0
        end if
        write(*,*) 'With delta ', bin_width

    else
        if ( bool_binw .eqv. .false. ) then
            bin_width = 6.0d0
            write(*,*) 'bwin default value ', bin_width
        end if
        write(*,*) 'Compute for a single bin_width: ', bin_width
    end if
    write(*,*) 'plot ', bplot

    !subroutine load_pdb ( psgrid, filename, opt_center_prot, opt_van, opt_return_center )
    call load_pdb ( p_sogrid, infname, .true. )
    write(*,*) 'nb of atoms: ', size ( p_sogrid % list_atom_surface ) / 3
    ! initialized by load_pdb
    write(*, *) 'distance_max: ', p_sogrid % distance_max
    dmax = p_sogrid % distance_max
    ! step is bin width
    !IF (step.GT.dmax) step = dmax

    !! open for plot
    if ( bplot ) then
        open ( 66, file='plot_coeff.dat', form='formatted', action='write', iostat = ierr )
    endif

    ! compute for one value
    if ( bscan .eqv. .false. ) then
        write(*,*) 'bscan false'
        bin1 = bin_width
        bin2 = bin_width
        dbin = bin_width
    !! scan from bin1
    else
        dbin = bin1
    end if

    !dbin = bin1 + bin_width
    do while ( dbin .le. bin2 )

        !dbin = bin1 + (ibin) * bin_width
        !write(*,*) 'value of dbin ', dbin
        !if ( dbin .gt. dmax ) dbin = dmax
        dbin_copy = dbin
        call compute_for_bin ( p_sogrid, dbin_copy, rad, dtrans, drot )

        ! print in output
        write (6,502) dmax,dbin,rad,dtrans,drot
 502 FORMAT ('dmax=',f8.3,' bin_witdh=',f8.3,'   h-radius=',f15.10, &
             '   dtrans=',f15.10,'   drot=',f15.10)

        if ( bplot ) then
            !! write header
            if ( dbin == bin1 ) then
                write (66,*) '#bin_width  h-radius   dtrans       drot'
            end if
            write (66, fmt='(f8.2, 3(3x,es11.4) )' ) dbin,rad,dtrans,drot
        end if


         ! update dbin
         dbin = dbin + bin_width
    end do

    call delete_sogrid ( p_sogrid  )
    deallocate ( p_sogrid )

END PROGRAM dcc

!> Retrun value of diffusion and h-radius for a given bin
subroutine compute_for_bin ( p_sogrid, bin, rad, dtrans, drot )
    use mod_setofgrid
    IMPLICIT NONE

    interface
        function distance_to_center ( position ) result ( distance )
            IMPLICIT NONE
            real(kind=8), dimension(:) :: position
            real(kind=8) :: distance
        end function distance_to_center
    end interface

    type ( sogrid ), pointer :: p_sogrid
    real(kind=8) :: bin
    real(kind=8), intent(out) :: rad, dtrans, drot

    real (kind=8) :: dmax
    real ( kind= 8 ), dimension(:), allocatable :: area, limit, ratio
    integer, dimension(:), allocatable :: counter

    integer :: i, ii, k, jcon, natom, ierr
    real(kind=8), parameter :: volcons = 4.d0/3.d0 * (3.14159265d0)
    real(kind=8) :: rtest, dst


    dmax = p_sogrid % distance_max
    ! check for not be larger
    if ( bin .gt. dmax ) bin = dmax

    jcon = nint( dmax/bin + .49999d0 )

    allocate ( area( jcon ), limit ( jcon ), ratio ( jcon ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot allocate area '
        stop 1
    end if
    area = 0.d0
    limit = 0.d0
    ratio = 0.d0

    allocate ( counter( jcon ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot allocate counter '
        stop 1
    end if
    counter = 0

    do i = 1, jcon
        if ( i .gt. 1) then
              area(i) = volcons* (bin*i)**3 - volcons* (bin* (i-1))**3
              limit(i) = bin*i
        ! if 1
        else
              area(i) = volcons* (bin*i)**3
              limit(i) = bin*i
        end if
    end do

    ! count atom in specifc range
    natom = size ( p_sogrid % list_atom_surface ) / 3
    !write(*,*) 'natom ', natom
    do i = 1, natom
          k = 1
           do while ( (distance_to_center ( p_sogrid % list_atom_surface(:,i) )) .gt. (limit( k )) )
              k = k + 1
          end do
          counter(k) = counter(k) + 1
    end do
    ! normalize
    do i = 1,jcon
       if ( counter(i) .NE. 0 ) then
           ratio(i) = counter(i) / area(i)
       else
           ratio(i) = 0.d0
       end if
       ratio(i) = ratio(i)*1000.d0
    end do

!  estimate a best value for "rad", the effective spherical radius:
     rtest = ratio(2)/3.0

     ! Test that central density is not zero 
     ! -- dcc works for globular proteins
     !    may not be well defined in all cases
     if (rtest.EQ.0.d0) then
       write (*,*) "ERROR: No atomic density found at centre of solute"
       write (*,*) "       dcc algorithm is not valid for this solute"
       STOP 1
     end if

     ii = 2
loop1: do i = 2, jcon
         if (ratio(i) .LT. rtest) then
              ii = i
              exit loop1
          END IF
    end do loop1

    !write(*,*) 'ii after stange loop ', i
    !write(*,*) 'rtest ', rtest
    dst = bin * (rtest-ratio(ii-1)) / (ratio(ii)-ratio(ii-1))
    rad = (ii-1)*bin - (bin/2.0) + dst
!c ??
    !irad = 10.*rad
    !rad = irad/10.0 !! ??

!  diffusion properties in bulk water at 293.15 k:
     dtrans = 0.218/rad
     drot = dtrans*6./8./rad**2

    deallocate ( counter )
    deallocate ( area, limit, ratio )

end subroutine compute_for_bin

function distance_to_center ( position ) result ( distance )
    IMPLICIT NONE
    real(kind=8), dimension(:) :: position
    real(kind=8) :: distance

    !distance = 0.d0
    !write(*,*) 'position ', position
    distance = sqrt( dot_product( position, position ) )
    !write(*,*) 'distance ', distance

end function distance_to_center


