!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file gen_init_box.f90
!! Generate initial configuration for sdamm

!> Generate initial configuration for sdamm
!! \ingroup tools
program genbox
 
      ! to get the functions like 'load_pdb' or types 'protein'
      use mod_sda_input
      use mod_array_protein
      use mod_ReactionCriteria
      use com_sda
      use genbox_subroutines
      USE Par_Zig_mod
      ! added, because error type with intel
      use mod_type_calc

      IMPLICIT NONE

!----------VARIABLES-----------------------------------------------

      character ( 128 ) :: input_f
      character ( 50 ) :: out_str
      ! replace 503, 504, bad output
      character ( 1280 ) :: str_density
      
      !! number of arguments given
      integer :: nargs
      !! indices for input and output info
      integer :: err_status, input_i, stdout_i, box_out_i, pdb_in_i
      !! counters for loops
      integer :: prot_type, i, k, prev_sum, j
      !! counter of how many attempts have been made to place a solute
      integer :: cur_iter !NJB
      !! number of atoms on the surface of molecule
      integer :: nb_surf_atoms
      !! how many proteins have already assigned coordinates and how many are left to assign
      integer :: nb_assigned_prot, curr_assigned
      !! if there is a surface, this is the last molecule that is a protein
      integer :: last_prot, single_type, cur_mol, prev_mol
      integer, dimension (:), allocatable :: nb_mol_types, sseeds, size_order, mols_order
      integer, dimension (1) :: max_tmp
      integer :: s
      
      !! coeff is constant: required for re-calculating the mass in g
      !! tmp_vol is the total volume of the box
      real ( kind = 8 ) :: coeff, tmp_vol
      real ( kind = 8 ), dimension(3) :: null_array = (/ 0., 0., 0. /)
      !! number of molecules per AA^3
      real ( kind = 8 ), dimension(:), allocatable :: nb_mol_vol
      !! box dimensions and coordinates of left lower corner
      real ( kind = 8 ), dimension(3) :: box_dim, box_beg
      !! how much space to leave, if there is a surface, so that proteins don't overlap
      real ( kind = 8 ) :: box_margin, min_c2c_dist
      !! needed when calculating the maximal distance from molecule center
      real ( kind = 8 ) :: cur_max_rad, global_max_rad, cur_dist
      !! coordinates of atoms on the surface of molecule
      real ( kind = 8 ), dimension(:,:), pointer :: surf_atom_coords
      !! tmp and final coordinates for the geometrical center of the protein
      real ( kind = 8 ), dimension(3) :: center
      real ( kind = 8 ), dimension(:), pointer :: center_protein
      !! 'radii' of molecules:
      real ( kind = 8 ), dimension(:), allocatable :: rad_per_type
      !! random number from range 0-1
      real ( kind = 8 ), dimension(:), allocatable :: rnd
      real ( kind = 8 ) :: rnd1
      !! final coordinates of centers of molecules
      real ( kind = 8 ), dimension(:, :), allocatable :: prot_cntr_crds
      !! final coords of internal x and y orientation of molecule centers
      real ( kind = 8 ), dimension(:, :), allocatable :: prot_cntr_rot
      !! temporary z coord
      real ( kind = 8 ) :: tmp_z
      
      !! is the currently considered molecule a surface
      logical this_surf
      !! is the newly selected z coord of protein center correct
      logical ok_z_coord
      
      !! list with all the input details
      type ( genbox_list ) :: in_list
      type ( array_protein_type ) :: tab_prot
      type ( sogrid ), pointer :: psgrid
      ! needed to read in pdb
      type ( probe_type ) :: probe_param
      type ( react_criter ) :: rcriter
      type ( type_calculation ) :: mtype_calc

!-----------------------------------------------------------------
      ! set the input/output indexes
      input_i = 5 ! input file index
      stdout_i = 6 ! stdout index
      box_out_i = 11 ! box output file index
      pdb_in_i = 12 ! pdbfile index
      str_density =''
      
      mtype_calc % enum_type_calc = sdamm

      coeff = 1.660538782e03

      nargs = command_argument_count ()
      if ( nargs .ne. 1 ) then
          write( stdout_i, * ) 'genbox    -    generate initial configuration for SDAMM runs'
          write( stdout_i, * ) ''
          write( stdout_i, * ) 'USAGE:'
          write( stdout_i, * ) '        ./genbox  input_file'
          write( stdout_i, * ) '      (see online documentation for an example of the input file)'
          write( stdout_i, * ) ''
          STOP
      end if
      
      ! get the input file name
      call getarg( 1, input_f )
!      write (*,*) 'input file name: ', trim( input_f ), '; opening...'
      err_status = 0
      open ( unit = input_i, file = input_f, form = 'formatted', status = 'old', iostat = err_status )
      if ( err_status .ne. 0 ) then
        write( stdout_i, * ) 'input file is missing or empty'
        STOP
      endif
      
      ! read the input file
      call set_def_list_values ( in_list )
      call read_genbox_input( input_i, in_list )

      close ( input_i )
!      write (*,*) 'finished with reading input file; now checking ...'
      
      ! check input
      out_str = ''
!      if ( ( in_list%nb_types .ne. 1 ) .and. ( in_list%nb_types .ne. 2 ) ) then
!        write ( out_str, '(i8)' ) in_list%nb_types
!        call wrong_input ( 'number of molecule types', out_str )
!        STOP
!      end if
      
      if ( ( in_list%nb_mol .lt. 1 ) .or. ( in_list%surf .and. in_list%nb_mol .lt. 2 ) ) then
        write ( out_str, '(i8)' ) in_list%nb_mol
        call wrong_input ( 'number of molecules', out_str )
        STOP
      end if
      
      if ( in_list%surf .and. test_all ( in_list%surf_dim, null_array ) ) then
        write ( out_str, '(f8.3, f8.3, f8.3)' ) ( in_list%surf_dim ( 1:3 ) )
        call wrong_input ( 'surface dimensions', out_str )
        STOP
      end if

      if (in_list%max_iter.LT.1) then
        write (out_str, '(i8)') in_list%max_iter
        call wrong_input('Number of iterations to attempt solute placement',out_str)
        write (*,*) "  This must be a positive integer"
        stop
      end if

      if ((in_list%fit_factor.LT.0.0).OR.(in_list%fit_factor.GT.1.0)) then
        write (out_str, '(f8.5)') in_list%fit_factor
        call wrong_input('fit_factor',out_str)
        write (*,*) "  This must be a between 0.0 and 1.0"
        stop
      end if

      if (in_list%fit_factor.LT.1.0) then
        write (*,*) "WARNING: Using a fit_factor of less than 1.0 may result in solute overlaps"
        write (*,*) " Check early stages of simulation to ensure equilibration removes these overlaps"
      end if

      do i = 1, in_list%nb_types
        if ( in_list%pdbs ( i ) .eq. '' ) then
          write ( out_str, '(a, i3, a)' ) 'pdb file for protein (',i,'): '
          call wrong_input ( out_str, '' )
          STOP
        end if
      
        if ( in_list%surf .and. i .eq. in_list%nb_types ) then
            EXIT
        end if
        
        if ( in_list%molw( i ) .le. 0.0 ) then
          write ( out_str, '(f8.3)' ) in_list%molw( i )
          call wrong_input ( 'molecular weight', out_str )
          write (*,*) '(NB: if you have a surface, include it as the last molecule!)'
          STOP
        end if
        if ( in_list%dens( i ) .le. 0.0 ) then
          write ( out_str, '(f8.3)' ) in_list%dens( i )
          call wrong_input ( 'density', out_str )
          write (*,*) '(NB: if you have a surface, include it as the last molecule!)'
          STOP
        end if
      end do
      
      allocate( nb_mol_vol( in_list%nb_types ), stat = err_status )
      if ( err_status /= 0 ) then
         write (*,*) 'ERROR could not allocate memory for nb_mol_vol ', in_list%nb_types
         STOP 1
      endif
      
      
      allocate( nb_mol_types( in_list%nb_types ), stat = err_status )
      if ( err_status /= 0 ) then
          write(*,*) 'ERROR could not allocate array for nb_mol_types ',in_list%nb_types
          stop 1
      end if
      nb_mol_types(:) = 0
!      write (*,*) 'input seems ok'
      
      ! start the random number generator
      if ( in_list%seed.eq.0 ) then
          ! does not compile with xlf, aix machine
          ! in_list % seed = abs( mod( ( s*181 )*( ( 147286 - 83 )*359 ), 104729 ) )
          call generate_seed_from_clock ( in_list % seed )

          write(*,*) "generate seed from timestep: ", in_list%seed
      else
          write(*,*) "seed: ", in_list%seed
      endif
      call random_seed( size = s )
      allocate( sseeds( s ) )
      do i = 1, s
          sseeds( i ) = in_list%seed + 37 * i
      end do
      call random_seed( put = sseeds )

      !-------------------------------------------
      ! set some default sda values (needed to create grids, proteins...:
      probe_param % probep = 1.7
      probe_param % probew = 1.4
      if ( in_list % surf ) then
        probe_param % thresold = 0.0
      else
        probe_param % thresold = 5.0
      end if
      probe_param % hexclusion = 0.0
      probe_param % save_excl = 0
      probe_param % save_access = .true.
      
      rcriter % type_reaction = -1
      rcriter % ntotal = 0
      !NJB: Pointer needs to be nulified here, as allocate_react_criter not
      !called.
      nullify (rcriter%p_assoc_rate)
      
      ! create the protein objects and read in pdbs
      call new_array_protein ( tab_prot, in_list%nb_types, in_list%nb_types )
      
      do prot_type = 1, in_list%nb_types
        this_surf = .false.
        if ( in_list % surf .and. prot_type .eq. in_list%nb_types ) then
            this_surf = .true.
        end if
        
        !! allocate the set of grid
        allocate ( psgrid, stat = err_status )
        if ( err_status /= 0 ) then
          write ( *,* ) 'cannot allocate the set of grid'
          STOP
        end if

        call allocate_sogrid ( psgrid )
        psgrid % alloc_which_at_surf = .true.
        center = null_array
        call prepare_atom_protein ( psgrid, rcriter, in_list%pdbs ( prot_type ), probe_param, center, &
                                prot_type, .false., .false., this_surf, mtype_calc ) ! intel do not accept 1 for type_cacl
        call new_protein ( tab_prot % array ( prot_type ), .false., .false., .false., .false., prot_type )
        
        tab_prot % array ( prot_type ) % p_sogrid => psgrid
        tab_prot % array ( prot_type ) % center_protein = center
        tab_prot % set_of_protein ( prot_type,: ) = (/ prot_type, prot_type /)
        nullify ( psgrid )
      end do
      !! after loading pdbs, tab_prot % array (prot_type) % p_sogrid points to info:
      !!   p_sogrid % list_atom_surface - coords of atoms on the prot.surface
      !!   p_sogrid % nb_atom_surface - number of atoms on the prot.surface
      
      if ( in_list%surf ) then
          tab_prot % array ( in_list%nb_types ) % surface = .true.
      end if
      
!      do prot_type = 1, in_list%nb_types
!         call info_protein ( tab_prot % array ( prot_type ) )
!      end do
!      write (*,*) '==== proteins reading complete ===='
      
      allocate( rad_per_type( in_list%nb_types ), stat = err_status )
      if ( err_status /= 0 ) then
         write (*,*) 'ERROR: could not allocate memory for rad_per_type; size: ', in_list%nb_types
         STOP
      endif
      
      global_max_rad = 0.0
      ! check the maximal radius of protein:
      do prot_type = 1, in_list%nb_types
          cur_max_rad = 0.0
          if ( .not. tab_prot % array ( prot_type ) % surface ) then
              nb_surf_atoms = tab_prot % array (prot_type) % p_sogrid % nb_atom_surface
              surf_atom_coords => tab_prot % array (prot_type) % p_sogrid % list_atom_surface
              center_protein => tab_prot % array (prot_type) % center_protein
              do i = 1, nb_surf_atoms
                call vect_dist_real8 ( surf_atom_coords ( :, i ), center_protein, cur_dist )
                if ( cur_dist .gt. cur_max_rad ) then
                    cur_max_rad = cur_dist
                end if
              end do
          end if
          rad_per_type( prot_type ) = cur_max_rad
          if ( cur_max_rad .gt. global_max_rad ) then
              global_max_rad = cur_max_rad
          end if
      end do
      
      ! change the order of the particle distribution, based on their radii
      if ( in_list%surf ) then
!        size_order = sort_radius( rad_per_type( 1:( in_list%nb_types - 1 ) ), in_list%nb_types - 1 )
        call sort_radius( rad_per_type( 1:( in_list%nb_types - 1 ) ), in_list%nb_types - 1 , size_order)
      else
!        size_order = sort_radius( rad_per_type, in_list%nb_types )
        call sort_radius( rad_per_type, in_list%nb_types , size_order )
      end if
      
!      write (*,*) 'maximal distance of atom from mol center: ', global_max_rad
!      write (*,*) 'rad_per_type: ', rad_per_type
!      write (*,*) 'order: ',size_order

      box_beg = (/ 0., 0., 0. /)
      
      !! find the dimensions of the simulation box
      last_prot = in_list%nb_types
      if ( in_list%surf ) then
          last_prot = in_list%nb_types - 1
          nb_mol_types( in_list%nb_types ) = 1
      end if
!     Number of molecules in 1 A^3      
      nb_mol_vol (:) = 0
      do i = 1, last_prot
         nb_mol_vol( i ) = in_list%dens( i ) / ( in_list%molw( i ) * coeff )
!         write (*,*) 'nb of molecules of type', i ,' per AA^3: ', nb_mol_vol( i )
      end do
      
      if ( in_list%surf ) then
    !     New volume [A^3]
          box_margin = ( global_max_rad / 2.0 )
          box_dim ( 1 ) = in_list%surf_dim( 1 ) - global_max_rad
          box_dim ( 2 ) = in_list%surf_dim( 2 ) - global_max_rad
          box_dim ( 3 ) = real( in_list%nb_mol - 1 ) / ( sum( nb_mol_vol ) * box_dim ( 1 ) * box_dim ( 2 ) )
          
          tmp_vol = real( in_list%nb_mol - 1 ) / sum( nb_mol_vol )
          box_beg = (/ in_list%surf_beg( 1 ) + box_margin, in_list%surf_beg( 2 ) + box_margin, in_list%surf_beg( 3 ) - 1.0 /)
!          write (*,*) 'box x-y determined by surface'
      else 
          ! assume cubic volume
          box_margin = ( global_max_rad / 2.0 ) + 3.0
          tmp_vol = in_list%nb_mol / sum( nb_mol_vol )
          box_dim ( 3 ) = ( tmp_vol )**(1.0/3.0) - box_margin
          box_dim ( 1 ) = box_dim ( 3 )
          box_dim ( 2 ) = box_dim ( 3 )
          
!          write (*,*) 'box cubic'
      end if
!      write (*,*) ' ====== BOX ( x y z ): ========= ( ', box_dim ( 1 ) ,' ', box_dim ( 2 ),' ', box_dim ( 3 ), ' )'
!      write (*,*) ' ====== BOX BEG ( x y z ): ========= ( ', box_beg ( 1 ) ,' ', box_beg ( 2 ),' ', box_beg ( 3 ), ' )'
      do i = 1, last_prot
          nb_mol_types( i ) = nint( tmp_vol * nb_mol_vol( i ) )
!          write(*,*) 'nb of mol of type: ',i,'; b4 rounding: ',tmp_vol * nb_mol_vol( i ),'; after: ',nb_mol_types( i )
      end do
      
      if( any( nb_mol_types .eq. 0 ) ) then
          write(*,*) 'ERROR: one of the molecule types would not have any molecules! Perhaps the density is too low?'
          write(*,*) '   ( number of mols per each type: ', nb_mol_types, ')'
          STOP
      end if
      
      if( sum( nb_mol_types(:) ).gt.in_list%nb_mol ) then
          max_tmp = maxloc( nb_mol_types )
          nb_mol_types( max_tmp( 1 ) ) = nb_mol_types( max_tmp( 1 ) ) - ( sum( nb_mol_types(:) ) - in_list%nb_mol )
      else if ( sum( nb_mol_types(:) ).lt.in_list%nb_mol ) then
          max_tmp = maxloc( nb_mol_types )
          nb_mol_types( max_tmp( 1 ) ) = nb_mol_types( max_tmp( 1 ) ) + ( in_list%nb_mol - sum( nb_mol_types(:) ) )
      end if
      write (*,*) 'number of mols per each type: ', nb_mol_types,'; sum: ',sum(nb_mol_types)
      
!      min_c2c_dist = ( global_max_rad * ( 3.0/2.0 ) ) + 0.1
!       min_c2c_dist = global_max_rad + 0.1
      allocate ( prot_cntr_crds ( in_list%nb_mol, 3 ), stat = err_status )
      if ( err_status /= 0 ) then
        write ( *,* ) 'cannot allocate array with final coordinates of protein centers'
        STOP
      end if

      !! choosing random positions in the box:
      !! first: choose random coords for x and y coordinates;
      !! then, take a random number one by one, create a new z coord and check whether the
      !! new position is far enough from any other already placed (i.e., center-to-center
      !! distance is larger than 2*max_prot_radius)
      last_prot = in_list%nb_mol
      nb_assigned_prot = 0
      if ( in_list%surf ) then
          last_prot = in_list%nb_mol - 1
          prot_cntr_crds ( in_list%nb_mol, : ) = tab_prot % array ( in_list%nb_types ) % center_protein
          nb_assigned_prot = 1
      end if
      curr_assigned = nb_assigned_prot
      
        !! first - coords x and y
        do k = 1, 2
!            write (*,*) '==== now starting random selection for coord: ',k,' ======'
            nb_assigned_prot = curr_assigned

            allocate ( rnd ( in_list%nb_mol ), stat = err_status )
            if ( err_status /= 0 ) then
              write ( *,* ) 'cannot allocate random number array; err: ', err_status
              STOP
            end if
            ! get the random numbers for all molecules at once
            call random_number ( rnd )
!           write(*,*) rnd
            do i = 1, last_prot
                prot_cntr_crds ( i, k ) = box_dim ( k )*rnd ( i ) + box_beg ( k )
!                write ( *,* ) 'new coord:               ', prot_cntr_crds ( i, k )
                nb_assigned_prot = nb_assigned_prot + 1

                if ( nb_assigned_prot .ge. in_list%nb_mol ) then
                    exit
                end if
            end do
            deallocate ( rnd, stat = err_status )
            if ( err_status /= 0 ) then
              write ( *,* ) 'cannot deallocate random number array; err: ', err_status
              STOP
            end if
        end do ! loop for x and y coords
        
        ! if one of the types has only one molecule - place it in the center
        single_type = -1
        if ( .not.in_list%surf ) then
            k = 1
            do i = 1, in_list%nb_types
                if ( nb_mol_types( i ).eq.1 ) then
                    prot_cntr_crds ( k, 1 ) = box_dim ( 1 ) / 2.0 + box_beg( 1 )
                    prot_cntr_crds ( k, 2 ) = box_dim ( 2 ) / 2.0 + box_beg( 2 )
                    single_type = i
                    exit
                end if
                k = k + nb_mol_types( i )
            end do
        end if

        !! choose z coords
        nb_assigned_prot = 0
!        write (*,*) '==== now starting random selection for coord: 3 ======'
        if ( single_type.ne.-1 ) then
            prot_cntr_crds ( k, 3 ) = box_dim ( 3 ) / 2.0 + box_beg( 3 )
            nb_assigned_prot = nb_assigned_prot + 1
        end if
        
        ! set the new order of assigning coords - based on size of the mols
        allocate( mols_order( last_prot ), stat = err_status )
        if ( err_status /= 0 ) then
          write ( *,* ) 'cannot allocate mols_order array; err: ', err_status, '; size: ', last_prot
          STOP
        end if

        k = 0
        mols_order = 0
        do i = 1, size( size_order )
            ! current molecule type
            prot_type = size_order( i )
            ! how many molecules are that have type number less than 'prot_type'
            prev_sum = sum( nb_mol_types( 1:( prot_type - 1 ) ) )
            ! fill the array
            do j = 1, nb_mol_types( prot_type )
                mols_order( k + j ) = prev_sum + j
            end do
            k = k + nb_mol_types( prot_type )
        end do
!        write ( *,* ) 'mols_order: ', mols_order
        
        do i = 1, last_prot
            cur_mol = mols_order( i )
            prot_type = get_mol_type( nb_mol_types, in_list%nb_types, cur_mol )
            if ( prot_type.eq.-1 ) then
              write(*,*) 'ERROR: prot_type = ',prot_type,'; cur_mol: ',cur_mol
              STOP
            end if
            if ( prot_type.eq.single_type ) then
                CYCLE
            end if
            cur_max_rad = rad_per_type( prot_type )
            ok_z_coord = .false.

            cur_iter = 0
            do while ( .not.ok_z_coord )
                cur_iter = cur_iter + 1
!               write (*,*) 'mol no. ',i
                call random_number ( rnd1 )
                tmp_z = box_dim ( 3 )*rnd1 + box_beg ( 3 )
                !! leave a small space closest to the surface
                if ( in_list % surf ) then
                    tmp_z = ( box_dim ( 3 ) - global_max_rad ) * rnd1 + box_beg ( 3 ) + in_list%surf_dim( 3 ) + global_max_rad
                end if
                ! temporary new center coordinates
                center = (/ prot_cntr_crds ( cur_mol, 1 ), prot_cntr_crds ( cur_mol, 2 ), tmp_z /)

                ! check distance to other located proteins
                ok_z_coord = .true.
!                if ( i .eq. 1 ) then
!                    CYCLE
!                end if
                do k = 1, nb_assigned_prot
                    prev_mol = mols_order( k )
                    call vect_dist_real8 ( prot_cntr_crds ( prev_mol, : ), center, cur_dist )
                    prot_type = get_mol_type( nb_mol_types, in_list%nb_types, prev_mol )
                    if ( prot_type.eq.-1 ) then
                      write(*,*) 'ERROR: prot_type = ',prot_type,'; cur_mol: ',prev_mol
                      STOP
                    end if
                    min_c2c_dist = in_list%fit_factor*(cur_max_rad + rad_per_type( prot_type ))

                    if ( cur_dist .lt. min_c2c_dist ) then
                        ok_z_coord = .false.
                        exit
                    end if
                end do ! checking the new z coord

                ! Check that placement is not stuck in loop
                if (cur_iter.GT.in_list%max_iter) then
                    write (*,*) ""
                    write (*,*) "Solute placement failed!"
                    write (*,'(" Reached ",I10, " iterations attempting to place solute: ",I6)') in_list%max_iter , i
                    write (*,*) "  Try using a value of fit_factor of less than 1.0"
                    write (*,*) "   -Reducing fit_factor may result in overlapping solutes in intial configuration,"
                    write (*,*) "    so early stages of simulations should be checked for problems."
                    ! Exit status of 8 used to catch failure in websda
                    stop 8
                end if            

            end do ! while loop
!write(*,*) "CUR_ITER = ", cur_iter
            prot_cntr_crds ( cur_mol, 3 ) = tmp_z
            nb_assigned_prot = nb_assigned_prot + 1
        end do ! loop for z coords

!       write (*,*) '===== DONE: number of assigned prot.: ', nb_assigned_prot
      
      !! randomize the orientation of the proteins
      ! in the beginning - all proteins have the same orientation: xRot = (1,0,0), yRot = (0,1,0)
      allocate ( prot_cntr_rot ( in_list%nb_mol, 6 ), stat = err_status )
      if ( err_status /= 0 ) then
        write ( *,* ) 'cannot allocate array with rotations of prot.centers; err: ', err_status
        STOP
      end if
      
      prot_cntr_rot ( :,: ) = 0.0
      do i = 1, last_prot
        prot_cntr_rot ( i, 1 ) = 1.0
        prot_cntr_rot ( i, 5 ) = 1.0
      
        !! first - random select three angles
        allocate ( rnd ( 3 ), stat = err_status )
        if ( err_status /= 0 ) then
          write ( *,* ) 'cannot allocate random number array; err: ', err_status
          STOP
        end if
        call random_number ( rnd )
        
        ! rotate the x,y axis around these angles
!        write (*,*) 'before rotation:        ( ', prot_cntr_rot ( i, : ), ' )'
        call rotate ( rnd, prot_cntr_rot ( i, : ) )
!        write (*,*) 'ROTATED VECTORS:        ( ', prot_cntr_rot ( i, : ), ' )'
        deallocate ( rnd, stat = err_status )
        if ( err_status /= 0 ) then
          write ( *,* ) 'cannot deallocate random number array; err: ', err_status
          STOP
        end if
      end do
      
      if ( in_list%surf ) then
          prot_cntr_rot ( in_list%nb_mol, 1 ) = 1.0
          prot_cntr_rot ( in_list%nb_mol, 5 ) = 1.0
      end if
      
      !! WRITE OUT
      open ( box_out_i, file = 'box_info.txt', form = 'formatted', status = 'replace')
502   format ( a, i4 )
503   format ( a, f8.3 )
!503   format ( a, 2f9.3 )
504   format ( a, 3f9.3 )

! create format specific for density (a25 for 'concentration [g/L]: ')
write (0,*) size( in_list % dens )
      write (str_density, '("(a25,", i2, "f9.3)")' ) size( in_list % dens )
!      write(*,*) 'str_density ', str_density, '!!'

      write ( box_out_i, * ) '--- INPUT ---'
      write ( box_out_i, 502 ) 'total number of molecules: ',in_list%nb_mol
      write ( box_out_i, 502 ) 'number of molecule types: ',in_list%nb_types

      !write(*,*) 'in_list % dens ', in_list % dens,'!!'
      !write(*,*) 'size dens ', size( in_list % dens )

! michael, commented. use str_density instead
!      if ( ( ( in_list%nb_types .gt. 1 ) .and. ( .not.in_list%surf ) ) .or. &
!           ( ( in_list%nb_types .gt. 2 ) .and. in_list%surf ) ) then
!          write(*,*) '504 '
!          write ( box_out_i, 504 ) 'concentration [g/L]: ',in_list%dens(:)
!      else
!          write(*,*) '503'
!          !write ( box_out_i, 503 ) 'concentration [g/L]: ',in_list%dens
!          write ( box_out_i, fmt = str_density ) 'concentration [g/L]: ', in_list % dens
!          !write ( box_out_i, fmt = str_density ) in_list % dens
!      end if

      ! should work for all size
      write ( box_out_i, fmt = str_density ) 'concentration [g/L]: ', in_list % dens

      write ( box_out_i, * ) '--- OUTPUT ---'
      write ( box_out_i, 503 ) 'maximal distance of atom from mol center: ', global_max_rad
      if ( ( ( in_list%nb_types .gt. 1 ) .and. ( .not.in_list%surf ) ) .or. &
           ( ( in_list%nb_types .gt. 2 ) .and. in_list%surf ) ) then
          write ( box_out_i, * ) 'number of molecules per each type: ', nb_mol_types
      end if
      if ( in_list%surf ) then
          write ( box_out_i, * ) '(!NB! the last molecule is the surface)'
      end if
      write ( box_out_i, 503 ) 'estimated volume fraction: ', 4/3 * 3.14159 * (global_max_rad + 0.1)**3 * in_list%nb_mol / tmp_vol
      write ( box_out_i, * ) '--- BOX INFO ---'
      write ( box_out_i, 504 ) 'box begining: ', box_beg (:)
      if ( in_list%surf ) then
          write ( box_out_i, 504 ) 'box end: ', box_beg (:) + box_dim (:)
          write ( box_out_i, 504 ) 'box dimensions: ', box_dim (:)
      else
          write ( box_out_i, 504 ) 'box end: ', ( ( box_beg (i) + box_dim (i) + box_margin ), i = 1,3 )
          write ( box_out_i, 504 ) 'box dimensions: ', ( ( box_dim (i) + box_margin ), i = 1,3 )
      end if
      
      close ( box_out_i )
      
      call write_init_pos ( in_list%nb_mol, in_list%nb_types, prot_cntr_crds, prot_cntr_rot )
      
      !! REMEMBER TO DEALLOCATE ALL !!
      call delete_array_protein( tab_prot )
      call delete_react_criter( rcriter )
      deallocate ( prot_cntr_crds, prot_cntr_rot, size_order )
      deallocate ( nb_mol_types, nb_mol_vol, in_list%molw, in_list%dens, in_list%pdbs )
      !NJB: Not deallocated previously
      deallocate(rad_per_type,sseeds,mols_order)
      
      call exit
end program genbox


