!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
program read_binary_mat
   IMPLICIT NONE
   
   character*128 :: pdb_filename, out_filename
   integer, dimension(:), allocatable :: cur_int
   integer :: i, j, k, f, nargs, err, dot_idx, int_size, mat_row_size
   integer :: in_file_p, out_file_p
   integer :: nb_mol, nb_frames, type1, type2
   logical :: first_out, first_in_frame
   
   nargs = 1
   int_size = bit_size( nargs )
   write(*,*) 'INFO: int_size = ', int_size
   
   if( iargc() .ne. 1 ) then
      write(*,*) 'get_clusters_per_frame - reading binary contact matrix from a file and writing out ascii format summary'
      write(*,*) ''
      write(*,*) 'USAGE:'
      write(*,*) '    ./get_clusters_per_frame file_with_matrix'
      write(*,*) ''
      STOP
   end if
   
   call getarg( 1, pdb_filename )
   write(*,*) 'input argument - input file: ', pdb_filename
   
   !--------------
   ! opening the files
   in_file_p = 101
   open( in_file_p, file = pdb_filename, form = "unformatted", status = "old", iostat = err )
   if ( err .ne. 0 ) then
      write(*,*) " ERROR: could not open the file ", pdb_filename
      STOP
   end if
   
   dot_idx = index( pdb_filename, '.', back = .true. )
   if( dot_idx .eq. 0 ) then
      dot_idx = len( pdb_filename ) + 1
   end if
   write( out_filename, '(a,a)' ) pdb_filename( 1:( dot_idx - 1 ) ), "_summary.dat"
   out_file_p = in_file_p + 1
   open( out_file_p, file = out_filename, form = "formatted", status = "new", iostat = err )
   if ( err .ne. 0 ) then
      write(*,*) " ERROR: could not open the file ", out_filename, " as new!"
      STOP
   end if
   !--------------
   
   ! first read: nb_mol, nb_frames, type1, type2
   read( in_file_p ) nb_mol, nb_frames, type1, type2
   write(*,*) 'nb_mol: ', nb_mol
   write(*,*) 'nb_frames: ', nb_frames
   write(*,*) 'type1: ', type1
   write(*,*) 'type1: ', type2
   
!   nb_frames = nb_frames / 200
   mat_row_size = int( nb_mol / int_size ) + 1

   write(*,*) 'INFO: nb_mol / int_size = ', nb_mol / int_size
   
   allocate( cur_int( mat_row_size ), stat = err )
   if ( err .ne. 0 ) then
      write(*,*) " ERROR: could not allocate cut_int of size: ", mat_row_size
      STOP
   end if

   first_out = .true.
   do f = 1, nb_frames
      first_in_frame = .true.
      do i = 1, nb_mol
         read( in_file_p, iostat = err ) cur_int
         if( err .ne. 0 ) then
            write(*,'(a,i2,a,i5,a,i4,a)') &
               "ERROR (",err,"): no more data to read!(f: ",f,", i: ",i,")"
            flush( out_file_p )
            close( in_file_p )
            close( out_file_p )
            STOP
         end if
         do j = 1, mat_row_size
            do k = 1,int_size
               if ( ( j - 1 )*int_size + k > i) then   !only search the half matrix (duplicates)
                  if( btest( cur_int(j), k - 1 ) ) then
                     if( first_in_frame ) then
                        write( out_file_p, '(a,i5)' ) "frame ",f
                        first_in_frame = .false.
                     end if
                     if( first_out ) then
                       write( out_file_p, '  (i4)', advance = "no" ) i
                       first_out = .false.
                     end if
                     write( out_file_p, '(i4)', advance = "no" ) ( j - 1 )*int_size + k
                  end if
               end if 
            end do !k - looping over the bits in integer
         end do !j
         if( .not.first_out ) then
            write( out_file_p, '(a)' ) ""
            flush( out_file_p )
            first_out = .true.
         end if
      end do !nb_mol
   end do !frames
   
   !--------------
   ! closing files
   close( in_file_p )
   close( out_file_p )
   
end program read_binary_mat
