!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> @file make_edhdlj_grid.f90
!! Create electrostatic desolvation, apolar desolvation and repulsive lennard-jones grids

!> Prgram make_edhdlj_grid
!!
!! \ingroup tools
!! Regroup all the programs into one executable\n
!! Accept identical input file as sda6\n
!! Parameters specific to this tools are defined in mod_input_parameter.f90, under the type \ref mod_input_parameter::param_list_tools "param_list_tools"\n
!! Parrallelisaed with OpenMP
program make_edhdlj_grid
    USE mod_setofgrid ! include mod_grid

    !! in mod_input_parameter declare a specific type for tool : param_list_tools
    USE mod_input_parameter
    USE mod_sda_input
 !$ USE omp_lib
    IMPLICIT NONE

    interface

        subroutine compute_ed ( grid, list_atom, a2, a3, parameters )
            USE mod_grid
            USE mod_input_parameter
            IMPLICIT NONE
            type ( type_grid ) :: grid
            real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
            real ( kind=8 ), dimension ( : ) :: a2, a3
            type ( param_list_tools ), intent ( in ) :: parameters
        end subroutine compute_ed

        subroutine compute_hd ( grid, list_atom, array_van, parameters )
            USE mod_grid
            USE mod_input_parameter
            IMPLICIT NONE
            type ( type_grid ) :: grid
            real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
            real ( kind = 8 ), dimension ( : ), intent ( in ) :: array_van
            type ( param_list_tools ), intent ( in ) :: parameters
        end subroutine compute_hd

        subroutine compute_ljrep ( grid, list_atom, array_van, parameters )
            USE mod_input_parameter
            USE mod_grid
            IMPLICIT NONE
            type ( type_grid ) :: grid
            real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
            real ( kind = 8 ), dimension ( : ), intent ( in ) :: array_van
            type ( param_list_tools ), intent ( in ) :: parameters
        end subroutine compute_ljrep

    end interface

    type ( type_grid ) :: grid

    !! declare constant
    real ( kind=8 ), parameter :: ais2db = 1.0807858E-04, fkcal = 332.12793d0, epsi = 4.0d0

    !! use enumeration for grid_type
    integer, parameter :: ed_grid = 1, hd_grid = 2, ljrep_grid = 3
    integer :: grid_type

    !! defined in mod_input_parameter
    type ( param_list_tools ) :: parameters

    !! van der walls radius
    real ( kind=8 ), dimension ( : ), allocatable :: array_van
    real ( kind=8 ) :: radius_average, fios
    real ( kind=8 ), dimension ( : ), allocatable :: a3, a2
    real ( kind =8 ), dimension ( 3 ) :: center_protein, tmp_size
    real ( kind =8 ), dimension ( 3 ) :: shift_center

    !! will contain the pdb in list_atom_surface
    !! need to compile with trajectory2dcd who defines load_pdb
    type ( sogrid ), pointer :: p_sogrid

    character *128 :: argum
    integer :: nb_argument, count_arg, i
    logical :: bool_ed, bool_hd, bool_ljrep

    !! name of argument
    character*128 ed_input, hd_input, ljrep_input

    !! initialize to 0 by default
    grid_type = 0
    center_protein = 0.d0
    radius_average = 0.d0
    shift_center = (/0.d0,0.d0,0.d0/)

    nullify ( p_sogrid )
    !! by default
    nullify ( grid % p_grid )
    bool_ed = .false.
    bool_hd = .false.
    bool_ljrep = .false.
    ed_input = ''
    hd_input = ''
    ljrep_input = ''

    !nb_argument = iargc ()
    nb_argument = command_argument_count()
    if ( nb_argument == 0 ) then
        write(*,*)
        write(*,*) 'HELP'
        write(*,*) 'make_edhd_grid -ed ed.in -hd hd.in -ljrep ljrep.in'
        write(*,*)
        write(*,*) 'Compute Electrsotatic and/or Hydrophobic desolvation grid'
        write(*,*)
        write(*,*) '-ed ed.in, input parameters file, ed.in has the same format as in sda6 '
        write(*,*) '-hd hd.in, input parameters file, hd.in has the same format as in sda6 '
        write(*,*) '-ljrep ljrep.in, input parameters file,same format'
        write(*,*) '-shift_center X Y Z move the center of the grid'
        write(*,*)
        stop
    end if

    count_arg = 1

    do while ( count_arg <= nb_argument )

        call getarg ( count_arg, argum )

        if ( argum(1:3) == "-ed" ) then
            bool_ed = .true.
            call getarg( count_arg + 1 , argum )
            ed_input = trim ( argum )
            count_arg = count_arg + 1

        else if ( argum(1:3) == "-hd" ) then
            bool_hd = .true.
            call getarg( count_arg + 1 , argum )
            hd_input = trim ( argum )
            count_arg = count_arg + 1

        else if ( argum(1:6) == "-ljrep" ) then
            bool_ljrep = .true.
            call getarg( count_arg + 1 , argum )
            ljrep_input = trim ( argum )
            count_arg = count_arg + 1
        !endif

        else if ( argum(1:13) == "-shift_center" ) then
            call getarg( count_arg + 1 , argum )
            !hift_center[0]= trim
            read ( argum,* ) shift_center(1)
            call getarg( count_arg + 2 , argum )
            read ( argum,* ) shift_center(2)
            call getarg( count_arg + 3 , argum )
            read ( argum,* ) shift_center(3)
            count_arg = count_arg + 3
        end if

        count_arg = count_arg + 1
    end do

    !always print
    write(*,*)
    if ( bool_ed .eqv. .true. ) then
        write(*,*) 'ed file name ', trim ( ed_input )
    end if
    if ( bool_hd .eqv. .true. ) then
        write(*,*) 'hd file name ', trim ( hd_input )
    end if
    if ( bool_ljrep .eqv. .true. ) then
        write(*,*) 'ljrep file name ', trim ( ljrep_input )
    end if

    write(*,*) 'Shift the center of the grid(s) :', shift_center(1:3)
    ! because call function shift_origin ( orig = orig - center )
    shift_center = -shift_center

    if ( ( bool_ed .eqv. .false. ) .and. ( bool_hd .eqv. .false. ) .and. ( bool_ljrep .eqv. .false. ) ) then
        write(*,*) 'ERROR missing arguments -ed or -hd -lj'
        stop 1
    end if

    !! compute electrostatic desolvation
    if ( bool_ed .eqv. .true. ) then
        grid_type = ed_grid
        call read_input ( ed_input, ed_grid, parameters )

        call allocate_grid ( grid, parameters % dimension ( 1 ), parameters % dimension ( 2 ), &
                             parameters % dimension ( 3 ) )

        grid % h = parameters % hscale
        grid % title ="Electrostatic desolvation per unit_square charge kcal/mol"
        grid % size_x = parameters % dimension ( 1 )
        grid % size_y = parameters % dimension ( 2 )
        grid % size_z = parameters % dimension ( 3 )
        grid % iform = parameters % iform
        ! UHBD scaling factor to get kcal/mole
        grid % scale= 0.239006E-02

        !! read pdb file function defined in utils_tools.f90
        !! opt_center false, do not center the protein
        !! fill array_van with radius
        call load_pdb ( p_sogrid, parameters % pdb_name, .false., array_van )

        ! note initialized in load_pdb
        write(*,*) 'distance_max ', p_sogrid % distance_max

        !! compute center, radius_average, and a2 and a3 array
        allocate ( a2 ( p_sogrid % nb_atom_surface ), a3 (  p_sogrid % nb_atom_surface ) )

        do i = 1, p_sogrid % nb_atom_surface
            radius_average = radius_average + array_van ( i )
            a3 ( i ) = ( array_van ( i ) )**3
            a2 ( i ) = ( array_van ( i ) + 0.3d0 )**2

            center_protein = center_protein + p_sogrid % list_atom_surface ( :,i )
        end do

        radius_average = radius_average / float ( p_sogrid % nb_atom_surface )
        !write(*,*) 'radius_average ', radius_average
        radius_average = 2.d0 * ( parameters % rion + radius_average )
        center_protein = center_protein / float ( p_sogrid % nb_atom_surface )

        tmp_size = ( float ( parameters % dimension - 1 ) / 2.d0 ) * parameters % hscale
        grid % origin = center_protein - tmp_size - parameters % hscale

        !shift the origin if required
        call shift_origin ( grid, shift_center )

        write(*,*) 'center_protein ', center_protein
        write(*,*) 'computed origin ', grid % origin
        !write(*,*) 'new radius_average ', radius_average

        !! van_array allocated in load_pdb, can be deleted here
        deallocate ( array_van )

        parameters % deb = sqrt ( parameters % iostr * ais2db )
        !write(*,*) 'eps epsi, fkcal ', parameters % epssol, epsi, fkcal
        parameters % factor = ( parameters % epssol - epsi ) / ( epsi + 2.d0 * parameters % epssol ) / &
                              parameters % epssol * fkcal
        !write(6,*) 'energy factor', parameters % factor
        fios = exp ( 2.d0 * radius_average * parameters % deb ) / &
               ( 1.d0 +  radius_average * parameters % deb )**2
        parameters % rfios = 1.d0 / fios

        !write(*,*) 'deb ', parameters % deb
        !write(6,*) 'corrected to ios by a factor',fios

        parameters % factor= parameters % factor * fios
        parameters % radius_average = radius_average

        call compute_ed ( grid, p_sogrid % list_atom_surface, a2, a3, parameters )
        call write_grid ( grid, parameters % grid_name )

        !! delete all
        deallocate ( a2, a3 )
        call delete_sogrid ( p_sogrid  )
        deallocate ( p_sogrid )
        call delete_grid ( grid )

    end if

    !! compute hydrphobic desolvation
    if ( bool_hd .eqv. .true. ) then
        grid_type = hd_grid
        call read_input ( hd_input, grid_type, parameters )

        call allocate_grid ( grid, parameters % dimension ( 1 ), parameters % dimension ( 2 ), &
                             parameters % dimension ( 3 ) )

        grid % h = parameters % hscale
        grid % title ="hydrophobic desolvation energy per square angstroem"
        !grid % title='hydrophobic desolvation energy per square angstroem'
        grid % size_x = parameters % dimension ( 1 )
        grid % size_y = parameters % dimension ( 2 )
        grid % size_z = parameters % dimension ( 3 )
        grid % iform = parameters % iform
        ! UHBD scaling factor to get kcal/mole
        grid % scale= 0.239006E-02


        !! function defined in utils_tools.f90, will allocate array_van
        call load_pdb ( p_sogrid, parameters % pdb_name, .false., array_van )


        do i = 1, p_sogrid % nb_atom_surface
            !radius_average = radius_average + array_van ( i )
            center_protein = center_protein + p_sogrid % list_atom_surface ( :,i )
        end do

        !radius_average = radius_average / float ( p_sogrid % nb_atom_surface )
        !write(*,*) 'radius_average ', radius_average
        center_protein = center_protein / float ( p_sogrid % nb_atom_surface )

        tmp_size = ( float ( parameters % dimension - 1 ) / 2.d0 ) * parameters % hscale
        grid % origin = center_protein - tmp_size - parameters % hscale

        !shift the origin if required
        call shift_origin ( grid, shift_center )

        write(*,*) 'computed center:', center_protein
        write(*,*) 'computed origin:  ', grid % origin

        call compute_hd ( grid, p_sogrid % list_atom_surface, array_van, parameters )
        call write_grid ( grid, parameters % grid_name )

        !! delete all
        deallocate ( array_van )
        call delete_sogrid ( p_sogrid  )
        deallocate ( p_sogrid )
        call delete_grid ( grid )

    end if

    if ( bool_ljrep .eqv. .true. ) then
        grid_type = ljrep_grid
        call read_input ( ljrep_input, grid_type, parameters )

        call allocate_grid ( grid, parameters % dimension ( 1 ), parameters % dimension ( 2 ), &
                             parameters % dimension ( 3 ) )

        grid % h = parameters % hscale
        grid % title ="repulsive lennard-jones energy per .."
        !grid % title='hydrophobic desolvation energy per square angstroem'
        grid % size_x = parameters % dimension ( 1 )
        grid % size_y = parameters % dimension ( 2 )
        grid % size_z = parameters % dimension ( 3 )
        grid % iform = parameters % iform
        ! UHBD scaling factor to get kcal/mole
        grid % scale= 0.239006E-02


        !! function defined in utils_tools.f90, will allocate array_van
        call load_pdb ( p_sogrid, parameters % pdb_name, .false., array_van )

        do i = 1, p_sogrid % nb_atom_surface
            radius_average = radius_average + array_van ( i )
            center_protein = center_protein + p_sogrid % list_atom_surface ( :,i )
            !.. factor 0.5 to count vor a volume intersection discarded
            ! original in the loop, not needed
            !ah( i ) = array_van ( i )/fraction
        end do

        center_protein = center_protein / float ( p_sogrid % nb_atom_surface )
        radius_average = radius_average / float ( p_sogrid % nb_atom_surface )
        !! not need a loop
        array_van = array_van / parameters % fraction
        !! get origin
        tmp_size = ( float ( parameters % dimension - 1 ) / 2.d0 ) * parameters % hscale
        grid % origin = center_protein - tmp_size - parameters % hscale

        !shift the origin if required
        call shift_origin ( grid, shift_center )

        write(6,*) 'read in dime:  ', grid % size_x, grid % size_y, grid % size_z, grid % h
        write(6,*) 'param ', parameters % factor, parameters % nexp, parameters % fraction
        write(6,*) 'computed center:', center_protein
        write(6,*) 'computed origin:  ',grid % origin
        write(6,*) 'started with', p_sogrid % nb_atom_surface,' av. radius:', radius_average

        ! can precompute array_van(i)**(parameters%nexp)
        do i = 1, size ( p_sogrid % list_atom_surface ( 1,: ) )
            array_van ( i ) = array_van(i)**(parameters%nexp)
        end do
        ! same no need explicit loop
        !array_van = array_van**(parameters%nexp)

        call compute_ljrep ( grid, p_sogrid % list_atom_surface, array_van, parameters )
        call write_grid ( grid, parameters % grid_name )

        !! delete all
        deallocate ( array_van )
        call delete_sogrid ( p_sogrid  )
        deallocate ( p_sogrid )
        call delete_grid ( grid )

    end if

end program make_edhdlj_grid

    !> Read specific input file for the grids
    !! @param name_input : file name of the input
    !! @param grid_type : integer for the grid_type, refer to \ref mod_gridtype "mod_gridtype"
    !! @param parameters : specific parameters for tools \ref mod_input_parameter::param_list_tools "param_list_tools"
    subroutine read_input ( name_input, grid_type, parameters )
        USE mod_input_parameter
        IMPLICIT NONE
        character *128, intent ( in ) :: name_input
        integer, intent ( in ) :: grid_type
        type ( param_list_tools ), intent ( out ) :: parameters

        integer :: ierr, io_unit = 50

        !write(*,*) 'read_input ', grid_type

        open (io_unit, file=name_input , form= 'formatted', status='old', iostat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'Cannot open file ', trim ( name_input )
            stop 1
        end if

        read(50,*)
        read(50,*) parameters % hscale, parameters % dimension
        read(50,*)

        !! check if ed_grid
        if ( grid_type == 1 ) then
            read(50,*) parameters % iostr, parameters % epssol, parameters % rion
        !! else check if hd_grid
        else if ( grid_type == 2 ) then
            read(50,*) parameters % a, parameters % b, parameters % grid_value
        else if ( grid_type == 3 ) then
            read(50,*) parameters % factor, parameters % nexp, parameters % fraction
        end if

        read(50,*)
        read(50, fmt='(a)' ) parameters % pdb_name
        read(50,*)
        read(50,fmt='(a)' ) parameters % grid_name
        read(50,fmt='(i1)') parameters % iform
        close ( io_unit )

    end subroutine read_input

    !> Compute electrostatic desolvation
    subroutine compute_ed ( grid, list_atom, a2, a3, parameters )
        USE mod_grid
        USE mod_input_parameter
        IMPLICIT NONE

        interface
            subroutine dipole ( list_atom, a2, a3, grid_pos, parameters, energy )
                USE mod_input_parameter
                IMPLICIT NONE
                real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
                real ( kind=8 ), dimension ( : ), intent ( in ) :: a2, a3
                real ( kind=8 ), intent ( out ) :: energy
                type ( param_list_tools ), intent ( in ) :: parameters
                real ( kind=8 ), dimension ( 3 ), intent ( in ) :: grid_pos
            end subroutine dipole
        end interface

        type ( type_grid ) :: grid
        real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
        real ( kind=8 ), dimension ( : ) :: a2, a3
        type ( param_list_tools ), intent ( in ) :: parameters

        real ( kind=8 ), dimension ( 3 ) :: grid_pos
        real ( kind=8 ) :: energy, energy_max

        integer :: i, j, k

        energy_max = -100000.d0

#ifdef DEBUG
        write(*,*) 'Entry compute ed'
#endif

        !$OMP PARALLEL DEFAULT (NONE) &
        !$OMP SHARED ( grid, list_atom, parameters, a2, a3  ) &
        !$OMP PRIVATE ( grid_pos, energy )

        !$OMP DO SCHEDULE ( DYNAMIC )
        do i = 1, grid % size_x
#ifdef DEBUG
            write(*,*) 'i :',i
#endif
            grid_pos ( 1 ) = grid % origin ( 1 ) + i * grid % h
            do j = 1, grid % size_y
                grid_pos ( 2 ) = grid % origin ( 2 ) + j * grid % h
                do k = 1, grid % size_z
                    grid_pos ( 3 ) = grid % origin ( 3 ) + k * grid % h

                    call dipole ( list_atom, a2, a3, grid_pos, parameters, energy )
                    grid % p_grid % pgrid_real ( i, j, k ) = energy

                end do
            end do
        end do
       !$OMP END DO
       !$OMP END PARALLEL

        do i = 1, grid % size_x
          do j =  1, grid % size_y
            do k = 1, grid % size_z

              energy = grid % p_grid % pgrid_real ( i, j, k )
              if ( energy > energy_max ) energy_max = energy

            end do
          end do
        end do
        write(6,*)' max energy outside probe acc surface', energy_max

    end subroutine compute_ed

    !> Compute apolar desolvation
    subroutine compute_hd ( grid, list_atom, array_van, parameters )
        USE mod_input_parameter
        USE mod_grid
        IMPLICIT NONE

        interface
            subroutine hdenergy ( list_atom, grid_pos, array_van, parameters, energy )
                USE mod_input_parameter
                IMPLICIT NONE
                real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
                type ( param_list_tools ) :: parameters
                real ( kind = 8 ), dimension ( : ), intent ( in ) :: array_van
                real ( kind=8 ), dimension ( 3 ), intent ( in ) :: grid_pos
                real ( kind=8 ), intent ( out ) :: energy
            end subroutine hdenergy
        end interface

        type ( type_grid ) :: grid
        real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
        real ( kind = 8 ), dimension ( : ), intent ( in ) :: array_van
        type ( param_list_tools ), intent ( in ) :: parameters

        real ( kind=8 ), dimension ( 3 ) :: grid_pos
        real ( kind=8 ) :: energy_max, energy_min, energy
        integer :: i, j, k

#ifdef DEBUG
        write(*,*) 'Entry compute hd'
        !write (*,*) 'probe ', parameters % a, ' skin ', parameters % b, &
        !            ' value ', parameters % grid_value
#endif

        energy_max = 0.d0
        energy_min = 0.d0

        !$OMP PARALLEL DEFAULT (NONE) &
        !$OMP SHARED ( grid, list_atom, parameters, array_van ) &
        !$OMP PRIVATE ( grid_pos, energy )

        !$OMP DO SCHEDULE ( DYNAMIC )
        do i = 1, grid % size_x
#ifdef DEBUG
            write(*,*) 'i :',i
#endif
            grid_pos ( 1 ) = grid % origin ( 1 ) + i * grid % h
            do j = 1, grid % size_y
                grid_pos ( 2 ) = grid % origin ( 2 ) + j * grid % h
                do k = 1, grid % size_z
                    grid_pos ( 3 ) = grid % origin ( 3 ) + k * grid % h

                    call hdenergy ( list_atom, grid_pos, array_van, parameters, energy )
                    grid % p_grid % pgrid_real ( i, j, k ) = energy

                end do
            end do
        end do
        !$OMP END DO
        !$OMP END PARALLEL

        ! move after the // loop
        do i = 1, grid % size_x
          do j =  1, grid % size_y
            do k = 1, grid % size_z
              energy = grid % p_grid % pgrid_real ( i, j, k )
              if ( energy > energy_max ) energy_max = energy
              if ( energy < energy_min ) energy_min = energy
            end do
          end do
        end do

        write(*,*)' max energy outside probe acc surface', energy_max
        write(*,*)' min energy outside probe acc surface', energy_min

    end subroutine compute_hd

    !> Compute soft-core (repulsive lennard-jones)
    subroutine compute_ljrep ( grid, list_atom, array_van, parameters )
        USE mod_input_parameter
        USE mod_grid
        IMPLICIT NONE

        interface
            subroutine ljrep_energy (list_atom, array_van, grid_pos, parameters, energy )
                USE mod_input_parameter
                IMPLICIT NONE
                real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
                real ( kind = 8 ), dimension ( : ), intent ( in ) :: array_van
                real ( kind=8 ), dimension ( 3 ), intent ( in ) :: grid_pos
                type ( param_list_tools ), intent ( in ) :: parameters
                real ( kind=8 ), intent ( out ) :: energy
            end subroutine ljrep_energy
        end interface

        type ( type_grid ) :: grid
        real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
        real ( kind = 8 ), dimension ( : ), intent ( in ) :: array_van
        type ( param_list_tools ), intent ( in ) :: parameters

        real ( kind=8 ), dimension ( 3 ) :: grid_pos
        real ( kind=8 ) :: energy_max, energy
        integer :: i, j, k

        energy_max = 0.d0
        grid % p_grid % pgrid_real = 0.d0

#ifdef DEBUG
        write(*,*) 'Entry compute repulsive-lj'
#endif

        !$OMP PARALLEL DEFAULT (NONE) &
        !$OMP SHARED ( grid, list_atom, parameters, array_van ) &
        !$OMP PRIVATE ( energy, grid_pos )

        !$OMP DO SCHEDULE ( DYNAMIC )
        do i = 1, grid % size_x
#ifdef DEBUG
            write(*,*) 'i :',i
#endif
            grid_pos ( 1 ) = grid % origin ( 1 ) + i * grid % h
            do j = 1, grid % size_y
                grid_pos ( 2 ) = grid % origin ( 2 ) + j * grid % h
                do k = 1, grid % size_z
                    grid_pos ( 3 ) = grid % origin ( 3 ) + k * grid % h

                    call ljrep_energy (list_atom, array_van, grid_pos, parameters, energy)
                    grid % p_grid % pgrid_real ( i, j, k ) = energy

               end do
           end do
        end do
        !$OMP END DO
        !$OMP END PARALLEL

        ! make after the parallel loop
        do i = 1, grid % size_x
          do j =  1, grid % size_y
            do k = 1, grid % size_z
              energy = grid % p_grid % pgrid_real ( i, j, k )
              if ( energy > energy_max ) energy_max = energy
            end do
          end do
        end do

        write(*,*) 'maximum energy for lj_rep ', energy_max

    end subroutine compute_ljrep

    !> Used for electrostic desovaltion computation
    subroutine dipole ( list_atom, a2, a3, grid_pos, parameters, energy )
        USE mod_input_parameter
        IMPLICIT NONE
        real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
        real ( kind=8 ), dimension ( : ), intent ( in ) :: a2, a3
        real ( kind=8 ), intent ( out ) :: energy
        type ( param_list_tools ), intent ( in ) :: parameters
        real ( kind=8 ), dimension ( 3 ), intent ( in ) :: grid_pos

        real ( kind=8 ), dimension ( : ), allocatable :: dist_square
        real ( kind = 8 ) :: energy_d, dist1, dist2, dist4, debr, expf
        real ( kind= 8 ) :: deb, rfios, radius_average
        integer :: nat, is

        nat = size ( list_atom ( 1,: ) )
        allocate ( dist_square ( nat ) )

        energy = 0.d0
        energy_d = 0.d0

        radius_average = parameters % radius_average
        deb = parameters % deb
        rfios = parameters % rfios

        ! tabulate distances and discard inner points
        do is = 1, nat
            dist_square ( is ) = (  list_atom ( 1, is ) - grid_pos ( 1 ) )**2 + &
                                 (  list_atom ( 2, is ) - grid_pos ( 2 ) )**2 + &
                                 (  list_atom ( 3, is ) - grid_pos ( 3 ) )**2

            if ( dist_square ( is ) < a2 ( is ) ) return
        end do

        !compute energy
        do is = 1, nat
            !! trick or bug ??
            !! trick if want to get value of the previous atom !!
            !dist1 = sqrt ( dist2 )
            !dist2 = dist_square ( is )

            dist2 = dist_square ( is )
            dist1 = sqrt ( dist2 )
            dist4 = dist2 * dist2

            if ( dist1 <= radius_average ) then
                energy_d = energy_d + rfios / dist4 * a3 ( is )
            else
                debr = deb * dist1
                expf = exp ( -debr * 2.d0 )
                energy_d = energy_d + ( 1.d0 + debr ) ** 2 * expf / dist4 * a3 ( is )
            end if
        end do

        energy = energy_d * parameters % factor
        deallocate ( dist_square )

    end subroutine dipole

    !> Use for apolar desolvation computation
    subroutine hdenergy ( list_atom, grid_pos, array_van, parameters, energy )
        USE mod_input_parameter
        IMPLICIT NONE
        real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
        type ( param_list_tools ) :: parameters
        real ( kind=8 ), dimension ( 3 ), intent ( in ) :: grid_pos
        real ( kind = 8 ), dimension ( : ), intent ( in ) :: array_van
        real ( kind=8 ), intent ( out ) :: energy

        real ( kind=8 ) :: d1, d2, efct, enerd, dmin, dist_square
        integer :: is, nat

        enerd = 0.d0
        energy = 0.d0
        d1 = parameters % a
        d2 = parameters % b
        efct = parameters % grid_value

        nat = size ( list_atom ( 1,: ) )

        ! tabulate distances and discard inner points
        dmin=9999.999

        do is = 1, nat
            dist_square=sqrt( (list_atom (1,is) - grid_pos (1))**2 + &
                              (list_atom (2,is) - grid_pos (2))**2 + &
                              (list_atom (3,is) - grid_pos (3))**2 ) &
                             - array_van ( is )

            if ( dist_square .lt. dmin  ) dmin = dist_square
        end do

        if ( dmin <= d1 ) then
            enerd = 1.d0
        else if ( dmin >= d2 ) then
            enerd = 0.d0
        else
            enerd = ( d2 - dmin ) / ( d2 - d1 )
        end if

        energy = enerd * efct

    end subroutine hdenergy

    !> Use for repulsive lennard-jones computation
    subroutine ljrep_energy ( list_atom, array_van, grid_pos, parameters, energy )
        USE mod_input_parameter
        IMPLICIT NONE
        real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: list_atom
        real ( kind = 8 ), dimension ( : ), intent ( in ) :: array_van
        real ( kind=8 ), dimension ( 3 ), intent ( in ) :: grid_pos
        type ( param_list_tools ), intent ( in ) :: parameters
        real ( kind=8 ), intent ( out ) :: energy

        real ( kind=8 ), dimension ( : ), allocatable :: dist_square
        real ( kind=8 ) :: dist
        integer :: nat, i

        energy = 0.d0

        nat = size ( list_atom ( 1,: ) )
        !! should allocated before, only once
        allocate ( dist_square ( nat ) )

        do i = 1, nat
            dist_square ( i ) = (  list_atom ( 1, i ) - grid_pos ( 1 ) )**2 + &
                                (  list_atom ( 2, i ) - grid_pos ( 2 ) )**2 + &
                                (  list_atom ( 3, i ) - grid_pos ( 3 ) )**2
        end do


        ! could precompute array_van(i)**(parameters%nexp)
        do i = 1, nat
            dist = sqrt ( dist_square ( i ) )
            energy = energy +  parameters % factor / &
                     ( array_van ( i ) + dist**(parameters%nexp) )
            ! original
            !energy = energy +  parameters % factor / &
            !          ( array_van ( i )**(parameters%nexp) + dist**(parameters%nexp) )
        end do

        deallocate ( dist_square )

    end subroutine ljrep_energy

