!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file nos2rates.f90
!! Estimate average and standard deviation of the association rates from multiple sda runs


!> Extract association rates, and compute average and dispersion from output files of sda_flex
!! \ingroup tools
!!
!! This is the original method in previous sda versions\n
!! It is updated with the bootstrap algorithm and the python auxiliary program
program nos2rates
    IMPLICIT NONE
    integer :: nb_contact, nb_infile
    integer :: io_change, counter, nwin
    logical :: not_found_key
!.
!. usage
!. copy first the sda outputs to fort.11 fort.12 fort.13 etc,
!. the first empty or invalid file will define the number of outputs to analyze
!. then run this: os2rates
!. stdout will give the table with
!. reaction_window -- rate -- deviation of rates derived from all runs.
!.
!. alternatively - give sda output file names in the command line
!.
    CHARACTER*80 string
    CHARACTER*80 fname
    CHARACTER*6 key
    CHARACTER*20 key0
    real ( kind=8 ), dimension (:), allocatable :: rate
    real ( kind=8 ) :: rav, rdisp, x
    DATA key/' Final'/
    DATA key0/'rxn windows start at'/

    integer :: irea, i, io, ii, idis

    nwin = 0
    nb_contact = 0
    not_found_key = .true.

    !if ( iargc() == 0 ) then
    if ( command_argument_count() == 0 ) then
        write (6,601)
  601   FORMAT ('os2rates - extract rates from SDA output',/, &
                'usage: nos2rates outf1 outf2 outf3 ...',/, &
                '       outf1 etc - SDA output files')
        stop 1
    end if
!
    !nb_infile = iargc()
    nb_infile = command_argument_count()
    !write(*,*) 'number of input file ', nb_infile

    !! init first integer descriptor to 11, and read size of array from the first file
    io_change = 11

    !! trick to avoid one variable
    call getarg ( io_change - 10, fname )
    !write(*,*) 'fname ', fname
    open (io_change, file=fname,form='formatted',status='UNKNOWN')

    do while ( not_found_key )
        read ( io_change, fmt='(a)', iostat=io ) string

        if ( io /= 0 ) then
            write(*,*) 'reach end of file io', io_change
            stop 1
        end if

        if ( string( 1:20 ) == key0 ) then
            not_found_key = .false.
        end if
    end do

    if ( not_found_key ) then
        write(*,*) 'keyword not found in the first file'
        stop 1
    end if

    !reset default value
    not_found_key = .true.

!association rate :
!rxn windows start at:       3.000 A
!rxn winndow increment:      1.000 A
!no of rxn windows:             40
!no of contact:                  4
! new format
!rxn windows start at:         3.000 A
!rxn winndow increment:        0.500 A
!no of rxn windows:               35
!no of contact:                    4


    !! read parameters in the first file
    read (io_change, fmt='(a)' ) string
    read (io_change, fmt='(a)' ) string
    read ( string(29:35),* ) nwin
    read (io_change, fmt='(a)' ) string
    read ( string(29:35),* ) nb_contact
    !write(*,*) 'nwin ', nwin
    !write(*,*) 'nb_contact ', nb_contact

    !! place all the file to the position for reading rate data
    !nouts = nb_infile

    counter = 1
    do while ( counter <= nb_infile )
        io_change = counter + 10
        if (io_change.EQ.61) then
           write (*,'("Maximum number of input files (",I3,") exceeded. Exiting.")') counter-1
           stop 1
        end if
        if ( counter > 1 ) then
            call getarg( io_change-10, fname )
            !write(*,*) 'open next file ', fname
            open (io_change,file=fname, form='formatted', status='UNKNOWN')

        end if
        !!look for ' Final'
        do while ( not_found_key )

            read (io_change, fmt='(a)', iostat = io ) string
            if ( io /= 0 ) then
                write(*,*) 'reach end of file io', io_change
                stop 1
            end if

            if ( string(1:6) == key ) then
                not_found_key = .false.
            end if

        end do
        !read one line more
        read (io_change,*)

        not_found_key = .true.
        counter = counter + 1

    end do

    !! allocate array for rates
    allocate ( rate ( nb_infile ) )
    rate = 0.d0

!
!.. the rate records are found, average "nouts" of them
!
    do irea = 1,nb_contact

        write (61,*)
        write (6,*)
        do i = 1,nb_infile
            io_change = i + 10
            read ( io_change, * )
        end do

        do idis = 1,nwin
            !! read data and compute average
            rav = 0.d0
            do i = 1,nb_infile
                io_change = i + 10
                read (io_change,501) x,rate(i)
                rav = rav + rate(i)
            end do
            rav = rav/float(nb_infile)

            !! compute dispersion
            rdisp = 0.d0
            do i = 1, nb_infile
                rdisp = rdisp + (rate(i)-rav)**2
            end do
            rdisp = sqrt( rdisp / float( nb_infile) )

            write (61,501) x,rav,rdisp, (rate(ii),ii=1,nb_infile)
            write (6,501) x,rav,rdisp

        end do ! idis

    end do ! irea

    if ( allocated ( rate ) ) deallocate ( rate )

  501 FORMAT (3e12.4,9e11.3)

end program
