!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> @file rot_diffusion.f90
!! Compute the autocorrelation function of the rotation and the rotational diffusion coefficients

!> Program rot_diffusion
!!
!! \ingroup tools
!! Correlation functions computed in parrallel with OpenMP\n
!! Different methods are implemented for fitting the diffusion coefficients
!!
!! Common function are implemented in utils_diffusion.f90
!!
!!  2 methods acf : short time diffusion\n
!!             msd : short and long time, it is an approximation, to try better\n
!!
!! sdamm tool : acf_multyprot.f , use for rotational diffusion constant\n
!!              rotdiff_msd_ensavg_multyprot.f : compute approximation to msd\n
!!              msd_timeensavg_multyprot.f, use for translational diffusion coefficient\n
!!
!! required :\n
!!   flexibility ? certianly not\n
!!   surface ? only z ? to do\n
!
!! can be done :\n
!!   group both rotation and translation diffusion coefficient, wait for paolo answer\n
!!  for all trajectory sda_2proteins/ sda_koff ? problem timestep\n
!!   option to read conformation by conformation or read all ( or large number of steps ) and use OpenMP for acf calculation\n
!!
!! For acf :\n
!!   Z axis is not needed, save computation (~ 20 % ) time and accuraty ( ACF(0) == 1. )\n
!!   Z only cross_product of X ^ Y\n
program rot_diffusion

    USE mod_record
    USE mod_input_parameter
    USE mod_sda_input
    USE AllTimer
!$  USE omp_lib
    IMPLICIT NONE

    interface
        subroutine all_acf_rotation ( orient_axis, tab_prot, ACF, nchunks, b_type, value_type ) !, dt ) ! parameter
            USE mod_array_protein
            IMPLICIT NONE
            ! x, y, z, nb_prot, time
            real ( kind=8 ), dimension (:,:,:,:) :: orient_axis
            type ( array_protein_type ) :: tab_prot
            ! Auto correlation function ( ntype, tau )
            real ( kind=8 ), dimension (:,:,:) :: ACF
            !type ( param_list ) :: parameters
            integer :: nchunks, value_type
            logical :: b_type
        end subroutine all_acf_rotation

        subroutine all_msd_rotation ( orient_axis, tab_prot, ACF, & !D_COEFF_DERIV, D_COEFF_SDERIV,
                                      nchunks, b_type, value_type, nb_region, dt, io_msd )
            USE mod_array_protein
            IMPLICIT NONE
            ! x, y, z, nb_prot, time
            real ( kind=8 ), dimension (:,:,:,:), intent ( in ) :: orient_axis
            type ( array_protein_type ), intent ( in ) :: tab_prot
            ! Auto correlation function ( ntype, tau )
            real ( kind=8 ), dimension (:,:,:), allocatable :: ACF !, D_COEFF_DERIV, D_COEFF_SDERIV
            !type ( param_list ) :: parameters
            integer, intent ( in ) :: nchunks, value_type, nb_region, io_msd
            logical, intent ( in ) :: b_type
            real ( kind=4 ), intent ( in ) :: dt
        end subroutine all_msd_rotation

        !> defined in trans_diffusion, not used yet maybe later
        subroutine print_acf ( ACF, time_snapshot, b_type, value_type, title )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:), intent ( in ) :: ACF
            character, intent( in ) :: title*(*)
            integer :: value_type
            logical :: b_type
            real (kind=4) :: time_snapshot
        end subroutine print_acf

        !subroutine print_coeffd_msd ( D_COEFF, b_type, value_type, b_trans, io_msd )
        !    IMPLICIT NONE
        !    real ( kind=8 ), dimension (:,:,:) :: D_COEFF
        !    integer :: value_type, io_msd
        !    logical :: b_type, b_trans
        !end subroutine print_coeffd_msd

        subroutine print_acfrot_sdamm ( ACF, time_snapshot, b_type, value_type, io_acf )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:) :: ACF
            integer :: value_type, io_acf
            logical :: b_type
            real (kind=4) :: time_snapshot
        end subroutine print_acfrot_sdamm

    end interface

    type ( record ) :: record_traj
    type ( array_protein_type ) :: tab_protein

    ! for reading record, even if not used
    !! keep 8 because binary wrote with 8, should change all to 4
    real ( kind = 8 ), dimension ( 2, 3 ) :: cm
    integer :: nb_line

    ! for reading input file, initialized in read_input_file
    type ( param_list ) :: list_parameter
    character*64, dimension(:,:), allocatable :: all_filename_pdb
    logical :: iflex

    !! for log files, to add for other method
    integer :: io_msd = 21, io_acf = 22

    type ( mytimer ) :: timer_read, timer_acf, timer_read2

    ! for reading argument
    character*64 str_arg, method_type
    character*64 fname_traj_input, filename_input
    integer :: nb_argument, count_arg
    integer :: input_traj_io
    logical :: bool_input
    logical :: bool_acf, bool_msd


    ! for loop over record
    !type ( one_complexe ), pointer :: p_tmp_onecomplexe
    ! number proteins and number of record (= total_line / nb_protein)
    integer :: nb_prot, nb_record !,nb_type
    integer :: nloop, tmp_prot
    integer :: ierr
    real ( kind=4 ) :: time_snapshot

    ! Auto correlation function ( nchunks, ntype, tau ), declared here, used in both methods
    real ( kind=8 ), dimension (:,:,:), allocatable :: ACF
    ! for storing mean and variance of diffusion coefficient at short and long term
    ! fit either the real derivative, or the simplified derivative, moved to function
    !real ( kind=8 ), dimension (:,:,:), allocatable :: D_COEFF_DERIV, D_COEFF_SDERIV

    ! array to store orientation of all proteins, at all time, first test maybe not needed
    ! ( 3, 3, nb_prot, time ) orient_x, orient_y, orient_z, time
    real ( kind=8 ), dimension (:,:,:,:), allocatable :: orient_axis
    real ( kind=8 ), dimension(3) :: orient_x, orient_y !, orient_z not necessary

    ! arguments
    integer :: nchunks, value_type, nb_region
    logical :: b_type

    !nullify ( p_tmp_onecomplexe )

    ! default initialisation
    fname_traj_input = ''
    filename_input = ''
    bool_input = .false.
    count_arg = 1
    input_traj_io = 155
    nb_line = 0

    ! default arguments
    nchunks = 1
    nb_region = 3
    b_type = .false.
    bool_acf = .true.
    bool_msd = .true.
    ! 0 by default means all
    value_type = 0

    !! initialise timer
    call init_mytimer ( timer_read )
    call init_mytimer ( timer_acf )
    call init_mytimer ( timer_read2 )

    !$  write(*,*) 'OpenMP activated'

    !! give available on hte machine, but not used here, only thread used
    !$  write(*,*) 'nb processor on this machine ', OMP_get_num_procs()

    !! give only the thread at this point
    !$  write(*,*) 'nb of threads ', OMP_get_num_threads()
    !$  write(*,*) 'max nb of threads ', OMP_get_max_threads()
    !$  write(*,*) 'dynamic activated ', OMP_get_dynamic()

    ! init record_traj
    nullify ( record_traj % pcomplexe, record_traj % ptrajectory, record_traj % ptab_prot )
    record_traj % io_compl = input_traj_io
    record_traj % header_str = ''

!!!!!!!!!!!!! Read arguments !!!!!!!!!!!!!!!!!!!!!!

    !nb_argument = IARGC ()
    nb_argument = command_argument_count()

    if ( nb_argument == 0 ) then
        write(*,*) ' rot_diffusion   - calculate the rotational diffusion from a SDAMM trajectory'
        write(*,*) ''
        write(*,*) 'USAGE:'
        write(*,*) '  acf  trajectory_input  [ -input sda.in ]  -method msd/acf  -nchunks 4  -solute_type'
        write(*,*)
        write(*,*) '-method      : msd or acf, rotational mean square displacement or autocorrelation function; default both'
        write(*,*)
        write(*,*) '-nchunks X   : split the trajectory into X equal part, and compute for the different parts; default = 1'
        write(*,*)
        write(*,*) '-nb_blocks   : split each chunk into nb_block for the least_square fitting; default = 3'
        write(*,*)
        write(*,*) '-solute_type [X] : optional value; '
        write(*,*) '               if X absent, compute for all types of solutes;'
        write(*,*) '               if X given, compute only for type X ( begin at 1 )'
        stop
    end if

    do while ( count_arg <= nb_argument )

      call getarg( count_arg, str_arg )

      if ( count_arg == 1 ) then
          !write ( *, * ) 'read file ', trim ( str_arg )
          fname_traj_input = str_arg

      else if ( str_arg ( 1:6 ) == "-input" ) then
          bool_input = .true.
          call getarg( count_arg +1, str_arg )
          filename_input = trim( str_arg )
          count_arg = count_arg + 1

      else if ( str_arg ( 1:8 ) == "-nchunks" ) then
          call getarg(  count_arg +1, str_arg )
          read ( str_arg, fmt='(i3)' ) nchunks
          count_arg = count_arg + 1

      else if  ( str_arg ( 1:8 ) == "-nblocks" ) then
          call getarg(  count_arg +1, str_arg )
          read ( str_arg, fmt='(i3)' ) nb_region
          count_arg = count_arg + 1

      else if  ( str_arg ( 1:12 ) == "-solute_type" ) then
          b_type = .true.
          call getarg(  count_arg +1, str_arg )
          read ( str_arg, fmt='(i3)', iostat = ierr ) value_type
          if ( ierr == 0 ) then
              count_arg = count_arg + 1
          !else
          !    write(*,*) 'use default value_type 0'
          !    stop 1
          end if

      else if  ( str_arg ( 1:12 ) == "-method" ) then
          bool_acf = .false.
          bool_msd = .false.
          call getarg(  count_arg +1, str_arg )
          read ( str_arg, fmt='(a)', iostat = ierr ) method_type
          if ( ierr == 0 ) then
              count_arg = count_arg + 1
          !else
          !    write(*,*) 'use default value_type 0'
          !    stop 1
          end if
          write(*,*) 'method_type ', trim ( method_type )
          if ( trim(method_type) == 'acf' ) then
              bool_acf = .true.
          else if ( trim(method_type) == 'msd' ) then
              bool_msd = .true.
          end if

      end if

      count_arg = count_arg + 1

    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    write(*,*)
    write(*,*) 'Entry Rotational Diffusion'
    write(*,*)

    if ( ( fname_traj_input == '' ) .or. ( filename_input == '' ) ) then
        write(*,*) 'arguments missing'
        stop 1
    end if

    write(*,*) 'filename ', trim(fname_traj_input)
    write(*,*) 'method acf :', bool_acf, ', msd :', bool_msd
    write(*,*) 'nchunks ', nchunks
    if ( b_type ) then
        write(*,*) 'solute_type ON value=', value_type
    else
        write(*,*) 'solute_type OFF'
    endif
    write(*,*)

    if ( bool_input ) write(*,*) 'sda_input=',trim( filename_input )

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                                                               ! all_filename_pdb, iflex, load_pdb, extend_array
    call read_input_sda ( tab_protein, filename_input, list_parameter, all_filename_pdb, iflex, .false., .false. )
    !write(*,*) 'list_parameter % total_solute ', list_parameter % total_solutes
    !write(*,*) 'list_parameter % dt ', list_parameter % dt1
    !write(*,*) 'list % freq ', list_parameter % freq_print

    ! time in nanosecond
    time_snapshot = real ( list_parameter % dt1 * list_parameter % freq_print / 1000.0 ,4)
     write(*,*) 'time_snapshot in ns. ', time_snapshot
    ! false, do not test if p_sogrid is associated
    call print_info_array_protein ( tab_protein, .false. )

  call start_mytimer( timer_read )
    ! read header, initialize record
    !! if option nb_line will compute number of record and setup variable field
    call read_header_record ( record_traj, fname_traj_input, cm, nb_line )
    write(*,*) 'nb_line ', nb_line
    !write(*,*) 'record_traj % nb_max_complexe ', record_traj % nb_max_complexe
    !write(*,*) 'record_traj % type_record ', record_traj % type_record
    !write(*,*) 'record_traj % record_type_calc % enum_type_calc ', record_traj % record_type_calc % enum_type_calc
    !nb_line = record_traj % nb_max_complexe

    !! check if sdamm and trajectory, may be less strict later ?
    if ( ( record_traj % record_type_calc % enum_type_calc == sdamm ) .and. &
         ( record_traj % type_record == trajectory_enum ) ) then
        !write(*,*) 'read sdamm trajectory '
        !nb_prot = record_traj % nb_protein
    else
        write(*,*) 'not a sdamm trajectory, not implemented '
        stop 1
    end if

    !! initialize trajectory
    if ( record_traj % type_record == trajectory_enum ) then
        allocate ( record_traj % ptrajectory )
        !! allocate array of new complexe with size array = record_traj % nb_max_complexe
        !! could add optional max_size, read conf by conf
        call initialize_trajectory ( record_traj % ptrajectory, record_traj )
        !write(*,*) 'Trajectory, init array with size ', record_traj % ptrajectory % size_array
        !write (*,*) 'size array for trajectory ', record_traj % ptrajectory % size_array

    ! a pririo, not used
    else if ( record_traj % type_record == complexe_enum ) then
        !! creates the list, no data yet
        !call new_listcomplexe ( this % pcomplexe, rmsd_min, one_run, one_conf )
        !call new_listcomplexe ( record_traj % pcomplexe, 0.d0, .false., .false. )
        write(*,*) ' not implemented for complexes file'
        stop 1
    else
        write(*,*) 'problem with record_traj % type_record, will stop '
        stop 1
    end if

    !! can read both trajectory or complexe, read all the trajectory into memory
    call read_file_record ( record_traj )
    ! rewind to the first element
    call reset_position ( record_traj, ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Record contains no data'
        stop 1
    end if
    ! point to the first
    !p_tmp_onecomplexe => record_traj % p_onecompl

  call stop_mytimer( timer_read )

    ! make loop
    nb_prot = record_traj % nb_protein
    !write(*,*) 'nb_prot ', nb_prot
    nb_record = record_traj % nb_max_complexe  / nb_prot
    !write(*,*) 'nb_record ', nb_record

    ! need global array orient_axis, efficient for storage and for parrallelisation
    ! vector x,y,z,nb_prot,time
    allocate ( orient_axis ( 3,3, nb_prot, nb_record ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating orient_axis'
        stop 1
    end if
    orient_axis = 0.d0

    ! global array, avoids to recompute cross many, many times !!
    nloop = 1
    ierr = 0

  call start_mytimer( timer_read2 )
    ! loop over all trajectory, do nloop = 1, nb_record, better for OpenMP
    do while ( ( ierr == 0 ) .and. ( nloop <= nb_record ) )
        !write(*,*) 'new record ', nloop, ' tmp_pos_in ', tmp_pos_in

        ! loop over all proteins
        do tmp_prot = 1, nb_prot

            ! read x,y orientation
            orient_x = record_traj % p_onecompl % orient_x
            orient_y = record_traj % p_onecompl % orient_y

            ! reconstruct z
            call norm ( orient_x )
            call norm ( orient_y )
            !call cross ( orient_z, orient_x, orient_y )

            ! assign to global array
            orient_axis (:, 1, tmp_prot, nloop ) = orient_x
            orient_axis (:, 2, tmp_prot, nloop ) = orient_y
            orient_axis (:, 3, tmp_prot, nloop ) = 0.d0 !orient_z

            ! update p_tmp_onecomple at the end of the loop only
            ! add option timestep for skipping frames
            call getNext(  record_traj, ierr )
            ! test ierr ??

        ! end do loop over proteins
        end do

        nloop = nloop + 1
    ! end loop over record
    end do

  call stop_mytimer( timer_read2 )

    ! can delete record, save memory
    call delete_record ( record_traj, .true. )

!!!!!!!!!!!!!!!!!!!!!!!!!! Compute autocorrelation and/or msd


  call start_mytimer ( timer_acf )

    if ( bool_acf .eqv. .true. ) then
        write(*,*) 'compute acf '
        call all_acf_rotation ( orient_axis, tab_protein, ACF, nchunks, b_type, value_type ) ! list_parameter
        call print_acfrot_sdamm ( ACF, time_snapshot, b_type, value_type, io_acf )
        ! standart output all in one file
        !call print_acf ( ACF, time_snapshot, b_type, value_type, 'Rot_ACF.dat' )

    end if

    !! may have been allocated in acf_calculation
    if ( allocated ( ACF ) ) then
        deallocate ( ACF, stat =ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'Cannot deallocate ACF'
        end if
    end if


    if ( bool_msd .eqv. .true. ) then
        !write(*,*) ''
        write(*,*) 'compute rot msd dt ', time_snapshot
        call all_msd_rotation ( orient_axis, tab_protein, ACF, nchunks, b_type, &
                                value_type, nb_region, time_snapshot, io_msd )
        ! close after all output done
        close ( io_msd )
    end if

  call stop_mytimer ( timer_acf )
    ! delete tab_protein
    call delete_array_protein ( tab_protein )

    write(*,*) 'timer read'
    call print_mytimer ( timer_read )
    write(*,*) 'timer read2'
    call print_mytimer ( timer_read2 )
    write(*,*) 'timer computation'
    call print_mytimer ( timer_acf )


end program rot_diffusion

!> Compute all autocorrelation function of rotation
!!
!! Can deal with splitting the trajectory with nchunks
!!
!! @param orient_axis : array containing the orientation of the solute during the trajectory
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param ACF : return, autocorrelation functions
!! @param nchunks : how many parts the simulation is splitted
!! @param b_type : logical, compute for different type of proteins if .true.
!! @param value_type : if 0 compute for all proteins AND for all types of proteins, if > 0 compute only for this type
!!                         could improve, read in orient_axis only for this type and forget about this argument
subroutine all_acf_rotation ( orient_axis, tab_prot, ACF, nchunks, b_type, value_type )
    USE mod_array_protein
    USE Alltimer
    IMPLICIT NONE

    interface
        subroutine autocorrelation_orient ( input_data, output_data, size_input )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:,:), intent ( in ) :: input_data
            real ( kind=8 ), dimension (:), intent ( out ) :: output_data
            integer, intent ( in ) :: size_input !, max_tau
        end subroutine autocorrelation_orient
    end interface

    ! x, y, z, nb_prot, time
    real ( kind=8 ), dimension (:,:,:,:) :: orient_axis
    type ( array_protein_type ) :: tab_prot
    ! Auto correlation function ( nchunks, ntype, tau )
    real ( kind=8 ), dimension (:,:,:), allocatable :: ACF
    !type ( param_list ) :: parameters
    integer :: nchunks, value_type
    logical :: b_type

    integer :: startstep, endstep, nacf, nstep_acf
    integer :: first_protein, last_protein, first_type, last_type
    !real ( kind=4 ) :: dt, max_tau

    ! local variable
    integer :: nb_prot, nb_type, size_input !itype !max_size
    integer :: ierr, itype, shift_acf

    type ( mytimer ) :: timer_acf

    call init_mytimer ( timer_acf )

    nb_type = -1
    !write(*,*) 'Entry all_acf_rotation '
    nb_prot = size ( tab_prot % array )
    !write(*,*) 'nb_prot ', nb_prot
    !write(*,*) 'nchunks ', nchunks

    !array_shape = shape ( orient_x )
    !max_size = array_shape(4)
    size_input = size ( orient_axis, 4 )
    !write(*,*) 'shape ', shape ( orient_axis )
    !write(*,*) 'max_size ', max_size

    ! want for all proteins and for the different types
    if ( ( b_type ) .and. ( value_type == 0 ) ) then
        nb_type = tab_prot % nb_grids + 1
    ! only one computation
    else if ( ( ( b_type ) .and. ( value_type > 0 ) ) .or. &
              ( b_type .eqv. .false. ) ) then
        nb_type = 1
    end if
    !write(*,*) 'nb_type ', nb_type
    ! account for nchunks
    nstep_acf = int ( size_input / nchunks )

    ! allocate to the correct size
    allocate ( ACF ( nchunks, nb_type, nstep_acf ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating ACF'
        stop 1
    end if
    ACF = 0.d0

    shift_acf = 0
    ! default, loop over all types
    first_type = 1
    last_type = tab_prot % nb_grids

LOOP_CHUNK: do  nacf = 1, nchunks

        startstep = nstep_acf * ( nacf - 1 ) + 1
        endstep = nstep_acf * ( nacf )

        write(*,*) 'startstep ', startstep
        write(*,*) 'endstep ', endstep
        write(*,*) 'nstep_acf ', nstep_acf
        !write(*,*) 'endstep-startstep+1 ', endstep-startstep+1

        ! compute acf for all proteins, send full array
        if ( (( b_type ) .and. (value_type == 0)) .or. &
              ( b_type .eqv. .false. ) ) then

             !write(*,*) 'b_type is true and compute for all protein types'
             call start_mytimer ( timer_acf )
             call autocorrelation_orient ( orient_axis(:,:,:,startstep:endstep), ACF( nacf, 1,: ), &
                                          nstep_acf )
                                          !endstep-startstep+1 , endstep-startstep+1 )
            call stop_mytimer ( timer_acf )
             ! if no type calculation required
             if ( b_type .eqv. .false. ) then
                CYCLE LOOP_CHUNK
             end if

            ! if add type acf, will be shifted by one
            shift_acf = 1

        ! else only acf for one type
        else if ( ( b_type ) .and. ( value_type > 0 ) ) then

            ! will limit the loop over the grid
            first_type = value_type
            last_type = value_type

            ! assur than ACF(nacf,1,:) is filled
            shift_acf = 1 - value_type

        end if ! end of tests


        !! calculations of the different types
        do itype = first_type, last_type

                first_protein = tab_prot % set_of_protein ( itype, 1 )
                last_protein  = tab_prot % set_of_protein ( itype, 2 )
                !write(*,*) 'first last prot ', first_protein, last_protein
                call start_mytimer ( timer_acf )

                call autocorrelation_orient ( orient_axis(:,:, first_protein:last_protein, startstep:endstep), &
                                          ACF( nacf, itype + shift_acf,: ), nstep_acf )
                                          !endstep-startstep+1 , endstep-startstep+1 )

                call stop_mytimer ( timer_acf )
        end do

    ! end do nchunks
    end do LOOP_CHUNK

    write(*,*)"timer autocorrelation"
    call print_mytimer ( timer_acf )


end subroutine all_acf_rotation

!> This method uses rotational mean square displacement ( rmsd )
!!  computed with a trick from above reference
subroutine all_msd_rotation ( orient_axis, tab_prot, ACF, nchunks, b_type, value_type, &
                              nb_region, dt, io_msd )
    USE mod_array_protein
    USE Alltimer
    IMPLICIT NONE

    interface

        subroutine rmsd_orient ( input_data, output_data, dt )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:,:), intent ( in ) :: input_data
            real ( kind=8 ), dimension (:), intent ( out ) :: output_data
            real ( kind=4 ), intent ( in ) :: dt
        end subroutine rmsd_orient

        ! defined in math.f
        subroutine derivative ( input, output, dt, opt_factor )
          IMPLICIT NONE
          real ( kind=8 ), dimension(:), intent ( in ) :: input
          real ( kind=8 ),dimension(:),intent( out ) :: output
          real ( kind=8 ) :: dt
          real ( kind=8 ), optional :: opt_factor
        end subroutine derivative

        subroutine simple_derivative ( input, output, dt, opt_factor )
          IMPLICIT NONE
          real ( kind=8 ), dimension(:), intent ( in ) :: input
          real ( kind=8 ),dimension(:),intent( out ) :: output
          real ( kind=8 ) :: dt
          real ( kind=8 ), optional :: opt_factor
        end subroutine simple_derivative

        ! defined in math.f
        subroutine fit_msd_by_regions ( input, nb_region, dt, b_trans, io_dt )
            IMPLICIT NONE
            real ( kind=8 ), dimension(:), intent ( in ) :: input
            real ( kind=4 ), intent ( in ) :: dt
            integer, intent ( in ) :: nb_region, io_dt
            logical, intent ( in ) :: b_trans
        end subroutine fit_msd_by_regions

        ! defined in utils.h
        subroutine print_acf ( ACF, time_snapshot, b_type, value_type, title )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:), intent ( in ) :: ACF
            character, intent( in ) :: title*(*)
            real ( kind=4 ), intent ( in ) :: time_snapshot
            integer, intent ( in ) :: value_type
            logical, intent ( in ) :: b_type
        end subroutine print_acf

        subroutine fit_msd_derivative ( input, D_COEFF, shift_step, array_limit )
            IMPLICIT NONE
            real ( kind=8 ), dimension(:), intent ( in ) :: input
            real ( kind=8 ), dimension(4) :: D_COEFF
            integer, dimension(4) :: array_limit
            integer :: shift_step
        end subroutine fit_msd_derivative

        subroutine print_coeffd_msd ( D_COEFF, region_fit, b_type, value_type, dt, b_trans, b_deriv, io_msd )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:) :: D_COEFF
            integer, dimension(:,: ) :: region_fit
            real ( kind=4 ) :: dt
            integer :: value_type, io_msd
            logical :: b_type, b_trans, b_deriv
        end subroutine print_coeffd_msd

    end interface

    ! x, y, z, nb_prot, time
    real ( kind=8 ), dimension (:,:,:,:), intent ( in ) :: orient_axis
    type ( array_protein_type ), intent ( in ) :: tab_prot
    ! Auto correlation function ( nchunk, ntype, tau )
    real ( kind=8 ), dimension (:,:,:), allocatable :: ACF
    real ( kind=4 ), intent ( in ) :: dt
    integer, intent ( in ) :: nchunks, value_type, nb_region, io_msd
    logical, intent ( in ) :: b_type

    !declared only inside function
    real ( kind=8 ), dimension (:,:,:), allocatable :: D_COEFF_DERIV, D_COEFF_SDERIV
    ! contains the simple derivative of acf / time, and "real derivative"
    real ( kind=8 ), dimension (:,:,:), allocatable :: deriv_acf, deriv_acf2
    ! to store regions where deriv are fitted, easier for output
    ! not dependent of type !nchumks,2
    integer, dimension(:,:), allocatable :: region_deriv_fit

    integer :: startstep, endstep, nacf, nstep_acf
    integer :: first_protein, last_protein, first_type, last_type

    ! local variable
    integer :: nb_prot, nb_type, size_input
    integer :: ierr, itype, shift

    type ( mytimer ) :: timer_rmsd

    call init_mytimer ( timer_rmsd )

    nb_type = -1
    write(*,*) 'Entry all_msd_rotation dt=', dt
    nb_prot = size ( tab_prot % array )

    !max_size = array_shape(4)
    size_input = size ( orient_axis, 4 )

    ! want for all proteins and for the different types
    if ( ( b_type ) .and. ( value_type == 0 ) ) then
        nb_type = tab_prot % nb_grids + 1
    ! only one computation
    else if ( ( ( b_type ) .and. ( value_type > 0 ) ) .or. &
              ( b_type .eqv. .false. ) ) then
        nb_type = 1
    end if
    !write(*,*) 'nb_type ', nb_type
    !write(*,*) 'nchunks ', nchunks

    ! account for nchunks
    nstep_acf = int ( size_input / nchunks )
    write(*,*) 'init nstep_acf ', nstep_acf

    ! allocate to the correct size
    ! if rmsd2, will be size - 1, because traj is reconstructed with - 1 for rotation
    allocate ( ACF ( nchunks, nb_type, nstep_acf - 1 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating ACF'
        stop 1
    end if
    ACF = 0.d0

    !! size of derivative - 1 than input
    allocate ( deriv_acf ( nchunks, nb_type, nstep_acf - 2 ), & !stat = ierr )
               deriv_acf2 ( nchunks, nb_type, nstep_acf - 2 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating deriv_acf'
        stop 1
    end if
    deriv_acf = 0.d0
    deriv_acf2 = 0.d0

    ! used only if fit of derivative of rmsd
    allocate ( D_COEFF_DERIV ( nchunks, nb_type, 4 ), &
               D_COEFF_SDERIV ( nchunks, nb_type, 4 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating D_COEFF'
        stop 1
    end if
    D_COEFF_DERIV = 0.d0
    D_COEFF_SDERIV = 0.d0

    !! not dependent of type ! nchumks,2
    allocate ( region_deriv_fit ( nchunks, 4 ), stat = ierr )
     if ( ierr /= 0 ) then
        write(*,*) 'ERROR In allocating ACF'
        stop 1
    end if
    region_deriv_fit = 0

    ! main log file
    open ( unit=io_msd,file='Rotational_MSD.log', status='replace' )
    write( io_msd, fmt='(a)' )
    write( io_msd, fmt='(a)' ) 'Rotational diffusion coefficient by mean-square displacement'
    write( io_msd, fmt='(a,i6)' ) 'Total number of steps ', size_input
    write( io_msd, fmt='(a,f5.2,a)' ) 'dt =', dt, ' ns'

    shift = 0
    ! default, loop over all types
    first_type = 1
    last_type = tab_prot % nb_grids

LOOP_CHUNK: do  nacf = 1, nchunks

        startstep = nstep_acf * ( nacf - 1 ) + 1
        endstep = nstep_acf * ( nacf )

        write(*,*) 'nacf ', nacf
        write(*,*) 'startstep ', startstep
        write(*,*) 'endstep ', endstep
        write(*,*) 'nstep_acf ', nstep_acf
        !write(*,*) 'nacf ', nacf

!        ! all in output, may change after
!        write( io_msd, fmt='(a)' )
!        write( io_msd, fmt='(a)' ) 'Definition of short and long time scale when fitting the derivative of the MSD'
!        write( io_msd, fmt='(a,i10)' ) 'For chunk ', nacf
!        write( io_msd, fmt='(a)' ) 'Short time scale '
!        write( io_msd, fmt='(a,i10,a,i10)' ) &
!                   'Average between step ',startstep,' and step ',startstep+(nstep_acf/5)
!        write( io_msd, fmt='(a,f10.2,a,f10.2,a)' ) &
!                   'Average between ',startstep*dt,' (ns) and ',(startstep+(nstep_acf/5))*dt,' (ns)'
!
!        write( io_msd, fmt='(a)' )
!        write( io_msd, fmt='(a)' ) 'Long time scale '
!        write( io_msd, fmt='(a,i10,a,i10)' ) &
!               !'Average between step ',startstep+(nstep_acf/5),' and step ',endstep
!               'Average between step ',startstep+(nstep_acf/5),' and step ',endstep-(nstep_acf*2)/3
!        write( io_msd, fmt='(a,f10.2,a,f10.2,a)' ) &
!               !'Average between ',(startstep+(nstep_acf/5))*dt,' (ns) and ',endstep*dt,' (ns)'
!               'Average between ',(startstep+(nstep_acf/5))*dt,' (ns) and ', (endstep-(nstep_acf*2)/3)*dt,' (ns)'

        !write(*,*) 'endstep-startstep+1 ', endstep-startstep+1

        ! compute acf for all proteins, send full array
        if ( (( b_type ) .and. (value_type == 0)) .or. &
             ( b_type .eqv. .false. ) ) then

            !write(*,*) 'b_type is true and compute for all protein types'
            call start_mytimer ( timer_rmsd )

            ! rmsd_orient fill always nstep_acf - 1
            ! simple derivative can be computed after...
            call rmsd_orient ( orient_axis(:,:,:,startstep:endstep), ACF( nacf, 1,: ), dt )

            ! make a least_square fit on blocks from the msd
            write( io_msd, fmt="(a)" )
            write( io_msd, fmt="(a)" ) 'Least square of the MSD of the average over all solutes '
            call fit_msd_by_regions ( ACF ( nacf, 1, : ), nb_region, dt, .false., io_msd )

            ! compute and fit simplified derivative
            call simple_derivative ( ACF ( nacf, 1, : ), deriv_acf( nacf, 1, : ), real(dt,8), 4.d0 )
            ! fit and store mean and std err and short and long-time
            call fit_msd_derivative ( deriv_acf( nacf, 1, : ), D_COEFF_SDERIV( nacf, 1, : ), &
                                      startstep, region_deriv_fit(nacf,:) )

            ! compute and fit real derivative, rotation need to divide by 4
            call derivative ( ACF ( nacf, 1, : ), deriv_acf2( nacf, 1, : ), real(dt,8), 4.d0 )
            call fit_msd_derivative ( deriv_acf2( nacf, 1, : ), D_COEFF_DERIV( nacf, 1, : ), &
                                      -1, region_deriv_fit(nacf,:) )

            call stop_mytimer ( timer_rmsd )
             ! if no type calculation required
             if ( b_type .eqv. .false. ) then
                CYCLE LOOP_CHUNK
             end if

            ! if add type acf, will be shifted by one
            shift = 1

        ! else only acf for one type
        else if ( ( b_type ) .and. ( value_type > 0 ) ) then

            ! will limit the loop over the grid
            first_type = value_type
            last_type = value_type

            ! assur than ACF(nacf,1,:) is filled
            shift = 1 - value_type

        end if ! end of tests


        !! calculations of the different types
        do itype = first_type, last_type

                first_protein = tab_prot % set_of_protein ( itype, 1 )
                last_protein  = tab_prot % set_of_protein ( itype, 2 )
                !write(*,*) 'first last prot ', first_protein, last_protein
                !write(*,*) 'itype+shift', itype+shift

                call start_mytimer ( timer_rmsd )

                call rmsd_orient ( orient_axis(:,:, first_protein:last_protein, startstep:endstep), &
                                   ACF( nacf, itype + shift,: ), dt )

                write( io_msd, fmt="(a)" )
                write( io_msd, fmt="(a,i3)" ) 'Least square fit of the MSD of the solutes type ', itype
                ! fit the msd by region
                call fit_msd_by_regions ( ACF ( nacf, itype + shift, : ), nb_region, dt, .false., io_msd )

                ! compute and fit simplified derivative, rotation need to divide by 4
                call simple_derivative ( ACF ( nacf, itype + shift, : ), deriv_acf( nacf, itype + shift, : ), real(dt,8), 4.d0 )
                ! fit and store mean and std err and short and long-time
                call fit_msd_derivative ( deriv_acf( nacf, itype + shift, : ), D_COEFF_SDERIV( nacf, itype + shift, : ), &
                                          startstep, region_deriv_fit(nacf,:) )

                ! compute and fit real derivative
                call derivative ( ACF ( nacf, itype + shift, : ), deriv_acf2( nacf, itype + shift, : ), real(dt,8), 4.d0 )
                call fit_msd_derivative ( deriv_acf2( nacf, itype + shift, : ), D_COEFF_DERIV( nacf, itype + shift, : ), &
                                          -1, region_deriv_fit(nacf,:) )

                call stop_mytimer ( timer_rmsd )
        end do

    ! end do nchunks
    end do LOOP_CHUNK

    ! print msd and derivative
    call print_acf ( acf, dt, b_type, value_type, 'rot_msd.dat' )
    call print_acf ( deriv_acf, dt, b_type, value_type, 'Dr_simpderiv.dat' )
    call print_acf ( deriv_acf2, dt, b_type, value_type, 'Dr_deriv.dat' )
    deallocate ( deriv_acf, deriv_acf2 )

    ! used in previous version                                                      b_trans, b_deriv
    call print_coeffd_msd ( D_COEFF_DERIV,  region_deriv_fit, b_type, value_type, dt, .false., .true., io_msd )
    call print_coeffd_msd ( D_COEFF_SDERIV, region_deriv_fit, b_type, value_type, dt, .false., .false., io_msd )

    deallocate ( D_COEFF_DERIV, D_COEFF_SDERIV, region_deriv_fit )

    write(*,*) " timer rot msd "
    call print_mytimer ( timer_rmsd )

end subroutine all_msd_rotation

!> Make autocorrelation of the orientation, simplest form
!!
!! Version using the maximum number of data points\n
!! Large tau have a poor statistic\n
!! Here can use OpenMP
!! @param input_data : trajectory of the orientation
!! @param output_data : return, 1 dimensional autocorrelation function
!! @param size_input : size of input
subroutine autocorrelation_orient ( input_data, output_data, size_input )
!$  USE omp_lib
    IMPLICIT NONE
    ! force to have the same shape
    real ( kind=8 ), dimension (:,:,:,:), intent ( in ) :: input_data
    ! only time
    real ( kind=8 ), dimension (:), intent ( out ) :: output_data
    integer, intent ( in ) :: size_input

    integer :: nprot, tau, t0, iprot, tmax, max_tau
    real ( kind=8 ) :: tmp_sum

    !write(*,*) 'autocorrelation_orient '
    nprot = size ( input_data, 3 )
    write(*,*) 'nprot ', nprot
    ! could get from the array
    write(*,*) 'size_input ', size_input
    !write(*,*) 'max_tau ', size_input - 1
    max_tau = size_input - 1
    tmax = 0

    !! can parrallelize this loop
    !$OMP PARALLEL DEFAULT ( NONE ) SHARED ( input_data, size_input, output_data, nprot, max_tau ) &
    !$OMP PRIVATE ( tmax, tmp_sum )

    !$OMP DO SCHEDULE( DYNAMIC )
    do tau = 0, max_tau !- 1

        tmp_sum = 0.d0
        tmax = size_input - tau

        !do t0 = 1, size_input / 2
        do t0 = 1, tmax ! size_input - tau

            !! loop over the proteins
            do iprot = 1, nprot

                !! dot_product of the axis
                ! To check, is it needed to include z axis ?? without get exactly 1.000
                tmp_sum = tmp_sum + &
                          dot_product ( input_data(:,1,iprot,t0 ), input_data(:,1,iprot,t0+tau ) ) + &
                          dot_product ( input_data(:,2,iprot,t0 ), input_data(:,2,iprot,t0+tau ) ) !+ &
                          !dot_product ( input_data(:,3,iprot,t0 ), input_data(:,3,iprot,t0+tau ) )

            ! end do iprot
            end do

        ! end do t0
        end do

        ! assign value to ACF at tau
        output_data ( tau+1 ) = tmp_sum / real ( nprot * tmax * 2, 8 )
        !output_data ( tau+1 ) = tmp_sum / real ( nprot * tmax * 3, 8 )
    ! end do tau
    end do
    !$OMP END DO
    !$OMP END PARALLEL

    !write(*,*) 'output_data ', output_data

end subroutine autocorrelation_orient

!> Different implementation, as an usual msd ( 2 loops over t )
!! dt only needed in check, could be deleted
subroutine rmsd_orient ( input_data, output_data, dt )
!$  USE omp_lib
    IMPLICIT NONE
    ! input contains orientation for all proteins for each time
    real ( kind=8 ), dimension (:,:,:,:), intent ( in ) :: input_data
    ! only time
    real ( kind=8 ), dimension (:), intent ( out ) :: output_data
    real ( kind=4 ), intent ( in ) :: dt

    integer :: ierr, nprot, iprot, t0, tmax , max_tau, tau, size_input
    integer ( kind=8 ) :: tmp_large_integer
    real ( kind=8 ), dimension(3) :: cpp0p1, tmp_vector
    real ( kind=8 ), dimension(:,:,:), allocatable :: reconst_traj
    real ( kind=8 ) :: tmp_sum

    write(*,*) 'rmsd_orient dt=', dt
    nprot = size ( input_data, 3 )
    size_input = size ( input_data, 4 )
    write(*,*) 'nprot ', nprot
    write(*,*) 'size_input ', size_input

    ! will be of size - 1
    allocate ( reconst_traj ( 3, nprot, size_input - 1 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot allocate reconst_traj '
        return
    end if
    reconst_traj = 0.d0
    !write(*,*) 'reconstr traj size ', size_input -1

    !write(*,*) 'reconstruct traj'
    !! can // over protein, ... but very quick anyway
    do iprot = 1, nprot
        tmp_vector = 0.d0
        do t0 = 1, size_input - 1
            call cross( cpp0p1, input_data(:,1, iprot ,t0 ),input_data(:,1, iprot ,t0+1) )

            !! value          0 to 0.3                                 0 to 0.03
            !write( 151,*) t0, sqrt ( dot_product ( cpp0p1, cpp0p1 ) ), sqrt ( dot_product ( cpp0p1*dt, cpp0p1*dt ) )
            !! how to choose ??
            if ( sqrt ( dot_product ( cpp0p1*dt, cpp0p1*dt ) ) > 0.25*dt ) then
                write(*,*) 'Large rotation prot ', iprot, sqrt ( dot_product ( cpp0p1*dt, cpp0p1*dt ) ), ' at t0 / time ', t0, t0*dt
                !make artificial data...tricky should have save last one
                !cpp0p1 = 0.d0
            end if

            tmp_vector = tmp_vector + cpp0p1 !* dt not here !!
            !! fill the trajectory
            reconst_traj ( :, iprot, t0 ) = tmp_vector
        end do
    end do
    !write(*,*) 'done reconst_traj(:,1,size_input -1 )', reconst_traj(:,1,size_input -1 )

    ! now size can reactualized, no modification of the algorithm
    size_input = size_input - 1
    max_tau = size_input - 1
    write(*,*) 'max_tau ', max_tau

    !! can parrallelize this loop
    !$OMP PARALLEL DEFAULT ( NONE ) SHARED ( input_data, reconst_traj, size_input ) &
    !$OMP SHARED ( output_data, dt, nprot, max_tau ) &
    !$OMP PRIVATE ( tmax, tmp_sum, t0, cpp0p1, tmp_vector, tmp_large_integer )

    tmp_sum = 0.d0
    tmp_vector = 0.d0
    cpp0p1 = 0.d0

    !$OMP DO SCHEDULE( DYNAMIC )
    !! // over tau, total_size size_input - 1,
    do tau = 0, max_tau

        tmax = size_input - tau
        tmp_sum = 0.d0
        !write(*,*) 'tau tmax ', tau, tmax

        do t0 = 1, tmax
    !  to make an increasing counter
    !  write (6, 1000) t/nprot,tstep(ttau)
 !1000    format("+","time:",i15," steps",f10.3," [ns]")
    !  write(*,fmt='("+","time:",i15," steps",f10.3," [ns]")' ) t0/nprot,real(t0,4)

          !! loop over the proteins
          do iprot = 1, nprot

            ! now possible to access directly
            tmp_vector = reconst_traj ( :, iprot, t0 + tau ) - reconst_traj ( :, iprot, t0 )
            tmp_sum = tmp_sum + dot_product ( tmp_vector, tmp_vector )

          ! end do iprot
          end do

        ! end do t0
        end do

        if ( MOD(tau, 2000) == 0 ) &
          write(*,*) 'computed integral tau', tau

        ! need intermediate integer*8 ?? certainly when tmax * prot * tau
        tmp_large_integer = tmax * nprot ! not needed anymore, derive in another function

        ! fill direct msd for least_square
        output_data ( tau + 1 ) = tmp_sum / real( int(nprot*tmax,4), 8 )
        !output_data ( tau + 1 ) = tmp_sum / real( tmp_large_integer, 8 )

    ! end do tau
    end do
    !$OMP END DO
    !$OMP END PARALLEL

    deallocate ( reconst_traj )

end subroutine rmsd_orient

!> Version with output similar to sdamm, need to produce more files
!!
!! For use with octave script
subroutine print_acfrot_sdamm ( ACF, time_snapshot, b_type, value_type, io_acf )
    IMPLICIT NONE
    ! Auto correlation function ( nchunks, ntype, tau )
    real ( kind=8 ), dimension (:,:,:) :: ACF
    integer :: value_type, io_acf
    logical :: b_type

    character*64 :: str_format
    character*3 :: tmp_str
    character*64 :: title_file
    integer :: nchunks
    integer :: ierr
    integer :: size_acf, i, j, nb_type
    real (kind=4) :: time_snapshot
    nchunks = size ( ACF, 1 )
    nb_type = size ( ACF, 2 )
    size_acf = size ( ACF, 3 )

    write(*,*) 'nb_type ', nb_type
    write(*,*) 'size_acf ', size_acf
    write(*,*) 'value_type ', value_type

    write ( str_format, '("(f10.2,",i2,"f15.7)")' ) nchunks

    do i = 1, nb_type


        ! will print only one type
        if ((b_type) .and. (value_type /= 0)) then
            write ( tmp_str, fmt='(i2)' ) value_type
        else
            write ( tmp_str, fmt='(i2)' ) i-1
        end if

        title_file = "ACFrot_type"//trim(adjustl(tmp_str))//".dat"
        write(*,*) 'title_file', trim ( title_file )

        if ( ( (b_type) .and. (value_type == 0) .and. (i==1) ) .or. &
             ( b_type .eqv. .false. ) ) then
            title_file="ACFrot_average_all_solutes.dat"
        end if
        open ( io_acf, file=title_file, form='formatted', iostat = ierr )

        write(io_acf,*) '# ACF rotation'
        !write(io_acf,*) '# Header of the column to create, include timestep '

        ! case of all proteins, first is average
        if ( (b_type) .and. (value_type == 0) .and. (i==1) ) then
            write (io_acf,*) '#    Average all solutes'
        else if ( (b_type) .and. (value_type > 0) ) then
            write (io_acf,*) '#    Solute ', i
        else
            write (io_acf,*) '#    Solute ', i-1 !value_type
        end if

        do j = 1, size_acf
            write ( io_acf, str_format ) time_snapshot*(j-1), ACF(:,i,j)
        end do
        write(io_acf,*)

        close ( io_acf )

    end do

end subroutine print_acfrot_sdamm

