!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file utils.f90
!!
!! Utilities functions

!> Extend dynamically an array of double
!!
!! array_in is extended to new_size\n
!! The previous data are copied at the beginning of the new array\n
!! RFirst test with real*8, later may make an interface.
!!
!! It is a test, used in case atom position here
!!
!! @param array_in : array to extend
!! @param new_size : size of the modified array
subroutine extend_array_double ( array_in, new_size )
    IMPLICIT NONE
    !real ( kind = 8 ), dimension ( :, : ), pointer :: array_in
    real ( kind = 8 ), dimension ( :, : ), allocatable :: array_in
    integer :: new_size

    !real ( kind = 8 ), dimension ( :,: ), allocatable :: tmp_array
    real ( kind = 8 ), dimension ( :,: ), pointer :: p_tmp_array
    integer :: old_size, i

    !write(*,*) 'entry extend array double'
    if ( size ( array_in ) == new_size * 3 ) then
        !write(*,*) 'nothing to change, correct size'
        return
    end if

    old_size = size ( array_in ) / 3
    write(*,*) 'extend from ', old_size, ' to ', new_size
    allocate ( p_tmp_array ( 3, old_size ) )

    !! copy all data
    p_tmp_array = array_in

    deallocate ( array_in )
    allocate ( array_in ( 3, new_size ) )

    array_in ( :, 1:old_size ) = p_tmp_array
    do i = old_size + 1, new_size
        array_in ( :, i ) = p_tmp_array ( :, old_size )
    end do

    deallocate ( p_tmp_array )

end subroutine extend_array_double

!> Extend a 1 dimensional array of float
!!
!! Similar to \ref extend_array_double
!!
!! @param array_in : array to extend
!! @param new_size : size of the modified array
subroutine extend_1Darray_float ( array_in, new_size )
    IMPLICIT NONE
    real ( kind = 4 ), dimension ( : ), pointer :: array_in
    integer :: new_size

    real ( kind = 4 ), dimension ( : ), pointer :: p_tmp_array
    integer :: old_size, i

    !write(*,*) 'entry extend array double'
    if ( size ( array_in ) == new_size ) then
        write(*,*) 'nothing to change, correct size'
        return
    end if

    old_size = size ( array_in )
    write(*,*) 'extend from ', old_size, ' to ', new_size
    allocate ( p_tmp_array ( old_size ) )

    !! copy all data
    p_tmp_array = array_in

    deallocate ( array_in )
    allocate ( array_in ( new_size ) )

    array_in ( 1:old_size ) = p_tmp_array
    do i = old_size + 1, new_size
        array_in ( i ) = p_tmp_array ( old_size )
    end do

    deallocate ( p_tmp_array )

end subroutine extend_1Darray_float


!> Create a filename with a suffix ( from toto.pdb get toto.new_suffix )
!!
!! Name is based on pdb file for accessibility, no need of additional variable, but could be more general with input of the old suffix
!! Add opt shift, can modify slightly the conserved part
!! Extend to any suffix, look back from the end to the first point
!!
!! @param pdb_filename : name of the file ( pdb not necessary )
!! @param suffix : string to append
!! @param out_filename : return new created name
!! @param opt_shift : optionaly add a shift to the part to replace. Default 0
subroutine create_filename ( pdb_filename, suffix, out_filename, opt_shift )
    IMPLICIT NONE
    character*128 :: pdb_filename, out_filename
    character, intent( in ) :: suffix*(*)
    integer, optional :: opt_shift

    integer :: shift
    integer :: pos1

    shift = 0
    out_filename = ''

    !write(*,*) 'entry create_filename'
    !write(*,*) trim( pdb_filename )

    if ( present ( opt_shift ) ) then
        shift = opt_shift
        !write (*,*) 'create_filename shift ', shift
    end if

    !pos1 = INDEX ( pdb_filename, ".pdb" )
    !! check from the back, but return the position from the left
    pos1 = INDEX ( pdb_filename, ".", .true. )
    !write(*,*) 'pos1', pos1
    out_filename = pdb_filename(1:pos1-shift) // trim(suffix)
    !write(*,*) 'out_filename:',trim ( out_filename )

end subroutine create_filename

!> Similar to \ref create_filename
!!
!! Here the entry name is modified
!! Without shift option
!! @param pdb_filename : name to be modified
!! @param suffix : string to append
subroutine create_filename_new ( pdb_filename, suffix )
    IMPLICIT NONE
    character*128 :: pdb_filename
    character, intent( in ) :: suffix*(*)

    integer :: shift
    integer :: pos1

    shift = 0
    !out_filename = ''
    !! check from the back, but return the position from the left
    pos1 = INDEX ( pdb_filename, ".", .true. )
    !write(*,*) 'pos1', pos1
    !write(*,*) 'suffix ', trim(suffix)
    !write(*,*) '1:8 ', trim(pdb_filename)
    !write(*,*) '1:8 ', trim(pdb_filename(1:8))

    !pdb_filename = pdb_filename(1:pos1-shift) // trim(suffix)
    pdb_filename = pdb_filename(1:pos1)//trim(suffix)
    !write(*,*) 'out_filename:',trim ( out_filename )
    !write(*,*) 'out pdb_filename:',trim ( pdb_filename )

end subroutine create_filename_new


!> Similar to other, developed for flexibility
!!
!! From toto.pdb to toto0001.pdb
!! \todo Merge this three functions.\n
!!
!! Make temporarly diffent function, to merge wih more general, maybe not
!!
!! Write output into input, may change later
!! @param input_filename : string to be modified
!! @param counter : append
subroutine create_filename_flex ( input_filename, counter )
    IMPLICIT NONE
    character :: input_filename*(*)
    integer :: counter

    character*4 :: tmp_string_counter

    character*64 :: suffix
    !! *(*) only as argument !!
    character*64 :: prefix
    integer :: pos1

    !! for test
    !write(*,*) 'counter ', counter
    !write ( str_numb_tmp, fmt='(i4)' ) i
    write ( tmp_string_counter, fmt='(i4)' ) counter
    !write ( *,*) 'tmp_string_counter ', trim ( tmp_string_counter )

    !! true for back, but pos1 from left to right
    pos1 = INDEX ( input_filename, ".", .true. )
    !write(*,*) 'pos1 ', pos1

    suffix = input_filename ( pos1:)
    !write(*,*) 'suffix!!',suffix,'!!'
    prefix  = input_filename ( 1:pos1-2)

    !write(*,*) 'create_filename_flex prefix!!',prefix,'!!'
    !write(*,*) 'test!!', trim ( prefix//suffix ),'!!'

    input_filename = trim ( trim(prefix) // trim ( adjustl( tmp_string_counter ) ) // trim (suffix ) )

end subroutine create_filename_flex

!> Add specific suffix for all grids
!!
!! Call \ref create_filename_flex function for all types of grids
!!
!! @param inp_list : instance of \ref mod_input_parameter::param_list "param_list"
!! @param counter : the number to append
subroutine modify_fname_implicit ( inp_list, counter )
    USE mod_input_parameter
    IMPLICIT NONE
    type ( param_list ) :: inp_list
    integer :: counter, i

    if ( len ( inp_list % pdb_filename ) > 1 ) then
        !write(*,*) 'modify pdb filename'
        call create_filename_flex ( inp_list % pdb_filename, counter )
        !write(*,*) 'out filename ', tmp_string_file
        !write(*,*)
        !write(*,*) 'new inp_list % pdb_filename ', trim ( inp_list % pdb_filename )

    end if

    if ( inp_list % qef /= '' ) then
        !write(*,*) 'modify qep filename'
        call create_filename_flex ( inp_list % qef, counter )
        !write(*,*) 'new inp_list % qef ', trim ( inp_list % qef )
    end if

    if ( inp_list % epf /= '' ) then
        !write(*,*) 'modify ep filename'
        call create_filename_flex ( inp_list % epf, counter )
        !write(*,*) 'new inp_list % epf ', trim ( inp_list % epf )
    end if

    if ( inp_list % edf /= '' ) then
        !write(*,*) 'modify ed filename'
        call create_filename_flex ( inp_list % edf, counter )
        !write(*,*) 'new inp_list % edf ', trim ( inp_list % edf )
    end if

    if ( inp_list % hdf /= '' ) then
        !write(*,*) 'modify ed filename'
        call create_filename_flex ( inp_list % hdf, counter )
        !write(*,*) 'new inp_list % hdf ', trim ( inp_list % hdf )
    end if


    if ( allocated ( inp_list % ljf ) ) then
        !if ( inp_list % ljf /= '' ) then
        do i = 1, size( inp_list % ljf )
            !write(*,*) 'modify ljf filename'
            call create_filename_flex ( inp_list % ljf(i), counter )
            !write(*,*) 'new inp_list % ljf ', trim ( inp_list % ljf )
        end do
    end if

end subroutine modify_fname_implicit

!> Add suffix for fparrallel runs.\n
!!
!! Modify the output filename, for allow different runs in the same directory ??
!! @param input_filename : name of input file
!! @param number : suffix to add
subroutine add_nbpar ( input_filename, number )
    IMPLICIT NONE
    character :: input_filename*(*)
    integer :: number

    character*4 :: tmp_string_counter

    write ( tmp_string_counter, fmt='(i4)' ) number

    input_filename = trim ( input_filename ) // trim ( adjustl ( tmp_string_counter) )

end subroutine add_nbpar
