!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file prepare_atom_protein.f90
!! call by prepare_grid.f90 for setting atom position and exclusion grid

! to allocate here, or pass as argument ?
! add new type_van for dealing with VdW parameters, can read external file provided by the user

! deleted completely zslice, never used, used only in solva_naccess but fixed in the code

! accessibility always computed from prep1p and prep2p original function
! for protein 1 needed to compute exclusion grid
! for protein 2, no exclusion grid, but needed to reduce the number of atom to keep in memory and check overlap
! exclusion_grid => logical
! add save_access and save_exclusion to write into file accessibility and exclusion_grid


! use to assign pdb, atom_surface and effective charges TO THE SOGRID, not to the protein
! add rcriteria to assign rcriter_pos
! many proteins, one grid ( sdamm )

! center will be computed here and returned
! distmax assigned to sogrid now

! proble with if read/write access is true and exclusion grid is false
! choice reread pdb, will be exactly the same as before
! normal use all the pdb, but if read access only atom_vdw of access are read. grid slightly different ( hole )

!> Read pdb file, assign vdw radius and compute atom_surface, exclusion grid if needed and setup the reaction criteria
!!
!! Assign all the data to a set_of_grid, not to proteins\n
!! Should be called only by prepare_grid ( never grid without pdbs, flex : each grid a pdb ), simple to handle\n
!! Log file for printing of eclusion.log and reaction criteria.log\n
!! Check if the accessibility file and/or the exclusion grid has already been computed. In this case the pdb does not need to be read again\n
!!
!! @bug if access is not read, center is not updated and exclusion grid is not correctly shifted
!!
!! @param sgrid : instance of \ref mod_setofgrid::sogrid "sogrid"
!! @param rcriter : instance of \ref mod_reactioncriteria::react_criter "react_criter"
!! @param pdb_filename :
!! @param param_probe : contains the different probe in \ref com_sda::probe_type "param_probe"
!! @param center : return the center of the solutes
!! @param counter : internal counter
!! @param bool_excl_grid : if exclusion grid is needed ( solute 1 in sda_2proteins )
!! @param iflex : if the solute is flexible
!! @param isurface :  if the solute is a surface
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
subroutine prepare_atom_protein ( sgrid, rcriter, pdb_filename, param_probe, &
                                  center, counter, bool_excl_grid, iflex, isurface, type_calc )

    USE mod_setofgrid
    USE mod_ReactionCriteria
    USE mod_type_calc
    USE mod_van
    USE com_sda ! needed for log output of access atoms
    IMPLICIT NONE

    interface
        subroutine van1( a, v )
            IMPLICIT NONE
            character*1, intent ( in ) :: a
            real ( kind=8 ), intent ( out ) :: v
        end subroutine van1

        subroutine read_pdb ( pdb_filename, rcriter, tmp_atom_letter, tmp_atom_position, tmp_rcriter_pos, counter_prot )
            USE mod_ReactionCriteria
            CHARACTER*128 pdb_filename
            type ( react_criter ) :: rcriter
            real ( kind=8 ), dimension(:,:), intent (inout), pointer :: tmp_atom_position, tmp_rcriter_pos
            character*2, dimension(:), intent(inout), pointer :: tmp_atom_letter
            integer, intent ( in ) :: counter_prot
        end subroutine read_pdb

        subroutine solva ( na,x,a,acc,probew )
            IMPLICIT NONE
            integer, intent ( in ) :: na
            real ( kind=8 ), dimension(:,:), intent ( in ) :: x ! atom_pos
            real ( kind=8 ), dimension(:), intent ( in ) :: a ! atom_vdw
            real ( kind=8 ), dimension(:), intent ( out ) :: acc ! atom access
            real ( kind=4 ), intent ( in ) :: probew
        end subroutine solva

        ! zslice is fixed at 0.05
        !subroutine solva_naccess ( nats, xyz, rads, accs, probew )
        !    IMPLICIT NONE
        !    integer, intent ( in ) :: nats
        !    real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: xyz
        !    real ( kind = 8 ), dimension ( : ), intent ( in ) :: rads
        !    real ( kind = 8 ), dimension ( : ), intent ( out ) :: accs
        !    real ( kind = 4 ), intent ( in ) :: probew
        !end subroutine solva_naccess

        subroutine make_exclusion_grid ( sgrid, tmp_atom_pos, tmp_nat, center, atom_vdw, max_vdw, probep, hexclusion )
            USE mod_setofgrid
            IMPLICIT NONE
            type ( sogrid ) :: sgrid
            real ( kind=8 ), dimension ( 3 ) :: center
            real ( kind=8 ), dimension(:) :: atom_vdw
            real ( kind=8 ) :: max_vdw
            real ( kind=4 ) :: probep, hexclusion
            !! original use all atoms, still with dynamic alloc
            real ( kind=8 ), dimension ( :, : ) :: tmp_atom_pos
            integer :: tmp_nat
        end subroutine make_exclusion_grid

        subroutine create_filename ( pdb_filename, suffix, out_filename, opt_shift )
            IMPLICIT NONE
            character*128 :: pdb_filename, out_filename
            character, intent( in ) :: suffix*(*)
            integer, optional :: opt_shift
        end subroutine create_filename

    end interface

    type ( sogrid ) :: sgrid
    type ( react_criter ) :: rcriter
    type ( type_calculation ) :: type_calc
    character ( 128 ), intent ( in ) :: pdb_filename
    type ( probe_type ), intent ( in ) :: param_probe
    logical, intent ( in ) :: iflex, isurface
    logical, intent ( in ) :: bool_excl_grid
    integer :: counter
    real ( kind=8 ), dimension(3), intent ( out ) :: center

    ! temporary allocatable array, if added here, be sure to deallocate before return !
    real ( kind=8 ), dimension(:,:), pointer :: tmp_atom_position, tmp_rcriter_pos
    character*2, dimension(:), pointer :: tmp_atom_letter
    ! keep the list of temporary acceessibility and vdw for all atoms in the pdb, same to delete before leaving !
    ! NJB: atom_vdw ordering was changed after accessibility calculation so that 
    !      correct radii were sent to accessibility file, but this caused errors
    !      in the exclusion grid as wrong radii used.
    !      FIX: new array of radii specifically for accessibility
    real ( kind=8 ), dimension(:), pointer :: tmp_atom_access, atom_vdw, access_atom_vdw 

    ! used if save_access, otherwise compute solva all the time
    character*128 :: filename_access, filename_excl
    logical :: exist_access_file, bool_save_read_access, exist_excl_file, bool_save_excl

    real ( kind=8) :: max_vdw, dist2, distmax2
    real ( kind=8 ) :: dist_max
    ! tmp_nat number of all atoms in the pdb, nat will be the reminding nb_atom_surface
    integer :: i, ierr, tmp_nat, nat
    integer :: nb_atom_surface

    ! to discard some part of the computation if read access file for instance
    logical :: bool_read_pdb, bool_vdw, bool_compute_center, compute_excl_grid

    ! with keyword save, value of the variable is maintened for next entry. print INFO only once
    logical, save :: bool_print_info = .true.

    ! certainly better as argument, can be loaded only once !
    ! make explicit test until now, delete and recreate every time
    type ( type_van ) :: tvan

    !! init variable
    nullify ( tmp_atom_position )
    nullify ( tmp_rcriter_pos )
    nullify ( tmp_atom_letter )
    nullify ( tmp_atom_access )
    nullify ( atom_vdw )
    nullify ( access_atom_vdw )

    !! set to false if read from access. in particular if flex need pdb for reaction criteria
    bool_read_pdb = .true.
    bool_vdw = .true.
    bool_compute_center = .true.


    filename_access = ''
    filename_excl = ''
    ! default value, input in param_probe
    exist_access_file = .false.
    bool_save_read_access = .false.
    exist_excl_file = .false.
    bool_save_excl = .false.

    ! default like bool_excl_grid, if already exists switch to off
    compute_excl_grid = bool_excl_grid

    ! need to initialize here
    max_vdw = 0.
    dist2 = 0.
    distmax2 = 0.
    dist_max = 0.
    center=(/0.d0 ,0.d0 ,0.d0/)

    !write(*,*)
    !write(*,*) 'Entry prepare_atom_protein counter ', counter, ' iflex ', iflex
    !write(*,*) 'bool_excl_grid ', bool_excl_grid
    !write(*,*) 'iflex ', iflex
    !write(*,*) 'isurface ', isurface
    !write(*,*) 'bool_excl ', bool_excl_grid
    !write(*,*) 'param_probe % hexclusion ', param_probe % hexclusion
    !write(*,*) 'param_probe % save_access ', param_probe % save_access


    !write(*,*) 'init of van ', allocated ( tvan % array_atom ), allocated ( tvan % array_tcharge )

    ! need to determine if reading of criteria should be done in again in reading pdb, problem with cm does not correspond to a real atom
    if ( ( (  counter == 1 ) .or. ( counter == 2 ) ) .and. &
         ( ( rcriter % type_reaction == docking ) .or. ( rcriter % type_reaction == assoc_rate ) ) .and. &
         ( iflex .eqv. .true. ) ) then
        !write(*,*) 'allocate tmp_rcriter_pos, may need to read_new_coordinates for criteria '

        allocate ( tmp_rcriter_pos ( 3, rcriter % ntotal ), stat =ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In allocating tmp_rcriter_pos '
            stop 1
        end if
        tmp_rcriter_pos = 0.
    end if

    ! test save_access
    if ( param_probe % save_access ) then

        ! flex or not save lots of time in reading accessibility
        call create_filename ( pdb_filename, "access", filename_access )
        inquire ( file = filename_access, exist =  exist_access_file )
        if ( exist_access_file ) then
            ! name is based on pdb file and unique !!
            ! write (*,*) 'file exists ', trim ( filename_access )

            call read_accessibility ( sgrid, filename_access, center, atom_vdw, max_vdw, dist_max, param_probe )
            ! if allocated after checking same parameter
            if ( allocated ( sgrid % accessibility ) ) then
                nb_atom_surface = sgrid % nb_atom_surface

            !write (*,*) 'after read accessibility center ', center
            !write (*,*) 'after read accessibility vdw(1) ', atom_vdw(1), atom_vdw( nb_atom_surface ), &
            !            ' size ', size ( atom_vdw )
            !write (*,*) 'max_vdw ', max_vdw
            !write (*,*) 'dist_max ', dist_max
            !write(*,*) 'center read ', center

                tmp_nat = size ( atom_vdw )
                ! whatever flex or not, vdw read from file
                bool_vdw = .false.
                bool_compute_center = .false.

                ! if iflex is true need pdb for getting reaction crieria position
                if ( iflex .eqv. .false. ) then
                    bool_read_pdb = .false.
                end if
           else
               ! will force to rewritte
               exist_access_file = .false.
           endif

            ! if iflex need pdb for getting reaction crieria position
            ! if ( iflex .eqv. .false. ) then
            !    bool_read_pdb = .false.
            !end if

        ! default, read all
        else
            write(*, fmt='("accessibility file does not exist, to write ",(a))' ) trim ( filename_access )
             ! use default, read and compute all

        !! end if read access file
        end if
    else
        ! use default, read and compute all
        ! print info only once
        if ( bool_print_info ) then
            write (*,*)
            write (*,*) 'INFO Set save_access = 1 to write the accessibility list atoms to a file'
            write (*,*) 'INFO It will avoid the recomputation to the next run,', &
                        ' and may save several hours if natoms > 100 000 and you not have the licencied naccess version'
            write (*,*) 'INFO You can also load the file as PDB and visualize it ( e.g. vmd)'
            write (*,*)
            bool_print_info = .false.
        end if

    end if

    ! Test if exclusion grid exists, if not need to read_pdb
    ! \bug if access is not read, center is not updated and exclusion grid is not correctly shifted
    if ( bool_excl_grid )  then

        if ( param_probe % save_excl /= 0 ) then
            !! create filename
            if ( param_probe % save_excl == 1 ) then
                call create_filename ( pdb_filename, "excl.bin.grd", filename_excl )
            else if  ( param_probe % save_excl == 2 ) then
                call create_filename ( pdb_filename, "excl.ascii.grd", filename_excl )
            else
                write(*,*) 'ERROR In option save_excl', param_probe % save_excl
                stop 1
            end if

            inquire ( file = filename_excl, exist =  exist_excl_file )
            if ( exist_excl_file ) then
                !write(*,*) 'exclusion file exists'
                allocate ( sgrid % p_excl_grid, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR Cannot allocate pointer on excl grid'
                    stop 1
                end if

                allocate ( sgrid % p_excl_grid % pgrid, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR Cannot allocate grid object in excl grid in prepare_atom_protein'
                    stop 1
                end if

                write (*, fmt='("Read exclusion grid from file ",(a))' ) trim ( filename_excl )

                !if ( center (/0.d0 ,0.d0 ,0.d0/) ) then
                if ( exist_access_file .eqv. .false. ) then
                    write(*,*) 'ERROR Center protein has not been set up before reading the exclusion grid'
                    write(*,*) 'ERROR This a bug in prepare_atom_protein'
                    write(*,*) 'ERROR Delete and force the recomputation of the exclusion grid with the flag save_access = 1'
                    stop 1
                endif
                ! use variable as status, check parameters in the file
                ! problem here, modify variable when reading
                ! bool_excl_grid = .false.
                call read_excl_grid ( sgrid % p_excl_grid % pgrid, filename_excl, center, param_probe, compute_excl_grid )

                ! if true, parameters differ
                if ( compute_excl_grid ) then
                    write(*,*)
                    !write(*,*) 'bool_excl_grid true the parameters in the file differ from your input probe or spacing'
                    write(*,*) 'status true the parameters in the file differ from your input probe or spacing'
                    write(*,*) 'will recompute the grid '
                    write(*,*)
                    !write(*,*) 'deallocate sgrid ??'
                    deallocate ( sgrid % p_excl_grid, stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR Cannot deallocate pointer on excl grid'
                        stop 1
                    end if
                    nullify ( sgrid % p_excl_grid )
                    bool_save_excl = .true.
                    bool_read_pdb = .true.
                    bool_vdw = .true.
                    bool_compute_center = .true.

                endif

            else
                write (*,*) 'Exclusion file does not exist, it will be computed and written to file'
                bool_save_excl = .true.
                bool_read_pdb = .true.
                bool_vdw = .true.
                bool_compute_center = .true.

            end if

        else
            !write (*,*) 'save_excl is false, need to read pdb and compute exclusion grid'
            !! could compute from access file, interior will be 0, but should work
            !! may change only pdb is strickly need
            bool_read_pdb = .true.
            bool_compute_center = .true.
            bool_vdw = .true.
        end if
    end if


#ifdef DEBUG
    write(*,*) 'prep atom_protein after bool setup:'
    write(*,*) 'bool_read_pdb ',bool_read_pdb
    write(*,*) 'bool_compute_center ', bool_compute_center
    write(*,*) 'bool_vdw ', bool_vdw    
#endif



    ! read pdb, get max van der walls and position of all atoms, and set cm of protein
    ! for solva need list of atom, error need list of atom_vdw (previously a(nat))
    ! needed for center, could be in access if save_access, except if flex !!! need to reread for reaction criteria
    ! need for vdw, but can be in the access file, as printed before

    ! if flex need for reaction criteria, if make_exclusion_grid need tmp_atom_position
    ! write(*,*) 'bool read pdb ',bool_read_pdb
    if ( bool_read_pdb ) then

        call read_pdb ( pdb_filename, rcriter, tmp_atom_letter, tmp_atom_position, tmp_rcriter_pos, counter )

        ! bug here, if compute exclusion grid bool_read_pdb is true, but tmp_nat is not updated
        ! reset tmp_nat
        ! still bug when computing maxdist, to check ??
        tmp_nat = size ( tmp_atom_letter )
        ! write(*,*) 'update tmp_nat size of tmp_atom_letter ', tmp_nat
        ! write(*,*) 'after reading pdb xyz(1),xyz(nat) ', tmp_atom_position(:,1), tmp_atom_position(:,tmp_nat)

        ! if flex, and access exists center already computed,  just need for reaction_criteria
        if ( bool_compute_center ) then

            tmp_nat = size ( tmp_atom_letter )
            ! write(*,*) 'read pdb and compute center assign tmp_nat', tmp_nat

            !write (*,*) 'bool_compute_center, need to recompute the center '
            !! need to reinitialize
            center = 0.
            do i=1, tmp_nat
                center = center + tmp_atom_position(:,i)
            end do
            !write(*,*) 'center before normalisation ', center
            !write(*,*) 'normalisation ', tmp_nat
            center = center / dfloat( tmp_nat )
            write ( *, fmt='("Center at:            ",3f8.3)' ) center
            !write(*,*) 'center ', center

#ifdef DEBUG
        else
            write (*,*) 'bool_compute_center is false, already computed center: ', center
            write (*,*) 'bool_compute_center is false, already computed dist_max: ', dist_max
#endif
        end if

        ! compute max vdw, otherwise read from access file
        if ( bool_vdw ) then
            ! write(*,*) 'bool vdw is true, allocate or reallocate atom_vdw size tmp_nat', tmp_nat

            ! why to reallocate ? to check
            ! may be reallocate if read_pdb, will be exactly the same
            if ( associated ( atom_vdw ) ) then
                !write(*,*) 'delete previous atom_vdw'
                deallocate ( atom_vdw )
            !else
                !write(*,*) 'atom_vdw no associated before '
            end if
            !NJB: Shouldn't be neccessary, not used previously.
            if ( associated ( access_atom_vdw ) ) then
                deallocate ( access_atom_vdw )
            end if

            ! tmp_nat = size ( tmp_atom_letter )
            ! allocate tabular to store the vdw of each atom
            allocate ( atom_vdw( tmp_nat ), stat=ierr )
            if ( ierr/=0 ) then
                write(*,*) ' problem with allocation of atom_vdw size ', tmp_nat
            endif

            ! test if mod_van has been allocated
            if ( allocated ( tvan % array_atom ) .eqv. .false. ) then
                !write(*,*) 'Call initialisatino of mod_van '
                call init_van ( tvan )
            else
                write(*,*) 'tvan already allocated, nothing more to do '
            end if

            ! rite(*,*) 'bool vdw call new module VAN'
            do i = 1, tmp_nat
                !call van1 ( tmp_atom_letter( i ), atom_vdw ( i ) )
                !write(*,*) i, tmp_atom_letter( i ), atom_vdw ( i )

                ! need to transform, expects 2 character
                !tmp_char2 = tmp_atom_letter ( i )
                !call get_VdW ( tmp_atom_letter ( i ), atom_vdw ( i ) )
                !call get_VdW ( tvan, tmp_char2, atom_vdw ( i ) )
                ! now fine tmp_letter is 2 characters
                call get_VdW ( tvan, tmp_atom_letter ( i ), atom_vdw ( i ) )

                if ( atom_vdw ( i ) > max_vdw ) max_vdw = atom_vdw ( i )
            end do

        end if
    ! else
    !    write(*,*) 'Do not need to read the pdbfile'
    !    write(*,*) 'value of tmp_nat ', tmp_nat
    end if

    write ( *, fmt='("Largest atom radius is:       ",f8.3)' ) max_vdw

    ! if read from file already allocated
    if ( allocated ( sgrid % accessibility ) .eqv. .false. ) then

#ifdef DEBUG
        write(*,*) 'sgrid % accessibility not allocated '
        write(*,*) 'will compute accessibility tmp_nat ', tmp_nat
#endif
        ! compute accessibility, always to define list_atom_surface
        write ( *, fmt='( "Compute atom_surface from ", i7, " atoms with probe water: ",f10.3)' ) tmp_nat, param_probe % probew

        allocate( tmp_atom_access( tmp_nat ), stat=ierr )
        if ( ierr/=0 ) then
            write(*,*) ' problem with allocation of tmp_atom_access'
            stop 1
        endif
        tmp_atom_access = 0.d0

        call solva ( tmp_nat, tmp_atom_position, atom_vdw, tmp_atom_access, param_probe % probew )
        ! bug version naccess, crash on 64 bits with gfortran and OpenMP ! not in debug mode (only on AMDx2)
        !call solva_naccess ( tmp_nat, tmp_atom_position, atom_vdw, tmp_atom_access, param_probe % probew )

        ! ckeck with threshold, could be in solva function..simpler, tmp_atom_access only in solva !
        ! printed if save_access = true, just add vdw
        nb_atom_surface = 0
        do i = 1, tmp_nat
            if ( tmp_atom_access ( i ) > param_probe % thresold ) then
                nb_atom_surface = nb_atom_surface + 1
            endif
        end do

        ! assign nb_atom_surface to sogrid
        sgrid % nb_atom_surface = nb_atom_surface
        write (*, fmt='("Nb atom surface :             ", i8)') nb_atom_surface

        ! allocate final array in set_ofgrid
        if ( nb_atom_surface > 0 ) then
            allocate ( sgrid % accessibility ( nb_atom_surface ), &
                       sgrid % list_atom_surface ( 3,nb_atom_surface ), stat=ierr )
            if ( ierr/= 0 ) then
                write (*,*) 'problem in allocation of grid accessibility and list_atom_surface'
            endif
            if ( sgrid % alloc_which_at_surf ) then
                allocate ( sgrid % which_atom_surface( nb_atom_surface ), stat=ierr )
                if ( ierr/= 0 ) then
                    write (*,*) 'problem in allocation of grid accessibility and list_atom_surface'
                end if
            endif
      
            ! NJB: Create array containing the vdW radii of the accessible atoms, avoids weird reindexing later
            !      - dealloc shouldn't be neccessary, not used previously.
            if ( associated ( access_atom_vdw ) ) then
                deallocate ( access_atom_vdw )
            end if
            allocate(access_atom_vdw(nb_atom_surface),stat=ierr)
            if (ierr.NE.0) then
                write (*,*) " Error allocating access_atom_vdw"
                stop 1
            end if

        else
            write (*,*) 'ERROR No atoms on the surface, certainly an error !'
            stop 1
        endif

        ! copy data with correct indice
        nat = 0
        do i = 1, tmp_nat
            if ( tmp_atom_access ( i ) > param_probe % thresold ) then
                nat = nat + 1
                ! keep the information which atoms from the entire PDB are the surface atoms (solvent accessible)
                if ( sgrid % alloc_which_at_surf ) then
                    sgrid % which_atom_surface( nat ) = i
                end if

                sgrid % accessibility ( nat ) = tmp_atom_access( i )
                !write(*,*) 'copy atom i', i, tmp_atom_access( i ), atom_vdw ( i )
                sgrid % list_atom_surface ( :, nat ) = tmp_atom_position( :,i )

                ! before forgot to copy atom_vdw at correct indice
                ! should be fine.. no data lost ?? to check
                access_atom_vdw ( nat ) = atom_vdw ( i )
            end if
        end do

        if ( nat /= nb_atom_surface ) then
            write(*,*) 'ERROR Problem in counting atom surface'
            stop 1
        end if

    else
        !write(*,*) 'not need to compute solva, assign nb_atom_surface= tmp_nat '
        nb_atom_surface = tmp_nat
    end if

    ! assign pos_criteria
    ! write(*,*) 'before assign rcriteria counter=', counter
    ! if iflex false assign default coordinates

    ! default assign to grid1 and 2, could keep reading of pdb by default ( but problem cm )
    if ( ( ( rcriter % type_reaction /= off_rc ) .and. ( rcriter % type_reaction /= all )  ) .and. &
         ( (  counter == 1 ) .or. ( counter == 2 ) ) ) then
        !write(*,*) 'allocate pos_criteria ', rcriter % ntotal

        allocate ( sgrid % reaction_criteria_pos ( 3, rcriter % ntotal ), stat = ierr )
        if ( ierr/= 0 ) then
            write (*,*) 'problem in allocation of sgrid % pos_criteria'
        endif
        sgrid % nb_reaction_criteria = rcriter % ntotal

        ! iflex false, so just copy data from rxna
        if ( iflex .eqv. .false. ) then
            !write(*,*) 'Grid is not flexible, copy default rcriter % pos atom with counter ',counter, ' and center ', center
            if ( counter == 1 ) then
                do i = 1, rcriter % ntotal
                    sgrid % reaction_criteria_pos (:, i) = real ( rcriter % pos_atom1 (:, i) - center, 4 )
                end do
            else
                do i = 1, rcriter % ntotal
                    sgrid % reaction_criteria_pos (:, i) = real ( rcriter % pos_atom2 (:, i) - center, 4 )
                end do
            end if

        ! if flex, copy data from tmp_rcriter_pos, has correct values for position, CANNOT USE CM in this scheme
        else
            ! write(*,*) 'grid is flexible, copy tmp_rcriter_pos counter =', counter, ' center= ',center
            do i = 1, rcriter % ntotal
                sgrid % reaction_criteria_pos (:, i) = real ( tmp_rcriter_pos (:, i) - center, 4 )
            end do
        end if
    end if

    ! compute max_distance, now threshold has been checked before
    ! not read from the file
    if ( exist_access_file .eqv. .false. ) then
        !write(*,*) 'before compute distmax center', center
        !write(*,*) 'nb_atom_surface', nb_atom_surface
        !write(*,*) 'isurface ', isurface

        do i = 1, nb_atom_surface
        ! add here calculation of distance_max,
        ! if isurface, compute only along z, lots of assumption, should be protein 1
        ! could check only > 0
            if ( isurface ) then
                if ( ( sgrid % list_atom_surface (3,i) - center(3) ) > 0 ) then
                    dist2 =  ( sgrid % list_atom_surface (3,i) - center(3) ) * &
                             ( sgrid % list_atom_surface (3,i) - center(3) )
                end if
            else
                dist2 = dot_product( sgrid % list_atom_surface (:,i) - center, &
                                     sgrid % list_atom_surface (:,i) - center )
            end if

            if ( dist2 > distmax2 ) distmax2 = dist2
        end do
        dist_max = sqrt ( distmax2 )

    end if

    ! assign to set of grid
    sgrid % distance_max = real ( dist_max, 4 )
    !write(*,*) 'distance max', sgrid % distance_max
    write(*, fmt='("Distance max of the solute : ",f12.3)' ) sgrid % distance_max

    ! change save_access until here, format is still not clear, may be just copy all text line. ( or rebecca asked number, name )
    if ( ( param_probe % save_access ) .and. ( exist_access_file .eqv. .false. ) ) then
        !write(*,*) 'call write_accessibility '
        call write_accessibility ( sgrid, filename_access, center, access_atom_vdw, max_vdw, dist_max, param_probe )
    end if
    
    ! check if associated, if read already done
    if ( compute_excl_grid ) then
        if ( associated ( sgrid % p_excl_grid ) .eqv. .false. ) then
            ! as original send tmp_atom all atoms in pbs
            ! write (*,*) 'p_excl_grid not associated '
            tmp_nat = size ( atom_vdw )

            ! original send all pdb position
            call make_exclusion_grid ( sgrid, tmp_atom_position, tmp_nat, center, atom_vdw, max_vdw, &
                                       param_probe % probep, param_probe % hexclusion )
            !! maybe later use list_atom_surface, do not need to reread the pdb if access file
            !call make_exclusion_grid ( sgrid, sgrid % list_atom_surface, tmp_nat, center, atom_vdw, max_vdw, &
            !                           param_probe % probep, param_probe % hexclusion )
        end if

        ! here printed whateever save_excl or not
        if ( type_calc % nb_par == 1 ) then
            !write (*,*) 'type_calc % nb_par == 1 print exclusion grid log'
            call print_excl_grid_log ( sgrid % p_excl_grid, iopoti )
        end if
        !write(*,*) 'save_excl ', param_probe % save_excl
        ! default false, true only if asked but not present
        ! add condition if fparralllel, only the first one, but should be present if fparrellel
        if ( ( bool_save_excl ) .and.  ( type_calc % nb_par == 1 ) ) then
            call write_excl_grid ( sgrid % p_excl_grid % pgrid, filename_excl, center, param_probe )
        end if
    end if

#ifdef DEBUG
    write(*,*) 'End prepare_atom_grid'
#endif

!!!!!!!! End exclusion, deallocate tmp arrays

    ! check because not associated if reading accessibility file or other cases
    if ( associated ( tmp_atom_access ) ) then
        deallocate ( tmp_atom_access, stat = ierr )
        if ( ierr/=0 ) then
            write(*,*) 'ERROR In deallocation of tmp_atomX when no accessibility'
            stop 1
        endif
        nullify ( tmp_atom_access )
    end if

    ! if pdb not read, not allocated
    if ( associated ( tmp_atom_position ) ) then
        deallocate ( tmp_atom_position, tmp_atom_letter, stat = ierr )
        if ( ierr/=0 ) then
            write(*,*) 'ERROR In deallocation of tmp_atomX when no accessibility'
            stop 1
        endif
        nullify ( tmp_atom_position, tmp_atom_letter )
    end if

    if ( associated ( atom_vdw ) ) then
        deallocate ( atom_vdw, stat = ierr )
        if ( ierr/=0 ) then
            write(*,*) 'ERROR In deallocation of atom_vdw '
            stop 1
        endif
        nullify ( atom_vdw )
    end if

    ! deallocate tmp_rcriter_pos if needed
    if ( associated ( tmp_rcriter_pos ) ) then
        deallocate ( tmp_rcriter_pos, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocation of tmp_rcriter_pos '
            stop 1
        end if
        nullify ( tmp_rcriter_pos )
    end if

    !! for test, allocate and deallocate everytime
    ! write(*,*) 'test if new to delete van'
    if ( allocated ( tvan % array_atom ) ) then
        ! write(*,*) 'call delete_van'
        call delete_van ( tvan )
    end if

end subroutine prepare_atom_protein
